/* Generated by makegtkalias.pl */

#ifndef DISABLE_VISIBILITY

#include <glib.h>

#ifdef G_HAVE_GNUC_VISIBILITY

#undef IN_FILE
#define IN_FILE defined

#undef IN_HEADER
#define IN_HEADER(x) 1

#if IN_HEADER(__GTK_ABOUT_DIALOG_H__)
#if IN_FILE(__GTK_ABOUT_DIALOG_C__)
#undef gtk_show_about_dialog
extern __typeof (gtk_show_about_dialog) gtk_show_about_dialog __attribute((alias("IA__gtk_show_about_dialog"), visibility("default"))) G_GNUC_NULL_TERMINATED;

#undef gtk_about_dialog_get_artists
extern __typeof (gtk_about_dialog_get_artists) gtk_about_dialog_get_artists __attribute((alias("IA__gtk_about_dialog_get_artists"), visibility("default")));

#undef gtk_about_dialog_get_authors
extern __typeof (gtk_about_dialog_get_authors) gtk_about_dialog_get_authors __attribute((alias("IA__gtk_about_dialog_get_authors"), visibility("default")));

#undef gtk_about_dialog_get_comments
extern __typeof (gtk_about_dialog_get_comments) gtk_about_dialog_get_comments __attribute((alias("IA__gtk_about_dialog_get_comments"), visibility("default")));

#undef gtk_about_dialog_get_copyright
extern __typeof (gtk_about_dialog_get_copyright) gtk_about_dialog_get_copyright __attribute((alias("IA__gtk_about_dialog_get_copyright"), visibility("default")));

#undef gtk_about_dialog_get_documenters
extern __typeof (gtk_about_dialog_get_documenters) gtk_about_dialog_get_documenters __attribute((alias("IA__gtk_about_dialog_get_documenters"), visibility("default")));

#undef gtk_about_dialog_get_license
extern __typeof (gtk_about_dialog_get_license) gtk_about_dialog_get_license __attribute((alias("IA__gtk_about_dialog_get_license"), visibility("default")));

#undef gtk_about_dialog_get_logo
extern __typeof (gtk_about_dialog_get_logo) gtk_about_dialog_get_logo __attribute((alias("IA__gtk_about_dialog_get_logo"), visibility("default")));

#undef gtk_about_dialog_get_logo_icon_name
extern __typeof (gtk_about_dialog_get_logo_icon_name) gtk_about_dialog_get_logo_icon_name __attribute((alias("IA__gtk_about_dialog_get_logo_icon_name"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_about_dialog_get_name
extern __typeof (gtk_about_dialog_get_name) gtk_about_dialog_get_name __attribute((alias("IA__gtk_about_dialog_get_name"), visibility("default")));

#undef gtk_about_dialog_set_name
extern __typeof (gtk_about_dialog_set_name) gtk_about_dialog_set_name __attribute((alias("IA__gtk_about_dialog_set_name"), visibility("default")));

#endif
#undef gtk_about_dialog_get_program_name
extern __typeof (gtk_about_dialog_get_program_name) gtk_about_dialog_get_program_name __attribute((alias("IA__gtk_about_dialog_get_program_name"), visibility("default")));

#undef gtk_about_dialog_set_program_name
extern __typeof (gtk_about_dialog_set_program_name) gtk_about_dialog_set_program_name __attribute((alias("IA__gtk_about_dialog_set_program_name"), visibility("default")));

#undef gtk_about_dialog_get_translator_credits
extern __typeof (gtk_about_dialog_get_translator_credits) gtk_about_dialog_get_translator_credits __attribute((alias("IA__gtk_about_dialog_get_translator_credits"), visibility("default")));

#undef gtk_about_dialog_get_type
extern __typeof (gtk_about_dialog_get_type) gtk_about_dialog_get_type __attribute((alias("IA__gtk_about_dialog_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_about_dialog_get_version
extern __typeof (gtk_about_dialog_get_version) gtk_about_dialog_get_version __attribute((alias("IA__gtk_about_dialog_get_version"), visibility("default")));

#undef gtk_about_dialog_get_website
extern __typeof (gtk_about_dialog_get_website) gtk_about_dialog_get_website __attribute((alias("IA__gtk_about_dialog_get_website"), visibility("default")));

#undef gtk_about_dialog_get_website_label
extern __typeof (gtk_about_dialog_get_website_label) gtk_about_dialog_get_website_label __attribute((alias("IA__gtk_about_dialog_get_website_label"), visibility("default")));

#undef gtk_about_dialog_get_wrap_license
extern __typeof (gtk_about_dialog_get_wrap_license) gtk_about_dialog_get_wrap_license __attribute((alias("IA__gtk_about_dialog_get_wrap_license"), visibility("default")));

#undef gtk_about_dialog_new
extern __typeof (gtk_about_dialog_new) gtk_about_dialog_new __attribute((alias("IA__gtk_about_dialog_new"), visibility("default")));

#undef gtk_about_dialog_set_artists
extern __typeof (gtk_about_dialog_set_artists) gtk_about_dialog_set_artists __attribute((alias("IA__gtk_about_dialog_set_artists"), visibility("default")));

#undef gtk_about_dialog_set_authors
extern __typeof (gtk_about_dialog_set_authors) gtk_about_dialog_set_authors __attribute((alias("IA__gtk_about_dialog_set_authors"), visibility("default")));

#undef gtk_about_dialog_set_comments
extern __typeof (gtk_about_dialog_set_comments) gtk_about_dialog_set_comments __attribute((alias("IA__gtk_about_dialog_set_comments"), visibility("default")));

#undef gtk_about_dialog_set_copyright
extern __typeof (gtk_about_dialog_set_copyright) gtk_about_dialog_set_copyright __attribute((alias("IA__gtk_about_dialog_set_copyright"), visibility("default")));

#undef gtk_about_dialog_set_documenters
extern __typeof (gtk_about_dialog_set_documenters) gtk_about_dialog_set_documenters __attribute((alias("IA__gtk_about_dialog_set_documenters"), visibility("default")));

#undef gtk_about_dialog_set_license
extern __typeof (gtk_about_dialog_set_license) gtk_about_dialog_set_license __attribute((alias("IA__gtk_about_dialog_set_license"), visibility("default")));

#undef gtk_about_dialog_set_logo
extern __typeof (gtk_about_dialog_set_logo) gtk_about_dialog_set_logo __attribute((alias("IA__gtk_about_dialog_set_logo"), visibility("default")));

#undef gtk_about_dialog_set_logo_icon_name
extern __typeof (gtk_about_dialog_set_logo_icon_name) gtk_about_dialog_set_logo_icon_name __attribute((alias("IA__gtk_about_dialog_set_logo_icon_name"), visibility("default")));

#undef gtk_about_dialog_set_translator_credits
extern __typeof (gtk_about_dialog_set_translator_credits) gtk_about_dialog_set_translator_credits __attribute((alias("IA__gtk_about_dialog_set_translator_credits"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_about_dialog_set_email_hook
extern __typeof (gtk_about_dialog_set_email_hook) gtk_about_dialog_set_email_hook __attribute((alias("IA__gtk_about_dialog_set_email_hook"), visibility("default")));

#undef gtk_about_dialog_set_url_hook
extern __typeof (gtk_about_dialog_set_url_hook) gtk_about_dialog_set_url_hook __attribute((alias("IA__gtk_about_dialog_set_url_hook"), visibility("default")));

#endif
#undef gtk_about_dialog_set_version
extern __typeof (gtk_about_dialog_set_version) gtk_about_dialog_set_version __attribute((alias("IA__gtk_about_dialog_set_version"), visibility("default")));

#undef gtk_about_dialog_set_website
extern __typeof (gtk_about_dialog_set_website) gtk_about_dialog_set_website __attribute((alias("IA__gtk_about_dialog_set_website"), visibility("default")));

#undef gtk_about_dialog_set_website_label
extern __typeof (gtk_about_dialog_set_website_label) gtk_about_dialog_set_website_label __attribute((alias("IA__gtk_about_dialog_set_website_label"), visibility("default")));

#undef gtk_about_dialog_set_wrap_license
extern __typeof (gtk_about_dialog_set_wrap_license) gtk_about_dialog_set_wrap_license __attribute((alias("IA__gtk_about_dialog_set_wrap_license"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_ACCEL_GROUP_H__)
#if IN_FILE(__GTK_ACCEL_GROUP_C__)
#undef gtk_accelerator_get_default_mod_mask
extern __typeof (gtk_accelerator_get_default_mod_mask) gtk_accelerator_get_default_mod_mask __attribute((alias("IA__gtk_accelerator_get_default_mod_mask"), visibility("default")));

#undef gtk_accelerator_get_label
extern __typeof (gtk_accelerator_get_label) gtk_accelerator_get_label __attribute((alias("IA__gtk_accelerator_get_label"), visibility("default")));

#undef gtk_accelerator_name
extern __typeof (gtk_accelerator_name) gtk_accelerator_name __attribute((alias("IA__gtk_accelerator_name"), visibility("default")));

#undef gtk_accelerator_parse
extern __typeof (gtk_accelerator_parse) gtk_accelerator_parse __attribute((alias("IA__gtk_accelerator_parse"), visibility("default")));

#undef gtk_accelerator_set_default_mod_mask
extern __typeof (gtk_accelerator_set_default_mod_mask) gtk_accelerator_set_default_mod_mask __attribute((alias("IA__gtk_accelerator_set_default_mod_mask"), visibility("default")));

#undef gtk_accelerator_valid
extern __typeof (gtk_accelerator_valid) gtk_accelerator_valid __attribute((alias("IA__gtk_accelerator_valid"), visibility("default"))) G_GNUC_CONST;

#undef gtk_accel_group_activate
extern __typeof (gtk_accel_group_activate) gtk_accel_group_activate __attribute((alias("IA__gtk_accel_group_activate"), visibility("default")));

#undef gtk_accel_group_connect
extern __typeof (gtk_accel_group_connect) gtk_accel_group_connect __attribute((alias("IA__gtk_accel_group_connect"), visibility("default")));

#undef gtk_accel_group_connect_by_path
extern __typeof (gtk_accel_group_connect_by_path) gtk_accel_group_connect_by_path __attribute((alias("IA__gtk_accel_group_connect_by_path"), visibility("default")));

#undef gtk_accel_group_disconnect
extern __typeof (gtk_accel_group_disconnect) gtk_accel_group_disconnect __attribute((alias("IA__gtk_accel_group_disconnect"), visibility("default")));

#undef gtk_accel_group_disconnect_key
extern __typeof (gtk_accel_group_disconnect_key) gtk_accel_group_disconnect_key __attribute((alias("IA__gtk_accel_group_disconnect_key"), visibility("default")));

#undef gtk_accel_group_find
extern __typeof (gtk_accel_group_find) gtk_accel_group_find __attribute((alias("IA__gtk_accel_group_find"), visibility("default")));

#undef gtk_accel_group_from_accel_closure
extern __typeof (gtk_accel_group_from_accel_closure) gtk_accel_group_from_accel_closure __attribute((alias("IA__gtk_accel_group_from_accel_closure"), visibility("default")));

#undef gtk_accel_group_get_type
extern __typeof (gtk_accel_group_get_type) gtk_accel_group_get_type __attribute((alias("IA__gtk_accel_group_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_accel_group_get_is_locked
extern __typeof (gtk_accel_group_get_is_locked) gtk_accel_group_get_is_locked __attribute((alias("IA__gtk_accel_group_get_is_locked"), visibility("default")));

#undef gtk_accel_group_get_modifier_mask
extern __typeof (gtk_accel_group_get_modifier_mask) gtk_accel_group_get_modifier_mask __attribute((alias("IA__gtk_accel_group_get_modifier_mask"), visibility("default")));

#undef gtk_accel_group_lock
extern __typeof (gtk_accel_group_lock) gtk_accel_group_lock __attribute((alias("IA__gtk_accel_group_lock"), visibility("default")));

#undef gtk_accel_group_new
extern __typeof (gtk_accel_group_new) gtk_accel_group_new __attribute((alias("IA__gtk_accel_group_new"), visibility("default")));

#undef gtk_accel_group_query
extern __typeof (gtk_accel_group_query) gtk_accel_group_query __attribute((alias("IA__gtk_accel_group_query"), visibility("default")));

#undef gtk_accel_groups_activate
extern __typeof (gtk_accel_groups_activate) gtk_accel_groups_activate __attribute((alias("IA__gtk_accel_groups_activate"), visibility("default")));

#undef gtk_accel_groups_from_object
extern __typeof (gtk_accel_groups_from_object) gtk_accel_groups_from_object __attribute((alias("IA__gtk_accel_groups_from_object"), visibility("default")));

#undef gtk_accel_group_unlock
extern __typeof (gtk_accel_group_unlock) gtk_accel_group_unlock __attribute((alias("IA__gtk_accel_group_unlock"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_ACCEL_LABEL_H__)
#if IN_FILE(__GTK_ACCEL_LABEL_C__)
#undef gtk_accel_label_get_accel_widget
extern __typeof (gtk_accel_label_get_accel_widget) gtk_accel_label_get_accel_widget __attribute((alias("IA__gtk_accel_label_get_accel_widget"), visibility("default")));

#undef gtk_accel_label_get_accel_width
extern __typeof (gtk_accel_label_get_accel_width) gtk_accel_label_get_accel_width __attribute((alias("IA__gtk_accel_label_get_accel_width"), visibility("default")));

#undef gtk_accel_label_get_type
extern __typeof (gtk_accel_label_get_type) gtk_accel_label_get_type __attribute((alias("IA__gtk_accel_label_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_accel_label_new
extern __typeof (gtk_accel_label_new) gtk_accel_label_new __attribute((alias("IA__gtk_accel_label_new"), visibility("default")));

#undef gtk_accel_label_refetch
extern __typeof (gtk_accel_label_refetch) gtk_accel_label_refetch __attribute((alias("IA__gtk_accel_label_refetch"), visibility("default")));

#undef gtk_accel_label_set_accel_closure
extern __typeof (gtk_accel_label_set_accel_closure) gtk_accel_label_set_accel_closure __attribute((alias("IA__gtk_accel_label_set_accel_closure"), visibility("default")));

#undef gtk_accel_label_set_accel_widget
extern __typeof (gtk_accel_label_set_accel_widget) gtk_accel_label_set_accel_widget __attribute((alias("IA__gtk_accel_label_set_accel_widget"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_ACCEL_MAP_H__)
#if IN_FILE(__GTK_ACCEL_MAP_C__)
#undef gtk_accel_map_add_entry
extern __typeof (gtk_accel_map_add_entry) gtk_accel_map_add_entry __attribute((alias("IA__gtk_accel_map_add_entry"), visibility("default")));

#undef gtk_accel_map_add_filter
extern __typeof (gtk_accel_map_add_filter) gtk_accel_map_add_filter __attribute((alias("IA__gtk_accel_map_add_filter"), visibility("default")));

#undef gtk_accel_map_change_entry
extern __typeof (gtk_accel_map_change_entry) gtk_accel_map_change_entry __attribute((alias("IA__gtk_accel_map_change_entry"), visibility("default")));

#undef gtk_accel_map_foreach
extern __typeof (gtk_accel_map_foreach) gtk_accel_map_foreach __attribute((alias("IA__gtk_accel_map_foreach"), visibility("default")));

#undef gtk_accel_map_foreach_unfiltered
extern __typeof (gtk_accel_map_foreach_unfiltered) gtk_accel_map_foreach_unfiltered __attribute((alias("IA__gtk_accel_map_foreach_unfiltered"), visibility("default")));

#undef gtk_accel_map_get
extern __typeof (gtk_accel_map_get) gtk_accel_map_get __attribute((alias("IA__gtk_accel_map_get"), visibility("default")));

#undef gtk_accel_map_get_type
extern __typeof (gtk_accel_map_get_type) gtk_accel_map_get_type __attribute((alias("IA__gtk_accel_map_get_type"), visibility("default"))) G_GNUC_CONST;

#ifndef _WIN64
#undef gtk_accel_map_load
extern __typeof (gtk_accel_map_load) gtk_accel_map_load __attribute((alias("IA__gtk_accel_map_load"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_accel_map_load_utf8
extern __typeof (gtk_accel_map_load_utf8) gtk_accel_map_load_utf8 __attribute((alias("IA__gtk_accel_map_load_utf8"), visibility("default")));

#endif
#undef gtk_accel_map_load_fd
extern __typeof (gtk_accel_map_load_fd) gtk_accel_map_load_fd __attribute((alias("IA__gtk_accel_map_load_fd"), visibility("default")));

#undef gtk_accel_map_load_scanner
extern __typeof (gtk_accel_map_load_scanner) gtk_accel_map_load_scanner __attribute((alias("IA__gtk_accel_map_load_scanner"), visibility("default")));

#undef gtk_accel_map_lock_path
extern __typeof (gtk_accel_map_lock_path) gtk_accel_map_lock_path __attribute((alias("IA__gtk_accel_map_lock_path"), visibility("default")));

#undef gtk_accel_map_lookup_entry
extern __typeof (gtk_accel_map_lookup_entry) gtk_accel_map_lookup_entry __attribute((alias("IA__gtk_accel_map_lookup_entry"), visibility("default")));

#ifndef _WIN64
#undef gtk_accel_map_save
extern __typeof (gtk_accel_map_save) gtk_accel_map_save __attribute((alias("IA__gtk_accel_map_save"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_accel_map_save_utf8
extern __typeof (gtk_accel_map_save_utf8) gtk_accel_map_save_utf8 __attribute((alias("IA__gtk_accel_map_save_utf8"), visibility("default")));

#endif
#undef gtk_accel_map_save_fd
extern __typeof (gtk_accel_map_save_fd) gtk_accel_map_save_fd __attribute((alias("IA__gtk_accel_map_save_fd"), visibility("default")));

#undef gtk_accel_map_unlock_path
extern __typeof (gtk_accel_map_unlock_path) gtk_accel_map_unlock_path __attribute((alias("IA__gtk_accel_map_unlock_path"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_ACCESSIBLE_H__)
#if IN_FILE(__GTK_ACCESSIBLE_C__)
#undef gtk_accessible_connect_widget_destroyed
extern __typeof (gtk_accessible_connect_widget_destroyed) gtk_accessible_connect_widget_destroyed __attribute((alias("IA__gtk_accessible_connect_widget_destroyed"), visibility("default")));

#undef gtk_accessible_get_type
extern __typeof (gtk_accessible_get_type) gtk_accessible_get_type __attribute((alias("IA__gtk_accessible_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_accessible_get_widget
extern __typeof (gtk_accessible_get_widget) gtk_accessible_get_widget __attribute((alias("IA__gtk_accessible_get_widget"), visibility("default")));

#undef gtk_accessible_set_widget
extern __typeof (gtk_accessible_set_widget) gtk_accessible_set_widget __attribute((alias("IA__gtk_accessible_set_widget"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_ACTION_H__)
#if IN_FILE(__GTK_ACTION_C__)
#undef gtk_action_activate
extern __typeof (gtk_action_activate) gtk_action_activate __attribute((alias("IA__gtk_action_activate"), visibility("default")));

#undef gtk_action_block_activate
extern __typeof (gtk_action_block_activate) gtk_action_block_activate __attribute((alias("IA__gtk_action_block_activate"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_action_block_activate_from
extern __typeof (gtk_action_block_activate_from) gtk_action_block_activate_from __attribute((alias("IA__gtk_action_block_activate_from"), visibility("default")));

#endif
#undef gtk_action_connect_accelerator
extern __typeof (gtk_action_connect_accelerator) gtk_action_connect_accelerator __attribute((alias("IA__gtk_action_connect_accelerator"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_action_connect_proxy
extern __typeof (gtk_action_connect_proxy) gtk_action_connect_proxy __attribute((alias("IA__gtk_action_connect_proxy"), visibility("default")));

#endif
#undef gtk_action_create_icon
extern __typeof (gtk_action_create_icon) gtk_action_create_icon __attribute((alias("IA__gtk_action_create_icon"), visibility("default")));

#undef gtk_action_create_menu_item
extern __typeof (gtk_action_create_menu_item) gtk_action_create_menu_item __attribute((alias("IA__gtk_action_create_menu_item"), visibility("default")));

#undef gtk_action_create_tool_item
extern __typeof (gtk_action_create_tool_item) gtk_action_create_tool_item __attribute((alias("IA__gtk_action_create_tool_item"), visibility("default")));

#undef gtk_action_create_menu
extern __typeof (gtk_action_create_menu) gtk_action_create_menu __attribute((alias("IA__gtk_action_create_menu"), visibility("default")));

#undef gtk_action_disconnect_accelerator
extern __typeof (gtk_action_disconnect_accelerator) gtk_action_disconnect_accelerator __attribute((alias("IA__gtk_action_disconnect_accelerator"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_action_disconnect_proxy
extern __typeof (gtk_action_disconnect_proxy) gtk_action_disconnect_proxy __attribute((alias("IA__gtk_action_disconnect_proxy"), visibility("default")));

#endif
#undef gtk_action_get_always_show_image
extern __typeof (gtk_action_get_always_show_image) gtk_action_get_always_show_image __attribute((alias("IA__gtk_action_get_always_show_image"), visibility("default")));

#undef gtk_action_get_accel_closure
extern __typeof (gtk_action_get_accel_closure) gtk_action_get_accel_closure __attribute((alias("IA__gtk_action_get_accel_closure"), visibility("default")));

#undef gtk_action_get_accel_path
extern __typeof (gtk_action_get_accel_path) gtk_action_get_accel_path __attribute((alias("IA__gtk_action_get_accel_path"), visibility("default")));

#undef gtk_action_get_gicon
extern __typeof (gtk_action_get_gicon) gtk_action_get_gicon __attribute((alias("IA__gtk_action_get_gicon"), visibility("default")));

#undef gtk_action_get_label
extern __typeof (gtk_action_get_label) gtk_action_get_label __attribute((alias("IA__gtk_action_get_label"), visibility("default")));

#undef gtk_action_get_name
extern __typeof (gtk_action_get_name) gtk_action_get_name __attribute((alias("IA__gtk_action_get_name"), visibility("default")));

#undef gtk_action_get_proxies
extern __typeof (gtk_action_get_proxies) gtk_action_get_proxies __attribute((alias("IA__gtk_action_get_proxies"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_widget_get_action
extern __typeof (gtk_widget_get_action) gtk_widget_get_action __attribute((alias("IA__gtk_widget_get_action"), visibility("default")));

#endif
#undef gtk_action_get_icon_name
extern __typeof (gtk_action_get_icon_name) gtk_action_get_icon_name __attribute((alias("IA__gtk_action_get_icon_name"), visibility("default")));

#undef gtk_action_get_is_important
extern __typeof (gtk_action_get_is_important) gtk_action_get_is_important __attribute((alias("IA__gtk_action_get_is_important"), visibility("default")));

#undef gtk_action_get_sensitive
extern __typeof (gtk_action_get_sensitive) gtk_action_get_sensitive __attribute((alias("IA__gtk_action_get_sensitive"), visibility("default")));

#undef gtk_action_get_short_label
extern __typeof (gtk_action_get_short_label) gtk_action_get_short_label __attribute((alias("IA__gtk_action_get_short_label"), visibility("default")));

#undef gtk_action_get_stock_id
extern __typeof (gtk_action_get_stock_id) gtk_action_get_stock_id __attribute((alias("IA__gtk_action_get_stock_id"), visibility("default")));

#undef gtk_action_get_tooltip
extern __typeof (gtk_action_get_tooltip) gtk_action_get_tooltip __attribute((alias("IA__gtk_action_get_tooltip"), visibility("default")));

#undef gtk_action_get_type
extern __typeof (gtk_action_get_type) gtk_action_get_type __attribute((alias("IA__gtk_action_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_action_get_visible
extern __typeof (gtk_action_get_visible) gtk_action_get_visible __attribute((alias("IA__gtk_action_get_visible"), visibility("default")));

#undef gtk_action_get_visible_horizontal
extern __typeof (gtk_action_get_visible_horizontal) gtk_action_get_visible_horizontal __attribute((alias("IA__gtk_action_get_visible_horizontal"), visibility("default")));

#undef gtk_action_get_visible_vertical
extern __typeof (gtk_action_get_visible_vertical) gtk_action_get_visible_vertical __attribute((alias("IA__gtk_action_get_visible_vertical"), visibility("default")));

#undef gtk_action_is_sensitive
extern __typeof (gtk_action_is_sensitive) gtk_action_is_sensitive __attribute((alias("IA__gtk_action_is_sensitive"), visibility("default")));

#undef gtk_action_is_visible
extern __typeof (gtk_action_is_visible) gtk_action_is_visible __attribute((alias("IA__gtk_action_is_visible"), visibility("default")));

#undef gtk_action_new
extern __typeof (gtk_action_new) gtk_action_new __attribute((alias("IA__gtk_action_new"), visibility("default")));

#undef gtk_action_set_accel_group
extern __typeof (gtk_action_set_accel_group) gtk_action_set_accel_group __attribute((alias("IA__gtk_action_set_accel_group"), visibility("default")));

#undef gtk_action_set_accel_path
extern __typeof (gtk_action_set_accel_path) gtk_action_set_accel_path __attribute((alias("IA__gtk_action_set_accel_path"), visibility("default")));

#undef gtk_action_set_always_show_image
extern __typeof (gtk_action_set_always_show_image) gtk_action_set_always_show_image __attribute((alias("IA__gtk_action_set_always_show_image"), visibility("default")));

#undef gtk_action_set_gicon
extern __typeof (gtk_action_set_gicon) gtk_action_set_gicon __attribute((alias("IA__gtk_action_set_gicon"), visibility("default")));

#undef gtk_action_set_icon_name
extern __typeof (gtk_action_set_icon_name) gtk_action_set_icon_name __attribute((alias("IA__gtk_action_set_icon_name"), visibility("default")));

#undef gtk_action_set_is_important
extern __typeof (gtk_action_set_is_important) gtk_action_set_is_important __attribute((alias("IA__gtk_action_set_is_important"), visibility("default")));

#undef gtk_action_set_label
extern __typeof (gtk_action_set_label) gtk_action_set_label __attribute((alias("IA__gtk_action_set_label"), visibility("default")));

#undef gtk_action_set_sensitive
extern __typeof (gtk_action_set_sensitive) gtk_action_set_sensitive __attribute((alias("IA__gtk_action_set_sensitive"), visibility("default")));

#undef gtk_action_set_short_label
extern __typeof (gtk_action_set_short_label) gtk_action_set_short_label __attribute((alias("IA__gtk_action_set_short_label"), visibility("default")));

#undef gtk_action_set_stock_id
extern __typeof (gtk_action_set_stock_id) gtk_action_set_stock_id __attribute((alias("IA__gtk_action_set_stock_id"), visibility("default")));

#undef gtk_action_set_tooltip
extern __typeof (gtk_action_set_tooltip) gtk_action_set_tooltip __attribute((alias("IA__gtk_action_set_tooltip"), visibility("default")));

#undef gtk_action_set_visible
extern __typeof (gtk_action_set_visible) gtk_action_set_visible __attribute((alias("IA__gtk_action_set_visible"), visibility("default")));

#undef gtk_action_set_visible_horizontal
extern __typeof (gtk_action_set_visible_horizontal) gtk_action_set_visible_horizontal __attribute((alias("IA__gtk_action_set_visible_horizontal"), visibility("default")));

#undef gtk_action_set_visible_vertical
extern __typeof (gtk_action_set_visible_vertical) gtk_action_set_visible_vertical __attribute((alias("IA__gtk_action_set_visible_vertical"), visibility("default")));

#undef gtk_action_unblock_activate
extern __typeof (gtk_action_unblock_activate) gtk_action_unblock_activate __attribute((alias("IA__gtk_action_unblock_activate"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_action_unblock_activate_from
extern __typeof (gtk_action_unblock_activate_from) gtk_action_unblock_activate_from __attribute((alias("IA__gtk_action_unblock_activate_from"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_ACTION_GROUP_H__)
#if IN_FILE(__GTK_ACTION_GROUP_C__)
#undef gtk_action_group_add_action
extern __typeof (gtk_action_group_add_action) gtk_action_group_add_action __attribute((alias("IA__gtk_action_group_add_action"), visibility("default")));

#undef gtk_action_group_add_actions
extern __typeof (gtk_action_group_add_actions) gtk_action_group_add_actions __attribute((alias("IA__gtk_action_group_add_actions"), visibility("default")));

#undef gtk_action_group_add_actions_full
extern __typeof (gtk_action_group_add_actions_full) gtk_action_group_add_actions_full __attribute((alias("IA__gtk_action_group_add_actions_full"), visibility("default")));

#undef gtk_action_group_add_action_with_accel
extern __typeof (gtk_action_group_add_action_with_accel) gtk_action_group_add_action_with_accel __attribute((alias("IA__gtk_action_group_add_action_with_accel"), visibility("default")));

#undef gtk_action_group_add_radio_actions
extern __typeof (gtk_action_group_add_radio_actions) gtk_action_group_add_radio_actions __attribute((alias("IA__gtk_action_group_add_radio_actions"), visibility("default")));

#undef gtk_action_group_add_radio_actions_full
extern __typeof (gtk_action_group_add_radio_actions_full) gtk_action_group_add_radio_actions_full __attribute((alias("IA__gtk_action_group_add_radio_actions_full"), visibility("default")));

#undef gtk_action_group_add_toggle_actions
extern __typeof (gtk_action_group_add_toggle_actions) gtk_action_group_add_toggle_actions __attribute((alias("IA__gtk_action_group_add_toggle_actions"), visibility("default")));

#undef gtk_action_group_add_toggle_actions_full
extern __typeof (gtk_action_group_add_toggle_actions_full) gtk_action_group_add_toggle_actions_full __attribute((alias("IA__gtk_action_group_add_toggle_actions_full"), visibility("default")));

#undef gtk_action_group_get_action
extern __typeof (gtk_action_group_get_action) gtk_action_group_get_action __attribute((alias("IA__gtk_action_group_get_action"), visibility("default")));

#undef gtk_action_group_get_name
extern __typeof (gtk_action_group_get_name) gtk_action_group_get_name __attribute((alias("IA__gtk_action_group_get_name"), visibility("default")));

#undef gtk_action_group_get_sensitive
extern __typeof (gtk_action_group_get_sensitive) gtk_action_group_get_sensitive __attribute((alias("IA__gtk_action_group_get_sensitive"), visibility("default")));

#undef gtk_action_group_get_type
extern __typeof (gtk_action_group_get_type) gtk_action_group_get_type __attribute((alias("IA__gtk_action_group_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_action_group_get_visible
extern __typeof (gtk_action_group_get_visible) gtk_action_group_get_visible __attribute((alias("IA__gtk_action_group_get_visible"), visibility("default")));

#undef gtk_action_group_list_actions
extern __typeof (gtk_action_group_list_actions) gtk_action_group_list_actions __attribute((alias("IA__gtk_action_group_list_actions"), visibility("default")));

#undef gtk_action_group_new
extern __typeof (gtk_action_group_new) gtk_action_group_new __attribute((alias("IA__gtk_action_group_new"), visibility("default")));

#undef gtk_action_group_remove_action
extern __typeof (gtk_action_group_remove_action) gtk_action_group_remove_action __attribute((alias("IA__gtk_action_group_remove_action"), visibility("default")));

#undef gtk_action_group_set_sensitive
extern __typeof (gtk_action_group_set_sensitive) gtk_action_group_set_sensitive __attribute((alias("IA__gtk_action_group_set_sensitive"), visibility("default")));

#undef gtk_action_group_set_translate_func
extern __typeof (gtk_action_group_set_translate_func) gtk_action_group_set_translate_func __attribute((alias("IA__gtk_action_group_set_translate_func"), visibility("default")));

#undef gtk_action_group_set_translation_domain
extern __typeof (gtk_action_group_set_translation_domain) gtk_action_group_set_translation_domain __attribute((alias("IA__gtk_action_group_set_translation_domain"), visibility("default")));

#undef gtk_action_group_set_visible
extern __typeof (gtk_action_group_set_visible) gtk_action_group_set_visible __attribute((alias("IA__gtk_action_group_set_visible"), visibility("default")));

#undef gtk_action_group_translate_string
extern __typeof (gtk_action_group_translate_string) gtk_action_group_translate_string __attribute((alias("IA__gtk_action_group_translate_string"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_ACTIVATABLE_H__)
#if IN_FILE(__GTK_ACTIVATABLE_C__)
#undef gtk_activatable_do_set_related_action
extern __typeof (gtk_activatable_do_set_related_action) gtk_activatable_do_set_related_action __attribute((alias("IA__gtk_activatable_do_set_related_action"), visibility("default")));

#undef gtk_activatable_get_related_action
extern __typeof (gtk_activatable_get_related_action) gtk_activatable_get_related_action __attribute((alias("IA__gtk_activatable_get_related_action"), visibility("default")));

#undef gtk_activatable_get_type
extern __typeof (gtk_activatable_get_type) gtk_activatable_get_type __attribute((alias("IA__gtk_activatable_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_activatable_get_use_action_appearance
extern __typeof (gtk_activatable_get_use_action_appearance) gtk_activatable_get_use_action_appearance __attribute((alias("IA__gtk_activatable_get_use_action_appearance"), visibility("default")));

#undef gtk_activatable_sync_action_properties
extern __typeof (gtk_activatable_sync_action_properties) gtk_activatable_sync_action_properties __attribute((alias("IA__gtk_activatable_sync_action_properties"), visibility("default")));

#undef gtk_activatable_set_related_action
extern __typeof (gtk_activatable_set_related_action) gtk_activatable_set_related_action __attribute((alias("IA__gtk_activatable_set_related_action"), visibility("default")));

#undef gtk_activatable_set_use_action_appearance
extern __typeof (gtk_activatable_set_use_action_appearance) gtk_activatable_set_use_action_appearance __attribute((alias("IA__gtk_activatable_set_use_action_appearance"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_ADJUSTMENT_H__)
#if IN_FILE(__GTK_ADJUSTMENT_C__)
#undef gtk_adjustment_changed
extern __typeof (gtk_adjustment_changed) gtk_adjustment_changed __attribute((alias("IA__gtk_adjustment_changed"), visibility("default")));

#undef gtk_adjustment_clamp_page
extern __typeof (gtk_adjustment_clamp_page) gtk_adjustment_clamp_page __attribute((alias("IA__gtk_adjustment_clamp_page"), visibility("default")));

#undef gtk_adjustment_configure
extern __typeof (gtk_adjustment_configure) gtk_adjustment_configure __attribute((alias("IA__gtk_adjustment_configure"), visibility("default")));

#undef gtk_adjustment_get_type
extern __typeof (gtk_adjustment_get_type) gtk_adjustment_get_type __attribute((alias("IA__gtk_adjustment_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_adjustment_get_value
extern __typeof (gtk_adjustment_get_value) gtk_adjustment_get_value __attribute((alias("IA__gtk_adjustment_get_value"), visibility("default")));

#undef gtk_adjustment_get_lower
extern __typeof (gtk_adjustment_get_lower) gtk_adjustment_get_lower __attribute((alias("IA__gtk_adjustment_get_lower"), visibility("default")));

#undef gtk_adjustment_get_upper
extern __typeof (gtk_adjustment_get_upper) gtk_adjustment_get_upper __attribute((alias("IA__gtk_adjustment_get_upper"), visibility("default")));

#undef gtk_adjustment_get_step_increment
extern __typeof (gtk_adjustment_get_step_increment) gtk_adjustment_get_step_increment __attribute((alias("IA__gtk_adjustment_get_step_increment"), visibility("default")));

#undef gtk_adjustment_get_page_increment
extern __typeof (gtk_adjustment_get_page_increment) gtk_adjustment_get_page_increment __attribute((alias("IA__gtk_adjustment_get_page_increment"), visibility("default")));

#undef gtk_adjustment_get_page_size
extern __typeof (gtk_adjustment_get_page_size) gtk_adjustment_get_page_size __attribute((alias("IA__gtk_adjustment_get_page_size"), visibility("default")));

#undef gtk_adjustment_new
extern __typeof (gtk_adjustment_new) gtk_adjustment_new __attribute((alias("IA__gtk_adjustment_new"), visibility("default")));

#undef gtk_adjustment_set_value
extern __typeof (gtk_adjustment_set_value) gtk_adjustment_set_value __attribute((alias("IA__gtk_adjustment_set_value"), visibility("default")));

#undef gtk_adjustment_set_lower
extern __typeof (gtk_adjustment_set_lower) gtk_adjustment_set_lower __attribute((alias("IA__gtk_adjustment_set_lower"), visibility("default")));

#undef gtk_adjustment_set_upper
extern __typeof (gtk_adjustment_set_upper) gtk_adjustment_set_upper __attribute((alias("IA__gtk_adjustment_set_upper"), visibility("default")));

#undef gtk_adjustment_set_step_increment
extern __typeof (gtk_adjustment_set_step_increment) gtk_adjustment_set_step_increment __attribute((alias("IA__gtk_adjustment_set_step_increment"), visibility("default")));

#undef gtk_adjustment_set_page_increment
extern __typeof (gtk_adjustment_set_page_increment) gtk_adjustment_set_page_increment __attribute((alias("IA__gtk_adjustment_set_page_increment"), visibility("default")));

#undef gtk_adjustment_set_page_size
extern __typeof (gtk_adjustment_set_page_size) gtk_adjustment_set_page_size __attribute((alias("IA__gtk_adjustment_set_page_size"), visibility("default")));

#undef gtk_adjustment_value_changed
extern __typeof (gtk_adjustment_value_changed) gtk_adjustment_value_changed __attribute((alias("IA__gtk_adjustment_value_changed"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_ALIGNMENT_H__)
#if IN_FILE(__GTK_ALIGNMENT_C__)
#undef gtk_alignment_get_padding
extern __typeof (gtk_alignment_get_padding) gtk_alignment_get_padding __attribute((alias("IA__gtk_alignment_get_padding"), visibility("default")));

#undef gtk_alignment_get_type
extern __typeof (gtk_alignment_get_type) gtk_alignment_get_type __attribute((alias("IA__gtk_alignment_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_alignment_new
extern __typeof (gtk_alignment_new) gtk_alignment_new __attribute((alias("IA__gtk_alignment_new"), visibility("default")));

#undef gtk_alignment_set
extern __typeof (gtk_alignment_set) gtk_alignment_set __attribute((alias("IA__gtk_alignment_set"), visibility("default")));

#undef gtk_alignment_set_padding
extern __typeof (gtk_alignment_set_padding) gtk_alignment_set_padding __attribute((alias("IA__gtk_alignment_set_padding"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_ASSISTANT_H__)
#if IN_FILE(__GTK_ASSISTANT_C__)
#undef gtk_assistant_get_type
extern __typeof (gtk_assistant_get_type) gtk_assistant_get_type __attribute((alias("IA__gtk_assistant_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_assistant_new
extern __typeof (gtk_assistant_new) gtk_assistant_new __attribute((alias("IA__gtk_assistant_new"), visibility("default")));

#undef gtk_assistant_get_current_page
extern __typeof (gtk_assistant_get_current_page) gtk_assistant_get_current_page __attribute((alias("IA__gtk_assistant_get_current_page"), visibility("default")));

#undef gtk_assistant_set_current_page
extern __typeof (gtk_assistant_set_current_page) gtk_assistant_set_current_page __attribute((alias("IA__gtk_assistant_set_current_page"), visibility("default")));

#undef gtk_assistant_get_n_pages
extern __typeof (gtk_assistant_get_n_pages) gtk_assistant_get_n_pages __attribute((alias("IA__gtk_assistant_get_n_pages"), visibility("default")));

#undef gtk_assistant_get_nth_page
extern __typeof (gtk_assistant_get_nth_page) gtk_assistant_get_nth_page __attribute((alias("IA__gtk_assistant_get_nth_page"), visibility("default")));

#undef gtk_assistant_prepend_page
extern __typeof (gtk_assistant_prepend_page) gtk_assistant_prepend_page __attribute((alias("IA__gtk_assistant_prepend_page"), visibility("default")));

#undef gtk_assistant_append_page
extern __typeof (gtk_assistant_append_page) gtk_assistant_append_page __attribute((alias("IA__gtk_assistant_append_page"), visibility("default")));

#undef gtk_assistant_insert_page
extern __typeof (gtk_assistant_insert_page) gtk_assistant_insert_page __attribute((alias("IA__gtk_assistant_insert_page"), visibility("default")));

#undef gtk_assistant_set_forward_page_func
extern __typeof (gtk_assistant_set_forward_page_func) gtk_assistant_set_forward_page_func __attribute((alias("IA__gtk_assistant_set_forward_page_func"), visibility("default")));

#undef gtk_assistant_set_page_type
extern __typeof (gtk_assistant_set_page_type) gtk_assistant_set_page_type __attribute((alias("IA__gtk_assistant_set_page_type"), visibility("default")));

#undef gtk_assistant_get_page_type
extern __typeof (gtk_assistant_get_page_type) gtk_assistant_get_page_type __attribute((alias("IA__gtk_assistant_get_page_type"), visibility("default")));

#undef gtk_assistant_set_page_title
extern __typeof (gtk_assistant_set_page_title) gtk_assistant_set_page_title __attribute((alias("IA__gtk_assistant_set_page_title"), visibility("default")));

#undef gtk_assistant_get_page_title
extern __typeof (gtk_assistant_get_page_title) gtk_assistant_get_page_title __attribute((alias("IA__gtk_assistant_get_page_title"), visibility("default")));

#undef gtk_assistant_set_page_header_image
extern __typeof (gtk_assistant_set_page_header_image) gtk_assistant_set_page_header_image __attribute((alias("IA__gtk_assistant_set_page_header_image"), visibility("default")));

#undef gtk_assistant_get_page_header_image
extern __typeof (gtk_assistant_get_page_header_image) gtk_assistant_get_page_header_image __attribute((alias("IA__gtk_assistant_get_page_header_image"), visibility("default")));

#undef gtk_assistant_set_page_side_image
extern __typeof (gtk_assistant_set_page_side_image) gtk_assistant_set_page_side_image __attribute((alias("IA__gtk_assistant_set_page_side_image"), visibility("default")));

#undef gtk_assistant_get_page_side_image
extern __typeof (gtk_assistant_get_page_side_image) gtk_assistant_get_page_side_image __attribute((alias("IA__gtk_assistant_get_page_side_image"), visibility("default")));

#undef gtk_assistant_set_page_complete
extern __typeof (gtk_assistant_set_page_complete) gtk_assistant_set_page_complete __attribute((alias("IA__gtk_assistant_set_page_complete"), visibility("default")));

#undef gtk_assistant_get_page_complete
extern __typeof (gtk_assistant_get_page_complete) gtk_assistant_get_page_complete __attribute((alias("IA__gtk_assistant_get_page_complete"), visibility("default")));

#undef gtk_assistant_add_action_widget
extern __typeof (gtk_assistant_add_action_widget) gtk_assistant_add_action_widget __attribute((alias("IA__gtk_assistant_add_action_widget"), visibility("default")));

#undef gtk_assistant_remove_action_widget
extern __typeof (gtk_assistant_remove_action_widget) gtk_assistant_remove_action_widget __attribute((alias("IA__gtk_assistant_remove_action_widget"), visibility("default")));

#undef gtk_assistant_update_buttons_state
extern __typeof (gtk_assistant_update_buttons_state) gtk_assistant_update_buttons_state __attribute((alias("IA__gtk_assistant_update_buttons_state"), visibility("default")));

#undef gtk_assistant_commit
extern __typeof (gtk_assistant_commit) gtk_assistant_commit __attribute((alias("IA__gtk_assistant_commit"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TYPE_BUILTINS_H__)
#if IN_FILE(__GTK_TYPE_BUILTINS_C__)
#undef gtk_pack_direction_get_type
extern __typeof (gtk_pack_direction_get_type) gtk_pack_direction_get_type __attribute((alias("IA__gtk_pack_direction_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_ui_manager_item_type_get_type
extern __typeof (gtk_ui_manager_item_type_get_type) gtk_ui_manager_item_type_get_type __attribute((alias("IA__gtk_ui_manager_item_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_spin_button_update_policy_get_type
extern __typeof (gtk_spin_button_update_policy_get_type) gtk_spin_button_update_policy_get_type __attribute((alias("IA__gtk_spin_button_update_policy_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_notebook_tab_get_type
extern __typeof (gtk_notebook_tab_get_type) gtk_notebook_tab_get_type __attribute((alias("IA__gtk_notebook_tab_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_number_up_layout_get_type
extern __typeof (gtk_number_up_layout_get_type) gtk_number_up_layout_get_type __attribute((alias("IA__gtk_number_up_layout_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_anchor_type_get_type
extern __typeof (gtk_anchor_type_get_type) gtk_anchor_type_get_type __attribute((alias("IA__gtk_anchor_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_arg_flags_get_type
extern __typeof (gtk_arg_flags_get_type) gtk_arg_flags_get_type __attribute((alias("IA__gtk_arg_flags_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_arrow_type_get_type
extern __typeof (gtk_arrow_type_get_type) gtk_arrow_type_get_type __attribute((alias("IA__gtk_arrow_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_arrow_placement_get_type
extern __typeof (gtk_arrow_placement_get_type) gtk_arrow_placement_get_type __attribute((alias("IA__gtk_arrow_placement_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_assistant_page_type_get_type
extern __typeof (gtk_assistant_page_type_get_type) gtk_assistant_page_type_get_type __attribute((alias("IA__gtk_assistant_page_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_attach_options_get_type
extern __typeof (gtk_attach_options_get_type) gtk_attach_options_get_type __attribute((alias("IA__gtk_attach_options_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_builder_error_get_type
extern __typeof (gtk_builder_error_get_type) gtk_builder_error_get_type __attribute((alias("IA__gtk_builder_error_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_button_action_get_type
extern __typeof (gtk_button_action_get_type) gtk_button_action_get_type __attribute((alias("IA__gtk_button_action_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_buttons_type_get_type
extern __typeof (gtk_buttons_type_get_type) gtk_buttons_type_get_type __attribute((alias("IA__gtk_buttons_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_calendar_display_options_get_type
extern __typeof (gtk_calendar_display_options_get_type) gtk_calendar_display_options_get_type __attribute((alias("IA__gtk_calendar_display_options_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_cell_renderer_mode_get_type
extern __typeof (gtk_cell_renderer_mode_get_type) gtk_cell_renderer_mode_get_type __attribute((alias("IA__gtk_cell_renderer_mode_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_cell_renderer_state_get_type
extern __typeof (gtk_cell_renderer_state_get_type) gtk_cell_renderer_state_get_type __attribute((alias("IA__gtk_cell_renderer_state_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_cell_renderer_accel_mode_get_type
extern __typeof (gtk_cell_renderer_accel_mode_get_type) gtk_cell_renderer_accel_mode_get_type __attribute((alias("IA__gtk_cell_renderer_accel_mode_get_type"), visibility("default")));

#undef gtk_corner_type_get_type
extern __typeof (gtk_corner_type_get_type) gtk_corner_type_get_type __attribute((alias("IA__gtk_corner_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_debug_flag_get_type
extern __typeof (gtk_debug_flag_get_type) gtk_debug_flag_get_type __attribute((alias("IA__gtk_debug_flag_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_delete_type_get_type
extern __typeof (gtk_delete_type_get_type) gtk_delete_type_get_type __attribute((alias("IA__gtk_delete_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_dest_defaults_get_type
extern __typeof (gtk_dest_defaults_get_type) gtk_dest_defaults_get_type __attribute((alias("IA__gtk_dest_defaults_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_drag_result_get_type
extern __typeof (gtk_drag_result_get_type) gtk_drag_result_get_type __attribute((alias("IA__gtk_drag_result_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_clist_drag_pos_get_type
extern __typeof (gtk_clist_drag_pos_get_type) gtk_clist_drag_pos_get_type __attribute((alias("IA__gtk_clist_drag_pos_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_dialog_flags_get_type
extern __typeof (gtk_dialog_flags_get_type) gtk_dialog_flags_get_type __attribute((alias("IA__gtk_dialog_flags_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_direction_type_get_type
extern __typeof (gtk_direction_type_get_type) gtk_direction_type_get_type __attribute((alias("IA__gtk_direction_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_entry_icon_position_get_type
extern __typeof (gtk_entry_icon_position_get_type) gtk_entry_icon_position_get_type __attribute((alias("IA__gtk_entry_icon_position_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_expander_style_get_type
extern __typeof (gtk_expander_style_get_type) gtk_expander_style_get_type __attribute((alias("IA__gtk_expander_style_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_file_chooser_action_get_type
extern __typeof (gtk_file_chooser_action_get_type) gtk_file_chooser_action_get_type __attribute((alias("IA__gtk_file_chooser_action_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_file_chooser_confirmation_get_type
extern __typeof (gtk_file_chooser_confirmation_get_type) gtk_file_chooser_confirmation_get_type __attribute((alias("IA__gtk_file_chooser_confirmation_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_file_filter_flags_get_type
extern __typeof (gtk_file_filter_flags_get_type) gtk_file_filter_flags_get_type __attribute((alias("IA__gtk_file_filter_flags_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_icon_view_drop_position_get_type
extern __typeof (gtk_icon_view_drop_position_get_type) gtk_icon_view_drop_position_get_type __attribute((alias("IA__gtk_icon_view_drop_position_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_sensitivity_type_get_type
extern __typeof (gtk_sensitivity_type_get_type) gtk_sensitivity_type_get_type __attribute((alias("IA__gtk_sensitivity_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_window_type_get_type
extern __typeof (gtk_window_type_get_type) gtk_window_type_get_type __attribute((alias("IA__gtk_window_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_wrap_mode_get_type
extern __typeof (gtk_wrap_mode_get_type) gtk_wrap_mode_get_type __attribute((alias("IA__gtk_wrap_mode_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_widget_flags_get_type
extern __typeof (gtk_widget_flags_get_type) gtk_widget_flags_get_type __attribute((alias("IA__gtk_widget_flags_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_widget_help_type_get_type
extern __typeof (gtk_widget_help_type_get_type) gtk_widget_help_type_get_type __attribute((alias("IA__gtk_widget_help_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_window_position_get_type
extern __typeof (gtk_window_position_get_type) gtk_window_position_get_type __attribute((alias("IA__gtk_window_position_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_view_column_sizing_get_type
extern __typeof (gtk_tree_view_column_sizing_get_type) gtk_tree_view_column_sizing_get_type __attribute((alias("IA__gtk_tree_view_column_sizing_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_view_drop_position_get_type
extern __typeof (gtk_tree_view_drop_position_get_type) gtk_tree_view_drop_position_get_type __attribute((alias("IA__gtk_tree_view_drop_position_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_button_box_style_get_type
extern __typeof (gtk_button_box_style_get_type) gtk_button_box_style_get_type __attribute((alias("IA__gtk_button_box_style_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_cell_type_get_type
extern __typeof (gtk_cell_type_get_type) gtk_cell_type_get_type __attribute((alias("IA__gtk_cell_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_ctree_expander_style_get_type
extern __typeof (gtk_ctree_expander_style_get_type) gtk_ctree_expander_style_get_type __attribute((alias("IA__gtk_ctree_expander_style_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_ctree_expansion_type_get_type
extern __typeof (gtk_ctree_expansion_type_get_type) gtk_ctree_expansion_type_get_type __attribute((alias("IA__gtk_ctree_expansion_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_ctree_line_style_get_type
extern __typeof (gtk_ctree_line_style_get_type) gtk_ctree_line_style_get_type __attribute((alias("IA__gtk_ctree_line_style_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_ctree_pos_get_type
extern __typeof (gtk_ctree_pos_get_type) gtk_ctree_pos_get_type __attribute((alias("IA__gtk_ctree_pos_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_curve_type_get_type
extern __typeof (gtk_curve_type_get_type) gtk_curve_type_get_type __attribute((alias("IA__gtk_curve_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_icon_lookup_flags_get_type
extern __typeof (gtk_icon_lookup_flags_get_type) gtk_icon_lookup_flags_get_type __attribute((alias("IA__gtk_icon_lookup_flags_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_image_type_get_type
extern __typeof (gtk_image_type_get_type) gtk_image_type_get_type __attribute((alias("IA__gtk_image_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_im_preedit_style_get_type
extern __typeof (gtk_im_preedit_style_get_type) gtk_im_preedit_style_get_type __attribute((alias("IA__gtk_im_preedit_style_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_im_status_style_get_type
extern __typeof (gtk_im_status_style_get_type) gtk_im_status_style_get_type __attribute((alias("IA__gtk_im_status_style_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_justification_get_type
extern __typeof (gtk_justification_get_type) gtk_justification_get_type __attribute((alias("IA__gtk_justification_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_match_type_get_type
extern __typeof (gtk_match_type_get_type) gtk_match_type_get_type __attribute((alias("IA__gtk_match_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_menu_direction_type_get_type
extern __typeof (gtk_menu_direction_type_get_type) gtk_menu_direction_type_get_type __attribute((alias("IA__gtk_menu_direction_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_message_type_get_type
extern __typeof (gtk_message_type_get_type) gtk_message_type_get_type __attribute((alias("IA__gtk_message_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_metric_type_get_type
extern __typeof (gtk_metric_type_get_type) gtk_metric_type_get_type __attribute((alias("IA__gtk_metric_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_movement_step_get_type
extern __typeof (gtk_movement_step_get_type) gtk_movement_step_get_type __attribute((alias("IA__gtk_movement_step_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_orientation_get_type
extern __typeof (gtk_orientation_get_type) gtk_orientation_get_type __attribute((alias("IA__gtk_orientation_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_pack_type_get_type
extern __typeof (gtk_pack_type_get_type) gtk_pack_type_get_type __attribute((alias("IA__gtk_pack_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_page_orientation_get_type
extern __typeof (gtk_page_orientation_get_type) gtk_page_orientation_get_type __attribute((alias("IA__gtk_page_orientation_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_print_operation_result_get_type
extern __typeof (gtk_print_operation_result_get_type) gtk_print_operation_result_get_type __attribute((alias("IA__gtk_print_operation_result_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_print_pages_get_type
extern __typeof (gtk_print_pages_get_type) gtk_print_pages_get_type __attribute((alias("IA__gtk_print_pages_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_print_quality_get_type
extern __typeof (gtk_print_quality_get_type) gtk_print_quality_get_type __attribute((alias("IA__gtk_print_quality_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_print_status_get_type
extern __typeof (gtk_print_status_get_type) gtk_print_status_get_type __attribute((alias("IA__gtk_print_status_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_page_set_get_type
extern __typeof (gtk_page_set_get_type) gtk_page_set_get_type __attribute((alias("IA__gtk_page_set_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_path_priority_type_get_type
extern __typeof (gtk_path_priority_type_get_type) gtk_path_priority_type_get_type __attribute((alias("IA__gtk_path_priority_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_path_type_get_type
extern __typeof (gtk_path_type_get_type) gtk_path_type_get_type __attribute((alias("IA__gtk_path_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_policy_type_get_type
extern __typeof (gtk_policy_type_get_type) gtk_policy_type_get_type __attribute((alias("IA__gtk_policy_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_position_type_get_type
extern __typeof (gtk_position_type_get_type) gtk_position_type_get_type __attribute((alias("IA__gtk_position_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_preview_type_get_type
extern __typeof (gtk_preview_type_get_type) gtk_preview_type_get_type __attribute((alias("IA__gtk_preview_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_print_duplex_get_type
extern __typeof (gtk_print_duplex_get_type) gtk_print_duplex_get_type __attribute((alias("IA__gtk_print_duplex_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_print_error_get_type
extern __typeof (gtk_print_error_get_type) gtk_print_error_get_type __attribute((alias("IA__gtk_print_error_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_recent_filter_flags_get_type
extern __typeof (gtk_recent_filter_flags_get_type) gtk_recent_filter_flags_get_type __attribute((alias("IA__gtk_recent_filter_flags_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_print_operation_action_get_type
extern __typeof (gtk_print_operation_action_get_type) gtk_print_operation_action_get_type __attribute((alias("IA__gtk_print_operation_action_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_private_flags_get_type
extern __typeof (gtk_private_flags_get_type) gtk_private_flags_get_type __attribute((alias("IA__gtk_private_flags_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_progress_bar_orientation_get_type
extern __typeof (gtk_progress_bar_orientation_get_type) gtk_progress_bar_orientation_get_type __attribute((alias("IA__gtk_progress_bar_orientation_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_progress_bar_style_get_type
extern __typeof (gtk_progress_bar_style_get_type) gtk_progress_bar_style_get_type __attribute((alias("IA__gtk_progress_bar_style_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_rc_flags_get_type
extern __typeof (gtk_rc_flags_get_type) gtk_rc_flags_get_type __attribute((alias("IA__gtk_rc_flags_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_rc_token_type_get_type
extern __typeof (gtk_rc_token_type_get_type) gtk_rc_token_type_get_type __attribute((alias("IA__gtk_rc_token_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_recent_chooser_error_get_type
extern __typeof (gtk_recent_chooser_error_get_type) gtk_recent_chooser_error_get_type __attribute((alias("IA__gtk_recent_chooser_error_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_recent_manager_error_get_type
extern __typeof (gtk_recent_manager_error_get_type) gtk_recent_manager_error_get_type __attribute((alias("IA__gtk_recent_manager_error_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_recent_sort_type_get_type
extern __typeof (gtk_recent_sort_type_get_type) gtk_recent_sort_type_get_type __attribute((alias("IA__gtk_recent_sort_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_unit_get_type
extern __typeof (gtk_unit_get_type) gtk_unit_get_type __attribute((alias("IA__gtk_unit_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_relief_style_get_type
extern __typeof (gtk_relief_style_get_type) gtk_relief_style_get_type __attribute((alias("IA__gtk_relief_style_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_resize_mode_get_type
extern __typeof (gtk_resize_mode_get_type) gtk_resize_mode_get_type __attribute((alias("IA__gtk_resize_mode_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_response_type_get_type
extern __typeof (gtk_response_type_get_type) gtk_response_type_get_type __attribute((alias("IA__gtk_response_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_scroll_step_get_type
extern __typeof (gtk_scroll_step_get_type) gtk_scroll_step_get_type __attribute((alias("IA__gtk_scroll_step_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_scroll_type_get_type
extern __typeof (gtk_scroll_type_get_type) gtk_scroll_type_get_type __attribute((alias("IA__gtk_scroll_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_selection_mode_get_type
extern __typeof (gtk_selection_mode_get_type) gtk_selection_mode_get_type __attribute((alias("IA__gtk_selection_mode_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_shadow_type_get_type
extern __typeof (gtk_shadow_type_get_type) gtk_shadow_type_get_type __attribute((alias("IA__gtk_shadow_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_side_type_get_type
extern __typeof (gtk_side_type_get_type) gtk_side_type_get_type __attribute((alias("IA__gtk_side_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_signal_run_type_get_type
extern __typeof (gtk_signal_run_type_get_type) gtk_signal_run_type_get_type __attribute((alias("IA__gtk_signal_run_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_size_group_mode_get_type
extern __typeof (gtk_size_group_mode_get_type) gtk_size_group_mode_get_type __attribute((alias("IA__gtk_size_group_mode_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_sort_type_get_type
extern __typeof (gtk_sort_type_get_type) gtk_sort_type_get_type __attribute((alias("IA__gtk_sort_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_spin_type_get_type
extern __typeof (gtk_spin_type_get_type) gtk_spin_type_get_type __attribute((alias("IA__gtk_spin_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_state_type_get_type
extern __typeof (gtk_state_type_get_type) gtk_state_type_get_type __attribute((alias("IA__gtk_state_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_submenu_direction_get_type
extern __typeof (gtk_submenu_direction_get_type) gtk_submenu_direction_get_type __attribute((alias("IA__gtk_submenu_direction_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_submenu_placement_get_type
extern __typeof (gtk_submenu_placement_get_type) gtk_submenu_placement_get_type __attribute((alias("IA__gtk_submenu_placement_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_target_flags_get_type
extern __typeof (gtk_target_flags_get_type) gtk_target_flags_get_type __attribute((alias("IA__gtk_target_flags_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_text_direction_get_type
extern __typeof (gtk_text_direction_get_type) gtk_text_direction_get_type __attribute((alias("IA__gtk_text_direction_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_text_buffer_target_info_get_type
extern __typeof (gtk_text_buffer_target_info_get_type) gtk_text_buffer_target_info_get_type __attribute((alias("IA__gtk_text_buffer_target_info_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_text_search_flags_get_type
extern __typeof (gtk_text_search_flags_get_type) gtk_text_search_flags_get_type __attribute((alias("IA__gtk_text_search_flags_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_text_window_type_get_type
extern __typeof (gtk_text_window_type_get_type) gtk_text_window_type_get_type __attribute((alias("IA__gtk_text_window_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tool_palette_drag_targets_get_type
extern __typeof (gtk_tool_palette_drag_targets_get_type) gtk_tool_palette_drag_targets_get_type __attribute((alias("IA__gtk_tool_palette_drag_targets_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_model_flags_get_type
extern __typeof (gtk_tree_model_flags_get_type) gtk_tree_model_flags_get_type __attribute((alias("IA__gtk_tree_model_flags_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_view_mode_get_type
extern __typeof (gtk_tree_view_mode_get_type) gtk_tree_view_mode_get_type __attribute((alias("IA__gtk_tree_view_mode_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_view_grid_lines_get_type
extern __typeof (gtk_tree_view_grid_lines_get_type) gtk_tree_view_grid_lines_get_type __attribute((alias("IA__gtk_tree_view_grid_lines_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_update_type_get_type
extern __typeof (gtk_update_type_get_type) gtk_update_type_get_type __attribute((alias("IA__gtk_update_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_visibility_get_type
extern __typeof (gtk_visibility_get_type) gtk_visibility_get_type __attribute((alias("IA__gtk_visibility_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_object_flags_get_type
extern __typeof (gtk_object_flags_get_type) gtk_object_flags_get_type __attribute((alias("IA__gtk_object_flags_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_accel_flags_get_type
extern __typeof (gtk_accel_flags_get_type) gtk_accel_flags_get_type __attribute((alias("IA__gtk_accel_flags_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_icon_size_get_type
extern __typeof (gtk_icon_size_get_type) gtk_icon_size_get_type __attribute((alias("IA__gtk_icon_size_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_file_chooser_error_get_type
extern __typeof (gtk_file_chooser_error_get_type) gtk_file_chooser_error_get_type __attribute((alias("IA__gtk_file_chooser_error_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_icon_theme_error_get_type
extern __typeof (gtk_icon_theme_error_get_type) gtk_icon_theme_error_get_type __attribute((alias("IA__gtk_icon_theme_error_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_toolbar_child_type_get_type
extern __typeof (gtk_toolbar_child_type_get_type) gtk_toolbar_child_type_get_type __attribute((alias("IA__gtk_toolbar_child_type_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_toolbar_space_style_get_type
extern __typeof (gtk_toolbar_space_style_get_type) gtk_toolbar_space_style_get_type __attribute((alias("IA__gtk_toolbar_space_style_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_toolbar_style_get_type
extern __typeof (gtk_toolbar_style_get_type) gtk_toolbar_style_get_type __attribute((alias("IA__gtk_toolbar_style_get_type"), visibility("default"))) G_GNUC_CONST;

#endif
#endif
#if IN_HEADER(__GTK_ARROW_H__)
#if IN_FILE(__GTK_ARROW_C__)
#undef gtk_arrow_get_type
extern __typeof (gtk_arrow_get_type) gtk_arrow_get_type __attribute((alias("IA__gtk_arrow_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_arrow_new
extern __typeof (gtk_arrow_new) gtk_arrow_new __attribute((alias("IA__gtk_arrow_new"), visibility("default")));

#undef gtk_arrow_set
extern __typeof (gtk_arrow_set) gtk_arrow_set __attribute((alias("IA__gtk_arrow_set"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_ASPECT_FRAME_H__)
#if IN_FILE(__GTK_ASPECT_FRAME_C__)
#undef gtk_aspect_frame_get_type
extern __typeof (gtk_aspect_frame_get_type) gtk_aspect_frame_get_type __attribute((alias("IA__gtk_aspect_frame_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_aspect_frame_new
extern __typeof (gtk_aspect_frame_new) gtk_aspect_frame_new __attribute((alias("IA__gtk_aspect_frame_new"), visibility("default")));

#undef gtk_aspect_frame_set
extern __typeof (gtk_aspect_frame_set) gtk_aspect_frame_set __attribute((alias("IA__gtk_aspect_frame_set"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_BINDINGS_H__)
#if IN_FILE(__GTK_BINDINGS_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_binding_entry_clear
extern __typeof (gtk_binding_entry_clear) gtk_binding_entry_clear __attribute((alias("IA__gtk_binding_entry_clear"), visibility("default")));

#undef gtk_binding_parse_binding
extern __typeof (gtk_binding_parse_binding) gtk_binding_parse_binding __attribute((alias("IA__gtk_binding_parse_binding"), visibility("default")));

#endif
#undef gtk_binding_entry_skip
extern __typeof (gtk_binding_entry_skip) gtk_binding_entry_skip __attribute((alias("IA__gtk_binding_entry_skip"), visibility("default")));

#undef gtk_binding_entry_add_signal
extern __typeof (gtk_binding_entry_add_signal) gtk_binding_entry_add_signal __attribute((alias("IA__gtk_binding_entry_add_signal"), visibility("default")));

#undef gtk_binding_entry_add_signall
extern __typeof (gtk_binding_entry_add_signall) gtk_binding_entry_add_signall __attribute((alias("IA__gtk_binding_entry_add_signall"), visibility("default")));

#undef gtk_binding_entry_remove
extern __typeof (gtk_binding_entry_remove) gtk_binding_entry_remove __attribute((alias("IA__gtk_binding_entry_remove"), visibility("default")));

#undef gtk_bindings_activate
extern __typeof (gtk_bindings_activate) gtk_bindings_activate __attribute((alias("IA__gtk_bindings_activate"), visibility("default")));

#undef gtk_bindings_activate_event
extern __typeof (gtk_bindings_activate_event) gtk_bindings_activate_event __attribute((alias("IA__gtk_bindings_activate_event"), visibility("default")));

#undef gtk_binding_set_activate
extern __typeof (gtk_binding_set_activate) gtk_binding_set_activate __attribute((alias("IA__gtk_binding_set_activate"), visibility("default")));

#undef gtk_binding_set_add_path
extern __typeof (gtk_binding_set_add_path) gtk_binding_set_add_path __attribute((alias("IA__gtk_binding_set_add_path"), visibility("default")));

#undef gtk_binding_set_by_class
extern __typeof (gtk_binding_set_by_class) gtk_binding_set_by_class __attribute((alias("IA__gtk_binding_set_by_class"), visibility("default")));

#undef gtk_binding_set_find
extern __typeof (gtk_binding_set_find) gtk_binding_set_find __attribute((alias("IA__gtk_binding_set_find"), visibility("default")));

#undef gtk_binding_set_new
extern __typeof (gtk_binding_set_new) gtk_binding_set_new __attribute((alias("IA__gtk_binding_set_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_BIN_H__)
#if IN_FILE(__GTK_BIN_C__)
#undef gtk_bin_get_child
extern __typeof (gtk_bin_get_child) gtk_bin_get_child __attribute((alias("IA__gtk_bin_get_child"), visibility("default")));

#undef gtk_bin_get_type
extern __typeof (gtk_bin_get_type) gtk_bin_get_type __attribute((alias("IA__gtk_bin_get_type"), visibility("default"))) G_GNUC_CONST;

#endif
#endif
#if IN_HEADER(__GTK_BOX_H__)
#if IN_FILE(__GTK_BOX_C__)
#undef gtk_box_get_homogeneous
extern __typeof (gtk_box_get_homogeneous) gtk_box_get_homogeneous __attribute((alias("IA__gtk_box_get_homogeneous"), visibility("default")));

#undef gtk_box_get_spacing
extern __typeof (gtk_box_get_spacing) gtk_box_get_spacing __attribute((alias("IA__gtk_box_get_spacing"), visibility("default")));

#undef gtk_box_get_type
extern __typeof (gtk_box_get_type) gtk_box_get_type __attribute((alias("IA__gtk_box_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_box_pack_end
extern __typeof (gtk_box_pack_end) gtk_box_pack_end __attribute((alias("IA__gtk_box_pack_end"), visibility("default")));

#undef gtk_box_pack_start
extern __typeof (gtk_box_pack_start) gtk_box_pack_start __attribute((alias("IA__gtk_box_pack_start"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_box_pack_end_defaults
extern __typeof (gtk_box_pack_end_defaults) gtk_box_pack_end_defaults __attribute((alias("IA__gtk_box_pack_end_defaults"), visibility("default")));

#undef gtk_box_pack_start_defaults
extern __typeof (gtk_box_pack_start_defaults) gtk_box_pack_start_defaults __attribute((alias("IA__gtk_box_pack_start_defaults"), visibility("default")));

#endif
#undef gtk_box_query_child_packing
extern __typeof (gtk_box_query_child_packing) gtk_box_query_child_packing __attribute((alias("IA__gtk_box_query_child_packing"), visibility("default")));

#undef gtk_box_reorder_child
extern __typeof (gtk_box_reorder_child) gtk_box_reorder_child __attribute((alias("IA__gtk_box_reorder_child"), visibility("default")));

#undef gtk_box_set_child_packing
extern __typeof (gtk_box_set_child_packing) gtk_box_set_child_packing __attribute((alias("IA__gtk_box_set_child_packing"), visibility("default")));

#undef gtk_box_set_homogeneous
extern __typeof (gtk_box_set_homogeneous) gtk_box_set_homogeneous __attribute((alias("IA__gtk_box_set_homogeneous"), visibility("default")));

#undef gtk_box_set_spacing
extern __typeof (gtk_box_set_spacing) gtk_box_set_spacing __attribute((alias("IA__gtk_box_set_spacing"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_BUILDABLE_H__)
#if IN_FILE(__GTK_BUILDABLE_C__)
#undef gtk_buildable_add_child
extern __typeof (gtk_buildable_add_child) gtk_buildable_add_child __attribute((alias("IA__gtk_buildable_add_child"), visibility("default")));

#undef gtk_buildable_construct_child
extern __typeof (gtk_buildable_construct_child) gtk_buildable_construct_child __attribute((alias("IA__gtk_buildable_construct_child"), visibility("default")));

#undef gtk_buildable_custom_tag_start
extern __typeof (gtk_buildable_custom_tag_start) gtk_buildable_custom_tag_start __attribute((alias("IA__gtk_buildable_custom_tag_start"), visibility("default")));

#undef gtk_buildable_custom_tag_end
extern __typeof (gtk_buildable_custom_tag_end) gtk_buildable_custom_tag_end __attribute((alias("IA__gtk_buildable_custom_tag_end"), visibility("default")));

#undef gtk_buildable_custom_finished
extern __typeof (gtk_buildable_custom_finished) gtk_buildable_custom_finished __attribute((alias("IA__gtk_buildable_custom_finished"), visibility("default")));

#undef gtk_buildable_get_internal_child
extern __typeof (gtk_buildable_get_internal_child) gtk_buildable_get_internal_child __attribute((alias("IA__gtk_buildable_get_internal_child"), visibility("default")));

#undef gtk_buildable_get_name
extern __typeof (gtk_buildable_get_name) gtk_buildable_get_name __attribute((alias("IA__gtk_buildable_get_name"), visibility("default")));

#undef gtk_buildable_get_type
extern __typeof (gtk_buildable_get_type) gtk_buildable_get_type __attribute((alias("IA__gtk_buildable_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_buildable_parser_finished
extern __typeof (gtk_buildable_parser_finished) gtk_buildable_parser_finished __attribute((alias("IA__gtk_buildable_parser_finished"), visibility("default")));

#undef gtk_buildable_set_name
extern __typeof (gtk_buildable_set_name) gtk_buildable_set_name __attribute((alias("IA__gtk_buildable_set_name"), visibility("default")));

#undef gtk_buildable_set_buildable_property
extern __typeof (gtk_buildable_set_buildable_property) gtk_buildable_set_buildable_property __attribute((alias("IA__gtk_buildable_set_buildable_property"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_BUILDER_H__)
#if IN_FILE(__GTK_BUILDER_C__)
#undef gtk_builder_add_from_file
extern __typeof (gtk_builder_add_from_file) gtk_builder_add_from_file __attribute((alias("IA__gtk_builder_add_from_file"), visibility("default")));

#undef gtk_builder_add_from_string
extern __typeof (gtk_builder_add_from_string) gtk_builder_add_from_string __attribute((alias("IA__gtk_builder_add_from_string"), visibility("default")));

#undef gtk_builder_add_objects_from_file
extern __typeof (gtk_builder_add_objects_from_file) gtk_builder_add_objects_from_file __attribute((alias("IA__gtk_builder_add_objects_from_file"), visibility("default")));

#undef gtk_builder_add_objects_from_string
extern __typeof (gtk_builder_add_objects_from_string) gtk_builder_add_objects_from_string __attribute((alias("IA__gtk_builder_add_objects_from_string"), visibility("default")));

#undef gtk_builder_error_quark
extern __typeof (gtk_builder_error_quark) gtk_builder_error_quark __attribute((alias("IA__gtk_builder_error_quark"), visibility("default")));

#undef gtk_builder_get_object
extern __typeof (gtk_builder_get_object) gtk_builder_get_object __attribute((alias("IA__gtk_builder_get_object"), visibility("default")));

#undef gtk_builder_get_objects
extern __typeof (gtk_builder_get_objects) gtk_builder_get_objects __attribute((alias("IA__gtk_builder_get_objects"), visibility("default")));

#undef gtk_builder_get_translation_domain
extern __typeof (gtk_builder_get_translation_domain) gtk_builder_get_translation_domain __attribute((alias("IA__gtk_builder_get_translation_domain"), visibility("default")));

#undef gtk_builder_get_type
extern __typeof (gtk_builder_get_type) gtk_builder_get_type __attribute((alias("IA__gtk_builder_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_builder_get_type_from_name
extern __typeof (gtk_builder_get_type_from_name) gtk_builder_get_type_from_name __attribute((alias("IA__gtk_builder_get_type_from_name"), visibility("default")));

#undef gtk_builder_new
extern __typeof (gtk_builder_new) gtk_builder_new __attribute((alias("IA__gtk_builder_new"), visibility("default")));

#undef gtk_builder_set_translation_domain
extern __typeof (gtk_builder_set_translation_domain) gtk_builder_set_translation_domain __attribute((alias("IA__gtk_builder_set_translation_domain"), visibility("default")));

#undef gtk_builder_connect_signals
extern __typeof (gtk_builder_connect_signals) gtk_builder_connect_signals __attribute((alias("IA__gtk_builder_connect_signals"), visibility("default")));

#undef gtk_builder_connect_signals_full
extern __typeof (gtk_builder_connect_signals_full) gtk_builder_connect_signals_full __attribute((alias("IA__gtk_builder_connect_signals_full"), visibility("default")));

#undef gtk_builder_value_from_string
extern __typeof (gtk_builder_value_from_string) gtk_builder_value_from_string __attribute((alias("IA__gtk_builder_value_from_string"), visibility("default")));

#undef gtk_builder_value_from_string_type
extern __typeof (gtk_builder_value_from_string_type) gtk_builder_value_from_string_type __attribute((alias("IA__gtk_builder_value_from_string_type"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_BUTTON_BOX_H__)
#if IN_FILE(__GTK_BUTTON_BOX_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_button_box_set_child_size
extern __typeof (gtk_button_box_set_child_size) gtk_button_box_set_child_size __attribute((alias("IA__gtk_button_box_set_child_size"), visibility("default")));

#undef gtk_button_box_set_child_ipadding
extern __typeof (gtk_button_box_set_child_ipadding) gtk_button_box_set_child_ipadding __attribute((alias("IA__gtk_button_box_set_child_ipadding"), visibility("default")));

#undef gtk_button_box_get_child_ipadding
extern __typeof (gtk_button_box_get_child_ipadding) gtk_button_box_get_child_ipadding __attribute((alias("IA__gtk_button_box_get_child_ipadding"), visibility("default")));

#undef gtk_button_box_get_child_size
extern __typeof (gtk_button_box_get_child_size) gtk_button_box_get_child_size __attribute((alias("IA__gtk_button_box_get_child_size"), visibility("default")));

#endif
#undef gtk_button_box_get_child_secondary
extern __typeof (gtk_button_box_get_child_secondary) gtk_button_box_get_child_secondary __attribute((alias("IA__gtk_button_box_get_child_secondary"), visibility("default")));

#undef gtk_button_box_get_layout
extern __typeof (gtk_button_box_get_layout) gtk_button_box_get_layout __attribute((alias("IA__gtk_button_box_get_layout"), visibility("default")));

#undef gtk_button_box_get_type
extern __typeof (gtk_button_box_get_type) gtk_button_box_get_type __attribute((alias("IA__gtk_button_box_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_button_box_set_child_secondary
extern __typeof (gtk_button_box_set_child_secondary) gtk_button_box_set_child_secondary __attribute((alias("IA__gtk_button_box_set_child_secondary"), visibility("default")));

#undef gtk_button_box_set_layout
extern __typeof (gtk_button_box_set_layout) gtk_button_box_set_layout __attribute((alias("IA__gtk_button_box_set_layout"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_BUTTON_H__)
#if IN_FILE(__GTK_BUTTON_C__)
#undef gtk_button_clicked
extern __typeof (gtk_button_clicked) gtk_button_clicked __attribute((alias("IA__gtk_button_clicked"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_button_enter
extern __typeof (gtk_button_enter) gtk_button_enter __attribute((alias("IA__gtk_button_enter"), visibility("default")));

#endif
#undef gtk_button_get_alignment
extern __typeof (gtk_button_get_alignment) gtk_button_get_alignment __attribute((alias("IA__gtk_button_get_alignment"), visibility("default")));

#undef gtk_button_get_event_window
extern __typeof (gtk_button_get_event_window) gtk_button_get_event_window __attribute((alias("IA__gtk_button_get_event_window"), visibility("default")));

#undef gtk_button_get_focus_on_click
extern __typeof (gtk_button_get_focus_on_click) gtk_button_get_focus_on_click __attribute((alias("IA__gtk_button_get_focus_on_click"), visibility("default")));

#undef gtk_button_get_image
extern __typeof (gtk_button_get_image) gtk_button_get_image __attribute((alias("IA__gtk_button_get_image"), visibility("default")));

#undef gtk_button_get_image_position
extern __typeof (gtk_button_get_image_position) gtk_button_get_image_position __attribute((alias("IA__gtk_button_get_image_position"), visibility("default")));

#undef gtk_button_get_label
extern __typeof (gtk_button_get_label) gtk_button_get_label __attribute((alias("IA__gtk_button_get_label"), visibility("default")));

#undef gtk_button_get_relief
extern __typeof (gtk_button_get_relief) gtk_button_get_relief __attribute((alias("IA__gtk_button_get_relief"), visibility("default")));

#undef gtk_button_get_type
extern __typeof (gtk_button_get_type) gtk_button_get_type __attribute((alias("IA__gtk_button_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_button_get_use_stock
extern __typeof (gtk_button_get_use_stock) gtk_button_get_use_stock __attribute((alias("IA__gtk_button_get_use_stock"), visibility("default")));

#undef gtk_button_get_use_underline
extern __typeof (gtk_button_get_use_underline) gtk_button_get_use_underline __attribute((alias("IA__gtk_button_get_use_underline"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_button_leave
extern __typeof (gtk_button_leave) gtk_button_leave __attribute((alias("IA__gtk_button_leave"), visibility("default")));

#endif
#undef gtk_button_new
extern __typeof (gtk_button_new) gtk_button_new __attribute((alias("IA__gtk_button_new"), visibility("default")));

#undef gtk_button_new_from_stock
extern __typeof (gtk_button_new_from_stock) gtk_button_new_from_stock __attribute((alias("IA__gtk_button_new_from_stock"), visibility("default")));

#undef gtk_button_new_with_label
extern __typeof (gtk_button_new_with_label) gtk_button_new_with_label __attribute((alias("IA__gtk_button_new_with_label"), visibility("default")));

#undef gtk_button_new_with_mnemonic
extern __typeof (gtk_button_new_with_mnemonic) gtk_button_new_with_mnemonic __attribute((alias("IA__gtk_button_new_with_mnemonic"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_button_pressed
extern __typeof (gtk_button_pressed) gtk_button_pressed __attribute((alias("IA__gtk_button_pressed"), visibility("default")));

#undef gtk_button_released
extern __typeof (gtk_button_released) gtk_button_released __attribute((alias("IA__gtk_button_released"), visibility("default")));

#endif
#undef gtk_button_set_alignment
extern __typeof (gtk_button_set_alignment) gtk_button_set_alignment __attribute((alias("IA__gtk_button_set_alignment"), visibility("default")));

#undef gtk_button_set_focus_on_click
extern __typeof (gtk_button_set_focus_on_click) gtk_button_set_focus_on_click __attribute((alias("IA__gtk_button_set_focus_on_click"), visibility("default")));

#undef gtk_button_set_image
extern __typeof (gtk_button_set_image) gtk_button_set_image __attribute((alias("IA__gtk_button_set_image"), visibility("default")));

#undef gtk_button_set_image_position
extern __typeof (gtk_button_set_image_position) gtk_button_set_image_position __attribute((alias("IA__gtk_button_set_image_position"), visibility("default")));

#undef gtk_button_set_label
extern __typeof (gtk_button_set_label) gtk_button_set_label __attribute((alias("IA__gtk_button_set_label"), visibility("default")));

#undef gtk_button_set_relief
extern __typeof (gtk_button_set_relief) gtk_button_set_relief __attribute((alias("IA__gtk_button_set_relief"), visibility("default")));

#undef gtk_button_set_use_stock
extern __typeof (gtk_button_set_use_stock) gtk_button_set_use_stock __attribute((alias("IA__gtk_button_set_use_stock"), visibility("default")));

#undef gtk_button_set_use_underline
extern __typeof (gtk_button_set_use_underline) gtk_button_set_use_underline __attribute((alias("IA__gtk_button_set_use_underline"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CALENDAR_H__)
#if IN_FILE(__GTK_CALENDAR_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_calendar_display_options
extern __typeof (gtk_calendar_display_options) gtk_calendar_display_options __attribute((alias("IA__gtk_calendar_display_options"), visibility("default")));

#undef gtk_calendar_freeze
extern __typeof (gtk_calendar_freeze) gtk_calendar_freeze __attribute((alias("IA__gtk_calendar_freeze"), visibility("default")));

#undef gtk_calendar_thaw
extern __typeof (gtk_calendar_thaw) gtk_calendar_thaw __attribute((alias("IA__gtk_calendar_thaw"), visibility("default")));

#endif
#undef gtk_calendar_clear_marks
extern __typeof (gtk_calendar_clear_marks) gtk_calendar_clear_marks __attribute((alias("IA__gtk_calendar_clear_marks"), visibility("default")));

#undef gtk_calendar_get_date
extern __typeof (gtk_calendar_get_date) gtk_calendar_get_date __attribute((alias("IA__gtk_calendar_get_date"), visibility("default")));

#undef gtk_calendar_get_detail_height_rows
extern __typeof (gtk_calendar_get_detail_height_rows) gtk_calendar_get_detail_height_rows __attribute((alias("IA__gtk_calendar_get_detail_height_rows"), visibility("default")));

#undef gtk_calendar_get_detail_width_chars
extern __typeof (gtk_calendar_get_detail_width_chars) gtk_calendar_get_detail_width_chars __attribute((alias("IA__gtk_calendar_get_detail_width_chars"), visibility("default")));

#undef gtk_calendar_get_display_options
extern __typeof (gtk_calendar_get_display_options) gtk_calendar_get_display_options __attribute((alias("IA__gtk_calendar_get_display_options"), visibility("default")));

#undef gtk_calendar_get_type
extern __typeof (gtk_calendar_get_type) gtk_calendar_get_type __attribute((alias("IA__gtk_calendar_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_calendar_mark_day
extern __typeof (gtk_calendar_mark_day) gtk_calendar_mark_day __attribute((alias("IA__gtk_calendar_mark_day"), visibility("default")));

#undef gtk_calendar_new
extern __typeof (gtk_calendar_new) gtk_calendar_new __attribute((alias("IA__gtk_calendar_new"), visibility("default")));

#undef gtk_calendar_select_day
extern __typeof (gtk_calendar_select_day) gtk_calendar_select_day __attribute((alias("IA__gtk_calendar_select_day"), visibility("default")));

#undef gtk_calendar_select_month
extern __typeof (gtk_calendar_select_month) gtk_calendar_select_month __attribute((alias("IA__gtk_calendar_select_month"), visibility("default")));

#undef gtk_calendar_set_detail_func
extern __typeof (gtk_calendar_set_detail_func) gtk_calendar_set_detail_func __attribute((alias("IA__gtk_calendar_set_detail_func"), visibility("default")));

#undef gtk_calendar_set_detail_height_rows
extern __typeof (gtk_calendar_set_detail_height_rows) gtk_calendar_set_detail_height_rows __attribute((alias("IA__gtk_calendar_set_detail_height_rows"), visibility("default")));

#undef gtk_calendar_set_detail_width_chars
extern __typeof (gtk_calendar_set_detail_width_chars) gtk_calendar_set_detail_width_chars __attribute((alias("IA__gtk_calendar_set_detail_width_chars"), visibility("default")));

#undef gtk_calendar_set_display_options
extern __typeof (gtk_calendar_set_display_options) gtk_calendar_set_display_options __attribute((alias("IA__gtk_calendar_set_display_options"), visibility("default")));

#undef gtk_calendar_unmark_day
extern __typeof (gtk_calendar_unmark_day) gtk_calendar_unmark_day __attribute((alias("IA__gtk_calendar_unmark_day"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CELL_EDITABLE_H__)
#if IN_FILE(__GTK_CELL_EDITABLE_C__)
#undef gtk_cell_editable_editing_done
extern __typeof (gtk_cell_editable_editing_done) gtk_cell_editable_editing_done __attribute((alias("IA__gtk_cell_editable_editing_done"), visibility("default")));

#undef gtk_cell_editable_get_type
extern __typeof (gtk_cell_editable_get_type) gtk_cell_editable_get_type __attribute((alias("IA__gtk_cell_editable_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_cell_editable_remove_widget
extern __typeof (gtk_cell_editable_remove_widget) gtk_cell_editable_remove_widget __attribute((alias("IA__gtk_cell_editable_remove_widget"), visibility("default")));

#undef gtk_cell_editable_start_editing
extern __typeof (gtk_cell_editable_start_editing) gtk_cell_editable_start_editing __attribute((alias("IA__gtk_cell_editable_start_editing"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CELL_LAYOUT_H__)
#if IN_FILE(__GTK_CELL_LAYOUT_C__)
#undef gtk_cell_layout_add_attribute
extern __typeof (gtk_cell_layout_add_attribute) gtk_cell_layout_add_attribute __attribute((alias("IA__gtk_cell_layout_add_attribute"), visibility("default")));

#undef gtk_cell_layout_clear
extern __typeof (gtk_cell_layout_clear) gtk_cell_layout_clear __attribute((alias("IA__gtk_cell_layout_clear"), visibility("default")));

#undef gtk_cell_layout_clear_attributes
extern __typeof (gtk_cell_layout_clear_attributes) gtk_cell_layout_clear_attributes __attribute((alias("IA__gtk_cell_layout_clear_attributes"), visibility("default")));

#undef gtk_cell_layout_get_type
extern __typeof (gtk_cell_layout_get_type) gtk_cell_layout_get_type __attribute((alias("IA__gtk_cell_layout_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_cell_layout_pack_end
extern __typeof (gtk_cell_layout_pack_end) gtk_cell_layout_pack_end __attribute((alias("IA__gtk_cell_layout_pack_end"), visibility("default")));

#undef gtk_cell_layout_pack_start
extern __typeof (gtk_cell_layout_pack_start) gtk_cell_layout_pack_start __attribute((alias("IA__gtk_cell_layout_pack_start"), visibility("default")));

#undef gtk_cell_layout_get_cells
extern __typeof (gtk_cell_layout_get_cells) gtk_cell_layout_get_cells __attribute((alias("IA__gtk_cell_layout_get_cells"), visibility("default")));

#undef gtk_cell_layout_reorder
extern __typeof (gtk_cell_layout_reorder) gtk_cell_layout_reorder __attribute((alias("IA__gtk_cell_layout_reorder"), visibility("default")));

#undef gtk_cell_layout_set_attributes
extern __typeof (gtk_cell_layout_set_attributes) gtk_cell_layout_set_attributes __attribute((alias("IA__gtk_cell_layout_set_attributes"), visibility("default"))) G_GNUC_NULL_TERMINATED;

#undef gtk_cell_layout_set_cell_data_func
extern __typeof (gtk_cell_layout_set_cell_data_func) gtk_cell_layout_set_cell_data_func __attribute((alias("IA__gtk_cell_layout_set_cell_data_func"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_H__)
#if IN_FILE(__GTK_CELL_RENDERER_C__)
#undef gtk_cell_renderer_activate
extern __typeof (gtk_cell_renderer_activate) gtk_cell_renderer_activate __attribute((alias("IA__gtk_cell_renderer_activate"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_cell_renderer_editing_canceled
extern __typeof (gtk_cell_renderer_editing_canceled) gtk_cell_renderer_editing_canceled __attribute((alias("IA__gtk_cell_renderer_editing_canceled"), visibility("default")));

#endif
#undef gtk_cell_renderer_get_alignment
extern __typeof (gtk_cell_renderer_get_alignment) gtk_cell_renderer_get_alignment __attribute((alias("IA__gtk_cell_renderer_get_alignment"), visibility("default")));

#undef gtk_cell_renderer_get_fixed_size
extern __typeof (gtk_cell_renderer_get_fixed_size) gtk_cell_renderer_get_fixed_size __attribute((alias("IA__gtk_cell_renderer_get_fixed_size"), visibility("default")));

#undef gtk_cell_renderer_get_padding
extern __typeof (gtk_cell_renderer_get_padding) gtk_cell_renderer_get_padding __attribute((alias("IA__gtk_cell_renderer_get_padding"), visibility("default")));

#undef gtk_cell_renderer_get_sensitive
extern __typeof (gtk_cell_renderer_get_sensitive) gtk_cell_renderer_get_sensitive __attribute((alias("IA__gtk_cell_renderer_get_sensitive"), visibility("default")));

#undef gtk_cell_renderer_get_size
extern __typeof (gtk_cell_renderer_get_size) gtk_cell_renderer_get_size __attribute((alias("IA__gtk_cell_renderer_get_size"), visibility("default")));

#undef gtk_cell_renderer_get_type
extern __typeof (gtk_cell_renderer_get_type) gtk_cell_renderer_get_type __attribute((alias("IA__gtk_cell_renderer_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_cell_renderer_get_visible
extern __typeof (gtk_cell_renderer_get_visible) gtk_cell_renderer_get_visible __attribute((alias("IA__gtk_cell_renderer_get_visible"), visibility("default")));

#undef gtk_cell_renderer_render
extern __typeof (gtk_cell_renderer_render) gtk_cell_renderer_render __attribute((alias("IA__gtk_cell_renderer_render"), visibility("default")));

#undef gtk_cell_renderer_set_alignment
extern __typeof (gtk_cell_renderer_set_alignment) gtk_cell_renderer_set_alignment __attribute((alias("IA__gtk_cell_renderer_set_alignment"), visibility("default")));

#undef gtk_cell_renderer_set_fixed_size
extern __typeof (gtk_cell_renderer_set_fixed_size) gtk_cell_renderer_set_fixed_size __attribute((alias("IA__gtk_cell_renderer_set_fixed_size"), visibility("default")));

#undef gtk_cell_renderer_set_padding
extern __typeof (gtk_cell_renderer_set_padding) gtk_cell_renderer_set_padding __attribute((alias("IA__gtk_cell_renderer_set_padding"), visibility("default")));

#undef gtk_cell_renderer_set_sensitive
extern __typeof (gtk_cell_renderer_set_sensitive) gtk_cell_renderer_set_sensitive __attribute((alias("IA__gtk_cell_renderer_set_sensitive"), visibility("default")));

#undef gtk_cell_renderer_set_visible
extern __typeof (gtk_cell_renderer_set_visible) gtk_cell_renderer_set_visible __attribute((alias("IA__gtk_cell_renderer_set_visible"), visibility("default")));

#undef gtk_cell_renderer_start_editing
extern __typeof (gtk_cell_renderer_start_editing) gtk_cell_renderer_start_editing __attribute((alias("IA__gtk_cell_renderer_start_editing"), visibility("default")));

#undef gtk_cell_renderer_stop_editing
extern __typeof (gtk_cell_renderer_stop_editing) gtk_cell_renderer_stop_editing __attribute((alias("IA__gtk_cell_renderer_stop_editing"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_ACCEL_H__)
#if IN_FILE(__GTK_CELL_RENDERER_ACCEL_C__)
#undef gtk_cell_renderer_accel_get_type
extern __typeof (gtk_cell_renderer_accel_get_type) gtk_cell_renderer_accel_get_type __attribute((alias("IA__gtk_cell_renderer_accel_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_cell_renderer_accel_new
extern __typeof (gtk_cell_renderer_accel_new) gtk_cell_renderer_accel_new __attribute((alias("IA__gtk_cell_renderer_accel_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_COMBO_H__)
#if IN_FILE(__GTK_CELL_RENDERER_COMBO_C__)
#undef gtk_cell_renderer_combo_get_type
extern __typeof (gtk_cell_renderer_combo_get_type) gtk_cell_renderer_combo_get_type __attribute((alias("IA__gtk_cell_renderer_combo_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_cell_renderer_combo_new
extern __typeof (gtk_cell_renderer_combo_new) gtk_cell_renderer_combo_new __attribute((alias("IA__gtk_cell_renderer_combo_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_PIXBUF_H__)
#if IN_FILE(__GTK_CELL_RENDERER_PIXBUF_C__)
#undef gtk_cell_renderer_pixbuf_get_type
extern __typeof (gtk_cell_renderer_pixbuf_get_type) gtk_cell_renderer_pixbuf_get_type __attribute((alias("IA__gtk_cell_renderer_pixbuf_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_cell_renderer_pixbuf_new
extern __typeof (gtk_cell_renderer_pixbuf_new) gtk_cell_renderer_pixbuf_new __attribute((alias("IA__gtk_cell_renderer_pixbuf_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_SPIN_H__)
#if IN_FILE(__GTK_CELL_RENDERER_SPIN_C__)
#undef gtk_cell_renderer_spin_get_type
extern __typeof (gtk_cell_renderer_spin_get_type) gtk_cell_renderer_spin_get_type __attribute((alias("IA__gtk_cell_renderer_spin_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_cell_renderer_spin_new
extern __typeof (gtk_cell_renderer_spin_new) gtk_cell_renderer_spin_new __attribute((alias("IA__gtk_cell_renderer_spin_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_SPINNER_H__)
#if IN_FILE(__GTK_CELL_RENDERER_SPINNER_C__)
#undef gtk_cell_renderer_spinner_get_type
extern __typeof (gtk_cell_renderer_spinner_get_type) gtk_cell_renderer_spinner_get_type __attribute((alias("IA__gtk_cell_renderer_spinner_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_cell_renderer_spinner_new
extern __typeof (gtk_cell_renderer_spinner_new) gtk_cell_renderer_spinner_new __attribute((alias("IA__gtk_cell_renderer_spinner_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_PROGRESS_H__)
#if IN_FILE(__GTK_CELL_RENDERER_PROGRESS_C__)
#undef gtk_cell_renderer_progress_get_type
extern __typeof (gtk_cell_renderer_progress_get_type) gtk_cell_renderer_progress_get_type __attribute((alias("IA__gtk_cell_renderer_progress_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_cell_renderer_progress_new
extern __typeof (gtk_cell_renderer_progress_new) gtk_cell_renderer_progress_new __attribute((alias("IA__gtk_cell_renderer_progress_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_TEXT_H__)
#if IN_FILE(__GTK_CELL_RENDERER_TEXT_C__)
#undef gtk_cell_renderer_text_get_type
extern __typeof (gtk_cell_renderer_text_get_type) gtk_cell_renderer_text_get_type __attribute((alias("IA__gtk_cell_renderer_text_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_cell_renderer_text_new
extern __typeof (gtk_cell_renderer_text_new) gtk_cell_renderer_text_new __attribute((alias("IA__gtk_cell_renderer_text_new"), visibility("default")));

#undef gtk_cell_renderer_text_set_fixed_height_from_font
extern __typeof (gtk_cell_renderer_text_set_fixed_height_from_font) gtk_cell_renderer_text_set_fixed_height_from_font __attribute((alias("IA__gtk_cell_renderer_text_set_fixed_height_from_font"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CELL_RENDERER_TOGGLE_H__)
#if IN_FILE(__GTK_CELL_RENDERER_TOGGLE_C__)
#undef gtk_cell_renderer_toggle_get_activatable
extern __typeof (gtk_cell_renderer_toggle_get_activatable) gtk_cell_renderer_toggle_get_activatable __attribute((alias("IA__gtk_cell_renderer_toggle_get_activatable"), visibility("default")));

#undef gtk_cell_renderer_toggle_get_active
extern __typeof (gtk_cell_renderer_toggle_get_active) gtk_cell_renderer_toggle_get_active __attribute((alias("IA__gtk_cell_renderer_toggle_get_active"), visibility("default")));

#undef gtk_cell_renderer_toggle_get_radio
extern __typeof (gtk_cell_renderer_toggle_get_radio) gtk_cell_renderer_toggle_get_radio __attribute((alias("IA__gtk_cell_renderer_toggle_get_radio"), visibility("default")));

#undef gtk_cell_renderer_toggle_get_type
extern __typeof (gtk_cell_renderer_toggle_get_type) gtk_cell_renderer_toggle_get_type __attribute((alias("IA__gtk_cell_renderer_toggle_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_cell_renderer_toggle_new
extern __typeof (gtk_cell_renderer_toggle_new) gtk_cell_renderer_toggle_new __attribute((alias("IA__gtk_cell_renderer_toggle_new"), visibility("default")));

#undef gtk_cell_renderer_toggle_set_activatable
extern __typeof (gtk_cell_renderer_toggle_set_activatable) gtk_cell_renderer_toggle_set_activatable __attribute((alias("IA__gtk_cell_renderer_toggle_set_activatable"), visibility("default")));

#undef gtk_cell_renderer_toggle_set_active
extern __typeof (gtk_cell_renderer_toggle_set_active) gtk_cell_renderer_toggle_set_active __attribute((alias("IA__gtk_cell_renderer_toggle_set_active"), visibility("default")));

#undef gtk_cell_renderer_toggle_set_radio
extern __typeof (gtk_cell_renderer_toggle_set_radio) gtk_cell_renderer_toggle_set_radio __attribute((alias("IA__gtk_cell_renderer_toggle_set_radio"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CELL_VIEW_H__)
#if IN_FILE(__GTK_CELL_VIEW_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_cell_view_get_cell_renderers
extern __typeof (gtk_cell_view_get_cell_renderers) gtk_cell_view_get_cell_renderers __attribute((alias("IA__gtk_cell_view_get_cell_renderers"), visibility("default")));

#endif
#undef gtk_cell_view_get_displayed_row
extern __typeof (gtk_cell_view_get_displayed_row) gtk_cell_view_get_displayed_row __attribute((alias("IA__gtk_cell_view_get_displayed_row"), visibility("default")));

#undef gtk_cell_view_get_model
extern __typeof (gtk_cell_view_get_model) gtk_cell_view_get_model __attribute((alias("IA__gtk_cell_view_get_model"), visibility("default")));

#undef gtk_cell_view_get_size_of_row
extern __typeof (gtk_cell_view_get_size_of_row) gtk_cell_view_get_size_of_row __attribute((alias("IA__gtk_cell_view_get_size_of_row"), visibility("default")));

#undef gtk_cell_view_get_type
extern __typeof (gtk_cell_view_get_type) gtk_cell_view_get_type __attribute((alias("IA__gtk_cell_view_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_cell_view_new
extern __typeof (gtk_cell_view_new) gtk_cell_view_new __attribute((alias("IA__gtk_cell_view_new"), visibility("default")));

#undef gtk_cell_view_new_with_markup
extern __typeof (gtk_cell_view_new_with_markup) gtk_cell_view_new_with_markup __attribute((alias("IA__gtk_cell_view_new_with_markup"), visibility("default")));

#undef gtk_cell_view_new_with_pixbuf
extern __typeof (gtk_cell_view_new_with_pixbuf) gtk_cell_view_new_with_pixbuf __attribute((alias("IA__gtk_cell_view_new_with_pixbuf"), visibility("default")));

#undef gtk_cell_view_new_with_text
extern __typeof (gtk_cell_view_new_with_text) gtk_cell_view_new_with_text __attribute((alias("IA__gtk_cell_view_new_with_text"), visibility("default")));

#undef gtk_cell_view_set_background_color
extern __typeof (gtk_cell_view_set_background_color) gtk_cell_view_set_background_color __attribute((alias("IA__gtk_cell_view_set_background_color"), visibility("default")));

#undef gtk_cell_view_set_displayed_row
extern __typeof (gtk_cell_view_set_displayed_row) gtk_cell_view_set_displayed_row __attribute((alias("IA__gtk_cell_view_set_displayed_row"), visibility("default")));

#undef gtk_cell_view_set_model
extern __typeof (gtk_cell_view_set_model) gtk_cell_view_set_model __attribute((alias("IA__gtk_cell_view_set_model"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CHECK_BUTTON_H__)
#if IN_FILE(__GTK_CHECK_BUTTON_C__)
#undef gtk_check_button_get_type
extern __typeof (gtk_check_button_get_type) gtk_check_button_get_type __attribute((alias("IA__gtk_check_button_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_check_button_new
extern __typeof (gtk_check_button_new) gtk_check_button_new __attribute((alias("IA__gtk_check_button_new"), visibility("default")));

#undef gtk_check_button_new_with_label
extern __typeof (gtk_check_button_new_with_label) gtk_check_button_new_with_label __attribute((alias("IA__gtk_check_button_new_with_label"), visibility("default")));

#undef gtk_check_button_new_with_mnemonic
extern __typeof (gtk_check_button_new_with_mnemonic) gtk_check_button_new_with_mnemonic __attribute((alias("IA__gtk_check_button_new_with_mnemonic"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CHECK_MENU_ITEM_H__)
#if IN_FILE(__GTK_CHECK_MENU_ITEM_C__)
#undef gtk_check_menu_item_get_active
extern __typeof (gtk_check_menu_item_get_active) gtk_check_menu_item_get_active __attribute((alias("IA__gtk_check_menu_item_get_active"), visibility("default")));

#undef gtk_check_menu_item_get_draw_as_radio
extern __typeof (gtk_check_menu_item_get_draw_as_radio) gtk_check_menu_item_get_draw_as_radio __attribute((alias("IA__gtk_check_menu_item_get_draw_as_radio"), visibility("default")));

#undef gtk_check_menu_item_get_inconsistent
extern __typeof (gtk_check_menu_item_get_inconsistent) gtk_check_menu_item_get_inconsistent __attribute((alias("IA__gtk_check_menu_item_get_inconsistent"), visibility("default")));

#undef gtk_check_menu_item_get_type
extern __typeof (gtk_check_menu_item_get_type) gtk_check_menu_item_get_type __attribute((alias("IA__gtk_check_menu_item_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_check_menu_item_new
extern __typeof (gtk_check_menu_item_new) gtk_check_menu_item_new __attribute((alias("IA__gtk_check_menu_item_new"), visibility("default")));

#undef gtk_check_menu_item_new_with_label
extern __typeof (gtk_check_menu_item_new_with_label) gtk_check_menu_item_new_with_label __attribute((alias("IA__gtk_check_menu_item_new_with_label"), visibility("default")));

#undef gtk_check_menu_item_new_with_mnemonic
extern __typeof (gtk_check_menu_item_new_with_mnemonic) gtk_check_menu_item_new_with_mnemonic __attribute((alias("IA__gtk_check_menu_item_new_with_mnemonic"), visibility("default")));

#undef gtk_check_menu_item_set_active
extern __typeof (gtk_check_menu_item_set_active) gtk_check_menu_item_set_active __attribute((alias("IA__gtk_check_menu_item_set_active"), visibility("default")));

#undef gtk_check_menu_item_set_draw_as_radio
extern __typeof (gtk_check_menu_item_set_draw_as_radio) gtk_check_menu_item_set_draw_as_radio __attribute((alias("IA__gtk_check_menu_item_set_draw_as_radio"), visibility("default")));

#undef gtk_check_menu_item_set_inconsistent
extern __typeof (gtk_check_menu_item_set_inconsistent) gtk_check_menu_item_set_inconsistent __attribute((alias("IA__gtk_check_menu_item_set_inconsistent"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_check_menu_item_set_show_toggle
extern __typeof (gtk_check_menu_item_set_show_toggle) gtk_check_menu_item_set_show_toggle __attribute((alias("IA__gtk_check_menu_item_set_show_toggle"), visibility("default")));

#endif
#undef gtk_check_menu_item_toggled
extern __typeof (gtk_check_menu_item_toggled) gtk_check_menu_item_toggled __attribute((alias("IA__gtk_check_menu_item_toggled"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CLIPBOARD_H__)
#if IN_FILE(__GTK_CLIPBOARD_C__)
#undef gtk_clipboard_clear
extern __typeof (gtk_clipboard_clear) gtk_clipboard_clear __attribute((alias("IA__gtk_clipboard_clear"), visibility("default")));

#undef gtk_clipboard_get
extern __typeof (gtk_clipboard_get) gtk_clipboard_get __attribute((alias("IA__gtk_clipboard_get"), visibility("default")));

#undef gtk_clipboard_get_display
extern __typeof (gtk_clipboard_get_display) gtk_clipboard_get_display __attribute((alias("IA__gtk_clipboard_get_display"), visibility("default")));

#undef gtk_clipboard_get_for_display
extern __typeof (gtk_clipboard_get_for_display) gtk_clipboard_get_for_display __attribute((alias("IA__gtk_clipboard_get_for_display"), visibility("default")));

#undef gtk_clipboard_get_owner
extern __typeof (gtk_clipboard_get_owner) gtk_clipboard_get_owner __attribute((alias("IA__gtk_clipboard_get_owner"), visibility("default")));

#undef gtk_clipboard_get_type
extern __typeof (gtk_clipboard_get_type) gtk_clipboard_get_type __attribute((alias("IA__gtk_clipboard_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_clipboard_request_contents
extern __typeof (gtk_clipboard_request_contents) gtk_clipboard_request_contents __attribute((alias("IA__gtk_clipboard_request_contents"), visibility("default")));

#undef gtk_clipboard_request_image
extern __typeof (gtk_clipboard_request_image) gtk_clipboard_request_image __attribute((alias("IA__gtk_clipboard_request_image"), visibility("default")));

#undef gtk_clipboard_request_rich_text
extern __typeof (gtk_clipboard_request_rich_text) gtk_clipboard_request_rich_text __attribute((alias("IA__gtk_clipboard_request_rich_text"), visibility("default")));

#undef gtk_clipboard_request_targets
extern __typeof (gtk_clipboard_request_targets) gtk_clipboard_request_targets __attribute((alias("IA__gtk_clipboard_request_targets"), visibility("default")));

#undef gtk_clipboard_request_text
extern __typeof (gtk_clipboard_request_text) gtk_clipboard_request_text __attribute((alias("IA__gtk_clipboard_request_text"), visibility("default")));

#undef gtk_clipboard_request_uris
extern __typeof (gtk_clipboard_request_uris) gtk_clipboard_request_uris __attribute((alias("IA__gtk_clipboard_request_uris"), visibility("default")));

#undef gtk_clipboard_set_can_store
extern __typeof (gtk_clipboard_set_can_store) gtk_clipboard_set_can_store __attribute((alias("IA__gtk_clipboard_set_can_store"), visibility("default")));

#undef gtk_clipboard_set_image
extern __typeof (gtk_clipboard_set_image) gtk_clipboard_set_image __attribute((alias("IA__gtk_clipboard_set_image"), visibility("default")));

#undef gtk_clipboard_set_text
extern __typeof (gtk_clipboard_set_text) gtk_clipboard_set_text __attribute((alias("IA__gtk_clipboard_set_text"), visibility("default")));

#undef gtk_clipboard_set_with_data
extern __typeof (gtk_clipboard_set_with_data) gtk_clipboard_set_with_data __attribute((alias("IA__gtk_clipboard_set_with_data"), visibility("default")));

#undef gtk_clipboard_set_with_owner
extern __typeof (gtk_clipboard_set_with_owner) gtk_clipboard_set_with_owner __attribute((alias("IA__gtk_clipboard_set_with_owner"), visibility("default")));

#undef gtk_clipboard_store
extern __typeof (gtk_clipboard_store) gtk_clipboard_store __attribute((alias("IA__gtk_clipboard_store"), visibility("default")));

#undef gtk_clipboard_wait_for_contents
extern __typeof (gtk_clipboard_wait_for_contents) gtk_clipboard_wait_for_contents __attribute((alias("IA__gtk_clipboard_wait_for_contents"), visibility("default")));

#undef gtk_clipboard_wait_for_image
extern __typeof (gtk_clipboard_wait_for_image) gtk_clipboard_wait_for_image __attribute((alias("IA__gtk_clipboard_wait_for_image"), visibility("default")));

#undef gtk_clipboard_wait_for_rich_text
extern __typeof (gtk_clipboard_wait_for_rich_text) gtk_clipboard_wait_for_rich_text __attribute((alias("IA__gtk_clipboard_wait_for_rich_text"), visibility("default")));

#undef gtk_clipboard_wait_for_uris
extern __typeof (gtk_clipboard_wait_for_uris) gtk_clipboard_wait_for_uris __attribute((alias("IA__gtk_clipboard_wait_for_uris"), visibility("default")));

#undef gtk_clipboard_wait_for_targets
extern __typeof (gtk_clipboard_wait_for_targets) gtk_clipboard_wait_for_targets __attribute((alias("IA__gtk_clipboard_wait_for_targets"), visibility("default")));

#undef gtk_clipboard_wait_for_text
extern __typeof (gtk_clipboard_wait_for_text) gtk_clipboard_wait_for_text __attribute((alias("IA__gtk_clipboard_wait_for_text"), visibility("default")));

#undef gtk_clipboard_wait_is_image_available
extern __typeof (gtk_clipboard_wait_is_image_available) gtk_clipboard_wait_is_image_available __attribute((alias("IA__gtk_clipboard_wait_is_image_available"), visibility("default")));

#undef gtk_clipboard_wait_is_rich_text_available
extern __typeof (gtk_clipboard_wait_is_rich_text_available) gtk_clipboard_wait_is_rich_text_available __attribute((alias("IA__gtk_clipboard_wait_is_rich_text_available"), visibility("default")));

#undef gtk_clipboard_wait_is_text_available
extern __typeof (gtk_clipboard_wait_is_text_available) gtk_clipboard_wait_is_text_available __attribute((alias("IA__gtk_clipboard_wait_is_text_available"), visibility("default")));

#undef gtk_clipboard_wait_is_uris_available
extern __typeof (gtk_clipboard_wait_is_uris_available) gtk_clipboard_wait_is_uris_available __attribute((alias("IA__gtk_clipboard_wait_is_uris_available"), visibility("default")));

#undef gtk_clipboard_wait_is_target_available
extern __typeof (gtk_clipboard_wait_is_target_available) gtk_clipboard_wait_is_target_available __attribute((alias("IA__gtk_clipboard_wait_is_target_available"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CLIST_H__)
#if IN_FILE(__GTK_CLIST_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_clist_append
extern __typeof (gtk_clist_append) gtk_clist_append __attribute((alias("IA__gtk_clist_append"), visibility("default")));

#undef gtk_clist_clear
extern __typeof (gtk_clist_clear) gtk_clist_clear __attribute((alias("IA__gtk_clist_clear"), visibility("default")));

#undef gtk_clist_columns_autosize
extern __typeof (gtk_clist_columns_autosize) gtk_clist_columns_autosize __attribute((alias("IA__gtk_clist_columns_autosize"), visibility("default")));

#undef gtk_clist_column_title_active
extern __typeof (gtk_clist_column_title_active) gtk_clist_column_title_active __attribute((alias("IA__gtk_clist_column_title_active"), visibility("default")));

#undef gtk_clist_column_title_passive
extern __typeof (gtk_clist_column_title_passive) gtk_clist_column_title_passive __attribute((alias("IA__gtk_clist_column_title_passive"), visibility("default")));

#undef gtk_clist_column_titles_active
extern __typeof (gtk_clist_column_titles_active) gtk_clist_column_titles_active __attribute((alias("IA__gtk_clist_column_titles_active"), visibility("default")));

#undef gtk_clist_column_titles_hide
extern __typeof (gtk_clist_column_titles_hide) gtk_clist_column_titles_hide __attribute((alias("IA__gtk_clist_column_titles_hide"), visibility("default")));

#undef gtk_clist_column_titles_passive
extern __typeof (gtk_clist_column_titles_passive) gtk_clist_column_titles_passive __attribute((alias("IA__gtk_clist_column_titles_passive"), visibility("default")));

#undef gtk_clist_column_titles_show
extern __typeof (gtk_clist_column_titles_show) gtk_clist_column_titles_show __attribute((alias("IA__gtk_clist_column_titles_show"), visibility("default")));

#undef gtk_clist_find_row_from_data
extern __typeof (gtk_clist_find_row_from_data) gtk_clist_find_row_from_data __attribute((alias("IA__gtk_clist_find_row_from_data"), visibility("default")));

#undef gtk_clist_freeze
extern __typeof (gtk_clist_freeze) gtk_clist_freeze __attribute((alias("IA__gtk_clist_freeze"), visibility("default")));

#undef gtk_clist_get_cell_style
extern __typeof (gtk_clist_get_cell_style) gtk_clist_get_cell_style __attribute((alias("IA__gtk_clist_get_cell_style"), visibility("default")));

#undef gtk_clist_get_cell_type
extern __typeof (gtk_clist_get_cell_type) gtk_clist_get_cell_type __attribute((alias("IA__gtk_clist_get_cell_type"), visibility("default")));

#undef gtk_clist_get_column_title
extern __typeof (gtk_clist_get_column_title) gtk_clist_get_column_title __attribute((alias("IA__gtk_clist_get_column_title"), visibility("default")));

#undef gtk_clist_get_column_widget
extern __typeof (gtk_clist_get_column_widget) gtk_clist_get_column_widget __attribute((alias("IA__gtk_clist_get_column_widget"), visibility("default")));

#undef gtk_clist_get_hadjustment
extern __typeof (gtk_clist_get_hadjustment) gtk_clist_get_hadjustment __attribute((alias("IA__gtk_clist_get_hadjustment"), visibility("default")));

#undef gtk_clist_get_pixmap
extern __typeof (gtk_clist_get_pixmap) gtk_clist_get_pixmap __attribute((alias("IA__gtk_clist_get_pixmap"), visibility("default")));

#undef gtk_clist_get_pixtext
extern __typeof (gtk_clist_get_pixtext) gtk_clist_get_pixtext __attribute((alias("IA__gtk_clist_get_pixtext"), visibility("default")));

#undef gtk_clist_get_row_data
extern __typeof (gtk_clist_get_row_data) gtk_clist_get_row_data __attribute((alias("IA__gtk_clist_get_row_data"), visibility("default")));

#undef gtk_clist_get_row_style
extern __typeof (gtk_clist_get_row_style) gtk_clist_get_row_style __attribute((alias("IA__gtk_clist_get_row_style"), visibility("default")));

#undef gtk_clist_get_selectable
extern __typeof (gtk_clist_get_selectable) gtk_clist_get_selectable __attribute((alias("IA__gtk_clist_get_selectable"), visibility("default")));

#undef gtk_clist_get_selection_info
extern __typeof (gtk_clist_get_selection_info) gtk_clist_get_selection_info __attribute((alias("IA__gtk_clist_get_selection_info"), visibility("default")));

#undef gtk_clist_get_text
extern __typeof (gtk_clist_get_text) gtk_clist_get_text __attribute((alias("IA__gtk_clist_get_text"), visibility("default")));

#undef gtk_clist_get_type
extern __typeof (gtk_clist_get_type) gtk_clist_get_type __attribute((alias("IA__gtk_clist_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_clist_get_vadjustment
extern __typeof (gtk_clist_get_vadjustment) gtk_clist_get_vadjustment __attribute((alias("IA__gtk_clist_get_vadjustment"), visibility("default")));

#undef gtk_clist_insert
extern __typeof (gtk_clist_insert) gtk_clist_insert __attribute((alias("IA__gtk_clist_insert"), visibility("default")));

#undef gtk_clist_moveto
extern __typeof (gtk_clist_moveto) gtk_clist_moveto __attribute((alias("IA__gtk_clist_moveto"), visibility("default")));

#undef gtk_clist_new
extern __typeof (gtk_clist_new) gtk_clist_new __attribute((alias("IA__gtk_clist_new"), visibility("default")));

#undef gtk_clist_new_with_titles
extern __typeof (gtk_clist_new_with_titles) gtk_clist_new_with_titles __attribute((alias("IA__gtk_clist_new_with_titles"), visibility("default")));

#undef gtk_clist_optimal_column_width
extern __typeof (gtk_clist_optimal_column_width) gtk_clist_optimal_column_width __attribute((alias("IA__gtk_clist_optimal_column_width"), visibility("default")));

#undef gtk_clist_prepend
extern __typeof (gtk_clist_prepend) gtk_clist_prepend __attribute((alias("IA__gtk_clist_prepend"), visibility("default")));

#undef gtk_clist_remove
extern __typeof (gtk_clist_remove) gtk_clist_remove __attribute((alias("IA__gtk_clist_remove"), visibility("default")));

#undef gtk_clist_row_is_visible
extern __typeof (gtk_clist_row_is_visible) gtk_clist_row_is_visible __attribute((alias("IA__gtk_clist_row_is_visible"), visibility("default")));

#undef gtk_clist_row_move
extern __typeof (gtk_clist_row_move) gtk_clist_row_move __attribute((alias("IA__gtk_clist_row_move"), visibility("default")));

#undef gtk_clist_select_all
extern __typeof (gtk_clist_select_all) gtk_clist_select_all __attribute((alias("IA__gtk_clist_select_all"), visibility("default")));

#undef gtk_clist_select_row
extern __typeof (gtk_clist_select_row) gtk_clist_select_row __attribute((alias("IA__gtk_clist_select_row"), visibility("default")));

#undef gtk_clist_set_auto_sort
extern __typeof (gtk_clist_set_auto_sort) gtk_clist_set_auto_sort __attribute((alias("IA__gtk_clist_set_auto_sort"), visibility("default")));

#undef gtk_clist_set_background
extern __typeof (gtk_clist_set_background) gtk_clist_set_background __attribute((alias("IA__gtk_clist_set_background"), visibility("default")));

#undef gtk_clist_set_button_actions
extern __typeof (gtk_clist_set_button_actions) gtk_clist_set_button_actions __attribute((alias("IA__gtk_clist_set_button_actions"), visibility("default")));

#undef gtk_clist_set_cell_style
extern __typeof (gtk_clist_set_cell_style) gtk_clist_set_cell_style __attribute((alias("IA__gtk_clist_set_cell_style"), visibility("default")));

#undef gtk_clist_set_column_auto_resize
extern __typeof (gtk_clist_set_column_auto_resize) gtk_clist_set_column_auto_resize __attribute((alias("IA__gtk_clist_set_column_auto_resize"), visibility("default")));

#undef gtk_clist_set_column_justification
extern __typeof (gtk_clist_set_column_justification) gtk_clist_set_column_justification __attribute((alias("IA__gtk_clist_set_column_justification"), visibility("default")));

#undef gtk_clist_set_column_max_width
extern __typeof (gtk_clist_set_column_max_width) gtk_clist_set_column_max_width __attribute((alias("IA__gtk_clist_set_column_max_width"), visibility("default")));

#undef gtk_clist_set_column_min_width
extern __typeof (gtk_clist_set_column_min_width) gtk_clist_set_column_min_width __attribute((alias("IA__gtk_clist_set_column_min_width"), visibility("default")));

#undef gtk_clist_set_column_resizeable
extern __typeof (gtk_clist_set_column_resizeable) gtk_clist_set_column_resizeable __attribute((alias("IA__gtk_clist_set_column_resizeable"), visibility("default")));

#undef gtk_clist_set_column_title
extern __typeof (gtk_clist_set_column_title) gtk_clist_set_column_title __attribute((alias("IA__gtk_clist_set_column_title"), visibility("default")));

#undef gtk_clist_set_column_visibility
extern __typeof (gtk_clist_set_column_visibility) gtk_clist_set_column_visibility __attribute((alias("IA__gtk_clist_set_column_visibility"), visibility("default")));

#undef gtk_clist_set_column_widget
extern __typeof (gtk_clist_set_column_widget) gtk_clist_set_column_widget __attribute((alias("IA__gtk_clist_set_column_widget"), visibility("default")));

#undef gtk_clist_set_column_width
extern __typeof (gtk_clist_set_column_width) gtk_clist_set_column_width __attribute((alias("IA__gtk_clist_set_column_width"), visibility("default")));

#undef gtk_clist_set_compare_func
extern __typeof (gtk_clist_set_compare_func) gtk_clist_set_compare_func __attribute((alias("IA__gtk_clist_set_compare_func"), visibility("default")));

#undef gtk_clist_set_foreground
extern __typeof (gtk_clist_set_foreground) gtk_clist_set_foreground __attribute((alias("IA__gtk_clist_set_foreground"), visibility("default")));

#undef gtk_clist_set_hadjustment
extern __typeof (gtk_clist_set_hadjustment) gtk_clist_set_hadjustment __attribute((alias("IA__gtk_clist_set_hadjustment"), visibility("default")));

#undef gtk_clist_set_pixmap
extern __typeof (gtk_clist_set_pixmap) gtk_clist_set_pixmap __attribute((alias("IA__gtk_clist_set_pixmap"), visibility("default")));

#undef gtk_clist_set_pixtext
extern __typeof (gtk_clist_set_pixtext) gtk_clist_set_pixtext __attribute((alias("IA__gtk_clist_set_pixtext"), visibility("default")));

#undef gtk_clist_set_reorderable
extern __typeof (gtk_clist_set_reorderable) gtk_clist_set_reorderable __attribute((alias("IA__gtk_clist_set_reorderable"), visibility("default")));

#undef gtk_clist_set_row_data
extern __typeof (gtk_clist_set_row_data) gtk_clist_set_row_data __attribute((alias("IA__gtk_clist_set_row_data"), visibility("default")));

#undef gtk_clist_set_row_data_full
extern __typeof (gtk_clist_set_row_data_full) gtk_clist_set_row_data_full __attribute((alias("IA__gtk_clist_set_row_data_full"), visibility("default")));

#undef gtk_clist_set_row_height
extern __typeof (gtk_clist_set_row_height) gtk_clist_set_row_height __attribute((alias("IA__gtk_clist_set_row_height"), visibility("default")));

#undef gtk_clist_set_row_style
extern __typeof (gtk_clist_set_row_style) gtk_clist_set_row_style __attribute((alias("IA__gtk_clist_set_row_style"), visibility("default")));

#undef gtk_clist_set_selectable
extern __typeof (gtk_clist_set_selectable) gtk_clist_set_selectable __attribute((alias("IA__gtk_clist_set_selectable"), visibility("default")));

#undef gtk_clist_set_selection_mode
extern __typeof (gtk_clist_set_selection_mode) gtk_clist_set_selection_mode __attribute((alias("IA__gtk_clist_set_selection_mode"), visibility("default")));

#undef gtk_clist_set_shadow_type
extern __typeof (gtk_clist_set_shadow_type) gtk_clist_set_shadow_type __attribute((alias("IA__gtk_clist_set_shadow_type"), visibility("default")));

#undef gtk_clist_set_shift
extern __typeof (gtk_clist_set_shift) gtk_clist_set_shift __attribute((alias("IA__gtk_clist_set_shift"), visibility("default")));

#undef gtk_clist_set_sort_column
extern __typeof (gtk_clist_set_sort_column) gtk_clist_set_sort_column __attribute((alias("IA__gtk_clist_set_sort_column"), visibility("default")));

#undef gtk_clist_set_sort_type
extern __typeof (gtk_clist_set_sort_type) gtk_clist_set_sort_type __attribute((alias("IA__gtk_clist_set_sort_type"), visibility("default")));

#undef gtk_clist_set_text
extern __typeof (gtk_clist_set_text) gtk_clist_set_text __attribute((alias("IA__gtk_clist_set_text"), visibility("default")));

#undef gtk_clist_set_use_drag_icons
extern __typeof (gtk_clist_set_use_drag_icons) gtk_clist_set_use_drag_icons __attribute((alias("IA__gtk_clist_set_use_drag_icons"), visibility("default")));

#undef gtk_clist_set_vadjustment
extern __typeof (gtk_clist_set_vadjustment) gtk_clist_set_vadjustment __attribute((alias("IA__gtk_clist_set_vadjustment"), visibility("default")));

#undef gtk_clist_sort
extern __typeof (gtk_clist_sort) gtk_clist_sort __attribute((alias("IA__gtk_clist_sort"), visibility("default")));

#undef gtk_clist_swap_rows
extern __typeof (gtk_clist_swap_rows) gtk_clist_swap_rows __attribute((alias("IA__gtk_clist_swap_rows"), visibility("default")));

#undef gtk_clist_thaw
extern __typeof (gtk_clist_thaw) gtk_clist_thaw __attribute((alias("IA__gtk_clist_thaw"), visibility("default")));

#undef gtk_clist_undo_selection
extern __typeof (gtk_clist_undo_selection) gtk_clist_undo_selection __attribute((alias("IA__gtk_clist_undo_selection"), visibility("default")));

#undef gtk_clist_unselect_all
extern __typeof (gtk_clist_unselect_all) gtk_clist_unselect_all __attribute((alias("IA__gtk_clist_unselect_all"), visibility("default")));

#undef gtk_clist_unselect_row
extern __typeof (gtk_clist_unselect_row) gtk_clist_unselect_row __attribute((alias("IA__gtk_clist_unselect_row"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_COLOR_BUTTON_H__)
#if IN_FILE(__GTK_COLOR_BUTTON_C__)
#undef gtk_color_button_get_alpha
extern __typeof (gtk_color_button_get_alpha) gtk_color_button_get_alpha __attribute((alias("IA__gtk_color_button_get_alpha"), visibility("default")));

#undef gtk_color_button_get_color
extern __typeof (gtk_color_button_get_color) gtk_color_button_get_color __attribute((alias("IA__gtk_color_button_get_color"), visibility("default")));

#undef gtk_color_button_get_title
extern __typeof (gtk_color_button_get_title) gtk_color_button_get_title __attribute((alias("IA__gtk_color_button_get_title"), visibility("default")));

#undef gtk_color_button_get_type
extern __typeof (gtk_color_button_get_type) gtk_color_button_get_type __attribute((alias("IA__gtk_color_button_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_color_button_get_use_alpha
extern __typeof (gtk_color_button_get_use_alpha) gtk_color_button_get_use_alpha __attribute((alias("IA__gtk_color_button_get_use_alpha"), visibility("default")));

#undef gtk_color_button_new
extern __typeof (gtk_color_button_new) gtk_color_button_new __attribute((alias("IA__gtk_color_button_new"), visibility("default")));

#undef gtk_color_button_new_with_color
extern __typeof (gtk_color_button_new_with_color) gtk_color_button_new_with_color __attribute((alias("IA__gtk_color_button_new_with_color"), visibility("default")));

#undef gtk_color_button_set_alpha
extern __typeof (gtk_color_button_set_alpha) gtk_color_button_set_alpha __attribute((alias("IA__gtk_color_button_set_alpha"), visibility("default")));

#undef gtk_color_button_set_color
extern __typeof (gtk_color_button_set_color) gtk_color_button_set_color __attribute((alias("IA__gtk_color_button_set_color"), visibility("default")));

#undef gtk_color_button_set_title
extern __typeof (gtk_color_button_set_title) gtk_color_button_set_title __attribute((alias("IA__gtk_color_button_set_title"), visibility("default")));

#undef gtk_color_button_set_use_alpha
extern __typeof (gtk_color_button_set_use_alpha) gtk_color_button_set_use_alpha __attribute((alias("IA__gtk_color_button_set_use_alpha"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_COLOR_SELECTION_H__)
#if IN_FILE(__GTK_COLOR_SELECTION_C__)
#undef gtk_color_selection_get_current_alpha
extern __typeof (gtk_color_selection_get_current_alpha) gtk_color_selection_get_current_alpha __attribute((alias("IA__gtk_color_selection_get_current_alpha"), visibility("default")));

#undef gtk_color_selection_get_current_color
extern __typeof (gtk_color_selection_get_current_color) gtk_color_selection_get_current_color __attribute((alias("IA__gtk_color_selection_get_current_color"), visibility("default")));

#undef gtk_color_selection_get_has_opacity_control
extern __typeof (gtk_color_selection_get_has_opacity_control) gtk_color_selection_get_has_opacity_control __attribute((alias("IA__gtk_color_selection_get_has_opacity_control"), visibility("default")));

#undef gtk_color_selection_get_has_palette
extern __typeof (gtk_color_selection_get_has_palette) gtk_color_selection_get_has_palette __attribute((alias("IA__gtk_color_selection_get_has_palette"), visibility("default")));

#undef gtk_color_selection_get_previous_alpha
extern __typeof (gtk_color_selection_get_previous_alpha) gtk_color_selection_get_previous_alpha __attribute((alias("IA__gtk_color_selection_get_previous_alpha"), visibility("default")));

#undef gtk_color_selection_get_previous_color
extern __typeof (gtk_color_selection_get_previous_color) gtk_color_selection_get_previous_color __attribute((alias("IA__gtk_color_selection_get_previous_color"), visibility("default")));

#undef gtk_color_selection_get_type
extern __typeof (gtk_color_selection_get_type) gtk_color_selection_get_type __attribute((alias("IA__gtk_color_selection_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_color_selection_is_adjusting
extern __typeof (gtk_color_selection_is_adjusting) gtk_color_selection_is_adjusting __attribute((alias("IA__gtk_color_selection_is_adjusting"), visibility("default")));

#undef gtk_color_selection_new
extern __typeof (gtk_color_selection_new) gtk_color_selection_new __attribute((alias("IA__gtk_color_selection_new"), visibility("default")));

#undef gtk_color_selection_palette_from_string
extern __typeof (gtk_color_selection_palette_from_string) gtk_color_selection_palette_from_string __attribute((alias("IA__gtk_color_selection_palette_from_string"), visibility("default")));

#undef gtk_color_selection_palette_to_string
extern __typeof (gtk_color_selection_palette_to_string) gtk_color_selection_palette_to_string __attribute((alias("IA__gtk_color_selection_palette_to_string"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_color_selection_set_change_palette_hook
extern __typeof (gtk_color_selection_set_change_palette_hook) gtk_color_selection_set_change_palette_hook __attribute((alias("IA__gtk_color_selection_set_change_palette_hook"), visibility("default")));

#undef gtk_color_selection_set_color
extern __typeof (gtk_color_selection_set_color) gtk_color_selection_set_color __attribute((alias("IA__gtk_color_selection_set_color"), visibility("default")));

#undef gtk_color_selection_get_color
extern __typeof (gtk_color_selection_get_color) gtk_color_selection_get_color __attribute((alias("IA__gtk_color_selection_get_color"), visibility("default")));

#undef gtk_color_selection_set_update_policy
extern __typeof (gtk_color_selection_set_update_policy) gtk_color_selection_set_update_policy __attribute((alias("IA__gtk_color_selection_set_update_policy"), visibility("default")));

#endif
#undef gtk_color_selection_set_change_palette_with_screen_hook
extern __typeof (gtk_color_selection_set_change_palette_with_screen_hook) gtk_color_selection_set_change_palette_with_screen_hook __attribute((alias("IA__gtk_color_selection_set_change_palette_with_screen_hook"), visibility("default")));

#undef gtk_color_selection_set_current_alpha
extern __typeof (gtk_color_selection_set_current_alpha) gtk_color_selection_set_current_alpha __attribute((alias("IA__gtk_color_selection_set_current_alpha"), visibility("default")));

#undef gtk_color_selection_set_current_color
extern __typeof (gtk_color_selection_set_current_color) gtk_color_selection_set_current_color __attribute((alias("IA__gtk_color_selection_set_current_color"), visibility("default")));

#undef gtk_color_selection_set_has_opacity_control
extern __typeof (gtk_color_selection_set_has_opacity_control) gtk_color_selection_set_has_opacity_control __attribute((alias("IA__gtk_color_selection_set_has_opacity_control"), visibility("default")));

#undef gtk_color_selection_set_has_palette
extern __typeof (gtk_color_selection_set_has_palette) gtk_color_selection_set_has_palette __attribute((alias("IA__gtk_color_selection_set_has_palette"), visibility("default")));

#undef gtk_color_selection_set_previous_alpha
extern __typeof (gtk_color_selection_set_previous_alpha) gtk_color_selection_set_previous_alpha __attribute((alias("IA__gtk_color_selection_set_previous_alpha"), visibility("default")));

#undef gtk_color_selection_set_previous_color
extern __typeof (gtk_color_selection_set_previous_color) gtk_color_selection_set_previous_color __attribute((alias("IA__gtk_color_selection_set_previous_color"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_COLOR_SELECTION_DIALOG_H__)
#if IN_FILE(__GTK_COLOR_SELECTION_DIALOG_C__)
#undef gtk_color_selection_dialog_get_type
extern __typeof (gtk_color_selection_dialog_get_type) gtk_color_selection_dialog_get_type __attribute((alias("IA__gtk_color_selection_dialog_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_color_selection_dialog_new
extern __typeof (gtk_color_selection_dialog_new) gtk_color_selection_dialog_new __attribute((alias("IA__gtk_color_selection_dialog_new"), visibility("default")));

#undef gtk_color_selection_dialog_get_color_selection
extern __typeof (gtk_color_selection_dialog_get_color_selection) gtk_color_selection_dialog_get_color_selection __attribute((alias("IA__gtk_color_selection_dialog_get_color_selection"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_COMBO_BOX_H__)
#if IN_FILE(__GTK_COMBO_BOX_C__)
#undef gtk_combo_box_append_text
extern __typeof (gtk_combo_box_append_text) gtk_combo_box_append_text __attribute((alias("IA__gtk_combo_box_append_text"), visibility("default")));

#undef gtk_combo_box_get_active
extern __typeof (gtk_combo_box_get_active) gtk_combo_box_get_active __attribute((alias("IA__gtk_combo_box_get_active"), visibility("default")));

#undef gtk_combo_box_get_active_iter
extern __typeof (gtk_combo_box_get_active_iter) gtk_combo_box_get_active_iter __attribute((alias("IA__gtk_combo_box_get_active_iter"), visibility("default")));

#undef gtk_combo_box_get_active_text
extern __typeof (gtk_combo_box_get_active_text) gtk_combo_box_get_active_text __attribute((alias("IA__gtk_combo_box_get_active_text"), visibility("default")));

#undef gtk_combo_box_get_add_tearoffs
extern __typeof (gtk_combo_box_get_add_tearoffs) gtk_combo_box_get_add_tearoffs __attribute((alias("IA__gtk_combo_box_get_add_tearoffs"), visibility("default")));

#undef gtk_combo_box_get_column_span_column
extern __typeof (gtk_combo_box_get_column_span_column) gtk_combo_box_get_column_span_column __attribute((alias("IA__gtk_combo_box_get_column_span_column"), visibility("default")));

#undef gtk_combo_box_get_entry_text_column
extern __typeof (gtk_combo_box_get_entry_text_column) gtk_combo_box_get_entry_text_column __attribute((alias("IA__gtk_combo_box_get_entry_text_column"), visibility("default")));

#undef gtk_combo_box_get_focus_on_click
extern __typeof (gtk_combo_box_get_focus_on_click) gtk_combo_box_get_focus_on_click __attribute((alias("IA__gtk_combo_box_get_focus_on_click"), visibility("default")));

#undef gtk_combo_box_get_has_entry
extern __typeof (gtk_combo_box_get_has_entry) gtk_combo_box_get_has_entry __attribute((alias("IA__gtk_combo_box_get_has_entry"), visibility("default")));

#undef gtk_combo_box_get_model
extern __typeof (gtk_combo_box_get_model) gtk_combo_box_get_model __attribute((alias("IA__gtk_combo_box_get_model"), visibility("default")));

#undef gtk_combo_box_get_popup_accessible
extern __typeof (gtk_combo_box_get_popup_accessible) gtk_combo_box_get_popup_accessible __attribute((alias("IA__gtk_combo_box_get_popup_accessible"), visibility("default")));

#undef gtk_combo_box_get_row_separator_func
extern __typeof (gtk_combo_box_get_row_separator_func) gtk_combo_box_get_row_separator_func __attribute((alias("IA__gtk_combo_box_get_row_separator_func"), visibility("default")));

#undef gtk_combo_box_get_row_span_column
extern __typeof (gtk_combo_box_get_row_span_column) gtk_combo_box_get_row_span_column __attribute((alias("IA__gtk_combo_box_get_row_span_column"), visibility("default")));

#undef gtk_combo_box_get_button_sensitivity
extern __typeof (gtk_combo_box_get_button_sensitivity) gtk_combo_box_get_button_sensitivity __attribute((alias("IA__gtk_combo_box_get_button_sensitivity"), visibility("default")));

#undef gtk_combo_box_get_title
extern __typeof (gtk_combo_box_get_title) gtk_combo_box_get_title __attribute((alias("IA__gtk_combo_box_get_title"), visibility("default")));

#undef gtk_combo_box_get_type
extern __typeof (gtk_combo_box_get_type) gtk_combo_box_get_type __attribute((alias("IA__gtk_combo_box_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_combo_box_get_wrap_width
extern __typeof (gtk_combo_box_get_wrap_width) gtk_combo_box_get_wrap_width __attribute((alias("IA__gtk_combo_box_get_wrap_width"), visibility("default")));

#undef gtk_combo_box_insert_text
extern __typeof (gtk_combo_box_insert_text) gtk_combo_box_insert_text __attribute((alias("IA__gtk_combo_box_insert_text"), visibility("default")));

#undef gtk_combo_box_new
extern __typeof (gtk_combo_box_new) gtk_combo_box_new __attribute((alias("IA__gtk_combo_box_new"), visibility("default")));

#undef gtk_combo_box_new_text
extern __typeof (gtk_combo_box_new_text) gtk_combo_box_new_text __attribute((alias("IA__gtk_combo_box_new_text"), visibility("default")));

#undef gtk_combo_box_new_with_entry
extern __typeof (gtk_combo_box_new_with_entry) gtk_combo_box_new_with_entry __attribute((alias("IA__gtk_combo_box_new_with_entry"), visibility("default")));

#undef gtk_combo_box_new_with_model
extern __typeof (gtk_combo_box_new_with_model) gtk_combo_box_new_with_model __attribute((alias("IA__gtk_combo_box_new_with_model"), visibility("default")));

#undef gtk_combo_box_new_with_model_and_entry
extern __typeof (gtk_combo_box_new_with_model_and_entry) gtk_combo_box_new_with_model_and_entry __attribute((alias("IA__gtk_combo_box_new_with_model_and_entry"), visibility("default")));

#undef gtk_combo_box_popdown
extern __typeof (gtk_combo_box_popdown) gtk_combo_box_popdown __attribute((alias("IA__gtk_combo_box_popdown"), visibility("default")));

#undef gtk_combo_box_popup
extern __typeof (gtk_combo_box_popup) gtk_combo_box_popup __attribute((alias("IA__gtk_combo_box_popup"), visibility("default")));

#undef gtk_combo_box_prepend_text
extern __typeof (gtk_combo_box_prepend_text) gtk_combo_box_prepend_text __attribute((alias("IA__gtk_combo_box_prepend_text"), visibility("default")));

#undef gtk_combo_box_remove_text
extern __typeof (gtk_combo_box_remove_text) gtk_combo_box_remove_text __attribute((alias("IA__gtk_combo_box_remove_text"), visibility("default")));

#undef gtk_combo_box_set_active
extern __typeof (gtk_combo_box_set_active) gtk_combo_box_set_active __attribute((alias("IA__gtk_combo_box_set_active"), visibility("default")));

#undef gtk_combo_box_set_active_iter
extern __typeof (gtk_combo_box_set_active_iter) gtk_combo_box_set_active_iter __attribute((alias("IA__gtk_combo_box_set_active_iter"), visibility("default")));

#undef gtk_combo_box_set_add_tearoffs
extern __typeof (gtk_combo_box_set_add_tearoffs) gtk_combo_box_set_add_tearoffs __attribute((alias("IA__gtk_combo_box_set_add_tearoffs"), visibility("default")));

#undef gtk_combo_box_set_column_span_column
extern __typeof (gtk_combo_box_set_column_span_column) gtk_combo_box_set_column_span_column __attribute((alias("IA__gtk_combo_box_set_column_span_column"), visibility("default")));

#undef gtk_combo_box_set_entry_text_column
extern __typeof (gtk_combo_box_set_entry_text_column) gtk_combo_box_set_entry_text_column __attribute((alias("IA__gtk_combo_box_set_entry_text_column"), visibility("default")));

#undef gtk_combo_box_set_focus_on_click
extern __typeof (gtk_combo_box_set_focus_on_click) gtk_combo_box_set_focus_on_click __attribute((alias("IA__gtk_combo_box_set_focus_on_click"), visibility("default")));

#undef gtk_combo_box_set_model
extern __typeof (gtk_combo_box_set_model) gtk_combo_box_set_model __attribute((alias("IA__gtk_combo_box_set_model"), visibility("default")));

#undef gtk_combo_box_set_row_separator_func
extern __typeof (gtk_combo_box_set_row_separator_func) gtk_combo_box_set_row_separator_func __attribute((alias("IA__gtk_combo_box_set_row_separator_func"), visibility("default")));

#undef gtk_combo_box_set_row_span_column
extern __typeof (gtk_combo_box_set_row_span_column) gtk_combo_box_set_row_span_column __attribute((alias("IA__gtk_combo_box_set_row_span_column"), visibility("default")));

#undef gtk_combo_box_set_button_sensitivity
extern __typeof (gtk_combo_box_set_button_sensitivity) gtk_combo_box_set_button_sensitivity __attribute((alias("IA__gtk_combo_box_set_button_sensitivity"), visibility("default")));

#undef gtk_combo_box_set_title
extern __typeof (gtk_combo_box_set_title) gtk_combo_box_set_title __attribute((alias("IA__gtk_combo_box_set_title"), visibility("default")));

#undef gtk_combo_box_set_wrap_width
extern __typeof (gtk_combo_box_set_wrap_width) gtk_combo_box_set_wrap_width __attribute((alias("IA__gtk_combo_box_set_wrap_width"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_COMBO_BOX_TEXT_H__)
#if IN_FILE(__GTK_COMBO_BOX_TEXT_C__)
#undef gtk_combo_box_text_append_text
extern __typeof (gtk_combo_box_text_append_text) gtk_combo_box_text_append_text __attribute((alias("IA__gtk_combo_box_text_append_text"), visibility("default")));

#undef gtk_combo_box_text_get_active_text
extern __typeof (gtk_combo_box_text_get_active_text) gtk_combo_box_text_get_active_text __attribute((alias("IA__gtk_combo_box_text_get_active_text"), visibility("default")));

#undef gtk_combo_box_text_get_type
extern __typeof (gtk_combo_box_text_get_type) gtk_combo_box_text_get_type __attribute((alias("IA__gtk_combo_box_text_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_combo_box_text_insert_text
extern __typeof (gtk_combo_box_text_insert_text) gtk_combo_box_text_insert_text __attribute((alias("IA__gtk_combo_box_text_insert_text"), visibility("default")));

#undef gtk_combo_box_text_new
extern __typeof (gtk_combo_box_text_new) gtk_combo_box_text_new __attribute((alias("IA__gtk_combo_box_text_new"), visibility("default")));

#undef gtk_combo_box_text_new_with_entry
extern __typeof (gtk_combo_box_text_new_with_entry) gtk_combo_box_text_new_with_entry __attribute((alias("IA__gtk_combo_box_text_new_with_entry"), visibility("default")));

#undef gtk_combo_box_text_prepend_text
extern __typeof (gtk_combo_box_text_prepend_text) gtk_combo_box_text_prepend_text __attribute((alias("IA__gtk_combo_box_text_prepend_text"), visibility("default")));

#undef gtk_combo_box_text_remove
extern __typeof (gtk_combo_box_text_remove) gtk_combo_box_text_remove __attribute((alias("IA__gtk_combo_box_text_remove"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_COMBO_BOX_ENTRY_H__)
#if IN_FILE(__GTK_COMBO_BOX_ENTRY_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_combo_box_entry_get_text_column
extern __typeof (gtk_combo_box_entry_get_text_column) gtk_combo_box_entry_get_text_column __attribute((alias("IA__gtk_combo_box_entry_get_text_column"), visibility("default")));

#undef gtk_combo_box_entry_get_type
extern __typeof (gtk_combo_box_entry_get_type) gtk_combo_box_entry_get_type __attribute((alias("IA__gtk_combo_box_entry_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_combo_box_entry_new
extern __typeof (gtk_combo_box_entry_new) gtk_combo_box_entry_new __attribute((alias("IA__gtk_combo_box_entry_new"), visibility("default")));

#undef gtk_combo_box_entry_new_text
extern __typeof (gtk_combo_box_entry_new_text) gtk_combo_box_entry_new_text __attribute((alias("IA__gtk_combo_box_entry_new_text"), visibility("default")));

#undef gtk_combo_box_entry_new_with_model
extern __typeof (gtk_combo_box_entry_new_with_model) gtk_combo_box_entry_new_with_model __attribute((alias("IA__gtk_combo_box_entry_new_with_model"), visibility("default")));

#undef gtk_combo_box_entry_set_text_column
extern __typeof (gtk_combo_box_entry_set_text_column) gtk_combo_box_entry_set_text_column __attribute((alias("IA__gtk_combo_box_entry_set_text_column"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_SMART_COMBO_H__)
#if IN_FILE(__GTK_SMART_COMBO_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_combo_disable_activate
extern __typeof (gtk_combo_disable_activate) gtk_combo_disable_activate __attribute((alias("IA__gtk_combo_disable_activate"), visibility("default")));

#undef gtk_combo_get_type
extern __typeof (gtk_combo_get_type) gtk_combo_get_type __attribute((alias("IA__gtk_combo_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_combo_new
extern __typeof (gtk_combo_new) gtk_combo_new __attribute((alias("IA__gtk_combo_new"), visibility("default")));

#undef gtk_combo_set_case_sensitive
extern __typeof (gtk_combo_set_case_sensitive) gtk_combo_set_case_sensitive __attribute((alias("IA__gtk_combo_set_case_sensitive"), visibility("default")));

#undef gtk_combo_set_item_string
extern __typeof (gtk_combo_set_item_string) gtk_combo_set_item_string __attribute((alias("IA__gtk_combo_set_item_string"), visibility("default")));

#undef gtk_combo_set_popdown_strings
extern __typeof (gtk_combo_set_popdown_strings) gtk_combo_set_popdown_strings __attribute((alias("IA__gtk_combo_set_popdown_strings"), visibility("default")));

#undef gtk_combo_set_use_arrows
extern __typeof (gtk_combo_set_use_arrows) gtk_combo_set_use_arrows __attribute((alias("IA__gtk_combo_set_use_arrows"), visibility("default")));

#undef gtk_combo_set_use_arrows_always
extern __typeof (gtk_combo_set_use_arrows_always) gtk_combo_set_use_arrows_always __attribute((alias("IA__gtk_combo_set_use_arrows_always"), visibility("default")));

#undef gtk_combo_set_value_in_list
extern __typeof (gtk_combo_set_value_in_list) gtk_combo_set_value_in_list __attribute((alias("IA__gtk_combo_set_value_in_list"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_CONTAINER_H__)
#if IN_FILE(__GTK_CONTAINER_C__)
#undef gtk_container_add
extern __typeof (gtk_container_add) gtk_container_add __attribute((alias("IA__gtk_container_add"), visibility("default")));

#undef gtk_container_add_with_properties
extern __typeof (gtk_container_add_with_properties) gtk_container_add_with_properties __attribute((alias("IA__gtk_container_add_with_properties"), visibility("default"))) G_GNUC_NULL_TERMINATED;

#undef gtk_container_check_resize
extern __typeof (gtk_container_check_resize) gtk_container_check_resize __attribute((alias("IA__gtk_container_check_resize"), visibility("default")));

#undef gtk_container_child_get
extern __typeof (gtk_container_child_get) gtk_container_child_get __attribute((alias("IA__gtk_container_child_get"), visibility("default"))) G_GNUC_NULL_TERMINATED;

#undef gtk_container_child_get_property
extern __typeof (gtk_container_child_get_property) gtk_container_child_get_property __attribute((alias("IA__gtk_container_child_get_property"), visibility("default")));

#undef gtk_container_child_get_valist
extern __typeof (gtk_container_child_get_valist) gtk_container_child_get_valist __attribute((alias("IA__gtk_container_child_get_valist"), visibility("default")));

#undef gtk_container_child_set
extern __typeof (gtk_container_child_set) gtk_container_child_set __attribute((alias("IA__gtk_container_child_set"), visibility("default"))) G_GNUC_NULL_TERMINATED;

#undef gtk_container_child_set_property
extern __typeof (gtk_container_child_set_property) gtk_container_child_set_property __attribute((alias("IA__gtk_container_child_set_property"), visibility("default")));

#undef gtk_container_child_set_valist
extern __typeof (gtk_container_child_set_valist) gtk_container_child_set_valist __attribute((alias("IA__gtk_container_child_set_valist"), visibility("default")));

#undef gtk_container_child_type
extern __typeof (gtk_container_child_type) gtk_container_child_type __attribute((alias("IA__gtk_container_child_type"), visibility("default")));

#undef gtk_container_class_find_child_property
extern __typeof (gtk_container_class_find_child_property) gtk_container_class_find_child_property __attribute((alias("IA__gtk_container_class_find_child_property"), visibility("default")));

#undef gtk_container_class_install_child_property
extern __typeof (gtk_container_class_install_child_property) gtk_container_class_install_child_property __attribute((alias("IA__gtk_container_class_install_child_property"), visibility("default")));

#undef gtk_container_class_list_child_properties
extern __typeof (gtk_container_class_list_child_properties) gtk_container_class_list_child_properties __attribute((alias("IA__gtk_container_class_list_child_properties"), visibility("default")));

#undef gtk_container_forall
extern __typeof (gtk_container_forall) gtk_container_forall __attribute((alias("IA__gtk_container_forall"), visibility("default")));

#undef gtk_container_foreach
extern __typeof (gtk_container_foreach) gtk_container_foreach __attribute((alias("IA__gtk_container_foreach"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_container_foreach_full
extern __typeof (gtk_container_foreach_full) gtk_container_foreach_full __attribute((alias("IA__gtk_container_foreach_full"), visibility("default")));

#endif
#undef gtk_container_get_border_width
extern __typeof (gtk_container_get_border_width) gtk_container_get_border_width __attribute((alias("IA__gtk_container_get_border_width"), visibility("default")));

#undef gtk_container_get_children
extern __typeof (gtk_container_get_children) gtk_container_get_children __attribute((alias("IA__gtk_container_get_children"), visibility("default")));

#undef gtk_container_get_focus_chain
extern __typeof (gtk_container_get_focus_chain) gtk_container_get_focus_chain __attribute((alias("IA__gtk_container_get_focus_chain"), visibility("default")));

#undef gtk_container_get_focus_child
extern __typeof (gtk_container_get_focus_child) gtk_container_get_focus_child __attribute((alias("IA__gtk_container_get_focus_child"), visibility("default")));

#undef gtk_container_get_focus_hadjustment
extern __typeof (gtk_container_get_focus_hadjustment) gtk_container_get_focus_hadjustment __attribute((alias("IA__gtk_container_get_focus_hadjustment"), visibility("default")));

#undef gtk_container_get_focus_vadjustment
extern __typeof (gtk_container_get_focus_vadjustment) gtk_container_get_focus_vadjustment __attribute((alias("IA__gtk_container_get_focus_vadjustment"), visibility("default")));

#undef gtk_container_get_resize_mode
extern __typeof (gtk_container_get_resize_mode) gtk_container_get_resize_mode __attribute((alias("IA__gtk_container_get_resize_mode"), visibility("default")));

#undef gtk_container_get_type
extern __typeof (gtk_container_get_type) gtk_container_get_type __attribute((alias("IA__gtk_container_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_container_propagate_expose
extern __typeof (gtk_container_propagate_expose) gtk_container_propagate_expose __attribute((alias("IA__gtk_container_propagate_expose"), visibility("default")));

#undef gtk_container_remove
extern __typeof (gtk_container_remove) gtk_container_remove __attribute((alias("IA__gtk_container_remove"), visibility("default")));

#undef gtk_container_resize_children
extern __typeof (gtk_container_resize_children) gtk_container_resize_children __attribute((alias("IA__gtk_container_resize_children"), visibility("default")));

#undef gtk_container_set_border_width
extern __typeof (gtk_container_set_border_width) gtk_container_set_border_width __attribute((alias("IA__gtk_container_set_border_width"), visibility("default")));

#undef gtk_container_set_focus_chain
extern __typeof (gtk_container_set_focus_chain) gtk_container_set_focus_chain __attribute((alias("IA__gtk_container_set_focus_chain"), visibility("default")));

#undef gtk_container_set_focus_child
extern __typeof (gtk_container_set_focus_child) gtk_container_set_focus_child __attribute((alias("IA__gtk_container_set_focus_child"), visibility("default")));

#undef gtk_container_set_focus_hadjustment
extern __typeof (gtk_container_set_focus_hadjustment) gtk_container_set_focus_hadjustment __attribute((alias("IA__gtk_container_set_focus_hadjustment"), visibility("default")));

#undef gtk_container_set_focus_vadjustment
extern __typeof (gtk_container_set_focus_vadjustment) gtk_container_set_focus_vadjustment __attribute((alias("IA__gtk_container_set_focus_vadjustment"), visibility("default")));

#undef gtk_container_set_reallocate_redraws
extern __typeof (gtk_container_set_reallocate_redraws) gtk_container_set_reallocate_redraws __attribute((alias("IA__gtk_container_set_reallocate_redraws"), visibility("default")));

#undef gtk_container_set_resize_mode
extern __typeof (gtk_container_set_resize_mode) gtk_container_set_resize_mode __attribute((alias("IA__gtk_container_set_resize_mode"), visibility("default")));

#undef gtk_container_unset_focus_chain
extern __typeof (gtk_container_unset_focus_chain) gtk_container_unset_focus_chain __attribute((alias("IA__gtk_container_unset_focus_chain"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CTREE_H__)
#if IN_FILE(__GTK_CTREE_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_ctree_collapse
extern __typeof (gtk_ctree_collapse) gtk_ctree_collapse __attribute((alias("IA__gtk_ctree_collapse"), visibility("default")));

#undef gtk_ctree_collapse_recursive
extern __typeof (gtk_ctree_collapse_recursive) gtk_ctree_collapse_recursive __attribute((alias("IA__gtk_ctree_collapse_recursive"), visibility("default")));

#undef gtk_ctree_collapse_to_depth
extern __typeof (gtk_ctree_collapse_to_depth) gtk_ctree_collapse_to_depth __attribute((alias("IA__gtk_ctree_collapse_to_depth"), visibility("default")));

#undef gtk_ctree_expand
extern __typeof (gtk_ctree_expand) gtk_ctree_expand __attribute((alias("IA__gtk_ctree_expand"), visibility("default")));

#undef gtk_ctree_expand_recursive
extern __typeof (gtk_ctree_expand_recursive) gtk_ctree_expand_recursive __attribute((alias("IA__gtk_ctree_expand_recursive"), visibility("default")));

#undef gtk_ctree_expand_to_depth
extern __typeof (gtk_ctree_expand_to_depth) gtk_ctree_expand_to_depth __attribute((alias("IA__gtk_ctree_expand_to_depth"), visibility("default")));

#undef gtk_ctree_export_to_gnode
extern __typeof (gtk_ctree_export_to_gnode) gtk_ctree_export_to_gnode __attribute((alias("IA__gtk_ctree_export_to_gnode"), visibility("default")));

#undef gtk_ctree_find
extern __typeof (gtk_ctree_find) gtk_ctree_find __attribute((alias("IA__gtk_ctree_find"), visibility("default")));

#undef gtk_ctree_find_all_by_row_data
extern __typeof (gtk_ctree_find_all_by_row_data) gtk_ctree_find_all_by_row_data __attribute((alias("IA__gtk_ctree_find_all_by_row_data"), visibility("default")));

#undef gtk_ctree_find_all_by_row_data_custom
extern __typeof (gtk_ctree_find_all_by_row_data_custom) gtk_ctree_find_all_by_row_data_custom __attribute((alias("IA__gtk_ctree_find_all_by_row_data_custom"), visibility("default")));

#undef gtk_ctree_find_by_row_data
extern __typeof (gtk_ctree_find_by_row_data) gtk_ctree_find_by_row_data __attribute((alias("IA__gtk_ctree_find_by_row_data"), visibility("default")));

#undef gtk_ctree_find_by_row_data_custom
extern __typeof (gtk_ctree_find_by_row_data_custom) gtk_ctree_find_by_row_data_custom __attribute((alias("IA__gtk_ctree_find_by_row_data_custom"), visibility("default")));

#undef gtk_ctree_find_node_ptr
extern __typeof (gtk_ctree_find_node_ptr) gtk_ctree_find_node_ptr __attribute((alias("IA__gtk_ctree_find_node_ptr"), visibility("default")));

#undef gtk_ctree_get_node_info
extern __typeof (gtk_ctree_get_node_info) gtk_ctree_get_node_info __attribute((alias("IA__gtk_ctree_get_node_info"), visibility("default")));

#undef gtk_ctree_get_type
extern __typeof (gtk_ctree_get_type) gtk_ctree_get_type __attribute((alias("IA__gtk_ctree_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_ctree_insert_gnode
extern __typeof (gtk_ctree_insert_gnode) gtk_ctree_insert_gnode __attribute((alias("IA__gtk_ctree_insert_gnode"), visibility("default")));

#undef gtk_ctree_insert_node
extern __typeof (gtk_ctree_insert_node) gtk_ctree_insert_node __attribute((alias("IA__gtk_ctree_insert_node"), visibility("default")));

#undef gtk_ctree_is_ancestor
extern __typeof (gtk_ctree_is_ancestor) gtk_ctree_is_ancestor __attribute((alias("IA__gtk_ctree_is_ancestor"), visibility("default")));

#undef gtk_ctree_is_hot_spot
extern __typeof (gtk_ctree_is_hot_spot) gtk_ctree_is_hot_spot __attribute((alias("IA__gtk_ctree_is_hot_spot"), visibility("default")));

#undef gtk_ctree_is_viewable
extern __typeof (gtk_ctree_is_viewable) gtk_ctree_is_viewable __attribute((alias("IA__gtk_ctree_is_viewable"), visibility("default")));

#undef gtk_ctree_last
extern __typeof (gtk_ctree_last) gtk_ctree_last __attribute((alias("IA__gtk_ctree_last"), visibility("default")));

#undef gtk_ctree_move
extern __typeof (gtk_ctree_move) gtk_ctree_move __attribute((alias("IA__gtk_ctree_move"), visibility("default")));

#undef gtk_ctree_new
extern __typeof (gtk_ctree_new) gtk_ctree_new __attribute((alias("IA__gtk_ctree_new"), visibility("default")));

#undef gtk_ctree_new_with_titles
extern __typeof (gtk_ctree_new_with_titles) gtk_ctree_new_with_titles __attribute((alias("IA__gtk_ctree_new_with_titles"), visibility("default")));

#undef gtk_ctree_node_get_cell_style
extern __typeof (gtk_ctree_node_get_cell_style) gtk_ctree_node_get_cell_style __attribute((alias("IA__gtk_ctree_node_get_cell_style"), visibility("default")));

#undef gtk_ctree_node_get_cell_type
extern __typeof (gtk_ctree_node_get_cell_type) gtk_ctree_node_get_cell_type __attribute((alias("IA__gtk_ctree_node_get_cell_type"), visibility("default")));

#undef gtk_ctree_node_get_pixmap
extern __typeof (gtk_ctree_node_get_pixmap) gtk_ctree_node_get_pixmap __attribute((alias("IA__gtk_ctree_node_get_pixmap"), visibility("default")));

#undef gtk_ctree_node_get_pixtext
extern __typeof (gtk_ctree_node_get_pixtext) gtk_ctree_node_get_pixtext __attribute((alias("IA__gtk_ctree_node_get_pixtext"), visibility("default")));

#undef gtk_ctree_node_get_row_data
extern __typeof (gtk_ctree_node_get_row_data) gtk_ctree_node_get_row_data __attribute((alias("IA__gtk_ctree_node_get_row_data"), visibility("default")));

#undef gtk_ctree_node_get_row_style
extern __typeof (gtk_ctree_node_get_row_style) gtk_ctree_node_get_row_style __attribute((alias("IA__gtk_ctree_node_get_row_style"), visibility("default")));

#undef gtk_ctree_node_get_selectable
extern __typeof (gtk_ctree_node_get_selectable) gtk_ctree_node_get_selectable __attribute((alias("IA__gtk_ctree_node_get_selectable"), visibility("default")));

#undef gtk_ctree_node_get_text
extern __typeof (gtk_ctree_node_get_text) gtk_ctree_node_get_text __attribute((alias("IA__gtk_ctree_node_get_text"), visibility("default")));

#undef gtk_ctree_node_get_type
extern __typeof (gtk_ctree_node_get_type) gtk_ctree_node_get_type __attribute((alias("IA__gtk_ctree_node_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_ctree_node_is_visible
extern __typeof (gtk_ctree_node_is_visible) gtk_ctree_node_is_visible __attribute((alias("IA__gtk_ctree_node_is_visible"), visibility("default")));

#undef gtk_ctree_node_moveto
extern __typeof (gtk_ctree_node_moveto) gtk_ctree_node_moveto __attribute((alias("IA__gtk_ctree_node_moveto"), visibility("default")));

#undef gtk_ctree_node_nth
extern __typeof (gtk_ctree_node_nth) gtk_ctree_node_nth __attribute((alias("IA__gtk_ctree_node_nth"), visibility("default")));

#undef gtk_ctree_node_set_background
extern __typeof (gtk_ctree_node_set_background) gtk_ctree_node_set_background __attribute((alias("IA__gtk_ctree_node_set_background"), visibility("default")));

#undef gtk_ctree_node_set_cell_style
extern __typeof (gtk_ctree_node_set_cell_style) gtk_ctree_node_set_cell_style __attribute((alias("IA__gtk_ctree_node_set_cell_style"), visibility("default")));

#undef gtk_ctree_node_set_foreground
extern __typeof (gtk_ctree_node_set_foreground) gtk_ctree_node_set_foreground __attribute((alias("IA__gtk_ctree_node_set_foreground"), visibility("default")));

#undef gtk_ctree_node_set_pixmap
extern __typeof (gtk_ctree_node_set_pixmap) gtk_ctree_node_set_pixmap __attribute((alias("IA__gtk_ctree_node_set_pixmap"), visibility("default")));

#undef gtk_ctree_node_set_pixtext
extern __typeof (gtk_ctree_node_set_pixtext) gtk_ctree_node_set_pixtext __attribute((alias("IA__gtk_ctree_node_set_pixtext"), visibility("default")));

#undef gtk_ctree_node_set_row_data
extern __typeof (gtk_ctree_node_set_row_data) gtk_ctree_node_set_row_data __attribute((alias("IA__gtk_ctree_node_set_row_data"), visibility("default")));

#undef gtk_ctree_node_set_row_data_full
extern __typeof (gtk_ctree_node_set_row_data_full) gtk_ctree_node_set_row_data_full __attribute((alias("IA__gtk_ctree_node_set_row_data_full"), visibility("default")));

#undef gtk_ctree_node_set_row_style
extern __typeof (gtk_ctree_node_set_row_style) gtk_ctree_node_set_row_style __attribute((alias("IA__gtk_ctree_node_set_row_style"), visibility("default")));

#undef gtk_ctree_node_set_selectable
extern __typeof (gtk_ctree_node_set_selectable) gtk_ctree_node_set_selectable __attribute((alias("IA__gtk_ctree_node_set_selectable"), visibility("default")));

#undef gtk_ctree_node_set_shift
extern __typeof (gtk_ctree_node_set_shift) gtk_ctree_node_set_shift __attribute((alias("IA__gtk_ctree_node_set_shift"), visibility("default")));

#undef gtk_ctree_node_set_text
extern __typeof (gtk_ctree_node_set_text) gtk_ctree_node_set_text __attribute((alias("IA__gtk_ctree_node_set_text"), visibility("default")));

#undef gtk_ctree_post_recursive
extern __typeof (gtk_ctree_post_recursive) gtk_ctree_post_recursive __attribute((alias("IA__gtk_ctree_post_recursive"), visibility("default")));

#undef gtk_ctree_post_recursive_to_depth
extern __typeof (gtk_ctree_post_recursive_to_depth) gtk_ctree_post_recursive_to_depth __attribute((alias("IA__gtk_ctree_post_recursive_to_depth"), visibility("default")));

#undef gtk_ctree_pre_recursive
extern __typeof (gtk_ctree_pre_recursive) gtk_ctree_pre_recursive __attribute((alias("IA__gtk_ctree_pre_recursive"), visibility("default")));

#undef gtk_ctree_pre_recursive_to_depth
extern __typeof (gtk_ctree_pre_recursive_to_depth) gtk_ctree_pre_recursive_to_depth __attribute((alias("IA__gtk_ctree_pre_recursive_to_depth"), visibility("default")));

#undef gtk_ctree_real_select_recursive
extern __typeof (gtk_ctree_real_select_recursive) gtk_ctree_real_select_recursive __attribute((alias("IA__gtk_ctree_real_select_recursive"), visibility("default")));

#undef gtk_ctree_remove_node
extern __typeof (gtk_ctree_remove_node) gtk_ctree_remove_node __attribute((alias("IA__gtk_ctree_remove_node"), visibility("default")));

#undef gtk_ctree_select
extern __typeof (gtk_ctree_select) gtk_ctree_select __attribute((alias("IA__gtk_ctree_select"), visibility("default")));

#undef gtk_ctree_select_recursive
extern __typeof (gtk_ctree_select_recursive) gtk_ctree_select_recursive __attribute((alias("IA__gtk_ctree_select_recursive"), visibility("default")));

#undef gtk_ctree_set_drag_compare_func
extern __typeof (gtk_ctree_set_drag_compare_func) gtk_ctree_set_drag_compare_func __attribute((alias("IA__gtk_ctree_set_drag_compare_func"), visibility("default")));

#undef gtk_ctree_set_expander_style
extern __typeof (gtk_ctree_set_expander_style) gtk_ctree_set_expander_style __attribute((alias("IA__gtk_ctree_set_expander_style"), visibility("default")));

#undef gtk_ctree_set_indent
extern __typeof (gtk_ctree_set_indent) gtk_ctree_set_indent __attribute((alias("IA__gtk_ctree_set_indent"), visibility("default")));

#undef gtk_ctree_set_line_style
extern __typeof (gtk_ctree_set_line_style) gtk_ctree_set_line_style __attribute((alias("IA__gtk_ctree_set_line_style"), visibility("default")));

#undef gtk_ctree_set_node_info
extern __typeof (gtk_ctree_set_node_info) gtk_ctree_set_node_info __attribute((alias("IA__gtk_ctree_set_node_info"), visibility("default")));

#undef gtk_ctree_set_show_stub
extern __typeof (gtk_ctree_set_show_stub) gtk_ctree_set_show_stub __attribute((alias("IA__gtk_ctree_set_show_stub"), visibility("default")));

#undef gtk_ctree_set_spacing
extern __typeof (gtk_ctree_set_spacing) gtk_ctree_set_spacing __attribute((alias("IA__gtk_ctree_set_spacing"), visibility("default")));

#undef gtk_ctree_sort_node
extern __typeof (gtk_ctree_sort_node) gtk_ctree_sort_node __attribute((alias("IA__gtk_ctree_sort_node"), visibility("default")));

#undef gtk_ctree_sort_recursive
extern __typeof (gtk_ctree_sort_recursive) gtk_ctree_sort_recursive __attribute((alias("IA__gtk_ctree_sort_recursive"), visibility("default")));

#undef gtk_ctree_toggle_expansion
extern __typeof (gtk_ctree_toggle_expansion) gtk_ctree_toggle_expansion __attribute((alias("IA__gtk_ctree_toggle_expansion"), visibility("default")));

#undef gtk_ctree_toggle_expansion_recursive
extern __typeof (gtk_ctree_toggle_expansion_recursive) gtk_ctree_toggle_expansion_recursive __attribute((alias("IA__gtk_ctree_toggle_expansion_recursive"), visibility("default")));

#undef gtk_ctree_unselect
extern __typeof (gtk_ctree_unselect) gtk_ctree_unselect __attribute((alias("IA__gtk_ctree_unselect"), visibility("default")));

#undef gtk_ctree_unselect_recursive
extern __typeof (gtk_ctree_unselect_recursive) gtk_ctree_unselect_recursive __attribute((alias("IA__gtk_ctree_unselect_recursive"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_CURVE_H__)
#if IN_FILE(__GTK_CURVE_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_curve_get_type
extern __typeof (gtk_curve_get_type) gtk_curve_get_type __attribute((alias("IA__gtk_curve_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_curve_get_vector
extern __typeof (gtk_curve_get_vector) gtk_curve_get_vector __attribute((alias("IA__gtk_curve_get_vector"), visibility("default")));

#undef gtk_curve_new
extern __typeof (gtk_curve_new) gtk_curve_new __attribute((alias("IA__gtk_curve_new"), visibility("default")));

#undef gtk_curve_reset
extern __typeof (gtk_curve_reset) gtk_curve_reset __attribute((alias("IA__gtk_curve_reset"), visibility("default")));

#undef gtk_curve_set_curve_type
extern __typeof (gtk_curve_set_curve_type) gtk_curve_set_curve_type __attribute((alias("IA__gtk_curve_set_curve_type"), visibility("default")));

#undef gtk_curve_set_gamma
extern __typeof (gtk_curve_set_gamma) gtk_curve_set_gamma __attribute((alias("IA__gtk_curve_set_gamma"), visibility("default")));

#undef gtk_curve_set_range
extern __typeof (gtk_curve_set_range) gtk_curve_set_range __attribute((alias("IA__gtk_curve_set_range"), visibility("default")));

#undef gtk_curve_set_vector
extern __typeof (gtk_curve_set_vector) gtk_curve_set_vector __attribute((alias("IA__gtk_curve_set_vector"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_WINDOW_DECORATE_H__)
#if IN_FILE(__GTK_WINDOW_DECORATE_C__)
#undef gtk_decorated_window_calculate_frame_size
extern __typeof (gtk_decorated_window_calculate_frame_size) gtk_decorated_window_calculate_frame_size __attribute((alias("IA__gtk_decorated_window_calculate_frame_size"), visibility("default")));

#undef gtk_decorated_window_init
extern __typeof (gtk_decorated_window_init) gtk_decorated_window_init __attribute((alias("IA__gtk_decorated_window_init"), visibility("default")));

#undef gtk_decorated_window_move_resize_window
extern __typeof (gtk_decorated_window_move_resize_window) gtk_decorated_window_move_resize_window __attribute((alias("IA__gtk_decorated_window_move_resize_window"), visibility("default")));

#undef gtk_decorated_window_set_title
extern __typeof (gtk_decorated_window_set_title) gtk_decorated_window_set_title __attribute((alias("IA__gtk_decorated_window_set_title"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_DIALOG_H__)
#if IN_FILE(__GTK_DIALOG_C__)
#undef gtk_dialog_add_action_widget
extern __typeof (gtk_dialog_add_action_widget) gtk_dialog_add_action_widget __attribute((alias("IA__gtk_dialog_add_action_widget"), visibility("default")));

#undef gtk_dialog_add_button
extern __typeof (gtk_dialog_add_button) gtk_dialog_add_button __attribute((alias("IA__gtk_dialog_add_button"), visibility("default")));

#undef gtk_dialog_add_buttons
extern __typeof (gtk_dialog_add_buttons) gtk_dialog_add_buttons __attribute((alias("IA__gtk_dialog_add_buttons"), visibility("default"))) G_GNUC_NULL_TERMINATED;

#undef gtk_dialog_get_action_area
extern __typeof (gtk_dialog_get_action_area) gtk_dialog_get_action_area __attribute((alias("IA__gtk_dialog_get_action_area"), visibility("default")));

#undef gtk_dialog_get_content_area
extern __typeof (gtk_dialog_get_content_area) gtk_dialog_get_content_area __attribute((alias("IA__gtk_dialog_get_content_area"), visibility("default")));

#undef gtk_dialog_get_widget_for_response
extern __typeof (gtk_dialog_get_widget_for_response) gtk_dialog_get_widget_for_response __attribute((alias("IA__gtk_dialog_get_widget_for_response"), visibility("default")));

#undef gtk_dialog_get_response_for_widget
extern __typeof (gtk_dialog_get_response_for_widget) gtk_dialog_get_response_for_widget __attribute((alias("IA__gtk_dialog_get_response_for_widget"), visibility("default")));

#undef gtk_dialog_get_type
extern __typeof (gtk_dialog_get_type) gtk_dialog_get_type __attribute((alias("IA__gtk_dialog_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_dialog_new
extern __typeof (gtk_dialog_new) gtk_dialog_new __attribute((alias("IA__gtk_dialog_new"), visibility("default")));

#undef gtk_dialog_new_with_buttons
extern __typeof (gtk_dialog_new_with_buttons) gtk_dialog_new_with_buttons __attribute((alias("IA__gtk_dialog_new_with_buttons"), visibility("default")));

#undef gtk_dialog_response
extern __typeof (gtk_dialog_response) gtk_dialog_response __attribute((alias("IA__gtk_dialog_response"), visibility("default")));

#undef gtk_dialog_run
extern __typeof (gtk_dialog_run) gtk_dialog_run __attribute((alias("IA__gtk_dialog_run"), visibility("default")));

#undef gtk_alternative_dialog_button_order
extern __typeof (gtk_alternative_dialog_button_order) gtk_alternative_dialog_button_order __attribute((alias("IA__gtk_alternative_dialog_button_order"), visibility("default")));

#undef gtk_dialog_set_alternative_button_order
extern __typeof (gtk_dialog_set_alternative_button_order) gtk_dialog_set_alternative_button_order __attribute((alias("IA__gtk_dialog_set_alternative_button_order"), visibility("default")));

#undef gtk_dialog_set_alternative_button_order_from_array
extern __typeof (gtk_dialog_set_alternative_button_order_from_array) gtk_dialog_set_alternative_button_order_from_array __attribute((alias("IA__gtk_dialog_set_alternative_button_order_from_array"), visibility("default")));

#undef gtk_dialog_set_default_response
extern __typeof (gtk_dialog_set_default_response) gtk_dialog_set_default_response __attribute((alias("IA__gtk_dialog_set_default_response"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_dialog_get_has_separator
extern __typeof (gtk_dialog_get_has_separator) gtk_dialog_get_has_separator __attribute((alias("IA__gtk_dialog_get_has_separator"), visibility("default")));

#undef gtk_dialog_set_has_separator
extern __typeof (gtk_dialog_set_has_separator) gtk_dialog_set_has_separator __attribute((alias("IA__gtk_dialog_set_has_separator"), visibility("default")));

#endif
#undef gtk_dialog_set_response_sensitive
extern __typeof (gtk_dialog_set_response_sensitive) gtk_dialog_set_response_sensitive __attribute((alias("IA__gtk_dialog_set_response_sensitive"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_DND_H__)
#if IN_FILE(__GTK_DND_C__)
#undef gtk_drag_begin
extern __typeof (gtk_drag_begin) gtk_drag_begin __attribute((alias("IA__gtk_drag_begin"), visibility("default")));

#undef gtk_drag_check_threshold
extern __typeof (gtk_drag_check_threshold) gtk_drag_check_threshold __attribute((alias("IA__gtk_drag_check_threshold"), visibility("default")));

#undef gtk_drag_dest_add_image_targets
extern __typeof (gtk_drag_dest_add_image_targets) gtk_drag_dest_add_image_targets __attribute((alias("IA__gtk_drag_dest_add_image_targets"), visibility("default")));

#undef gtk_drag_dest_add_text_targets
extern __typeof (gtk_drag_dest_add_text_targets) gtk_drag_dest_add_text_targets __attribute((alias("IA__gtk_drag_dest_add_text_targets"), visibility("default")));

#undef gtk_drag_dest_add_uri_targets
extern __typeof (gtk_drag_dest_add_uri_targets) gtk_drag_dest_add_uri_targets __attribute((alias("IA__gtk_drag_dest_add_uri_targets"), visibility("default")));

#undef gtk_drag_dest_find_target
extern __typeof (gtk_drag_dest_find_target) gtk_drag_dest_find_target __attribute((alias("IA__gtk_drag_dest_find_target"), visibility("default")));

#undef gtk_drag_dest_get_target_list
extern __typeof (gtk_drag_dest_get_target_list) gtk_drag_dest_get_target_list __attribute((alias("IA__gtk_drag_dest_get_target_list"), visibility("default")));

#undef gtk_drag_dest_set
extern __typeof (gtk_drag_dest_set) gtk_drag_dest_set __attribute((alias("IA__gtk_drag_dest_set"), visibility("default")));

#undef gtk_drag_dest_set_proxy
extern __typeof (gtk_drag_dest_set_proxy) gtk_drag_dest_set_proxy __attribute((alias("IA__gtk_drag_dest_set_proxy"), visibility("default")));

#undef gtk_drag_dest_set_target_list
extern __typeof (gtk_drag_dest_set_target_list) gtk_drag_dest_set_target_list __attribute((alias("IA__gtk_drag_dest_set_target_list"), visibility("default")));

#undef gtk_drag_dest_set_track_motion
extern __typeof (gtk_drag_dest_set_track_motion) gtk_drag_dest_set_track_motion __attribute((alias("IA__gtk_drag_dest_set_track_motion"), visibility("default")));

#undef gtk_drag_dest_get_track_motion
extern __typeof (gtk_drag_dest_get_track_motion) gtk_drag_dest_get_track_motion __attribute((alias("IA__gtk_drag_dest_get_track_motion"), visibility("default")));

#undef gtk_drag_dest_unset
extern __typeof (gtk_drag_dest_unset) gtk_drag_dest_unset __attribute((alias("IA__gtk_drag_dest_unset"), visibility("default")));

#undef gtk_drag_finish
extern __typeof (gtk_drag_finish) gtk_drag_finish __attribute((alias("IA__gtk_drag_finish"), visibility("default")));

#undef gtk_drag_get_data
extern __typeof (gtk_drag_get_data) gtk_drag_get_data __attribute((alias("IA__gtk_drag_get_data"), visibility("default")));

#undef gtk_drag_get_source_widget
extern __typeof (gtk_drag_get_source_widget) gtk_drag_get_source_widget __attribute((alias("IA__gtk_drag_get_source_widget"), visibility("default")));

#undef gtk_drag_highlight
extern __typeof (gtk_drag_highlight) gtk_drag_highlight __attribute((alias("IA__gtk_drag_highlight"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_drag_set_default_icon
extern __typeof (gtk_drag_set_default_icon) gtk_drag_set_default_icon __attribute((alias("IA__gtk_drag_set_default_icon"), visibility("default")));

#endif
#undef gtk_drag_set_icon_default
extern __typeof (gtk_drag_set_icon_default) gtk_drag_set_icon_default __attribute((alias("IA__gtk_drag_set_icon_default"), visibility("default")));

#undef gtk_drag_set_icon_name
extern __typeof (gtk_drag_set_icon_name) gtk_drag_set_icon_name __attribute((alias("IA__gtk_drag_set_icon_name"), visibility("default")));

#undef gtk_drag_set_icon_pixbuf
extern __typeof (gtk_drag_set_icon_pixbuf) gtk_drag_set_icon_pixbuf __attribute((alias("IA__gtk_drag_set_icon_pixbuf"), visibility("default")));

#undef gtk_drag_set_icon_pixmap
extern __typeof (gtk_drag_set_icon_pixmap) gtk_drag_set_icon_pixmap __attribute((alias("IA__gtk_drag_set_icon_pixmap"), visibility("default")));

#undef gtk_drag_set_icon_stock
extern __typeof (gtk_drag_set_icon_stock) gtk_drag_set_icon_stock __attribute((alias("IA__gtk_drag_set_icon_stock"), visibility("default")));

#undef gtk_drag_set_icon_widget
extern __typeof (gtk_drag_set_icon_widget) gtk_drag_set_icon_widget __attribute((alias("IA__gtk_drag_set_icon_widget"), visibility("default")));

#undef gtk_drag_source_add_image_targets
extern __typeof (gtk_drag_source_add_image_targets) gtk_drag_source_add_image_targets __attribute((alias("IA__gtk_drag_source_add_image_targets"), visibility("default")));

#undef gtk_drag_source_add_text_targets
extern __typeof (gtk_drag_source_add_text_targets) gtk_drag_source_add_text_targets __attribute((alias("IA__gtk_drag_source_add_text_targets"), visibility("default")));

#undef gtk_drag_source_add_uri_targets
extern __typeof (gtk_drag_source_add_uri_targets) gtk_drag_source_add_uri_targets __attribute((alias("IA__gtk_drag_source_add_uri_targets"), visibility("default")));

#undef gtk_drag_source_get_target_list
extern __typeof (gtk_drag_source_get_target_list) gtk_drag_source_get_target_list __attribute((alias("IA__gtk_drag_source_get_target_list"), visibility("default")));

#undef gtk_drag_source_set
extern __typeof (gtk_drag_source_set) gtk_drag_source_set __attribute((alias("IA__gtk_drag_source_set"), visibility("default")));

#undef gtk_drag_source_set_icon
extern __typeof (gtk_drag_source_set_icon) gtk_drag_source_set_icon __attribute((alias("IA__gtk_drag_source_set_icon"), visibility("default")));

#undef gtk_drag_source_set_icon_name
extern __typeof (gtk_drag_source_set_icon_name) gtk_drag_source_set_icon_name __attribute((alias("IA__gtk_drag_source_set_icon_name"), visibility("default")));

#undef gtk_drag_source_set_icon_pixbuf
extern __typeof (gtk_drag_source_set_icon_pixbuf) gtk_drag_source_set_icon_pixbuf __attribute((alias("IA__gtk_drag_source_set_icon_pixbuf"), visibility("default")));

#undef gtk_drag_source_set_icon_stock
extern __typeof (gtk_drag_source_set_icon_stock) gtk_drag_source_set_icon_stock __attribute((alias("IA__gtk_drag_source_set_icon_stock"), visibility("default")));

#undef gtk_drag_source_set_target_list
extern __typeof (gtk_drag_source_set_target_list) gtk_drag_source_set_target_list __attribute((alias("IA__gtk_drag_source_set_target_list"), visibility("default")));

#undef gtk_drag_source_unset
extern __typeof (gtk_drag_source_unset) gtk_drag_source_unset __attribute((alias("IA__gtk_drag_source_unset"), visibility("default")));

#undef gtk_drag_unhighlight
extern __typeof (gtk_drag_unhighlight) gtk_drag_unhighlight __attribute((alias("IA__gtk_drag_unhighlight"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_STATUS_ICON_H__)
#if IN_FILE(__GTK_STATUS_ICON_C__)
#undef gtk_status_icon_get_type
extern __typeof (gtk_status_icon_get_type) gtk_status_icon_get_type __attribute((alias("IA__gtk_status_icon_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_status_icon_new
extern __typeof (gtk_status_icon_new) gtk_status_icon_new __attribute((alias("IA__gtk_status_icon_new"), visibility("default")));

#undef gtk_status_icon_new_from_pixbuf
extern __typeof (gtk_status_icon_new_from_pixbuf) gtk_status_icon_new_from_pixbuf __attribute((alias("IA__gtk_status_icon_new_from_pixbuf"), visibility("default")));

#undef gtk_status_icon_new_from_file
extern __typeof (gtk_status_icon_new_from_file) gtk_status_icon_new_from_file __attribute((alias("IA__gtk_status_icon_new_from_file"), visibility("default")));

#undef gtk_status_icon_new_from_stock
extern __typeof (gtk_status_icon_new_from_stock) gtk_status_icon_new_from_stock __attribute((alias("IA__gtk_status_icon_new_from_stock"), visibility("default")));

#undef gtk_status_icon_new_from_icon_name
extern __typeof (gtk_status_icon_new_from_icon_name) gtk_status_icon_new_from_icon_name __attribute((alias("IA__gtk_status_icon_new_from_icon_name"), visibility("default")));

#undef gtk_status_icon_new_from_gicon
extern __typeof (gtk_status_icon_new_from_gicon) gtk_status_icon_new_from_gicon __attribute((alias("IA__gtk_status_icon_new_from_gicon"), visibility("default")));

#undef gtk_status_icon_set_from_pixbuf
extern __typeof (gtk_status_icon_set_from_pixbuf) gtk_status_icon_set_from_pixbuf __attribute((alias("IA__gtk_status_icon_set_from_pixbuf"), visibility("default")));

#undef gtk_status_icon_set_from_file
extern __typeof (gtk_status_icon_set_from_file) gtk_status_icon_set_from_file __attribute((alias("IA__gtk_status_icon_set_from_file"), visibility("default")));

#undef gtk_status_icon_set_from_stock
extern __typeof (gtk_status_icon_set_from_stock) gtk_status_icon_set_from_stock __attribute((alias("IA__gtk_status_icon_set_from_stock"), visibility("default")));

#undef gtk_status_icon_set_from_icon_name
extern __typeof (gtk_status_icon_set_from_icon_name) gtk_status_icon_set_from_icon_name __attribute((alias("IA__gtk_status_icon_set_from_icon_name"), visibility("default")));

#undef gtk_status_icon_set_from_gicon
extern __typeof (gtk_status_icon_set_from_gicon) gtk_status_icon_set_from_gicon __attribute((alias("IA__gtk_status_icon_set_from_gicon"), visibility("default")));

#undef gtk_status_icon_get_storage_type
extern __typeof (gtk_status_icon_get_storage_type) gtk_status_icon_get_storage_type __attribute((alias("IA__gtk_status_icon_get_storage_type"), visibility("default")));

#undef gtk_status_icon_get_pixbuf
extern __typeof (gtk_status_icon_get_pixbuf) gtk_status_icon_get_pixbuf __attribute((alias("IA__gtk_status_icon_get_pixbuf"), visibility("default")));

#undef gtk_status_icon_get_screen
extern __typeof (gtk_status_icon_get_screen) gtk_status_icon_get_screen __attribute((alias("IA__gtk_status_icon_get_screen"), visibility("default")));

#undef gtk_status_icon_get_stock
extern __typeof (gtk_status_icon_get_stock) gtk_status_icon_get_stock __attribute((alias("IA__gtk_status_icon_get_stock"), visibility("default")));

#undef gtk_status_icon_get_icon_name
extern __typeof (gtk_status_icon_get_icon_name) gtk_status_icon_get_icon_name __attribute((alias("IA__gtk_status_icon_get_icon_name"), visibility("default")));

#undef gtk_status_icon_get_gicon
extern __typeof (gtk_status_icon_get_gicon) gtk_status_icon_get_gicon __attribute((alias("IA__gtk_status_icon_get_gicon"), visibility("default")));

#undef gtk_status_icon_get_size
extern __typeof (gtk_status_icon_get_size) gtk_status_icon_get_size __attribute((alias("IA__gtk_status_icon_get_size"), visibility("default")));

#undef gtk_status_icon_set_screen
extern __typeof (gtk_status_icon_set_screen) gtk_status_icon_set_screen __attribute((alias("IA__gtk_status_icon_set_screen"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_status_icon_set_tooltip
extern __typeof (gtk_status_icon_set_tooltip) gtk_status_icon_set_tooltip __attribute((alias("IA__gtk_status_icon_set_tooltip"), visibility("default")));

#endif
#undef gtk_status_icon_set_tooltip_text
extern __typeof (gtk_status_icon_set_tooltip_text) gtk_status_icon_set_tooltip_text __attribute((alias("IA__gtk_status_icon_set_tooltip_text"), visibility("default")));

#undef gtk_status_icon_get_tooltip_text
extern __typeof (gtk_status_icon_get_tooltip_text) gtk_status_icon_get_tooltip_text __attribute((alias("IA__gtk_status_icon_get_tooltip_text"), visibility("default")));

#undef gtk_status_icon_set_tooltip_markup
extern __typeof (gtk_status_icon_set_tooltip_markup) gtk_status_icon_set_tooltip_markup __attribute((alias("IA__gtk_status_icon_set_tooltip_markup"), visibility("default")));

#undef gtk_status_icon_get_tooltip_markup
extern __typeof (gtk_status_icon_get_tooltip_markup) gtk_status_icon_get_tooltip_markup __attribute((alias("IA__gtk_status_icon_get_tooltip_markup"), visibility("default")));

#undef gtk_status_icon_set_has_tooltip
extern __typeof (gtk_status_icon_set_has_tooltip) gtk_status_icon_set_has_tooltip __attribute((alias("IA__gtk_status_icon_set_has_tooltip"), visibility("default")));

#undef gtk_status_icon_get_has_tooltip
extern __typeof (gtk_status_icon_get_has_tooltip) gtk_status_icon_get_has_tooltip __attribute((alias("IA__gtk_status_icon_get_has_tooltip"), visibility("default")));

#undef gtk_status_icon_set_visible
extern __typeof (gtk_status_icon_set_visible) gtk_status_icon_set_visible __attribute((alias("IA__gtk_status_icon_set_visible"), visibility("default")));

#undef gtk_status_icon_get_visible
extern __typeof (gtk_status_icon_get_visible) gtk_status_icon_get_visible __attribute((alias("IA__gtk_status_icon_get_visible"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_status_icon_set_blinking
extern __typeof (gtk_status_icon_set_blinking) gtk_status_icon_set_blinking __attribute((alias("IA__gtk_status_icon_set_blinking"), visibility("default")));

#undef gtk_status_icon_get_blinking
extern __typeof (gtk_status_icon_get_blinking) gtk_status_icon_get_blinking __attribute((alias("IA__gtk_status_icon_get_blinking"), visibility("default")));

#endif
#undef gtk_status_icon_is_embedded
extern __typeof (gtk_status_icon_is_embedded) gtk_status_icon_is_embedded __attribute((alias("IA__gtk_status_icon_is_embedded"), visibility("default")));

#undef gtk_status_icon_position_menu
extern __typeof (gtk_status_icon_position_menu) gtk_status_icon_position_menu __attribute((alias("IA__gtk_status_icon_position_menu"), visibility("default")));

#undef gtk_status_icon_get_geometry
extern __typeof (gtk_status_icon_get_geometry) gtk_status_icon_get_geometry __attribute((alias("IA__gtk_status_icon_get_geometry"), visibility("default")));

#undef gtk_status_icon_get_x11_window_id
extern __typeof (gtk_status_icon_get_x11_window_id) gtk_status_icon_get_x11_window_id __attribute((alias("IA__gtk_status_icon_get_x11_window_id"), visibility("default")));

#undef gtk_status_icon_get_title
extern __typeof (gtk_status_icon_get_title) gtk_status_icon_get_title __attribute((alias("IA__gtk_status_icon_get_title"), visibility("default")));

#undef gtk_status_icon_set_title
extern __typeof (gtk_status_icon_set_title) gtk_status_icon_set_title __attribute((alias("IA__gtk_status_icon_set_title"), visibility("default")));

#undef gtk_status_icon_set_name
extern __typeof (gtk_status_icon_set_name) gtk_status_icon_set_name __attribute((alias("IA__gtk_status_icon_set_name"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_STYLE_H__)
#if IN_FILE(__GTK_STYLE_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_draw_arrow
extern __typeof (gtk_draw_arrow) gtk_draw_arrow __attribute((alias("IA__gtk_draw_arrow"), visibility("default")));

#undef gtk_draw_box
extern __typeof (gtk_draw_box) gtk_draw_box __attribute((alias("IA__gtk_draw_box"), visibility("default")));

#undef gtk_draw_box_gap
extern __typeof (gtk_draw_box_gap) gtk_draw_box_gap __attribute((alias("IA__gtk_draw_box_gap"), visibility("default")));

#undef gtk_draw_check
extern __typeof (gtk_draw_check) gtk_draw_check __attribute((alias("IA__gtk_draw_check"), visibility("default")));

#undef gtk_draw_diamond
extern __typeof (gtk_draw_diamond) gtk_draw_diamond __attribute((alias("IA__gtk_draw_diamond"), visibility("default")));

#undef gtk_draw_expander
extern __typeof (gtk_draw_expander) gtk_draw_expander __attribute((alias("IA__gtk_draw_expander"), visibility("default")));

#undef gtk_draw_extension
extern __typeof (gtk_draw_extension) gtk_draw_extension __attribute((alias("IA__gtk_draw_extension"), visibility("default")));

#undef gtk_draw_flat_box
extern __typeof (gtk_draw_flat_box) gtk_draw_flat_box __attribute((alias("IA__gtk_draw_flat_box"), visibility("default")));

#undef gtk_draw_focus
extern __typeof (gtk_draw_focus) gtk_draw_focus __attribute((alias("IA__gtk_draw_focus"), visibility("default")));

#undef gtk_draw_handle
extern __typeof (gtk_draw_handle) gtk_draw_handle __attribute((alias("IA__gtk_draw_handle"), visibility("default")));

#undef gtk_draw_hline
extern __typeof (gtk_draw_hline) gtk_draw_hline __attribute((alias("IA__gtk_draw_hline"), visibility("default")));

#undef gtk_draw_layout
extern __typeof (gtk_draw_layout) gtk_draw_layout __attribute((alias("IA__gtk_draw_layout"), visibility("default")));

#undef gtk_draw_option
extern __typeof (gtk_draw_option) gtk_draw_option __attribute((alias("IA__gtk_draw_option"), visibility("default")));

#undef gtk_draw_polygon
extern __typeof (gtk_draw_polygon) gtk_draw_polygon __attribute((alias("IA__gtk_draw_polygon"), visibility("default")));

#undef gtk_draw_resize_grip
extern __typeof (gtk_draw_resize_grip) gtk_draw_resize_grip __attribute((alias("IA__gtk_draw_resize_grip"), visibility("default")));

#undef gtk_draw_shadow
extern __typeof (gtk_draw_shadow) gtk_draw_shadow __attribute((alias("IA__gtk_draw_shadow"), visibility("default")));

#undef gtk_draw_shadow_gap
extern __typeof (gtk_draw_shadow_gap) gtk_draw_shadow_gap __attribute((alias("IA__gtk_draw_shadow_gap"), visibility("default")));

#undef gtk_draw_slider
extern __typeof (gtk_draw_slider) gtk_draw_slider __attribute((alias("IA__gtk_draw_slider"), visibility("default")));

#undef gtk_draw_string
extern __typeof (gtk_draw_string) gtk_draw_string __attribute((alias("IA__gtk_draw_string"), visibility("default")));

#undef gtk_draw_tab
extern __typeof (gtk_draw_tab) gtk_draw_tab __attribute((alias("IA__gtk_draw_tab"), visibility("default")));

#undef gtk_draw_vline
extern __typeof (gtk_draw_vline) gtk_draw_vline __attribute((alias("IA__gtk_draw_vline"), visibility("default")));

#undef gtk_style_set_font
extern __typeof (gtk_style_set_font) gtk_style_set_font __attribute((alias("IA__gtk_style_set_font"), visibility("default")));

#undef gtk_style_get_font
extern __typeof (gtk_style_get_font) gtk_style_get_font __attribute((alias("IA__gtk_style_get_font"), visibility("default")));

#undef gtk_style_ref
extern __typeof (gtk_style_ref) gtk_style_ref __attribute((alias("IA__gtk_style_ref"), visibility("default")));

#undef gtk_style_unref
extern __typeof (gtk_style_unref) gtk_style_unref __attribute((alias("IA__gtk_style_unref"), visibility("default")));

#undef gtk_paint_string
extern __typeof (gtk_paint_string) gtk_paint_string __attribute((alias("IA__gtk_paint_string"), visibility("default")));

#endif
#undef gtk_paint_arrow
extern __typeof (gtk_paint_arrow) gtk_paint_arrow __attribute((alias("IA__gtk_paint_arrow"), visibility("default")));

#undef gtk_paint_box
extern __typeof (gtk_paint_box) gtk_paint_box __attribute((alias("IA__gtk_paint_box"), visibility("default")));

#undef gtk_paint_box_gap
extern __typeof (gtk_paint_box_gap) gtk_paint_box_gap __attribute((alias("IA__gtk_paint_box_gap"), visibility("default")));

#undef gtk_paint_check
extern __typeof (gtk_paint_check) gtk_paint_check __attribute((alias("IA__gtk_paint_check"), visibility("default")));

#undef gtk_paint_diamond
extern __typeof (gtk_paint_diamond) gtk_paint_diamond __attribute((alias("IA__gtk_paint_diamond"), visibility("default")));

#undef gtk_paint_expander
extern __typeof (gtk_paint_expander) gtk_paint_expander __attribute((alias("IA__gtk_paint_expander"), visibility("default")));

#undef gtk_paint_extension
extern __typeof (gtk_paint_extension) gtk_paint_extension __attribute((alias("IA__gtk_paint_extension"), visibility("default")));

#undef gtk_paint_flat_box
extern __typeof (gtk_paint_flat_box) gtk_paint_flat_box __attribute((alias("IA__gtk_paint_flat_box"), visibility("default")));

#undef gtk_paint_focus
extern __typeof (gtk_paint_focus) gtk_paint_focus __attribute((alias("IA__gtk_paint_focus"), visibility("default")));

#undef gtk_paint_handle
extern __typeof (gtk_paint_handle) gtk_paint_handle __attribute((alias("IA__gtk_paint_handle"), visibility("default")));

#undef gtk_paint_hline
extern __typeof (gtk_paint_hline) gtk_paint_hline __attribute((alias("IA__gtk_paint_hline"), visibility("default")));

#undef gtk_paint_layout
extern __typeof (gtk_paint_layout) gtk_paint_layout __attribute((alias("IA__gtk_paint_layout"), visibility("default")));

#undef gtk_paint_option
extern __typeof (gtk_paint_option) gtk_paint_option __attribute((alias("IA__gtk_paint_option"), visibility("default")));

#undef gtk_paint_polygon
extern __typeof (gtk_paint_polygon) gtk_paint_polygon __attribute((alias("IA__gtk_paint_polygon"), visibility("default")));

#undef gtk_paint_resize_grip
extern __typeof (gtk_paint_resize_grip) gtk_paint_resize_grip __attribute((alias("IA__gtk_paint_resize_grip"), visibility("default")));

#undef gtk_paint_shadow
extern __typeof (gtk_paint_shadow) gtk_paint_shadow __attribute((alias("IA__gtk_paint_shadow"), visibility("default")));

#undef gtk_paint_shadow_gap
extern __typeof (gtk_paint_shadow_gap) gtk_paint_shadow_gap __attribute((alias("IA__gtk_paint_shadow_gap"), visibility("default")));

#undef gtk_paint_slider
extern __typeof (gtk_paint_slider) gtk_paint_slider __attribute((alias("IA__gtk_paint_slider"), visibility("default")));

#undef gtk_paint_spinner
extern __typeof (gtk_paint_spinner) gtk_paint_spinner __attribute((alias("IA__gtk_paint_spinner"), visibility("default")));

#undef gtk_paint_tab
extern __typeof (gtk_paint_tab) gtk_paint_tab __attribute((alias("IA__gtk_paint_tab"), visibility("default")));

#undef gtk_paint_vline
extern __typeof (gtk_paint_vline) gtk_paint_vline __attribute((alias("IA__gtk_paint_vline"), visibility("default")));

#undef gtk_border_new
extern __typeof (gtk_border_new) gtk_border_new __attribute((alias("IA__gtk_border_new"), visibility("default"))) G_GNUC_MALLOC;

#undef gtk_border_copy
extern __typeof (gtk_border_copy) gtk_border_copy __attribute((alias("IA__gtk_border_copy"), visibility("default")));

#undef gtk_border_free
extern __typeof (gtk_border_free) gtk_border_free __attribute((alias("IA__gtk_border_free"), visibility("default")));

#undef gtk_border_get_type
extern __typeof (gtk_border_get_type) gtk_border_get_type __attribute((alias("IA__gtk_border_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_style_apply_default_background
extern __typeof (gtk_style_apply_default_background) gtk_style_apply_default_background __attribute((alias("IA__gtk_style_apply_default_background"), visibility("default")));

#undef gtk_style_attach
extern __typeof (gtk_style_attach) gtk_style_attach __attribute((alias("IA__gtk_style_attach"), visibility("default")));

#undef gtk_style_copy
extern __typeof (gtk_style_copy) gtk_style_copy __attribute((alias("IA__gtk_style_copy"), visibility("default")));

#undef gtk_style_detach
extern __typeof (gtk_style_detach) gtk_style_detach __attribute((alias("IA__gtk_style_detach"), visibility("default")));

#undef gtk_style_get_type
extern __typeof (gtk_style_get_type) gtk_style_get_type __attribute((alias("IA__gtk_style_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_style_lookup_icon_set
extern __typeof (gtk_style_lookup_icon_set) gtk_style_lookup_icon_set __attribute((alias("IA__gtk_style_lookup_icon_set"), visibility("default")));

#undef gtk_style_lookup_color
extern __typeof (gtk_style_lookup_color) gtk_style_lookup_color __attribute((alias("IA__gtk_style_lookup_color"), visibility("default")));

#undef gtk_style_new
extern __typeof (gtk_style_new) gtk_style_new __attribute((alias("IA__gtk_style_new"), visibility("default")));

#undef gtk_style_render_icon
extern __typeof (gtk_style_render_icon) gtk_style_render_icon __attribute((alias("IA__gtk_style_render_icon"), visibility("default")));

#undef gtk_style_set_background
extern __typeof (gtk_style_set_background) gtk_style_set_background __attribute((alias("IA__gtk_style_set_background"), visibility("default")));

#undef gtk_draw_insertion_cursor
extern __typeof (gtk_draw_insertion_cursor) gtk_draw_insertion_cursor __attribute((alias("IA__gtk_draw_insertion_cursor"), visibility("default")));

#undef gtk_style_get_style_property
extern __typeof (gtk_style_get_style_property) gtk_style_get_style_property __attribute((alias("IA__gtk_style_get_style_property"), visibility("default")));

#undef gtk_style_get_valist
extern __typeof (gtk_style_get_valist) gtk_style_get_valist __attribute((alias("IA__gtk_style_get_valist"), visibility("default")));

#undef gtk_style_get
extern __typeof (gtk_style_get) gtk_style_get __attribute((alias("IA__gtk_style_get"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_DRAWING_AREA_H__)
#if IN_FILE(__GTK_DRAWING_AREA_C__)
#undef gtk_drawing_area_get_type
extern __typeof (gtk_drawing_area_get_type) gtk_drawing_area_get_type __attribute((alias("IA__gtk_drawing_area_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_drawing_area_new
extern __typeof (gtk_drawing_area_new) gtk_drawing_area_new __attribute((alias("IA__gtk_drawing_area_new"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_drawing_area_size
extern __typeof (gtk_drawing_area_size) gtk_drawing_area_size __attribute((alias("IA__gtk_drawing_area_size"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_EDITABLE_H__)
#if IN_FILE(__GTK_EDITABLE_C__)
#undef gtk_editable_copy_clipboard
extern __typeof (gtk_editable_copy_clipboard) gtk_editable_copy_clipboard __attribute((alias("IA__gtk_editable_copy_clipboard"), visibility("default")));

#undef gtk_editable_cut_clipboard
extern __typeof (gtk_editable_cut_clipboard) gtk_editable_cut_clipboard __attribute((alias("IA__gtk_editable_cut_clipboard"), visibility("default")));

#undef gtk_editable_delete_selection
extern __typeof (gtk_editable_delete_selection) gtk_editable_delete_selection __attribute((alias("IA__gtk_editable_delete_selection"), visibility("default")));

#undef gtk_editable_delete_text
extern __typeof (gtk_editable_delete_text) gtk_editable_delete_text __attribute((alias("IA__gtk_editable_delete_text"), visibility("default")));

#undef gtk_editable_get_chars
extern __typeof (gtk_editable_get_chars) gtk_editable_get_chars __attribute((alias("IA__gtk_editable_get_chars"), visibility("default")));

#undef gtk_editable_get_editable
extern __typeof (gtk_editable_get_editable) gtk_editable_get_editable __attribute((alias("IA__gtk_editable_get_editable"), visibility("default")));

#undef gtk_editable_get_position
extern __typeof (gtk_editable_get_position) gtk_editable_get_position __attribute((alias("IA__gtk_editable_get_position"), visibility("default")));

#undef gtk_editable_get_selection_bounds
extern __typeof (gtk_editable_get_selection_bounds) gtk_editable_get_selection_bounds __attribute((alias("IA__gtk_editable_get_selection_bounds"), visibility("default")));

#undef gtk_editable_get_type
extern __typeof (gtk_editable_get_type) gtk_editable_get_type __attribute((alias("IA__gtk_editable_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_editable_insert_text
extern __typeof (gtk_editable_insert_text) gtk_editable_insert_text __attribute((alias("IA__gtk_editable_insert_text"), visibility("default")));

#undef gtk_editable_paste_clipboard
extern __typeof (gtk_editable_paste_clipboard) gtk_editable_paste_clipboard __attribute((alias("IA__gtk_editable_paste_clipboard"), visibility("default")));

#undef gtk_editable_select_region
extern __typeof (gtk_editable_select_region) gtk_editable_select_region __attribute((alias("IA__gtk_editable_select_region"), visibility("default")));

#undef gtk_editable_set_editable
extern __typeof (gtk_editable_set_editable) gtk_editable_set_editable __attribute((alias("IA__gtk_editable_set_editable"), visibility("default")));

#undef gtk_editable_set_position
extern __typeof (gtk_editable_set_position) gtk_editable_set_position __attribute((alias("IA__gtk_editable_set_position"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_ENTRY_H__)
#if IN_FILE(__GTK_ENTRY_C__)
#undef gtk_entry_get_activates_default
extern __typeof (gtk_entry_get_activates_default) gtk_entry_get_activates_default __attribute((alias("IA__gtk_entry_get_activates_default"), visibility("default")));

#undef gtk_entry_get_alignment
extern __typeof (gtk_entry_get_alignment) gtk_entry_get_alignment __attribute((alias("IA__gtk_entry_get_alignment"), visibility("default")));

#undef gtk_entry_get_buffer
extern __typeof (gtk_entry_get_buffer) gtk_entry_get_buffer __attribute((alias("IA__gtk_entry_get_buffer"), visibility("default")));

#undef gtk_entry_get_completion
extern __typeof (gtk_entry_get_completion) gtk_entry_get_completion __attribute((alias("IA__gtk_entry_get_completion"), visibility("default")));

#undef gtk_entry_get_current_icon_drag_source
extern __typeof (gtk_entry_get_current_icon_drag_source) gtk_entry_get_current_icon_drag_source __attribute((alias("IA__gtk_entry_get_current_icon_drag_source"), visibility("default")));

#undef gtk_entry_get_cursor_hadjustment
extern __typeof (gtk_entry_get_cursor_hadjustment) gtk_entry_get_cursor_hadjustment __attribute((alias("IA__gtk_entry_get_cursor_hadjustment"), visibility("default")));

#undef gtk_entry_get_has_frame
extern __typeof (gtk_entry_get_has_frame) gtk_entry_get_has_frame __attribute((alias("IA__gtk_entry_get_has_frame"), visibility("default")));

#undef gtk_entry_get_icon_activatable
extern __typeof (gtk_entry_get_icon_activatable) gtk_entry_get_icon_activatable __attribute((alias("IA__gtk_entry_get_icon_activatable"), visibility("default")));

#undef gtk_entry_get_icon_at_pos
extern __typeof (gtk_entry_get_icon_at_pos) gtk_entry_get_icon_at_pos __attribute((alias("IA__gtk_entry_get_icon_at_pos"), visibility("default")));

#undef gtk_entry_get_icon_gicon
extern __typeof (gtk_entry_get_icon_gicon) gtk_entry_get_icon_gicon __attribute((alias("IA__gtk_entry_get_icon_gicon"), visibility("default")));

#undef gtk_entry_get_icon_name
extern __typeof (gtk_entry_get_icon_name) gtk_entry_get_icon_name __attribute((alias("IA__gtk_entry_get_icon_name"), visibility("default")));

#undef gtk_entry_get_icon_pixbuf
extern __typeof (gtk_entry_get_icon_pixbuf) gtk_entry_get_icon_pixbuf __attribute((alias("IA__gtk_entry_get_icon_pixbuf"), visibility("default")));

#undef gtk_entry_get_icon_sensitive
extern __typeof (gtk_entry_get_icon_sensitive) gtk_entry_get_icon_sensitive __attribute((alias("IA__gtk_entry_get_icon_sensitive"), visibility("default")));

#undef gtk_entry_get_icon_stock
extern __typeof (gtk_entry_get_icon_stock) gtk_entry_get_icon_stock __attribute((alias("IA__gtk_entry_get_icon_stock"), visibility("default")));

#undef gtk_entry_get_icon_storage_type
extern __typeof (gtk_entry_get_icon_storage_type) gtk_entry_get_icon_storage_type __attribute((alias("IA__gtk_entry_get_icon_storage_type"), visibility("default")));

#undef gtk_entry_get_icon_tooltip_markup
extern __typeof (gtk_entry_get_icon_tooltip_markup) gtk_entry_get_icon_tooltip_markup __attribute((alias("IA__gtk_entry_get_icon_tooltip_markup"), visibility("default")));

#undef gtk_entry_get_icon_tooltip_text
extern __typeof (gtk_entry_get_icon_tooltip_text) gtk_entry_get_icon_tooltip_text __attribute((alias("IA__gtk_entry_get_icon_tooltip_text"), visibility("default")));

#undef gtk_entry_get_icon_window
extern __typeof (gtk_entry_get_icon_window) gtk_entry_get_icon_window __attribute((alias("IA__gtk_entry_get_icon_window"), visibility("default")));

#undef gtk_entry_get_inner_border
extern __typeof (gtk_entry_get_inner_border) gtk_entry_get_inner_border __attribute((alias("IA__gtk_entry_get_inner_border"), visibility("default")));

#undef gtk_entry_get_invisible_char
extern __typeof (gtk_entry_get_invisible_char) gtk_entry_get_invisible_char __attribute((alias("IA__gtk_entry_get_invisible_char"), visibility("default")));

#undef gtk_entry_get_layout
extern __typeof (gtk_entry_get_layout) gtk_entry_get_layout __attribute((alias("IA__gtk_entry_get_layout"), visibility("default")));

#undef gtk_entry_get_layout_offsets
extern __typeof (gtk_entry_get_layout_offsets) gtk_entry_get_layout_offsets __attribute((alias("IA__gtk_entry_get_layout_offsets"), visibility("default")));

#undef gtk_entry_get_max_length
extern __typeof (gtk_entry_get_max_length) gtk_entry_get_max_length __attribute((alias("IA__gtk_entry_get_max_length"), visibility("default")));

#undef gtk_entry_get_overwrite_mode
extern __typeof (gtk_entry_get_overwrite_mode) gtk_entry_get_overwrite_mode __attribute((alias("IA__gtk_entry_get_overwrite_mode"), visibility("default")));

#undef gtk_entry_get_progress_fraction
extern __typeof (gtk_entry_get_progress_fraction) gtk_entry_get_progress_fraction __attribute((alias("IA__gtk_entry_get_progress_fraction"), visibility("default")));

#undef gtk_entry_get_progress_pulse_step
extern __typeof (gtk_entry_get_progress_pulse_step) gtk_entry_get_progress_pulse_step __attribute((alias("IA__gtk_entry_get_progress_pulse_step"), visibility("default")));

#undef gtk_entry_get_text
extern __typeof (gtk_entry_get_text) gtk_entry_get_text __attribute((alias("IA__gtk_entry_get_text"), visibility("default")));

#undef gtk_entry_get_text_length
extern __typeof (gtk_entry_get_text_length) gtk_entry_get_text_length __attribute((alias("IA__gtk_entry_get_text_length"), visibility("default")));

#undef gtk_entry_get_type
extern __typeof (gtk_entry_get_type) gtk_entry_get_type __attribute((alias("IA__gtk_entry_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_entry_get_visibility
extern __typeof (gtk_entry_get_visibility) gtk_entry_get_visibility __attribute((alias("IA__gtk_entry_get_visibility"), visibility("default")));

#undef gtk_entry_get_width_chars
extern __typeof (gtk_entry_get_width_chars) gtk_entry_get_width_chars __attribute((alias("IA__gtk_entry_get_width_chars"), visibility("default")));

#undef gtk_entry_get_text_window
extern __typeof (gtk_entry_get_text_window) gtk_entry_get_text_window __attribute((alias("IA__gtk_entry_get_text_window"), visibility("default")));

#undef gtk_entry_im_context_filter_keypress
extern __typeof (gtk_entry_im_context_filter_keypress) gtk_entry_im_context_filter_keypress __attribute((alias("IA__gtk_entry_im_context_filter_keypress"), visibility("default")));

#undef gtk_entry_layout_index_to_text_index
extern __typeof (gtk_entry_layout_index_to_text_index) gtk_entry_layout_index_to_text_index __attribute((alias("IA__gtk_entry_layout_index_to_text_index"), visibility("default")));

#undef gtk_entry_new
extern __typeof (gtk_entry_new) gtk_entry_new __attribute((alias("IA__gtk_entry_new"), visibility("default")));

#undef gtk_entry_new_with_buffer
extern __typeof (gtk_entry_new_with_buffer) gtk_entry_new_with_buffer __attribute((alias("IA__gtk_entry_new_with_buffer"), visibility("default")));

#undef gtk_entry_progress_pulse
extern __typeof (gtk_entry_progress_pulse) gtk_entry_progress_pulse __attribute((alias("IA__gtk_entry_progress_pulse"), visibility("default")));

#undef gtk_entry_reset_im_context
extern __typeof (gtk_entry_reset_im_context) gtk_entry_reset_im_context __attribute((alias("IA__gtk_entry_reset_im_context"), visibility("default")));

#undef gtk_entry_set_activates_default
extern __typeof (gtk_entry_set_activates_default) gtk_entry_set_activates_default __attribute((alias("IA__gtk_entry_set_activates_default"), visibility("default")));

#undef gtk_entry_set_alignment
extern __typeof (gtk_entry_set_alignment) gtk_entry_set_alignment __attribute((alias("IA__gtk_entry_set_alignment"), visibility("default")));

#undef gtk_entry_set_buffer
extern __typeof (gtk_entry_set_buffer) gtk_entry_set_buffer __attribute((alias("IA__gtk_entry_set_buffer"), visibility("default")));

#undef gtk_entry_set_completion
extern __typeof (gtk_entry_set_completion) gtk_entry_set_completion __attribute((alias("IA__gtk_entry_set_completion"), visibility("default")));

#undef gtk_entry_set_cursor_hadjustment
extern __typeof (gtk_entry_set_cursor_hadjustment) gtk_entry_set_cursor_hadjustment __attribute((alias("IA__gtk_entry_set_cursor_hadjustment"), visibility("default")));

#undef gtk_entry_set_has_frame
extern __typeof (gtk_entry_set_has_frame) gtk_entry_set_has_frame __attribute((alias("IA__gtk_entry_set_has_frame"), visibility("default")));

#undef gtk_entry_set_icon_activatable
extern __typeof (gtk_entry_set_icon_activatable) gtk_entry_set_icon_activatable __attribute((alias("IA__gtk_entry_set_icon_activatable"), visibility("default")));

#undef gtk_entry_set_icon_drag_source
extern __typeof (gtk_entry_set_icon_drag_source) gtk_entry_set_icon_drag_source __attribute((alias("IA__gtk_entry_set_icon_drag_source"), visibility("default")));

#undef gtk_entry_set_icon_from_gicon
extern __typeof (gtk_entry_set_icon_from_gicon) gtk_entry_set_icon_from_gicon __attribute((alias("IA__gtk_entry_set_icon_from_gicon"), visibility("default")));

#undef gtk_entry_set_icon_from_icon_name
extern __typeof (gtk_entry_set_icon_from_icon_name) gtk_entry_set_icon_from_icon_name __attribute((alias("IA__gtk_entry_set_icon_from_icon_name"), visibility("default")));

#undef gtk_entry_set_icon_from_pixbuf
extern __typeof (gtk_entry_set_icon_from_pixbuf) gtk_entry_set_icon_from_pixbuf __attribute((alias("IA__gtk_entry_set_icon_from_pixbuf"), visibility("default")));

#undef gtk_entry_set_icon_from_stock
extern __typeof (gtk_entry_set_icon_from_stock) gtk_entry_set_icon_from_stock __attribute((alias("IA__gtk_entry_set_icon_from_stock"), visibility("default")));

#undef gtk_entry_set_icon_sensitive
extern __typeof (gtk_entry_set_icon_sensitive) gtk_entry_set_icon_sensitive __attribute((alias("IA__gtk_entry_set_icon_sensitive"), visibility("default")));

#undef gtk_entry_set_icon_tooltip_markup
extern __typeof (gtk_entry_set_icon_tooltip_markup) gtk_entry_set_icon_tooltip_markup __attribute((alias("IA__gtk_entry_set_icon_tooltip_markup"), visibility("default")));

#undef gtk_entry_set_icon_tooltip_text
extern __typeof (gtk_entry_set_icon_tooltip_text) gtk_entry_set_icon_tooltip_text __attribute((alias("IA__gtk_entry_set_icon_tooltip_text"), visibility("default")));

#undef gtk_entry_set_inner_border
extern __typeof (gtk_entry_set_inner_border) gtk_entry_set_inner_border __attribute((alias("IA__gtk_entry_set_inner_border"), visibility("default")));

#undef gtk_entry_set_invisible_char
extern __typeof (gtk_entry_set_invisible_char) gtk_entry_set_invisible_char __attribute((alias("IA__gtk_entry_set_invisible_char"), visibility("default")));

#undef gtk_entry_set_max_length
extern __typeof (gtk_entry_set_max_length) gtk_entry_set_max_length __attribute((alias("IA__gtk_entry_set_max_length"), visibility("default")));

#undef gtk_entry_set_overwrite_mode
extern __typeof (gtk_entry_set_overwrite_mode) gtk_entry_set_overwrite_mode __attribute((alias("IA__gtk_entry_set_overwrite_mode"), visibility("default")));

#undef gtk_entry_set_progress_fraction
extern __typeof (gtk_entry_set_progress_fraction) gtk_entry_set_progress_fraction __attribute((alias("IA__gtk_entry_set_progress_fraction"), visibility("default")));

#undef gtk_entry_set_progress_pulse_step
extern __typeof (gtk_entry_set_progress_pulse_step) gtk_entry_set_progress_pulse_step __attribute((alias("IA__gtk_entry_set_progress_pulse_step"), visibility("default")));

#undef gtk_entry_set_text
extern __typeof (gtk_entry_set_text) gtk_entry_set_text __attribute((alias("IA__gtk_entry_set_text"), visibility("default")));

#undef gtk_entry_set_visibility
extern __typeof (gtk_entry_set_visibility) gtk_entry_set_visibility __attribute((alias("IA__gtk_entry_set_visibility"), visibility("default")));

#undef gtk_entry_set_width_chars
extern __typeof (gtk_entry_set_width_chars) gtk_entry_set_width_chars __attribute((alias("IA__gtk_entry_set_width_chars"), visibility("default")));

#undef gtk_entry_text_index_to_layout_index
extern __typeof (gtk_entry_text_index_to_layout_index) gtk_entry_text_index_to_layout_index __attribute((alias("IA__gtk_entry_text_index_to_layout_index"), visibility("default")));

#undef gtk_entry_unset_invisible_char
extern __typeof (gtk_entry_unset_invisible_char) gtk_entry_unset_invisible_char __attribute((alias("IA__gtk_entry_unset_invisible_char"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_entry_append_text
extern __typeof (gtk_entry_append_text) gtk_entry_append_text __attribute((alias("IA__gtk_entry_append_text"), visibility("default")));

#undef gtk_entry_new_with_max_length
extern __typeof (gtk_entry_new_with_max_length) gtk_entry_new_with_max_length __attribute((alias("IA__gtk_entry_new_with_max_length"), visibility("default")));

#undef gtk_entry_prepend_text
extern __typeof (gtk_entry_prepend_text) gtk_entry_prepend_text __attribute((alias("IA__gtk_entry_prepend_text"), visibility("default")));

#undef gtk_entry_select_region
extern __typeof (gtk_entry_select_region) gtk_entry_select_region __attribute((alias("IA__gtk_entry_select_region"), visibility("default")));

#undef gtk_entry_set_editable
extern __typeof (gtk_entry_set_editable) gtk_entry_set_editable __attribute((alias("IA__gtk_entry_set_editable"), visibility("default")));

#undef gtk_entry_set_position
extern __typeof (gtk_entry_set_position) gtk_entry_set_position __attribute((alias("IA__gtk_entry_set_position"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_ENTRY_COMPLETION_H__)
#if IN_FILE(__GTK_ENTRY_COMPLETION_C__)
#undef gtk_entry_completion_complete
extern __typeof (gtk_entry_completion_complete) gtk_entry_completion_complete __attribute((alias("IA__gtk_entry_completion_complete"), visibility("default")));

#undef gtk_entry_completion_delete_action
extern __typeof (gtk_entry_completion_delete_action) gtk_entry_completion_delete_action __attribute((alias("IA__gtk_entry_completion_delete_action"), visibility("default")));

#undef gtk_entry_completion_get_entry
extern __typeof (gtk_entry_completion_get_entry) gtk_entry_completion_get_entry __attribute((alias("IA__gtk_entry_completion_get_entry"), visibility("default")));

#undef gtk_entry_completion_get_inline_completion
extern __typeof (gtk_entry_completion_get_inline_completion) gtk_entry_completion_get_inline_completion __attribute((alias("IA__gtk_entry_completion_get_inline_completion"), visibility("default")));

#undef gtk_entry_completion_get_inline_selection
extern __typeof (gtk_entry_completion_get_inline_selection) gtk_entry_completion_get_inline_selection __attribute((alias("IA__gtk_entry_completion_get_inline_selection"), visibility("default")));

#undef gtk_entry_completion_get_minimum_key_length
extern __typeof (gtk_entry_completion_get_minimum_key_length) gtk_entry_completion_get_minimum_key_length __attribute((alias("IA__gtk_entry_completion_get_minimum_key_length"), visibility("default")));

#undef gtk_entry_completion_get_completion_prefix
extern __typeof (gtk_entry_completion_get_completion_prefix) gtk_entry_completion_get_completion_prefix __attribute((alias("IA__gtk_entry_completion_get_completion_prefix"), visibility("default")));

#undef gtk_entry_completion_get_model
extern __typeof (gtk_entry_completion_get_model) gtk_entry_completion_get_model __attribute((alias("IA__gtk_entry_completion_get_model"), visibility("default")));

#undef gtk_entry_completion_get_popup_completion
extern __typeof (gtk_entry_completion_get_popup_completion) gtk_entry_completion_get_popup_completion __attribute((alias("IA__gtk_entry_completion_get_popup_completion"), visibility("default")));

#undef gtk_entry_completion_get_popup_set_width
extern __typeof (gtk_entry_completion_get_popup_set_width) gtk_entry_completion_get_popup_set_width __attribute((alias("IA__gtk_entry_completion_get_popup_set_width"), visibility("default")));

#undef gtk_entry_completion_get_popup_single_match
extern __typeof (gtk_entry_completion_get_popup_single_match) gtk_entry_completion_get_popup_single_match __attribute((alias("IA__gtk_entry_completion_get_popup_single_match"), visibility("default")));

#undef gtk_entry_completion_get_text_column
extern __typeof (gtk_entry_completion_get_text_column) gtk_entry_completion_get_text_column __attribute((alias("IA__gtk_entry_completion_get_text_column"), visibility("default")));

#undef gtk_entry_completion_get_type
extern __typeof (gtk_entry_completion_get_type) gtk_entry_completion_get_type __attribute((alias("IA__gtk_entry_completion_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_entry_completion_insert_action_markup
extern __typeof (gtk_entry_completion_insert_action_markup) gtk_entry_completion_insert_action_markup __attribute((alias("IA__gtk_entry_completion_insert_action_markup"), visibility("default")));

#undef gtk_entry_completion_insert_action_text
extern __typeof (gtk_entry_completion_insert_action_text) gtk_entry_completion_insert_action_text __attribute((alias("IA__gtk_entry_completion_insert_action_text"), visibility("default")));

#undef gtk_entry_completion_insert_prefix
extern __typeof (gtk_entry_completion_insert_prefix) gtk_entry_completion_insert_prefix __attribute((alias("IA__gtk_entry_completion_insert_prefix"), visibility("default")));

#undef gtk_entry_completion_new
extern __typeof (gtk_entry_completion_new) gtk_entry_completion_new __attribute((alias("IA__gtk_entry_completion_new"), visibility("default")));

#undef gtk_entry_completion_set_inline_completion
extern __typeof (gtk_entry_completion_set_inline_completion) gtk_entry_completion_set_inline_completion __attribute((alias("IA__gtk_entry_completion_set_inline_completion"), visibility("default")));

#undef gtk_entry_completion_set_inline_selection
extern __typeof (gtk_entry_completion_set_inline_selection) gtk_entry_completion_set_inline_selection __attribute((alias("IA__gtk_entry_completion_set_inline_selection"), visibility("default")));

#undef gtk_entry_completion_set_match_func
extern __typeof (gtk_entry_completion_set_match_func) gtk_entry_completion_set_match_func __attribute((alias("IA__gtk_entry_completion_set_match_func"), visibility("default")));

#undef gtk_entry_completion_set_minimum_key_length
extern __typeof (gtk_entry_completion_set_minimum_key_length) gtk_entry_completion_set_minimum_key_length __attribute((alias("IA__gtk_entry_completion_set_minimum_key_length"), visibility("default")));

#undef gtk_entry_completion_set_model
extern __typeof (gtk_entry_completion_set_model) gtk_entry_completion_set_model __attribute((alias("IA__gtk_entry_completion_set_model"), visibility("default")));

#undef gtk_entry_completion_set_popup_completion
extern __typeof (gtk_entry_completion_set_popup_completion) gtk_entry_completion_set_popup_completion __attribute((alias("IA__gtk_entry_completion_set_popup_completion"), visibility("default")));

#undef gtk_entry_completion_set_popup_set_width
extern __typeof (gtk_entry_completion_set_popup_set_width) gtk_entry_completion_set_popup_set_width __attribute((alias("IA__gtk_entry_completion_set_popup_set_width"), visibility("default")));

#undef gtk_entry_completion_set_popup_single_match
extern __typeof (gtk_entry_completion_set_popup_single_match) gtk_entry_completion_set_popup_single_match __attribute((alias("IA__gtk_entry_completion_set_popup_single_match"), visibility("default")));

#undef gtk_entry_completion_set_text_column
extern __typeof (gtk_entry_completion_set_text_column) gtk_entry_completion_set_text_column __attribute((alias("IA__gtk_entry_completion_set_text_column"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_ENTRY_BUFFER_H__)
#if IN_FILE(__GTK_ENTRY_BUFFER_C__)
#undef gtk_entry_buffer_get_type
extern __typeof (gtk_entry_buffer_get_type) gtk_entry_buffer_get_type __attribute((alias("IA__gtk_entry_buffer_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_entry_buffer_new
extern __typeof (gtk_entry_buffer_new) gtk_entry_buffer_new __attribute((alias("IA__gtk_entry_buffer_new"), visibility("default")));

#undef gtk_entry_buffer_get_bytes
extern __typeof (gtk_entry_buffer_get_bytes) gtk_entry_buffer_get_bytes __attribute((alias("IA__gtk_entry_buffer_get_bytes"), visibility("default")));

#undef gtk_entry_buffer_get_length
extern __typeof (gtk_entry_buffer_get_length) gtk_entry_buffer_get_length __attribute((alias("IA__gtk_entry_buffer_get_length"), visibility("default")));

#undef gtk_entry_buffer_get_text
extern __typeof (gtk_entry_buffer_get_text) gtk_entry_buffer_get_text __attribute((alias("IA__gtk_entry_buffer_get_text"), visibility("default")));

#undef gtk_entry_buffer_set_text
extern __typeof (gtk_entry_buffer_set_text) gtk_entry_buffer_set_text __attribute((alias("IA__gtk_entry_buffer_set_text"), visibility("default")));

#undef gtk_entry_buffer_set_max_length
extern __typeof (gtk_entry_buffer_set_max_length) gtk_entry_buffer_set_max_length __attribute((alias("IA__gtk_entry_buffer_set_max_length"), visibility("default")));

#undef gtk_entry_buffer_get_max_length
extern __typeof (gtk_entry_buffer_get_max_length) gtk_entry_buffer_get_max_length __attribute((alias("IA__gtk_entry_buffer_get_max_length"), visibility("default")));

#undef gtk_entry_buffer_insert_text
extern __typeof (gtk_entry_buffer_insert_text) gtk_entry_buffer_insert_text __attribute((alias("IA__gtk_entry_buffer_insert_text"), visibility("default")));

#undef gtk_entry_buffer_delete_text
extern __typeof (gtk_entry_buffer_delete_text) gtk_entry_buffer_delete_text __attribute((alias("IA__gtk_entry_buffer_delete_text"), visibility("default")));

#undef gtk_entry_buffer_emit_inserted_text
extern __typeof (gtk_entry_buffer_emit_inserted_text) gtk_entry_buffer_emit_inserted_text __attribute((alias("IA__gtk_entry_buffer_emit_inserted_text"), visibility("default")));

#undef gtk_entry_buffer_emit_deleted_text
extern __typeof (gtk_entry_buffer_emit_deleted_text) gtk_entry_buffer_emit_deleted_text __attribute((alias("IA__gtk_entry_buffer_emit_deleted_text"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_EVENT_BOX_H__)
#if IN_FILE(__GTK_EVENT_BOX_C__)
#undef gtk_event_box_get_above_child
extern __typeof (gtk_event_box_get_above_child) gtk_event_box_get_above_child __attribute((alias("IA__gtk_event_box_get_above_child"), visibility("default")));

#undef gtk_event_box_get_type
extern __typeof (gtk_event_box_get_type) gtk_event_box_get_type __attribute((alias("IA__gtk_event_box_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_event_box_get_visible_window
extern __typeof (gtk_event_box_get_visible_window) gtk_event_box_get_visible_window __attribute((alias("IA__gtk_event_box_get_visible_window"), visibility("default")));

#undef gtk_event_box_new
extern __typeof (gtk_event_box_new) gtk_event_box_new __attribute((alias("IA__gtk_event_box_new"), visibility("default")));

#undef gtk_event_box_set_above_child
extern __typeof (gtk_event_box_set_above_child) gtk_event_box_set_above_child __attribute((alias("IA__gtk_event_box_set_above_child"), visibility("default")));

#undef gtk_event_box_set_visible_window
extern __typeof (gtk_event_box_set_visible_window) gtk_event_box_set_visible_window __attribute((alias("IA__gtk_event_box_set_visible_window"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_EXPANDER_H__)
#if IN_FILE(__GTK_EXPANDER_C__)
#undef gtk_expander_get_expanded
extern __typeof (gtk_expander_get_expanded) gtk_expander_get_expanded __attribute((alias("IA__gtk_expander_get_expanded"), visibility("default")));

#undef gtk_expander_get_label
extern __typeof (gtk_expander_get_label) gtk_expander_get_label __attribute((alias("IA__gtk_expander_get_label"), visibility("default")));

#undef gtk_expander_get_label_fill
extern __typeof (gtk_expander_get_label_fill) gtk_expander_get_label_fill __attribute((alias("IA__gtk_expander_get_label_fill"), visibility("default")));

#undef gtk_expander_get_label_widget
extern __typeof (gtk_expander_get_label_widget) gtk_expander_get_label_widget __attribute((alias("IA__gtk_expander_get_label_widget"), visibility("default")));

#undef gtk_expander_get_spacing
extern __typeof (gtk_expander_get_spacing) gtk_expander_get_spacing __attribute((alias("IA__gtk_expander_get_spacing"), visibility("default")));

#undef gtk_expander_get_type
extern __typeof (gtk_expander_get_type) gtk_expander_get_type __attribute((alias("IA__gtk_expander_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_expander_get_use_markup
extern __typeof (gtk_expander_get_use_markup) gtk_expander_get_use_markup __attribute((alias("IA__gtk_expander_get_use_markup"), visibility("default")));

#undef gtk_expander_get_use_underline
extern __typeof (gtk_expander_get_use_underline) gtk_expander_get_use_underline __attribute((alias("IA__gtk_expander_get_use_underline"), visibility("default")));

#undef gtk_expander_new
extern __typeof (gtk_expander_new) gtk_expander_new __attribute((alias("IA__gtk_expander_new"), visibility("default")));

#undef gtk_expander_new_with_mnemonic
extern __typeof (gtk_expander_new_with_mnemonic) gtk_expander_new_with_mnemonic __attribute((alias("IA__gtk_expander_new_with_mnemonic"), visibility("default")));

#undef gtk_expander_set_expanded
extern __typeof (gtk_expander_set_expanded) gtk_expander_set_expanded __attribute((alias("IA__gtk_expander_set_expanded"), visibility("default")));

#undef gtk_expander_set_label
extern __typeof (gtk_expander_set_label) gtk_expander_set_label __attribute((alias("IA__gtk_expander_set_label"), visibility("default")));

#undef gtk_expander_set_label_fill
extern __typeof (gtk_expander_set_label_fill) gtk_expander_set_label_fill __attribute((alias("IA__gtk_expander_set_label_fill"), visibility("default")));

#undef gtk_expander_set_label_widget
extern __typeof (gtk_expander_set_label_widget) gtk_expander_set_label_widget __attribute((alias("IA__gtk_expander_set_label_widget"), visibility("default")));

#undef gtk_expander_set_spacing
extern __typeof (gtk_expander_set_spacing) gtk_expander_set_spacing __attribute((alias("IA__gtk_expander_set_spacing"), visibility("default")));

#undef gtk_expander_set_use_markup
extern __typeof (gtk_expander_set_use_markup) gtk_expander_set_use_markup __attribute((alias("IA__gtk_expander_set_use_markup"), visibility("default")));

#undef gtk_expander_set_use_underline
extern __typeof (gtk_expander_set_use_underline) gtk_expander_set_use_underline __attribute((alias("IA__gtk_expander_set_use_underline"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_FILE_CHOOSER_H__)
#if IN_FILE(__GTK_FILE_CHOOSER_C__)
#undef gtk_file_chooser_add_filter
extern __typeof (gtk_file_chooser_add_filter) gtk_file_chooser_add_filter __attribute((alias("IA__gtk_file_chooser_add_filter"), visibility("default")));

#ifndef _WIN64
#undef gtk_file_chooser_add_shortcut_folder
extern __typeof (gtk_file_chooser_add_shortcut_folder) gtk_file_chooser_add_shortcut_folder __attribute((alias("IA__gtk_file_chooser_add_shortcut_folder"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_file_chooser_add_shortcut_folder_utf8
extern __typeof (gtk_file_chooser_add_shortcut_folder_utf8) gtk_file_chooser_add_shortcut_folder_utf8 __attribute((alias("IA__gtk_file_chooser_add_shortcut_folder_utf8"), visibility("default")));

#endif
#undef gtk_file_chooser_add_shortcut_folder_uri
extern __typeof (gtk_file_chooser_add_shortcut_folder_uri) gtk_file_chooser_add_shortcut_folder_uri __attribute((alias("IA__gtk_file_chooser_add_shortcut_folder_uri"), visibility("default")));

#undef gtk_file_chooser_error_quark
extern __typeof (gtk_file_chooser_error_quark) gtk_file_chooser_error_quark __attribute((alias("IA__gtk_file_chooser_error_quark"), visibility("default")));

#undef gtk_file_chooser_get_action
extern __typeof (gtk_file_chooser_get_action) gtk_file_chooser_get_action __attribute((alias("IA__gtk_file_chooser_get_action"), visibility("default")));

#ifndef _WIN64
#undef gtk_file_chooser_get_current_folder
extern __typeof (gtk_file_chooser_get_current_folder) gtk_file_chooser_get_current_folder __attribute((alias("IA__gtk_file_chooser_get_current_folder"), visibility("default")));

#endif
#undef gtk_file_chooser_get_current_folder_file
extern __typeof (gtk_file_chooser_get_current_folder_file) gtk_file_chooser_get_current_folder_file __attribute((alias("IA__gtk_file_chooser_get_current_folder_file"), visibility("default")));

#ifdef G_OS_WIN32
#undef gtk_file_chooser_get_current_folder_utf8
extern __typeof (gtk_file_chooser_get_current_folder_utf8) gtk_file_chooser_get_current_folder_utf8 __attribute((alias("IA__gtk_file_chooser_get_current_folder_utf8"), visibility("default")));

#endif
#undef gtk_file_chooser_get_current_folder_uri
extern __typeof (gtk_file_chooser_get_current_folder_uri) gtk_file_chooser_get_current_folder_uri __attribute((alias("IA__gtk_file_chooser_get_current_folder_uri"), visibility("default")));

#undef gtk_file_chooser_get_extra_widget
extern __typeof (gtk_file_chooser_get_extra_widget) gtk_file_chooser_get_extra_widget __attribute((alias("IA__gtk_file_chooser_get_extra_widget"), visibility("default")));

#undef gtk_file_chooser_get_file
extern __typeof (gtk_file_chooser_get_file) gtk_file_chooser_get_file __attribute((alias("IA__gtk_file_chooser_get_file"), visibility("default")));

#ifndef _WIN64
#undef gtk_file_chooser_get_filename
extern __typeof (gtk_file_chooser_get_filename) gtk_file_chooser_get_filename __attribute((alias("IA__gtk_file_chooser_get_filename"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_file_chooser_get_filename_utf8
extern __typeof (gtk_file_chooser_get_filename_utf8) gtk_file_chooser_get_filename_utf8 __attribute((alias("IA__gtk_file_chooser_get_filename_utf8"), visibility("default")));

#endif
#ifndef _WIN64
#undef gtk_file_chooser_get_filenames
extern __typeof (gtk_file_chooser_get_filenames) gtk_file_chooser_get_filenames __attribute((alias("IA__gtk_file_chooser_get_filenames"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_file_chooser_get_filenames_utf8
extern __typeof (gtk_file_chooser_get_filenames_utf8) gtk_file_chooser_get_filenames_utf8 __attribute((alias("IA__gtk_file_chooser_get_filenames_utf8"), visibility("default")));

#endif
#undef gtk_file_chooser_get_files
extern __typeof (gtk_file_chooser_get_files) gtk_file_chooser_get_files __attribute((alias("IA__gtk_file_chooser_get_files"), visibility("default")));

#undef gtk_file_chooser_get_filter
extern __typeof (gtk_file_chooser_get_filter) gtk_file_chooser_get_filter __attribute((alias("IA__gtk_file_chooser_get_filter"), visibility("default")));

#undef gtk_file_chooser_get_local_only
extern __typeof (gtk_file_chooser_get_local_only) gtk_file_chooser_get_local_only __attribute((alias("IA__gtk_file_chooser_get_local_only"), visibility("default")));

#undef gtk_file_chooser_get_preview_file
extern __typeof (gtk_file_chooser_get_preview_file) gtk_file_chooser_get_preview_file __attribute((alias("IA__gtk_file_chooser_get_preview_file"), visibility("default")));

#ifndef _WIN64
#undef gtk_file_chooser_get_preview_filename
extern __typeof (gtk_file_chooser_get_preview_filename) gtk_file_chooser_get_preview_filename __attribute((alias("IA__gtk_file_chooser_get_preview_filename"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_file_chooser_get_preview_filename_utf8
extern __typeof (gtk_file_chooser_get_preview_filename_utf8) gtk_file_chooser_get_preview_filename_utf8 __attribute((alias("IA__gtk_file_chooser_get_preview_filename_utf8"), visibility("default")));

#endif
#undef gtk_file_chooser_get_preview_uri
extern __typeof (gtk_file_chooser_get_preview_uri) gtk_file_chooser_get_preview_uri __attribute((alias("IA__gtk_file_chooser_get_preview_uri"), visibility("default")));

#undef gtk_file_chooser_get_preview_widget
extern __typeof (gtk_file_chooser_get_preview_widget) gtk_file_chooser_get_preview_widget __attribute((alias("IA__gtk_file_chooser_get_preview_widget"), visibility("default")));

#undef gtk_file_chooser_get_preview_widget_active
extern __typeof (gtk_file_chooser_get_preview_widget_active) gtk_file_chooser_get_preview_widget_active __attribute((alias("IA__gtk_file_chooser_get_preview_widget_active"), visibility("default")));

#undef gtk_file_chooser_get_select_multiple
extern __typeof (gtk_file_chooser_get_select_multiple) gtk_file_chooser_get_select_multiple __attribute((alias("IA__gtk_file_chooser_get_select_multiple"), visibility("default")));

#undef gtk_file_chooser_get_show_hidden
extern __typeof (gtk_file_chooser_get_show_hidden) gtk_file_chooser_get_show_hidden __attribute((alias("IA__gtk_file_chooser_get_show_hidden"), visibility("default")));

#undef gtk_file_chooser_set_do_overwrite_confirmation
extern __typeof (gtk_file_chooser_set_do_overwrite_confirmation) gtk_file_chooser_set_do_overwrite_confirmation __attribute((alias("IA__gtk_file_chooser_set_do_overwrite_confirmation"), visibility("default")));

#undef gtk_file_chooser_get_do_overwrite_confirmation
extern __typeof (gtk_file_chooser_get_do_overwrite_confirmation) gtk_file_chooser_get_do_overwrite_confirmation __attribute((alias("IA__gtk_file_chooser_get_do_overwrite_confirmation"), visibility("default")));

#undef gtk_file_chooser_set_create_folders
extern __typeof (gtk_file_chooser_set_create_folders) gtk_file_chooser_set_create_folders __attribute((alias("IA__gtk_file_chooser_set_create_folders"), visibility("default")));

#undef gtk_file_chooser_get_create_folders
extern __typeof (gtk_file_chooser_get_create_folders) gtk_file_chooser_get_create_folders __attribute((alias("IA__gtk_file_chooser_get_create_folders"), visibility("default")));

#undef gtk_file_chooser_get_type
extern __typeof (gtk_file_chooser_get_type) gtk_file_chooser_get_type __attribute((alias("IA__gtk_file_chooser_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_file_chooser_get_uri
extern __typeof (gtk_file_chooser_get_uri) gtk_file_chooser_get_uri __attribute((alias("IA__gtk_file_chooser_get_uri"), visibility("default")));

#undef gtk_file_chooser_get_uris
extern __typeof (gtk_file_chooser_get_uris) gtk_file_chooser_get_uris __attribute((alias("IA__gtk_file_chooser_get_uris"), visibility("default")));

#undef gtk_file_chooser_get_use_preview_label
extern __typeof (gtk_file_chooser_get_use_preview_label) gtk_file_chooser_get_use_preview_label __attribute((alias("IA__gtk_file_chooser_get_use_preview_label"), visibility("default")));

#undef gtk_file_chooser_list_filters
extern __typeof (gtk_file_chooser_list_filters) gtk_file_chooser_list_filters __attribute((alias("IA__gtk_file_chooser_list_filters"), visibility("default")));

#ifndef _WIN64
#undef gtk_file_chooser_list_shortcut_folders
extern __typeof (gtk_file_chooser_list_shortcut_folders) gtk_file_chooser_list_shortcut_folders __attribute((alias("IA__gtk_file_chooser_list_shortcut_folders"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_file_chooser_list_shortcut_folders_utf8
extern __typeof (gtk_file_chooser_list_shortcut_folders_utf8) gtk_file_chooser_list_shortcut_folders_utf8 __attribute((alias("IA__gtk_file_chooser_list_shortcut_folders_utf8"), visibility("default")));

#endif
#undef gtk_file_chooser_list_shortcut_folder_uris
extern __typeof (gtk_file_chooser_list_shortcut_folder_uris) gtk_file_chooser_list_shortcut_folder_uris __attribute((alias("IA__gtk_file_chooser_list_shortcut_folder_uris"), visibility("default")));

#undef gtk_file_chooser_remove_filter
extern __typeof (gtk_file_chooser_remove_filter) gtk_file_chooser_remove_filter __attribute((alias("IA__gtk_file_chooser_remove_filter"), visibility("default")));

#ifndef _WIN64
#undef gtk_file_chooser_remove_shortcut_folder
extern __typeof (gtk_file_chooser_remove_shortcut_folder) gtk_file_chooser_remove_shortcut_folder __attribute((alias("IA__gtk_file_chooser_remove_shortcut_folder"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_file_chooser_remove_shortcut_folder_utf8
extern __typeof (gtk_file_chooser_remove_shortcut_folder_utf8) gtk_file_chooser_remove_shortcut_folder_utf8 __attribute((alias("IA__gtk_file_chooser_remove_shortcut_folder_utf8"), visibility("default")));

#endif
#undef gtk_file_chooser_remove_shortcut_folder_uri
extern __typeof (gtk_file_chooser_remove_shortcut_folder_uri) gtk_file_chooser_remove_shortcut_folder_uri __attribute((alias("IA__gtk_file_chooser_remove_shortcut_folder_uri"), visibility("default")));

#undef gtk_file_chooser_select_all
extern __typeof (gtk_file_chooser_select_all) gtk_file_chooser_select_all __attribute((alias("IA__gtk_file_chooser_select_all"), visibility("default")));

#undef gtk_file_chooser_select_file
extern __typeof (gtk_file_chooser_select_file) gtk_file_chooser_select_file __attribute((alias("IA__gtk_file_chooser_select_file"), visibility("default")));

#ifndef _WIN64
#undef gtk_file_chooser_select_filename
extern __typeof (gtk_file_chooser_select_filename) gtk_file_chooser_select_filename __attribute((alias("IA__gtk_file_chooser_select_filename"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_file_chooser_select_filename_utf8
extern __typeof (gtk_file_chooser_select_filename_utf8) gtk_file_chooser_select_filename_utf8 __attribute((alias("IA__gtk_file_chooser_select_filename_utf8"), visibility("default")));

#endif
#undef gtk_file_chooser_select_uri
extern __typeof (gtk_file_chooser_select_uri) gtk_file_chooser_select_uri __attribute((alias("IA__gtk_file_chooser_select_uri"), visibility("default")));

#undef gtk_file_chooser_set_action
extern __typeof (gtk_file_chooser_set_action) gtk_file_chooser_set_action __attribute((alias("IA__gtk_file_chooser_set_action"), visibility("default")));

#ifndef _WIN64
#undef gtk_file_chooser_set_current_folder
extern __typeof (gtk_file_chooser_set_current_folder) gtk_file_chooser_set_current_folder __attribute((alias("IA__gtk_file_chooser_set_current_folder"), visibility("default")));

#endif
#undef gtk_file_chooser_set_current_folder_file
extern __typeof (gtk_file_chooser_set_current_folder_file) gtk_file_chooser_set_current_folder_file __attribute((alias("IA__gtk_file_chooser_set_current_folder_file"), visibility("default")));

#ifdef G_OS_WIN32
#undef gtk_file_chooser_set_current_folder_utf8
extern __typeof (gtk_file_chooser_set_current_folder_utf8) gtk_file_chooser_set_current_folder_utf8 __attribute((alias("IA__gtk_file_chooser_set_current_folder_utf8"), visibility("default")));

#endif
#undef gtk_file_chooser_set_current_folder_uri
extern __typeof (gtk_file_chooser_set_current_folder_uri) gtk_file_chooser_set_current_folder_uri __attribute((alias("IA__gtk_file_chooser_set_current_folder_uri"), visibility("default")));

#undef gtk_file_chooser_set_current_name
extern __typeof (gtk_file_chooser_set_current_name) gtk_file_chooser_set_current_name __attribute((alias("IA__gtk_file_chooser_set_current_name"), visibility("default")));

#undef gtk_file_chooser_set_extra_widget
extern __typeof (gtk_file_chooser_set_extra_widget) gtk_file_chooser_set_extra_widget __attribute((alias("IA__gtk_file_chooser_set_extra_widget"), visibility("default")));

#undef gtk_file_chooser_set_file
extern __typeof (gtk_file_chooser_set_file) gtk_file_chooser_set_file __attribute((alias("IA__gtk_file_chooser_set_file"), visibility("default")));

#ifndef _WIN64
#undef gtk_file_chooser_set_filename
extern __typeof (gtk_file_chooser_set_filename) gtk_file_chooser_set_filename __attribute((alias("IA__gtk_file_chooser_set_filename"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_file_chooser_set_filename_utf8
extern __typeof (gtk_file_chooser_set_filename_utf8) gtk_file_chooser_set_filename_utf8 __attribute((alias("IA__gtk_file_chooser_set_filename_utf8"), visibility("default")));

#endif
#undef gtk_file_chooser_set_filter
extern __typeof (gtk_file_chooser_set_filter) gtk_file_chooser_set_filter __attribute((alias("IA__gtk_file_chooser_set_filter"), visibility("default")));

#undef gtk_file_chooser_set_local_only
extern __typeof (gtk_file_chooser_set_local_only) gtk_file_chooser_set_local_only __attribute((alias("IA__gtk_file_chooser_set_local_only"), visibility("default")));

#undef gtk_file_chooser_set_preview_widget
extern __typeof (gtk_file_chooser_set_preview_widget) gtk_file_chooser_set_preview_widget __attribute((alias("IA__gtk_file_chooser_set_preview_widget"), visibility("default")));

#undef gtk_file_chooser_set_preview_widget_active
extern __typeof (gtk_file_chooser_set_preview_widget_active) gtk_file_chooser_set_preview_widget_active __attribute((alias("IA__gtk_file_chooser_set_preview_widget_active"), visibility("default")));

#undef gtk_file_chooser_set_select_multiple
extern __typeof (gtk_file_chooser_set_select_multiple) gtk_file_chooser_set_select_multiple __attribute((alias("IA__gtk_file_chooser_set_select_multiple"), visibility("default")));

#undef gtk_file_chooser_set_show_hidden
extern __typeof (gtk_file_chooser_set_show_hidden) gtk_file_chooser_set_show_hidden __attribute((alias("IA__gtk_file_chooser_set_show_hidden"), visibility("default")));

#undef gtk_file_chooser_set_uri
extern __typeof (gtk_file_chooser_set_uri) gtk_file_chooser_set_uri __attribute((alias("IA__gtk_file_chooser_set_uri"), visibility("default")));

#undef gtk_file_chooser_set_use_preview_label
extern __typeof (gtk_file_chooser_set_use_preview_label) gtk_file_chooser_set_use_preview_label __attribute((alias("IA__gtk_file_chooser_set_use_preview_label"), visibility("default")));

#undef gtk_file_chooser_unselect_all
extern __typeof (gtk_file_chooser_unselect_all) gtk_file_chooser_unselect_all __attribute((alias("IA__gtk_file_chooser_unselect_all"), visibility("default")));

#undef gtk_file_chooser_unselect_file
extern __typeof (gtk_file_chooser_unselect_file) gtk_file_chooser_unselect_file __attribute((alias("IA__gtk_file_chooser_unselect_file"), visibility("default")));

#ifndef _WIN64
#undef gtk_file_chooser_unselect_filename
extern __typeof (gtk_file_chooser_unselect_filename) gtk_file_chooser_unselect_filename __attribute((alias("IA__gtk_file_chooser_unselect_filename"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_file_chooser_unselect_filename_utf8
extern __typeof (gtk_file_chooser_unselect_filename_utf8) gtk_file_chooser_unselect_filename_utf8 __attribute((alias("IA__gtk_file_chooser_unselect_filename_utf8"), visibility("default")));

#endif
#undef gtk_file_chooser_unselect_uri
extern __typeof (gtk_file_chooser_unselect_uri) gtk_file_chooser_unselect_uri __attribute((alias("IA__gtk_file_chooser_unselect_uri"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_FILE_CHOOSER_DIALOG_H__)
#if IN_FILE(__GTK_FILE_CHOOSER_DIALOG_C__)
#undef gtk_file_chooser_dialog_get_type
extern __typeof (gtk_file_chooser_dialog_get_type) gtk_file_chooser_dialog_get_type __attribute((alias("IA__gtk_file_chooser_dialog_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_file_chooser_dialog_new
extern __typeof (gtk_file_chooser_dialog_new) gtk_file_chooser_dialog_new __attribute((alias("IA__gtk_file_chooser_dialog_new"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_file_chooser_dialog_new_with_backend
extern __typeof (gtk_file_chooser_dialog_new_with_backend) gtk_file_chooser_dialog_new_with_backend __attribute((alias("IA__gtk_file_chooser_dialog_new_with_backend"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_FILE_CHOOSER_WIDGET_H__)
#if IN_FILE(__GTK_FILE_CHOOSER_WIDGET_C__)
#undef gtk_file_chooser_widget_get_type
extern __typeof (gtk_file_chooser_widget_get_type) gtk_file_chooser_widget_get_type __attribute((alias("IA__gtk_file_chooser_widget_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_file_chooser_widget_new
extern __typeof (gtk_file_chooser_widget_new) gtk_file_chooser_widget_new __attribute((alias("IA__gtk_file_chooser_widget_new"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_file_chooser_widget_new_with_backend
extern __typeof (gtk_file_chooser_widget_new_with_backend) gtk_file_chooser_widget_new_with_backend __attribute((alias("IA__gtk_file_chooser_widget_new_with_backend"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_FILE_CHOOSER_BUTTON_H__)
#if IN_FILE(__GTK_FILE_CHOOSER_BUTTON_C__)
#undef gtk_file_chooser_button_get_title
extern __typeof (gtk_file_chooser_button_get_title) gtk_file_chooser_button_get_title __attribute((alias("IA__gtk_file_chooser_button_get_title"), visibility("default")));

#undef gtk_file_chooser_button_get_type
extern __typeof (gtk_file_chooser_button_get_type) gtk_file_chooser_button_get_type __attribute((alias("IA__gtk_file_chooser_button_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_file_chooser_button_get_width_chars
extern __typeof (gtk_file_chooser_button_get_width_chars) gtk_file_chooser_button_get_width_chars __attribute((alias("IA__gtk_file_chooser_button_get_width_chars"), visibility("default")));

#undef gtk_file_chooser_button_new
extern __typeof (gtk_file_chooser_button_new) gtk_file_chooser_button_new __attribute((alias("IA__gtk_file_chooser_button_new"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_file_chooser_button_new_with_backend
extern __typeof (gtk_file_chooser_button_new_with_backend) gtk_file_chooser_button_new_with_backend __attribute((alias("IA__gtk_file_chooser_button_new_with_backend"), visibility("default")));

#endif
#undef gtk_file_chooser_button_new_with_dialog
extern __typeof (gtk_file_chooser_button_new_with_dialog) gtk_file_chooser_button_new_with_dialog __attribute((alias("IA__gtk_file_chooser_button_new_with_dialog"), visibility("default")));

#undef gtk_file_chooser_button_set_title
extern __typeof (gtk_file_chooser_button_set_title) gtk_file_chooser_button_set_title __attribute((alias("IA__gtk_file_chooser_button_set_title"), visibility("default")));

#undef gtk_file_chooser_button_set_width_chars
extern __typeof (gtk_file_chooser_button_set_width_chars) gtk_file_chooser_button_set_width_chars __attribute((alias("IA__gtk_file_chooser_button_set_width_chars"), visibility("default")));

#undef gtk_file_chooser_button_get_focus_on_click
extern __typeof (gtk_file_chooser_button_get_focus_on_click) gtk_file_chooser_button_get_focus_on_click __attribute((alias("IA__gtk_file_chooser_button_get_focus_on_click"), visibility("default")));

#undef gtk_file_chooser_button_set_focus_on_click
extern __typeof (gtk_file_chooser_button_set_focus_on_click) gtk_file_chooser_button_set_focus_on_click __attribute((alias("IA__gtk_file_chooser_button_set_focus_on_click"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_FILE_FILTER_H__)
#if IN_FILE(__GTK_FILE_FILTER_C__)
#undef gtk_file_filter_add_custom
extern __typeof (gtk_file_filter_add_custom) gtk_file_filter_add_custom __attribute((alias("IA__gtk_file_filter_add_custom"), visibility("default")));

#undef gtk_file_filter_add_mime_type
extern __typeof (gtk_file_filter_add_mime_type) gtk_file_filter_add_mime_type __attribute((alias("IA__gtk_file_filter_add_mime_type"), visibility("default")));

#undef gtk_file_filter_add_pattern
extern __typeof (gtk_file_filter_add_pattern) gtk_file_filter_add_pattern __attribute((alias("IA__gtk_file_filter_add_pattern"), visibility("default")));

#undef gtk_file_filter_add_pixbuf_formats
extern __typeof (gtk_file_filter_add_pixbuf_formats) gtk_file_filter_add_pixbuf_formats __attribute((alias("IA__gtk_file_filter_add_pixbuf_formats"), visibility("default")));

#undef gtk_file_filter_filter
extern __typeof (gtk_file_filter_filter) gtk_file_filter_filter __attribute((alias("IA__gtk_file_filter_filter"), visibility("default")));

#undef gtk_file_filter_get_name
extern __typeof (gtk_file_filter_get_name) gtk_file_filter_get_name __attribute((alias("IA__gtk_file_filter_get_name"), visibility("default")));

#undef gtk_file_filter_get_needed
extern __typeof (gtk_file_filter_get_needed) gtk_file_filter_get_needed __attribute((alias("IA__gtk_file_filter_get_needed"), visibility("default")));

#undef gtk_file_filter_get_type
extern __typeof (gtk_file_filter_get_type) gtk_file_filter_get_type __attribute((alias("IA__gtk_file_filter_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_file_filter_new
extern __typeof (gtk_file_filter_new) gtk_file_filter_new __attribute((alias("IA__gtk_file_filter_new"), visibility("default")));

#undef gtk_file_filter_set_name
extern __typeof (gtk_file_filter_set_name) gtk_file_filter_set_name __attribute((alias("IA__gtk_file_filter_set_name"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_FILESEL_H__)
#if IN_FILE(__GTK_FILESEL_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_file_selection_complete
extern __typeof (gtk_file_selection_complete) gtk_file_selection_complete __attribute((alias("IA__gtk_file_selection_complete"), visibility("default")));

#ifndef _WIN64
#undef gtk_file_selection_get_filename
extern __typeof (gtk_file_selection_get_filename) gtk_file_selection_get_filename __attribute((alias("IA__gtk_file_selection_get_filename"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_file_selection_get_filename_utf8
extern __typeof (gtk_file_selection_get_filename_utf8) gtk_file_selection_get_filename_utf8 __attribute((alias("IA__gtk_file_selection_get_filename_utf8"), visibility("default")));

#endif
#ifndef _WIN64
#undef gtk_file_selection_get_selections
extern __typeof (gtk_file_selection_get_selections) gtk_file_selection_get_selections __attribute((alias("IA__gtk_file_selection_get_selections"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_file_selection_get_selections_utf8
extern __typeof (gtk_file_selection_get_selections_utf8) gtk_file_selection_get_selections_utf8 __attribute((alias("IA__gtk_file_selection_get_selections_utf8"), visibility("default")));

#endif
#undef gtk_file_selection_get_select_multiple
extern __typeof (gtk_file_selection_get_select_multiple) gtk_file_selection_get_select_multiple __attribute((alias("IA__gtk_file_selection_get_select_multiple"), visibility("default")));

#undef gtk_file_selection_get_type
extern __typeof (gtk_file_selection_get_type) gtk_file_selection_get_type __attribute((alias("IA__gtk_file_selection_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_file_selection_hide_fileop_buttons
extern __typeof (gtk_file_selection_hide_fileop_buttons) gtk_file_selection_hide_fileop_buttons __attribute((alias("IA__gtk_file_selection_hide_fileop_buttons"), visibility("default")));

#undef gtk_file_selection_new
extern __typeof (gtk_file_selection_new) gtk_file_selection_new __attribute((alias("IA__gtk_file_selection_new"), visibility("default")));

#ifndef _WIN64
#undef gtk_file_selection_set_filename
extern __typeof (gtk_file_selection_set_filename) gtk_file_selection_set_filename __attribute((alias("IA__gtk_file_selection_set_filename"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_file_selection_set_filename_utf8
extern __typeof (gtk_file_selection_set_filename_utf8) gtk_file_selection_set_filename_utf8 __attribute((alias("IA__gtk_file_selection_set_filename_utf8"), visibility("default")));

#endif
#undef gtk_file_selection_set_select_multiple
extern __typeof (gtk_file_selection_set_select_multiple) gtk_file_selection_set_select_multiple __attribute((alias("IA__gtk_file_selection_set_select_multiple"), visibility("default")));

#undef gtk_file_selection_show_fileop_buttons
extern __typeof (gtk_file_selection_show_fileop_buttons) gtk_file_selection_show_fileop_buttons __attribute((alias("IA__gtk_file_selection_show_fileop_buttons"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_FIXED_H__)
#if IN_FILE(__GTK_FIXED_C__)
#undef gtk_fixed_get_type
extern __typeof (gtk_fixed_get_type) gtk_fixed_get_type __attribute((alias("IA__gtk_fixed_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_fixed_move
extern __typeof (gtk_fixed_move) gtk_fixed_move __attribute((alias("IA__gtk_fixed_move"), visibility("default")));

#undef gtk_fixed_new
extern __typeof (gtk_fixed_new) gtk_fixed_new __attribute((alias("IA__gtk_fixed_new"), visibility("default")));

#undef gtk_fixed_put
extern __typeof (gtk_fixed_put) gtk_fixed_put __attribute((alias("IA__gtk_fixed_put"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_fixed_set_has_window
extern __typeof (gtk_fixed_set_has_window) gtk_fixed_set_has_window __attribute((alias("IA__gtk_fixed_set_has_window"), visibility("default")));

#undef gtk_fixed_get_has_window
extern __typeof (gtk_fixed_get_has_window) gtk_fixed_get_has_window __attribute((alias("IA__gtk_fixed_get_has_window"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_FONT_BUTTON_H__)
#if IN_FILE(__GTK_FONT_BUTTON_C__)
#undef gtk_font_button_get_font_name
extern __typeof (gtk_font_button_get_font_name) gtk_font_button_get_font_name __attribute((alias("IA__gtk_font_button_get_font_name"), visibility("default")));

#undef gtk_font_button_get_show_size
extern __typeof (gtk_font_button_get_show_size) gtk_font_button_get_show_size __attribute((alias("IA__gtk_font_button_get_show_size"), visibility("default")));

#undef gtk_font_button_get_show_style
extern __typeof (gtk_font_button_get_show_style) gtk_font_button_get_show_style __attribute((alias("IA__gtk_font_button_get_show_style"), visibility("default")));

#undef gtk_font_button_get_title
extern __typeof (gtk_font_button_get_title) gtk_font_button_get_title __attribute((alias("IA__gtk_font_button_get_title"), visibility("default")));

#undef gtk_font_button_get_type
extern __typeof (gtk_font_button_get_type) gtk_font_button_get_type __attribute((alias("IA__gtk_font_button_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_font_button_get_use_font
extern __typeof (gtk_font_button_get_use_font) gtk_font_button_get_use_font __attribute((alias("IA__gtk_font_button_get_use_font"), visibility("default")));

#undef gtk_font_button_get_use_size
extern __typeof (gtk_font_button_get_use_size) gtk_font_button_get_use_size __attribute((alias("IA__gtk_font_button_get_use_size"), visibility("default")));

#undef gtk_font_button_new
extern __typeof (gtk_font_button_new) gtk_font_button_new __attribute((alias("IA__gtk_font_button_new"), visibility("default")));

#undef gtk_font_button_new_with_font
extern __typeof (gtk_font_button_new_with_font) gtk_font_button_new_with_font __attribute((alias("IA__gtk_font_button_new_with_font"), visibility("default")));

#undef gtk_font_button_set_font_name
extern __typeof (gtk_font_button_set_font_name) gtk_font_button_set_font_name __attribute((alias("IA__gtk_font_button_set_font_name"), visibility("default")));

#undef gtk_font_button_set_show_size
extern __typeof (gtk_font_button_set_show_size) gtk_font_button_set_show_size __attribute((alias("IA__gtk_font_button_set_show_size"), visibility("default")));

#undef gtk_font_button_set_show_style
extern __typeof (gtk_font_button_set_show_style) gtk_font_button_set_show_style __attribute((alias("IA__gtk_font_button_set_show_style"), visibility("default")));

#undef gtk_font_button_set_title
extern __typeof (gtk_font_button_set_title) gtk_font_button_set_title __attribute((alias("IA__gtk_font_button_set_title"), visibility("default")));

#undef gtk_font_button_set_use_font
extern __typeof (gtk_font_button_set_use_font) gtk_font_button_set_use_font __attribute((alias("IA__gtk_font_button_set_use_font"), visibility("default")));

#undef gtk_font_button_set_use_size
extern __typeof (gtk_font_button_set_use_size) gtk_font_button_set_use_size __attribute((alias("IA__gtk_font_button_set_use_size"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_FONTSEL_H__)
#if IN_FILE(__GTK_FONTSEL_C__)
#undef gtk_font_selection_dialog_get_cancel_button
extern __typeof (gtk_font_selection_dialog_get_cancel_button) gtk_font_selection_dialog_get_cancel_button __attribute((alias("IA__gtk_font_selection_dialog_get_cancel_button"), visibility("default")));

#undef gtk_font_selection_dialog_get_font_name
extern __typeof (gtk_font_selection_dialog_get_font_name) gtk_font_selection_dialog_get_font_name __attribute((alias("IA__gtk_font_selection_dialog_get_font_name"), visibility("default")));

#undef gtk_font_selection_dialog_get_font_selection
extern __typeof (gtk_font_selection_dialog_get_font_selection) gtk_font_selection_dialog_get_font_selection __attribute((alias("IA__gtk_font_selection_dialog_get_font_selection"), visibility("default")));

#undef gtk_font_selection_dialog_get_ok_button
extern __typeof (gtk_font_selection_dialog_get_ok_button) gtk_font_selection_dialog_get_ok_button __attribute((alias("IA__gtk_font_selection_dialog_get_ok_button"), visibility("default")));

#undef gtk_font_selection_dialog_get_preview_text
extern __typeof (gtk_font_selection_dialog_get_preview_text) gtk_font_selection_dialog_get_preview_text __attribute((alias("IA__gtk_font_selection_dialog_get_preview_text"), visibility("default")));

#undef gtk_font_selection_dialog_get_type
extern __typeof (gtk_font_selection_dialog_get_type) gtk_font_selection_dialog_get_type __attribute((alias("IA__gtk_font_selection_dialog_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_font_selection_dialog_new
extern __typeof (gtk_font_selection_dialog_new) gtk_font_selection_dialog_new __attribute((alias("IA__gtk_font_selection_dialog_new"), visibility("default")));

#undef gtk_font_selection_dialog_set_font_name
extern __typeof (gtk_font_selection_dialog_set_font_name) gtk_font_selection_dialog_set_font_name __attribute((alias("IA__gtk_font_selection_dialog_set_font_name"), visibility("default")));

#undef gtk_font_selection_dialog_set_preview_text
extern __typeof (gtk_font_selection_dialog_set_preview_text) gtk_font_selection_dialog_set_preview_text __attribute((alias("IA__gtk_font_selection_dialog_set_preview_text"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_font_selection_dialog_get_apply_button
extern __typeof (gtk_font_selection_dialog_get_apply_button) gtk_font_selection_dialog_get_apply_button __attribute((alias("IA__gtk_font_selection_dialog_get_apply_button"), visibility("default")));

#undef gtk_font_selection_get_font
extern __typeof (gtk_font_selection_get_font) gtk_font_selection_get_font __attribute((alias("IA__gtk_font_selection_get_font"), visibility("default")));

#undef gtk_font_selection_dialog_get_font
extern __typeof (gtk_font_selection_dialog_get_font) gtk_font_selection_dialog_get_font __attribute((alias("IA__gtk_font_selection_dialog_get_font"), visibility("default")));

#endif
#undef gtk_font_selection_get_face
extern __typeof (gtk_font_selection_get_face) gtk_font_selection_get_face __attribute((alias("IA__gtk_font_selection_get_face"), visibility("default")));

#undef gtk_font_selection_get_face_list
extern __typeof (gtk_font_selection_get_face_list) gtk_font_selection_get_face_list __attribute((alias("IA__gtk_font_selection_get_face_list"), visibility("default")));

#undef gtk_font_selection_get_family
extern __typeof (gtk_font_selection_get_family) gtk_font_selection_get_family __attribute((alias("IA__gtk_font_selection_get_family"), visibility("default")));

#undef gtk_font_selection_get_family_list
extern __typeof (gtk_font_selection_get_family_list) gtk_font_selection_get_family_list __attribute((alias("IA__gtk_font_selection_get_family_list"), visibility("default")));

#undef gtk_font_selection_get_font_name
extern __typeof (gtk_font_selection_get_font_name) gtk_font_selection_get_font_name __attribute((alias("IA__gtk_font_selection_get_font_name"), visibility("default")));

#undef gtk_font_selection_get_preview_entry
extern __typeof (gtk_font_selection_get_preview_entry) gtk_font_selection_get_preview_entry __attribute((alias("IA__gtk_font_selection_get_preview_entry"), visibility("default")));

#undef gtk_font_selection_get_preview_text
extern __typeof (gtk_font_selection_get_preview_text) gtk_font_selection_get_preview_text __attribute((alias("IA__gtk_font_selection_get_preview_text"), visibility("default")));

#undef gtk_font_selection_get_size
extern __typeof (gtk_font_selection_get_size) gtk_font_selection_get_size __attribute((alias("IA__gtk_font_selection_get_size"), visibility("default")));

#undef gtk_font_selection_get_size_entry
extern __typeof (gtk_font_selection_get_size_entry) gtk_font_selection_get_size_entry __attribute((alias("IA__gtk_font_selection_get_size_entry"), visibility("default")));

#undef gtk_font_selection_get_size_list
extern __typeof (gtk_font_selection_get_size_list) gtk_font_selection_get_size_list __attribute((alias("IA__gtk_font_selection_get_size_list"), visibility("default")));

#undef gtk_font_selection_get_type
extern __typeof (gtk_font_selection_get_type) gtk_font_selection_get_type __attribute((alias("IA__gtk_font_selection_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_font_selection_new
extern __typeof (gtk_font_selection_new) gtk_font_selection_new __attribute((alias("IA__gtk_font_selection_new"), visibility("default")));

#undef gtk_font_selection_set_font_name
extern __typeof (gtk_font_selection_set_font_name) gtk_font_selection_set_font_name __attribute((alias("IA__gtk_font_selection_set_font_name"), visibility("default")));

#undef gtk_font_selection_set_preview_text
extern __typeof (gtk_font_selection_set_preview_text) gtk_font_selection_set_preview_text __attribute((alias("IA__gtk_font_selection_set_preview_text"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_FRAME_H__)
#if IN_FILE(__GTK_FRAME_C__)
#undef gtk_frame_get_label
extern __typeof (gtk_frame_get_label) gtk_frame_get_label __attribute((alias("IA__gtk_frame_get_label"), visibility("default")));

#undef gtk_frame_get_label_align
extern __typeof (gtk_frame_get_label_align) gtk_frame_get_label_align __attribute((alias("IA__gtk_frame_get_label_align"), visibility("default")));

#undef gtk_frame_get_label_widget
extern __typeof (gtk_frame_get_label_widget) gtk_frame_get_label_widget __attribute((alias("IA__gtk_frame_get_label_widget"), visibility("default")));

#undef gtk_frame_get_shadow_type
extern __typeof (gtk_frame_get_shadow_type) gtk_frame_get_shadow_type __attribute((alias("IA__gtk_frame_get_shadow_type"), visibility("default")));

#undef gtk_frame_get_type
extern __typeof (gtk_frame_get_type) gtk_frame_get_type __attribute((alias("IA__gtk_frame_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_frame_new
extern __typeof (gtk_frame_new) gtk_frame_new __attribute((alias("IA__gtk_frame_new"), visibility("default")));

#undef gtk_frame_set_label
extern __typeof (gtk_frame_set_label) gtk_frame_set_label __attribute((alias("IA__gtk_frame_set_label"), visibility("default")));

#undef gtk_frame_set_label_align
extern __typeof (gtk_frame_set_label_align) gtk_frame_set_label_align __attribute((alias("IA__gtk_frame_set_label_align"), visibility("default")));

#undef gtk_frame_set_label_widget
extern __typeof (gtk_frame_set_label_widget) gtk_frame_set_label_widget __attribute((alias("IA__gtk_frame_set_label_widget"), visibility("default")));

#undef gtk_frame_set_shadow_type
extern __typeof (gtk_frame_set_shadow_type) gtk_frame_set_shadow_type __attribute((alias("IA__gtk_frame_set_shadow_type"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_GAMMA_CURVE_H__)
#if IN_FILE(__GTK_GAMMA_CURVE_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_gamma_curve_get_type
extern __typeof (gtk_gamma_curve_get_type) gtk_gamma_curve_get_type __attribute((alias("IA__gtk_gamma_curve_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_gamma_curve_new
extern __typeof (gtk_gamma_curve_new) gtk_gamma_curve_new __attribute((alias("IA__gtk_gamma_curve_new"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_GC_H__)
#if IN_FILE(__GTK_GC_C__)
#undef gtk_gc_get
extern __typeof (gtk_gc_get) gtk_gc_get __attribute((alias("IA__gtk_gc_get"), visibility("default")));

#undef gtk_gc_release
extern __typeof (gtk_gc_release) gtk_gc_release __attribute((alias("IA__gtk_gc_release"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_HANDLE_BOX_H__)
#if IN_FILE(__GTK_HANDLE_BOX_C__)
#undef gtk_handle_box_get_handle_position
extern __typeof (gtk_handle_box_get_handle_position) gtk_handle_box_get_handle_position __attribute((alias("IA__gtk_handle_box_get_handle_position"), visibility("default")));

#undef gtk_handle_box_get_shadow_type
extern __typeof (gtk_handle_box_get_shadow_type) gtk_handle_box_get_shadow_type __attribute((alias("IA__gtk_handle_box_get_shadow_type"), visibility("default")));

#undef gtk_handle_box_get_snap_edge
extern __typeof (gtk_handle_box_get_snap_edge) gtk_handle_box_get_snap_edge __attribute((alias("IA__gtk_handle_box_get_snap_edge"), visibility("default")));

#undef gtk_handle_box_get_child_detached
extern __typeof (gtk_handle_box_get_child_detached) gtk_handle_box_get_child_detached __attribute((alias("IA__gtk_handle_box_get_child_detached"), visibility("default")));

#undef gtk_handle_box_get_type
extern __typeof (gtk_handle_box_get_type) gtk_handle_box_get_type __attribute((alias("IA__gtk_handle_box_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_handle_box_new
extern __typeof (gtk_handle_box_new) gtk_handle_box_new __attribute((alias("IA__gtk_handle_box_new"), visibility("default")));

#undef gtk_handle_box_set_handle_position
extern __typeof (gtk_handle_box_set_handle_position) gtk_handle_box_set_handle_position __attribute((alias("IA__gtk_handle_box_set_handle_position"), visibility("default")));

#undef gtk_handle_box_set_shadow_type
extern __typeof (gtk_handle_box_set_shadow_type) gtk_handle_box_set_shadow_type __attribute((alias("IA__gtk_handle_box_set_shadow_type"), visibility("default")));

#undef gtk_handle_box_set_snap_edge
extern __typeof (gtk_handle_box_set_snap_edge) gtk_handle_box_set_snap_edge __attribute((alias("IA__gtk_handle_box_set_snap_edge"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_HBOX_H__)
#if IN_FILE(__GTK_HBOX_C__)
#undef gtk_hbox_get_type
extern __typeof (gtk_hbox_get_type) gtk_hbox_get_type __attribute((alias("IA__gtk_hbox_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_hbox_new
extern __typeof (gtk_hbox_new) gtk_hbox_new __attribute((alias("IA__gtk_hbox_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_HBUTTON_BOX_H__)
#if IN_FILE(__GTK_HBUTTON_BOX_C__)
#undef gtk_hbutton_box_new
extern __typeof (gtk_hbutton_box_new) gtk_hbutton_box_new __attribute((alias("IA__gtk_hbutton_box_new"), visibility("default")));

#undef gtk_hbutton_box_get_type
extern __typeof (gtk_hbutton_box_get_type) gtk_hbutton_box_get_type __attribute((alias("IA__gtk_hbutton_box_get_type"), visibility("default"))) G_GNUC_CONST;

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_hbutton_box_get_layout_default
extern __typeof (gtk_hbutton_box_get_layout_default) gtk_hbutton_box_get_layout_default __attribute((alias("IA__gtk_hbutton_box_get_layout_default"), visibility("default")));

#undef gtk_hbutton_box_get_spacing_default
extern __typeof (gtk_hbutton_box_get_spacing_default) gtk_hbutton_box_get_spacing_default __attribute((alias("IA__gtk_hbutton_box_get_spacing_default"), visibility("default")));

#undef gtk_hbutton_box_set_layout_default
extern __typeof (gtk_hbutton_box_set_layout_default) gtk_hbutton_box_set_layout_default __attribute((alias("IA__gtk_hbutton_box_set_layout_default"), visibility("default")));

#undef gtk_hbutton_box_set_spacing_default
extern __typeof (gtk_hbutton_box_set_spacing_default) gtk_hbutton_box_set_spacing_default __attribute((alias("IA__gtk_hbutton_box_set_spacing_default"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_HPANED_H__)
#if IN_FILE(__GTK_HPANED_C__)
#undef gtk_hpaned_get_type
extern __typeof (gtk_hpaned_get_type) gtk_hpaned_get_type __attribute((alias("IA__gtk_hpaned_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_hpaned_new
extern __typeof (gtk_hpaned_new) gtk_hpaned_new __attribute((alias("IA__gtk_hpaned_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_HRULER_H__)
#if IN_FILE(__GTK_HRULER_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_hruler_get_type
extern __typeof (gtk_hruler_get_type) gtk_hruler_get_type __attribute((alias("IA__gtk_hruler_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_hruler_new
extern __typeof (gtk_hruler_new) gtk_hruler_new __attribute((alias("IA__gtk_hruler_new"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_HSCALE_H__)
#if IN_FILE(__GTK_HSCALE_C__)
#undef gtk_hscale_get_type
extern __typeof (gtk_hscale_get_type) gtk_hscale_get_type __attribute((alias("IA__gtk_hscale_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_hscale_new
extern __typeof (gtk_hscale_new) gtk_hscale_new __attribute((alias("IA__gtk_hscale_new"), visibility("default")));

#undef gtk_hscale_new_with_range
extern __typeof (gtk_hscale_new_with_range) gtk_hscale_new_with_range __attribute((alias("IA__gtk_hscale_new_with_range"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_HSCROLLBAR_H__)
#if IN_FILE(__GTK_HSCROLLBAR_C__)
#undef gtk_hscrollbar_get_type
extern __typeof (gtk_hscrollbar_get_type) gtk_hscrollbar_get_type __attribute((alias("IA__gtk_hscrollbar_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_hscrollbar_new
extern __typeof (gtk_hscrollbar_new) gtk_hscrollbar_new __attribute((alias("IA__gtk_hscrollbar_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_HSEPARATOR_H__)
#if IN_FILE(__GTK_HSEPARATOR_C__)
#undef gtk_hseparator_get_type
extern __typeof (gtk_hseparator_get_type) gtk_hseparator_get_type __attribute((alias("IA__gtk_hseparator_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_hseparator_new
extern __typeof (gtk_hseparator_new) gtk_hseparator_new __attribute((alias("IA__gtk_hseparator_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_HSV_H__)
#if IN_FILE(__GTK_HSV_C__)
#undef gtk_hsv_get_color
extern __typeof (gtk_hsv_get_color) gtk_hsv_get_color __attribute((alias("IA__gtk_hsv_get_color"), visibility("default")));

#undef gtk_hsv_get_metrics
extern __typeof (gtk_hsv_get_metrics) gtk_hsv_get_metrics __attribute((alias("IA__gtk_hsv_get_metrics"), visibility("default")));

#undef gtk_hsv_get_type
extern __typeof (gtk_hsv_get_type) gtk_hsv_get_type __attribute((alias("IA__gtk_hsv_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_hsv_is_adjusting
extern __typeof (gtk_hsv_is_adjusting) gtk_hsv_is_adjusting __attribute((alias("IA__gtk_hsv_is_adjusting"), visibility("default")));

#undef gtk_hsv_new
extern __typeof (gtk_hsv_new) gtk_hsv_new __attribute((alias("IA__gtk_hsv_new"), visibility("default")));

#undef gtk_hsv_set_color
extern __typeof (gtk_hsv_set_color) gtk_hsv_set_color __attribute((alias("IA__gtk_hsv_set_color"), visibility("default")));

#undef gtk_hsv_set_metrics
extern __typeof (gtk_hsv_set_metrics) gtk_hsv_set_metrics __attribute((alias("IA__gtk_hsv_set_metrics"), visibility("default")));

#undef gtk_hsv_to_rgb
extern __typeof (gtk_hsv_to_rgb) gtk_hsv_to_rgb __attribute((alias("IA__gtk_hsv_to_rgb"), visibility("default")));

#undef gtk_rgb_to_hsv
extern __typeof (gtk_rgb_to_hsv) gtk_rgb_to_hsv __attribute((alias("IA__gtk_rgb_to_hsv"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_ICON_FACTORY_H__)
#if IN_FILE(__GTK_ICON_FACTORY_C__)
#undef gtk_icon_factory_add
extern __typeof (gtk_icon_factory_add) gtk_icon_factory_add __attribute((alias("IA__gtk_icon_factory_add"), visibility("default")));

#undef gtk_icon_factory_add_default
extern __typeof (gtk_icon_factory_add_default) gtk_icon_factory_add_default __attribute((alias("IA__gtk_icon_factory_add_default"), visibility("default")));

#undef gtk_icon_factory_get_type
extern __typeof (gtk_icon_factory_get_type) gtk_icon_factory_get_type __attribute((alias("IA__gtk_icon_factory_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_icon_factory_lookup
extern __typeof (gtk_icon_factory_lookup) gtk_icon_factory_lookup __attribute((alias("IA__gtk_icon_factory_lookup"), visibility("default")));

#undef gtk_icon_factory_lookup_default
extern __typeof (gtk_icon_factory_lookup_default) gtk_icon_factory_lookup_default __attribute((alias("IA__gtk_icon_factory_lookup_default"), visibility("default")));

#undef gtk_icon_factory_new
extern __typeof (gtk_icon_factory_new) gtk_icon_factory_new __attribute((alias("IA__gtk_icon_factory_new"), visibility("default")));

#undef gtk_icon_factory_remove_default
extern __typeof (gtk_icon_factory_remove_default) gtk_icon_factory_remove_default __attribute((alias("IA__gtk_icon_factory_remove_default"), visibility("default")));

#undef gtk_icon_size_from_name
extern __typeof (gtk_icon_size_from_name) gtk_icon_size_from_name __attribute((alias("IA__gtk_icon_size_from_name"), visibility("default")));

#undef gtk_icon_size_get_name
extern __typeof (gtk_icon_size_get_name) gtk_icon_size_get_name __attribute((alias("IA__gtk_icon_size_get_name"), visibility("default")));

#undef gtk_icon_size_lookup
extern __typeof (gtk_icon_size_lookup) gtk_icon_size_lookup __attribute((alias("IA__gtk_icon_size_lookup"), visibility("default")));

#undef gtk_icon_size_lookup_for_settings
extern __typeof (gtk_icon_size_lookup_for_settings) gtk_icon_size_lookup_for_settings __attribute((alias("IA__gtk_icon_size_lookup_for_settings"), visibility("default")));

#undef gtk_icon_size_register
extern __typeof (gtk_icon_size_register) gtk_icon_size_register __attribute((alias("IA__gtk_icon_size_register"), visibility("default")));

#undef gtk_icon_size_register_alias
extern __typeof (gtk_icon_size_register_alias) gtk_icon_size_register_alias __attribute((alias("IA__gtk_icon_size_register_alias"), visibility("default")));

#undef gtk_icon_set_add_source
extern __typeof (gtk_icon_set_add_source) gtk_icon_set_add_source __attribute((alias("IA__gtk_icon_set_add_source"), visibility("default")));

#undef gtk_icon_set_copy
extern __typeof (gtk_icon_set_copy) gtk_icon_set_copy __attribute((alias("IA__gtk_icon_set_copy"), visibility("default")));

#undef gtk_icon_set_get_sizes
extern __typeof (gtk_icon_set_get_sizes) gtk_icon_set_get_sizes __attribute((alias("IA__gtk_icon_set_get_sizes"), visibility("default")));

#undef gtk_icon_set_get_type
extern __typeof (gtk_icon_set_get_type) gtk_icon_set_get_type __attribute((alias("IA__gtk_icon_set_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_icon_set_new
extern __typeof (gtk_icon_set_new) gtk_icon_set_new __attribute((alias("IA__gtk_icon_set_new"), visibility("default")));

#undef gtk_icon_set_new_from_pixbuf
extern __typeof (gtk_icon_set_new_from_pixbuf) gtk_icon_set_new_from_pixbuf __attribute((alias("IA__gtk_icon_set_new_from_pixbuf"), visibility("default")));

#undef gtk_icon_set_ref
extern __typeof (gtk_icon_set_ref) gtk_icon_set_ref __attribute((alias("IA__gtk_icon_set_ref"), visibility("default")));

#undef gtk_icon_set_render_icon
extern __typeof (gtk_icon_set_render_icon) gtk_icon_set_render_icon __attribute((alias("IA__gtk_icon_set_render_icon"), visibility("default")));

#undef gtk_icon_set_unref
extern __typeof (gtk_icon_set_unref) gtk_icon_set_unref __attribute((alias("IA__gtk_icon_set_unref"), visibility("default")));

#undef gtk_icon_source_copy
extern __typeof (gtk_icon_source_copy) gtk_icon_source_copy __attribute((alias("IA__gtk_icon_source_copy"), visibility("default")));

#undef gtk_icon_source_free
extern __typeof (gtk_icon_source_free) gtk_icon_source_free __attribute((alias("IA__gtk_icon_source_free"), visibility("default")));

#undef gtk_icon_source_get_direction
extern __typeof (gtk_icon_source_get_direction) gtk_icon_source_get_direction __attribute((alias("IA__gtk_icon_source_get_direction"), visibility("default")));

#undef gtk_icon_source_get_direction_wildcarded
extern __typeof (gtk_icon_source_get_direction_wildcarded) gtk_icon_source_get_direction_wildcarded __attribute((alias("IA__gtk_icon_source_get_direction_wildcarded"), visibility("default")));

#ifndef _WIN64
#undef gtk_icon_source_get_filename
extern __typeof (gtk_icon_source_get_filename) gtk_icon_source_get_filename __attribute((alias("IA__gtk_icon_source_get_filename"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_icon_source_get_filename_utf8
extern __typeof (gtk_icon_source_get_filename_utf8) gtk_icon_source_get_filename_utf8 __attribute((alias("IA__gtk_icon_source_get_filename_utf8"), visibility("default")));

#endif
#undef gtk_icon_source_get_icon_name
extern __typeof (gtk_icon_source_get_icon_name) gtk_icon_source_get_icon_name __attribute((alias("IA__gtk_icon_source_get_icon_name"), visibility("default")));

#undef gtk_icon_source_get_pixbuf
extern __typeof (gtk_icon_source_get_pixbuf) gtk_icon_source_get_pixbuf __attribute((alias("IA__gtk_icon_source_get_pixbuf"), visibility("default")));

#undef gtk_icon_source_get_size
extern __typeof (gtk_icon_source_get_size) gtk_icon_source_get_size __attribute((alias("IA__gtk_icon_source_get_size"), visibility("default")));

#undef gtk_icon_source_get_size_wildcarded
extern __typeof (gtk_icon_source_get_size_wildcarded) gtk_icon_source_get_size_wildcarded __attribute((alias("IA__gtk_icon_source_get_size_wildcarded"), visibility("default")));

#undef gtk_icon_source_get_state
extern __typeof (gtk_icon_source_get_state) gtk_icon_source_get_state __attribute((alias("IA__gtk_icon_source_get_state"), visibility("default")));

#undef gtk_icon_source_get_state_wildcarded
extern __typeof (gtk_icon_source_get_state_wildcarded) gtk_icon_source_get_state_wildcarded __attribute((alias("IA__gtk_icon_source_get_state_wildcarded"), visibility("default")));

#undef gtk_icon_source_get_type
extern __typeof (gtk_icon_source_get_type) gtk_icon_source_get_type __attribute((alias("IA__gtk_icon_source_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_icon_source_new
extern __typeof (gtk_icon_source_new) gtk_icon_source_new __attribute((alias("IA__gtk_icon_source_new"), visibility("default")));

#undef gtk_icon_source_set_direction
extern __typeof (gtk_icon_source_set_direction) gtk_icon_source_set_direction __attribute((alias("IA__gtk_icon_source_set_direction"), visibility("default")));

#undef gtk_icon_source_set_direction_wildcarded
extern __typeof (gtk_icon_source_set_direction_wildcarded) gtk_icon_source_set_direction_wildcarded __attribute((alias("IA__gtk_icon_source_set_direction_wildcarded"), visibility("default")));

#ifndef _WIN64
#undef gtk_icon_source_set_filename
extern __typeof (gtk_icon_source_set_filename) gtk_icon_source_set_filename __attribute((alias("IA__gtk_icon_source_set_filename"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_icon_source_set_filename_utf8
extern __typeof (gtk_icon_source_set_filename_utf8) gtk_icon_source_set_filename_utf8 __attribute((alias("IA__gtk_icon_source_set_filename_utf8"), visibility("default")));

#endif
#undef gtk_icon_source_set_icon_name
extern __typeof (gtk_icon_source_set_icon_name) gtk_icon_source_set_icon_name __attribute((alias("IA__gtk_icon_source_set_icon_name"), visibility("default")));

#undef gtk_icon_source_set_pixbuf
extern __typeof (gtk_icon_source_set_pixbuf) gtk_icon_source_set_pixbuf __attribute((alias("IA__gtk_icon_source_set_pixbuf"), visibility("default")));

#undef gtk_icon_source_set_size
extern __typeof (gtk_icon_source_set_size) gtk_icon_source_set_size __attribute((alias("IA__gtk_icon_source_set_size"), visibility("default")));

#undef gtk_icon_source_set_size_wildcarded
extern __typeof (gtk_icon_source_set_size_wildcarded) gtk_icon_source_set_size_wildcarded __attribute((alias("IA__gtk_icon_source_set_size_wildcarded"), visibility("default")));

#undef gtk_icon_source_set_state
extern __typeof (gtk_icon_source_set_state) gtk_icon_source_set_state __attribute((alias("IA__gtk_icon_source_set_state"), visibility("default")));

#undef gtk_icon_source_set_state_wildcarded
extern __typeof (gtk_icon_source_set_state_wildcarded) gtk_icon_source_set_state_wildcarded __attribute((alias("IA__gtk_icon_source_set_state_wildcarded"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_ICON_THEME_H__)
#if IN_FILE(__GTK_ICON_THEME_C__)
#undef gtk_icon_info_copy
extern __typeof (gtk_icon_info_copy) gtk_icon_info_copy __attribute((alias("IA__gtk_icon_info_copy"), visibility("default")));

#undef gtk_icon_info_free
extern __typeof (gtk_icon_info_free) gtk_icon_info_free __attribute((alias("IA__gtk_icon_info_free"), visibility("default")));

#undef gtk_icon_info_new_for_pixbuf
extern __typeof (gtk_icon_info_new_for_pixbuf) gtk_icon_info_new_for_pixbuf __attribute((alias("IA__gtk_icon_info_new_for_pixbuf"), visibility("default")));

#undef gtk_icon_info_get_attach_points
extern __typeof (gtk_icon_info_get_attach_points) gtk_icon_info_get_attach_points __attribute((alias("IA__gtk_icon_info_get_attach_points"), visibility("default")));

#undef gtk_icon_info_get_base_size
extern __typeof (gtk_icon_info_get_base_size) gtk_icon_info_get_base_size __attribute((alias("IA__gtk_icon_info_get_base_size"), visibility("default")));

#undef gtk_icon_info_get_builtin_pixbuf
extern __typeof (gtk_icon_info_get_builtin_pixbuf) gtk_icon_info_get_builtin_pixbuf __attribute((alias("IA__gtk_icon_info_get_builtin_pixbuf"), visibility("default")));

#undef gtk_icon_info_get_display_name
extern __typeof (gtk_icon_info_get_display_name) gtk_icon_info_get_display_name __attribute((alias("IA__gtk_icon_info_get_display_name"), visibility("default")));

#undef gtk_icon_info_get_embedded_rect
extern __typeof (gtk_icon_info_get_embedded_rect) gtk_icon_info_get_embedded_rect __attribute((alias("IA__gtk_icon_info_get_embedded_rect"), visibility("default")));

#ifndef _WIN64
#undef gtk_icon_info_get_filename
extern __typeof (gtk_icon_info_get_filename) gtk_icon_info_get_filename __attribute((alias("IA__gtk_icon_info_get_filename"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_icon_info_get_filename_utf8
extern __typeof (gtk_icon_info_get_filename_utf8) gtk_icon_info_get_filename_utf8 __attribute((alias("IA__gtk_icon_info_get_filename_utf8"), visibility("default")));

#endif
#undef gtk_icon_info_get_type
extern __typeof (gtk_icon_info_get_type) gtk_icon_info_get_type __attribute((alias("IA__gtk_icon_info_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_icon_info_load_icon
extern __typeof (gtk_icon_info_load_icon) gtk_icon_info_load_icon __attribute((alias("IA__gtk_icon_info_load_icon"), visibility("default")));

#undef gtk_icon_info_set_raw_coordinates
extern __typeof (gtk_icon_info_set_raw_coordinates) gtk_icon_info_set_raw_coordinates __attribute((alias("IA__gtk_icon_info_set_raw_coordinates"), visibility("default")));

#undef gtk_icon_theme_add_builtin_icon
extern __typeof (gtk_icon_theme_add_builtin_icon) gtk_icon_theme_add_builtin_icon __attribute((alias("IA__gtk_icon_theme_add_builtin_icon"), visibility("default")));

#ifndef _WIN64
#undef gtk_icon_theme_append_search_path
extern __typeof (gtk_icon_theme_append_search_path) gtk_icon_theme_append_search_path __attribute((alias("IA__gtk_icon_theme_append_search_path"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_icon_theme_append_search_path_utf8
extern __typeof (gtk_icon_theme_append_search_path_utf8) gtk_icon_theme_append_search_path_utf8 __attribute((alias("IA__gtk_icon_theme_append_search_path_utf8"), visibility("default")));

#endif
#undef gtk_icon_theme_error_quark
extern __typeof (gtk_icon_theme_error_quark) gtk_icon_theme_error_quark __attribute((alias("IA__gtk_icon_theme_error_quark"), visibility("default")));

#undef gtk_icon_theme_get_default
extern __typeof (gtk_icon_theme_get_default) gtk_icon_theme_get_default __attribute((alias("IA__gtk_icon_theme_get_default"), visibility("default")));

#undef gtk_icon_theme_get_example_icon_name
extern __typeof (gtk_icon_theme_get_example_icon_name) gtk_icon_theme_get_example_icon_name __attribute((alias("IA__gtk_icon_theme_get_example_icon_name"), visibility("default")));

#undef gtk_icon_theme_get_for_screen
extern __typeof (gtk_icon_theme_get_for_screen) gtk_icon_theme_get_for_screen __attribute((alias("IA__gtk_icon_theme_get_for_screen"), visibility("default")));

#undef gtk_icon_theme_get_icon_sizes
extern __typeof (gtk_icon_theme_get_icon_sizes) gtk_icon_theme_get_icon_sizes __attribute((alias("IA__gtk_icon_theme_get_icon_sizes"), visibility("default")));

#ifndef _WIN64
#undef gtk_icon_theme_get_search_path
extern __typeof (gtk_icon_theme_get_search_path) gtk_icon_theme_get_search_path __attribute((alias("IA__gtk_icon_theme_get_search_path"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_icon_theme_get_search_path_utf8
extern __typeof (gtk_icon_theme_get_search_path_utf8) gtk_icon_theme_get_search_path_utf8 __attribute((alias("IA__gtk_icon_theme_get_search_path_utf8"), visibility("default")));

#endif
#undef gtk_icon_theme_get_type
extern __typeof (gtk_icon_theme_get_type) gtk_icon_theme_get_type __attribute((alias("IA__gtk_icon_theme_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_icon_theme_has_icon
extern __typeof (gtk_icon_theme_has_icon) gtk_icon_theme_has_icon __attribute((alias("IA__gtk_icon_theme_has_icon"), visibility("default")));

#undef gtk_icon_theme_list_contexts
extern __typeof (gtk_icon_theme_list_contexts) gtk_icon_theme_list_contexts __attribute((alias("IA__gtk_icon_theme_list_contexts"), visibility("default")));

#undef gtk_icon_theme_list_icons
extern __typeof (gtk_icon_theme_list_icons) gtk_icon_theme_list_icons __attribute((alias("IA__gtk_icon_theme_list_icons"), visibility("default")));

#undef gtk_icon_theme_load_icon
extern __typeof (gtk_icon_theme_load_icon) gtk_icon_theme_load_icon __attribute((alias("IA__gtk_icon_theme_load_icon"), visibility("default")));

#undef gtk_icon_theme_lookup_icon
extern __typeof (gtk_icon_theme_lookup_icon) gtk_icon_theme_lookup_icon __attribute((alias("IA__gtk_icon_theme_lookup_icon"), visibility("default")));

#undef gtk_icon_theme_lookup_by_gicon
extern __typeof (gtk_icon_theme_lookup_by_gicon) gtk_icon_theme_lookup_by_gicon __attribute((alias("IA__gtk_icon_theme_lookup_by_gicon"), visibility("default")));

#undef gtk_icon_theme_choose_icon
extern __typeof (gtk_icon_theme_choose_icon) gtk_icon_theme_choose_icon __attribute((alias("IA__gtk_icon_theme_choose_icon"), visibility("default")));

#undef gtk_icon_theme_new
extern __typeof (gtk_icon_theme_new) gtk_icon_theme_new __attribute((alias("IA__gtk_icon_theme_new"), visibility("default")));

#ifndef _WIN64
#undef gtk_icon_theme_prepend_search_path
extern __typeof (gtk_icon_theme_prepend_search_path) gtk_icon_theme_prepend_search_path __attribute((alias("IA__gtk_icon_theme_prepend_search_path"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_icon_theme_prepend_search_path_utf8
extern __typeof (gtk_icon_theme_prepend_search_path_utf8) gtk_icon_theme_prepend_search_path_utf8 __attribute((alias("IA__gtk_icon_theme_prepend_search_path_utf8"), visibility("default")));

#endif
#undef gtk_icon_theme_rescan_if_needed
extern __typeof (gtk_icon_theme_rescan_if_needed) gtk_icon_theme_rescan_if_needed __attribute((alias("IA__gtk_icon_theme_rescan_if_needed"), visibility("default")));

#undef gtk_icon_theme_set_custom_theme
extern __typeof (gtk_icon_theme_set_custom_theme) gtk_icon_theme_set_custom_theme __attribute((alias("IA__gtk_icon_theme_set_custom_theme"), visibility("default")));

#undef gtk_icon_theme_set_screen
extern __typeof (gtk_icon_theme_set_screen) gtk_icon_theme_set_screen __attribute((alias("IA__gtk_icon_theme_set_screen"), visibility("default")));

#ifndef _WIN64
#undef gtk_icon_theme_set_search_path
extern __typeof (gtk_icon_theme_set_search_path) gtk_icon_theme_set_search_path __attribute((alias("IA__gtk_icon_theme_set_search_path"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_icon_theme_set_search_path_utf8
extern __typeof (gtk_icon_theme_set_search_path_utf8) gtk_icon_theme_set_search_path_utf8 __attribute((alias("IA__gtk_icon_theme_set_search_path_utf8"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_ICON_VIEW_H__)
#if IN_FILE(__GTK_ICON_VIEW_C__)
#undef gtk_icon_view_get_column_spacing
extern __typeof (gtk_icon_view_get_column_spacing) gtk_icon_view_get_column_spacing __attribute((alias("IA__gtk_icon_view_get_column_spacing"), visibility("default")));

#undef gtk_icon_view_get_columns
extern __typeof (gtk_icon_view_get_columns) gtk_icon_view_get_columns __attribute((alias("IA__gtk_icon_view_get_columns"), visibility("default")));

#undef gtk_icon_view_get_cursor
extern __typeof (gtk_icon_view_get_cursor) gtk_icon_view_get_cursor __attribute((alias("IA__gtk_icon_view_get_cursor"), visibility("default")));

#undef gtk_icon_view_get_item_padding
extern __typeof (gtk_icon_view_get_item_padding) gtk_icon_view_get_item_padding __attribute((alias("IA__gtk_icon_view_get_item_padding"), visibility("default")));

#undef gtk_icon_view_get_item_width
extern __typeof (gtk_icon_view_get_item_width) gtk_icon_view_get_item_width __attribute((alias("IA__gtk_icon_view_get_item_width"), visibility("default")));

#undef gtk_icon_view_get_margin
extern __typeof (gtk_icon_view_get_margin) gtk_icon_view_get_margin __attribute((alias("IA__gtk_icon_view_get_margin"), visibility("default")));

#undef gtk_icon_view_get_markup_column
extern __typeof (gtk_icon_view_get_markup_column) gtk_icon_view_get_markup_column __attribute((alias("IA__gtk_icon_view_get_markup_column"), visibility("default")));

#undef gtk_icon_view_get_model
extern __typeof (gtk_icon_view_get_model) gtk_icon_view_get_model __attribute((alias("IA__gtk_icon_view_get_model"), visibility("default")));

#undef gtk_icon_view_get_orientation
extern __typeof (gtk_icon_view_get_orientation) gtk_icon_view_get_orientation __attribute((alias("IA__gtk_icon_view_get_orientation"), visibility("default")));

#undef gtk_icon_view_get_item_orientation
extern __typeof (gtk_icon_view_get_item_orientation) gtk_icon_view_get_item_orientation __attribute((alias("IA__gtk_icon_view_get_item_orientation"), visibility("default")));

#undef gtk_icon_view_get_path_at_pos
extern __typeof (gtk_icon_view_get_path_at_pos) gtk_icon_view_get_path_at_pos __attribute((alias("IA__gtk_icon_view_get_path_at_pos"), visibility("default")));

#undef gtk_icon_view_get_item_at_pos
extern __typeof (gtk_icon_view_get_item_at_pos) gtk_icon_view_get_item_at_pos __attribute((alias("IA__gtk_icon_view_get_item_at_pos"), visibility("default")));

#undef gtk_icon_view_convert_widget_to_bin_window_coords
extern __typeof (gtk_icon_view_convert_widget_to_bin_window_coords) gtk_icon_view_convert_widget_to_bin_window_coords __attribute((alias("IA__gtk_icon_view_convert_widget_to_bin_window_coords"), visibility("default")));

#undef gtk_icon_view_get_pixbuf_column
extern __typeof (gtk_icon_view_get_pixbuf_column) gtk_icon_view_get_pixbuf_column __attribute((alias("IA__gtk_icon_view_get_pixbuf_column"), visibility("default")));

#undef gtk_icon_view_get_row_spacing
extern __typeof (gtk_icon_view_get_row_spacing) gtk_icon_view_get_row_spacing __attribute((alias("IA__gtk_icon_view_get_row_spacing"), visibility("default")));

#undef gtk_icon_view_get_selected_items
extern __typeof (gtk_icon_view_get_selected_items) gtk_icon_view_get_selected_items __attribute((alias("IA__gtk_icon_view_get_selected_items"), visibility("default")));

#undef gtk_icon_view_get_selection_mode
extern __typeof (gtk_icon_view_get_selection_mode) gtk_icon_view_get_selection_mode __attribute((alias("IA__gtk_icon_view_get_selection_mode"), visibility("default")));

#undef gtk_icon_view_get_spacing
extern __typeof (gtk_icon_view_get_spacing) gtk_icon_view_get_spacing __attribute((alias("IA__gtk_icon_view_get_spacing"), visibility("default")));

#undef gtk_icon_view_get_text_column
extern __typeof (gtk_icon_view_get_text_column) gtk_icon_view_get_text_column __attribute((alias("IA__gtk_icon_view_get_text_column"), visibility("default")));

#undef gtk_icon_view_get_type
extern __typeof (gtk_icon_view_get_type) gtk_icon_view_get_type __attribute((alias("IA__gtk_icon_view_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_icon_view_get_visible_range
extern __typeof (gtk_icon_view_get_visible_range) gtk_icon_view_get_visible_range __attribute((alias("IA__gtk_icon_view_get_visible_range"), visibility("default")));

#undef gtk_icon_view_item_activated
extern __typeof (gtk_icon_view_item_activated) gtk_icon_view_item_activated __attribute((alias("IA__gtk_icon_view_item_activated"), visibility("default")));

#undef gtk_icon_view_new
extern __typeof (gtk_icon_view_new) gtk_icon_view_new __attribute((alias("IA__gtk_icon_view_new"), visibility("default")));

#undef gtk_icon_view_new_with_model
extern __typeof (gtk_icon_view_new_with_model) gtk_icon_view_new_with_model __attribute((alias("IA__gtk_icon_view_new_with_model"), visibility("default")));

#undef gtk_icon_view_path_is_selected
extern __typeof (gtk_icon_view_path_is_selected) gtk_icon_view_path_is_selected __attribute((alias("IA__gtk_icon_view_path_is_selected"), visibility("default")));

#undef gtk_icon_view_scroll_to_path
extern __typeof (gtk_icon_view_scroll_to_path) gtk_icon_view_scroll_to_path __attribute((alias("IA__gtk_icon_view_scroll_to_path"), visibility("default")));

#undef gtk_icon_view_select_all
extern __typeof (gtk_icon_view_select_all) gtk_icon_view_select_all __attribute((alias("IA__gtk_icon_view_select_all"), visibility("default")));

#undef gtk_icon_view_selected_foreach
extern __typeof (gtk_icon_view_selected_foreach) gtk_icon_view_selected_foreach __attribute((alias("IA__gtk_icon_view_selected_foreach"), visibility("default")));

#undef gtk_icon_view_select_path
extern __typeof (gtk_icon_view_select_path) gtk_icon_view_select_path __attribute((alias("IA__gtk_icon_view_select_path"), visibility("default")));

#undef gtk_icon_view_set_column_spacing
extern __typeof (gtk_icon_view_set_column_spacing) gtk_icon_view_set_column_spacing __attribute((alias("IA__gtk_icon_view_set_column_spacing"), visibility("default")));

#undef gtk_icon_view_set_columns
extern __typeof (gtk_icon_view_set_columns) gtk_icon_view_set_columns __attribute((alias("IA__gtk_icon_view_set_columns"), visibility("default")));

#undef gtk_icon_view_set_cursor
extern __typeof (gtk_icon_view_set_cursor) gtk_icon_view_set_cursor __attribute((alias("IA__gtk_icon_view_set_cursor"), visibility("default")));

#undef gtk_icon_view_set_item_padding
extern __typeof (gtk_icon_view_set_item_padding) gtk_icon_view_set_item_padding __attribute((alias("IA__gtk_icon_view_set_item_padding"), visibility("default")));

#undef gtk_icon_view_set_item_width
extern __typeof (gtk_icon_view_set_item_width) gtk_icon_view_set_item_width __attribute((alias("IA__gtk_icon_view_set_item_width"), visibility("default")));

#undef gtk_icon_view_set_margin
extern __typeof (gtk_icon_view_set_margin) gtk_icon_view_set_margin __attribute((alias("IA__gtk_icon_view_set_margin"), visibility("default")));

#undef gtk_icon_view_set_markup_column
extern __typeof (gtk_icon_view_set_markup_column) gtk_icon_view_set_markup_column __attribute((alias("IA__gtk_icon_view_set_markup_column"), visibility("default")));

#undef gtk_icon_view_set_model
extern __typeof (gtk_icon_view_set_model) gtk_icon_view_set_model __attribute((alias("IA__gtk_icon_view_set_model"), visibility("default")));

#undef gtk_icon_view_set_orientation
extern __typeof (gtk_icon_view_set_orientation) gtk_icon_view_set_orientation __attribute((alias("IA__gtk_icon_view_set_orientation"), visibility("default")));

#undef gtk_icon_view_set_item_orientation
extern __typeof (gtk_icon_view_set_item_orientation) gtk_icon_view_set_item_orientation __attribute((alias("IA__gtk_icon_view_set_item_orientation"), visibility("default")));

#undef gtk_icon_view_set_pixbuf_column
extern __typeof (gtk_icon_view_set_pixbuf_column) gtk_icon_view_set_pixbuf_column __attribute((alias("IA__gtk_icon_view_set_pixbuf_column"), visibility("default")));

#undef gtk_icon_view_set_row_spacing
extern __typeof (gtk_icon_view_set_row_spacing) gtk_icon_view_set_row_spacing __attribute((alias("IA__gtk_icon_view_set_row_spacing"), visibility("default")));

#undef gtk_icon_view_set_selection_mode
extern __typeof (gtk_icon_view_set_selection_mode) gtk_icon_view_set_selection_mode __attribute((alias("IA__gtk_icon_view_set_selection_mode"), visibility("default")));

#undef gtk_icon_view_set_spacing
extern __typeof (gtk_icon_view_set_spacing) gtk_icon_view_set_spacing __attribute((alias("IA__gtk_icon_view_set_spacing"), visibility("default")));

#undef gtk_icon_view_set_text_column
extern __typeof (gtk_icon_view_set_text_column) gtk_icon_view_set_text_column __attribute((alias("IA__gtk_icon_view_set_text_column"), visibility("default")));

#undef gtk_icon_view_unselect_all
extern __typeof (gtk_icon_view_unselect_all) gtk_icon_view_unselect_all __attribute((alias("IA__gtk_icon_view_unselect_all"), visibility("default")));

#undef gtk_icon_view_unselect_path
extern __typeof (gtk_icon_view_unselect_path) gtk_icon_view_unselect_path __attribute((alias("IA__gtk_icon_view_unselect_path"), visibility("default")));

#undef gtk_icon_view_enable_model_drag_source
extern __typeof (gtk_icon_view_enable_model_drag_source) gtk_icon_view_enable_model_drag_source __attribute((alias("IA__gtk_icon_view_enable_model_drag_source"), visibility("default")));

#undef gtk_icon_view_enable_model_drag_dest
extern __typeof (gtk_icon_view_enable_model_drag_dest) gtk_icon_view_enable_model_drag_dest __attribute((alias("IA__gtk_icon_view_enable_model_drag_dest"), visibility("default")));

#undef gtk_icon_view_unset_model_drag_source
extern __typeof (gtk_icon_view_unset_model_drag_source) gtk_icon_view_unset_model_drag_source __attribute((alias("IA__gtk_icon_view_unset_model_drag_source"), visibility("default")));

#undef gtk_icon_view_unset_model_drag_dest
extern __typeof (gtk_icon_view_unset_model_drag_dest) gtk_icon_view_unset_model_drag_dest __attribute((alias("IA__gtk_icon_view_unset_model_drag_dest"), visibility("default")));

#undef gtk_icon_view_set_reorderable
extern __typeof (gtk_icon_view_set_reorderable) gtk_icon_view_set_reorderable __attribute((alias("IA__gtk_icon_view_set_reorderable"), visibility("default")));

#undef gtk_icon_view_get_reorderable
extern __typeof (gtk_icon_view_get_reorderable) gtk_icon_view_get_reorderable __attribute((alias("IA__gtk_icon_view_get_reorderable"), visibility("default")));

#undef gtk_icon_view_set_drag_dest_item
extern __typeof (gtk_icon_view_set_drag_dest_item) gtk_icon_view_set_drag_dest_item __attribute((alias("IA__gtk_icon_view_set_drag_dest_item"), visibility("default")));

#undef gtk_icon_view_get_drag_dest_item
extern __typeof (gtk_icon_view_get_drag_dest_item) gtk_icon_view_get_drag_dest_item __attribute((alias("IA__gtk_icon_view_get_drag_dest_item"), visibility("default")));

#undef gtk_icon_view_get_dest_item_at_pos
extern __typeof (gtk_icon_view_get_dest_item_at_pos) gtk_icon_view_get_dest_item_at_pos __attribute((alias("IA__gtk_icon_view_get_dest_item_at_pos"), visibility("default")));

#undef gtk_icon_view_create_drag_icon
extern __typeof (gtk_icon_view_create_drag_icon) gtk_icon_view_create_drag_icon __attribute((alias("IA__gtk_icon_view_create_drag_icon"), visibility("default")));

#undef gtk_icon_view_set_tooltip_item
extern __typeof (gtk_icon_view_set_tooltip_item) gtk_icon_view_set_tooltip_item __attribute((alias("IA__gtk_icon_view_set_tooltip_item"), visibility("default")));

#undef gtk_icon_view_set_tooltip_cell
extern __typeof (gtk_icon_view_set_tooltip_cell) gtk_icon_view_set_tooltip_cell __attribute((alias("IA__gtk_icon_view_set_tooltip_cell"), visibility("default")));

#undef gtk_icon_view_get_tooltip_context
extern __typeof (gtk_icon_view_get_tooltip_context) gtk_icon_view_get_tooltip_context __attribute((alias("IA__gtk_icon_view_get_tooltip_context"), visibility("default")));

#undef gtk_icon_view_set_tooltip_column
extern __typeof (gtk_icon_view_set_tooltip_column) gtk_icon_view_set_tooltip_column __attribute((alias("IA__gtk_icon_view_set_tooltip_column"), visibility("default")));

#undef gtk_icon_view_get_tooltip_column
extern __typeof (gtk_icon_view_get_tooltip_column) gtk_icon_view_get_tooltip_column __attribute((alias("IA__gtk_icon_view_get_tooltip_column"), visibility("default")));

#undef gtk_icon_view_get_item_row
extern __typeof (gtk_icon_view_get_item_row) gtk_icon_view_get_item_row __attribute((alias("IA__gtk_icon_view_get_item_row"), visibility("default")));

#undef gtk_icon_view_get_item_column
extern __typeof (gtk_icon_view_get_item_column) gtk_icon_view_get_item_column __attribute((alias("IA__gtk_icon_view_get_item_column"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_IMAGE_H__)
#if IN_FILE(__GTK_IMAGE_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_image_get
extern __typeof (gtk_image_get) gtk_image_get __attribute((alias("IA__gtk_image_get"), visibility("default")));

#undef gtk_image_set
extern __typeof (gtk_image_set) gtk_image_set __attribute((alias("IA__gtk_image_set"), visibility("default")));

#endif
#undef gtk_image_clear
extern __typeof (gtk_image_clear) gtk_image_clear __attribute((alias("IA__gtk_image_clear"), visibility("default")));

#undef gtk_image_get_animation
extern __typeof (gtk_image_get_animation) gtk_image_get_animation __attribute((alias("IA__gtk_image_get_animation"), visibility("default")));

#undef gtk_image_get_icon_name
extern __typeof (gtk_image_get_icon_name) gtk_image_get_icon_name __attribute((alias("IA__gtk_image_get_icon_name"), visibility("default")));

#undef gtk_image_get_icon_set
extern __typeof (gtk_image_get_icon_set) gtk_image_get_icon_set __attribute((alias("IA__gtk_image_get_icon_set"), visibility("default")));

#undef gtk_image_get_image
extern __typeof (gtk_image_get_image) gtk_image_get_image __attribute((alias("IA__gtk_image_get_image"), visibility("default")));

#undef gtk_image_get_pixbuf
extern __typeof (gtk_image_get_pixbuf) gtk_image_get_pixbuf __attribute((alias("IA__gtk_image_get_pixbuf"), visibility("default")));

#undef gtk_image_get_pixel_size
extern __typeof (gtk_image_get_pixel_size) gtk_image_get_pixel_size __attribute((alias("IA__gtk_image_get_pixel_size"), visibility("default")));

#undef gtk_image_get_pixmap
extern __typeof (gtk_image_get_pixmap) gtk_image_get_pixmap __attribute((alias("IA__gtk_image_get_pixmap"), visibility("default")));

#undef gtk_image_get_stock
extern __typeof (gtk_image_get_stock) gtk_image_get_stock __attribute((alias("IA__gtk_image_get_stock"), visibility("default")));

#undef gtk_image_get_gicon
extern __typeof (gtk_image_get_gicon) gtk_image_get_gicon __attribute((alias("IA__gtk_image_get_gicon"), visibility("default")));

#undef gtk_image_get_storage_type
extern __typeof (gtk_image_get_storage_type) gtk_image_get_storage_type __attribute((alias("IA__gtk_image_get_storage_type"), visibility("default")));

#undef gtk_image_get_type
extern __typeof (gtk_image_get_type) gtk_image_get_type __attribute((alias("IA__gtk_image_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_image_new
extern __typeof (gtk_image_new) gtk_image_new __attribute((alias("IA__gtk_image_new"), visibility("default")));

#undef gtk_image_new_from_animation
extern __typeof (gtk_image_new_from_animation) gtk_image_new_from_animation __attribute((alias("IA__gtk_image_new_from_animation"), visibility("default")));

#ifndef _WIN64
#undef gtk_image_new_from_file
extern __typeof (gtk_image_new_from_file) gtk_image_new_from_file __attribute((alias("IA__gtk_image_new_from_file"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_image_new_from_file_utf8
extern __typeof (gtk_image_new_from_file_utf8) gtk_image_new_from_file_utf8 __attribute((alias("IA__gtk_image_new_from_file_utf8"), visibility("default")));

#endif
#undef gtk_image_new_from_icon_name
extern __typeof (gtk_image_new_from_icon_name) gtk_image_new_from_icon_name __attribute((alias("IA__gtk_image_new_from_icon_name"), visibility("default")));

#undef gtk_image_new_from_icon_set
extern __typeof (gtk_image_new_from_icon_set) gtk_image_new_from_icon_set __attribute((alias("IA__gtk_image_new_from_icon_set"), visibility("default")));

#undef gtk_image_new_from_image
extern __typeof (gtk_image_new_from_image) gtk_image_new_from_image __attribute((alias("IA__gtk_image_new_from_image"), visibility("default")));

#undef gtk_image_new_from_pixbuf
extern __typeof (gtk_image_new_from_pixbuf) gtk_image_new_from_pixbuf __attribute((alias("IA__gtk_image_new_from_pixbuf"), visibility("default")));

#undef gtk_image_new_from_pixmap
extern __typeof (gtk_image_new_from_pixmap) gtk_image_new_from_pixmap __attribute((alias("IA__gtk_image_new_from_pixmap"), visibility("default")));

#undef gtk_image_new_from_stock
extern __typeof (gtk_image_new_from_stock) gtk_image_new_from_stock __attribute((alias("IA__gtk_image_new_from_stock"), visibility("default")));

#undef gtk_image_new_from_gicon
extern __typeof (gtk_image_new_from_gicon) gtk_image_new_from_gicon __attribute((alias("IA__gtk_image_new_from_gicon"), visibility("default")));

#undef gtk_image_set_from_animation
extern __typeof (gtk_image_set_from_animation) gtk_image_set_from_animation __attribute((alias("IA__gtk_image_set_from_animation"), visibility("default")));

#ifndef _WIN64
#undef gtk_image_set_from_file
extern __typeof (gtk_image_set_from_file) gtk_image_set_from_file __attribute((alias("IA__gtk_image_set_from_file"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_image_set_from_file_utf8
extern __typeof (gtk_image_set_from_file_utf8) gtk_image_set_from_file_utf8 __attribute((alias("IA__gtk_image_set_from_file_utf8"), visibility("default")));

#endif
#undef gtk_image_set_from_icon_name
extern __typeof (gtk_image_set_from_icon_name) gtk_image_set_from_icon_name __attribute((alias("IA__gtk_image_set_from_icon_name"), visibility("default")));

#undef gtk_image_set_from_icon_set
extern __typeof (gtk_image_set_from_icon_set) gtk_image_set_from_icon_set __attribute((alias("IA__gtk_image_set_from_icon_set"), visibility("default")));

#undef gtk_image_set_from_image
extern __typeof (gtk_image_set_from_image) gtk_image_set_from_image __attribute((alias("IA__gtk_image_set_from_image"), visibility("default")));

#undef gtk_image_set_from_pixbuf
extern __typeof (gtk_image_set_from_pixbuf) gtk_image_set_from_pixbuf __attribute((alias("IA__gtk_image_set_from_pixbuf"), visibility("default")));

#undef gtk_image_set_from_pixmap
extern __typeof (gtk_image_set_from_pixmap) gtk_image_set_from_pixmap __attribute((alias("IA__gtk_image_set_from_pixmap"), visibility("default")));

#undef gtk_image_set_from_stock
extern __typeof (gtk_image_set_from_stock) gtk_image_set_from_stock __attribute((alias("IA__gtk_image_set_from_stock"), visibility("default")));

#undef gtk_image_set_from_gicon
extern __typeof (gtk_image_set_from_gicon) gtk_image_set_from_gicon __attribute((alias("IA__gtk_image_set_from_gicon"), visibility("default")));

#undef gtk_image_set_pixel_size
extern __typeof (gtk_image_set_pixel_size) gtk_image_set_pixel_size __attribute((alias("IA__gtk_image_set_pixel_size"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_IMAGE_MENU_ITEM_H__)
#if IN_FILE(__GTK_IMAGE_MENU_ITEM_C__)
#undef gtk_image_menu_item_get_always_show_image
extern __typeof (gtk_image_menu_item_get_always_show_image) gtk_image_menu_item_get_always_show_image __attribute((alias("IA__gtk_image_menu_item_get_always_show_image"), visibility("default")));

#undef gtk_image_menu_item_get_image
extern __typeof (gtk_image_menu_item_get_image) gtk_image_menu_item_get_image __attribute((alias("IA__gtk_image_menu_item_get_image"), visibility("default")));

#undef gtk_image_menu_item_get_type
extern __typeof (gtk_image_menu_item_get_type) gtk_image_menu_item_get_type __attribute((alias("IA__gtk_image_menu_item_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_image_menu_item_get_use_stock
extern __typeof (gtk_image_menu_item_get_use_stock) gtk_image_menu_item_get_use_stock __attribute((alias("IA__gtk_image_menu_item_get_use_stock"), visibility("default")));

#undef gtk_image_menu_item_new
extern __typeof (gtk_image_menu_item_new) gtk_image_menu_item_new __attribute((alias("IA__gtk_image_menu_item_new"), visibility("default")));

#undef gtk_image_menu_item_new_from_stock
extern __typeof (gtk_image_menu_item_new_from_stock) gtk_image_menu_item_new_from_stock __attribute((alias("IA__gtk_image_menu_item_new_from_stock"), visibility("default")));

#undef gtk_image_menu_item_new_with_label
extern __typeof (gtk_image_menu_item_new_with_label) gtk_image_menu_item_new_with_label __attribute((alias("IA__gtk_image_menu_item_new_with_label"), visibility("default")));

#undef gtk_image_menu_item_new_with_mnemonic
extern __typeof (gtk_image_menu_item_new_with_mnemonic) gtk_image_menu_item_new_with_mnemonic __attribute((alias("IA__gtk_image_menu_item_new_with_mnemonic"), visibility("default")));

#undef gtk_image_menu_item_set_accel_group
extern __typeof (gtk_image_menu_item_set_accel_group) gtk_image_menu_item_set_accel_group __attribute((alias("IA__gtk_image_menu_item_set_accel_group"), visibility("default")));

#undef gtk_image_menu_item_set_always_show_image
extern __typeof (gtk_image_menu_item_set_always_show_image) gtk_image_menu_item_set_always_show_image __attribute((alias("IA__gtk_image_menu_item_set_always_show_image"), visibility("default")));

#undef gtk_image_menu_item_set_image
extern __typeof (gtk_image_menu_item_set_image) gtk_image_menu_item_set_image __attribute((alias("IA__gtk_image_menu_item_set_image"), visibility("default")));

#undef gtk_image_menu_item_set_use_stock
extern __typeof (gtk_image_menu_item_set_use_stock) gtk_image_menu_item_set_use_stock __attribute((alias("IA__gtk_image_menu_item_set_use_stock"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_IM_CONTEXT_H__)
#if IN_FILE(__GTK_IM_CONTEXT_C__)
#undef gtk_im_context_delete_surrounding
extern __typeof (gtk_im_context_delete_surrounding) gtk_im_context_delete_surrounding __attribute((alias("IA__gtk_im_context_delete_surrounding"), visibility("default")));

#undef gtk_im_context_filter_keypress
extern __typeof (gtk_im_context_filter_keypress) gtk_im_context_filter_keypress __attribute((alias("IA__gtk_im_context_filter_keypress"), visibility("default")));

#undef gtk_im_context_focus_in
extern __typeof (gtk_im_context_focus_in) gtk_im_context_focus_in __attribute((alias("IA__gtk_im_context_focus_in"), visibility("default")));

#undef gtk_im_context_focus_out
extern __typeof (gtk_im_context_focus_out) gtk_im_context_focus_out __attribute((alias("IA__gtk_im_context_focus_out"), visibility("default")));

#undef gtk_im_context_get_preedit_string
extern __typeof (gtk_im_context_get_preedit_string) gtk_im_context_get_preedit_string __attribute((alias("IA__gtk_im_context_get_preedit_string"), visibility("default")));

#undef gtk_im_context_get_surrounding
extern __typeof (gtk_im_context_get_surrounding) gtk_im_context_get_surrounding __attribute((alias("IA__gtk_im_context_get_surrounding"), visibility("default")));

#undef gtk_im_context_get_type
extern __typeof (gtk_im_context_get_type) gtk_im_context_get_type __attribute((alias("IA__gtk_im_context_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_im_context_reset
extern __typeof (gtk_im_context_reset) gtk_im_context_reset __attribute((alias("IA__gtk_im_context_reset"), visibility("default")));

#undef gtk_im_context_set_client_window
extern __typeof (gtk_im_context_set_client_window) gtk_im_context_set_client_window __attribute((alias("IA__gtk_im_context_set_client_window"), visibility("default")));

#undef gtk_im_context_set_cursor_location
extern __typeof (gtk_im_context_set_cursor_location) gtk_im_context_set_cursor_location __attribute((alias("IA__gtk_im_context_set_cursor_location"), visibility("default")));

#undef gtk_im_context_set_surrounding
extern __typeof (gtk_im_context_set_surrounding) gtk_im_context_set_surrounding __attribute((alias("IA__gtk_im_context_set_surrounding"), visibility("default")));

#undef gtk_im_context_set_use_preedit
extern __typeof (gtk_im_context_set_use_preedit) gtk_im_context_set_use_preedit __attribute((alias("IA__gtk_im_context_set_use_preedit"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_IM_CONTEXT_SIMPLE_H__)
#if IN_FILE(__GTK_IM_CONTEXT_SIMPLE_C__)
#undef gtk_im_context_simple_add_table
extern __typeof (gtk_im_context_simple_add_table) gtk_im_context_simple_add_table __attribute((alias("IA__gtk_im_context_simple_add_table"), visibility("default")));

#undef gtk_im_context_simple_get_type
extern __typeof (gtk_im_context_simple_get_type) gtk_im_context_simple_get_type __attribute((alias("IA__gtk_im_context_simple_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_im_context_simple_new
extern __typeof (gtk_im_context_simple_new) gtk_im_context_simple_new __attribute((alias("IA__gtk_im_context_simple_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_IM_MULTICONTEXT_H__)
#if IN_FILE(__GTK_IM_MULTICONTEXT_C__)
#undef gtk_im_multicontext_append_menuitems
extern __typeof (gtk_im_multicontext_append_menuitems) gtk_im_multicontext_append_menuitems __attribute((alias("IA__gtk_im_multicontext_append_menuitems"), visibility("default")));

#undef gtk_im_multicontext_get_context_id
extern __typeof (gtk_im_multicontext_get_context_id) gtk_im_multicontext_get_context_id __attribute((alias("IA__gtk_im_multicontext_get_context_id"), visibility("default")));

#undef gtk_im_multicontext_set_context_id
extern __typeof (gtk_im_multicontext_set_context_id) gtk_im_multicontext_set_context_id __attribute((alias("IA__gtk_im_multicontext_set_context_id"), visibility("default")));

#undef gtk_im_multicontext_get_type
extern __typeof (gtk_im_multicontext_get_type) gtk_im_multicontext_get_type __attribute((alias("IA__gtk_im_multicontext_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_im_multicontext_new
extern __typeof (gtk_im_multicontext_new) gtk_im_multicontext_new __attribute((alias("IA__gtk_im_multicontext_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_INVISIBLE_H__)
#if IN_FILE(__GTK_INVISIBLE_C__)
#undef gtk_invisible_get_screen
extern __typeof (gtk_invisible_get_screen) gtk_invisible_get_screen __attribute((alias("IA__gtk_invisible_get_screen"), visibility("default")));

#undef gtk_invisible_get_type
extern __typeof (gtk_invisible_get_type) gtk_invisible_get_type __attribute((alias("IA__gtk_invisible_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_invisible_new
extern __typeof (gtk_invisible_new) gtk_invisible_new __attribute((alias("IA__gtk_invisible_new"), visibility("default")));

#undef gtk_invisible_new_for_screen
extern __typeof (gtk_invisible_new_for_screen) gtk_invisible_new_for_screen __attribute((alias("IA__gtk_invisible_new_for_screen"), visibility("default")));

#undef gtk_invisible_set_screen
extern __typeof (gtk_invisible_set_screen) gtk_invisible_set_screen __attribute((alias("IA__gtk_invisible_set_screen"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_ITEM_H__)
#if IN_FILE(__GTK_ITEM_C__)
#undef gtk_item_get_type
extern __typeof (gtk_item_get_type) gtk_item_get_type __attribute((alias("IA__gtk_item_get_type"), visibility("default"))) G_GNUC_CONST;

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_item_deselect
extern __typeof (gtk_item_deselect) gtk_item_deselect __attribute((alias("IA__gtk_item_deselect"), visibility("default")));

#undef gtk_item_select
extern __typeof (gtk_item_select) gtk_item_select __attribute((alias("IA__gtk_item_select"), visibility("default")));

#undef gtk_item_toggle
extern __typeof (gtk_item_toggle) gtk_item_toggle __attribute((alias("IA__gtk_item_toggle"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_ITEM_FACTORY_H__)
#if IN_FILE(__GTK_ITEM_FACTORY_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_item_factories_path_delete
extern __typeof (gtk_item_factories_path_delete) gtk_item_factories_path_delete __attribute((alias("IA__gtk_item_factories_path_delete"), visibility("default")));

#undef gtk_item_factory_add_foreign
extern __typeof (gtk_item_factory_add_foreign) gtk_item_factory_add_foreign __attribute((alias("IA__gtk_item_factory_add_foreign"), visibility("default")));

#undef gtk_item_factory_construct
extern __typeof (gtk_item_factory_construct) gtk_item_factory_construct __attribute((alias("IA__gtk_item_factory_construct"), visibility("default")));

#undef gtk_item_factory_create_item
extern __typeof (gtk_item_factory_create_item) gtk_item_factory_create_item __attribute((alias("IA__gtk_item_factory_create_item"), visibility("default")));

#undef gtk_item_factory_create_items
extern __typeof (gtk_item_factory_create_items) gtk_item_factory_create_items __attribute((alias("IA__gtk_item_factory_create_items"), visibility("default")));

#undef gtk_item_factory_create_items_ac
extern __typeof (gtk_item_factory_create_items_ac) gtk_item_factory_create_items_ac __attribute((alias("IA__gtk_item_factory_create_items_ac"), visibility("default")));

#undef gtk_item_factory_create_menu_entries
extern __typeof (gtk_item_factory_create_menu_entries) gtk_item_factory_create_menu_entries __attribute((alias("IA__gtk_item_factory_create_menu_entries"), visibility("default")));

#undef gtk_item_factory_delete_entries
extern __typeof (gtk_item_factory_delete_entries) gtk_item_factory_delete_entries __attribute((alias("IA__gtk_item_factory_delete_entries"), visibility("default")));

#undef gtk_item_factory_delete_entry
extern __typeof (gtk_item_factory_delete_entry) gtk_item_factory_delete_entry __attribute((alias("IA__gtk_item_factory_delete_entry"), visibility("default")));

#undef gtk_item_factory_delete_item
extern __typeof (gtk_item_factory_delete_item) gtk_item_factory_delete_item __attribute((alias("IA__gtk_item_factory_delete_item"), visibility("default")));

#undef gtk_item_factory_from_path
extern __typeof (gtk_item_factory_from_path) gtk_item_factory_from_path __attribute((alias("IA__gtk_item_factory_from_path"), visibility("default")));

#undef gtk_item_factory_from_widget
extern __typeof (gtk_item_factory_from_widget) gtk_item_factory_from_widget __attribute((alias("IA__gtk_item_factory_from_widget"), visibility("default")));

#undef gtk_item_factory_get_item
extern __typeof (gtk_item_factory_get_item) gtk_item_factory_get_item __attribute((alias("IA__gtk_item_factory_get_item"), visibility("default")));

#undef gtk_item_factory_get_item_by_action
extern __typeof (gtk_item_factory_get_item_by_action) gtk_item_factory_get_item_by_action __attribute((alias("IA__gtk_item_factory_get_item_by_action"), visibility("default")));

#undef gtk_item_factory_get_type
extern __typeof (gtk_item_factory_get_type) gtk_item_factory_get_type __attribute((alias("IA__gtk_item_factory_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_item_factory_get_widget
extern __typeof (gtk_item_factory_get_widget) gtk_item_factory_get_widget __attribute((alias("IA__gtk_item_factory_get_widget"), visibility("default")));

#undef gtk_item_factory_get_widget_by_action
extern __typeof (gtk_item_factory_get_widget_by_action) gtk_item_factory_get_widget_by_action __attribute((alias("IA__gtk_item_factory_get_widget_by_action"), visibility("default")));

#undef gtk_item_factory_new
extern __typeof (gtk_item_factory_new) gtk_item_factory_new __attribute((alias("IA__gtk_item_factory_new"), visibility("default")));

#undef gtk_item_factory_path_from_widget
extern __typeof (gtk_item_factory_path_from_widget) gtk_item_factory_path_from_widget __attribute((alias("IA__gtk_item_factory_path_from_widget"), visibility("default")));

#undef gtk_item_factory_popup
extern __typeof (gtk_item_factory_popup) gtk_item_factory_popup __attribute((alias("IA__gtk_item_factory_popup"), visibility("default")));

#undef gtk_item_factory_popup_data
extern __typeof (gtk_item_factory_popup_data) gtk_item_factory_popup_data __attribute((alias("IA__gtk_item_factory_popup_data"), visibility("default")));

#undef gtk_item_factory_popup_data_from_widget
extern __typeof (gtk_item_factory_popup_data_from_widget) gtk_item_factory_popup_data_from_widget __attribute((alias("IA__gtk_item_factory_popup_data_from_widget"), visibility("default")));

#undef gtk_item_factory_popup_with_data
extern __typeof (gtk_item_factory_popup_with_data) gtk_item_factory_popup_with_data __attribute((alias("IA__gtk_item_factory_popup_with_data"), visibility("default")));

#undef gtk_item_factory_set_translate_func
extern __typeof (gtk_item_factory_set_translate_func) gtk_item_factory_set_translate_func __attribute((alias("IA__gtk_item_factory_set_translate_func"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_LABEL_H__)
#if IN_FILE(__GTK_LABEL_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_label_get
extern __typeof (gtk_label_get) gtk_label_get __attribute((alias("IA__gtk_label_get"), visibility("default")));

#undef gtk_label_parse_uline
extern __typeof (gtk_label_parse_uline) gtk_label_parse_uline __attribute((alias("IA__gtk_label_parse_uline"), visibility("default")));

#endif
#undef gtk_label_get_angle
extern __typeof (gtk_label_get_angle) gtk_label_get_angle __attribute((alias("IA__gtk_label_get_angle"), visibility("default")));

#undef gtk_label_get_attributes
extern __typeof (gtk_label_get_attributes) gtk_label_get_attributes __attribute((alias("IA__gtk_label_get_attributes"), visibility("default")));

#undef gtk_label_get_ellipsize
extern __typeof (gtk_label_get_ellipsize) gtk_label_get_ellipsize __attribute((alias("IA__gtk_label_get_ellipsize"), visibility("default")));

#undef gtk_label_get_justify
extern __typeof (gtk_label_get_justify) gtk_label_get_justify __attribute((alias("IA__gtk_label_get_justify"), visibility("default")));

#undef gtk_label_get_label
extern __typeof (gtk_label_get_label) gtk_label_get_label __attribute((alias("IA__gtk_label_get_label"), visibility("default")));

#undef gtk_label_get_layout
extern __typeof (gtk_label_get_layout) gtk_label_get_layout __attribute((alias("IA__gtk_label_get_layout"), visibility("default")));

#undef gtk_label_get_layout_offsets
extern __typeof (gtk_label_get_layout_offsets) gtk_label_get_layout_offsets __attribute((alias("IA__gtk_label_get_layout_offsets"), visibility("default")));

#undef gtk_label_get_line_wrap
extern __typeof (gtk_label_get_line_wrap) gtk_label_get_line_wrap __attribute((alias("IA__gtk_label_get_line_wrap"), visibility("default")));

#undef gtk_label_get_line_wrap_mode
extern __typeof (gtk_label_get_line_wrap_mode) gtk_label_get_line_wrap_mode __attribute((alias("IA__gtk_label_get_line_wrap_mode"), visibility("default")));

#undef gtk_label_get_max_width_chars
extern __typeof (gtk_label_get_max_width_chars) gtk_label_get_max_width_chars __attribute((alias("IA__gtk_label_get_max_width_chars"), visibility("default")));

#undef gtk_label_get_mnemonic_keyval
extern __typeof (gtk_label_get_mnemonic_keyval) gtk_label_get_mnemonic_keyval __attribute((alias("IA__gtk_label_get_mnemonic_keyval"), visibility("default")));

#undef gtk_label_get_mnemonic_widget
extern __typeof (gtk_label_get_mnemonic_widget) gtk_label_get_mnemonic_widget __attribute((alias("IA__gtk_label_get_mnemonic_widget"), visibility("default")));

#undef gtk_label_get_selectable
extern __typeof (gtk_label_get_selectable) gtk_label_get_selectable __attribute((alias("IA__gtk_label_get_selectable"), visibility("default")));

#undef gtk_label_get_selection_bounds
extern __typeof (gtk_label_get_selection_bounds) gtk_label_get_selection_bounds __attribute((alias("IA__gtk_label_get_selection_bounds"), visibility("default")));

#undef gtk_label_get_single_line_mode
extern __typeof (gtk_label_get_single_line_mode) gtk_label_get_single_line_mode __attribute((alias("IA__gtk_label_get_single_line_mode"), visibility("default")));

#undef gtk_label_get_text
extern __typeof (gtk_label_get_text) gtk_label_get_text __attribute((alias("IA__gtk_label_get_text"), visibility("default")));

#undef gtk_label_get_type
extern __typeof (gtk_label_get_type) gtk_label_get_type __attribute((alias("IA__gtk_label_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_label_get_use_markup
extern __typeof (gtk_label_get_use_markup) gtk_label_get_use_markup __attribute((alias("IA__gtk_label_get_use_markup"), visibility("default")));

#undef gtk_label_get_use_underline
extern __typeof (gtk_label_get_use_underline) gtk_label_get_use_underline __attribute((alias("IA__gtk_label_get_use_underline"), visibility("default")));

#undef gtk_label_get_width_chars
extern __typeof (gtk_label_get_width_chars) gtk_label_get_width_chars __attribute((alias("IA__gtk_label_get_width_chars"), visibility("default")));

#undef gtk_label_new
extern __typeof (gtk_label_new) gtk_label_new __attribute((alias("IA__gtk_label_new"), visibility("default")));

#undef gtk_label_new_with_mnemonic
extern __typeof (gtk_label_new_with_mnemonic) gtk_label_new_with_mnemonic __attribute((alias("IA__gtk_label_new_with_mnemonic"), visibility("default")));

#undef gtk_label_select_region
extern __typeof (gtk_label_select_region) gtk_label_select_region __attribute((alias("IA__gtk_label_select_region"), visibility("default")));

#undef gtk_label_set_angle
extern __typeof (gtk_label_set_angle) gtk_label_set_angle __attribute((alias("IA__gtk_label_set_angle"), visibility("default")));

#undef gtk_label_set_attributes
extern __typeof (gtk_label_set_attributes) gtk_label_set_attributes __attribute((alias("IA__gtk_label_set_attributes"), visibility("default")));

#undef gtk_label_set_ellipsize
extern __typeof (gtk_label_set_ellipsize) gtk_label_set_ellipsize __attribute((alias("IA__gtk_label_set_ellipsize"), visibility("default")));

#undef gtk_label_set_justify
extern __typeof (gtk_label_set_justify) gtk_label_set_justify __attribute((alias("IA__gtk_label_set_justify"), visibility("default")));

#undef gtk_label_set_label
extern __typeof (gtk_label_set_label) gtk_label_set_label __attribute((alias("IA__gtk_label_set_label"), visibility("default")));

#undef gtk_label_set_line_wrap
extern __typeof (gtk_label_set_line_wrap) gtk_label_set_line_wrap __attribute((alias("IA__gtk_label_set_line_wrap"), visibility("default")));

#undef gtk_label_set_line_wrap_mode
extern __typeof (gtk_label_set_line_wrap_mode) gtk_label_set_line_wrap_mode __attribute((alias("IA__gtk_label_set_line_wrap_mode"), visibility("default")));

#undef gtk_label_set_markup
extern __typeof (gtk_label_set_markup) gtk_label_set_markup __attribute((alias("IA__gtk_label_set_markup"), visibility("default")));

#undef gtk_label_set_markup_with_mnemonic
extern __typeof (gtk_label_set_markup_with_mnemonic) gtk_label_set_markup_with_mnemonic __attribute((alias("IA__gtk_label_set_markup_with_mnemonic"), visibility("default")));

#undef gtk_label_set_max_width_chars
extern __typeof (gtk_label_set_max_width_chars) gtk_label_set_max_width_chars __attribute((alias("IA__gtk_label_set_max_width_chars"), visibility("default")));

#undef gtk_label_set_mnemonic_widget
extern __typeof (gtk_label_set_mnemonic_widget) gtk_label_set_mnemonic_widget __attribute((alias("IA__gtk_label_set_mnemonic_widget"), visibility("default")));

#undef gtk_label_set_pattern
extern __typeof (gtk_label_set_pattern) gtk_label_set_pattern __attribute((alias("IA__gtk_label_set_pattern"), visibility("default")));

#undef gtk_label_set_selectable
extern __typeof (gtk_label_set_selectable) gtk_label_set_selectable __attribute((alias("IA__gtk_label_set_selectable"), visibility("default")));

#undef gtk_label_set_single_line_mode
extern __typeof (gtk_label_set_single_line_mode) gtk_label_set_single_line_mode __attribute((alias("IA__gtk_label_set_single_line_mode"), visibility("default")));

#undef gtk_label_set_text
extern __typeof (gtk_label_set_text) gtk_label_set_text __attribute((alias("IA__gtk_label_set_text"), visibility("default")));

#undef gtk_label_set_text_with_mnemonic
extern __typeof (gtk_label_set_text_with_mnemonic) gtk_label_set_text_with_mnemonic __attribute((alias("IA__gtk_label_set_text_with_mnemonic"), visibility("default")));

#undef gtk_label_set_use_markup
extern __typeof (gtk_label_set_use_markup) gtk_label_set_use_markup __attribute((alias("IA__gtk_label_set_use_markup"), visibility("default")));

#undef gtk_label_set_use_underline
extern __typeof (gtk_label_set_use_underline) gtk_label_set_use_underline __attribute((alias("IA__gtk_label_set_use_underline"), visibility("default")));

#undef gtk_label_set_width_chars
extern __typeof (gtk_label_set_width_chars) gtk_label_set_width_chars __attribute((alias("IA__gtk_label_set_width_chars"), visibility("default")));

#undef gtk_label_get_current_uri
extern __typeof (gtk_label_get_current_uri) gtk_label_get_current_uri __attribute((alias("IA__gtk_label_get_current_uri"), visibility("default")));

#undef gtk_label_set_track_visited_links
extern __typeof (gtk_label_set_track_visited_links) gtk_label_set_track_visited_links __attribute((alias("IA__gtk_label_set_track_visited_links"), visibility("default")));

#undef gtk_label_get_track_visited_links
extern __typeof (gtk_label_get_track_visited_links) gtk_label_get_track_visited_links __attribute((alias("IA__gtk_label_get_track_visited_links"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_LAYOUT_H__)
#if IN_FILE(__GTK_LAYOUT_C__)
#undef gtk_layout_get_bin_window
extern __typeof (gtk_layout_get_bin_window) gtk_layout_get_bin_window __attribute((alias("IA__gtk_layout_get_bin_window"), visibility("default")));

#undef gtk_layout_get_hadjustment
extern __typeof (gtk_layout_get_hadjustment) gtk_layout_get_hadjustment __attribute((alias("IA__gtk_layout_get_hadjustment"), visibility("default")));

#undef gtk_layout_get_size
extern __typeof (gtk_layout_get_size) gtk_layout_get_size __attribute((alias("IA__gtk_layout_get_size"), visibility("default")));

#undef gtk_layout_get_type
extern __typeof (gtk_layout_get_type) gtk_layout_get_type __attribute((alias("IA__gtk_layout_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_layout_get_vadjustment
extern __typeof (gtk_layout_get_vadjustment) gtk_layout_get_vadjustment __attribute((alias("IA__gtk_layout_get_vadjustment"), visibility("default")));

#undef gtk_layout_move
extern __typeof (gtk_layout_move) gtk_layout_move __attribute((alias("IA__gtk_layout_move"), visibility("default")));

#undef gtk_layout_new
extern __typeof (gtk_layout_new) gtk_layout_new __attribute((alias("IA__gtk_layout_new"), visibility("default")));

#undef gtk_layout_put
extern __typeof (gtk_layout_put) gtk_layout_put __attribute((alias("IA__gtk_layout_put"), visibility("default")));

#undef gtk_layout_set_hadjustment
extern __typeof (gtk_layout_set_hadjustment) gtk_layout_set_hadjustment __attribute((alias("IA__gtk_layout_set_hadjustment"), visibility("default")));

#undef gtk_layout_set_size
extern __typeof (gtk_layout_set_size) gtk_layout_set_size __attribute((alias("IA__gtk_layout_set_size"), visibility("default")));

#undef gtk_layout_set_vadjustment
extern __typeof (gtk_layout_set_vadjustment) gtk_layout_set_vadjustment __attribute((alias("IA__gtk_layout_set_vadjustment"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_layout_freeze
extern __typeof (gtk_layout_freeze) gtk_layout_freeze __attribute((alias("IA__gtk_layout_freeze"), visibility("default")));

#undef gtk_layout_thaw
extern __typeof (gtk_layout_thaw) gtk_layout_thaw __attribute((alias("IA__gtk_layout_thaw"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_LINK_BUTTON_H__)
#if IN_FILE(__GTK_LINK_BUTTON_C__)
#undef gtk_link_button_get_type
extern __typeof (gtk_link_button_get_type) gtk_link_button_get_type __attribute((alias("IA__gtk_link_button_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_link_button_new
extern __typeof (gtk_link_button_new) gtk_link_button_new __attribute((alias("IA__gtk_link_button_new"), visibility("default")));

#undef gtk_link_button_new_with_label
extern __typeof (gtk_link_button_new_with_label) gtk_link_button_new_with_label __attribute((alias("IA__gtk_link_button_new_with_label"), visibility("default")));

#undef gtk_link_button_get_uri
extern __typeof (gtk_link_button_get_uri) gtk_link_button_get_uri __attribute((alias("IA__gtk_link_button_get_uri"), visibility("default")));

#undef gtk_link_button_set_uri
extern __typeof (gtk_link_button_set_uri) gtk_link_button_set_uri __attribute((alias("IA__gtk_link_button_set_uri"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_link_button_set_uri_hook
extern __typeof (gtk_link_button_set_uri_hook) gtk_link_button_set_uri_hook __attribute((alias("IA__gtk_link_button_set_uri_hook"), visibility("default")));

#endif
#undef gtk_link_button_get_visited
extern __typeof (gtk_link_button_get_visited) gtk_link_button_get_visited __attribute((alias("IA__gtk_link_button_get_visited"), visibility("default")));

#undef gtk_link_button_set_visited
extern __typeof (gtk_link_button_set_visited) gtk_link_button_set_visited __attribute((alias("IA__gtk_link_button_set_visited"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_LIST_H__)
#if IN_FILE(__GTK_LIST_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_list_append_items
extern __typeof (gtk_list_append_items) gtk_list_append_items __attribute((alias("IA__gtk_list_append_items"), visibility("default")));

#undef gtk_list_child_position
extern __typeof (gtk_list_child_position) gtk_list_child_position __attribute((alias("IA__gtk_list_child_position"), visibility("default")));

#undef gtk_list_clear_items
extern __typeof (gtk_list_clear_items) gtk_list_clear_items __attribute((alias("IA__gtk_list_clear_items"), visibility("default")));

#undef gtk_list_end_drag_selection
extern __typeof (gtk_list_end_drag_selection) gtk_list_end_drag_selection __attribute((alias("IA__gtk_list_end_drag_selection"), visibility("default")));

#undef gtk_list_end_selection
extern __typeof (gtk_list_end_selection) gtk_list_end_selection __attribute((alias("IA__gtk_list_end_selection"), visibility("default")));

#undef gtk_list_extend_selection
extern __typeof (gtk_list_extend_selection) gtk_list_extend_selection __attribute((alias("IA__gtk_list_extend_selection"), visibility("default")));

#undef gtk_list_get_type
extern __typeof (gtk_list_get_type) gtk_list_get_type __attribute((alias("IA__gtk_list_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_list_insert_items
extern __typeof (gtk_list_insert_items) gtk_list_insert_items __attribute((alias("IA__gtk_list_insert_items"), visibility("default")));

#undef gtk_list_new
extern __typeof (gtk_list_new) gtk_list_new __attribute((alias("IA__gtk_list_new"), visibility("default")));

#undef gtk_list_prepend_items
extern __typeof (gtk_list_prepend_items) gtk_list_prepend_items __attribute((alias("IA__gtk_list_prepend_items"), visibility("default")));

#undef gtk_list_remove_items
extern __typeof (gtk_list_remove_items) gtk_list_remove_items __attribute((alias("IA__gtk_list_remove_items"), visibility("default")));

#undef gtk_list_remove_items_no_unref
extern __typeof (gtk_list_remove_items_no_unref) gtk_list_remove_items_no_unref __attribute((alias("IA__gtk_list_remove_items_no_unref"), visibility("default")));

#undef gtk_list_scroll_horizontal
extern __typeof (gtk_list_scroll_horizontal) gtk_list_scroll_horizontal __attribute((alias("IA__gtk_list_scroll_horizontal"), visibility("default")));

#undef gtk_list_scroll_vertical
extern __typeof (gtk_list_scroll_vertical) gtk_list_scroll_vertical __attribute((alias("IA__gtk_list_scroll_vertical"), visibility("default")));

#undef gtk_list_select_all
extern __typeof (gtk_list_select_all) gtk_list_select_all __attribute((alias("IA__gtk_list_select_all"), visibility("default")));

#undef gtk_list_select_child
extern __typeof (gtk_list_select_child) gtk_list_select_child __attribute((alias("IA__gtk_list_select_child"), visibility("default")));

#undef gtk_list_select_item
extern __typeof (gtk_list_select_item) gtk_list_select_item __attribute((alias("IA__gtk_list_select_item"), visibility("default")));

#undef gtk_list_set_selection_mode
extern __typeof (gtk_list_set_selection_mode) gtk_list_set_selection_mode __attribute((alias("IA__gtk_list_set_selection_mode"), visibility("default")));

#undef gtk_list_start_selection
extern __typeof (gtk_list_start_selection) gtk_list_start_selection __attribute((alias("IA__gtk_list_start_selection"), visibility("default")));

#undef gtk_list_toggle_add_mode
extern __typeof (gtk_list_toggle_add_mode) gtk_list_toggle_add_mode __attribute((alias("IA__gtk_list_toggle_add_mode"), visibility("default")));

#undef gtk_list_toggle_focus_row
extern __typeof (gtk_list_toggle_focus_row) gtk_list_toggle_focus_row __attribute((alias("IA__gtk_list_toggle_focus_row"), visibility("default")));

#undef gtk_list_toggle_row
extern __typeof (gtk_list_toggle_row) gtk_list_toggle_row __attribute((alias("IA__gtk_list_toggle_row"), visibility("default")));

#undef gtk_list_undo_selection
extern __typeof (gtk_list_undo_selection) gtk_list_undo_selection __attribute((alias("IA__gtk_list_undo_selection"), visibility("default")));

#undef gtk_list_unselect_all
extern __typeof (gtk_list_unselect_all) gtk_list_unselect_all __attribute((alias("IA__gtk_list_unselect_all"), visibility("default")));

#undef gtk_list_unselect_child
extern __typeof (gtk_list_unselect_child) gtk_list_unselect_child __attribute((alias("IA__gtk_list_unselect_child"), visibility("default")));

#undef gtk_list_unselect_item
extern __typeof (gtk_list_unselect_item) gtk_list_unselect_item __attribute((alias("IA__gtk_list_unselect_item"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_LIST_ITEM_H__)
#if IN_FILE(__GTK_LIST_ITEM_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_list_item_deselect
extern __typeof (gtk_list_item_deselect) gtk_list_item_deselect __attribute((alias("IA__gtk_list_item_deselect"), visibility("default")));

#undef gtk_list_item_get_type
extern __typeof (gtk_list_item_get_type) gtk_list_item_get_type __attribute((alias("IA__gtk_list_item_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_list_item_new
extern __typeof (gtk_list_item_new) gtk_list_item_new __attribute((alias("IA__gtk_list_item_new"), visibility("default")));

#undef gtk_list_item_new_with_label
extern __typeof (gtk_list_item_new_with_label) gtk_list_item_new_with_label __attribute((alias("IA__gtk_list_item_new_with_label"), visibility("default")));

#undef gtk_list_item_select
extern __typeof (gtk_list_item_select) gtk_list_item_select __attribute((alias("IA__gtk_list_item_select"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_LIST_STORE_H__)
#if IN_FILE(__GTK_LIST_STORE_C__)
#undef gtk_list_store_append
extern __typeof (gtk_list_store_append) gtk_list_store_append __attribute((alias("IA__gtk_list_store_append"), visibility("default")));

#undef gtk_list_store_clear
extern __typeof (gtk_list_store_clear) gtk_list_store_clear __attribute((alias("IA__gtk_list_store_clear"), visibility("default")));

#undef gtk_list_store_get_type
extern __typeof (gtk_list_store_get_type) gtk_list_store_get_type __attribute((alias("IA__gtk_list_store_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_list_store_insert
extern __typeof (gtk_list_store_insert) gtk_list_store_insert __attribute((alias("IA__gtk_list_store_insert"), visibility("default")));

#undef gtk_list_store_insert_after
extern __typeof (gtk_list_store_insert_after) gtk_list_store_insert_after __attribute((alias("IA__gtk_list_store_insert_after"), visibility("default")));

#undef gtk_list_store_insert_before
extern __typeof (gtk_list_store_insert_before) gtk_list_store_insert_before __attribute((alias("IA__gtk_list_store_insert_before"), visibility("default")));

#undef gtk_list_store_insert_with_values
extern __typeof (gtk_list_store_insert_with_values) gtk_list_store_insert_with_values __attribute((alias("IA__gtk_list_store_insert_with_values"), visibility("default")));

#undef gtk_list_store_insert_with_valuesv
extern __typeof (gtk_list_store_insert_with_valuesv) gtk_list_store_insert_with_valuesv __attribute((alias("IA__gtk_list_store_insert_with_valuesv"), visibility("default")));

#undef gtk_list_store_iter_is_valid
extern __typeof (gtk_list_store_iter_is_valid) gtk_list_store_iter_is_valid __attribute((alias("IA__gtk_list_store_iter_is_valid"), visibility("default")));

#undef gtk_list_store_move_after
extern __typeof (gtk_list_store_move_after) gtk_list_store_move_after __attribute((alias("IA__gtk_list_store_move_after"), visibility("default")));

#undef gtk_list_store_move_before
extern __typeof (gtk_list_store_move_before) gtk_list_store_move_before __attribute((alias("IA__gtk_list_store_move_before"), visibility("default")));

#undef gtk_list_store_new
extern __typeof (gtk_list_store_new) gtk_list_store_new __attribute((alias("IA__gtk_list_store_new"), visibility("default")));

#undef gtk_list_store_newv
extern __typeof (gtk_list_store_newv) gtk_list_store_newv __attribute((alias("IA__gtk_list_store_newv"), visibility("default")));

#undef gtk_list_store_prepend
extern __typeof (gtk_list_store_prepend) gtk_list_store_prepend __attribute((alias("IA__gtk_list_store_prepend"), visibility("default")));

#undef gtk_list_store_remove
extern __typeof (gtk_list_store_remove) gtk_list_store_remove __attribute((alias("IA__gtk_list_store_remove"), visibility("default")));

#undef gtk_list_store_reorder
extern __typeof (gtk_list_store_reorder) gtk_list_store_reorder __attribute((alias("IA__gtk_list_store_reorder"), visibility("default")));

#undef gtk_list_store_set
extern __typeof (gtk_list_store_set) gtk_list_store_set __attribute((alias("IA__gtk_list_store_set"), visibility("default")));

#undef gtk_list_store_set_column_types
extern __typeof (gtk_list_store_set_column_types) gtk_list_store_set_column_types __attribute((alias("IA__gtk_list_store_set_column_types"), visibility("default")));

#undef gtk_list_store_set_valist
extern __typeof (gtk_list_store_set_valist) gtk_list_store_set_valist __attribute((alias("IA__gtk_list_store_set_valist"), visibility("default")));

#undef gtk_list_store_set_value
extern __typeof (gtk_list_store_set_value) gtk_list_store_set_value __attribute((alias("IA__gtk_list_store_set_value"), visibility("default")));

#undef gtk_list_store_set_valuesv
extern __typeof (gtk_list_store_set_valuesv) gtk_list_store_set_valuesv __attribute((alias("IA__gtk_list_store_set_valuesv"), visibility("default")));

#undef gtk_list_store_swap
extern __typeof (gtk_list_store_swap) gtk_list_store_swap __attribute((alias("IA__gtk_list_store_swap"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_MAIN_H__)
#if IN_FILE(__GTK_MAIN_C__)
#undef gtk_get_option_group
extern __typeof (gtk_get_option_group) gtk_get_option_group __attribute((alias("IA__gtk_get_option_group"), visibility("default")));

#undef gtk_get_current_event
extern __typeof (gtk_get_current_event) gtk_get_current_event __attribute((alias("IA__gtk_get_current_event"), visibility("default")));

#undef gtk_get_current_event_state
extern __typeof (gtk_get_current_event_state) gtk_get_current_event_state __attribute((alias("IA__gtk_get_current_event_state"), visibility("default")));

#undef gtk_get_current_event_time
extern __typeof (gtk_get_current_event_time) gtk_get_current_event_time __attribute((alias("IA__gtk_get_current_event_time"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_exit
extern __typeof (gtk_exit) gtk_exit __attribute((alias("IA__gtk_exit"), visibility("default")));

#undef gtk_set_locale
extern __typeof (gtk_set_locale) gtk_set_locale __attribute((alias("IA__gtk_set_locale"), visibility("default")));

#undef gtk_idle_add
extern __typeof (gtk_idle_add) gtk_idle_add __attribute((alias("IA__gtk_idle_add"), visibility("default")));

#undef gtk_idle_add_full
extern __typeof (gtk_idle_add_full) gtk_idle_add_full __attribute((alias("IA__gtk_idle_add_full"), visibility("default")));

#undef gtk_idle_add_priority
extern __typeof (gtk_idle_add_priority) gtk_idle_add_priority __attribute((alias("IA__gtk_idle_add_priority"), visibility("default")));

#undef gtk_idle_remove
extern __typeof (gtk_idle_remove) gtk_idle_remove __attribute((alias("IA__gtk_idle_remove"), visibility("default")));

#undef gtk_idle_remove_by_data
extern __typeof (gtk_idle_remove_by_data) gtk_idle_remove_by_data __attribute((alias("IA__gtk_idle_remove_by_data"), visibility("default")));

#undef gtk_timeout_add
extern __typeof (gtk_timeout_add) gtk_timeout_add __attribute((alias("IA__gtk_timeout_add"), visibility("default")));

#undef gtk_timeout_add_full
extern __typeof (gtk_timeout_add_full) gtk_timeout_add_full __attribute((alias("IA__gtk_timeout_add_full"), visibility("default")));

#undef gtk_timeout_remove
extern __typeof (gtk_timeout_remove) gtk_timeout_remove __attribute((alias("IA__gtk_timeout_remove"), visibility("default")));

#undef gtk_input_add_full
extern __typeof (gtk_input_add_full) gtk_input_add_full __attribute((alias("IA__gtk_input_add_full"), visibility("default")));

#undef gtk_input_remove
extern __typeof (gtk_input_remove) gtk_input_remove __attribute((alias("IA__gtk_input_remove"), visibility("default")));

#endif
#undef gtk_false
extern __typeof (gtk_false) gtk_false __attribute((alias("IA__gtk_false"), visibility("default"))) G_GNUC_CONST;

#undef gtk_true
extern __typeof (gtk_true) gtk_true __attribute((alias("IA__gtk_true"), visibility("default"))) G_GNUC_CONST;

#undef gtk_events_pending
extern __typeof (gtk_events_pending) gtk_events_pending __attribute((alias("IA__gtk_events_pending"), visibility("default")));

#undef gtk_disable_setlocale
extern __typeof (gtk_disable_setlocale) gtk_disable_setlocale __attribute((alias("IA__gtk_disable_setlocale"), visibility("default")));

#undef gtk_check_version
extern __typeof (gtk_check_version) gtk_check_version __attribute((alias("IA__gtk_check_version"), visibility("default")));

#undef gtk_get_default_language
extern __typeof (gtk_get_default_language) gtk_get_default_language __attribute((alias("IA__gtk_get_default_language"), visibility("default")));

#undef gtk_get_event_widget
extern __typeof (gtk_get_event_widget) gtk_get_event_widget __attribute((alias("IA__gtk_get_event_widget"), visibility("default")));

#undef gtk_grab_add
extern __typeof (gtk_grab_add) gtk_grab_add __attribute((alias("IA__gtk_grab_add"), visibility("default")));

#undef gtk_grab_get_current
extern __typeof (gtk_grab_get_current) gtk_grab_get_current __attribute((alias("IA__gtk_grab_get_current"), visibility("default")));

#undef gtk_grab_remove
extern __typeof (gtk_grab_remove) gtk_grab_remove __attribute((alias("IA__gtk_grab_remove"), visibility("default")));

#undef gtk_propagate_event
extern __typeof (gtk_propagate_event) gtk_propagate_event __attribute((alias("IA__gtk_propagate_event"), visibility("default")));

#undef gtk_quit_add
extern __typeof (gtk_quit_add) gtk_quit_add __attribute((alias("IA__gtk_quit_add"), visibility("default")));

#undef gtk_quit_add_destroy
extern __typeof (gtk_quit_add_destroy) gtk_quit_add_destroy __attribute((alias("IA__gtk_quit_add_destroy"), visibility("default")));

#undef gtk_quit_add_full
extern __typeof (gtk_quit_add_full) gtk_quit_add_full __attribute((alias("IA__gtk_quit_add_full"), visibility("default")));

#undef gtk_quit_remove
extern __typeof (gtk_quit_remove) gtk_quit_remove __attribute((alias("IA__gtk_quit_remove"), visibility("default")));

#undef gtk_quit_remove_by_data
extern __typeof (gtk_quit_remove_by_data) gtk_quit_remove_by_data __attribute((alias("IA__gtk_quit_remove_by_data"), visibility("default")));

#undef gtk_key_snooper_install
extern __typeof (gtk_key_snooper_install) gtk_key_snooper_install __attribute((alias("IA__gtk_key_snooper_install"), visibility("default")));

#undef gtk_key_snooper_remove
extern __typeof (gtk_key_snooper_remove) gtk_key_snooper_remove __attribute((alias("IA__gtk_key_snooper_remove"), visibility("default")));

#undef gtk_init
extern __typeof (gtk_init) gtk_init __attribute((alias("IA__gtk_init"), visibility("default")));

#ifdef G_OS_WIN32
#undef gtk_init_abi_check
extern __typeof (gtk_init_abi_check) gtk_init_abi_check __attribute((alias("IA__gtk_init_abi_check"), visibility("default")));

#endif
#undef gtk_init_add
extern __typeof (gtk_init_add) gtk_init_add __attribute((alias("IA__gtk_init_add"), visibility("default")));

#undef gtk_init_check
extern __typeof (gtk_init_check) gtk_init_check __attribute((alias("IA__gtk_init_check"), visibility("default")));

#ifdef G_OS_WIN32
#undef gtk_init_check_abi_check
extern __typeof (gtk_init_check_abi_check) gtk_init_check_abi_check __attribute((alias("IA__gtk_init_check_abi_check"), visibility("default")));

#endif
#undef gtk_init_with_args
extern __typeof (gtk_init_with_args) gtk_init_with_args __attribute((alias("IA__gtk_init_with_args"), visibility("default")));

#undef gtk_parse_args
extern __typeof (gtk_parse_args) gtk_parse_args __attribute((alias("IA__gtk_parse_args"), visibility("default")));

#undef gtk_main
extern __typeof (gtk_main) gtk_main __attribute((alias("IA__gtk_main"), visibility("default")));

#undef gtk_main_do_event
extern __typeof (gtk_main_do_event) gtk_main_do_event __attribute((alias("IA__gtk_main_do_event"), visibility("default")));

#undef gtk_main_iteration
extern __typeof (gtk_main_iteration) gtk_main_iteration __attribute((alias("IA__gtk_main_iteration"), visibility("default")));

#undef gtk_main_iteration_do
extern __typeof (gtk_main_iteration_do) gtk_main_iteration_do __attribute((alias("IA__gtk_main_iteration_do"), visibility("default")));

#undef gtk_main_level
extern __typeof (gtk_main_level) gtk_main_level __attribute((alias("IA__gtk_main_level"), visibility("default")));

#undef gtk_main_quit
extern __typeof (gtk_main_quit) gtk_main_quit __attribute((alias("IA__gtk_main_quit"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_INPUTDIALOG_H__)
#if IN_FILE(__GTK_INPUTDIALOG_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_input_dialog_get_type
extern __typeof (gtk_input_dialog_get_type) gtk_input_dialog_get_type __attribute((alias("IA__gtk_input_dialog_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_input_dialog_new
extern __typeof (gtk_input_dialog_new) gtk_input_dialog_new __attribute((alias("IA__gtk_input_dialog_new"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__gtk_marshal_MARSHAL_H__) || IN_HEADER(__GTK_MARSHAL_MARSHAL_H__)
#if IN_FILE(__gtk_marshal_MARSHAL_C__) || IN_FILE(__GTK_MARSHAL_MARSHAL_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_marshal_BOOLEAN__POINTER
extern __typeof (gtk_marshal_BOOLEAN__POINTER) gtk_marshal_BOOLEAN__POINTER __attribute((alias("IA__gtk_marshal_BOOLEAN__POINTER"), visibility("default")));

#undef gtk_marshal_BOOLEAN__POINTER_INT_INT
extern __typeof (gtk_marshal_BOOLEAN__POINTER_INT_INT) gtk_marshal_BOOLEAN__POINTER_INT_INT __attribute((alias("IA__gtk_marshal_BOOLEAN__POINTER_INT_INT"), visibility("default")));

#undef gtk_marshal_BOOLEAN__POINTER_INT_INT_UINT
extern __typeof (gtk_marshal_BOOLEAN__POINTER_INT_INT_UINT) gtk_marshal_BOOLEAN__POINTER_INT_INT_UINT __attribute((alias("IA__gtk_marshal_BOOLEAN__POINTER_INT_INT_UINT"), visibility("default")));

#undef gtk_marshal_BOOLEAN__POINTER_POINTER_INT_INT
extern __typeof (gtk_marshal_BOOLEAN__POINTER_POINTER_INT_INT) gtk_marshal_BOOLEAN__POINTER_POINTER_INT_INT __attribute((alias("IA__gtk_marshal_BOOLEAN__POINTER_POINTER_INT_INT"), visibility("default")));

#undef gtk_marshal_BOOLEAN__POINTER_STRING_STRING_POINTER
extern __typeof (gtk_marshal_BOOLEAN__POINTER_STRING_STRING_POINTER) gtk_marshal_BOOLEAN__POINTER_STRING_STRING_POINTER __attribute((alias("IA__gtk_marshal_BOOLEAN__POINTER_STRING_STRING_POINTER"), visibility("default")));

#undef gtk_marshal_BOOLEAN__VOID
extern __typeof (gtk_marshal_BOOLEAN__VOID) gtk_marshal_BOOLEAN__VOID __attribute((alias("IA__gtk_marshal_BOOLEAN__VOID"), visibility("default")));

#undef gtk_marshal_ENUM__ENUM
extern __typeof (gtk_marshal_ENUM__ENUM) gtk_marshal_ENUM__ENUM __attribute((alias("IA__gtk_marshal_ENUM__ENUM"), visibility("default")));

#undef gtk_marshal_INT__POINTER
extern __typeof (gtk_marshal_INT__POINTER) gtk_marshal_INT__POINTER __attribute((alias("IA__gtk_marshal_INT__POINTER"), visibility("default")));

#undef gtk_marshal_INT__POINTER_CHAR_CHAR
extern __typeof (gtk_marshal_INT__POINTER_CHAR_CHAR) gtk_marshal_INT__POINTER_CHAR_CHAR __attribute((alias("IA__gtk_marshal_INT__POINTER_CHAR_CHAR"), visibility("default")));

#undef gtk_marshal_VOID__ENUM_FLOAT
extern __typeof (gtk_marshal_VOID__ENUM_FLOAT) gtk_marshal_VOID__ENUM_FLOAT __attribute((alias("IA__gtk_marshal_VOID__ENUM_FLOAT"), visibility("default")));

#undef gtk_marshal_VOID__ENUM_FLOAT_BOOLEAN
extern __typeof (gtk_marshal_VOID__ENUM_FLOAT_BOOLEAN) gtk_marshal_VOID__ENUM_FLOAT_BOOLEAN __attribute((alias("IA__gtk_marshal_VOID__ENUM_FLOAT_BOOLEAN"), visibility("default")));

#undef gtk_marshal_VOID__INT_INT
extern __typeof (gtk_marshal_VOID__INT_INT) gtk_marshal_VOID__INT_INT __attribute((alias("IA__gtk_marshal_VOID__INT_INT"), visibility("default")));

#undef gtk_marshal_VOID__INT_INT_POINTER
extern __typeof (gtk_marshal_VOID__INT_INT_POINTER) gtk_marshal_VOID__INT_INT_POINTER __attribute((alias("IA__gtk_marshal_VOID__INT_INT_POINTER"), visibility("default")));

#undef gtk_marshal_VOID__POINTER_INT
extern __typeof (gtk_marshal_VOID__POINTER_INT) gtk_marshal_VOID__POINTER_INT __attribute((alias("IA__gtk_marshal_VOID__POINTER_INT"), visibility("default")));

#undef gtk_marshal_VOID__POINTER_INT_INT_POINTER_UINT_UINT
extern __typeof (gtk_marshal_VOID__POINTER_INT_INT_POINTER_UINT_UINT) gtk_marshal_VOID__POINTER_INT_INT_POINTER_UINT_UINT __attribute((alias("IA__gtk_marshal_VOID__POINTER_INT_INT_POINTER_UINT_UINT"), visibility("default")));

#undef gtk_marshal_VOID__POINTER_POINTER
extern __typeof (gtk_marshal_VOID__POINTER_POINTER) gtk_marshal_VOID__POINTER_POINTER __attribute((alias("IA__gtk_marshal_VOID__POINTER_POINTER"), visibility("default")));

#undef gtk_marshal_VOID__POINTER_POINTER_POINTER
extern __typeof (gtk_marshal_VOID__POINTER_POINTER_POINTER) gtk_marshal_VOID__POINTER_POINTER_POINTER __attribute((alias("IA__gtk_marshal_VOID__POINTER_POINTER_POINTER"), visibility("default")));

#undef gtk_marshal_VOID__POINTER_POINTER_UINT_UINT
extern __typeof (gtk_marshal_VOID__POINTER_POINTER_UINT_UINT) gtk_marshal_VOID__POINTER_POINTER_UINT_UINT __attribute((alias("IA__gtk_marshal_VOID__POINTER_POINTER_UINT_UINT"), visibility("default")));

#undef gtk_marshal_VOID__POINTER_STRING_STRING
extern __typeof (gtk_marshal_VOID__POINTER_STRING_STRING) gtk_marshal_VOID__POINTER_STRING_STRING __attribute((alias("IA__gtk_marshal_VOID__POINTER_STRING_STRING"), visibility("default")));

#undef gtk_marshal_VOID__POINTER_UINT
extern __typeof (gtk_marshal_VOID__POINTER_UINT) gtk_marshal_VOID__POINTER_UINT __attribute((alias("IA__gtk_marshal_VOID__POINTER_UINT"), visibility("default")));

#undef gtk_marshal_VOID__POINTER_UINT_ENUM
extern __typeof (gtk_marshal_VOID__POINTER_UINT_ENUM) gtk_marshal_VOID__POINTER_UINT_ENUM __attribute((alias("IA__gtk_marshal_VOID__POINTER_UINT_ENUM"), visibility("default")));

#undef gtk_marshal_VOID__POINTER_UINT_UINT
extern __typeof (gtk_marshal_VOID__POINTER_UINT_UINT) gtk_marshal_VOID__POINTER_UINT_UINT __attribute((alias("IA__gtk_marshal_VOID__POINTER_UINT_UINT"), visibility("default")));

#undef gtk_marshal_VOID__STRING_INT_POINTER
extern __typeof (gtk_marshal_VOID__STRING_INT_POINTER) gtk_marshal_VOID__STRING_INT_POINTER __attribute((alias("IA__gtk_marshal_VOID__STRING_INT_POINTER"), visibility("default")));

#undef gtk_marshal_VOID__UINT_POINTER_UINT_ENUM_ENUM_POINTER
extern __typeof (gtk_marshal_VOID__UINT_POINTER_UINT_ENUM_ENUM_POINTER) gtk_marshal_VOID__UINT_POINTER_UINT_ENUM_ENUM_POINTER __attribute((alias("IA__gtk_marshal_VOID__UINT_POINTER_UINT_ENUM_ENUM_POINTER"), visibility("default")));

#undef gtk_marshal_VOID__UINT_POINTER_UINT_UINT_ENUM
extern __typeof (gtk_marshal_VOID__UINT_POINTER_UINT_UINT_ENUM) gtk_marshal_VOID__UINT_POINTER_UINT_UINT_ENUM __attribute((alias("IA__gtk_marshal_VOID__UINT_POINTER_UINT_UINT_ENUM"), visibility("default")));

#undef gtk_marshal_VOID__UINT_STRING
extern __typeof (gtk_marshal_VOID__UINT_STRING) gtk_marshal_VOID__UINT_STRING __attribute((alias("IA__gtk_marshal_VOID__UINT_STRING"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_MENU_BAR_H__)
#if IN_FILE(__GTK_MENU_BAR_C__)
#undef gtk_menu_bar_get_type
extern __typeof (gtk_menu_bar_get_type) gtk_menu_bar_get_type __attribute((alias("IA__gtk_menu_bar_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_menu_bar_new
extern __typeof (gtk_menu_bar_new) gtk_menu_bar_new __attribute((alias("IA__gtk_menu_bar_new"), visibility("default")));

#undef gtk_menu_bar_get_child_pack_direction
extern __typeof (gtk_menu_bar_get_child_pack_direction) gtk_menu_bar_get_child_pack_direction __attribute((alias("IA__gtk_menu_bar_get_child_pack_direction"), visibility("default")));

#undef gtk_menu_bar_set_child_pack_direction
extern __typeof (gtk_menu_bar_set_child_pack_direction) gtk_menu_bar_set_child_pack_direction __attribute((alias("IA__gtk_menu_bar_set_child_pack_direction"), visibility("default")));

#undef gtk_menu_bar_get_pack_direction
extern __typeof (gtk_menu_bar_get_pack_direction) gtk_menu_bar_get_pack_direction __attribute((alias("IA__gtk_menu_bar_get_pack_direction"), visibility("default")));

#undef gtk_menu_bar_set_pack_direction
extern __typeof (gtk_menu_bar_set_pack_direction) gtk_menu_bar_set_pack_direction __attribute((alias("IA__gtk_menu_bar_set_pack_direction"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_MENU_H__)
#if IN_FILE(__GTK_MENU_C__)
#undef gtk_menu_attach
extern __typeof (gtk_menu_attach) gtk_menu_attach __attribute((alias("IA__gtk_menu_attach"), visibility("default")));

#undef gtk_menu_attach_to_widget
extern __typeof (gtk_menu_attach_to_widget) gtk_menu_attach_to_widget __attribute((alias("IA__gtk_menu_attach_to_widget"), visibility("default")));

#undef gtk_menu_detach
extern __typeof (gtk_menu_detach) gtk_menu_detach __attribute((alias("IA__gtk_menu_detach"), visibility("default")));

#undef gtk_menu_get_accel_group
extern __typeof (gtk_menu_get_accel_group) gtk_menu_get_accel_group __attribute((alias("IA__gtk_menu_get_accel_group"), visibility("default")));

#undef gtk_menu_get_accel_path
extern __typeof (gtk_menu_get_accel_path) gtk_menu_get_accel_path __attribute((alias("IA__gtk_menu_get_accel_path"), visibility("default")));

#undef gtk_menu_get_active
extern __typeof (gtk_menu_get_active) gtk_menu_get_active __attribute((alias("IA__gtk_menu_get_active"), visibility("default")));

#undef gtk_menu_get_monitor
extern __typeof (gtk_menu_get_monitor) gtk_menu_get_monitor __attribute((alias("IA__gtk_menu_get_monitor"), visibility("default")));

#undef gtk_menu_get_attach_widget
extern __typeof (gtk_menu_get_attach_widget) gtk_menu_get_attach_widget __attribute((alias("IA__gtk_menu_get_attach_widget"), visibility("default")));

#undef gtk_menu_get_for_attach_widget
extern __typeof (gtk_menu_get_for_attach_widget) gtk_menu_get_for_attach_widget __attribute((alias("IA__gtk_menu_get_for_attach_widget"), visibility("default")));

#undef gtk_menu_get_reserve_toggle_size
extern __typeof (gtk_menu_get_reserve_toggle_size) gtk_menu_get_reserve_toggle_size __attribute((alias("IA__gtk_menu_get_reserve_toggle_size"), visibility("default")));

#undef gtk_menu_get_tearoff_state
extern __typeof (gtk_menu_get_tearoff_state) gtk_menu_get_tearoff_state __attribute((alias("IA__gtk_menu_get_tearoff_state"), visibility("default")));

#undef gtk_menu_get_title
extern __typeof (gtk_menu_get_title) gtk_menu_get_title __attribute((alias("IA__gtk_menu_get_title"), visibility("default")));

#undef gtk_menu_get_type
extern __typeof (gtk_menu_get_type) gtk_menu_get_type __attribute((alias("IA__gtk_menu_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_menu_new
extern __typeof (gtk_menu_new) gtk_menu_new __attribute((alias("IA__gtk_menu_new"), visibility("default")));

#undef gtk_menu_popdown
extern __typeof (gtk_menu_popdown) gtk_menu_popdown __attribute((alias("IA__gtk_menu_popdown"), visibility("default")));

#undef gtk_menu_popup
extern __typeof (gtk_menu_popup) gtk_menu_popup __attribute((alias("IA__gtk_menu_popup"), visibility("default")));

#undef gtk_menu_reorder_child
extern __typeof (gtk_menu_reorder_child) gtk_menu_reorder_child __attribute((alias("IA__gtk_menu_reorder_child"), visibility("default")));

#undef gtk_menu_reposition
extern __typeof (gtk_menu_reposition) gtk_menu_reposition __attribute((alias("IA__gtk_menu_reposition"), visibility("default")));

#undef gtk_menu_set_accel_group
extern __typeof (gtk_menu_set_accel_group) gtk_menu_set_accel_group __attribute((alias("IA__gtk_menu_set_accel_group"), visibility("default")));

#undef gtk_menu_set_accel_path
extern __typeof (gtk_menu_set_accel_path) gtk_menu_set_accel_path __attribute((alias("IA__gtk_menu_set_accel_path"), visibility("default")));

#undef gtk_menu_set_active
extern __typeof (gtk_menu_set_active) gtk_menu_set_active __attribute((alias("IA__gtk_menu_set_active"), visibility("default")));

#undef gtk_menu_set_monitor
extern __typeof (gtk_menu_set_monitor) gtk_menu_set_monitor __attribute((alias("IA__gtk_menu_set_monitor"), visibility("default")));

#undef gtk_menu_set_reserve_toggle_size
extern __typeof (gtk_menu_set_reserve_toggle_size) gtk_menu_set_reserve_toggle_size __attribute((alias("IA__gtk_menu_set_reserve_toggle_size"), visibility("default")));

#undef gtk_menu_set_screen
extern __typeof (gtk_menu_set_screen) gtk_menu_set_screen __attribute((alias("IA__gtk_menu_set_screen"), visibility("default")));

#undef gtk_menu_set_tearoff_state
extern __typeof (gtk_menu_set_tearoff_state) gtk_menu_set_tearoff_state __attribute((alias("IA__gtk_menu_set_tearoff_state"), visibility("default")));

#undef gtk_menu_set_title
extern __typeof (gtk_menu_set_title) gtk_menu_set_title __attribute((alias("IA__gtk_menu_set_title"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_MENU_ITEM_H__)
#if IN_FILE(__GTK_MENU_ITEM_C__)
#undef gtk_menu_item_activate
extern __typeof (gtk_menu_item_activate) gtk_menu_item_activate __attribute((alias("IA__gtk_menu_item_activate"), visibility("default")));

#undef gtk_menu_item_deselect
extern __typeof (gtk_menu_item_deselect) gtk_menu_item_deselect __attribute((alias("IA__gtk_menu_item_deselect"), visibility("default")));

#undef gtk_menu_item_get_accel_path
extern __typeof (gtk_menu_item_get_accel_path) gtk_menu_item_get_accel_path __attribute((alias("IA__gtk_menu_item_get_accel_path"), visibility("default")));

#undef gtk_menu_item_get_label
extern __typeof (gtk_menu_item_get_label) gtk_menu_item_get_label __attribute((alias("IA__gtk_menu_item_get_label"), visibility("default")));

#undef gtk_menu_item_get_right_justified
extern __typeof (gtk_menu_item_get_right_justified) gtk_menu_item_get_right_justified __attribute((alias("IA__gtk_menu_item_get_right_justified"), visibility("default")));

#undef gtk_menu_item_get_submenu
extern __typeof (gtk_menu_item_get_submenu) gtk_menu_item_get_submenu __attribute((alias("IA__gtk_menu_item_get_submenu"), visibility("default")));

#undef gtk_menu_item_get_type
extern __typeof (gtk_menu_item_get_type) gtk_menu_item_get_type __attribute((alias("IA__gtk_menu_item_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_menu_item_get_use_underline
extern __typeof (gtk_menu_item_get_use_underline) gtk_menu_item_get_use_underline __attribute((alias("IA__gtk_menu_item_get_use_underline"), visibility("default")));

#undef gtk_menu_item_new
extern __typeof (gtk_menu_item_new) gtk_menu_item_new __attribute((alias("IA__gtk_menu_item_new"), visibility("default")));

#undef gtk_menu_item_new_with_label
extern __typeof (gtk_menu_item_new_with_label) gtk_menu_item_new_with_label __attribute((alias("IA__gtk_menu_item_new_with_label"), visibility("default")));

#undef gtk_menu_item_new_with_mnemonic
extern __typeof (gtk_menu_item_new_with_mnemonic) gtk_menu_item_new_with_mnemonic __attribute((alias("IA__gtk_menu_item_new_with_mnemonic"), visibility("default")));

#undef gtk_menu_item_select
extern __typeof (gtk_menu_item_select) gtk_menu_item_select __attribute((alias("IA__gtk_menu_item_select"), visibility("default")));

#undef gtk_menu_item_set_accel_path
extern __typeof (gtk_menu_item_set_accel_path) gtk_menu_item_set_accel_path __attribute((alias("IA__gtk_menu_item_set_accel_path"), visibility("default")));

#undef gtk_menu_item_set_label
extern __typeof (gtk_menu_item_set_label) gtk_menu_item_set_label __attribute((alias("IA__gtk_menu_item_set_label"), visibility("default")));

#undef gtk_menu_item_set_right_justified
extern __typeof (gtk_menu_item_set_right_justified) gtk_menu_item_set_right_justified __attribute((alias("IA__gtk_menu_item_set_right_justified"), visibility("default")));

#undef gtk_menu_item_set_submenu
extern __typeof (gtk_menu_item_set_submenu) gtk_menu_item_set_submenu __attribute((alias("IA__gtk_menu_item_set_submenu"), visibility("default")));

#undef gtk_menu_item_set_use_underline
extern __typeof (gtk_menu_item_set_use_underline) gtk_menu_item_set_use_underline __attribute((alias("IA__gtk_menu_item_set_use_underline"), visibility("default")));

#undef gtk_menu_item_toggle_size_allocate
extern __typeof (gtk_menu_item_toggle_size_allocate) gtk_menu_item_toggle_size_allocate __attribute((alias("IA__gtk_menu_item_toggle_size_allocate"), visibility("default")));

#undef gtk_menu_item_toggle_size_request
extern __typeof (gtk_menu_item_toggle_size_request) gtk_menu_item_toggle_size_request __attribute((alias("IA__gtk_menu_item_toggle_size_request"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_menu_item_remove_submenu
extern __typeof (gtk_menu_item_remove_submenu) gtk_menu_item_remove_submenu __attribute((alias("IA__gtk_menu_item_remove_submenu"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_MENU_SHELL_H__)
#if IN_FILE(__GTK_MENU_SHELL_C__)
#undef gtk_menu_shell_activate_item
extern __typeof (gtk_menu_shell_activate_item) gtk_menu_shell_activate_item __attribute((alias("IA__gtk_menu_shell_activate_item"), visibility("default")));

#undef gtk_menu_shell_append
extern __typeof (gtk_menu_shell_append) gtk_menu_shell_append __attribute((alias("IA__gtk_menu_shell_append"), visibility("default")));

#undef gtk_menu_shell_cancel
extern __typeof (gtk_menu_shell_cancel) gtk_menu_shell_cancel __attribute((alias("IA__gtk_menu_shell_cancel"), visibility("default")));

#undef gtk_menu_shell_deactivate
extern __typeof (gtk_menu_shell_deactivate) gtk_menu_shell_deactivate __attribute((alias("IA__gtk_menu_shell_deactivate"), visibility("default")));

#undef gtk_menu_shell_deselect
extern __typeof (gtk_menu_shell_deselect) gtk_menu_shell_deselect __attribute((alias("IA__gtk_menu_shell_deselect"), visibility("default")));

#undef gtk_menu_shell_get_type
extern __typeof (gtk_menu_shell_get_type) gtk_menu_shell_get_type __attribute((alias("IA__gtk_menu_shell_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_menu_shell_insert
extern __typeof (gtk_menu_shell_insert) gtk_menu_shell_insert __attribute((alias("IA__gtk_menu_shell_insert"), visibility("default")));

#undef gtk_menu_shell_prepend
extern __typeof (gtk_menu_shell_prepend) gtk_menu_shell_prepend __attribute((alias("IA__gtk_menu_shell_prepend"), visibility("default")));

#undef gtk_menu_shell_select_first
extern __typeof (gtk_menu_shell_select_first) gtk_menu_shell_select_first __attribute((alias("IA__gtk_menu_shell_select_first"), visibility("default")));

#undef gtk_menu_shell_select_item
extern __typeof (gtk_menu_shell_select_item) gtk_menu_shell_select_item __attribute((alias("IA__gtk_menu_shell_select_item"), visibility("default")));

#undef gtk_menu_shell_set_take_focus
extern __typeof (gtk_menu_shell_set_take_focus) gtk_menu_shell_set_take_focus __attribute((alias("IA__gtk_menu_shell_set_take_focus"), visibility("default")));

#undef gtk_menu_shell_get_take_focus
extern __typeof (gtk_menu_shell_get_take_focus) gtk_menu_shell_get_take_focus __attribute((alias("IA__gtk_menu_shell_get_take_focus"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_MENU_TOOL_BUTTON_H__)
#if IN_FILE(__GTK_MENU_TOOL_BUTTON_C__)
#undef gtk_menu_tool_button_get_menu
extern __typeof (gtk_menu_tool_button_get_menu) gtk_menu_tool_button_get_menu __attribute((alias("IA__gtk_menu_tool_button_get_menu"), visibility("default")));

#undef gtk_menu_tool_button_get_type
extern __typeof (gtk_menu_tool_button_get_type) gtk_menu_tool_button_get_type __attribute((alias("IA__gtk_menu_tool_button_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_menu_tool_button_new
extern __typeof (gtk_menu_tool_button_new) gtk_menu_tool_button_new __attribute((alias("IA__gtk_menu_tool_button_new"), visibility("default")));

#undef gtk_menu_tool_button_new_from_stock
extern __typeof (gtk_menu_tool_button_new_from_stock) gtk_menu_tool_button_new_from_stock __attribute((alias("IA__gtk_menu_tool_button_new_from_stock"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_menu_tool_button_set_arrow_tooltip
extern __typeof (gtk_menu_tool_button_set_arrow_tooltip) gtk_menu_tool_button_set_arrow_tooltip __attribute((alias("IA__gtk_menu_tool_button_set_arrow_tooltip"), visibility("default")));

#endif
#undef gtk_menu_tool_button_set_arrow_tooltip_markup
extern __typeof (gtk_menu_tool_button_set_arrow_tooltip_markup) gtk_menu_tool_button_set_arrow_tooltip_markup __attribute((alias("IA__gtk_menu_tool_button_set_arrow_tooltip_markup"), visibility("default")));

#undef gtk_menu_tool_button_set_arrow_tooltip_text
extern __typeof (gtk_menu_tool_button_set_arrow_tooltip_text) gtk_menu_tool_button_set_arrow_tooltip_text __attribute((alias("IA__gtk_menu_tool_button_set_arrow_tooltip_text"), visibility("default")));

#undef gtk_menu_tool_button_set_menu
extern __typeof (gtk_menu_tool_button_set_menu) gtk_menu_tool_button_set_menu __attribute((alias("IA__gtk_menu_tool_button_set_menu"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_MESSAGE_DIALOG_H__)
#if IN_FILE(__GTK_MESSAGE_DIALOG_C__)
#undef gtk_message_dialog_format_secondary_markup
extern __typeof (gtk_message_dialog_format_secondary_markup) gtk_message_dialog_format_secondary_markup __attribute((alias("IA__gtk_message_dialog_format_secondary_markup"), visibility("default"))) G_GNUC_PRINTF(2,3);

#undef gtk_message_dialog_format_secondary_text
extern __typeof (gtk_message_dialog_format_secondary_text) gtk_message_dialog_format_secondary_text __attribute((alias("IA__gtk_message_dialog_format_secondary_text"), visibility("default"))) G_GNUC_PRINTF(2,3);

#undef gtk_message_dialog_get_message_area
extern __typeof (gtk_message_dialog_get_message_area) gtk_message_dialog_get_message_area __attribute((alias("IA__gtk_message_dialog_get_message_area"), visibility("default")));

#undef gtk_message_dialog_get_type
extern __typeof (gtk_message_dialog_get_type) gtk_message_dialog_get_type __attribute((alias("IA__gtk_message_dialog_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_message_dialog_new
extern __typeof (gtk_message_dialog_new) gtk_message_dialog_new __attribute((alias("IA__gtk_message_dialog_new"), visibility("default"))) G_GNUC_PRINTF(5,6);

#undef gtk_message_dialog_new_with_markup
extern __typeof (gtk_message_dialog_new_with_markup) gtk_message_dialog_new_with_markup __attribute((alias("IA__gtk_message_dialog_new_with_markup"), visibility("default"))) G_GNUC_PRINTF(5,6);

#undef gtk_message_dialog_set_markup
extern __typeof (gtk_message_dialog_set_markup) gtk_message_dialog_set_markup __attribute((alias("IA__gtk_message_dialog_set_markup"), visibility("default")));

#undef gtk_message_dialog_set_image
extern __typeof (gtk_message_dialog_set_image) gtk_message_dialog_set_image __attribute((alias("IA__gtk_message_dialog_set_image"), visibility("default")));

#undef gtk_message_dialog_get_image
extern __typeof (gtk_message_dialog_get_image) gtk_message_dialog_get_image __attribute((alias("IA__gtk_message_dialog_get_image"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_MISC_H__)
#if IN_FILE(__GTK_MISC_C__)
#undef gtk_misc_get_alignment
extern __typeof (gtk_misc_get_alignment) gtk_misc_get_alignment __attribute((alias("IA__gtk_misc_get_alignment"), visibility("default")));

#undef gtk_misc_get_padding
extern __typeof (gtk_misc_get_padding) gtk_misc_get_padding __attribute((alias("IA__gtk_misc_get_padding"), visibility("default")));

#undef gtk_misc_get_type
extern __typeof (gtk_misc_get_type) gtk_misc_get_type __attribute((alias("IA__gtk_misc_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_misc_set_alignment
extern __typeof (gtk_misc_set_alignment) gtk_misc_set_alignment __attribute((alias("IA__gtk_misc_set_alignment"), visibility("default")));

#undef gtk_misc_set_padding
extern __typeof (gtk_misc_set_padding) gtk_misc_set_padding __attribute((alias("IA__gtk_misc_set_padding"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_MOUNT_OPERATION_H__)
#if IN_FILE(__GTK_MOUNT_OPERATION_C__)
#undef gtk_mount_operation_get_type
extern __typeof (gtk_mount_operation_get_type) gtk_mount_operation_get_type __attribute((alias("IA__gtk_mount_operation_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_mount_operation_new
extern __typeof (gtk_mount_operation_new) gtk_mount_operation_new __attribute((alias("IA__gtk_mount_operation_new"), visibility("default")));

#undef gtk_mount_operation_is_showing
extern __typeof (gtk_mount_operation_is_showing) gtk_mount_operation_is_showing __attribute((alias("IA__gtk_mount_operation_is_showing"), visibility("default")));

#undef gtk_mount_operation_set_parent
extern __typeof (gtk_mount_operation_set_parent) gtk_mount_operation_set_parent __attribute((alias("IA__gtk_mount_operation_set_parent"), visibility("default")));

#undef gtk_mount_operation_get_parent
extern __typeof (gtk_mount_operation_get_parent) gtk_mount_operation_get_parent __attribute((alias("IA__gtk_mount_operation_get_parent"), visibility("default")));

#undef gtk_mount_operation_set_screen
extern __typeof (gtk_mount_operation_set_screen) gtk_mount_operation_set_screen __attribute((alias("IA__gtk_mount_operation_set_screen"), visibility("default")));

#undef gtk_mount_operation_get_screen
extern __typeof (gtk_mount_operation_get_screen) gtk_mount_operation_get_screen __attribute((alias("IA__gtk_mount_operation_get_screen"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_NOTEBOOK_H__)
#if IN_FILE(__GTK_NOTEBOOK_C__)
#undef gtk_notebook_append_page
extern __typeof (gtk_notebook_append_page) gtk_notebook_append_page __attribute((alias("IA__gtk_notebook_append_page"), visibility("default")));

#undef gtk_notebook_append_page_menu
extern __typeof (gtk_notebook_append_page_menu) gtk_notebook_append_page_menu __attribute((alias("IA__gtk_notebook_append_page_menu"), visibility("default")));

#undef gtk_notebook_get_current_page
extern __typeof (gtk_notebook_get_current_page) gtk_notebook_get_current_page __attribute((alias("IA__gtk_notebook_get_current_page"), visibility("default")));

#undef gtk_notebook_get_menu_label
extern __typeof (gtk_notebook_get_menu_label) gtk_notebook_get_menu_label __attribute((alias("IA__gtk_notebook_get_menu_label"), visibility("default")));

#undef gtk_notebook_get_menu_label_text
extern __typeof (gtk_notebook_get_menu_label_text) gtk_notebook_get_menu_label_text __attribute((alias("IA__gtk_notebook_get_menu_label_text"), visibility("default")));

#undef gtk_notebook_get_n_pages
extern __typeof (gtk_notebook_get_n_pages) gtk_notebook_get_n_pages __attribute((alias("IA__gtk_notebook_get_n_pages"), visibility("default")));

#undef gtk_notebook_get_nth_page
extern __typeof (gtk_notebook_get_nth_page) gtk_notebook_get_nth_page __attribute((alias("IA__gtk_notebook_get_nth_page"), visibility("default")));

#undef gtk_notebook_get_scrollable
extern __typeof (gtk_notebook_get_scrollable) gtk_notebook_get_scrollable __attribute((alias("IA__gtk_notebook_get_scrollable"), visibility("default")));

#undef gtk_notebook_get_show_border
extern __typeof (gtk_notebook_get_show_border) gtk_notebook_get_show_border __attribute((alias("IA__gtk_notebook_get_show_border"), visibility("default")));

#undef gtk_notebook_get_show_tabs
extern __typeof (gtk_notebook_get_show_tabs) gtk_notebook_get_show_tabs __attribute((alias("IA__gtk_notebook_get_show_tabs"), visibility("default")));

#undef gtk_notebook_get_tab_label
extern __typeof (gtk_notebook_get_tab_label) gtk_notebook_get_tab_label __attribute((alias("IA__gtk_notebook_get_tab_label"), visibility("default")));

#undef gtk_notebook_get_tab_label_text
extern __typeof (gtk_notebook_get_tab_label_text) gtk_notebook_get_tab_label_text __attribute((alias("IA__gtk_notebook_get_tab_label_text"), visibility("default")));

#undef gtk_notebook_get_tab_pos
extern __typeof (gtk_notebook_get_tab_pos) gtk_notebook_get_tab_pos __attribute((alias("IA__gtk_notebook_get_tab_pos"), visibility("default")));

#undef gtk_notebook_get_type
extern __typeof (gtk_notebook_get_type) gtk_notebook_get_type __attribute((alias("IA__gtk_notebook_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_notebook_insert_page
extern __typeof (gtk_notebook_insert_page) gtk_notebook_insert_page __attribute((alias("IA__gtk_notebook_insert_page"), visibility("default")));

#undef gtk_notebook_insert_page_menu
extern __typeof (gtk_notebook_insert_page_menu) gtk_notebook_insert_page_menu __attribute((alias("IA__gtk_notebook_insert_page_menu"), visibility("default")));

#undef gtk_notebook_new
extern __typeof (gtk_notebook_new) gtk_notebook_new __attribute((alias("IA__gtk_notebook_new"), visibility("default")));

#undef gtk_notebook_next_page
extern __typeof (gtk_notebook_next_page) gtk_notebook_next_page __attribute((alias("IA__gtk_notebook_next_page"), visibility("default")));

#undef gtk_notebook_page_num
extern __typeof (gtk_notebook_page_num) gtk_notebook_page_num __attribute((alias("IA__gtk_notebook_page_num"), visibility("default")));

#undef gtk_notebook_popup_disable
extern __typeof (gtk_notebook_popup_disable) gtk_notebook_popup_disable __attribute((alias("IA__gtk_notebook_popup_disable"), visibility("default")));

#undef gtk_notebook_popup_enable
extern __typeof (gtk_notebook_popup_enable) gtk_notebook_popup_enable __attribute((alias("IA__gtk_notebook_popup_enable"), visibility("default")));

#undef gtk_notebook_prepend_page
extern __typeof (gtk_notebook_prepend_page) gtk_notebook_prepend_page __attribute((alias("IA__gtk_notebook_prepend_page"), visibility("default")));

#undef gtk_notebook_prepend_page_menu
extern __typeof (gtk_notebook_prepend_page_menu) gtk_notebook_prepend_page_menu __attribute((alias("IA__gtk_notebook_prepend_page_menu"), visibility("default")));

#undef gtk_notebook_prev_page
extern __typeof (gtk_notebook_prev_page) gtk_notebook_prev_page __attribute((alias("IA__gtk_notebook_prev_page"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_notebook_query_tab_label_packing
extern __typeof (gtk_notebook_query_tab_label_packing) gtk_notebook_query_tab_label_packing __attribute((alias("IA__gtk_notebook_query_tab_label_packing"), visibility("default")));

#endif
#undef gtk_notebook_remove_page
extern __typeof (gtk_notebook_remove_page) gtk_notebook_remove_page __attribute((alias("IA__gtk_notebook_remove_page"), visibility("default")));

#undef gtk_notebook_reorder_child
extern __typeof (gtk_notebook_reorder_child) gtk_notebook_reorder_child __attribute((alias("IA__gtk_notebook_reorder_child"), visibility("default")));

#undef gtk_notebook_set_current_page
extern __typeof (gtk_notebook_set_current_page) gtk_notebook_set_current_page __attribute((alias("IA__gtk_notebook_set_current_page"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_notebook_set_homogeneous_tabs
extern __typeof (gtk_notebook_set_homogeneous_tabs) gtk_notebook_set_homogeneous_tabs __attribute((alias("IA__gtk_notebook_set_homogeneous_tabs"), visibility("default")));

#undef gtk_notebook_set_tab_border
extern __typeof (gtk_notebook_set_tab_border) gtk_notebook_set_tab_border __attribute((alias("IA__gtk_notebook_set_tab_border"), visibility("default")));

#undef gtk_notebook_set_tab_hborder
extern __typeof (gtk_notebook_set_tab_hborder) gtk_notebook_set_tab_hborder __attribute((alias("IA__gtk_notebook_set_tab_hborder"), visibility("default")));

#undef gtk_notebook_set_tab_vborder
extern __typeof (gtk_notebook_set_tab_vborder) gtk_notebook_set_tab_vborder __attribute((alias("IA__gtk_notebook_set_tab_vborder"), visibility("default")));

#endif
#undef gtk_notebook_set_menu_label
extern __typeof (gtk_notebook_set_menu_label) gtk_notebook_set_menu_label __attribute((alias("IA__gtk_notebook_set_menu_label"), visibility("default")));

#undef gtk_notebook_set_menu_label_text
extern __typeof (gtk_notebook_set_menu_label_text) gtk_notebook_set_menu_label_text __attribute((alias("IA__gtk_notebook_set_menu_label_text"), visibility("default")));

#undef gtk_notebook_set_scrollable
extern __typeof (gtk_notebook_set_scrollable) gtk_notebook_set_scrollable __attribute((alias("IA__gtk_notebook_set_scrollable"), visibility("default")));

#undef gtk_notebook_set_show_border
extern __typeof (gtk_notebook_set_show_border) gtk_notebook_set_show_border __attribute((alias("IA__gtk_notebook_set_show_border"), visibility("default")));

#undef gtk_notebook_set_show_tabs
extern __typeof (gtk_notebook_set_show_tabs) gtk_notebook_set_show_tabs __attribute((alias("IA__gtk_notebook_set_show_tabs"), visibility("default")));

#undef gtk_notebook_set_tab_label
extern __typeof (gtk_notebook_set_tab_label) gtk_notebook_set_tab_label __attribute((alias("IA__gtk_notebook_set_tab_label"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_notebook_set_tab_label_packing
extern __typeof (gtk_notebook_set_tab_label_packing) gtk_notebook_set_tab_label_packing __attribute((alias("IA__gtk_notebook_set_tab_label_packing"), visibility("default")));

#endif
#undef gtk_notebook_set_tab_label_text
extern __typeof (gtk_notebook_set_tab_label_text) gtk_notebook_set_tab_label_text __attribute((alias("IA__gtk_notebook_set_tab_label_text"), visibility("default")));

#undef gtk_notebook_set_tab_pos
extern __typeof (gtk_notebook_set_tab_pos) gtk_notebook_set_tab_pos __attribute((alias("IA__gtk_notebook_set_tab_pos"), visibility("default")));

#undef gtk_notebook_get_tab_hborder
extern __typeof (gtk_notebook_get_tab_hborder) gtk_notebook_get_tab_hborder __attribute((alias("IA__gtk_notebook_get_tab_hborder"), visibility("default")));

#undef gtk_notebook_get_tab_vborder
extern __typeof (gtk_notebook_get_tab_vborder) gtk_notebook_get_tab_vborder __attribute((alias("IA__gtk_notebook_get_tab_vborder"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_notebook_set_window_creation_hook
extern __typeof (gtk_notebook_set_window_creation_hook) gtk_notebook_set_window_creation_hook __attribute((alias("IA__gtk_notebook_set_window_creation_hook"), visibility("default")));

#undef gtk_notebook_get_group_id
extern __typeof (gtk_notebook_get_group_id) gtk_notebook_get_group_id __attribute((alias("IA__gtk_notebook_get_group_id"), visibility("default")));

#undef gtk_notebook_set_group_id
extern __typeof (gtk_notebook_set_group_id) gtk_notebook_set_group_id __attribute((alias("IA__gtk_notebook_set_group_id"), visibility("default")));

#undef gtk_notebook_get_group
extern __typeof (gtk_notebook_get_group) gtk_notebook_get_group __attribute((alias("IA__gtk_notebook_get_group"), visibility("default")));

#undef gtk_notebook_set_group
extern __typeof (gtk_notebook_set_group) gtk_notebook_set_group __attribute((alias("IA__gtk_notebook_set_group"), visibility("default")));

#endif
#undef gtk_notebook_get_group_name
extern __typeof (gtk_notebook_get_group_name) gtk_notebook_get_group_name __attribute((alias("IA__gtk_notebook_get_group_name"), visibility("default")));

#undef gtk_notebook_set_group_name
extern __typeof (gtk_notebook_set_group_name) gtk_notebook_set_group_name __attribute((alias("IA__gtk_notebook_set_group_name"), visibility("default")));

#undef gtk_notebook_get_tab_reorderable
extern __typeof (gtk_notebook_get_tab_reorderable) gtk_notebook_get_tab_reorderable __attribute((alias("IA__gtk_notebook_get_tab_reorderable"), visibility("default")));

#undef gtk_notebook_set_tab_reorderable
extern __typeof (gtk_notebook_set_tab_reorderable) gtk_notebook_set_tab_reorderable __attribute((alias("IA__gtk_notebook_set_tab_reorderable"), visibility("default")));

#undef gtk_notebook_get_tab_detachable
extern __typeof (gtk_notebook_get_tab_detachable) gtk_notebook_get_tab_detachable __attribute((alias("IA__gtk_notebook_get_tab_detachable"), visibility("default")));

#undef gtk_notebook_set_tab_detachable
extern __typeof (gtk_notebook_set_tab_detachable) gtk_notebook_set_tab_detachable __attribute((alias("IA__gtk_notebook_set_tab_detachable"), visibility("default")));

#undef gtk_notebook_get_action_widget
extern __typeof (gtk_notebook_get_action_widget) gtk_notebook_get_action_widget __attribute((alias("IA__gtk_notebook_get_action_widget"), visibility("default")));

#undef gtk_notebook_set_action_widget
extern __typeof (gtk_notebook_set_action_widget) gtk_notebook_set_action_widget __attribute((alias("IA__gtk_notebook_set_action_widget"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_OBJECT_H__)
#if IN_FILE(__GTK_OBJECT_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_object_add_arg_type
extern __typeof (gtk_object_add_arg_type) gtk_object_add_arg_type __attribute((alias("IA__gtk_object_add_arg_type"), visibility("default")));

#undef gtk_object_get
extern __typeof (gtk_object_get) gtk_object_get __attribute((alias("IA__gtk_object_get"), visibility("default"))) G_GNUC_NULL_TERMINATED;

#undef gtk_object_set
extern __typeof (gtk_object_set) gtk_object_set __attribute((alias("IA__gtk_object_set"), visibility("default"))) G_GNUC_NULL_TERMINATED;

#undef gtk_object_get_data
extern __typeof (gtk_object_get_data) gtk_object_get_data __attribute((alias("IA__gtk_object_get_data"), visibility("default")));

#undef gtk_object_get_data_by_id
extern __typeof (gtk_object_get_data_by_id) gtk_object_get_data_by_id __attribute((alias("IA__gtk_object_get_data_by_id"), visibility("default")));

#undef gtk_object_get_user_data
extern __typeof (gtk_object_get_user_data) gtk_object_get_user_data __attribute((alias("IA__gtk_object_get_user_data"), visibility("default")));

#undef gtk_object_new
extern __typeof (gtk_object_new) gtk_object_new __attribute((alias("IA__gtk_object_new"), visibility("default")));

#undef gtk_object_remove_data
extern __typeof (gtk_object_remove_data) gtk_object_remove_data __attribute((alias("IA__gtk_object_remove_data"), visibility("default")));

#undef gtk_object_remove_data_by_id
extern __typeof (gtk_object_remove_data_by_id) gtk_object_remove_data_by_id __attribute((alias("IA__gtk_object_remove_data_by_id"), visibility("default")));

#undef gtk_object_remove_no_notify
extern __typeof (gtk_object_remove_no_notify) gtk_object_remove_no_notify __attribute((alias("IA__gtk_object_remove_no_notify"), visibility("default")));

#undef gtk_object_remove_no_notify_by_id
extern __typeof (gtk_object_remove_no_notify_by_id) gtk_object_remove_no_notify_by_id __attribute((alias("IA__gtk_object_remove_no_notify_by_id"), visibility("default")));

#undef gtk_object_ref
extern __typeof (gtk_object_ref) gtk_object_ref __attribute((alias("IA__gtk_object_ref"), visibility("default")));

#undef gtk_object_unref
extern __typeof (gtk_object_unref) gtk_object_unref __attribute((alias("IA__gtk_object_unref"), visibility("default")));

#undef gtk_object_weakref
extern __typeof (gtk_object_weakref) gtk_object_weakref __attribute((alias("IA__gtk_object_weakref"), visibility("default")));

#undef gtk_object_weakunref
extern __typeof (gtk_object_weakunref) gtk_object_weakunref __attribute((alias("IA__gtk_object_weakunref"), visibility("default")));

#undef gtk_object_set_data
extern __typeof (gtk_object_set_data) gtk_object_set_data __attribute((alias("IA__gtk_object_set_data"), visibility("default")));

#undef gtk_object_set_data_by_id
extern __typeof (gtk_object_set_data_by_id) gtk_object_set_data_by_id __attribute((alias("IA__gtk_object_set_data_by_id"), visibility("default")));

#undef gtk_object_set_data_by_id_full
extern __typeof (gtk_object_set_data_by_id_full) gtk_object_set_data_by_id_full __attribute((alias("IA__gtk_object_set_data_by_id_full"), visibility("default")));

#undef gtk_object_set_data_full
extern __typeof (gtk_object_set_data_full) gtk_object_set_data_full __attribute((alias("IA__gtk_object_set_data_full"), visibility("default")));

#undef gtk_object_set_user_data
extern __typeof (gtk_object_set_user_data) gtk_object_set_user_data __attribute((alias("IA__gtk_object_set_user_data"), visibility("default")));

#undef gtk_object_sink
extern __typeof (gtk_object_sink) gtk_object_sink __attribute((alias("IA__gtk_object_sink"), visibility("default")));

#endif
#undef gtk_object_destroy
extern __typeof (gtk_object_destroy) gtk_object_destroy __attribute((alias("IA__gtk_object_destroy"), visibility("default")));

#undef gtk_object_get_type
extern __typeof (gtk_object_get_type) gtk_object_get_type __attribute((alias("IA__gtk_object_get_type"), visibility("default"))) G_GNUC_CONST;

#endif
#endif
#if IN_HEADER(__GTK_OFFSCREEN_WINDOW_H__)
#if IN_FILE(__GTK_OFFSCREEN_WINDOW_C__)
#undef gtk_offscreen_window_get_type
extern __typeof (gtk_offscreen_window_get_type) gtk_offscreen_window_get_type __attribute((alias("IA__gtk_offscreen_window_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_offscreen_window_new
extern __typeof (gtk_offscreen_window_new) gtk_offscreen_window_new __attribute((alias("IA__gtk_offscreen_window_new"), visibility("default")));

#undef gtk_offscreen_window_get_pixmap
extern __typeof (gtk_offscreen_window_get_pixmap) gtk_offscreen_window_get_pixmap __attribute((alias("IA__gtk_offscreen_window_get_pixmap"), visibility("default")));

#undef gtk_offscreen_window_get_pixbuf
extern __typeof (gtk_offscreen_window_get_pixbuf) gtk_offscreen_window_get_pixbuf __attribute((alias("IA__gtk_offscreen_window_get_pixbuf"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_ORIENTABLE_H__)
#if IN_FILE(__GTK_ORIENTABLE_C__)
#undef gtk_orientable_get_orientation
extern __typeof (gtk_orientable_get_orientation) gtk_orientable_get_orientation __attribute((alias("IA__gtk_orientable_get_orientation"), visibility("default")));

#undef gtk_orientable_get_type
extern __typeof (gtk_orientable_get_type) gtk_orientable_get_type __attribute((alias("IA__gtk_orientable_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_orientable_set_orientation
extern __typeof (gtk_orientable_set_orientation) gtk_orientable_set_orientation __attribute((alias("IA__gtk_orientable_set_orientation"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_OLD_EDITABLE_H__)
#if IN_FILE(__GTK_OLD_EDITABLE_C__)
#undef gtk_old_editable_changed
extern __typeof (gtk_old_editable_changed) gtk_old_editable_changed __attribute((alias("IA__gtk_old_editable_changed"), visibility("default")));

#undef gtk_old_editable_claim_selection
extern __typeof (gtk_old_editable_claim_selection) gtk_old_editable_claim_selection __attribute((alias("IA__gtk_old_editable_claim_selection"), visibility("default")));

#undef gtk_old_editable_get_type
extern __typeof (gtk_old_editable_get_type) gtk_old_editable_get_type __attribute((alias("IA__gtk_old_editable_get_type"), visibility("default"))) G_GNUC_CONST;

#endif
#endif
#if IN_HEADER(__GTK_OPTION_MENU_H__)
#if IN_FILE(__GTK_OPTION_MENU_C__)
#undef gtk_option_menu_get_history
extern __typeof (gtk_option_menu_get_history) gtk_option_menu_get_history __attribute((alias("IA__gtk_option_menu_get_history"), visibility("default")));

#undef gtk_option_menu_get_menu
extern __typeof (gtk_option_menu_get_menu) gtk_option_menu_get_menu __attribute((alias("IA__gtk_option_menu_get_menu"), visibility("default")));

#undef gtk_option_menu_get_type
extern __typeof (gtk_option_menu_get_type) gtk_option_menu_get_type __attribute((alias("IA__gtk_option_menu_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_option_menu_new
extern __typeof (gtk_option_menu_new) gtk_option_menu_new __attribute((alias("IA__gtk_option_menu_new"), visibility("default")));

#undef gtk_option_menu_remove_menu
extern __typeof (gtk_option_menu_remove_menu) gtk_option_menu_remove_menu __attribute((alias("IA__gtk_option_menu_remove_menu"), visibility("default")));

#undef gtk_option_menu_set_history
extern __typeof (gtk_option_menu_set_history) gtk_option_menu_set_history __attribute((alias("IA__gtk_option_menu_set_history"), visibility("default")));

#undef gtk_option_menu_set_menu
extern __typeof (gtk_option_menu_set_menu) gtk_option_menu_set_menu __attribute((alias("IA__gtk_option_menu_set_menu"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_PANED_H__)
#if IN_FILE(__GTK_PANED_C__)
#undef gtk_paned_add1
extern __typeof (gtk_paned_add1) gtk_paned_add1 __attribute((alias("IA__gtk_paned_add1"), visibility("default")));

#undef gtk_paned_add2
extern __typeof (gtk_paned_add2) gtk_paned_add2 __attribute((alias("IA__gtk_paned_add2"), visibility("default")));

#undef gtk_paned_get_child1
extern __typeof (gtk_paned_get_child1) gtk_paned_get_child1 __attribute((alias("IA__gtk_paned_get_child1"), visibility("default")));

#undef gtk_paned_get_child2
extern __typeof (gtk_paned_get_child2) gtk_paned_get_child2 __attribute((alias("IA__gtk_paned_get_child2"), visibility("default")));

#undef gtk_paned_get_handle_window
extern __typeof (gtk_paned_get_handle_window) gtk_paned_get_handle_window __attribute((alias("IA__gtk_paned_get_handle_window"), visibility("default")));

#undef gtk_paned_get_position
extern __typeof (gtk_paned_get_position) gtk_paned_get_position __attribute((alias("IA__gtk_paned_get_position"), visibility("default")));

#undef gtk_paned_get_type
extern __typeof (gtk_paned_get_type) gtk_paned_get_type __attribute((alias("IA__gtk_paned_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_paned_pack1
extern __typeof (gtk_paned_pack1) gtk_paned_pack1 __attribute((alias("IA__gtk_paned_pack1"), visibility("default")));

#undef gtk_paned_pack2
extern __typeof (gtk_paned_pack2) gtk_paned_pack2 __attribute((alias("IA__gtk_paned_pack2"), visibility("default")));

#undef gtk_paned_set_position
extern __typeof (gtk_paned_set_position) gtk_paned_set_position __attribute((alias("IA__gtk_paned_set_position"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_paned_compute_position
extern __typeof (gtk_paned_compute_position) gtk_paned_compute_position __attribute((alias("IA__gtk_paned_compute_position"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_PATH_BAR_H__)
#if IN_FILE(__GTK_PATH_BAR_C__)
#undef gtk_path_bar_get_type
extern __typeof (gtk_path_bar_get_type) gtk_path_bar_get_type __attribute((alias("IA__gtk_path_bar_get_type"), visibility("default"))) G_GNUC_CONST;

#endif
#endif
#if IN_HEADER(__GTK_PIXMAP_H__)
#if IN_FILE(__GTK_PIXMAP_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_pixmap_get
extern __typeof (gtk_pixmap_get) gtk_pixmap_get __attribute((alias("IA__gtk_pixmap_get"), visibility("default")));

#undef gtk_pixmap_get_type
extern __typeof (gtk_pixmap_get_type) gtk_pixmap_get_type __attribute((alias("IA__gtk_pixmap_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_pixmap_new
extern __typeof (gtk_pixmap_new) gtk_pixmap_new __attribute((alias("IA__gtk_pixmap_new"), visibility("default")));

#undef gtk_pixmap_set
extern __typeof (gtk_pixmap_set) gtk_pixmap_set __attribute((alias("IA__gtk_pixmap_set"), visibility("default")));

#undef gtk_pixmap_set_build_insensitive
extern __typeof (gtk_pixmap_set_build_insensitive) gtk_pixmap_set_build_insensitive __attribute((alias("IA__gtk_pixmap_set_build_insensitive"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_PLUG_H__)
#if IN_FILE(__GTK_PLUG_C__)
#undef gtk_plug_construct
extern __typeof (gtk_plug_construct) gtk_plug_construct __attribute((alias("IA__gtk_plug_construct"), visibility("default")));

#undef gtk_plug_construct_for_display
extern __typeof (gtk_plug_construct_for_display) gtk_plug_construct_for_display __attribute((alias("IA__gtk_plug_construct_for_display"), visibility("default")));

#undef gtk_plug_get_id
extern __typeof (gtk_plug_get_id) gtk_plug_get_id __attribute((alias("IA__gtk_plug_get_id"), visibility("default")));

#undef gtk_plug_get_embedded
extern __typeof (gtk_plug_get_embedded) gtk_plug_get_embedded __attribute((alias("IA__gtk_plug_get_embedded"), visibility("default")));

#undef gtk_plug_get_socket_window
extern __typeof (gtk_plug_get_socket_window) gtk_plug_get_socket_window __attribute((alias("IA__gtk_plug_get_socket_window"), visibility("default")));

#undef gtk_plug_get_type
extern __typeof (gtk_plug_get_type) gtk_plug_get_type __attribute((alias("IA__gtk_plug_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_plug_new
extern __typeof (gtk_plug_new) gtk_plug_new __attribute((alias("IA__gtk_plug_new"), visibility("default")));

#undef gtk_plug_new_for_display
extern __typeof (gtk_plug_new_for_display) gtk_plug_new_for_display __attribute((alias("IA__gtk_plug_new_for_display"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_PREVIEW_H__)
#if IN_FILE(__GTK_PREVIEW_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_preview_draw_row
extern __typeof (gtk_preview_draw_row) gtk_preview_draw_row __attribute((alias("IA__gtk_preview_draw_row"), visibility("default")));

#undef gtk_preview_get_cmap
extern __typeof (gtk_preview_get_cmap) gtk_preview_get_cmap __attribute((alias("IA__gtk_preview_get_cmap"), visibility("default")));

#undef gtk_preview_get_info
extern __typeof (gtk_preview_get_info) gtk_preview_get_info __attribute((alias("IA__gtk_preview_get_info"), visibility("default")));

#undef gtk_preview_get_type
extern __typeof (gtk_preview_get_type) gtk_preview_get_type __attribute((alias("IA__gtk_preview_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_preview_get_visual
extern __typeof (gtk_preview_get_visual) gtk_preview_get_visual __attribute((alias("IA__gtk_preview_get_visual"), visibility("default")));

#undef gtk_preview_new
extern __typeof (gtk_preview_new) gtk_preview_new __attribute((alias("IA__gtk_preview_new"), visibility("default")));

#undef gtk_preview_put
extern __typeof (gtk_preview_put) gtk_preview_put __attribute((alias("IA__gtk_preview_put"), visibility("default")));

#undef gtk_preview_reset
extern __typeof (gtk_preview_reset) gtk_preview_reset __attribute((alias("IA__gtk_preview_reset"), visibility("default")));

#undef gtk_preview_set_color_cube
extern __typeof (gtk_preview_set_color_cube) gtk_preview_set_color_cube __attribute((alias("IA__gtk_preview_set_color_cube"), visibility("default")));

#undef gtk_preview_set_dither
extern __typeof (gtk_preview_set_dither) gtk_preview_set_dither __attribute((alias("IA__gtk_preview_set_dither"), visibility("default")));

#undef gtk_preview_set_expand
extern __typeof (gtk_preview_set_expand) gtk_preview_set_expand __attribute((alias("IA__gtk_preview_set_expand"), visibility("default")));

#undef gtk_preview_set_gamma
extern __typeof (gtk_preview_set_gamma) gtk_preview_set_gamma __attribute((alias("IA__gtk_preview_set_gamma"), visibility("default")));

#undef gtk_preview_set_install_cmap
extern __typeof (gtk_preview_set_install_cmap) gtk_preview_set_install_cmap __attribute((alias("IA__gtk_preview_set_install_cmap"), visibility("default")));

#undef gtk_preview_set_reserved
extern __typeof (gtk_preview_set_reserved) gtk_preview_set_reserved __attribute((alias("IA__gtk_preview_set_reserved"), visibility("default")));

#undef gtk_preview_size
extern __typeof (gtk_preview_size) gtk_preview_size __attribute((alias("IA__gtk_preview_size"), visibility("default")));

#undef gtk_preview_uninit
extern __typeof (gtk_preview_uninit) gtk_preview_uninit __attribute((alias("IA__gtk_preview_uninit"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_PAGE_SETUP_H__)
#if IN_FILE(__GTK_PAGE_SETUP_C__)
#undef gtk_page_setup_get_type
extern __typeof (gtk_page_setup_get_type) gtk_page_setup_get_type __attribute((alias("IA__gtk_page_setup_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_page_setup_new
extern __typeof (gtk_page_setup_new) gtk_page_setup_new __attribute((alias("IA__gtk_page_setup_new"), visibility("default")));

#undef gtk_page_setup_new_from_file
extern __typeof (gtk_page_setup_new_from_file) gtk_page_setup_new_from_file __attribute((alias("IA__gtk_page_setup_new_from_file"), visibility("default")));

#undef gtk_page_setup_new_from_key_file
extern __typeof (gtk_page_setup_new_from_key_file) gtk_page_setup_new_from_key_file __attribute((alias("IA__gtk_page_setup_new_from_key_file"), visibility("default")));

#undef gtk_page_setup_load_file
extern __typeof (gtk_page_setup_load_file) gtk_page_setup_load_file __attribute((alias("IA__gtk_page_setup_load_file"), visibility("default")));

#undef gtk_page_setup_load_key_file
extern __typeof (gtk_page_setup_load_key_file) gtk_page_setup_load_key_file __attribute((alias("IA__gtk_page_setup_load_key_file"), visibility("default")));

#undef gtk_page_setup_to_file
extern __typeof (gtk_page_setup_to_file) gtk_page_setup_to_file __attribute((alias("IA__gtk_page_setup_to_file"), visibility("default")));

#undef gtk_page_setup_to_key_file
extern __typeof (gtk_page_setup_to_key_file) gtk_page_setup_to_key_file __attribute((alias("IA__gtk_page_setup_to_key_file"), visibility("default")));

#undef gtk_page_setup_copy
extern __typeof (gtk_page_setup_copy) gtk_page_setup_copy __attribute((alias("IA__gtk_page_setup_copy"), visibility("default")));

#undef gtk_page_setup_get_orientation
extern __typeof (gtk_page_setup_get_orientation) gtk_page_setup_get_orientation __attribute((alias("IA__gtk_page_setup_get_orientation"), visibility("default")));

#undef gtk_page_setup_set_orientation
extern __typeof (gtk_page_setup_set_orientation) gtk_page_setup_set_orientation __attribute((alias("IA__gtk_page_setup_set_orientation"), visibility("default")));

#undef gtk_page_setup_get_paper_size
extern __typeof (gtk_page_setup_get_paper_size) gtk_page_setup_get_paper_size __attribute((alias("IA__gtk_page_setup_get_paper_size"), visibility("default")));

#undef gtk_page_setup_set_paper_size
extern __typeof (gtk_page_setup_set_paper_size) gtk_page_setup_set_paper_size __attribute((alias("IA__gtk_page_setup_set_paper_size"), visibility("default")));

#undef gtk_page_setup_get_top_margin
extern __typeof (gtk_page_setup_get_top_margin) gtk_page_setup_get_top_margin __attribute((alias("IA__gtk_page_setup_get_top_margin"), visibility("default")));

#undef gtk_page_setup_set_top_margin
extern __typeof (gtk_page_setup_set_top_margin) gtk_page_setup_set_top_margin __attribute((alias("IA__gtk_page_setup_set_top_margin"), visibility("default")));

#undef gtk_page_setup_get_bottom_margin
extern __typeof (gtk_page_setup_get_bottom_margin) gtk_page_setup_get_bottom_margin __attribute((alias("IA__gtk_page_setup_get_bottom_margin"), visibility("default")));

#undef gtk_page_setup_set_bottom_margin
extern __typeof (gtk_page_setup_set_bottom_margin) gtk_page_setup_set_bottom_margin __attribute((alias("IA__gtk_page_setup_set_bottom_margin"), visibility("default")));

#undef gtk_page_setup_get_left_margin
extern __typeof (gtk_page_setup_get_left_margin) gtk_page_setup_get_left_margin __attribute((alias("IA__gtk_page_setup_get_left_margin"), visibility("default")));

#undef gtk_page_setup_set_left_margin
extern __typeof (gtk_page_setup_set_left_margin) gtk_page_setup_set_left_margin __attribute((alias("IA__gtk_page_setup_set_left_margin"), visibility("default")));

#undef gtk_page_setup_get_right_margin
extern __typeof (gtk_page_setup_get_right_margin) gtk_page_setup_get_right_margin __attribute((alias("IA__gtk_page_setup_get_right_margin"), visibility("default")));

#undef gtk_page_setup_set_right_margin
extern __typeof (gtk_page_setup_set_right_margin) gtk_page_setup_set_right_margin __attribute((alias("IA__gtk_page_setup_set_right_margin"), visibility("default")));

#undef gtk_page_setup_set_paper_size_and_default_margins
extern __typeof (gtk_page_setup_set_paper_size_and_default_margins) gtk_page_setup_set_paper_size_and_default_margins __attribute((alias("IA__gtk_page_setup_set_paper_size_and_default_margins"), visibility("default")));

#undef gtk_page_setup_get_paper_width
extern __typeof (gtk_page_setup_get_paper_width) gtk_page_setup_get_paper_width __attribute((alias("IA__gtk_page_setup_get_paper_width"), visibility("default")));

#undef gtk_page_setup_get_paper_height
extern __typeof (gtk_page_setup_get_paper_height) gtk_page_setup_get_paper_height __attribute((alias("IA__gtk_page_setup_get_paper_height"), visibility("default")));

#undef gtk_page_setup_get_page_width
extern __typeof (gtk_page_setup_get_page_width) gtk_page_setup_get_page_width __attribute((alias("IA__gtk_page_setup_get_page_width"), visibility("default")));

#undef gtk_page_setup_get_page_height
extern __typeof (gtk_page_setup_get_page_height) gtk_page_setup_get_page_height __attribute((alias("IA__gtk_page_setup_get_page_height"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_PAGE_SETUP_UNIX_DIALOG_H__)
#if IN_FILE(__GTK_PAGE_SETUP_UNIX_DIALOG_C__)
#ifdef G_OS_UNIX
#undef gtk_page_setup_unix_dialog_get_type
extern __typeof (gtk_page_setup_unix_dialog_get_type) gtk_page_setup_unix_dialog_get_type __attribute((alias("IA__gtk_page_setup_unix_dialog_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_page_setup_unix_dialog_new
extern __typeof (gtk_page_setup_unix_dialog_new) gtk_page_setup_unix_dialog_new __attribute((alias("IA__gtk_page_setup_unix_dialog_new"), visibility("default")));

#undef gtk_page_setup_unix_dialog_set_page_setup
extern __typeof (gtk_page_setup_unix_dialog_set_page_setup) gtk_page_setup_unix_dialog_set_page_setup __attribute((alias("IA__gtk_page_setup_unix_dialog_set_page_setup"), visibility("default")));

#undef gtk_page_setup_unix_dialog_get_page_setup
extern __typeof (gtk_page_setup_unix_dialog_get_page_setup) gtk_page_setup_unix_dialog_get_page_setup __attribute((alias("IA__gtk_page_setup_unix_dialog_get_page_setup"), visibility("default")));

#undef gtk_page_setup_unix_dialog_set_print_settings
extern __typeof (gtk_page_setup_unix_dialog_set_print_settings) gtk_page_setup_unix_dialog_set_print_settings __attribute((alias("IA__gtk_page_setup_unix_dialog_set_print_settings"), visibility("default")));

#undef gtk_page_setup_unix_dialog_get_print_settings
extern __typeof (gtk_page_setup_unix_dialog_get_print_settings) gtk_page_setup_unix_dialog_get_print_settings __attribute((alias("IA__gtk_page_setup_unix_dialog_get_print_settings"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_PAPER_SIZE_H__)
#if IN_FILE(__GTK_PAPER_SIZE_C__)
#undef gtk_paper_size_get_type
extern __typeof (gtk_paper_size_get_type) gtk_paper_size_get_type __attribute((alias("IA__gtk_paper_size_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_paper_size_new
extern __typeof (gtk_paper_size_new) gtk_paper_size_new __attribute((alias("IA__gtk_paper_size_new"), visibility("default")));

#undef gtk_paper_size_new_from_ppd
extern __typeof (gtk_paper_size_new_from_ppd) gtk_paper_size_new_from_ppd __attribute((alias("IA__gtk_paper_size_new_from_ppd"), visibility("default")));

#undef gtk_paper_size_new_custom
extern __typeof (gtk_paper_size_new_custom) gtk_paper_size_new_custom __attribute((alias("IA__gtk_paper_size_new_custom"), visibility("default")));

#undef gtk_paper_size_new_from_key_file
extern __typeof (gtk_paper_size_new_from_key_file) gtk_paper_size_new_from_key_file __attribute((alias("IA__gtk_paper_size_new_from_key_file"), visibility("default")));

#undef gtk_paper_size_to_key_file
extern __typeof (gtk_paper_size_to_key_file) gtk_paper_size_to_key_file __attribute((alias("IA__gtk_paper_size_to_key_file"), visibility("default")));

#undef gtk_paper_size_copy
extern __typeof (gtk_paper_size_copy) gtk_paper_size_copy __attribute((alias("IA__gtk_paper_size_copy"), visibility("default")));

#undef gtk_paper_size_free
extern __typeof (gtk_paper_size_free) gtk_paper_size_free __attribute((alias("IA__gtk_paper_size_free"), visibility("default")));

#undef gtk_paper_size_is_equal
extern __typeof (gtk_paper_size_is_equal) gtk_paper_size_is_equal __attribute((alias("IA__gtk_paper_size_is_equal"), visibility("default")));

#undef gtk_paper_size_get_paper_sizes
extern __typeof (gtk_paper_size_get_paper_sizes) gtk_paper_size_get_paper_sizes __attribute((alias("IA__gtk_paper_size_get_paper_sizes"), visibility("default")));

#undef gtk_paper_size_get_name
extern __typeof (gtk_paper_size_get_name) gtk_paper_size_get_name __attribute((alias("IA__gtk_paper_size_get_name"), visibility("default")));

#undef gtk_paper_size_get_display_name
extern __typeof (gtk_paper_size_get_display_name) gtk_paper_size_get_display_name __attribute((alias("IA__gtk_paper_size_get_display_name"), visibility("default")));

#undef gtk_paper_size_get_ppd_name
extern __typeof (gtk_paper_size_get_ppd_name) gtk_paper_size_get_ppd_name __attribute((alias("IA__gtk_paper_size_get_ppd_name"), visibility("default")));

#undef gtk_paper_size_get_width
extern __typeof (gtk_paper_size_get_width) gtk_paper_size_get_width __attribute((alias("IA__gtk_paper_size_get_width"), visibility("default")));

#undef gtk_paper_size_get_height
extern __typeof (gtk_paper_size_get_height) gtk_paper_size_get_height __attribute((alias("IA__gtk_paper_size_get_height"), visibility("default")));

#undef gtk_paper_size_is_custom
extern __typeof (gtk_paper_size_is_custom) gtk_paper_size_is_custom __attribute((alias("IA__gtk_paper_size_is_custom"), visibility("default")));

#undef gtk_paper_size_set_size
extern __typeof (gtk_paper_size_set_size) gtk_paper_size_set_size __attribute((alias("IA__gtk_paper_size_set_size"), visibility("default")));

#undef gtk_paper_size_get_default_top_margin
extern __typeof (gtk_paper_size_get_default_top_margin) gtk_paper_size_get_default_top_margin __attribute((alias("IA__gtk_paper_size_get_default_top_margin"), visibility("default")));

#undef gtk_paper_size_get_default_bottom_margin
extern __typeof (gtk_paper_size_get_default_bottom_margin) gtk_paper_size_get_default_bottom_margin __attribute((alias("IA__gtk_paper_size_get_default_bottom_margin"), visibility("default")));

#undef gtk_paper_size_get_default_left_margin
extern __typeof (gtk_paper_size_get_default_left_margin) gtk_paper_size_get_default_left_margin __attribute((alias("IA__gtk_paper_size_get_default_left_margin"), visibility("default")));

#undef gtk_paper_size_get_default_right_margin
extern __typeof (gtk_paper_size_get_default_right_margin) gtk_paper_size_get_default_right_margin __attribute((alias("IA__gtk_paper_size_get_default_right_margin"), visibility("default")));

#undef gtk_paper_size_get_default
extern __typeof (gtk_paper_size_get_default) gtk_paper_size_get_default __attribute((alias("IA__gtk_paper_size_get_default"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_CUSTOM_PAPER_UNIX_DIALOG_H__)
#if IN_FILE(__GTK_CUSTOM_PAPER_UNIX_DIALOG_C__)
#ifdef G_OS_UNIX
#undef gtk_custom_paper_unix_dialog_get_type
extern __typeof (gtk_custom_paper_unix_dialog_get_type) gtk_custom_paper_unix_dialog_get_type __attribute((alias("IA__gtk_custom_paper_unix_dialog_get_type"), visibility("default"))) G_GNUC_CONST;

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINT_BACKEND_H__)
#if IN_FILE(__GTK_PRINT_BACKEND_C__)
#ifdef G_OS_UNIX
#undef gtk_print_backend_error_quark
extern __typeof (gtk_print_backend_error_quark) gtk_print_backend_error_quark __attribute((alias("IA__gtk_print_backend_error_quark"), visibility("default")));

#undef gtk_print_backend_get_type
extern __typeof (gtk_print_backend_get_type) gtk_print_backend_get_type __attribute((alias("IA__gtk_print_backend_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_print_backend_add_printer
extern __typeof (gtk_print_backend_add_printer) gtk_print_backend_add_printer __attribute((alias("IA__gtk_print_backend_add_printer"), visibility("default")));

#undef gtk_print_backend_remove_printer
extern __typeof (gtk_print_backend_remove_printer) gtk_print_backend_remove_printer __attribute((alias("IA__gtk_print_backend_remove_printer"), visibility("default")));

#undef gtk_print_backend_set_list_done
extern __typeof (gtk_print_backend_set_list_done) gtk_print_backend_set_list_done __attribute((alias("IA__gtk_print_backend_set_list_done"), visibility("default")));

#undef gtk_print_backend_destroy
extern __typeof (gtk_print_backend_destroy) gtk_print_backend_destroy __attribute((alias("IA__gtk_print_backend_destroy"), visibility("default")));

#undef gtk_print_backend_get_printer_list
extern __typeof (gtk_print_backend_get_printer_list) gtk_print_backend_get_printer_list __attribute((alias("IA__gtk_print_backend_get_printer_list"), visibility("default")));

#undef gtk_print_backend_printer_list_is_done
extern __typeof (gtk_print_backend_printer_list_is_done) gtk_print_backend_printer_list_is_done __attribute((alias("IA__gtk_print_backend_printer_list_is_done"), visibility("default")));

#undef gtk_print_backend_find_printer
extern __typeof (gtk_print_backend_find_printer) gtk_print_backend_find_printer __attribute((alias("IA__gtk_print_backend_find_printer"), visibility("default")));

#undef gtk_print_backend_print_stream
extern __typeof (gtk_print_backend_print_stream) gtk_print_backend_print_stream __attribute((alias("IA__gtk_print_backend_print_stream"), visibility("default")));

#undef gtk_print_backend_load_modules
extern __typeof (gtk_print_backend_load_modules) gtk_print_backend_load_modules __attribute((alias("IA__gtk_print_backend_load_modules"), visibility("default")));

#undef gtk_print_backend_set_password
extern __typeof (gtk_print_backend_set_password) gtk_print_backend_set_password __attribute((alias("IA__gtk_print_backend_set_password"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINT_CONTEXT_H__)
#if IN_FILE(__GTK_PRINT_CONTEXT_C__)
#undef gtk_print_context_get_type
extern __typeof (gtk_print_context_get_type) gtk_print_context_get_type __attribute((alias("IA__gtk_print_context_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_print_context_get_cairo_context
extern __typeof (gtk_print_context_get_cairo_context) gtk_print_context_get_cairo_context __attribute((alias("IA__gtk_print_context_get_cairo_context"), visibility("default")));

#undef gtk_print_context_get_page_setup
extern __typeof (gtk_print_context_get_page_setup) gtk_print_context_get_page_setup __attribute((alias("IA__gtk_print_context_get_page_setup"), visibility("default")));

#undef gtk_print_context_get_width
extern __typeof (gtk_print_context_get_width) gtk_print_context_get_width __attribute((alias("IA__gtk_print_context_get_width"), visibility("default")));

#undef gtk_print_context_get_height
extern __typeof (gtk_print_context_get_height) gtk_print_context_get_height __attribute((alias("IA__gtk_print_context_get_height"), visibility("default")));

#undef gtk_print_context_get_dpi_x
extern __typeof (gtk_print_context_get_dpi_x) gtk_print_context_get_dpi_x __attribute((alias("IA__gtk_print_context_get_dpi_x"), visibility("default")));

#undef gtk_print_context_get_dpi_y
extern __typeof (gtk_print_context_get_dpi_y) gtk_print_context_get_dpi_y __attribute((alias("IA__gtk_print_context_get_dpi_y"), visibility("default")));

#undef gtk_print_context_get_hard_margins
extern __typeof (gtk_print_context_get_hard_margins) gtk_print_context_get_hard_margins __attribute((alias("IA__gtk_print_context_get_hard_margins"), visibility("default")));

#undef gtk_print_context_get_pango_fontmap
extern __typeof (gtk_print_context_get_pango_fontmap) gtk_print_context_get_pango_fontmap __attribute((alias("IA__gtk_print_context_get_pango_fontmap"), visibility("default")));

#undef gtk_print_context_create_pango_context
extern __typeof (gtk_print_context_create_pango_context) gtk_print_context_create_pango_context __attribute((alias("IA__gtk_print_context_create_pango_context"), visibility("default")));

#undef gtk_print_context_create_pango_layout
extern __typeof (gtk_print_context_create_pango_layout) gtk_print_context_create_pango_layout __attribute((alias("IA__gtk_print_context_create_pango_layout"), visibility("default")));

#undef gtk_print_context_set_cairo_context
extern __typeof (gtk_print_context_set_cairo_context) gtk_print_context_set_cairo_context __attribute((alias("IA__gtk_print_context_set_cairo_context"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_PRINTER_H__)
#if IN_FILE(__GTK_PRINTER_C__)
#ifdef G_OS_UNIX
#undef gtk_printer_get_type
extern __typeof (gtk_printer_get_type) gtk_printer_get_type __attribute((alias("IA__gtk_printer_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_printer_get_backend
extern __typeof (gtk_printer_get_backend) gtk_printer_get_backend __attribute((alias("IA__gtk_printer_get_backend"), visibility("default")));

#undef gtk_printer_get_name
extern __typeof (gtk_printer_get_name) gtk_printer_get_name __attribute((alias("IA__gtk_printer_get_name"), visibility("default")));

#undef gtk_printer_get_description
extern __typeof (gtk_printer_get_description) gtk_printer_get_description __attribute((alias("IA__gtk_printer_get_description"), visibility("default")));

#undef gtk_printer_get_default_page_size
extern __typeof (gtk_printer_get_default_page_size) gtk_printer_get_default_page_size __attribute((alias("IA__gtk_printer_get_default_page_size"), visibility("default")));

#undef gtk_printer_get_state_message
extern __typeof (gtk_printer_get_state_message) gtk_printer_get_state_message __attribute((alias("IA__gtk_printer_get_state_message"), visibility("default")));

#undef gtk_printer_get_location
extern __typeof (gtk_printer_get_location) gtk_printer_get_location __attribute((alias("IA__gtk_printer_get_location"), visibility("default")));

#undef gtk_printer_get_icon_name
extern __typeof (gtk_printer_get_icon_name) gtk_printer_get_icon_name __attribute((alias("IA__gtk_printer_get_icon_name"), visibility("default")));

#undef gtk_printer_get_job_count
extern __typeof (gtk_printer_get_job_count) gtk_printer_get_job_count __attribute((alias("IA__gtk_printer_get_job_count"), visibility("default")));

#undef gtk_printer_is_virtual
extern __typeof (gtk_printer_is_virtual) gtk_printer_is_virtual __attribute((alias("IA__gtk_printer_is_virtual"), visibility("default")));

#undef gtk_printer_list_papers
extern __typeof (gtk_printer_list_papers) gtk_printer_list_papers __attribute((alias("IA__gtk_printer_list_papers"), visibility("default")));

#undef gtk_printer_accepts_pdf
extern __typeof (gtk_printer_accepts_pdf) gtk_printer_accepts_pdf __attribute((alias("IA__gtk_printer_accepts_pdf"), visibility("default")));

#undef gtk_printer_accepts_ps
extern __typeof (gtk_printer_accepts_ps) gtk_printer_accepts_ps __attribute((alias("IA__gtk_printer_accepts_ps"), visibility("default")));

#undef gtk_printer_compare
extern __typeof (gtk_printer_compare) gtk_printer_compare __attribute((alias("IA__gtk_printer_compare"), visibility("default")));

#undef gtk_printer_has_details
extern __typeof (gtk_printer_has_details) gtk_printer_has_details __attribute((alias("IA__gtk_printer_has_details"), visibility("default")));

#undef gtk_printer_request_details
extern __typeof (gtk_printer_request_details) gtk_printer_request_details __attribute((alias("IA__gtk_printer_request_details"), visibility("default")));

#undef gtk_printer_get_capabilities
extern __typeof (gtk_printer_get_capabilities) gtk_printer_get_capabilities __attribute((alias("IA__gtk_printer_get_capabilities"), visibility("default")));

#undef gtk_printer_get_hard_margins
extern __typeof (gtk_printer_get_hard_margins) gtk_printer_get_hard_margins __attribute((alias("IA__gtk_printer_get_hard_margins"), visibility("default")));

#undef gtk_enumerate_printers
extern __typeof (gtk_enumerate_printers) gtk_enumerate_printers __attribute((alias("IA__gtk_enumerate_printers"), visibility("default")));

#undef gtk_print_capabilities_get_type
extern __typeof (gtk_print_capabilities_get_type) gtk_print_capabilities_get_type __attribute((alias("IA__gtk_print_capabilities_get_type"), visibility("default"))) G_GNUC_CONST;

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINT_BACKEND_H__)
#if IN_FILE(__GTK_PRINTER_C__)
#ifdef G_OS_UNIX
#undef gtk_printer_new
extern __typeof (gtk_printer_new) gtk_printer_new __attribute((alias("IA__gtk_printer_new"), visibility("default")));

#undef gtk_printer_set_accepts_pdf
extern __typeof (gtk_printer_set_accepts_pdf) gtk_printer_set_accepts_pdf __attribute((alias("IA__gtk_printer_set_accepts_pdf"), visibility("default")));

#undef gtk_printer_set_accepts_ps
extern __typeof (gtk_printer_set_accepts_ps) gtk_printer_set_accepts_ps __attribute((alias("IA__gtk_printer_set_accepts_ps"), visibility("default")));

#undef gtk_printer_set_description
extern __typeof (gtk_printer_set_description) gtk_printer_set_description __attribute((alias("IA__gtk_printer_set_description"), visibility("default")));

#undef gtk_printer_set_has_details
extern __typeof (gtk_printer_set_has_details) gtk_printer_set_has_details __attribute((alias("IA__gtk_printer_set_has_details"), visibility("default")));

#undef gtk_printer_set_icon_name
extern __typeof (gtk_printer_set_icon_name) gtk_printer_set_icon_name __attribute((alias("IA__gtk_printer_set_icon_name"), visibility("default")));

#undef gtk_printer_set_is_active
extern __typeof (gtk_printer_set_is_active) gtk_printer_set_is_active __attribute((alias("IA__gtk_printer_set_is_active"), visibility("default")));

#undef gtk_printer_set_is_paused
extern __typeof (gtk_printer_set_is_paused) gtk_printer_set_is_paused __attribute((alias("IA__gtk_printer_set_is_paused"), visibility("default")));

#undef gtk_printer_set_is_accepting_jobs
extern __typeof (gtk_printer_set_is_accepting_jobs) gtk_printer_set_is_accepting_jobs __attribute((alias("IA__gtk_printer_set_is_accepting_jobs"), visibility("default")));

#undef gtk_printer_set_is_default
extern __typeof (gtk_printer_set_is_default) gtk_printer_set_is_default __attribute((alias("IA__gtk_printer_set_is_default"), visibility("default")));

#undef gtk_printer_set_is_new
extern __typeof (gtk_printer_set_is_new) gtk_printer_set_is_new __attribute((alias("IA__gtk_printer_set_is_new"), visibility("default")));

#undef gtk_printer_set_job_count
extern __typeof (gtk_printer_set_job_count) gtk_printer_set_job_count __attribute((alias("IA__gtk_printer_set_job_count"), visibility("default")));

#undef gtk_printer_set_location
extern __typeof (gtk_printer_set_location) gtk_printer_set_location __attribute((alias("IA__gtk_printer_set_location"), visibility("default")));

#undef gtk_printer_set_state_message
extern __typeof (gtk_printer_set_state_message) gtk_printer_set_state_message __attribute((alias("IA__gtk_printer_set_state_message"), visibility("default")));

#undef gtk_printer_is_active
extern __typeof (gtk_printer_is_active) gtk_printer_is_active __attribute((alias("IA__gtk_printer_is_active"), visibility("default")));

#undef gtk_printer_is_paused
extern __typeof (gtk_printer_is_paused) gtk_printer_is_paused __attribute((alias("IA__gtk_printer_is_paused"), visibility("default")));

#undef gtk_printer_is_accepting_jobs
extern __typeof (gtk_printer_is_accepting_jobs) gtk_printer_is_accepting_jobs __attribute((alias("IA__gtk_printer_is_accepting_jobs"), visibility("default")));

#undef gtk_printer_is_new
extern __typeof (gtk_printer_is_new) gtk_printer_is_new __attribute((alias("IA__gtk_printer_is_new"), visibility("default")));

#undef gtk_printer_is_default
extern __typeof (gtk_printer_is_default) gtk_printer_is_default __attribute((alias("IA__gtk_printer_is_default"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINTER_OPTION_H__)
#if IN_FILE(__GTK_PRINTER_OPTION_C__)
#ifdef G_OS_UNIX
#undef gtk_printer_option_get_type
extern __typeof (gtk_printer_option_get_type) gtk_printer_option_get_type __attribute((alias("IA__gtk_printer_option_get_type"), visibility("default")));

#undef gtk_printer_option_has_choice
extern __typeof (gtk_printer_option_has_choice) gtk_printer_option_has_choice __attribute((alias("IA__gtk_printer_option_has_choice"), visibility("default")));

#undef gtk_printer_option_new
extern __typeof (gtk_printer_option_new) gtk_printer_option_new __attribute((alias("IA__gtk_printer_option_new"), visibility("default")));

#undef gtk_printer_option_set
extern __typeof (gtk_printer_option_set) gtk_printer_option_set __attribute((alias("IA__gtk_printer_option_set"), visibility("default")));

#undef gtk_printer_option_set_has_conflict
extern __typeof (gtk_printer_option_set_has_conflict) gtk_printer_option_set_has_conflict __attribute((alias("IA__gtk_printer_option_set_has_conflict"), visibility("default")));

#undef gtk_printer_option_clear_has_conflict
extern __typeof (gtk_printer_option_clear_has_conflict) gtk_printer_option_clear_has_conflict __attribute((alias("IA__gtk_printer_option_clear_has_conflict"), visibility("default")));

#undef gtk_printer_option_set_boolean
extern __typeof (gtk_printer_option_set_boolean) gtk_printer_option_set_boolean __attribute((alias("IA__gtk_printer_option_set_boolean"), visibility("default")));

#undef gtk_printer_option_allocate_choices
extern __typeof (gtk_printer_option_allocate_choices) gtk_printer_option_allocate_choices __attribute((alias("IA__gtk_printer_option_allocate_choices"), visibility("default")));

#undef gtk_printer_option_choices_from_array
extern __typeof (gtk_printer_option_choices_from_array) gtk_printer_option_choices_from_array __attribute((alias("IA__gtk_printer_option_choices_from_array"), visibility("default")));

#undef gtk_printer_option_set_activates_default
extern __typeof (gtk_printer_option_set_activates_default) gtk_printer_option_set_activates_default __attribute((alias("IA__gtk_printer_option_set_activates_default"), visibility("default")));

#undef gtk_printer_option_get_activates_default
extern __typeof (gtk_printer_option_get_activates_default) gtk_printer_option_get_activates_default __attribute((alias("IA__gtk_printer_option_get_activates_default"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINTER_OPTION_SET_H__)
#if IN_FILE(__GTK_PRINTER_OPTION_SET_C__)
#ifdef G_OS_UNIX
#undef gtk_printer_option_set_get_type
extern __typeof (gtk_printer_option_set_get_type) gtk_printer_option_set_get_type __attribute((alias("IA__gtk_printer_option_set_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_printer_option_set_new
extern __typeof (gtk_printer_option_set_new) gtk_printer_option_set_new __attribute((alias("IA__gtk_printer_option_set_new"), visibility("default")));

#undef gtk_printer_option_set_add
extern __typeof (gtk_printer_option_set_add) gtk_printer_option_set_add __attribute((alias("IA__gtk_printer_option_set_add"), visibility("default")));

#undef gtk_printer_option_set_lookup
extern __typeof (gtk_printer_option_set_lookup) gtk_printer_option_set_lookup __attribute((alias("IA__gtk_printer_option_set_lookup"), visibility("default")));

#undef gtk_printer_option_set_remove
extern __typeof (gtk_printer_option_set_remove) gtk_printer_option_set_remove __attribute((alias("IA__gtk_printer_option_set_remove"), visibility("default")));

#undef gtk_printer_option_set_foreach
extern __typeof (gtk_printer_option_set_foreach) gtk_printer_option_set_foreach __attribute((alias("IA__gtk_printer_option_set_foreach"), visibility("default")));

#undef gtk_printer_option_set_clear_conflicts
extern __typeof (gtk_printer_option_set_clear_conflicts) gtk_printer_option_set_clear_conflicts __attribute((alias("IA__gtk_printer_option_set_clear_conflicts"), visibility("default")));

#undef gtk_printer_option_set_get_groups
extern __typeof (gtk_printer_option_set_get_groups) gtk_printer_option_set_get_groups __attribute((alias("IA__gtk_printer_option_set_get_groups"), visibility("default")));

#undef gtk_printer_option_set_foreach_in_group
extern __typeof (gtk_printer_option_set_foreach_in_group) gtk_printer_option_set_foreach_in_group __attribute((alias("IA__gtk_printer_option_set_foreach_in_group"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINTER_OPTION_WIDGET_H__)
#if IN_FILE(__GTK_PRINTER_OPTION_WIDGET_C__)
#ifdef G_OS_UNIX
#undef gtk_printer_option_widget_get_type
extern __typeof (gtk_printer_option_widget_get_type) gtk_printer_option_widget_get_type __attribute((alias("IA__gtk_printer_option_widget_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_printer_option_widget_new
extern __typeof (gtk_printer_option_widget_new) gtk_printer_option_widget_new __attribute((alias("IA__gtk_printer_option_widget_new"), visibility("default")));

#undef gtk_printer_option_widget_set_source
extern __typeof (gtk_printer_option_widget_set_source) gtk_printer_option_widget_set_source __attribute((alias("IA__gtk_printer_option_widget_set_source"), visibility("default")));

#undef gtk_printer_option_widget_has_external_label
extern __typeof (gtk_printer_option_widget_has_external_label) gtk_printer_option_widget_has_external_label __attribute((alias("IA__gtk_printer_option_widget_has_external_label"), visibility("default")));

#undef gtk_printer_option_widget_get_external_label
extern __typeof (gtk_printer_option_widget_get_external_label) gtk_printer_option_widget_get_external_label __attribute((alias("IA__gtk_printer_option_widget_get_external_label"), visibility("default")));

#undef gtk_printer_option_widget_get_value
extern __typeof (gtk_printer_option_widget_get_value) gtk_printer_option_widget_get_value __attribute((alias("IA__gtk_printer_option_widget_get_value"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINT_JOB_H__)
#if IN_FILE(__GTK_PRINT_JOB_C__)
#ifdef G_OS_UNIX
#undef gtk_print_job_get_type
extern __typeof (gtk_print_job_get_type) gtk_print_job_get_type __attribute((alias("IA__gtk_print_job_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_print_job_new
extern __typeof (gtk_print_job_new) gtk_print_job_new __attribute((alias("IA__gtk_print_job_new"), visibility("default")));

#undef gtk_print_job_get_settings
extern __typeof (gtk_print_job_get_settings) gtk_print_job_get_settings __attribute((alias("IA__gtk_print_job_get_settings"), visibility("default")));

#undef gtk_print_job_get_printer
extern __typeof (gtk_print_job_get_printer) gtk_print_job_get_printer __attribute((alias("IA__gtk_print_job_get_printer"), visibility("default")));

#undef gtk_print_job_get_title
extern __typeof (gtk_print_job_get_title) gtk_print_job_get_title __attribute((alias("IA__gtk_print_job_get_title"), visibility("default")));

#undef gtk_print_job_get_status
extern __typeof (gtk_print_job_get_status) gtk_print_job_get_status __attribute((alias("IA__gtk_print_job_get_status"), visibility("default")));

#undef gtk_print_job_set_source_file
extern __typeof (gtk_print_job_set_source_file) gtk_print_job_set_source_file __attribute((alias("IA__gtk_print_job_set_source_file"), visibility("default")));

#undef gtk_print_job_get_surface
extern __typeof (gtk_print_job_get_surface) gtk_print_job_get_surface __attribute((alias("IA__gtk_print_job_get_surface"), visibility("default")));

#undef gtk_print_job_send
extern __typeof (gtk_print_job_send) gtk_print_job_send __attribute((alias("IA__gtk_print_job_send"), visibility("default")));

#undef gtk_print_job_set_track_print_status
extern __typeof (gtk_print_job_set_track_print_status) gtk_print_job_set_track_print_status __attribute((alias("IA__gtk_print_job_set_track_print_status"), visibility("default")));

#undef gtk_print_job_get_track_print_status
extern __typeof (gtk_print_job_get_track_print_status) gtk_print_job_get_track_print_status __attribute((alias("IA__gtk_print_job_get_track_print_status"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINTER_PRIVATE_H__)
#if IN_FILE(__GTK_PRINT_JOB_C__)
#ifdef G_OS_UNIX
#undef gtk_print_job_set_status
extern __typeof (gtk_print_job_set_status) gtk_print_job_set_status __attribute((alias("IA__gtk_print_job_set_status"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINT_OPERATION_H__)
#if IN_FILE(__GTK_PRINT_OPERATION_C__)
#undef gtk_print_error_quark
extern __typeof (gtk_print_error_quark) gtk_print_error_quark __attribute((alias("IA__gtk_print_error_quark"), visibility("default")));

#undef gtk_print_operation_get_type
extern __typeof (gtk_print_operation_get_type) gtk_print_operation_get_type __attribute((alias("IA__gtk_print_operation_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_print_operation_new
extern __typeof (gtk_print_operation_new) gtk_print_operation_new __attribute((alias("IA__gtk_print_operation_new"), visibility("default")));

#undef gtk_print_operation_set_allow_async
extern __typeof (gtk_print_operation_set_allow_async) gtk_print_operation_set_allow_async __attribute((alias("IA__gtk_print_operation_set_allow_async"), visibility("default")));

#undef gtk_print_operation_set_default_page_setup
extern __typeof (gtk_print_operation_set_default_page_setup) gtk_print_operation_set_default_page_setup __attribute((alias("IA__gtk_print_operation_set_default_page_setup"), visibility("default")));

#undef gtk_print_operation_get_default_page_setup
extern __typeof (gtk_print_operation_get_default_page_setup) gtk_print_operation_get_default_page_setup __attribute((alias("IA__gtk_print_operation_get_default_page_setup"), visibility("default")));

#undef gtk_print_operation_set_print_settings
extern __typeof (gtk_print_operation_set_print_settings) gtk_print_operation_set_print_settings __attribute((alias("IA__gtk_print_operation_set_print_settings"), visibility("default")));

#undef gtk_print_operation_get_n_pages_to_print
extern __typeof (gtk_print_operation_get_n_pages_to_print) gtk_print_operation_get_n_pages_to_print __attribute((alias("IA__gtk_print_operation_get_n_pages_to_print"), visibility("default")));

#undef gtk_print_operation_get_print_settings
extern __typeof (gtk_print_operation_get_print_settings) gtk_print_operation_get_print_settings __attribute((alias("IA__gtk_print_operation_get_print_settings"), visibility("default")));

#undef gtk_print_operation_set_job_name
extern __typeof (gtk_print_operation_set_job_name) gtk_print_operation_set_job_name __attribute((alias("IA__gtk_print_operation_set_job_name"), visibility("default")));

#undef gtk_print_operation_set_n_pages
extern __typeof (gtk_print_operation_set_n_pages) gtk_print_operation_set_n_pages __attribute((alias("IA__gtk_print_operation_set_n_pages"), visibility("default")));

#undef gtk_print_operation_set_current_page
extern __typeof (gtk_print_operation_set_current_page) gtk_print_operation_set_current_page __attribute((alias("IA__gtk_print_operation_set_current_page"), visibility("default")));

#undef gtk_print_operation_set_use_full_page
extern __typeof (gtk_print_operation_set_use_full_page) gtk_print_operation_set_use_full_page __attribute((alias("IA__gtk_print_operation_set_use_full_page"), visibility("default")));

#undef gtk_print_operation_set_unit
extern __typeof (gtk_print_operation_set_unit) gtk_print_operation_set_unit __attribute((alias("IA__gtk_print_operation_set_unit"), visibility("default")));

#undef gtk_print_operation_set_export_filename
extern __typeof (gtk_print_operation_set_export_filename) gtk_print_operation_set_export_filename __attribute((alias("IA__gtk_print_operation_set_export_filename"), visibility("default")));

#undef gtk_print_operation_set_track_print_status
extern __typeof (gtk_print_operation_set_track_print_status) gtk_print_operation_set_track_print_status __attribute((alias("IA__gtk_print_operation_set_track_print_status"), visibility("default")));

#undef gtk_print_operation_set_show_progress
extern __typeof (gtk_print_operation_set_show_progress) gtk_print_operation_set_show_progress __attribute((alias("IA__gtk_print_operation_set_show_progress"), visibility("default")));

#undef gtk_print_operation_set_custom_tab_label
extern __typeof (gtk_print_operation_set_custom_tab_label) gtk_print_operation_set_custom_tab_label __attribute((alias("IA__gtk_print_operation_set_custom_tab_label"), visibility("default")));

#undef gtk_print_operation_get_error
extern __typeof (gtk_print_operation_get_error) gtk_print_operation_get_error __attribute((alias("IA__gtk_print_operation_get_error"), visibility("default")));

#undef gtk_print_operation_run
extern __typeof (gtk_print_operation_run) gtk_print_operation_run __attribute((alias("IA__gtk_print_operation_run"), visibility("default")));

#undef gtk_print_operation_get_status
extern __typeof (gtk_print_operation_get_status) gtk_print_operation_get_status __attribute((alias("IA__gtk_print_operation_get_status"), visibility("default")));

#undef gtk_print_operation_get_status_string
extern __typeof (gtk_print_operation_get_status_string) gtk_print_operation_get_status_string __attribute((alias("IA__gtk_print_operation_get_status_string"), visibility("default")));

#undef gtk_print_operation_is_finished
extern __typeof (gtk_print_operation_is_finished) gtk_print_operation_is_finished __attribute((alias("IA__gtk_print_operation_is_finished"), visibility("default")));

#undef gtk_print_operation_cancel
extern __typeof (gtk_print_operation_cancel) gtk_print_operation_cancel __attribute((alias("IA__gtk_print_operation_cancel"), visibility("default")));

#undef gtk_print_operation_draw_page_finish
extern __typeof (gtk_print_operation_draw_page_finish) gtk_print_operation_draw_page_finish __attribute((alias("IA__gtk_print_operation_draw_page_finish"), visibility("default")));

#undef gtk_print_operation_set_defer_drawing
extern __typeof (gtk_print_operation_set_defer_drawing) gtk_print_operation_set_defer_drawing __attribute((alias("IA__gtk_print_operation_set_defer_drawing"), visibility("default")));

#undef gtk_print_operation_set_support_selection
extern __typeof (gtk_print_operation_set_support_selection) gtk_print_operation_set_support_selection __attribute((alias("IA__gtk_print_operation_set_support_selection"), visibility("default")));

#undef gtk_print_operation_get_support_selection
extern __typeof (gtk_print_operation_get_support_selection) gtk_print_operation_get_support_selection __attribute((alias("IA__gtk_print_operation_get_support_selection"), visibility("default")));

#undef gtk_print_operation_set_has_selection
extern __typeof (gtk_print_operation_set_has_selection) gtk_print_operation_set_has_selection __attribute((alias("IA__gtk_print_operation_set_has_selection"), visibility("default")));

#undef gtk_print_operation_get_has_selection
extern __typeof (gtk_print_operation_get_has_selection) gtk_print_operation_get_has_selection __attribute((alias("IA__gtk_print_operation_get_has_selection"), visibility("default")));

#undef gtk_print_operation_set_embed_page_setup
extern __typeof (gtk_print_operation_set_embed_page_setup) gtk_print_operation_set_embed_page_setup __attribute((alias("IA__gtk_print_operation_set_embed_page_setup"), visibility("default")));

#undef gtk_print_operation_get_embed_page_setup
extern __typeof (gtk_print_operation_get_embed_page_setup) gtk_print_operation_get_embed_page_setup __attribute((alias("IA__gtk_print_operation_get_embed_page_setup"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_PRINT_OPERATION_PREVIEW_H__)
#if IN_FILE(__GTK_PRINT_OPERATION_PREVIEW_C__)
#undef gtk_print_operation_preview_end_preview
extern __typeof (gtk_print_operation_preview_end_preview) gtk_print_operation_preview_end_preview __attribute((alias("IA__gtk_print_operation_preview_end_preview"), visibility("default")));

#undef gtk_print_operation_preview_get_type
extern __typeof (gtk_print_operation_preview_get_type) gtk_print_operation_preview_get_type __attribute((alias("IA__gtk_print_operation_preview_get_type"), visibility("default")));

#undef gtk_print_operation_preview_is_selected
extern __typeof (gtk_print_operation_preview_is_selected) gtk_print_operation_preview_is_selected __attribute((alias("IA__gtk_print_operation_preview_is_selected"), visibility("default")));

#undef gtk_print_operation_preview_render_page
extern __typeof (gtk_print_operation_preview_render_page) gtk_print_operation_preview_render_page __attribute((alias("IA__gtk_print_operation_preview_render_page"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_PRINT_OPERATION_H__)
#if IN_FILE(__GTK_PRINT_OPERATION_UNIX_C__)
#ifdef G_OS_UNIX
#undef gtk_print_run_page_setup_dialog
extern __typeof (gtk_print_run_page_setup_dialog) gtk_print_run_page_setup_dialog __attribute((alias("IA__gtk_print_run_page_setup_dialog"), visibility("default")));

#undef gtk_print_run_page_setup_dialog_async
extern __typeof (gtk_print_run_page_setup_dialog_async) gtk_print_run_page_setup_dialog_async __attribute((alias("IA__gtk_print_run_page_setup_dialog_async"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINT_OPERATION_H__)
#if IN_FILE(__GTK_PRINT_OPERATION_WIN32_C__)
#ifdef G_OS_WIN32
#undef gtk_print_run_page_setup_dialog
extern __typeof (gtk_print_run_page_setup_dialog) gtk_print_run_page_setup_dialog __attribute((alias("IA__gtk_print_run_page_setup_dialog"), visibility("default")));

#undef gtk_print_run_page_setup_dialog_async
extern __typeof (gtk_print_run_page_setup_dialog_async) gtk_print_run_page_setup_dialog_async __attribute((alias("IA__gtk_print_run_page_setup_dialog_async"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINT_SETTINGS_H__)
#if IN_FILE(__GTK_PRINT_SETTINGS_C__)
#undef gtk_print_settings_get_type
extern __typeof (gtk_print_settings_get_type) gtk_print_settings_get_type __attribute((alias("IA__gtk_print_settings_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_print_settings_new
extern __typeof (gtk_print_settings_new) gtk_print_settings_new __attribute((alias("IA__gtk_print_settings_new"), visibility("default")));

#undef gtk_print_settings_new_from_file
extern __typeof (gtk_print_settings_new_from_file) gtk_print_settings_new_from_file __attribute((alias("IA__gtk_print_settings_new_from_file"), visibility("default")));

#undef gtk_print_settings_new_from_key_file
extern __typeof (gtk_print_settings_new_from_key_file) gtk_print_settings_new_from_key_file __attribute((alias("IA__gtk_print_settings_new_from_key_file"), visibility("default")));

#undef gtk_print_settings_load_file
extern __typeof (gtk_print_settings_load_file) gtk_print_settings_load_file __attribute((alias("IA__gtk_print_settings_load_file"), visibility("default")));

#undef gtk_print_settings_load_key_file
extern __typeof (gtk_print_settings_load_key_file) gtk_print_settings_load_key_file __attribute((alias("IA__gtk_print_settings_load_key_file"), visibility("default")));

#undef gtk_print_settings_to_file
extern __typeof (gtk_print_settings_to_file) gtk_print_settings_to_file __attribute((alias("IA__gtk_print_settings_to_file"), visibility("default")));

#undef gtk_print_settings_to_key_file
extern __typeof (gtk_print_settings_to_key_file) gtk_print_settings_to_key_file __attribute((alias("IA__gtk_print_settings_to_key_file"), visibility("default")));

#undef gtk_print_settings_copy
extern __typeof (gtk_print_settings_copy) gtk_print_settings_copy __attribute((alias("IA__gtk_print_settings_copy"), visibility("default")));

#undef gtk_print_settings_has_key
extern __typeof (gtk_print_settings_has_key) gtk_print_settings_has_key __attribute((alias("IA__gtk_print_settings_has_key"), visibility("default")));

#undef gtk_print_settings_get
extern __typeof (gtk_print_settings_get) gtk_print_settings_get __attribute((alias("IA__gtk_print_settings_get"), visibility("default")));

#undef gtk_print_settings_set
extern __typeof (gtk_print_settings_set) gtk_print_settings_set __attribute((alias("IA__gtk_print_settings_set"), visibility("default")));

#undef gtk_print_settings_unset
extern __typeof (gtk_print_settings_unset) gtk_print_settings_unset __attribute((alias("IA__gtk_print_settings_unset"), visibility("default")));

#undef gtk_print_settings_foreach
extern __typeof (gtk_print_settings_foreach) gtk_print_settings_foreach __attribute((alias("IA__gtk_print_settings_foreach"), visibility("default")));

#undef gtk_print_settings_get_bool
extern __typeof (gtk_print_settings_get_bool) gtk_print_settings_get_bool __attribute((alias("IA__gtk_print_settings_get_bool"), visibility("default")));

#undef gtk_print_settings_set_bool
extern __typeof (gtk_print_settings_set_bool) gtk_print_settings_set_bool __attribute((alias("IA__gtk_print_settings_set_bool"), visibility("default")));

#undef gtk_print_settings_get_double
extern __typeof (gtk_print_settings_get_double) gtk_print_settings_get_double __attribute((alias("IA__gtk_print_settings_get_double"), visibility("default")));

#undef gtk_print_settings_get_double_with_default
extern __typeof (gtk_print_settings_get_double_with_default) gtk_print_settings_get_double_with_default __attribute((alias("IA__gtk_print_settings_get_double_with_default"), visibility("default")));

#undef gtk_print_settings_set_double
extern __typeof (gtk_print_settings_set_double) gtk_print_settings_set_double __attribute((alias("IA__gtk_print_settings_set_double"), visibility("default")));

#undef gtk_print_settings_get_length
extern __typeof (gtk_print_settings_get_length) gtk_print_settings_get_length __attribute((alias("IA__gtk_print_settings_get_length"), visibility("default")));

#undef gtk_print_settings_set_length
extern __typeof (gtk_print_settings_set_length) gtk_print_settings_set_length __attribute((alias("IA__gtk_print_settings_set_length"), visibility("default")));

#undef gtk_print_settings_get_int
extern __typeof (gtk_print_settings_get_int) gtk_print_settings_get_int __attribute((alias("IA__gtk_print_settings_get_int"), visibility("default")));

#undef gtk_print_settings_get_int_with_default
extern __typeof (gtk_print_settings_get_int_with_default) gtk_print_settings_get_int_with_default __attribute((alias("IA__gtk_print_settings_get_int_with_default"), visibility("default")));

#undef gtk_print_settings_set_int
extern __typeof (gtk_print_settings_set_int) gtk_print_settings_set_int __attribute((alias("IA__gtk_print_settings_set_int"), visibility("default")));

#undef gtk_print_settings_get_printer
extern __typeof (gtk_print_settings_get_printer) gtk_print_settings_get_printer __attribute((alias("IA__gtk_print_settings_get_printer"), visibility("default")));

#undef gtk_print_settings_set_printer
extern __typeof (gtk_print_settings_set_printer) gtk_print_settings_set_printer __attribute((alias("IA__gtk_print_settings_set_printer"), visibility("default")));

#undef gtk_print_settings_get_orientation
extern __typeof (gtk_print_settings_get_orientation) gtk_print_settings_get_orientation __attribute((alias("IA__gtk_print_settings_get_orientation"), visibility("default")));

#undef gtk_print_settings_set_orientation
extern __typeof (gtk_print_settings_set_orientation) gtk_print_settings_set_orientation __attribute((alias("IA__gtk_print_settings_set_orientation"), visibility("default")));

#undef gtk_print_settings_get_paper_size
extern __typeof (gtk_print_settings_get_paper_size) gtk_print_settings_get_paper_size __attribute((alias("IA__gtk_print_settings_get_paper_size"), visibility("default")));

#undef gtk_print_settings_set_paper_size
extern __typeof (gtk_print_settings_set_paper_size) gtk_print_settings_set_paper_size __attribute((alias("IA__gtk_print_settings_set_paper_size"), visibility("default")));

#undef gtk_print_settings_get_paper_width
extern __typeof (gtk_print_settings_get_paper_width) gtk_print_settings_get_paper_width __attribute((alias("IA__gtk_print_settings_get_paper_width"), visibility("default")));

#undef gtk_print_settings_set_paper_width
extern __typeof (gtk_print_settings_set_paper_width) gtk_print_settings_set_paper_width __attribute((alias("IA__gtk_print_settings_set_paper_width"), visibility("default")));

#undef gtk_print_settings_get_paper_height
extern __typeof (gtk_print_settings_get_paper_height) gtk_print_settings_get_paper_height __attribute((alias("IA__gtk_print_settings_get_paper_height"), visibility("default")));

#undef gtk_print_settings_set_paper_height
extern __typeof (gtk_print_settings_set_paper_height) gtk_print_settings_set_paper_height __attribute((alias("IA__gtk_print_settings_set_paper_height"), visibility("default")));

#undef gtk_print_settings_get_use_color
extern __typeof (gtk_print_settings_get_use_color) gtk_print_settings_get_use_color __attribute((alias("IA__gtk_print_settings_get_use_color"), visibility("default")));

#undef gtk_print_settings_set_use_color
extern __typeof (gtk_print_settings_set_use_color) gtk_print_settings_set_use_color __attribute((alias("IA__gtk_print_settings_set_use_color"), visibility("default")));

#undef gtk_print_settings_get_collate
extern __typeof (gtk_print_settings_get_collate) gtk_print_settings_get_collate __attribute((alias("IA__gtk_print_settings_get_collate"), visibility("default")));

#undef gtk_print_settings_set_collate
extern __typeof (gtk_print_settings_set_collate) gtk_print_settings_set_collate __attribute((alias("IA__gtk_print_settings_set_collate"), visibility("default")));

#undef gtk_print_settings_get_reverse
extern __typeof (gtk_print_settings_get_reverse) gtk_print_settings_get_reverse __attribute((alias("IA__gtk_print_settings_get_reverse"), visibility("default")));

#undef gtk_print_settings_set_reverse
extern __typeof (gtk_print_settings_set_reverse) gtk_print_settings_set_reverse __attribute((alias("IA__gtk_print_settings_set_reverse"), visibility("default")));

#undef gtk_print_settings_get_duplex
extern __typeof (gtk_print_settings_get_duplex) gtk_print_settings_get_duplex __attribute((alias("IA__gtk_print_settings_get_duplex"), visibility("default")));

#undef gtk_print_settings_set_duplex
extern __typeof (gtk_print_settings_set_duplex) gtk_print_settings_set_duplex __attribute((alias("IA__gtk_print_settings_set_duplex"), visibility("default")));

#undef gtk_print_settings_get_quality
extern __typeof (gtk_print_settings_get_quality) gtk_print_settings_get_quality __attribute((alias("IA__gtk_print_settings_get_quality"), visibility("default")));

#undef gtk_print_settings_set_quality
extern __typeof (gtk_print_settings_set_quality) gtk_print_settings_set_quality __attribute((alias("IA__gtk_print_settings_set_quality"), visibility("default")));

#undef gtk_print_settings_get_n_copies
extern __typeof (gtk_print_settings_get_n_copies) gtk_print_settings_get_n_copies __attribute((alias("IA__gtk_print_settings_get_n_copies"), visibility("default")));

#undef gtk_print_settings_set_n_copies
extern __typeof (gtk_print_settings_set_n_copies) gtk_print_settings_set_n_copies __attribute((alias("IA__gtk_print_settings_set_n_copies"), visibility("default")));

#undef gtk_print_settings_get_number_up
extern __typeof (gtk_print_settings_get_number_up) gtk_print_settings_get_number_up __attribute((alias("IA__gtk_print_settings_get_number_up"), visibility("default")));

#undef gtk_print_settings_set_number_up
extern __typeof (gtk_print_settings_set_number_up) gtk_print_settings_set_number_up __attribute((alias("IA__gtk_print_settings_set_number_up"), visibility("default")));

#undef gtk_print_settings_get_number_up_layout
extern __typeof (gtk_print_settings_get_number_up_layout) gtk_print_settings_get_number_up_layout __attribute((alias("IA__gtk_print_settings_get_number_up_layout"), visibility("default")));

#undef gtk_print_settings_set_number_up_layout
extern __typeof (gtk_print_settings_set_number_up_layout) gtk_print_settings_set_number_up_layout __attribute((alias("IA__gtk_print_settings_set_number_up_layout"), visibility("default")));

#undef gtk_print_settings_get_resolution
extern __typeof (gtk_print_settings_get_resolution) gtk_print_settings_get_resolution __attribute((alias("IA__gtk_print_settings_get_resolution"), visibility("default")));

#undef gtk_print_settings_set_resolution
extern __typeof (gtk_print_settings_set_resolution) gtk_print_settings_set_resolution __attribute((alias("IA__gtk_print_settings_set_resolution"), visibility("default")));

#undef gtk_print_settings_get_resolution_x
extern __typeof (gtk_print_settings_get_resolution_x) gtk_print_settings_get_resolution_x __attribute((alias("IA__gtk_print_settings_get_resolution_x"), visibility("default")));

#undef gtk_print_settings_get_resolution_y
extern __typeof (gtk_print_settings_get_resolution_y) gtk_print_settings_get_resolution_y __attribute((alias("IA__gtk_print_settings_get_resolution_y"), visibility("default")));

#undef gtk_print_settings_set_resolution_xy
extern __typeof (gtk_print_settings_set_resolution_xy) gtk_print_settings_set_resolution_xy __attribute((alias("IA__gtk_print_settings_set_resolution_xy"), visibility("default")));

#undef gtk_print_settings_get_printer_lpi
extern __typeof (gtk_print_settings_get_printer_lpi) gtk_print_settings_get_printer_lpi __attribute((alias("IA__gtk_print_settings_get_printer_lpi"), visibility("default")));

#undef gtk_print_settings_set_printer_lpi
extern __typeof (gtk_print_settings_set_printer_lpi) gtk_print_settings_set_printer_lpi __attribute((alias("IA__gtk_print_settings_set_printer_lpi"), visibility("default")));

#undef gtk_print_settings_get_scale
extern __typeof (gtk_print_settings_get_scale) gtk_print_settings_get_scale __attribute((alias("IA__gtk_print_settings_get_scale"), visibility("default")));

#undef gtk_print_settings_set_scale
extern __typeof (gtk_print_settings_set_scale) gtk_print_settings_set_scale __attribute((alias("IA__gtk_print_settings_set_scale"), visibility("default")));

#undef gtk_print_settings_get_print_pages
extern __typeof (gtk_print_settings_get_print_pages) gtk_print_settings_get_print_pages __attribute((alias("IA__gtk_print_settings_get_print_pages"), visibility("default")));

#undef gtk_print_settings_set_print_pages
extern __typeof (gtk_print_settings_set_print_pages) gtk_print_settings_set_print_pages __attribute((alias("IA__gtk_print_settings_set_print_pages"), visibility("default")));

#undef gtk_print_settings_get_page_ranges
extern __typeof (gtk_print_settings_get_page_ranges) gtk_print_settings_get_page_ranges __attribute((alias("IA__gtk_print_settings_get_page_ranges"), visibility("default")));

#undef gtk_print_settings_set_page_ranges
extern __typeof (gtk_print_settings_set_page_ranges) gtk_print_settings_set_page_ranges __attribute((alias("IA__gtk_print_settings_set_page_ranges"), visibility("default")));

#undef gtk_print_settings_get_page_set
extern __typeof (gtk_print_settings_get_page_set) gtk_print_settings_get_page_set __attribute((alias("IA__gtk_print_settings_get_page_set"), visibility("default")));

#undef gtk_print_settings_set_page_set
extern __typeof (gtk_print_settings_set_page_set) gtk_print_settings_set_page_set __attribute((alias("IA__gtk_print_settings_set_page_set"), visibility("default")));

#undef gtk_print_settings_get_default_source
extern __typeof (gtk_print_settings_get_default_source) gtk_print_settings_get_default_source __attribute((alias("IA__gtk_print_settings_get_default_source"), visibility("default")));

#undef gtk_print_settings_set_default_source
extern __typeof (gtk_print_settings_set_default_source) gtk_print_settings_set_default_source __attribute((alias("IA__gtk_print_settings_set_default_source"), visibility("default")));

#undef gtk_print_settings_get_media_type
extern __typeof (gtk_print_settings_get_media_type) gtk_print_settings_get_media_type __attribute((alias("IA__gtk_print_settings_get_media_type"), visibility("default")));

#undef gtk_print_settings_set_media_type
extern __typeof (gtk_print_settings_set_media_type) gtk_print_settings_set_media_type __attribute((alias("IA__gtk_print_settings_set_media_type"), visibility("default")));

#undef gtk_print_settings_get_dither
extern __typeof (gtk_print_settings_get_dither) gtk_print_settings_get_dither __attribute((alias("IA__gtk_print_settings_get_dither"), visibility("default")));

#undef gtk_print_settings_set_dither
extern __typeof (gtk_print_settings_set_dither) gtk_print_settings_set_dither __attribute((alias("IA__gtk_print_settings_set_dither"), visibility("default")));

#undef gtk_print_settings_get_finishings
extern __typeof (gtk_print_settings_get_finishings) gtk_print_settings_get_finishings __attribute((alias("IA__gtk_print_settings_get_finishings"), visibility("default")));

#undef gtk_print_settings_set_finishings
extern __typeof (gtk_print_settings_set_finishings) gtk_print_settings_set_finishings __attribute((alias("IA__gtk_print_settings_set_finishings"), visibility("default")));

#undef gtk_print_settings_get_output_bin
extern __typeof (gtk_print_settings_get_output_bin) gtk_print_settings_get_output_bin __attribute((alias("IA__gtk_print_settings_get_output_bin"), visibility("default")));

#undef gtk_print_settings_set_output_bin
extern __typeof (gtk_print_settings_set_output_bin) gtk_print_settings_set_output_bin __attribute((alias("IA__gtk_print_settings_set_output_bin"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_PRINT_UNIX_DIALOG_H__)
#if IN_FILE(__GTK_PRINT_UNIX_DIALOG_C__)
#ifdef G_OS_UNIX
#undef gtk_print_unix_dialog_get_type
extern __typeof (gtk_print_unix_dialog_get_type) gtk_print_unix_dialog_get_type __attribute((alias("IA__gtk_print_unix_dialog_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_print_unix_dialog_new
extern __typeof (gtk_print_unix_dialog_new) gtk_print_unix_dialog_new __attribute((alias("IA__gtk_print_unix_dialog_new"), visibility("default")));

#undef gtk_print_unix_dialog_set_page_setup
extern __typeof (gtk_print_unix_dialog_set_page_setup) gtk_print_unix_dialog_set_page_setup __attribute((alias("IA__gtk_print_unix_dialog_set_page_setup"), visibility("default")));

#undef gtk_print_unix_dialog_get_page_setup
extern __typeof (gtk_print_unix_dialog_get_page_setup) gtk_print_unix_dialog_get_page_setup __attribute((alias("IA__gtk_print_unix_dialog_get_page_setup"), visibility("default")));

#undef gtk_print_unix_dialog_set_current_page
extern __typeof (gtk_print_unix_dialog_set_current_page) gtk_print_unix_dialog_set_current_page __attribute((alias("IA__gtk_print_unix_dialog_set_current_page"), visibility("default")));

#undef gtk_print_unix_dialog_get_current_page
extern __typeof (gtk_print_unix_dialog_get_current_page) gtk_print_unix_dialog_get_current_page __attribute((alias("IA__gtk_print_unix_dialog_get_current_page"), visibility("default")));

#undef gtk_print_unix_dialog_set_settings
extern __typeof (gtk_print_unix_dialog_set_settings) gtk_print_unix_dialog_set_settings __attribute((alias("IA__gtk_print_unix_dialog_set_settings"), visibility("default")));

#undef gtk_print_unix_dialog_get_settings
extern __typeof (gtk_print_unix_dialog_get_settings) gtk_print_unix_dialog_get_settings __attribute((alias("IA__gtk_print_unix_dialog_get_settings"), visibility("default")));

#undef gtk_print_unix_dialog_get_selected_printer
extern __typeof (gtk_print_unix_dialog_get_selected_printer) gtk_print_unix_dialog_get_selected_printer __attribute((alias("IA__gtk_print_unix_dialog_get_selected_printer"), visibility("default")));

#undef gtk_print_unix_dialog_add_custom_tab
extern __typeof (gtk_print_unix_dialog_add_custom_tab) gtk_print_unix_dialog_add_custom_tab __attribute((alias("IA__gtk_print_unix_dialog_add_custom_tab"), visibility("default")));

#undef gtk_print_unix_dialog_set_manual_capabilities
extern __typeof (gtk_print_unix_dialog_set_manual_capabilities) gtk_print_unix_dialog_set_manual_capabilities __attribute((alias("IA__gtk_print_unix_dialog_set_manual_capabilities"), visibility("default")));

#undef gtk_print_unix_dialog_get_manual_capabilities
extern __typeof (gtk_print_unix_dialog_get_manual_capabilities) gtk_print_unix_dialog_get_manual_capabilities __attribute((alias("IA__gtk_print_unix_dialog_get_manual_capabilities"), visibility("default")));

#undef gtk_print_unix_dialog_set_support_selection
extern __typeof (gtk_print_unix_dialog_set_support_selection) gtk_print_unix_dialog_set_support_selection __attribute((alias("IA__gtk_print_unix_dialog_set_support_selection"), visibility("default")));

#undef gtk_print_unix_dialog_get_support_selection
extern __typeof (gtk_print_unix_dialog_get_support_selection) gtk_print_unix_dialog_get_support_selection __attribute((alias("IA__gtk_print_unix_dialog_get_support_selection"), visibility("default")));

#undef gtk_print_unix_dialog_set_has_selection
extern __typeof (gtk_print_unix_dialog_set_has_selection) gtk_print_unix_dialog_set_has_selection __attribute((alias("IA__gtk_print_unix_dialog_set_has_selection"), visibility("default")));

#undef gtk_print_unix_dialog_get_has_selection
extern __typeof (gtk_print_unix_dialog_get_has_selection) gtk_print_unix_dialog_get_has_selection __attribute((alias("IA__gtk_print_unix_dialog_get_has_selection"), visibility("default")));

#undef gtk_print_unix_dialog_set_embed_page_setup
extern __typeof (gtk_print_unix_dialog_set_embed_page_setup) gtk_print_unix_dialog_set_embed_page_setup __attribute((alias("IA__gtk_print_unix_dialog_set_embed_page_setup"), visibility("default")));

#undef gtk_print_unix_dialog_get_embed_page_setup
extern __typeof (gtk_print_unix_dialog_get_embed_page_setup) gtk_print_unix_dialog_get_embed_page_setup __attribute((alias("IA__gtk_print_unix_dialog_get_embed_page_setup"), visibility("default")));

#undef gtk_print_unix_dialog_get_page_setup_set
extern __typeof (gtk_print_unix_dialog_get_page_setup_set) gtk_print_unix_dialog_get_page_setup_set __attribute((alias("IA__gtk_print_unix_dialog_get_page_setup_set"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_PRINT_WIN32_H__)
#if IN_FILE(__GTK_PRINT_WIN32_C__)
#ifdef G_OS_WIN32
#undef gtk_print_win32_devnames_free
extern __typeof (gtk_print_win32_devnames_free) gtk_print_win32_devnames_free __attribute((alias("IA__gtk_print_win32_devnames_free"), visibility("default")));

#undef gtk_print_win32_devnames_from_win32
extern __typeof (gtk_print_win32_devnames_from_win32) gtk_print_win32_devnames_from_win32 __attribute((alias("IA__gtk_print_win32_devnames_from_win32"), visibility("default")));

#undef gtk_print_win32_devnames_to_win32
extern __typeof (gtk_print_win32_devnames_to_win32) gtk_print_win32_devnames_to_win32 __attribute((alias("IA__gtk_print_win32_devnames_to_win32"), visibility("default")));

#undef gtk_print_win32_devnames_from_printer_name
extern __typeof (gtk_print_win32_devnames_from_printer_name) gtk_print_win32_devnames_from_printer_name __attribute((alias("IA__gtk_print_win32_devnames_from_printer_name"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_PROGRESS_BAR_H__)
#if IN_FILE(__GTK_PROGRESS_BAR_C__)
#undef gtk_progress_bar_get_fraction
extern __typeof (gtk_progress_bar_get_fraction) gtk_progress_bar_get_fraction __attribute((alias("IA__gtk_progress_bar_get_fraction"), visibility("default")));

#undef gtk_progress_bar_get_orientation
extern __typeof (gtk_progress_bar_get_orientation) gtk_progress_bar_get_orientation __attribute((alias("IA__gtk_progress_bar_get_orientation"), visibility("default")));

#undef gtk_progress_bar_get_pulse_step
extern __typeof (gtk_progress_bar_get_pulse_step) gtk_progress_bar_get_pulse_step __attribute((alias("IA__gtk_progress_bar_get_pulse_step"), visibility("default")));

#undef gtk_progress_bar_get_text
extern __typeof (gtk_progress_bar_get_text) gtk_progress_bar_get_text __attribute((alias("IA__gtk_progress_bar_get_text"), visibility("default")));

#undef gtk_progress_bar_get_ellipsize
extern __typeof (gtk_progress_bar_get_ellipsize) gtk_progress_bar_get_ellipsize __attribute((alias("IA__gtk_progress_bar_get_ellipsize"), visibility("default")));

#undef gtk_progress_bar_get_type
extern __typeof (gtk_progress_bar_get_type) gtk_progress_bar_get_type __attribute((alias("IA__gtk_progress_bar_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_progress_bar_new
extern __typeof (gtk_progress_bar_new) gtk_progress_bar_new __attribute((alias("IA__gtk_progress_bar_new"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_progress_bar_new_with_adjustment
extern __typeof (gtk_progress_bar_new_with_adjustment) gtk_progress_bar_new_with_adjustment __attribute((alias("IA__gtk_progress_bar_new_with_adjustment"), visibility("default")));

#undef gtk_progress_bar_set_bar_style
extern __typeof (gtk_progress_bar_set_bar_style) gtk_progress_bar_set_bar_style __attribute((alias("IA__gtk_progress_bar_set_bar_style"), visibility("default")));

#undef gtk_progress_bar_set_discrete_blocks
extern __typeof (gtk_progress_bar_set_discrete_blocks) gtk_progress_bar_set_discrete_blocks __attribute((alias("IA__gtk_progress_bar_set_discrete_blocks"), visibility("default")));

#undef gtk_progress_bar_set_activity_blocks
extern __typeof (gtk_progress_bar_set_activity_blocks) gtk_progress_bar_set_activity_blocks __attribute((alias("IA__gtk_progress_bar_set_activity_blocks"), visibility("default")));

#undef gtk_progress_bar_set_activity_step
extern __typeof (gtk_progress_bar_set_activity_step) gtk_progress_bar_set_activity_step __attribute((alias("IA__gtk_progress_bar_set_activity_step"), visibility("default")));

#undef gtk_progress_bar_update
extern __typeof (gtk_progress_bar_update) gtk_progress_bar_update __attribute((alias("IA__gtk_progress_bar_update"), visibility("default")));

#endif
#undef gtk_progress_bar_pulse
extern __typeof (gtk_progress_bar_pulse) gtk_progress_bar_pulse __attribute((alias("IA__gtk_progress_bar_pulse"), visibility("default")));

#undef gtk_progress_bar_set_fraction
extern __typeof (gtk_progress_bar_set_fraction) gtk_progress_bar_set_fraction __attribute((alias("IA__gtk_progress_bar_set_fraction"), visibility("default")));

#undef gtk_progress_bar_set_orientation
extern __typeof (gtk_progress_bar_set_orientation) gtk_progress_bar_set_orientation __attribute((alias("IA__gtk_progress_bar_set_orientation"), visibility("default")));

#undef gtk_progress_bar_set_pulse_step
extern __typeof (gtk_progress_bar_set_pulse_step) gtk_progress_bar_set_pulse_step __attribute((alias("IA__gtk_progress_bar_set_pulse_step"), visibility("default")));

#undef gtk_progress_bar_set_text
extern __typeof (gtk_progress_bar_set_text) gtk_progress_bar_set_text __attribute((alias("IA__gtk_progress_bar_set_text"), visibility("default")));

#undef gtk_progress_bar_set_ellipsize
extern __typeof (gtk_progress_bar_set_ellipsize) gtk_progress_bar_set_ellipsize __attribute((alias("IA__gtk_progress_bar_set_ellipsize"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_PROGRESS_H__)
#if IN_FILE(__GTK_PROGRESS_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_progress_configure
extern __typeof (gtk_progress_configure) gtk_progress_configure __attribute((alias("IA__gtk_progress_configure"), visibility("default")));

#undef gtk_progress_get_current_percentage
extern __typeof (gtk_progress_get_current_percentage) gtk_progress_get_current_percentage __attribute((alias("IA__gtk_progress_get_current_percentage"), visibility("default")));

#undef gtk_progress_get_current_text
extern __typeof (gtk_progress_get_current_text) gtk_progress_get_current_text __attribute((alias("IA__gtk_progress_get_current_text"), visibility("default")));

#undef gtk_progress_get_percentage_from_value
extern __typeof (gtk_progress_get_percentage_from_value) gtk_progress_get_percentage_from_value __attribute((alias("IA__gtk_progress_get_percentage_from_value"), visibility("default")));

#undef gtk_progress_get_text_from_value
extern __typeof (gtk_progress_get_text_from_value) gtk_progress_get_text_from_value __attribute((alias("IA__gtk_progress_get_text_from_value"), visibility("default")));

#undef gtk_progress_get_type
extern __typeof (gtk_progress_get_type) gtk_progress_get_type __attribute((alias("IA__gtk_progress_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_progress_get_value
extern __typeof (gtk_progress_get_value) gtk_progress_get_value __attribute((alias("IA__gtk_progress_get_value"), visibility("default")));

#undef gtk_progress_set_activity_mode
extern __typeof (gtk_progress_set_activity_mode) gtk_progress_set_activity_mode __attribute((alias("IA__gtk_progress_set_activity_mode"), visibility("default")));

#undef gtk_progress_set_adjustment
extern __typeof (gtk_progress_set_adjustment) gtk_progress_set_adjustment __attribute((alias("IA__gtk_progress_set_adjustment"), visibility("default")));

#undef gtk_progress_set_format_string
extern __typeof (gtk_progress_set_format_string) gtk_progress_set_format_string __attribute((alias("IA__gtk_progress_set_format_string"), visibility("default")));

#undef gtk_progress_set_percentage
extern __typeof (gtk_progress_set_percentage) gtk_progress_set_percentage __attribute((alias("IA__gtk_progress_set_percentage"), visibility("default")));

#undef gtk_progress_set_show_text
extern __typeof (gtk_progress_set_show_text) gtk_progress_set_show_text __attribute((alias("IA__gtk_progress_set_show_text"), visibility("default")));

#undef gtk_progress_set_text_alignment
extern __typeof (gtk_progress_set_text_alignment) gtk_progress_set_text_alignment __attribute((alias("IA__gtk_progress_set_text_alignment"), visibility("default")));

#undef gtk_progress_set_value
extern __typeof (gtk_progress_set_value) gtk_progress_set_value __attribute((alias("IA__gtk_progress_set_value"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_RADIO_ACTION_H__)
#if IN_FILE(__GTK_RADIO_ACTION_C__)
#undef gtk_radio_action_get_current_value
extern __typeof (gtk_radio_action_get_current_value) gtk_radio_action_get_current_value __attribute((alias("IA__gtk_radio_action_get_current_value"), visibility("default")));

#undef gtk_radio_action_get_group
extern __typeof (gtk_radio_action_get_group) gtk_radio_action_get_group __attribute((alias("IA__gtk_radio_action_get_group"), visibility("default")));

#undef gtk_radio_action_get_type
extern __typeof (gtk_radio_action_get_type) gtk_radio_action_get_type __attribute((alias("IA__gtk_radio_action_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_radio_action_new
extern __typeof (gtk_radio_action_new) gtk_radio_action_new __attribute((alias("IA__gtk_radio_action_new"), visibility("default")));

#undef gtk_radio_action_set_current_value
extern __typeof (gtk_radio_action_set_current_value) gtk_radio_action_set_current_value __attribute((alias("IA__gtk_radio_action_set_current_value"), visibility("default")));

#undef gtk_radio_action_set_group
extern __typeof (gtk_radio_action_set_group) gtk_radio_action_set_group __attribute((alias("IA__gtk_radio_action_set_group"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_RADIO_BUTTON_H__)
#if IN_FILE(__GTK_RADIO_BUTTON_C__)
#undef gtk_radio_button_get_group
extern __typeof (gtk_radio_button_get_group) gtk_radio_button_get_group __attribute((alias("IA__gtk_radio_button_get_group"), visibility("default")));

#undef gtk_radio_button_get_type
extern __typeof (gtk_radio_button_get_type) gtk_radio_button_get_type __attribute((alias("IA__gtk_radio_button_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_radio_button_new
extern __typeof (gtk_radio_button_new) gtk_radio_button_new __attribute((alias("IA__gtk_radio_button_new"), visibility("default")));

#undef gtk_radio_button_new_from_widget
extern __typeof (gtk_radio_button_new_from_widget) gtk_radio_button_new_from_widget __attribute((alias("IA__gtk_radio_button_new_from_widget"), visibility("default")));

#undef gtk_radio_button_new_with_label
extern __typeof (gtk_radio_button_new_with_label) gtk_radio_button_new_with_label __attribute((alias("IA__gtk_radio_button_new_with_label"), visibility("default")));

#undef gtk_radio_button_new_with_label_from_widget
extern __typeof (gtk_radio_button_new_with_label_from_widget) gtk_radio_button_new_with_label_from_widget __attribute((alias("IA__gtk_radio_button_new_with_label_from_widget"), visibility("default")));

#undef gtk_radio_button_new_with_mnemonic
extern __typeof (gtk_radio_button_new_with_mnemonic) gtk_radio_button_new_with_mnemonic __attribute((alias("IA__gtk_radio_button_new_with_mnemonic"), visibility("default")));

#undef gtk_radio_button_new_with_mnemonic_from_widget
extern __typeof (gtk_radio_button_new_with_mnemonic_from_widget) gtk_radio_button_new_with_mnemonic_from_widget __attribute((alias("IA__gtk_radio_button_new_with_mnemonic_from_widget"), visibility("default")));

#undef gtk_radio_button_set_group
extern __typeof (gtk_radio_button_set_group) gtk_radio_button_set_group __attribute((alias("IA__gtk_radio_button_set_group"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_RADIO_MENU_ITEM_H__)
#if IN_FILE(__GTK_RADIO_MENU_ITEM_C__)
#undef gtk_radio_menu_item_get_group
extern __typeof (gtk_radio_menu_item_get_group) gtk_radio_menu_item_get_group __attribute((alias("IA__gtk_radio_menu_item_get_group"), visibility("default")));

#undef gtk_radio_menu_item_get_type
extern __typeof (gtk_radio_menu_item_get_type) gtk_radio_menu_item_get_type __attribute((alias("IA__gtk_radio_menu_item_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_radio_menu_item_new
extern __typeof (gtk_radio_menu_item_new) gtk_radio_menu_item_new __attribute((alias("IA__gtk_radio_menu_item_new"), visibility("default")));

#undef gtk_radio_menu_item_new_from_widget
extern __typeof (gtk_radio_menu_item_new_from_widget) gtk_radio_menu_item_new_from_widget __attribute((alias("IA__gtk_radio_menu_item_new_from_widget"), visibility("default")));

#undef gtk_radio_menu_item_new_with_label
extern __typeof (gtk_radio_menu_item_new_with_label) gtk_radio_menu_item_new_with_label __attribute((alias("IA__gtk_radio_menu_item_new_with_label"), visibility("default")));

#undef gtk_radio_menu_item_new_with_label_from_widget
extern __typeof (gtk_radio_menu_item_new_with_label_from_widget) gtk_radio_menu_item_new_with_label_from_widget __attribute((alias("IA__gtk_radio_menu_item_new_with_label_from_widget"), visibility("default")));

#undef gtk_radio_menu_item_new_with_mnemonic
extern __typeof (gtk_radio_menu_item_new_with_mnemonic) gtk_radio_menu_item_new_with_mnemonic __attribute((alias("IA__gtk_radio_menu_item_new_with_mnemonic"), visibility("default")));

#undef gtk_radio_menu_item_new_with_mnemonic_from_widget
extern __typeof (gtk_radio_menu_item_new_with_mnemonic_from_widget) gtk_radio_menu_item_new_with_mnemonic_from_widget __attribute((alias("IA__gtk_radio_menu_item_new_with_mnemonic_from_widget"), visibility("default")));

#undef gtk_radio_menu_item_set_group
extern __typeof (gtk_radio_menu_item_set_group) gtk_radio_menu_item_set_group __attribute((alias("IA__gtk_radio_menu_item_set_group"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_RADIO_TOOL_BUTTON_H__)
#if IN_FILE(__GTK_RADIO_TOOL_BUTTON_C__)
#undef gtk_radio_tool_button_get_group
extern __typeof (gtk_radio_tool_button_get_group) gtk_radio_tool_button_get_group __attribute((alias("IA__gtk_radio_tool_button_get_group"), visibility("default")));

#undef gtk_radio_tool_button_get_type
extern __typeof (gtk_radio_tool_button_get_type) gtk_radio_tool_button_get_type __attribute((alias("IA__gtk_radio_tool_button_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_radio_tool_button_new
extern __typeof (gtk_radio_tool_button_new) gtk_radio_tool_button_new __attribute((alias("IA__gtk_radio_tool_button_new"), visibility("default")));

#undef gtk_radio_tool_button_new_from_stock
extern __typeof (gtk_radio_tool_button_new_from_stock) gtk_radio_tool_button_new_from_stock __attribute((alias("IA__gtk_radio_tool_button_new_from_stock"), visibility("default")));

#undef gtk_radio_tool_button_new_from_widget
extern __typeof (gtk_radio_tool_button_new_from_widget) gtk_radio_tool_button_new_from_widget __attribute((alias("IA__gtk_radio_tool_button_new_from_widget"), visibility("default")));

#undef gtk_radio_tool_button_new_with_stock_from_widget
extern __typeof (gtk_radio_tool_button_new_with_stock_from_widget) gtk_radio_tool_button_new_with_stock_from_widget __attribute((alias("IA__gtk_radio_tool_button_new_with_stock_from_widget"), visibility("default")));

#undef gtk_radio_tool_button_set_group
extern __typeof (gtk_radio_tool_button_set_group) gtk_radio_tool_button_set_group __attribute((alias("IA__gtk_radio_tool_button_set_group"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_RANGE_H__)
#if IN_FILE(__GTK_RANGE_C__)
#undef gtk_range_get_adjustment
extern __typeof (gtk_range_get_adjustment) gtk_range_get_adjustment __attribute((alias("IA__gtk_range_get_adjustment"), visibility("default")));

#undef gtk_range_get_fill_level
extern __typeof (gtk_range_get_fill_level) gtk_range_get_fill_level __attribute((alias("IA__gtk_range_get_fill_level"), visibility("default")));

#undef gtk_range_get_flippable
extern __typeof (gtk_range_get_flippable) gtk_range_get_flippable __attribute((alias("IA__gtk_range_get_flippable"), visibility("default")));

#undef gtk_range_get_inverted
extern __typeof (gtk_range_get_inverted) gtk_range_get_inverted __attribute((alias("IA__gtk_range_get_inverted"), visibility("default")));

#undef gtk_range_get_lower_stepper_sensitivity
extern __typeof (gtk_range_get_lower_stepper_sensitivity) gtk_range_get_lower_stepper_sensitivity __attribute((alias("IA__gtk_range_get_lower_stepper_sensitivity"), visibility("default")));

#undef gtk_range_get_min_slider_size
extern __typeof (gtk_range_get_min_slider_size) gtk_range_get_min_slider_size __attribute((alias("IA__gtk_range_get_min_slider_size"), visibility("default")));

#undef gtk_range_get_range_rect
extern __typeof (gtk_range_get_range_rect) gtk_range_get_range_rect __attribute((alias("IA__gtk_range_get_range_rect"), visibility("default")));

#undef gtk_range_get_restrict_to_fill_level
extern __typeof (gtk_range_get_restrict_to_fill_level) gtk_range_get_restrict_to_fill_level __attribute((alias("IA__gtk_range_get_restrict_to_fill_level"), visibility("default")));

#undef gtk_range_get_show_fill_level
extern __typeof (gtk_range_get_show_fill_level) gtk_range_get_show_fill_level __attribute((alias("IA__gtk_range_get_show_fill_level"), visibility("default")));

#undef gtk_range_get_slider_range
extern __typeof (gtk_range_get_slider_range) gtk_range_get_slider_range __attribute((alias("IA__gtk_range_get_slider_range"), visibility("default")));

#undef gtk_range_get_slider_size_fixed
extern __typeof (gtk_range_get_slider_size_fixed) gtk_range_get_slider_size_fixed __attribute((alias("IA__gtk_range_get_slider_size_fixed"), visibility("default")));

#undef gtk_range_get_type
extern __typeof (gtk_range_get_type) gtk_range_get_type __attribute((alias("IA__gtk_range_get_type"), visibility("default"))) G_GNUC_CONST;

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_range_get_update_policy
extern __typeof (gtk_range_get_update_policy) gtk_range_get_update_policy __attribute((alias("IA__gtk_range_get_update_policy"), visibility("default")));

#endif
#undef gtk_range_get_upper_stepper_sensitivity
extern __typeof (gtk_range_get_upper_stepper_sensitivity) gtk_range_get_upper_stepper_sensitivity __attribute((alias("IA__gtk_range_get_upper_stepper_sensitivity"), visibility("default")));

#undef gtk_range_get_value
extern __typeof (gtk_range_get_value) gtk_range_get_value __attribute((alias("IA__gtk_range_get_value"), visibility("default")));

#undef gtk_range_set_adjustment
extern __typeof (gtk_range_set_adjustment) gtk_range_set_adjustment __attribute((alias("IA__gtk_range_set_adjustment"), visibility("default")));

#undef gtk_range_set_fill_level
extern __typeof (gtk_range_set_fill_level) gtk_range_set_fill_level __attribute((alias("IA__gtk_range_set_fill_level"), visibility("default")));

#undef gtk_range_set_flippable
extern __typeof (gtk_range_set_flippable) gtk_range_set_flippable __attribute((alias("IA__gtk_range_set_flippable"), visibility("default")));

#undef gtk_range_set_increments
extern __typeof (gtk_range_set_increments) gtk_range_set_increments __attribute((alias("IA__gtk_range_set_increments"), visibility("default")));

#undef gtk_range_set_inverted
extern __typeof (gtk_range_set_inverted) gtk_range_set_inverted __attribute((alias("IA__gtk_range_set_inverted"), visibility("default")));

#undef gtk_range_set_lower_stepper_sensitivity
extern __typeof (gtk_range_set_lower_stepper_sensitivity) gtk_range_set_lower_stepper_sensitivity __attribute((alias("IA__gtk_range_set_lower_stepper_sensitivity"), visibility("default")));

#undef gtk_range_set_min_slider_size
extern __typeof (gtk_range_set_min_slider_size) gtk_range_set_min_slider_size __attribute((alias("IA__gtk_range_set_min_slider_size"), visibility("default")));

#undef gtk_range_set_range
extern __typeof (gtk_range_set_range) gtk_range_set_range __attribute((alias("IA__gtk_range_set_range"), visibility("default")));

#undef gtk_range_set_restrict_to_fill_level
extern __typeof (gtk_range_set_restrict_to_fill_level) gtk_range_set_restrict_to_fill_level __attribute((alias("IA__gtk_range_set_restrict_to_fill_level"), visibility("default")));

#undef gtk_range_set_show_fill_level
extern __typeof (gtk_range_set_show_fill_level) gtk_range_set_show_fill_level __attribute((alias("IA__gtk_range_set_show_fill_level"), visibility("default")));

#undef gtk_range_set_slider_size_fixed
extern __typeof (gtk_range_set_slider_size_fixed) gtk_range_set_slider_size_fixed __attribute((alias("IA__gtk_range_set_slider_size_fixed"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_range_set_update_policy
extern __typeof (gtk_range_set_update_policy) gtk_range_set_update_policy __attribute((alias("IA__gtk_range_set_update_policy"), visibility("default")));

#endif
#undef gtk_range_set_upper_stepper_sensitivity
extern __typeof (gtk_range_set_upper_stepper_sensitivity) gtk_range_set_upper_stepper_sensitivity __attribute((alias("IA__gtk_range_set_upper_stepper_sensitivity"), visibility("default")));

#undef gtk_range_set_value
extern __typeof (gtk_range_set_value) gtk_range_set_value __attribute((alias("IA__gtk_range_set_value"), visibility("default")));

#undef gtk_range_set_round_digits
extern __typeof (gtk_range_set_round_digits) gtk_range_set_round_digits __attribute((alias("IA__gtk_range_set_round_digits"), visibility("default")));

#undef gtk_range_get_round_digits
extern __typeof (gtk_range_get_round_digits) gtk_range_get_round_digits __attribute((alias("IA__gtk_range_get_round_digits"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_RC_H__)
#if IN_FILE(__GTK_RC_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_rc_add_class_style
extern __typeof (gtk_rc_add_class_style) gtk_rc_add_class_style __attribute((alias("IA__gtk_rc_add_class_style"), visibility("default")));

#undef gtk_rc_add_widget_class_style
extern __typeof (gtk_rc_add_widget_class_style) gtk_rc_add_widget_class_style __attribute((alias("IA__gtk_rc_add_widget_class_style"), visibility("default")));

#undef gtk_rc_add_widget_name_style
extern __typeof (gtk_rc_add_widget_name_style) gtk_rc_add_widget_name_style __attribute((alias("IA__gtk_rc_add_widget_name_style"), visibility("default")));

#endif
#ifndef _WIN64
#undef gtk_rc_add_default_file
extern __typeof (gtk_rc_add_default_file) gtk_rc_add_default_file __attribute((alias("IA__gtk_rc_add_default_file"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_rc_add_default_file_utf8
extern __typeof (gtk_rc_add_default_file_utf8) gtk_rc_add_default_file_utf8 __attribute((alias("IA__gtk_rc_add_default_file_utf8"), visibility("default")));

#endif
#undef gtk_rc_find_module_in_path
extern __typeof (gtk_rc_find_module_in_path) gtk_rc_find_module_in_path __attribute((alias("IA__gtk_rc_find_module_in_path"), visibility("default")));

#undef gtk_rc_find_pixmap_in_path
extern __typeof (gtk_rc_find_pixmap_in_path) gtk_rc_find_pixmap_in_path __attribute((alias("IA__gtk_rc_find_pixmap_in_path"), visibility("default")));

#undef gtk_rc_get_default_files
extern __typeof (gtk_rc_get_default_files) gtk_rc_get_default_files __attribute((alias("IA__gtk_rc_get_default_files"), visibility("default")));

#undef gtk_rc_get_im_module_file
extern __typeof (gtk_rc_get_im_module_file) gtk_rc_get_im_module_file __attribute((alias("IA__gtk_rc_get_im_module_file"), visibility("default")));

#undef gtk_rc_get_im_module_path
extern __typeof (gtk_rc_get_im_module_path) gtk_rc_get_im_module_path __attribute((alias("IA__gtk_rc_get_im_module_path"), visibility("default")));

#undef gtk_rc_get_module_dir
extern __typeof (gtk_rc_get_module_dir) gtk_rc_get_module_dir __attribute((alias("IA__gtk_rc_get_module_dir"), visibility("default")));

#undef gtk_rc_get_style
extern __typeof (gtk_rc_get_style) gtk_rc_get_style __attribute((alias("IA__gtk_rc_get_style"), visibility("default")));

#undef gtk_rc_get_style_by_paths
extern __typeof (gtk_rc_get_style_by_paths) gtk_rc_get_style_by_paths __attribute((alias("IA__gtk_rc_get_style_by_paths"), visibility("default")));

#undef gtk_rc_get_theme_dir
extern __typeof (gtk_rc_get_theme_dir) gtk_rc_get_theme_dir __attribute((alias("IA__gtk_rc_get_theme_dir"), visibility("default")));

#ifndef _WIN64
#undef gtk_rc_parse
extern __typeof (gtk_rc_parse) gtk_rc_parse __attribute((alias("IA__gtk_rc_parse"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_rc_parse_utf8
extern __typeof (gtk_rc_parse_utf8) gtk_rc_parse_utf8 __attribute((alias("IA__gtk_rc_parse_utf8"), visibility("default")));

#endif
#undef gtk_rc_parse_color
extern __typeof (gtk_rc_parse_color) gtk_rc_parse_color __attribute((alias("IA__gtk_rc_parse_color"), visibility("default")));

#undef gtk_rc_parse_color_full
extern __typeof (gtk_rc_parse_color_full) gtk_rc_parse_color_full __attribute((alias("IA__gtk_rc_parse_color_full"), visibility("default")));

#undef gtk_rc_parse_priority
extern __typeof (gtk_rc_parse_priority) gtk_rc_parse_priority __attribute((alias("IA__gtk_rc_parse_priority"), visibility("default")));

#undef gtk_rc_parse_state
extern __typeof (gtk_rc_parse_state) gtk_rc_parse_state __attribute((alias("IA__gtk_rc_parse_state"), visibility("default")));

#undef gtk_rc_parse_string
extern __typeof (gtk_rc_parse_string) gtk_rc_parse_string __attribute((alias("IA__gtk_rc_parse_string"), visibility("default")));

#undef gtk_rc_reparse_all
extern __typeof (gtk_rc_reparse_all) gtk_rc_reparse_all __attribute((alias("IA__gtk_rc_reparse_all"), visibility("default")));

#undef gtk_rc_reparse_all_for_settings
extern __typeof (gtk_rc_reparse_all_for_settings) gtk_rc_reparse_all_for_settings __attribute((alias("IA__gtk_rc_reparse_all_for_settings"), visibility("default")));

#undef gtk_rc_reset_styles
extern __typeof (gtk_rc_reset_styles) gtk_rc_reset_styles __attribute((alias("IA__gtk_rc_reset_styles"), visibility("default")));

#undef gtk_rc_scanner_new
extern __typeof (gtk_rc_scanner_new) gtk_rc_scanner_new __attribute((alias("IA__gtk_rc_scanner_new"), visibility("default")));

#ifndef _WIN64
#undef gtk_rc_set_default_files
extern __typeof (gtk_rc_set_default_files) gtk_rc_set_default_files __attribute((alias("IA__gtk_rc_set_default_files"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_rc_set_default_files_utf8
extern __typeof (gtk_rc_set_default_files_utf8) gtk_rc_set_default_files_utf8 __attribute((alias("IA__gtk_rc_set_default_files_utf8"), visibility("default")));

#endif
#undef gtk_rc_style_copy
extern __typeof (gtk_rc_style_copy) gtk_rc_style_copy __attribute((alias("IA__gtk_rc_style_copy"), visibility("default")));

#undef gtk_rc_style_get_type
extern __typeof (gtk_rc_style_get_type) gtk_rc_style_get_type __attribute((alias("IA__gtk_rc_style_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_rc_style_new
extern __typeof (gtk_rc_style_new) gtk_rc_style_new __attribute((alias("IA__gtk_rc_style_new"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_rc_style_ref
extern __typeof (gtk_rc_style_ref) gtk_rc_style_ref __attribute((alias("IA__gtk_rc_style_ref"), visibility("default")));

#undef gtk_rc_style_unref
extern __typeof (gtk_rc_style_unref) gtk_rc_style_unref __attribute((alias("IA__gtk_rc_style_unref"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_RECENT_ACTION_H__)
#if IN_FILE(__GTK_RECENT_ACTION_C__)
#undef gtk_recent_action_get_type
extern __typeof (gtk_recent_action_get_type) gtk_recent_action_get_type __attribute((alias("IA__gtk_recent_action_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_recent_action_new
extern __typeof (gtk_recent_action_new) gtk_recent_action_new __attribute((alias("IA__gtk_recent_action_new"), visibility("default")));

#undef gtk_recent_action_new_for_manager
extern __typeof (gtk_recent_action_new_for_manager) gtk_recent_action_new_for_manager __attribute((alias("IA__gtk_recent_action_new_for_manager"), visibility("default")));

#undef gtk_recent_action_set_show_numbers
extern __typeof (gtk_recent_action_set_show_numbers) gtk_recent_action_set_show_numbers __attribute((alias("IA__gtk_recent_action_set_show_numbers"), visibility("default")));

#undef gtk_recent_action_get_show_numbers
extern __typeof (gtk_recent_action_get_show_numbers) gtk_recent_action_get_show_numbers __attribute((alias("IA__gtk_recent_action_get_show_numbers"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_RECENT_CHOOSER_H__)
#if IN_FILE(__GTK_RECENT_CHOOSER_C__)
#undef gtk_recent_chooser_set_show_private
extern __typeof (gtk_recent_chooser_set_show_private) gtk_recent_chooser_set_show_private __attribute((alias("IA__gtk_recent_chooser_set_show_private"), visibility("default")));

#undef gtk_recent_chooser_get_show_private
extern __typeof (gtk_recent_chooser_get_show_private) gtk_recent_chooser_get_show_private __attribute((alias("IA__gtk_recent_chooser_get_show_private"), visibility("default")));

#undef gtk_recent_chooser_set_show_not_found
extern __typeof (gtk_recent_chooser_set_show_not_found) gtk_recent_chooser_set_show_not_found __attribute((alias("IA__gtk_recent_chooser_set_show_not_found"), visibility("default")));

#undef gtk_recent_chooser_get_show_not_found
extern __typeof (gtk_recent_chooser_get_show_not_found) gtk_recent_chooser_get_show_not_found __attribute((alias("IA__gtk_recent_chooser_get_show_not_found"), visibility("default")));

#undef gtk_recent_chooser_set_show_icons
extern __typeof (gtk_recent_chooser_set_show_icons) gtk_recent_chooser_set_show_icons __attribute((alias("IA__gtk_recent_chooser_set_show_icons"), visibility("default")));

#undef gtk_recent_chooser_get_show_icons
extern __typeof (gtk_recent_chooser_get_show_icons) gtk_recent_chooser_get_show_icons __attribute((alias("IA__gtk_recent_chooser_get_show_icons"), visibility("default")));

#undef gtk_recent_chooser_set_select_multiple
extern __typeof (gtk_recent_chooser_set_select_multiple) gtk_recent_chooser_set_select_multiple __attribute((alias("IA__gtk_recent_chooser_set_select_multiple"), visibility("default")));

#undef gtk_recent_chooser_get_select_multiple
extern __typeof (gtk_recent_chooser_get_select_multiple) gtk_recent_chooser_get_select_multiple __attribute((alias("IA__gtk_recent_chooser_get_select_multiple"), visibility("default")));

#undef gtk_recent_chooser_set_local_only
extern __typeof (gtk_recent_chooser_set_local_only) gtk_recent_chooser_set_local_only __attribute((alias("IA__gtk_recent_chooser_set_local_only"), visibility("default")));

#undef gtk_recent_chooser_get_local_only
extern __typeof (gtk_recent_chooser_get_local_only) gtk_recent_chooser_get_local_only __attribute((alias("IA__gtk_recent_chooser_get_local_only"), visibility("default")));

#undef gtk_recent_chooser_set_limit
extern __typeof (gtk_recent_chooser_set_limit) gtk_recent_chooser_set_limit __attribute((alias("IA__gtk_recent_chooser_set_limit"), visibility("default")));

#undef gtk_recent_chooser_get_limit
extern __typeof (gtk_recent_chooser_get_limit) gtk_recent_chooser_get_limit __attribute((alias("IA__gtk_recent_chooser_get_limit"), visibility("default")));

#undef gtk_recent_chooser_set_show_tips
extern __typeof (gtk_recent_chooser_set_show_tips) gtk_recent_chooser_set_show_tips __attribute((alias("IA__gtk_recent_chooser_set_show_tips"), visibility("default")));

#undef gtk_recent_chooser_get_show_tips
extern __typeof (gtk_recent_chooser_get_show_tips) gtk_recent_chooser_get_show_tips __attribute((alias("IA__gtk_recent_chooser_get_show_tips"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_recent_chooser_set_show_numbers
extern __typeof (gtk_recent_chooser_set_show_numbers) gtk_recent_chooser_set_show_numbers __attribute((alias("IA__gtk_recent_chooser_set_show_numbers"), visibility("default")));

#undef gtk_recent_chooser_get_show_numbers
extern __typeof (gtk_recent_chooser_get_show_numbers) gtk_recent_chooser_get_show_numbers __attribute((alias("IA__gtk_recent_chooser_get_show_numbers"), visibility("default")));

#endif
#undef gtk_recent_chooser_set_sort_type
extern __typeof (gtk_recent_chooser_set_sort_type) gtk_recent_chooser_set_sort_type __attribute((alias("IA__gtk_recent_chooser_set_sort_type"), visibility("default")));

#undef gtk_recent_chooser_get_sort_type
extern __typeof (gtk_recent_chooser_get_sort_type) gtk_recent_chooser_get_sort_type __attribute((alias("IA__gtk_recent_chooser_get_sort_type"), visibility("default")));

#undef gtk_recent_chooser_set_sort_func
extern __typeof (gtk_recent_chooser_set_sort_func) gtk_recent_chooser_set_sort_func __attribute((alias("IA__gtk_recent_chooser_set_sort_func"), visibility("default")));

#undef gtk_recent_chooser_set_current_uri
extern __typeof (gtk_recent_chooser_set_current_uri) gtk_recent_chooser_set_current_uri __attribute((alias("IA__gtk_recent_chooser_set_current_uri"), visibility("default")));

#undef gtk_recent_chooser_get_current_uri
extern __typeof (gtk_recent_chooser_get_current_uri) gtk_recent_chooser_get_current_uri __attribute((alias("IA__gtk_recent_chooser_get_current_uri"), visibility("default")));

#undef gtk_recent_chooser_get_current_item
extern __typeof (gtk_recent_chooser_get_current_item) gtk_recent_chooser_get_current_item __attribute((alias("IA__gtk_recent_chooser_get_current_item"), visibility("default")));

#undef gtk_recent_chooser_select_uri
extern __typeof (gtk_recent_chooser_select_uri) gtk_recent_chooser_select_uri __attribute((alias("IA__gtk_recent_chooser_select_uri"), visibility("default")));

#undef gtk_recent_chooser_unselect_uri
extern __typeof (gtk_recent_chooser_unselect_uri) gtk_recent_chooser_unselect_uri __attribute((alias("IA__gtk_recent_chooser_unselect_uri"), visibility("default")));

#undef gtk_recent_chooser_select_all
extern __typeof (gtk_recent_chooser_select_all) gtk_recent_chooser_select_all __attribute((alias("IA__gtk_recent_chooser_select_all"), visibility("default")));

#undef gtk_recent_chooser_unselect_all
extern __typeof (gtk_recent_chooser_unselect_all) gtk_recent_chooser_unselect_all __attribute((alias("IA__gtk_recent_chooser_unselect_all"), visibility("default")));

#undef gtk_recent_chooser_get_items
extern __typeof (gtk_recent_chooser_get_items) gtk_recent_chooser_get_items __attribute((alias("IA__gtk_recent_chooser_get_items"), visibility("default")));

#undef gtk_recent_chooser_get_uris
extern __typeof (gtk_recent_chooser_get_uris) gtk_recent_chooser_get_uris __attribute((alias("IA__gtk_recent_chooser_get_uris"), visibility("default")));

#undef gtk_recent_chooser_add_filter
extern __typeof (gtk_recent_chooser_add_filter) gtk_recent_chooser_add_filter __attribute((alias("IA__gtk_recent_chooser_add_filter"), visibility("default")));

#undef gtk_recent_chooser_remove_filter
extern __typeof (gtk_recent_chooser_remove_filter) gtk_recent_chooser_remove_filter __attribute((alias("IA__gtk_recent_chooser_remove_filter"), visibility("default")));

#undef gtk_recent_chooser_list_filters
extern __typeof (gtk_recent_chooser_list_filters) gtk_recent_chooser_list_filters __attribute((alias("IA__gtk_recent_chooser_list_filters"), visibility("default")));

#undef gtk_recent_chooser_set_filter
extern __typeof (gtk_recent_chooser_set_filter) gtk_recent_chooser_set_filter __attribute((alias("IA__gtk_recent_chooser_set_filter"), visibility("default")));

#undef gtk_recent_chooser_get_filter
extern __typeof (gtk_recent_chooser_get_filter) gtk_recent_chooser_get_filter __attribute((alias("IA__gtk_recent_chooser_get_filter"), visibility("default")));

#undef gtk_recent_chooser_get_type
extern __typeof (gtk_recent_chooser_get_type) gtk_recent_chooser_get_type __attribute((alias("IA__gtk_recent_chooser_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_recent_chooser_error_quark
extern __typeof (gtk_recent_chooser_error_quark) gtk_recent_chooser_error_quark __attribute((alias("IA__gtk_recent_chooser_error_quark"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_RECENT_CHOOSER_DIALOG_H__)
#if IN_FILE(__GTK_RECENT_CHOOSER_DIALOG_C__)
#undef gtk_recent_chooser_dialog_get_type
extern __typeof (gtk_recent_chooser_dialog_get_type) gtk_recent_chooser_dialog_get_type __attribute((alias("IA__gtk_recent_chooser_dialog_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_recent_chooser_dialog_new
extern __typeof (gtk_recent_chooser_dialog_new) gtk_recent_chooser_dialog_new __attribute((alias("IA__gtk_recent_chooser_dialog_new"), visibility("default"))) G_GNUC_NULL_TERMINATED;

#undef gtk_recent_chooser_dialog_new_for_manager
extern __typeof (gtk_recent_chooser_dialog_new_for_manager) gtk_recent_chooser_dialog_new_for_manager __attribute((alias("IA__gtk_recent_chooser_dialog_new_for_manager"), visibility("default"))) G_GNUC_NULL_TERMINATED;

#endif
#endif
#if IN_HEADER(__GTK_RECENT_CHOOSER_MENU_H__)
#if IN_FILE(__GTK_RECENT_CHOOSER_MENU_C__)
#undef gtk_recent_chooser_menu_get_type
extern __typeof (gtk_recent_chooser_menu_get_type) gtk_recent_chooser_menu_get_type __attribute((alias("IA__gtk_recent_chooser_menu_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_recent_chooser_menu_new
extern __typeof (gtk_recent_chooser_menu_new) gtk_recent_chooser_menu_new __attribute((alias("IA__gtk_recent_chooser_menu_new"), visibility("default")));

#undef gtk_recent_chooser_menu_new_for_manager
extern __typeof (gtk_recent_chooser_menu_new_for_manager) gtk_recent_chooser_menu_new_for_manager __attribute((alias("IA__gtk_recent_chooser_menu_new_for_manager"), visibility("default")));

#undef gtk_recent_chooser_menu_get_show_numbers
extern __typeof (gtk_recent_chooser_menu_get_show_numbers) gtk_recent_chooser_menu_get_show_numbers __attribute((alias("IA__gtk_recent_chooser_menu_get_show_numbers"), visibility("default")));

#undef gtk_recent_chooser_menu_set_show_numbers
extern __typeof (gtk_recent_chooser_menu_set_show_numbers) gtk_recent_chooser_menu_set_show_numbers __attribute((alias("IA__gtk_recent_chooser_menu_set_show_numbers"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_RECENT_CHOOSER_WIDGET_H__)
#if IN_FILE(__GTK_RECENT_CHOOSER_WIDGET_C__)
#undef gtk_recent_chooser_widget_get_type
extern __typeof (gtk_recent_chooser_widget_get_type) gtk_recent_chooser_widget_get_type __attribute((alias("IA__gtk_recent_chooser_widget_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_recent_chooser_widget_new
extern __typeof (gtk_recent_chooser_widget_new) gtk_recent_chooser_widget_new __attribute((alias("IA__gtk_recent_chooser_widget_new"), visibility("default")));

#undef gtk_recent_chooser_widget_new_for_manager
extern __typeof (gtk_recent_chooser_widget_new_for_manager) gtk_recent_chooser_widget_new_for_manager __attribute((alias("IA__gtk_recent_chooser_widget_new_for_manager"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_RECENT_FILTER_H__)
#if IN_FILE(__GTK_RECENT_FILTER_C__)
#undef gtk_recent_filter_get_type
extern __typeof (gtk_recent_filter_get_type) gtk_recent_filter_get_type __attribute((alias("IA__gtk_recent_filter_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_recent_filter_new
extern __typeof (gtk_recent_filter_new) gtk_recent_filter_new __attribute((alias("IA__gtk_recent_filter_new"), visibility("default")));

#undef gtk_recent_filter_set_name
extern __typeof (gtk_recent_filter_set_name) gtk_recent_filter_set_name __attribute((alias("IA__gtk_recent_filter_set_name"), visibility("default")));

#undef gtk_recent_filter_get_name
extern __typeof (gtk_recent_filter_get_name) gtk_recent_filter_get_name __attribute((alias("IA__gtk_recent_filter_get_name"), visibility("default")));

#undef gtk_recent_filter_add_mime_type
extern __typeof (gtk_recent_filter_add_mime_type) gtk_recent_filter_add_mime_type __attribute((alias("IA__gtk_recent_filter_add_mime_type"), visibility("default")));

#undef gtk_recent_filter_add_pattern
extern __typeof (gtk_recent_filter_add_pattern) gtk_recent_filter_add_pattern __attribute((alias("IA__gtk_recent_filter_add_pattern"), visibility("default")));

#undef gtk_recent_filter_add_pixbuf_formats
extern __typeof (gtk_recent_filter_add_pixbuf_formats) gtk_recent_filter_add_pixbuf_formats __attribute((alias("IA__gtk_recent_filter_add_pixbuf_formats"), visibility("default")));

#undef gtk_recent_filter_add_application
extern __typeof (gtk_recent_filter_add_application) gtk_recent_filter_add_application __attribute((alias("IA__gtk_recent_filter_add_application"), visibility("default")));

#undef gtk_recent_filter_add_group
extern __typeof (gtk_recent_filter_add_group) gtk_recent_filter_add_group __attribute((alias("IA__gtk_recent_filter_add_group"), visibility("default")));

#undef gtk_recent_filter_add_age
extern __typeof (gtk_recent_filter_add_age) gtk_recent_filter_add_age __attribute((alias("IA__gtk_recent_filter_add_age"), visibility("default")));

#undef gtk_recent_filter_add_custom
extern __typeof (gtk_recent_filter_add_custom) gtk_recent_filter_add_custom __attribute((alias("IA__gtk_recent_filter_add_custom"), visibility("default")));

#undef gtk_recent_filter_get_needed
extern __typeof (gtk_recent_filter_get_needed) gtk_recent_filter_get_needed __attribute((alias("IA__gtk_recent_filter_get_needed"), visibility("default")));

#undef gtk_recent_filter_filter
extern __typeof (gtk_recent_filter_filter) gtk_recent_filter_filter __attribute((alias("IA__gtk_recent_filter_filter"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_RECENT_MANAGER_H__)
#if IN_FILE(__GTK_RECENT_MANAGER_C__)
#undef gtk_recent_manager_error_quark
extern __typeof (gtk_recent_manager_error_quark) gtk_recent_manager_error_quark __attribute((alias("IA__gtk_recent_manager_error_quark"), visibility("default")));

#undef gtk_recent_manager_get_type
extern __typeof (gtk_recent_manager_get_type) gtk_recent_manager_get_type __attribute((alias("IA__gtk_recent_manager_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_recent_manager_new
extern __typeof (gtk_recent_manager_new) gtk_recent_manager_new __attribute((alias("IA__gtk_recent_manager_new"), visibility("default")));

#undef gtk_recent_manager_get_items
extern __typeof (gtk_recent_manager_get_items) gtk_recent_manager_get_items __attribute((alias("IA__gtk_recent_manager_get_items"), visibility("default")));

#undef gtk_recent_manager_get_default
extern __typeof (gtk_recent_manager_get_default) gtk_recent_manager_get_default __attribute((alias("IA__gtk_recent_manager_get_default"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_recent_manager_get_for_screen
extern __typeof (gtk_recent_manager_get_for_screen) gtk_recent_manager_get_for_screen __attribute((alias("IA__gtk_recent_manager_get_for_screen"), visibility("default")));

#undef gtk_recent_manager_set_screen
extern __typeof (gtk_recent_manager_set_screen) gtk_recent_manager_set_screen __attribute((alias("IA__gtk_recent_manager_set_screen"), visibility("default")));

#endif
#undef gtk_recent_manager_add_item
extern __typeof (gtk_recent_manager_add_item) gtk_recent_manager_add_item __attribute((alias("IA__gtk_recent_manager_add_item"), visibility("default")));

#undef gtk_recent_manager_add_full
extern __typeof (gtk_recent_manager_add_full) gtk_recent_manager_add_full __attribute((alias("IA__gtk_recent_manager_add_full"), visibility("default")));

#undef gtk_recent_manager_remove_item
extern __typeof (gtk_recent_manager_remove_item) gtk_recent_manager_remove_item __attribute((alias("IA__gtk_recent_manager_remove_item"), visibility("default")));

#undef gtk_recent_manager_lookup_item
extern __typeof (gtk_recent_manager_lookup_item) gtk_recent_manager_lookup_item __attribute((alias("IA__gtk_recent_manager_lookup_item"), visibility("default")));

#undef gtk_recent_manager_has_item
extern __typeof (gtk_recent_manager_has_item) gtk_recent_manager_has_item __attribute((alias("IA__gtk_recent_manager_has_item"), visibility("default")));

#undef gtk_recent_manager_move_item
extern __typeof (gtk_recent_manager_move_item) gtk_recent_manager_move_item __attribute((alias("IA__gtk_recent_manager_move_item"), visibility("default")));

#undef gtk_recent_manager_set_limit
extern __typeof (gtk_recent_manager_set_limit) gtk_recent_manager_set_limit __attribute((alias("IA__gtk_recent_manager_set_limit"), visibility("default")));

#undef gtk_recent_manager_get_limit
extern __typeof (gtk_recent_manager_get_limit) gtk_recent_manager_get_limit __attribute((alias("IA__gtk_recent_manager_get_limit"), visibility("default")));

#undef gtk_recent_manager_purge_items
extern __typeof (gtk_recent_manager_purge_items) gtk_recent_manager_purge_items __attribute((alias("IA__gtk_recent_manager_purge_items"), visibility("default")));

#undef gtk_recent_info_get_type
extern __typeof (gtk_recent_info_get_type) gtk_recent_info_get_type __attribute((alias("IA__gtk_recent_info_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_recent_info_ref
extern __typeof (gtk_recent_info_ref) gtk_recent_info_ref __attribute((alias("IA__gtk_recent_info_ref"), visibility("default")));

#undef gtk_recent_info_unref
extern __typeof (gtk_recent_info_unref) gtk_recent_info_unref __attribute((alias("IA__gtk_recent_info_unref"), visibility("default")));

#undef gtk_recent_info_get_uri
extern __typeof (gtk_recent_info_get_uri) gtk_recent_info_get_uri __attribute((alias("IA__gtk_recent_info_get_uri"), visibility("default")));

#undef gtk_recent_info_get_display_name
extern __typeof (gtk_recent_info_get_display_name) gtk_recent_info_get_display_name __attribute((alias("IA__gtk_recent_info_get_display_name"), visibility("default")));

#undef gtk_recent_info_get_description
extern __typeof (gtk_recent_info_get_description) gtk_recent_info_get_description __attribute((alias("IA__gtk_recent_info_get_description"), visibility("default")));

#undef gtk_recent_info_get_mime_type
extern __typeof (gtk_recent_info_get_mime_type) gtk_recent_info_get_mime_type __attribute((alias("IA__gtk_recent_info_get_mime_type"), visibility("default")));

#undef gtk_recent_info_get_added
extern __typeof (gtk_recent_info_get_added) gtk_recent_info_get_added __attribute((alias("IA__gtk_recent_info_get_added"), visibility("default")));

#undef gtk_recent_info_get_modified
extern __typeof (gtk_recent_info_get_modified) gtk_recent_info_get_modified __attribute((alias("IA__gtk_recent_info_get_modified"), visibility("default")));

#undef gtk_recent_info_get_visited
extern __typeof (gtk_recent_info_get_visited) gtk_recent_info_get_visited __attribute((alias("IA__gtk_recent_info_get_visited"), visibility("default")));

#undef gtk_recent_info_get_private_hint
extern __typeof (gtk_recent_info_get_private_hint) gtk_recent_info_get_private_hint __attribute((alias("IA__gtk_recent_info_get_private_hint"), visibility("default")));

#undef gtk_recent_info_get_application_info
extern __typeof (gtk_recent_info_get_application_info) gtk_recent_info_get_application_info __attribute((alias("IA__gtk_recent_info_get_application_info"), visibility("default")));

#undef gtk_recent_info_get_applications
extern __typeof (gtk_recent_info_get_applications) gtk_recent_info_get_applications __attribute((alias("IA__gtk_recent_info_get_applications"), visibility("default"))) G_GNUC_MALLOC;

#undef gtk_recent_info_last_application
extern __typeof (gtk_recent_info_last_application) gtk_recent_info_last_application __attribute((alias("IA__gtk_recent_info_last_application"), visibility("default"))) G_GNUC_MALLOC;

#undef gtk_recent_info_has_application
extern __typeof (gtk_recent_info_has_application) gtk_recent_info_has_application __attribute((alias("IA__gtk_recent_info_has_application"), visibility("default")));

#undef gtk_recent_info_get_groups
extern __typeof (gtk_recent_info_get_groups) gtk_recent_info_get_groups __attribute((alias("IA__gtk_recent_info_get_groups"), visibility("default"))) G_GNUC_MALLOC;

#undef gtk_recent_info_has_group
extern __typeof (gtk_recent_info_has_group) gtk_recent_info_has_group __attribute((alias("IA__gtk_recent_info_has_group"), visibility("default")));

#undef gtk_recent_info_get_icon
extern __typeof (gtk_recent_info_get_icon) gtk_recent_info_get_icon __attribute((alias("IA__gtk_recent_info_get_icon"), visibility("default")));

#undef gtk_recent_info_get_short_name
extern __typeof (gtk_recent_info_get_short_name) gtk_recent_info_get_short_name __attribute((alias("IA__gtk_recent_info_get_short_name"), visibility("default"))) G_GNUC_MALLOC;

#undef gtk_recent_info_get_uri_display
extern __typeof (gtk_recent_info_get_uri_display) gtk_recent_info_get_uri_display __attribute((alias("IA__gtk_recent_info_get_uri_display"), visibility("default"))) G_GNUC_MALLOC;

#undef gtk_recent_info_get_age
extern __typeof (gtk_recent_info_get_age) gtk_recent_info_get_age __attribute((alias("IA__gtk_recent_info_get_age"), visibility("default")));

#undef gtk_recent_info_is_local
extern __typeof (gtk_recent_info_is_local) gtk_recent_info_is_local __attribute((alias("IA__gtk_recent_info_is_local"), visibility("default")));

#undef gtk_recent_info_exists
extern __typeof (gtk_recent_info_exists) gtk_recent_info_exists __attribute((alias("IA__gtk_recent_info_exists"), visibility("default")));

#undef gtk_recent_info_match
extern __typeof (gtk_recent_info_match) gtk_recent_info_match __attribute((alias("IA__gtk_recent_info_match"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEST_UTILS_H__)
#if IN_FILE(__GTK_TEST_UTILS_C__)
#undef gtk_test_create_simple_window
extern __typeof (gtk_test_create_simple_window) gtk_test_create_simple_window __attribute((alias("IA__gtk_test_create_simple_window"), visibility("default")));

#undef gtk_test_create_widget
extern __typeof (gtk_test_create_widget) gtk_test_create_widget __attribute((alias("IA__gtk_test_create_widget"), visibility("default")));

#undef gtk_test_display_button_window
extern __typeof (gtk_test_display_button_window) gtk_test_display_button_window __attribute((alias("IA__gtk_test_display_button_window"), visibility("default")));

#undef gtk_test_find_label
extern __typeof (gtk_test_find_label) gtk_test_find_label __attribute((alias("IA__gtk_test_find_label"), visibility("default")));

#undef gtk_test_find_sibling
extern __typeof (gtk_test_find_sibling) gtk_test_find_sibling __attribute((alias("IA__gtk_test_find_sibling"), visibility("default")));

#undef gtk_test_find_widget
extern __typeof (gtk_test_find_widget) gtk_test_find_widget __attribute((alias("IA__gtk_test_find_widget"), visibility("default")));

#undef gtk_test_init
extern __typeof (gtk_test_init) gtk_test_init __attribute((alias("IA__gtk_test_init"), visibility("default")));

#undef gtk_test_register_all_types
extern __typeof (gtk_test_register_all_types) gtk_test_register_all_types __attribute((alias("IA__gtk_test_register_all_types"), visibility("default")));

#undef gtk_test_list_all_types
extern __typeof (gtk_test_list_all_types) gtk_test_list_all_types __attribute((alias("IA__gtk_test_list_all_types"), visibility("default")));

#undef gtk_test_slider_get_value
extern __typeof (gtk_test_slider_get_value) gtk_test_slider_get_value __attribute((alias("IA__gtk_test_slider_get_value"), visibility("default")));

#undef gtk_test_slider_set_perc
extern __typeof (gtk_test_slider_set_perc) gtk_test_slider_set_perc __attribute((alias("IA__gtk_test_slider_set_perc"), visibility("default")));

#undef gtk_test_spin_button_click
extern __typeof (gtk_test_spin_button_click) gtk_test_spin_button_click __attribute((alias("IA__gtk_test_spin_button_click"), visibility("default")));

#undef gtk_test_text_get
extern __typeof (gtk_test_text_get) gtk_test_text_get __attribute((alias("IA__gtk_test_text_get"), visibility("default")));

#undef gtk_test_text_set
extern __typeof (gtk_test_text_set) gtk_test_text_set __attribute((alias("IA__gtk_test_text_set"), visibility("default")));

#undef gtk_test_widget_click
extern __typeof (gtk_test_widget_click) gtk_test_widget_click __attribute((alias("IA__gtk_test_widget_click"), visibility("default")));

#undef gtk_test_widget_send_key
extern __typeof (gtk_test_widget_send_key) gtk_test_widget_send_key __attribute((alias("IA__gtk_test_widget_send_key"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEXT_BUFFER_RICH_TEXT_H__)
#if IN_FILE(__GTK_TEXT_BUFFER_RICH_TEXT_C__)
#undef gtk_text_buffer_deserialize
extern __typeof (gtk_text_buffer_deserialize) gtk_text_buffer_deserialize __attribute((alias("IA__gtk_text_buffer_deserialize"), visibility("default")));

#undef gtk_text_buffer_deserialize_get_can_create_tags
extern __typeof (gtk_text_buffer_deserialize_get_can_create_tags) gtk_text_buffer_deserialize_get_can_create_tags __attribute((alias("IA__gtk_text_buffer_deserialize_get_can_create_tags"), visibility("default")));

#undef gtk_text_buffer_deserialize_set_can_create_tags
extern __typeof (gtk_text_buffer_deserialize_set_can_create_tags) gtk_text_buffer_deserialize_set_can_create_tags __attribute((alias("IA__gtk_text_buffer_deserialize_set_can_create_tags"), visibility("default")));

#undef gtk_text_buffer_get_deserialize_formats
extern __typeof (gtk_text_buffer_get_deserialize_formats) gtk_text_buffer_get_deserialize_formats __attribute((alias("IA__gtk_text_buffer_get_deserialize_formats"), visibility("default")));

#undef gtk_text_buffer_get_serialize_formats
extern __typeof (gtk_text_buffer_get_serialize_formats) gtk_text_buffer_get_serialize_formats __attribute((alias("IA__gtk_text_buffer_get_serialize_formats"), visibility("default")));

#undef gtk_text_buffer_register_deserialize_format
extern __typeof (gtk_text_buffer_register_deserialize_format) gtk_text_buffer_register_deserialize_format __attribute((alias("IA__gtk_text_buffer_register_deserialize_format"), visibility("default")));

#undef gtk_text_buffer_register_deserialize_tagset
extern __typeof (gtk_text_buffer_register_deserialize_tagset) gtk_text_buffer_register_deserialize_tagset __attribute((alias("IA__gtk_text_buffer_register_deserialize_tagset"), visibility("default")));

#undef gtk_text_buffer_register_serialize_format
extern __typeof (gtk_text_buffer_register_serialize_format) gtk_text_buffer_register_serialize_format __attribute((alias("IA__gtk_text_buffer_register_serialize_format"), visibility("default")));

#undef gtk_text_buffer_register_serialize_tagset
extern __typeof (gtk_text_buffer_register_serialize_tagset) gtk_text_buffer_register_serialize_tagset __attribute((alias("IA__gtk_text_buffer_register_serialize_tagset"), visibility("default")));

#undef gtk_text_buffer_serialize
extern __typeof (gtk_text_buffer_serialize) gtk_text_buffer_serialize __attribute((alias("IA__gtk_text_buffer_serialize"), visibility("default")));

#undef gtk_text_buffer_unregister_deserialize_format
extern __typeof (gtk_text_buffer_unregister_deserialize_format) gtk_text_buffer_unregister_deserialize_format __attribute((alias("IA__gtk_text_buffer_unregister_deserialize_format"), visibility("default")));

#undef gtk_text_buffer_unregister_serialize_format
extern __typeof (gtk_text_buffer_unregister_serialize_format) gtk_text_buffer_unregister_serialize_format __attribute((alias("IA__gtk_text_buffer_unregister_serialize_format"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_RULER_H__)
#if IN_FILE(__GTK_RULER_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_ruler_draw_pos
extern __typeof (gtk_ruler_draw_pos) gtk_ruler_draw_pos __attribute((alias("IA__gtk_ruler_draw_pos"), visibility("default")));

#undef gtk_ruler_draw_ticks
extern __typeof (gtk_ruler_draw_ticks) gtk_ruler_draw_ticks __attribute((alias("IA__gtk_ruler_draw_ticks"), visibility("default")));

#undef gtk_ruler_get_metric
extern __typeof (gtk_ruler_get_metric) gtk_ruler_get_metric __attribute((alias("IA__gtk_ruler_get_metric"), visibility("default")));

#undef gtk_ruler_get_range
extern __typeof (gtk_ruler_get_range) gtk_ruler_get_range __attribute((alias("IA__gtk_ruler_get_range"), visibility("default")));

#undef gtk_ruler_get_type
extern __typeof (gtk_ruler_get_type) gtk_ruler_get_type __attribute((alias("IA__gtk_ruler_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_ruler_set_metric
extern __typeof (gtk_ruler_set_metric) gtk_ruler_set_metric __attribute((alias("IA__gtk_ruler_set_metric"), visibility("default")));

#undef gtk_ruler_set_range
extern __typeof (gtk_ruler_set_range) gtk_ruler_set_range __attribute((alias("IA__gtk_ruler_set_range"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_SCALE_H__)
#if IN_FILE(__GTK_SCALE_C__)
#undef gtk_scale_get_digits
extern __typeof (gtk_scale_get_digits) gtk_scale_get_digits __attribute((alias("IA__gtk_scale_get_digits"), visibility("default")));

#undef gtk_scale_get_draw_value
extern __typeof (gtk_scale_get_draw_value) gtk_scale_get_draw_value __attribute((alias("IA__gtk_scale_get_draw_value"), visibility("default")));

#undef gtk_scale_get_layout
extern __typeof (gtk_scale_get_layout) gtk_scale_get_layout __attribute((alias("IA__gtk_scale_get_layout"), visibility("default")));

#undef gtk_scale_get_layout_offsets
extern __typeof (gtk_scale_get_layout_offsets) gtk_scale_get_layout_offsets __attribute((alias("IA__gtk_scale_get_layout_offsets"), visibility("default")));

#undef gtk_scale_get_type
extern __typeof (gtk_scale_get_type) gtk_scale_get_type __attribute((alias("IA__gtk_scale_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_scale_get_value_pos
extern __typeof (gtk_scale_get_value_pos) gtk_scale_get_value_pos __attribute((alias("IA__gtk_scale_get_value_pos"), visibility("default")));

#undef gtk_scale_set_digits
extern __typeof (gtk_scale_set_digits) gtk_scale_set_digits __attribute((alias("IA__gtk_scale_set_digits"), visibility("default")));

#undef gtk_scale_set_draw_value
extern __typeof (gtk_scale_set_draw_value) gtk_scale_set_draw_value __attribute((alias("IA__gtk_scale_set_draw_value"), visibility("default")));

#undef gtk_scale_set_value_pos
extern __typeof (gtk_scale_set_value_pos) gtk_scale_set_value_pos __attribute((alias("IA__gtk_scale_set_value_pos"), visibility("default")));

#undef gtk_scale_add_mark
extern __typeof (gtk_scale_add_mark) gtk_scale_add_mark __attribute((alias("IA__gtk_scale_add_mark"), visibility("default")));

#undef gtk_scale_clear_marks
extern __typeof (gtk_scale_clear_marks) gtk_scale_clear_marks __attribute((alias("IA__gtk_scale_clear_marks"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_SCALE_BUTTON_H__)
#if IN_FILE(__GTK_SCALE_BUTTON_C__)
#undef gtk_scale_button_get_type
extern __typeof (gtk_scale_button_get_type) gtk_scale_button_get_type __attribute((alias("IA__gtk_scale_button_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_scale_button_get_value
extern __typeof (gtk_scale_button_get_value) gtk_scale_button_get_value __attribute((alias("IA__gtk_scale_button_get_value"), visibility("default")));

#undef gtk_scale_button_new
extern __typeof (gtk_scale_button_new) gtk_scale_button_new __attribute((alias("IA__gtk_scale_button_new"), visibility("default")));

#undef gtk_scale_button_set_icons
extern __typeof (gtk_scale_button_set_icons) gtk_scale_button_set_icons __attribute((alias("IA__gtk_scale_button_set_icons"), visibility("default")));

#undef gtk_scale_button_set_value
extern __typeof (gtk_scale_button_set_value) gtk_scale_button_set_value __attribute((alias("IA__gtk_scale_button_set_value"), visibility("default")));

#undef gtk_scale_button_get_adjustment
extern __typeof (gtk_scale_button_get_adjustment) gtk_scale_button_get_adjustment __attribute((alias("IA__gtk_scale_button_get_adjustment"), visibility("default")));

#undef gtk_scale_button_set_adjustment
extern __typeof (gtk_scale_button_set_adjustment) gtk_scale_button_set_adjustment __attribute((alias("IA__gtk_scale_button_set_adjustment"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_scale_button_get_orientation
extern __typeof (gtk_scale_button_get_orientation) gtk_scale_button_get_orientation __attribute((alias("IA__gtk_scale_button_get_orientation"), visibility("default")));

#undef gtk_scale_button_set_orientation
extern __typeof (gtk_scale_button_set_orientation) gtk_scale_button_set_orientation __attribute((alias("IA__gtk_scale_button_set_orientation"), visibility("default")));

#endif
#undef gtk_scale_button_get_plus_button
extern __typeof (gtk_scale_button_get_plus_button) gtk_scale_button_get_plus_button __attribute((alias("IA__gtk_scale_button_get_plus_button"), visibility("default")));

#undef gtk_scale_button_get_minus_button
extern __typeof (gtk_scale_button_get_minus_button) gtk_scale_button_get_minus_button __attribute((alias("IA__gtk_scale_button_get_minus_button"), visibility("default")));

#undef gtk_scale_button_get_popup
extern __typeof (gtk_scale_button_get_popup) gtk_scale_button_get_popup __attribute((alias("IA__gtk_scale_button_get_popup"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_SCROLLBAR_H__)
#if IN_FILE(__GTK_SCROLLBAR_C__)
#undef gtk_scrollbar_get_type
extern __typeof (gtk_scrollbar_get_type) gtk_scrollbar_get_type __attribute((alias("IA__gtk_scrollbar_get_type"), visibility("default"))) G_GNUC_CONST;

#endif
#endif
#if IN_HEADER(__GTK_SCROLLED_WINDOW_H__)
#if IN_FILE(__GTK_SCROLLED_WINDOW_C__)
#undef gtk_scrolled_window_add_with_viewport
extern __typeof (gtk_scrolled_window_add_with_viewport) gtk_scrolled_window_add_with_viewport __attribute((alias("IA__gtk_scrolled_window_add_with_viewport"), visibility("default")));

#undef gtk_scrolled_window_get_hadjustment
extern __typeof (gtk_scrolled_window_get_hadjustment) gtk_scrolled_window_get_hadjustment __attribute((alias("IA__gtk_scrolled_window_get_hadjustment"), visibility("default")));

#undef gtk_scrolled_window_get_hscrollbar
extern __typeof (gtk_scrolled_window_get_hscrollbar) gtk_scrolled_window_get_hscrollbar __attribute((alias("IA__gtk_scrolled_window_get_hscrollbar"), visibility("default")));

#undef gtk_scrolled_window_get_placement
extern __typeof (gtk_scrolled_window_get_placement) gtk_scrolled_window_get_placement __attribute((alias("IA__gtk_scrolled_window_get_placement"), visibility("default")));

#undef gtk_scrolled_window_get_policy
extern __typeof (gtk_scrolled_window_get_policy) gtk_scrolled_window_get_policy __attribute((alias("IA__gtk_scrolled_window_get_policy"), visibility("default")));

#undef gtk_scrolled_window_get_shadow_type
extern __typeof (gtk_scrolled_window_get_shadow_type) gtk_scrolled_window_get_shadow_type __attribute((alias("IA__gtk_scrolled_window_get_shadow_type"), visibility("default")));

#undef gtk_scrolled_window_get_type
extern __typeof (gtk_scrolled_window_get_type) gtk_scrolled_window_get_type __attribute((alias("IA__gtk_scrolled_window_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_scrolled_window_get_vadjustment
extern __typeof (gtk_scrolled_window_get_vadjustment) gtk_scrolled_window_get_vadjustment __attribute((alias("IA__gtk_scrolled_window_get_vadjustment"), visibility("default")));

#undef gtk_scrolled_window_get_vscrollbar
extern __typeof (gtk_scrolled_window_get_vscrollbar) gtk_scrolled_window_get_vscrollbar __attribute((alias("IA__gtk_scrolled_window_get_vscrollbar"), visibility("default")));

#undef gtk_scrolled_window_new
extern __typeof (gtk_scrolled_window_new) gtk_scrolled_window_new __attribute((alias("IA__gtk_scrolled_window_new"), visibility("default")));

#undef gtk_scrolled_window_set_hadjustment
extern __typeof (gtk_scrolled_window_set_hadjustment) gtk_scrolled_window_set_hadjustment __attribute((alias("IA__gtk_scrolled_window_set_hadjustment"), visibility("default")));

#undef gtk_scrolled_window_set_placement
extern __typeof (gtk_scrolled_window_set_placement) gtk_scrolled_window_set_placement __attribute((alias("IA__gtk_scrolled_window_set_placement"), visibility("default")));

#undef gtk_scrolled_window_set_policy
extern __typeof (gtk_scrolled_window_set_policy) gtk_scrolled_window_set_policy __attribute((alias("IA__gtk_scrolled_window_set_policy"), visibility("default")));

#undef gtk_scrolled_window_set_shadow_type
extern __typeof (gtk_scrolled_window_set_shadow_type) gtk_scrolled_window_set_shadow_type __attribute((alias("IA__gtk_scrolled_window_set_shadow_type"), visibility("default")));

#undef gtk_scrolled_window_set_vadjustment
extern __typeof (gtk_scrolled_window_set_vadjustment) gtk_scrolled_window_set_vadjustment __attribute((alias("IA__gtk_scrolled_window_set_vadjustment"), visibility("default")));

#undef gtk_scrolled_window_unset_placement
extern __typeof (gtk_scrolled_window_unset_placement) gtk_scrolled_window_unset_placement __attribute((alias("IA__gtk_scrolled_window_unset_placement"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_SELECTION_H__)
#if IN_FILE(__GTK_SELECTION_C__)
#undef gtk_selection_add_target
extern __typeof (gtk_selection_add_target) gtk_selection_add_target __attribute((alias("IA__gtk_selection_add_target"), visibility("default")));

#undef gtk_selection_add_targets
extern __typeof (gtk_selection_add_targets) gtk_selection_add_targets __attribute((alias("IA__gtk_selection_add_targets"), visibility("default")));

#undef gtk_selection_clear
extern __typeof (gtk_selection_clear) gtk_selection_clear __attribute((alias("IA__gtk_selection_clear"), visibility("default")));

#undef gtk_selection_clear_targets
extern __typeof (gtk_selection_clear_targets) gtk_selection_clear_targets __attribute((alias("IA__gtk_selection_clear_targets"), visibility("default")));

#undef gtk_selection_convert
extern __typeof (gtk_selection_convert) gtk_selection_convert __attribute((alias("IA__gtk_selection_convert"), visibility("default")));

#undef gtk_selection_data_copy
extern __typeof (gtk_selection_data_copy) gtk_selection_data_copy __attribute((alias("IA__gtk_selection_data_copy"), visibility("default")));

#undef gtk_selection_data_free
extern __typeof (gtk_selection_data_free) gtk_selection_data_free __attribute((alias("IA__gtk_selection_data_free"), visibility("default")));

#undef gtk_selection_data_get_selection
extern __typeof (gtk_selection_data_get_selection) gtk_selection_data_get_selection __attribute((alias("IA__gtk_selection_data_get_selection"), visibility("default")));

#undef gtk_selection_data_get_target
extern __typeof (gtk_selection_data_get_target) gtk_selection_data_get_target __attribute((alias("IA__gtk_selection_data_get_target"), visibility("default")));

#undef gtk_selection_data_get_data_type
extern __typeof (gtk_selection_data_get_data_type) gtk_selection_data_get_data_type __attribute((alias("IA__gtk_selection_data_get_data_type"), visibility("default")));

#undef gtk_selection_data_get_display
extern __typeof (gtk_selection_data_get_display) gtk_selection_data_get_display __attribute((alias("IA__gtk_selection_data_get_display"), visibility("default")));

#undef gtk_selection_data_get_format
extern __typeof (gtk_selection_data_get_format) gtk_selection_data_get_format __attribute((alias("IA__gtk_selection_data_get_format"), visibility("default")));

#undef gtk_selection_data_get_data
extern __typeof (gtk_selection_data_get_data) gtk_selection_data_get_data __attribute((alias("IA__gtk_selection_data_get_data"), visibility("default")));

#undef gtk_selection_data_get_length
extern __typeof (gtk_selection_data_get_length) gtk_selection_data_get_length __attribute((alias("IA__gtk_selection_data_get_length"), visibility("default")));

#undef gtk_selection_data_get_pixbuf
extern __typeof (gtk_selection_data_get_pixbuf) gtk_selection_data_get_pixbuf __attribute((alias("IA__gtk_selection_data_get_pixbuf"), visibility("default")));

#undef gtk_selection_data_get_targets
extern __typeof (gtk_selection_data_get_targets) gtk_selection_data_get_targets __attribute((alias("IA__gtk_selection_data_get_targets"), visibility("default")));

#undef gtk_selection_data_get_text
extern __typeof (gtk_selection_data_get_text) gtk_selection_data_get_text __attribute((alias("IA__gtk_selection_data_get_text"), visibility("default")));

#undef gtk_selection_data_get_type
extern __typeof (gtk_selection_data_get_type) gtk_selection_data_get_type __attribute((alias("IA__gtk_selection_data_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_selection_data_get_uris
extern __typeof (gtk_selection_data_get_uris) gtk_selection_data_get_uris __attribute((alias("IA__gtk_selection_data_get_uris"), visibility("default")));

#undef gtk_selection_data_set
extern __typeof (gtk_selection_data_set) gtk_selection_data_set __attribute((alias("IA__gtk_selection_data_set"), visibility("default")));

#undef gtk_selection_data_set_pixbuf
extern __typeof (gtk_selection_data_set_pixbuf) gtk_selection_data_set_pixbuf __attribute((alias("IA__gtk_selection_data_set_pixbuf"), visibility("default")));

#undef gtk_selection_data_set_text
extern __typeof (gtk_selection_data_set_text) gtk_selection_data_set_text __attribute((alias("IA__gtk_selection_data_set_text"), visibility("default")));

#undef gtk_selection_data_set_uris
extern __typeof (gtk_selection_data_set_uris) gtk_selection_data_set_uris __attribute((alias("IA__gtk_selection_data_set_uris"), visibility("default")));

#undef gtk_selection_data_targets_include_image
extern __typeof (gtk_selection_data_targets_include_image) gtk_selection_data_targets_include_image __attribute((alias("IA__gtk_selection_data_targets_include_image"), visibility("default")));

#undef gtk_selection_data_targets_include_rich_text
extern __typeof (gtk_selection_data_targets_include_rich_text) gtk_selection_data_targets_include_rich_text __attribute((alias("IA__gtk_selection_data_targets_include_rich_text"), visibility("default")));

#undef gtk_selection_data_targets_include_text
extern __typeof (gtk_selection_data_targets_include_text) gtk_selection_data_targets_include_text __attribute((alias("IA__gtk_selection_data_targets_include_text"), visibility("default")));

#undef gtk_selection_data_targets_include_uri
extern __typeof (gtk_selection_data_targets_include_uri) gtk_selection_data_targets_include_uri __attribute((alias("IA__gtk_selection_data_targets_include_uri"), visibility("default")));

#undef gtk_targets_include_image
extern __typeof (gtk_targets_include_image) gtk_targets_include_image __attribute((alias("IA__gtk_targets_include_image"), visibility("default")));

#undef gtk_targets_include_rich_text
extern __typeof (gtk_targets_include_rich_text) gtk_targets_include_rich_text __attribute((alias("IA__gtk_targets_include_rich_text"), visibility("default")));

#undef gtk_targets_include_text
extern __typeof (gtk_targets_include_text) gtk_targets_include_text __attribute((alias("IA__gtk_targets_include_text"), visibility("default")));

#undef gtk_targets_include_uri
extern __typeof (gtk_targets_include_uri) gtk_targets_include_uri __attribute((alias("IA__gtk_targets_include_uri"), visibility("default")));

#undef gtk_selection_owner_set
extern __typeof (gtk_selection_owner_set) gtk_selection_owner_set __attribute((alias("IA__gtk_selection_owner_set"), visibility("default")));

#undef gtk_selection_owner_set_for_display
extern __typeof (gtk_selection_owner_set_for_display) gtk_selection_owner_set_for_display __attribute((alias("IA__gtk_selection_owner_set_for_display"), visibility("default")));

#undef gtk_selection_remove_all
extern __typeof (gtk_selection_remove_all) gtk_selection_remove_all __attribute((alias("IA__gtk_selection_remove_all"), visibility("default")));

#undef gtk_target_list_add
extern __typeof (gtk_target_list_add) gtk_target_list_add __attribute((alias("IA__gtk_target_list_add"), visibility("default")));

#undef gtk_target_list_add_image_targets
extern __typeof (gtk_target_list_add_image_targets) gtk_target_list_add_image_targets __attribute((alias("IA__gtk_target_list_add_image_targets"), visibility("default")));

#undef gtk_target_list_add_rich_text_targets
extern __typeof (gtk_target_list_add_rich_text_targets) gtk_target_list_add_rich_text_targets __attribute((alias("IA__gtk_target_list_add_rich_text_targets"), visibility("default")));

#undef gtk_target_list_add_table
extern __typeof (gtk_target_list_add_table) gtk_target_list_add_table __attribute((alias("IA__gtk_target_list_add_table"), visibility("default")));

#undef gtk_target_list_add_text_targets
extern __typeof (gtk_target_list_add_text_targets) gtk_target_list_add_text_targets __attribute((alias("IA__gtk_target_list_add_text_targets"), visibility("default")));

#undef gtk_target_list_add_uri_targets
extern __typeof (gtk_target_list_add_uri_targets) gtk_target_list_add_uri_targets __attribute((alias("IA__gtk_target_list_add_uri_targets"), visibility("default")));

#undef gtk_target_list_find
extern __typeof (gtk_target_list_find) gtk_target_list_find __attribute((alias("IA__gtk_target_list_find"), visibility("default")));

#undef gtk_target_list_get_type
extern __typeof (gtk_target_list_get_type) gtk_target_list_get_type __attribute((alias("IA__gtk_target_list_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_target_list_new
extern __typeof (gtk_target_list_new) gtk_target_list_new __attribute((alias("IA__gtk_target_list_new"), visibility("default")));

#undef gtk_target_list_ref
extern __typeof (gtk_target_list_ref) gtk_target_list_ref __attribute((alias("IA__gtk_target_list_ref"), visibility("default")));

#undef gtk_target_list_remove
extern __typeof (gtk_target_list_remove) gtk_target_list_remove __attribute((alias("IA__gtk_target_list_remove"), visibility("default")));

#undef gtk_target_list_unref
extern __typeof (gtk_target_list_unref) gtk_target_list_unref __attribute((alias("IA__gtk_target_list_unref"), visibility("default")));

#undef gtk_target_table_new_from_list
extern __typeof (gtk_target_table_new_from_list) gtk_target_table_new_from_list __attribute((alias("IA__gtk_target_table_new_from_list"), visibility("default")));

#undef gtk_target_table_free
extern __typeof (gtk_target_table_free) gtk_target_table_free __attribute((alias("IA__gtk_target_table_free"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_SEPARATOR_H__)
#if IN_FILE(__GTK_SEPARATOR_C__)
#undef gtk_separator_get_type
extern __typeof (gtk_separator_get_type) gtk_separator_get_type __attribute((alias("IA__gtk_separator_get_type"), visibility("default"))) G_GNUC_CONST;

#endif
#endif
#if IN_HEADER(__GTK_SEPARATOR_MENU_ITEM_H__)
#if IN_FILE(__GTK_SEPARATOR_MENU_ITEM_C__)
#undef gtk_separator_menu_item_get_type
extern __typeof (gtk_separator_menu_item_get_type) gtk_separator_menu_item_get_type __attribute((alias("IA__gtk_separator_menu_item_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_separator_menu_item_new
extern __typeof (gtk_separator_menu_item_new) gtk_separator_menu_item_new __attribute((alias("IA__gtk_separator_menu_item_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_SEPARATOR_TOOL_ITEM_H__)
#if IN_FILE(__GTK_SEPARATOR_TOOL_ITEM_C__)
#undef gtk_separator_tool_item_get_draw
extern __typeof (gtk_separator_tool_item_get_draw) gtk_separator_tool_item_get_draw __attribute((alias("IA__gtk_separator_tool_item_get_draw"), visibility("default")));

#undef gtk_separator_tool_item_get_type
extern __typeof (gtk_separator_tool_item_get_type) gtk_separator_tool_item_get_type __attribute((alias("IA__gtk_separator_tool_item_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_separator_tool_item_new
extern __typeof (gtk_separator_tool_item_new) gtk_separator_tool_item_new __attribute((alias("IA__gtk_separator_tool_item_new"), visibility("default")));

#undef gtk_separator_tool_item_set_draw
extern __typeof (gtk_separator_tool_item_set_draw) gtk_separator_tool_item_set_draw __attribute((alias("IA__gtk_separator_tool_item_set_draw"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_SETTINGS_H__)
#if IN_FILE(__GTK_SETTINGS_C__)
#undef gtk_settings_get_default
extern __typeof (gtk_settings_get_default) gtk_settings_get_default __attribute((alias("IA__gtk_settings_get_default"), visibility("default")));

#undef gtk_settings_get_for_screen
extern __typeof (gtk_settings_get_for_screen) gtk_settings_get_for_screen __attribute((alias("IA__gtk_settings_get_for_screen"), visibility("default")));

#undef gtk_settings_get_type
extern __typeof (gtk_settings_get_type) gtk_settings_get_type __attribute((alias("IA__gtk_settings_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_settings_install_property
extern __typeof (gtk_settings_install_property) gtk_settings_install_property __attribute((alias("IA__gtk_settings_install_property"), visibility("default")));

#undef gtk_settings_install_property_parser
extern __typeof (gtk_settings_install_property_parser) gtk_settings_install_property_parser __attribute((alias("IA__gtk_settings_install_property_parser"), visibility("default")));

#undef gtk_settings_set_double_property
extern __typeof (gtk_settings_set_double_property) gtk_settings_set_double_property __attribute((alias("IA__gtk_settings_set_double_property"), visibility("default")));

#undef gtk_settings_set_long_property
extern __typeof (gtk_settings_set_long_property) gtk_settings_set_long_property __attribute((alias("IA__gtk_settings_set_long_property"), visibility("default")));

#undef gtk_settings_set_property_value
extern __typeof (gtk_settings_set_property_value) gtk_settings_set_property_value __attribute((alias("IA__gtk_settings_set_property_value"), visibility("default")));

#undef gtk_settings_set_string_property
extern __typeof (gtk_settings_set_string_property) gtk_settings_set_string_property __attribute((alias("IA__gtk_settings_set_string_property"), visibility("default")));

#undef gtk_rc_property_parse_border
extern __typeof (gtk_rc_property_parse_border) gtk_rc_property_parse_border __attribute((alias("IA__gtk_rc_property_parse_border"), visibility("default")));

#undef gtk_rc_property_parse_color
extern __typeof (gtk_rc_property_parse_color) gtk_rc_property_parse_color __attribute((alias("IA__gtk_rc_property_parse_color"), visibility("default")));

#undef gtk_rc_property_parse_enum
extern __typeof (gtk_rc_property_parse_enum) gtk_rc_property_parse_enum __attribute((alias("IA__gtk_rc_property_parse_enum"), visibility("default")));

#undef gtk_rc_property_parse_flags
extern __typeof (gtk_rc_property_parse_flags) gtk_rc_property_parse_flags __attribute((alias("IA__gtk_rc_property_parse_flags"), visibility("default")));

#undef gtk_rc_property_parse_requisition
extern __typeof (gtk_rc_property_parse_requisition) gtk_rc_property_parse_requisition __attribute((alias("IA__gtk_rc_property_parse_requisition"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_SIGNAL_H__)
#if IN_FILE(__GTK_SIGNAL_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_signal_compat_matched
extern __typeof (gtk_signal_compat_matched) gtk_signal_compat_matched __attribute((alias("IA__gtk_signal_compat_matched"), visibility("default")));

#undef gtk_signal_connect_full
extern __typeof (gtk_signal_connect_full) gtk_signal_connect_full __attribute((alias("IA__gtk_signal_connect_full"), visibility("default")));

#undef gtk_signal_connect_object_while_alive
extern __typeof (gtk_signal_connect_object_while_alive) gtk_signal_connect_object_while_alive __attribute((alias("IA__gtk_signal_connect_object_while_alive"), visibility("default")));

#undef gtk_signal_connect_while_alive
extern __typeof (gtk_signal_connect_while_alive) gtk_signal_connect_while_alive __attribute((alias("IA__gtk_signal_connect_while_alive"), visibility("default")));

#undef gtk_signal_emit
extern __typeof (gtk_signal_emit) gtk_signal_emit __attribute((alias("IA__gtk_signal_emit"), visibility("default")));

#undef gtk_signal_emit_by_name
extern __typeof (gtk_signal_emit_by_name) gtk_signal_emit_by_name __attribute((alias("IA__gtk_signal_emit_by_name"), visibility("default")));

#undef gtk_signal_emit_stop_by_name
extern __typeof (gtk_signal_emit_stop_by_name) gtk_signal_emit_stop_by_name __attribute((alias("IA__gtk_signal_emit_stop_by_name"), visibility("default")));

#undef gtk_signal_emitv
extern __typeof (gtk_signal_emitv) gtk_signal_emitv __attribute((alias("IA__gtk_signal_emitv"), visibility("default")));

#undef gtk_signal_emitv_by_name
extern __typeof (gtk_signal_emitv_by_name) gtk_signal_emitv_by_name __attribute((alias("IA__gtk_signal_emitv_by_name"), visibility("default")));

#undef gtk_signal_new
extern __typeof (gtk_signal_new) gtk_signal_new __attribute((alias("IA__gtk_signal_new"), visibility("default")));

#undef gtk_signal_newv
extern __typeof (gtk_signal_newv) gtk_signal_newv __attribute((alias("IA__gtk_signal_newv"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_SIZE_GROUP_H__)
#if IN_FILE(__GTK_SIZE_GROUP_C__)
#undef gtk_size_group_add_widget
extern __typeof (gtk_size_group_add_widget) gtk_size_group_add_widget __attribute((alias("IA__gtk_size_group_add_widget"), visibility("default")));

#undef gtk_size_group_get_ignore_hidden
extern __typeof (gtk_size_group_get_ignore_hidden) gtk_size_group_get_ignore_hidden __attribute((alias("IA__gtk_size_group_get_ignore_hidden"), visibility("default")));

#undef gtk_size_group_get_mode
extern __typeof (gtk_size_group_get_mode) gtk_size_group_get_mode __attribute((alias("IA__gtk_size_group_get_mode"), visibility("default")));

#undef gtk_size_group_get_type
extern __typeof (gtk_size_group_get_type) gtk_size_group_get_type __attribute((alias("IA__gtk_size_group_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_size_group_get_widgets
extern __typeof (gtk_size_group_get_widgets) gtk_size_group_get_widgets __attribute((alias("IA__gtk_size_group_get_widgets"), visibility("default")));

#undef gtk_size_group_new
extern __typeof (gtk_size_group_new) gtk_size_group_new __attribute((alias("IA__gtk_size_group_new"), visibility("default")));

#undef gtk_size_group_remove_widget
extern __typeof (gtk_size_group_remove_widget) gtk_size_group_remove_widget __attribute((alias("IA__gtk_size_group_remove_widget"), visibility("default")));

#undef gtk_size_group_set_ignore_hidden
extern __typeof (gtk_size_group_set_ignore_hidden) gtk_size_group_set_ignore_hidden __attribute((alias("IA__gtk_size_group_set_ignore_hidden"), visibility("default")));

#undef gtk_size_group_set_mode
extern __typeof (gtk_size_group_set_mode) gtk_size_group_set_mode __attribute((alias("IA__gtk_size_group_set_mode"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_SHOW_H__)
#if IN_FILE(__GTK_SHOW_C__)
#undef gtk_show_uri
extern __typeof (gtk_show_uri) gtk_show_uri __attribute((alias("IA__gtk_show_uri"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_SOCKET_H__)
#if IN_FILE(__GTK_SOCKET_C__)
#undef gtk_socket_add_id
extern __typeof (gtk_socket_add_id) gtk_socket_add_id __attribute((alias("IA__gtk_socket_add_id"), visibility("default")));

#undef gtk_socket_get_id
extern __typeof (gtk_socket_get_id) gtk_socket_get_id __attribute((alias("IA__gtk_socket_get_id"), visibility("default")));

#undef gtk_socket_get_plug_window
extern __typeof (gtk_socket_get_plug_window) gtk_socket_get_plug_window __attribute((alias("IA__gtk_socket_get_plug_window"), visibility("default")));

#undef gtk_socket_get_type
extern __typeof (gtk_socket_get_type) gtk_socket_get_type __attribute((alias("IA__gtk_socket_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_socket_new
extern __typeof (gtk_socket_new) gtk_socket_new __attribute((alias("IA__gtk_socket_new"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_socket_steal
extern __typeof (gtk_socket_steal) gtk_socket_steal __attribute((alias("IA__gtk_socket_steal"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_SPIN_BUTTON_H__)
#if IN_FILE(__GTK_SPIN_BUTTON_C__)
#undef gtk_spin_button_configure
extern __typeof (gtk_spin_button_configure) gtk_spin_button_configure __attribute((alias("IA__gtk_spin_button_configure"), visibility("default")));

#undef gtk_spin_button_get_adjustment
extern __typeof (gtk_spin_button_get_adjustment) gtk_spin_button_get_adjustment __attribute((alias("IA__gtk_spin_button_get_adjustment"), visibility("default")));

#undef gtk_spin_button_get_digits
extern __typeof (gtk_spin_button_get_digits) gtk_spin_button_get_digits __attribute((alias("IA__gtk_spin_button_get_digits"), visibility("default")));

#undef gtk_spin_button_get_increments
extern __typeof (gtk_spin_button_get_increments) gtk_spin_button_get_increments __attribute((alias("IA__gtk_spin_button_get_increments"), visibility("default")));

#undef gtk_spin_button_get_numeric
extern __typeof (gtk_spin_button_get_numeric) gtk_spin_button_get_numeric __attribute((alias("IA__gtk_spin_button_get_numeric"), visibility("default")));

#undef gtk_spin_button_get_range
extern __typeof (gtk_spin_button_get_range) gtk_spin_button_get_range __attribute((alias("IA__gtk_spin_button_get_range"), visibility("default")));

#undef gtk_spin_button_get_snap_to_ticks
extern __typeof (gtk_spin_button_get_snap_to_ticks) gtk_spin_button_get_snap_to_ticks __attribute((alias("IA__gtk_spin_button_get_snap_to_ticks"), visibility("default")));

#undef gtk_spin_button_get_type
extern __typeof (gtk_spin_button_get_type) gtk_spin_button_get_type __attribute((alias("IA__gtk_spin_button_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_spin_button_get_update_policy
extern __typeof (gtk_spin_button_get_update_policy) gtk_spin_button_get_update_policy __attribute((alias("IA__gtk_spin_button_get_update_policy"), visibility("default")));

#undef gtk_spin_button_get_value
extern __typeof (gtk_spin_button_get_value) gtk_spin_button_get_value __attribute((alias("IA__gtk_spin_button_get_value"), visibility("default")));

#undef gtk_spin_button_get_value_as_int
extern __typeof (gtk_spin_button_get_value_as_int) gtk_spin_button_get_value_as_int __attribute((alias("IA__gtk_spin_button_get_value_as_int"), visibility("default")));

#undef gtk_spin_button_get_wrap
extern __typeof (gtk_spin_button_get_wrap) gtk_spin_button_get_wrap __attribute((alias("IA__gtk_spin_button_get_wrap"), visibility("default")));

#undef gtk_spin_button_new
extern __typeof (gtk_spin_button_new) gtk_spin_button_new __attribute((alias("IA__gtk_spin_button_new"), visibility("default")));

#undef gtk_spin_button_new_with_range
extern __typeof (gtk_spin_button_new_with_range) gtk_spin_button_new_with_range __attribute((alias("IA__gtk_spin_button_new_with_range"), visibility("default")));

#undef gtk_spin_button_set_adjustment
extern __typeof (gtk_spin_button_set_adjustment) gtk_spin_button_set_adjustment __attribute((alias("IA__gtk_spin_button_set_adjustment"), visibility("default")));

#undef gtk_spin_button_set_digits
extern __typeof (gtk_spin_button_set_digits) gtk_spin_button_set_digits __attribute((alias("IA__gtk_spin_button_set_digits"), visibility("default")));

#undef gtk_spin_button_set_increments
extern __typeof (gtk_spin_button_set_increments) gtk_spin_button_set_increments __attribute((alias("IA__gtk_spin_button_set_increments"), visibility("default")));

#undef gtk_spin_button_set_numeric
extern __typeof (gtk_spin_button_set_numeric) gtk_spin_button_set_numeric __attribute((alias("IA__gtk_spin_button_set_numeric"), visibility("default")));

#undef gtk_spin_button_set_range
extern __typeof (gtk_spin_button_set_range) gtk_spin_button_set_range __attribute((alias("IA__gtk_spin_button_set_range"), visibility("default")));

#undef gtk_spin_button_set_snap_to_ticks
extern __typeof (gtk_spin_button_set_snap_to_ticks) gtk_spin_button_set_snap_to_ticks __attribute((alias("IA__gtk_spin_button_set_snap_to_ticks"), visibility("default")));

#undef gtk_spin_button_set_update_policy
extern __typeof (gtk_spin_button_set_update_policy) gtk_spin_button_set_update_policy __attribute((alias("IA__gtk_spin_button_set_update_policy"), visibility("default")));

#undef gtk_spin_button_set_value
extern __typeof (gtk_spin_button_set_value) gtk_spin_button_set_value __attribute((alias("IA__gtk_spin_button_set_value"), visibility("default")));

#undef gtk_spin_button_set_wrap
extern __typeof (gtk_spin_button_set_wrap) gtk_spin_button_set_wrap __attribute((alias("IA__gtk_spin_button_set_wrap"), visibility("default")));

#undef gtk_spin_button_spin
extern __typeof (gtk_spin_button_spin) gtk_spin_button_spin __attribute((alias("IA__gtk_spin_button_spin"), visibility("default")));

#undef gtk_spin_button_update
extern __typeof (gtk_spin_button_update) gtk_spin_button_update __attribute((alias("IA__gtk_spin_button_update"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_SPINNER_H__)
#if IN_FILE(__GTK_SPINNER_C__)
#undef gtk_spinner_get_type
extern __typeof (gtk_spinner_get_type) gtk_spinner_get_type __attribute((alias("IA__gtk_spinner_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_spinner_new
extern __typeof (gtk_spinner_new) gtk_spinner_new __attribute((alias("IA__gtk_spinner_new"), visibility("default")));

#undef gtk_spinner_start
extern __typeof (gtk_spinner_start) gtk_spinner_start __attribute((alias("IA__gtk_spinner_start"), visibility("default")));

#undef gtk_spinner_stop
extern __typeof (gtk_spinner_stop) gtk_spinner_stop __attribute((alias("IA__gtk_spinner_stop"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_STATUSBAR_H__)
#if IN_FILE(__GTK_STATUSBAR_C__)
#undef gtk_statusbar_get_context_id
extern __typeof (gtk_statusbar_get_context_id) gtk_statusbar_get_context_id __attribute((alias("IA__gtk_statusbar_get_context_id"), visibility("default")));

#undef gtk_statusbar_get_has_resize_grip
extern __typeof (gtk_statusbar_get_has_resize_grip) gtk_statusbar_get_has_resize_grip __attribute((alias("IA__gtk_statusbar_get_has_resize_grip"), visibility("default")));

#undef gtk_statusbar_get_message_area
extern __typeof (gtk_statusbar_get_message_area) gtk_statusbar_get_message_area __attribute((alias("IA__gtk_statusbar_get_message_area"), visibility("default")));

#undef gtk_statusbar_get_type
extern __typeof (gtk_statusbar_get_type) gtk_statusbar_get_type __attribute((alias("IA__gtk_statusbar_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_statusbar_new
extern __typeof (gtk_statusbar_new) gtk_statusbar_new __attribute((alias("IA__gtk_statusbar_new"), visibility("default")));

#undef gtk_statusbar_pop
extern __typeof (gtk_statusbar_pop) gtk_statusbar_pop __attribute((alias("IA__gtk_statusbar_pop"), visibility("default")));

#undef gtk_statusbar_push
extern __typeof (gtk_statusbar_push) gtk_statusbar_push __attribute((alias("IA__gtk_statusbar_push"), visibility("default")));

#undef gtk_statusbar_remove
extern __typeof (gtk_statusbar_remove) gtk_statusbar_remove __attribute((alias("IA__gtk_statusbar_remove"), visibility("default")));

#undef gtk_statusbar_remove_all
extern __typeof (gtk_statusbar_remove_all) gtk_statusbar_remove_all __attribute((alias("IA__gtk_statusbar_remove_all"), visibility("default")));

#undef gtk_statusbar_set_has_resize_grip
extern __typeof (gtk_statusbar_set_has_resize_grip) gtk_statusbar_set_has_resize_grip __attribute((alias("IA__gtk_statusbar_set_has_resize_grip"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_STOCK_H__)
#if IN_FILE(__GTK_STOCK_C__)
#undef gtk_stock_add
extern __typeof (gtk_stock_add) gtk_stock_add __attribute((alias("IA__gtk_stock_add"), visibility("default")));

#undef gtk_stock_add_static
extern __typeof (gtk_stock_add_static) gtk_stock_add_static __attribute((alias("IA__gtk_stock_add_static"), visibility("default")));

#undef gtk_stock_item_get_type
extern __typeof (gtk_stock_item_get_type) gtk_stock_item_get_type __attribute((alias("IA__gtk_stock_item_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_stock_item_copy
extern __typeof (gtk_stock_item_copy) gtk_stock_item_copy __attribute((alias("IA__gtk_stock_item_copy"), visibility("default")));

#undef gtk_stock_item_free
extern __typeof (gtk_stock_item_free) gtk_stock_item_free __attribute((alias("IA__gtk_stock_item_free"), visibility("default")));

#undef gtk_stock_list_ids
extern __typeof (gtk_stock_list_ids) gtk_stock_list_ids __attribute((alias("IA__gtk_stock_list_ids"), visibility("default")));

#undef gtk_stock_lookup
extern __typeof (gtk_stock_lookup) gtk_stock_lookup __attribute((alias("IA__gtk_stock_lookup"), visibility("default")));

#undef gtk_stock_set_translate_func
extern __typeof (gtk_stock_set_translate_func) gtk_stock_set_translate_func __attribute((alias("IA__gtk_stock_set_translate_func"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TABLE_H__)
#if IN_FILE(__GTK_TABLE_C__)
#undef gtk_table_attach
extern __typeof (gtk_table_attach) gtk_table_attach __attribute((alias("IA__gtk_table_attach"), visibility("default")));

#undef gtk_table_attach_defaults
extern __typeof (gtk_table_attach_defaults) gtk_table_attach_defaults __attribute((alias("IA__gtk_table_attach_defaults"), visibility("default")));

#undef gtk_table_get_col_spacing
extern __typeof (gtk_table_get_col_spacing) gtk_table_get_col_spacing __attribute((alias("IA__gtk_table_get_col_spacing"), visibility("default")));

#undef gtk_table_get_default_col_spacing
extern __typeof (gtk_table_get_default_col_spacing) gtk_table_get_default_col_spacing __attribute((alias("IA__gtk_table_get_default_col_spacing"), visibility("default")));

#undef gtk_table_get_default_row_spacing
extern __typeof (gtk_table_get_default_row_spacing) gtk_table_get_default_row_spacing __attribute((alias("IA__gtk_table_get_default_row_spacing"), visibility("default")));

#undef gtk_table_get_homogeneous
extern __typeof (gtk_table_get_homogeneous) gtk_table_get_homogeneous __attribute((alias("IA__gtk_table_get_homogeneous"), visibility("default")));

#undef gtk_table_get_row_spacing
extern __typeof (gtk_table_get_row_spacing) gtk_table_get_row_spacing __attribute((alias("IA__gtk_table_get_row_spacing"), visibility("default")));

#undef gtk_table_get_size
extern __typeof (gtk_table_get_size) gtk_table_get_size __attribute((alias("IA__gtk_table_get_size"), visibility("default")));

#undef gtk_table_get_type
extern __typeof (gtk_table_get_type) gtk_table_get_type __attribute((alias("IA__gtk_table_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_table_new
extern __typeof (gtk_table_new) gtk_table_new __attribute((alias("IA__gtk_table_new"), visibility("default")));

#undef gtk_table_resize
extern __typeof (gtk_table_resize) gtk_table_resize __attribute((alias("IA__gtk_table_resize"), visibility("default")));

#undef gtk_table_set_col_spacing
extern __typeof (gtk_table_set_col_spacing) gtk_table_set_col_spacing __attribute((alias("IA__gtk_table_set_col_spacing"), visibility("default")));

#undef gtk_table_set_col_spacings
extern __typeof (gtk_table_set_col_spacings) gtk_table_set_col_spacings __attribute((alias("IA__gtk_table_set_col_spacings"), visibility("default")));

#undef gtk_table_set_homogeneous
extern __typeof (gtk_table_set_homogeneous) gtk_table_set_homogeneous __attribute((alias("IA__gtk_table_set_homogeneous"), visibility("default")));

#undef gtk_table_set_row_spacing
extern __typeof (gtk_table_set_row_spacing) gtk_table_set_row_spacing __attribute((alias("IA__gtk_table_set_row_spacing"), visibility("default")));

#undef gtk_table_set_row_spacings
extern __typeof (gtk_table_set_row_spacings) gtk_table_set_row_spacings __attribute((alias("IA__gtk_table_set_row_spacings"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEAROFF_MENU_ITEM_H__)
#if IN_FILE(__GTK_TEAROFF_MENU_ITEM_C__)
#undef gtk_tearoff_menu_item_get_type
extern __typeof (gtk_tearoff_menu_item_get_type) gtk_tearoff_menu_item_get_type __attribute((alias("IA__gtk_tearoff_menu_item_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tearoff_menu_item_new
extern __typeof (gtk_tearoff_menu_item_new) gtk_tearoff_menu_item_new __attribute((alias("IA__gtk_tearoff_menu_item_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEXT_TAG_H__)
#if IN_FILE(__GTK_TEXT_TAG_C__)
#undef gtk_text_attributes_copy
extern __typeof (gtk_text_attributes_copy) gtk_text_attributes_copy __attribute((alias("IA__gtk_text_attributes_copy"), visibility("default")));

#undef gtk_text_attributes_copy_values
extern __typeof (gtk_text_attributes_copy_values) gtk_text_attributes_copy_values __attribute((alias("IA__gtk_text_attributes_copy_values"), visibility("default")));

#undef gtk_text_attributes_get_type
extern __typeof (gtk_text_attributes_get_type) gtk_text_attributes_get_type __attribute((alias("IA__gtk_text_attributes_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_text_attributes_new
extern __typeof (gtk_text_attributes_new) gtk_text_attributes_new __attribute((alias("IA__gtk_text_attributes_new"), visibility("default")));

#undef gtk_text_attributes_ref
extern __typeof (gtk_text_attributes_ref) gtk_text_attributes_ref __attribute((alias("IA__gtk_text_attributes_ref"), visibility("default")));

#undef gtk_text_attributes_unref
extern __typeof (gtk_text_attributes_unref) gtk_text_attributes_unref __attribute((alias("IA__gtk_text_attributes_unref"), visibility("default")));

#undef gtk_text_tag_event
extern __typeof (gtk_text_tag_event) gtk_text_tag_event __attribute((alias("IA__gtk_text_tag_event"), visibility("default")));

#undef gtk_text_tag_get_priority
extern __typeof (gtk_text_tag_get_priority) gtk_text_tag_get_priority __attribute((alias("IA__gtk_text_tag_get_priority"), visibility("default")));

#undef gtk_text_tag_get_type
extern __typeof (gtk_text_tag_get_type) gtk_text_tag_get_type __attribute((alias("IA__gtk_text_tag_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_text_tag_new
extern __typeof (gtk_text_tag_new) gtk_text_tag_new __attribute((alias("IA__gtk_text_tag_new"), visibility("default")));

#undef gtk_text_tag_set_priority
extern __typeof (gtk_text_tag_set_priority) gtk_text_tag_set_priority __attribute((alias("IA__gtk_text_tag_set_priority"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEXT_TAG_TABLE_H__)
#if IN_FILE(__GTK_TEXT_TAG_TABLE_C__)
#undef gtk_text_tag_table_add
extern __typeof (gtk_text_tag_table_add) gtk_text_tag_table_add __attribute((alias("IA__gtk_text_tag_table_add"), visibility("default")));

#undef gtk_text_tag_table_foreach
extern __typeof (gtk_text_tag_table_foreach) gtk_text_tag_table_foreach __attribute((alias("IA__gtk_text_tag_table_foreach"), visibility("default")));

#undef gtk_text_tag_table_get_size
extern __typeof (gtk_text_tag_table_get_size) gtk_text_tag_table_get_size __attribute((alias("IA__gtk_text_tag_table_get_size"), visibility("default")));

#undef gtk_text_tag_table_get_type
extern __typeof (gtk_text_tag_table_get_type) gtk_text_tag_table_get_type __attribute((alias("IA__gtk_text_tag_table_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_text_tag_table_lookup
extern __typeof (gtk_text_tag_table_lookup) gtk_text_tag_table_lookup __attribute((alias("IA__gtk_text_tag_table_lookup"), visibility("default")));

#undef gtk_text_tag_table_new
extern __typeof (gtk_text_tag_table_new) gtk_text_tag_table_new __attribute((alias("IA__gtk_text_tag_table_new"), visibility("default")));

#undef gtk_text_tag_table_remove
extern __typeof (gtk_text_tag_table_remove) gtk_text_tag_table_remove __attribute((alias("IA__gtk_text_tag_table_remove"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEXT_H__)
#if IN_FILE(__GTK_TEXT_C__)
#undef gtk_text_backward_delete
extern __typeof (gtk_text_backward_delete) gtk_text_backward_delete __attribute((alias("IA__gtk_text_backward_delete"), visibility("default")));

#undef gtk_text_forward_delete
extern __typeof (gtk_text_forward_delete) gtk_text_forward_delete __attribute((alias("IA__gtk_text_forward_delete"), visibility("default")));

#undef gtk_text_freeze
extern __typeof (gtk_text_freeze) gtk_text_freeze __attribute((alias("IA__gtk_text_freeze"), visibility("default")));

#undef gtk_text_get_length
extern __typeof (gtk_text_get_length) gtk_text_get_length __attribute((alias("IA__gtk_text_get_length"), visibility("default")));

#undef gtk_text_get_point
extern __typeof (gtk_text_get_point) gtk_text_get_point __attribute((alias("IA__gtk_text_get_point"), visibility("default")));

#undef gtk_text_get_type
extern __typeof (gtk_text_get_type) gtk_text_get_type __attribute((alias("IA__gtk_text_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_text_insert
extern __typeof (gtk_text_insert) gtk_text_insert __attribute((alias("IA__gtk_text_insert"), visibility("default")));

#undef gtk_text_new
extern __typeof (gtk_text_new) gtk_text_new __attribute((alias("IA__gtk_text_new"), visibility("default")));

#undef gtk_text_set_adjustments
extern __typeof (gtk_text_set_adjustments) gtk_text_set_adjustments __attribute((alias("IA__gtk_text_set_adjustments"), visibility("default")));

#undef gtk_text_set_editable
extern __typeof (gtk_text_set_editable) gtk_text_set_editable __attribute((alias("IA__gtk_text_set_editable"), visibility("default")));

#undef gtk_text_set_line_wrap
extern __typeof (gtk_text_set_line_wrap) gtk_text_set_line_wrap __attribute((alias("IA__gtk_text_set_line_wrap"), visibility("default")));

#undef gtk_text_set_point
extern __typeof (gtk_text_set_point) gtk_text_set_point __attribute((alias("IA__gtk_text_set_point"), visibility("default")));

#undef gtk_text_set_word_wrap
extern __typeof (gtk_text_set_word_wrap) gtk_text_set_word_wrap __attribute((alias("IA__gtk_text_set_word_wrap"), visibility("default")));

#undef gtk_text_thaw
extern __typeof (gtk_text_thaw) gtk_text_thaw __attribute((alias("IA__gtk_text_thaw"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEXT_BUFFER_H__)
#if IN_FILE(__GTK_TEXT_BUFFER_C__)
#undef gtk_text_buffer_add_selection_clipboard
extern __typeof (gtk_text_buffer_add_selection_clipboard) gtk_text_buffer_add_selection_clipboard __attribute((alias("IA__gtk_text_buffer_add_selection_clipboard"), visibility("default")));

#undef gtk_text_buffer_apply_tag
extern __typeof (gtk_text_buffer_apply_tag) gtk_text_buffer_apply_tag __attribute((alias("IA__gtk_text_buffer_apply_tag"), visibility("default")));

#undef gtk_text_buffer_apply_tag_by_name
extern __typeof (gtk_text_buffer_apply_tag_by_name) gtk_text_buffer_apply_tag_by_name __attribute((alias("IA__gtk_text_buffer_apply_tag_by_name"), visibility("default")));

#undef gtk_text_buffer_backspace
extern __typeof (gtk_text_buffer_backspace) gtk_text_buffer_backspace __attribute((alias("IA__gtk_text_buffer_backspace"), visibility("default")));

#undef gtk_text_buffer_begin_user_action
extern __typeof (gtk_text_buffer_begin_user_action) gtk_text_buffer_begin_user_action __attribute((alias("IA__gtk_text_buffer_begin_user_action"), visibility("default")));

#undef gtk_text_buffer_copy_clipboard
extern __typeof (gtk_text_buffer_copy_clipboard) gtk_text_buffer_copy_clipboard __attribute((alias("IA__gtk_text_buffer_copy_clipboard"), visibility("default")));

#undef gtk_text_buffer_create_child_anchor
extern __typeof (gtk_text_buffer_create_child_anchor) gtk_text_buffer_create_child_anchor __attribute((alias("IA__gtk_text_buffer_create_child_anchor"), visibility("default")));

#undef gtk_text_buffer_add_mark
extern __typeof (gtk_text_buffer_add_mark) gtk_text_buffer_add_mark __attribute((alias("IA__gtk_text_buffer_add_mark"), visibility("default")));

#undef gtk_text_buffer_create_mark
extern __typeof (gtk_text_buffer_create_mark) gtk_text_buffer_create_mark __attribute((alias("IA__gtk_text_buffer_create_mark"), visibility("default")));

#undef gtk_text_buffer_create_tag
extern __typeof (gtk_text_buffer_create_tag) gtk_text_buffer_create_tag __attribute((alias("IA__gtk_text_buffer_create_tag"), visibility("default")));

#undef gtk_text_buffer_cut_clipboard
extern __typeof (gtk_text_buffer_cut_clipboard) gtk_text_buffer_cut_clipboard __attribute((alias("IA__gtk_text_buffer_cut_clipboard"), visibility("default")));

#undef gtk_text_buffer_delete
extern __typeof (gtk_text_buffer_delete) gtk_text_buffer_delete __attribute((alias("IA__gtk_text_buffer_delete"), visibility("default")));

#undef gtk_text_buffer_delete_interactive
extern __typeof (gtk_text_buffer_delete_interactive) gtk_text_buffer_delete_interactive __attribute((alias("IA__gtk_text_buffer_delete_interactive"), visibility("default")));

#undef gtk_text_buffer_delete_mark
extern __typeof (gtk_text_buffer_delete_mark) gtk_text_buffer_delete_mark __attribute((alias("IA__gtk_text_buffer_delete_mark"), visibility("default")));

#undef gtk_text_buffer_delete_mark_by_name
extern __typeof (gtk_text_buffer_delete_mark_by_name) gtk_text_buffer_delete_mark_by_name __attribute((alias("IA__gtk_text_buffer_delete_mark_by_name"), visibility("default")));

#undef gtk_text_buffer_delete_selection
extern __typeof (gtk_text_buffer_delete_selection) gtk_text_buffer_delete_selection __attribute((alias("IA__gtk_text_buffer_delete_selection"), visibility("default")));

#undef gtk_text_buffer_end_user_action
extern __typeof (gtk_text_buffer_end_user_action) gtk_text_buffer_end_user_action __attribute((alias("IA__gtk_text_buffer_end_user_action"), visibility("default")));

#undef gtk_text_buffer_get_bounds
extern __typeof (gtk_text_buffer_get_bounds) gtk_text_buffer_get_bounds __attribute((alias("IA__gtk_text_buffer_get_bounds"), visibility("default")));

#undef gtk_text_buffer_get_char_count
extern __typeof (gtk_text_buffer_get_char_count) gtk_text_buffer_get_char_count __attribute((alias("IA__gtk_text_buffer_get_char_count"), visibility("default")));

#undef gtk_text_buffer_get_copy_target_list
extern __typeof (gtk_text_buffer_get_copy_target_list) gtk_text_buffer_get_copy_target_list __attribute((alias("IA__gtk_text_buffer_get_copy_target_list"), visibility("default")));

#undef gtk_text_buffer_get_end_iter
extern __typeof (gtk_text_buffer_get_end_iter) gtk_text_buffer_get_end_iter __attribute((alias("IA__gtk_text_buffer_get_end_iter"), visibility("default")));

#undef gtk_text_buffer_get_has_selection
extern __typeof (gtk_text_buffer_get_has_selection) gtk_text_buffer_get_has_selection __attribute((alias("IA__gtk_text_buffer_get_has_selection"), visibility("default")));

#undef gtk_text_buffer_get_insert
extern __typeof (gtk_text_buffer_get_insert) gtk_text_buffer_get_insert __attribute((alias("IA__gtk_text_buffer_get_insert"), visibility("default")));

#undef gtk_text_buffer_get_iter_at_child_anchor
extern __typeof (gtk_text_buffer_get_iter_at_child_anchor) gtk_text_buffer_get_iter_at_child_anchor __attribute((alias("IA__gtk_text_buffer_get_iter_at_child_anchor"), visibility("default")));

#undef gtk_text_buffer_get_iter_at_line
extern __typeof (gtk_text_buffer_get_iter_at_line) gtk_text_buffer_get_iter_at_line __attribute((alias("IA__gtk_text_buffer_get_iter_at_line"), visibility("default")));

#undef gtk_text_buffer_get_iter_at_line_index
extern __typeof (gtk_text_buffer_get_iter_at_line_index) gtk_text_buffer_get_iter_at_line_index __attribute((alias("IA__gtk_text_buffer_get_iter_at_line_index"), visibility("default")));

#undef gtk_text_buffer_get_iter_at_line_offset
extern __typeof (gtk_text_buffer_get_iter_at_line_offset) gtk_text_buffer_get_iter_at_line_offset __attribute((alias("IA__gtk_text_buffer_get_iter_at_line_offset"), visibility("default")));

#undef gtk_text_buffer_get_iter_at_mark
extern __typeof (gtk_text_buffer_get_iter_at_mark) gtk_text_buffer_get_iter_at_mark __attribute((alias("IA__gtk_text_buffer_get_iter_at_mark"), visibility("default")));

#undef gtk_text_buffer_get_iter_at_offset
extern __typeof (gtk_text_buffer_get_iter_at_offset) gtk_text_buffer_get_iter_at_offset __attribute((alias("IA__gtk_text_buffer_get_iter_at_offset"), visibility("default")));

#undef gtk_text_buffer_get_line_count
extern __typeof (gtk_text_buffer_get_line_count) gtk_text_buffer_get_line_count __attribute((alias("IA__gtk_text_buffer_get_line_count"), visibility("default")));

#undef gtk_text_buffer_get_mark
extern __typeof (gtk_text_buffer_get_mark) gtk_text_buffer_get_mark __attribute((alias("IA__gtk_text_buffer_get_mark"), visibility("default")));

#undef gtk_text_buffer_get_modified
extern __typeof (gtk_text_buffer_get_modified) gtk_text_buffer_get_modified __attribute((alias("IA__gtk_text_buffer_get_modified"), visibility("default")));

#undef gtk_text_buffer_get_paste_target_list
extern __typeof (gtk_text_buffer_get_paste_target_list) gtk_text_buffer_get_paste_target_list __attribute((alias("IA__gtk_text_buffer_get_paste_target_list"), visibility("default")));

#undef gtk_text_buffer_get_selection_bound
extern __typeof (gtk_text_buffer_get_selection_bound) gtk_text_buffer_get_selection_bound __attribute((alias("IA__gtk_text_buffer_get_selection_bound"), visibility("default")));

#undef gtk_text_buffer_get_selection_bounds
extern __typeof (gtk_text_buffer_get_selection_bounds) gtk_text_buffer_get_selection_bounds __attribute((alias("IA__gtk_text_buffer_get_selection_bounds"), visibility("default")));

#undef gtk_text_buffer_get_slice
extern __typeof (gtk_text_buffer_get_slice) gtk_text_buffer_get_slice __attribute((alias("IA__gtk_text_buffer_get_slice"), visibility("default")));

#undef gtk_text_buffer_get_start_iter
extern __typeof (gtk_text_buffer_get_start_iter) gtk_text_buffer_get_start_iter __attribute((alias("IA__gtk_text_buffer_get_start_iter"), visibility("default")));

#undef gtk_text_buffer_get_tag_table
extern __typeof (gtk_text_buffer_get_tag_table) gtk_text_buffer_get_tag_table __attribute((alias("IA__gtk_text_buffer_get_tag_table"), visibility("default")));

#undef gtk_text_buffer_get_text
extern __typeof (gtk_text_buffer_get_text) gtk_text_buffer_get_text __attribute((alias("IA__gtk_text_buffer_get_text"), visibility("default")));

#undef gtk_text_buffer_get_type
extern __typeof (gtk_text_buffer_get_type) gtk_text_buffer_get_type __attribute((alias("IA__gtk_text_buffer_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_text_buffer_insert
extern __typeof (gtk_text_buffer_insert) gtk_text_buffer_insert __attribute((alias("IA__gtk_text_buffer_insert"), visibility("default")));

#undef gtk_text_buffer_insert_at_cursor
extern __typeof (gtk_text_buffer_insert_at_cursor) gtk_text_buffer_insert_at_cursor __attribute((alias("IA__gtk_text_buffer_insert_at_cursor"), visibility("default")));

#undef gtk_text_buffer_insert_child_anchor
extern __typeof (gtk_text_buffer_insert_child_anchor) gtk_text_buffer_insert_child_anchor __attribute((alias("IA__gtk_text_buffer_insert_child_anchor"), visibility("default")));

#undef gtk_text_buffer_insert_interactive
extern __typeof (gtk_text_buffer_insert_interactive) gtk_text_buffer_insert_interactive __attribute((alias("IA__gtk_text_buffer_insert_interactive"), visibility("default")));

#undef gtk_text_buffer_insert_interactive_at_cursor
extern __typeof (gtk_text_buffer_insert_interactive_at_cursor) gtk_text_buffer_insert_interactive_at_cursor __attribute((alias("IA__gtk_text_buffer_insert_interactive_at_cursor"), visibility("default")));

#undef gtk_text_buffer_insert_pixbuf
extern __typeof (gtk_text_buffer_insert_pixbuf) gtk_text_buffer_insert_pixbuf __attribute((alias("IA__gtk_text_buffer_insert_pixbuf"), visibility("default")));

#undef gtk_text_buffer_insert_range
extern __typeof (gtk_text_buffer_insert_range) gtk_text_buffer_insert_range __attribute((alias("IA__gtk_text_buffer_insert_range"), visibility("default")));

#undef gtk_text_buffer_insert_range_interactive
extern __typeof (gtk_text_buffer_insert_range_interactive) gtk_text_buffer_insert_range_interactive __attribute((alias("IA__gtk_text_buffer_insert_range_interactive"), visibility("default")));

#undef gtk_text_buffer_insert_with_tags
extern __typeof (gtk_text_buffer_insert_with_tags) gtk_text_buffer_insert_with_tags __attribute((alias("IA__gtk_text_buffer_insert_with_tags"), visibility("default"))) G_GNUC_NULL_TERMINATED;

#undef gtk_text_buffer_insert_with_tags_by_name
extern __typeof (gtk_text_buffer_insert_with_tags_by_name) gtk_text_buffer_insert_with_tags_by_name __attribute((alias("IA__gtk_text_buffer_insert_with_tags_by_name"), visibility("default"))) G_GNUC_NULL_TERMINATED;

#undef gtk_text_buffer_move_mark
extern __typeof (gtk_text_buffer_move_mark) gtk_text_buffer_move_mark __attribute((alias("IA__gtk_text_buffer_move_mark"), visibility("default")));

#undef gtk_text_buffer_move_mark_by_name
extern __typeof (gtk_text_buffer_move_mark_by_name) gtk_text_buffer_move_mark_by_name __attribute((alias("IA__gtk_text_buffer_move_mark_by_name"), visibility("default")));

#undef gtk_text_buffer_new
extern __typeof (gtk_text_buffer_new) gtk_text_buffer_new __attribute((alias("IA__gtk_text_buffer_new"), visibility("default")));

#undef gtk_text_buffer_paste_clipboard
extern __typeof (gtk_text_buffer_paste_clipboard) gtk_text_buffer_paste_clipboard __attribute((alias("IA__gtk_text_buffer_paste_clipboard"), visibility("default")));

#undef gtk_text_buffer_place_cursor
extern __typeof (gtk_text_buffer_place_cursor) gtk_text_buffer_place_cursor __attribute((alias("IA__gtk_text_buffer_place_cursor"), visibility("default")));

#undef gtk_text_buffer_remove_all_tags
extern __typeof (gtk_text_buffer_remove_all_tags) gtk_text_buffer_remove_all_tags __attribute((alias("IA__gtk_text_buffer_remove_all_tags"), visibility("default")));

#undef gtk_text_buffer_remove_selection_clipboard
extern __typeof (gtk_text_buffer_remove_selection_clipboard) gtk_text_buffer_remove_selection_clipboard __attribute((alias("IA__gtk_text_buffer_remove_selection_clipboard"), visibility("default")));

#undef gtk_text_buffer_remove_tag
extern __typeof (gtk_text_buffer_remove_tag) gtk_text_buffer_remove_tag __attribute((alias("IA__gtk_text_buffer_remove_tag"), visibility("default")));

#undef gtk_text_buffer_remove_tag_by_name
extern __typeof (gtk_text_buffer_remove_tag_by_name) gtk_text_buffer_remove_tag_by_name __attribute((alias("IA__gtk_text_buffer_remove_tag_by_name"), visibility("default")));

#undef gtk_text_buffer_select_range
extern __typeof (gtk_text_buffer_select_range) gtk_text_buffer_select_range __attribute((alias("IA__gtk_text_buffer_select_range"), visibility("default")));

#undef gtk_text_buffer_set_modified
extern __typeof (gtk_text_buffer_set_modified) gtk_text_buffer_set_modified __attribute((alias("IA__gtk_text_buffer_set_modified"), visibility("default")));

#undef gtk_text_buffer_set_text
extern __typeof (gtk_text_buffer_set_text) gtk_text_buffer_set_text __attribute((alias("IA__gtk_text_buffer_set_text"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEXT_TYPES_H__)
#if IN_FILE(__GTK_TEXT_TYPES_C__)
#undef gtk_text_byte_begins_utf8_char
extern __typeof (gtk_text_byte_begins_utf8_char) gtk_text_byte_begins_utf8_char __attribute((alias("IA__gtk_text_byte_begins_utf8_char"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEXT_ITER_H__)
#if IN_FILE(__GTK_TEXT_ITER_C__)
#undef gtk_text_iter_backward_char
extern __typeof (gtk_text_iter_backward_char) gtk_text_iter_backward_char __attribute((alias("IA__gtk_text_iter_backward_char"), visibility("default")));

#undef gtk_text_iter_backward_chars
extern __typeof (gtk_text_iter_backward_chars) gtk_text_iter_backward_chars __attribute((alias("IA__gtk_text_iter_backward_chars"), visibility("default")));

#undef gtk_text_iter_backward_cursor_position
extern __typeof (gtk_text_iter_backward_cursor_position) gtk_text_iter_backward_cursor_position __attribute((alias("IA__gtk_text_iter_backward_cursor_position"), visibility("default")));

#undef gtk_text_iter_backward_cursor_positions
extern __typeof (gtk_text_iter_backward_cursor_positions) gtk_text_iter_backward_cursor_positions __attribute((alias("IA__gtk_text_iter_backward_cursor_positions"), visibility("default")));

#undef gtk_text_iter_backward_find_char
extern __typeof (gtk_text_iter_backward_find_char) gtk_text_iter_backward_find_char __attribute((alias("IA__gtk_text_iter_backward_find_char"), visibility("default")));

#undef gtk_text_iter_backward_line
extern __typeof (gtk_text_iter_backward_line) gtk_text_iter_backward_line __attribute((alias("IA__gtk_text_iter_backward_line"), visibility("default")));

#undef gtk_text_iter_backward_lines
extern __typeof (gtk_text_iter_backward_lines) gtk_text_iter_backward_lines __attribute((alias("IA__gtk_text_iter_backward_lines"), visibility("default")));

#undef gtk_text_iter_backward_search
extern __typeof (gtk_text_iter_backward_search) gtk_text_iter_backward_search __attribute((alias("IA__gtk_text_iter_backward_search"), visibility("default")));

#undef gtk_text_iter_backward_sentence_start
extern __typeof (gtk_text_iter_backward_sentence_start) gtk_text_iter_backward_sentence_start __attribute((alias("IA__gtk_text_iter_backward_sentence_start"), visibility("default")));

#undef gtk_text_iter_backward_sentence_starts
extern __typeof (gtk_text_iter_backward_sentence_starts) gtk_text_iter_backward_sentence_starts __attribute((alias("IA__gtk_text_iter_backward_sentence_starts"), visibility("default")));

#undef gtk_text_iter_backward_to_tag_toggle
extern __typeof (gtk_text_iter_backward_to_tag_toggle) gtk_text_iter_backward_to_tag_toggle __attribute((alias("IA__gtk_text_iter_backward_to_tag_toggle"), visibility("default")));

#undef gtk_text_iter_backward_visible_cursor_position
extern __typeof (gtk_text_iter_backward_visible_cursor_position) gtk_text_iter_backward_visible_cursor_position __attribute((alias("IA__gtk_text_iter_backward_visible_cursor_position"), visibility("default")));

#undef gtk_text_iter_backward_visible_cursor_positions
extern __typeof (gtk_text_iter_backward_visible_cursor_positions) gtk_text_iter_backward_visible_cursor_positions __attribute((alias("IA__gtk_text_iter_backward_visible_cursor_positions"), visibility("default")));

#undef gtk_text_iter_backward_visible_line
extern __typeof (gtk_text_iter_backward_visible_line) gtk_text_iter_backward_visible_line __attribute((alias("IA__gtk_text_iter_backward_visible_line"), visibility("default")));

#undef gtk_text_iter_backward_visible_lines
extern __typeof (gtk_text_iter_backward_visible_lines) gtk_text_iter_backward_visible_lines __attribute((alias("IA__gtk_text_iter_backward_visible_lines"), visibility("default")));

#undef gtk_text_iter_backward_visible_word_start
extern __typeof (gtk_text_iter_backward_visible_word_start) gtk_text_iter_backward_visible_word_start __attribute((alias("IA__gtk_text_iter_backward_visible_word_start"), visibility("default")));

#undef gtk_text_iter_backward_visible_word_starts
extern __typeof (gtk_text_iter_backward_visible_word_starts) gtk_text_iter_backward_visible_word_starts __attribute((alias("IA__gtk_text_iter_backward_visible_word_starts"), visibility("default")));

#undef gtk_text_iter_backward_word_start
extern __typeof (gtk_text_iter_backward_word_start) gtk_text_iter_backward_word_start __attribute((alias("IA__gtk_text_iter_backward_word_start"), visibility("default")));

#undef gtk_text_iter_backward_word_starts
extern __typeof (gtk_text_iter_backward_word_starts) gtk_text_iter_backward_word_starts __attribute((alias("IA__gtk_text_iter_backward_word_starts"), visibility("default")));

#undef gtk_text_iter_begins_tag
extern __typeof (gtk_text_iter_begins_tag) gtk_text_iter_begins_tag __attribute((alias("IA__gtk_text_iter_begins_tag"), visibility("default")));

#undef gtk_text_iter_can_insert
extern __typeof (gtk_text_iter_can_insert) gtk_text_iter_can_insert __attribute((alias("IA__gtk_text_iter_can_insert"), visibility("default")));

#undef gtk_text_iter_compare
extern __typeof (gtk_text_iter_compare) gtk_text_iter_compare __attribute((alias("IA__gtk_text_iter_compare"), visibility("default")));

#undef gtk_text_iter_copy
extern __typeof (gtk_text_iter_copy) gtk_text_iter_copy __attribute((alias("IA__gtk_text_iter_copy"), visibility("default")));

#undef gtk_text_iter_editable
extern __typeof (gtk_text_iter_editable) gtk_text_iter_editable __attribute((alias("IA__gtk_text_iter_editable"), visibility("default")));

#undef gtk_text_iter_ends_line
extern __typeof (gtk_text_iter_ends_line) gtk_text_iter_ends_line __attribute((alias("IA__gtk_text_iter_ends_line"), visibility("default")));

#undef gtk_text_iter_ends_sentence
extern __typeof (gtk_text_iter_ends_sentence) gtk_text_iter_ends_sentence __attribute((alias("IA__gtk_text_iter_ends_sentence"), visibility("default")));

#undef gtk_text_iter_ends_tag
extern __typeof (gtk_text_iter_ends_tag) gtk_text_iter_ends_tag __attribute((alias("IA__gtk_text_iter_ends_tag"), visibility("default")));

#undef gtk_text_iter_ends_word
extern __typeof (gtk_text_iter_ends_word) gtk_text_iter_ends_word __attribute((alias("IA__gtk_text_iter_ends_word"), visibility("default")));

#undef gtk_text_iter_equal
extern __typeof (gtk_text_iter_equal) gtk_text_iter_equal __attribute((alias("IA__gtk_text_iter_equal"), visibility("default")));

#undef gtk_text_iter_forward_char
extern __typeof (gtk_text_iter_forward_char) gtk_text_iter_forward_char __attribute((alias("IA__gtk_text_iter_forward_char"), visibility("default")));

#undef gtk_text_iter_forward_chars
extern __typeof (gtk_text_iter_forward_chars) gtk_text_iter_forward_chars __attribute((alias("IA__gtk_text_iter_forward_chars"), visibility("default")));

#undef gtk_text_iter_forward_cursor_position
extern __typeof (gtk_text_iter_forward_cursor_position) gtk_text_iter_forward_cursor_position __attribute((alias("IA__gtk_text_iter_forward_cursor_position"), visibility("default")));

#undef gtk_text_iter_forward_cursor_positions
extern __typeof (gtk_text_iter_forward_cursor_positions) gtk_text_iter_forward_cursor_positions __attribute((alias("IA__gtk_text_iter_forward_cursor_positions"), visibility("default")));

#undef gtk_text_iter_forward_find_char
extern __typeof (gtk_text_iter_forward_find_char) gtk_text_iter_forward_find_char __attribute((alias("IA__gtk_text_iter_forward_find_char"), visibility("default")));

#undef gtk_text_iter_forward_line
extern __typeof (gtk_text_iter_forward_line) gtk_text_iter_forward_line __attribute((alias("IA__gtk_text_iter_forward_line"), visibility("default")));

#undef gtk_text_iter_forward_lines
extern __typeof (gtk_text_iter_forward_lines) gtk_text_iter_forward_lines __attribute((alias("IA__gtk_text_iter_forward_lines"), visibility("default")));

#undef gtk_text_iter_forward_search
extern __typeof (gtk_text_iter_forward_search) gtk_text_iter_forward_search __attribute((alias("IA__gtk_text_iter_forward_search"), visibility("default")));

#undef gtk_text_iter_forward_sentence_end
extern __typeof (gtk_text_iter_forward_sentence_end) gtk_text_iter_forward_sentence_end __attribute((alias("IA__gtk_text_iter_forward_sentence_end"), visibility("default")));

#undef gtk_text_iter_forward_sentence_ends
extern __typeof (gtk_text_iter_forward_sentence_ends) gtk_text_iter_forward_sentence_ends __attribute((alias("IA__gtk_text_iter_forward_sentence_ends"), visibility("default")));

#undef gtk_text_iter_forward_to_end
extern __typeof (gtk_text_iter_forward_to_end) gtk_text_iter_forward_to_end __attribute((alias("IA__gtk_text_iter_forward_to_end"), visibility("default")));

#undef gtk_text_iter_forward_to_line_end
extern __typeof (gtk_text_iter_forward_to_line_end) gtk_text_iter_forward_to_line_end __attribute((alias("IA__gtk_text_iter_forward_to_line_end"), visibility("default")));

#undef gtk_text_iter_forward_to_tag_toggle
extern __typeof (gtk_text_iter_forward_to_tag_toggle) gtk_text_iter_forward_to_tag_toggle __attribute((alias("IA__gtk_text_iter_forward_to_tag_toggle"), visibility("default")));

#undef gtk_text_iter_forward_visible_cursor_position
extern __typeof (gtk_text_iter_forward_visible_cursor_position) gtk_text_iter_forward_visible_cursor_position __attribute((alias("IA__gtk_text_iter_forward_visible_cursor_position"), visibility("default")));

#undef gtk_text_iter_forward_visible_cursor_positions
extern __typeof (gtk_text_iter_forward_visible_cursor_positions) gtk_text_iter_forward_visible_cursor_positions __attribute((alias("IA__gtk_text_iter_forward_visible_cursor_positions"), visibility("default")));

#undef gtk_text_iter_forward_visible_line
extern __typeof (gtk_text_iter_forward_visible_line) gtk_text_iter_forward_visible_line __attribute((alias("IA__gtk_text_iter_forward_visible_line"), visibility("default")));

#undef gtk_text_iter_forward_visible_lines
extern __typeof (gtk_text_iter_forward_visible_lines) gtk_text_iter_forward_visible_lines __attribute((alias("IA__gtk_text_iter_forward_visible_lines"), visibility("default")));

#undef gtk_text_iter_forward_visible_word_end
extern __typeof (gtk_text_iter_forward_visible_word_end) gtk_text_iter_forward_visible_word_end __attribute((alias("IA__gtk_text_iter_forward_visible_word_end"), visibility("default")));

#undef gtk_text_iter_forward_visible_word_ends
extern __typeof (gtk_text_iter_forward_visible_word_ends) gtk_text_iter_forward_visible_word_ends __attribute((alias("IA__gtk_text_iter_forward_visible_word_ends"), visibility("default")));

#undef gtk_text_iter_forward_word_end
extern __typeof (gtk_text_iter_forward_word_end) gtk_text_iter_forward_word_end __attribute((alias("IA__gtk_text_iter_forward_word_end"), visibility("default")));

#undef gtk_text_iter_forward_word_ends
extern __typeof (gtk_text_iter_forward_word_ends) gtk_text_iter_forward_word_ends __attribute((alias("IA__gtk_text_iter_forward_word_ends"), visibility("default")));

#undef gtk_text_iter_free
extern __typeof (gtk_text_iter_free) gtk_text_iter_free __attribute((alias("IA__gtk_text_iter_free"), visibility("default")));

#undef gtk_text_iter_get_attributes
extern __typeof (gtk_text_iter_get_attributes) gtk_text_iter_get_attributes __attribute((alias("IA__gtk_text_iter_get_attributes"), visibility("default")));

#undef gtk_text_iter_get_buffer
extern __typeof (gtk_text_iter_get_buffer) gtk_text_iter_get_buffer __attribute((alias("IA__gtk_text_iter_get_buffer"), visibility("default")));

#undef gtk_text_iter_get_bytes_in_line
extern __typeof (gtk_text_iter_get_bytes_in_line) gtk_text_iter_get_bytes_in_line __attribute((alias("IA__gtk_text_iter_get_bytes_in_line"), visibility("default")));

#undef gtk_text_iter_get_char
extern __typeof (gtk_text_iter_get_char) gtk_text_iter_get_char __attribute((alias("IA__gtk_text_iter_get_char"), visibility("default")));

#undef gtk_text_iter_get_chars_in_line
extern __typeof (gtk_text_iter_get_chars_in_line) gtk_text_iter_get_chars_in_line __attribute((alias("IA__gtk_text_iter_get_chars_in_line"), visibility("default")));

#undef gtk_text_iter_get_child_anchor
extern __typeof (gtk_text_iter_get_child_anchor) gtk_text_iter_get_child_anchor __attribute((alias("IA__gtk_text_iter_get_child_anchor"), visibility("default")));

#undef gtk_text_iter_get_language
extern __typeof (gtk_text_iter_get_language) gtk_text_iter_get_language __attribute((alias("IA__gtk_text_iter_get_language"), visibility("default")));

#undef gtk_text_iter_get_line
extern __typeof (gtk_text_iter_get_line) gtk_text_iter_get_line __attribute((alias("IA__gtk_text_iter_get_line"), visibility("default")));

#undef gtk_text_iter_get_line_index
extern __typeof (gtk_text_iter_get_line_index) gtk_text_iter_get_line_index __attribute((alias("IA__gtk_text_iter_get_line_index"), visibility("default")));

#undef gtk_text_iter_get_line_offset
extern __typeof (gtk_text_iter_get_line_offset) gtk_text_iter_get_line_offset __attribute((alias("IA__gtk_text_iter_get_line_offset"), visibility("default")));

#undef gtk_text_iter_get_marks
extern __typeof (gtk_text_iter_get_marks) gtk_text_iter_get_marks __attribute((alias("IA__gtk_text_iter_get_marks"), visibility("default")));

#undef gtk_text_iter_get_offset
extern __typeof (gtk_text_iter_get_offset) gtk_text_iter_get_offset __attribute((alias("IA__gtk_text_iter_get_offset"), visibility("default")));

#undef gtk_text_iter_get_pixbuf
extern __typeof (gtk_text_iter_get_pixbuf) gtk_text_iter_get_pixbuf __attribute((alias("IA__gtk_text_iter_get_pixbuf"), visibility("default")));

#undef gtk_text_iter_get_slice
extern __typeof (gtk_text_iter_get_slice) gtk_text_iter_get_slice __attribute((alias("IA__gtk_text_iter_get_slice"), visibility("default")));

#undef gtk_text_iter_get_tags
extern __typeof (gtk_text_iter_get_tags) gtk_text_iter_get_tags __attribute((alias("IA__gtk_text_iter_get_tags"), visibility("default")));

#undef gtk_text_iter_get_text
extern __typeof (gtk_text_iter_get_text) gtk_text_iter_get_text __attribute((alias("IA__gtk_text_iter_get_text"), visibility("default")));

#undef gtk_text_iter_get_toggled_tags
extern __typeof (gtk_text_iter_get_toggled_tags) gtk_text_iter_get_toggled_tags __attribute((alias("IA__gtk_text_iter_get_toggled_tags"), visibility("default")));

#undef gtk_text_iter_get_type
extern __typeof (gtk_text_iter_get_type) gtk_text_iter_get_type __attribute((alias("IA__gtk_text_iter_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_text_iter_get_visible_line_index
extern __typeof (gtk_text_iter_get_visible_line_index) gtk_text_iter_get_visible_line_index __attribute((alias("IA__gtk_text_iter_get_visible_line_index"), visibility("default")));

#undef gtk_text_iter_get_visible_line_offset
extern __typeof (gtk_text_iter_get_visible_line_offset) gtk_text_iter_get_visible_line_offset __attribute((alias("IA__gtk_text_iter_get_visible_line_offset"), visibility("default")));

#undef gtk_text_iter_get_visible_slice
extern __typeof (gtk_text_iter_get_visible_slice) gtk_text_iter_get_visible_slice __attribute((alias("IA__gtk_text_iter_get_visible_slice"), visibility("default")));

#undef gtk_text_iter_get_visible_text
extern __typeof (gtk_text_iter_get_visible_text) gtk_text_iter_get_visible_text __attribute((alias("IA__gtk_text_iter_get_visible_text"), visibility("default")));

#undef gtk_text_iter_has_tag
extern __typeof (gtk_text_iter_has_tag) gtk_text_iter_has_tag __attribute((alias("IA__gtk_text_iter_has_tag"), visibility("default")));

#undef gtk_text_iter_in_range
extern __typeof (gtk_text_iter_in_range) gtk_text_iter_in_range __attribute((alias("IA__gtk_text_iter_in_range"), visibility("default")));

#undef gtk_text_iter_inside_sentence
extern __typeof (gtk_text_iter_inside_sentence) gtk_text_iter_inside_sentence __attribute((alias("IA__gtk_text_iter_inside_sentence"), visibility("default")));

#undef gtk_text_iter_inside_word
extern __typeof (gtk_text_iter_inside_word) gtk_text_iter_inside_word __attribute((alias("IA__gtk_text_iter_inside_word"), visibility("default")));

#undef gtk_text_iter_is_cursor_position
extern __typeof (gtk_text_iter_is_cursor_position) gtk_text_iter_is_cursor_position __attribute((alias("IA__gtk_text_iter_is_cursor_position"), visibility("default")));

#undef gtk_text_iter_is_end
extern __typeof (gtk_text_iter_is_end) gtk_text_iter_is_end __attribute((alias("IA__gtk_text_iter_is_end"), visibility("default")));

#undef gtk_text_iter_is_start
extern __typeof (gtk_text_iter_is_start) gtk_text_iter_is_start __attribute((alias("IA__gtk_text_iter_is_start"), visibility("default")));

#undef gtk_text_iter_order
extern __typeof (gtk_text_iter_order) gtk_text_iter_order __attribute((alias("IA__gtk_text_iter_order"), visibility("default")));

#undef gtk_text_iter_set_line
extern __typeof (gtk_text_iter_set_line) gtk_text_iter_set_line __attribute((alias("IA__gtk_text_iter_set_line"), visibility("default")));

#undef gtk_text_iter_set_line_index
extern __typeof (gtk_text_iter_set_line_index) gtk_text_iter_set_line_index __attribute((alias("IA__gtk_text_iter_set_line_index"), visibility("default")));

#undef gtk_text_iter_set_line_offset
extern __typeof (gtk_text_iter_set_line_offset) gtk_text_iter_set_line_offset __attribute((alias("IA__gtk_text_iter_set_line_offset"), visibility("default")));

#undef gtk_text_iter_set_offset
extern __typeof (gtk_text_iter_set_offset) gtk_text_iter_set_offset __attribute((alias("IA__gtk_text_iter_set_offset"), visibility("default")));

#undef gtk_text_iter_set_visible_line_index
extern __typeof (gtk_text_iter_set_visible_line_index) gtk_text_iter_set_visible_line_index __attribute((alias("IA__gtk_text_iter_set_visible_line_index"), visibility("default")));

#undef gtk_text_iter_set_visible_line_offset
extern __typeof (gtk_text_iter_set_visible_line_offset) gtk_text_iter_set_visible_line_offset __attribute((alias("IA__gtk_text_iter_set_visible_line_offset"), visibility("default")));

#undef gtk_text_iter_starts_line
extern __typeof (gtk_text_iter_starts_line) gtk_text_iter_starts_line __attribute((alias("IA__gtk_text_iter_starts_line"), visibility("default")));

#undef gtk_text_iter_starts_sentence
extern __typeof (gtk_text_iter_starts_sentence) gtk_text_iter_starts_sentence __attribute((alias("IA__gtk_text_iter_starts_sentence"), visibility("default")));

#undef gtk_text_iter_starts_word
extern __typeof (gtk_text_iter_starts_word) gtk_text_iter_starts_word __attribute((alias("IA__gtk_text_iter_starts_word"), visibility("default")));

#undef gtk_text_iter_toggles_tag
extern __typeof (gtk_text_iter_toggles_tag) gtk_text_iter_toggles_tag __attribute((alias("IA__gtk_text_iter_toggles_tag"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEXT_LAYOUT_H__)
#if IN_FILE(__GTK_TEXT_LAYOUT_C__)
#undef gtk_text_layout_changed
extern __typeof (gtk_text_layout_changed) gtk_text_layout_changed __attribute((alias("IA__gtk_text_layout_changed"), visibility("default")));

#undef gtk_text_layout_cursors_changed
extern __typeof (gtk_text_layout_cursors_changed) gtk_text_layout_cursors_changed __attribute((alias("IA__gtk_text_layout_cursors_changed"), visibility("default")));

#undef gtk_text_layout_clamp_iter_to_vrange
extern __typeof (gtk_text_layout_clamp_iter_to_vrange) gtk_text_layout_clamp_iter_to_vrange __attribute((alias("IA__gtk_text_layout_clamp_iter_to_vrange"), visibility("default")));

#undef gtk_text_layout_default_style_changed
extern __typeof (gtk_text_layout_default_style_changed) gtk_text_layout_default_style_changed __attribute((alias("IA__gtk_text_layout_default_style_changed"), visibility("default")));

#undef gtk_text_layout_free_line_data
extern __typeof (gtk_text_layout_free_line_data) gtk_text_layout_free_line_data __attribute((alias("IA__gtk_text_layout_free_line_data"), visibility("default")));

#undef gtk_text_layout_free_line_display
extern __typeof (gtk_text_layout_free_line_display) gtk_text_layout_free_line_display __attribute((alias("IA__gtk_text_layout_free_line_display"), visibility("default")));

#undef gtk_text_layout_get_buffer
extern __typeof (gtk_text_layout_get_buffer) gtk_text_layout_get_buffer __attribute((alias("IA__gtk_text_layout_get_buffer"), visibility("default")));

#undef gtk_text_layout_get_cursor_locations
extern __typeof (gtk_text_layout_get_cursor_locations) gtk_text_layout_get_cursor_locations __attribute((alias("IA__gtk_text_layout_get_cursor_locations"), visibility("default")));

#undef gtk_text_layout_get_cursor_visible
extern __typeof (gtk_text_layout_get_cursor_visible) gtk_text_layout_get_cursor_visible __attribute((alias("IA__gtk_text_layout_get_cursor_visible"), visibility("default")));

#undef gtk_text_layout_get_iter_at_line
extern __typeof (gtk_text_layout_get_iter_at_line) gtk_text_layout_get_iter_at_line __attribute((alias("IA__gtk_text_layout_get_iter_at_line"), visibility("default")));

#undef gtk_text_layout_get_iter_at_pixel
extern __typeof (gtk_text_layout_get_iter_at_pixel) gtk_text_layout_get_iter_at_pixel __attribute((alias("IA__gtk_text_layout_get_iter_at_pixel"), visibility("default")));

#undef gtk_text_layout_get_iter_at_position
extern __typeof (gtk_text_layout_get_iter_at_position) gtk_text_layout_get_iter_at_position __attribute((alias("IA__gtk_text_layout_get_iter_at_position"), visibility("default")));

#undef gtk_text_layout_get_iter_location
extern __typeof (gtk_text_layout_get_iter_location) gtk_text_layout_get_iter_location __attribute((alias("IA__gtk_text_layout_get_iter_location"), visibility("default")));

#undef gtk_text_layout_get_line_at_y
extern __typeof (gtk_text_layout_get_line_at_y) gtk_text_layout_get_line_at_y __attribute((alias("IA__gtk_text_layout_get_line_at_y"), visibility("default")));

#undef gtk_text_layout_get_line_display
extern __typeof (gtk_text_layout_get_line_display) gtk_text_layout_get_line_display __attribute((alias("IA__gtk_text_layout_get_line_display"), visibility("default")));

#undef gtk_text_layout_get_lines
extern __typeof (gtk_text_layout_get_lines) gtk_text_layout_get_lines __attribute((alias("IA__gtk_text_layout_get_lines"), visibility("default")));

#undef gtk_text_layout_get_line_yrange
extern __typeof (gtk_text_layout_get_line_yrange) gtk_text_layout_get_line_yrange __attribute((alias("IA__gtk_text_layout_get_line_yrange"), visibility("default")));

#undef gtk_text_layout_get_size
extern __typeof (gtk_text_layout_get_size) gtk_text_layout_get_size __attribute((alias("IA__gtk_text_layout_get_size"), visibility("default")));

#undef gtk_text_layout_get_type
extern __typeof (gtk_text_layout_get_type) gtk_text_layout_get_type __attribute((alias("IA__gtk_text_layout_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_text_layout_invalidate
extern __typeof (gtk_text_layout_invalidate) gtk_text_layout_invalidate __attribute((alias("IA__gtk_text_layout_invalidate"), visibility("default")));

#undef gtk_text_layout_invalidate_cursors
extern __typeof (gtk_text_layout_invalidate_cursors) gtk_text_layout_invalidate_cursors __attribute((alias("IA__gtk_text_layout_invalidate_cursors"), visibility("default")));

#undef gtk_text_layout_is_valid
extern __typeof (gtk_text_layout_is_valid) gtk_text_layout_is_valid __attribute((alias("IA__gtk_text_layout_is_valid"), visibility("default")));

#undef gtk_text_layout_iter_starts_line
extern __typeof (gtk_text_layout_iter_starts_line) gtk_text_layout_iter_starts_line __attribute((alias("IA__gtk_text_layout_iter_starts_line"), visibility("default")));

#undef gtk_text_layout_move_iter_to_line_end
extern __typeof (gtk_text_layout_move_iter_to_line_end) gtk_text_layout_move_iter_to_line_end __attribute((alias("IA__gtk_text_layout_move_iter_to_line_end"), visibility("default")));

#undef gtk_text_layout_move_iter_to_next_line
extern __typeof (gtk_text_layout_move_iter_to_next_line) gtk_text_layout_move_iter_to_next_line __attribute((alias("IA__gtk_text_layout_move_iter_to_next_line"), visibility("default")));

#undef gtk_text_layout_move_iter_to_previous_line
extern __typeof (gtk_text_layout_move_iter_to_previous_line) gtk_text_layout_move_iter_to_previous_line __attribute((alias("IA__gtk_text_layout_move_iter_to_previous_line"), visibility("default")));

#undef gtk_text_layout_move_iter_to_x
extern __typeof (gtk_text_layout_move_iter_to_x) gtk_text_layout_move_iter_to_x __attribute((alias("IA__gtk_text_layout_move_iter_to_x"), visibility("default")));

#undef gtk_text_layout_move_iter_visually
extern __typeof (gtk_text_layout_move_iter_visually) gtk_text_layout_move_iter_visually __attribute((alias("IA__gtk_text_layout_move_iter_visually"), visibility("default")));

#undef gtk_text_layout_new
extern __typeof (gtk_text_layout_new) gtk_text_layout_new __attribute((alias("IA__gtk_text_layout_new"), visibility("default")));

#undef gtk_text_layout_set_buffer
extern __typeof (gtk_text_layout_set_buffer) gtk_text_layout_set_buffer __attribute((alias("IA__gtk_text_layout_set_buffer"), visibility("default")));

#undef gtk_text_layout_set_contexts
extern __typeof (gtk_text_layout_set_contexts) gtk_text_layout_set_contexts __attribute((alias("IA__gtk_text_layout_set_contexts"), visibility("default")));

#undef gtk_text_layout_set_cursor_direction
extern __typeof (gtk_text_layout_set_cursor_direction) gtk_text_layout_set_cursor_direction __attribute((alias("IA__gtk_text_layout_set_cursor_direction"), visibility("default")));

#undef gtk_text_layout_set_cursor_visible
extern __typeof (gtk_text_layout_set_cursor_visible) gtk_text_layout_set_cursor_visible __attribute((alias("IA__gtk_text_layout_set_cursor_visible"), visibility("default")));

#undef gtk_text_layout_set_default_style
extern __typeof (gtk_text_layout_set_default_style) gtk_text_layout_set_default_style __attribute((alias("IA__gtk_text_layout_set_default_style"), visibility("default")));

#undef gtk_text_layout_set_keyboard_direction
extern __typeof (gtk_text_layout_set_keyboard_direction) gtk_text_layout_set_keyboard_direction __attribute((alias("IA__gtk_text_layout_set_keyboard_direction"), visibility("default")));

#undef gtk_text_layout_set_overwrite_mode
extern __typeof (gtk_text_layout_set_overwrite_mode) gtk_text_layout_set_overwrite_mode __attribute((alias("IA__gtk_text_layout_set_overwrite_mode"), visibility("default")));

#undef gtk_text_layout_set_preedit_string
extern __typeof (gtk_text_layout_set_preedit_string) gtk_text_layout_set_preedit_string __attribute((alias("IA__gtk_text_layout_set_preedit_string"), visibility("default")));

#undef gtk_text_layout_set_screen_width
extern __typeof (gtk_text_layout_set_screen_width) gtk_text_layout_set_screen_width __attribute((alias("IA__gtk_text_layout_set_screen_width"), visibility("default")));

#undef gtk_text_layout_spew
extern __typeof (gtk_text_layout_spew) gtk_text_layout_spew __attribute((alias("IA__gtk_text_layout_spew"), visibility("default")));

#undef gtk_text_layout_validate
extern __typeof (gtk_text_layout_validate) gtk_text_layout_validate __attribute((alias("IA__gtk_text_layout_validate"), visibility("default")));

#undef gtk_text_layout_validate_yrange
extern __typeof (gtk_text_layout_validate_yrange) gtk_text_layout_validate_yrange __attribute((alias("IA__gtk_text_layout_validate_yrange"), visibility("default")));

#undef gtk_text_layout_wrap
extern __typeof (gtk_text_layout_wrap) gtk_text_layout_wrap __attribute((alias("IA__gtk_text_layout_wrap"), visibility("default")));

#undef gtk_text_layout_wrap_loop_end
extern __typeof (gtk_text_layout_wrap_loop_end) gtk_text_layout_wrap_loop_end __attribute((alias("IA__gtk_text_layout_wrap_loop_end"), visibility("default")));

#undef gtk_text_layout_wrap_loop_start
extern __typeof (gtk_text_layout_wrap_loop_start) gtk_text_layout_wrap_loop_start __attribute((alias("IA__gtk_text_layout_wrap_loop_start"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEXT_CHILD_H__)
#if IN_FILE(__GTK_TEXT_CHILD_C__)
#undef gtk_text_child_anchor_get_deleted
extern __typeof (gtk_text_child_anchor_get_deleted) gtk_text_child_anchor_get_deleted __attribute((alias("IA__gtk_text_child_anchor_get_deleted"), visibility("default")));

#undef gtk_text_child_anchor_get_type
extern __typeof (gtk_text_child_anchor_get_type) gtk_text_child_anchor_get_type __attribute((alias("IA__gtk_text_child_anchor_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_text_child_anchor_get_widgets
extern __typeof (gtk_text_child_anchor_get_widgets) gtk_text_child_anchor_get_widgets __attribute((alias("IA__gtk_text_child_anchor_get_widgets"), visibility("default")));

#undef gtk_text_child_anchor_new
extern __typeof (gtk_text_child_anchor_new) gtk_text_child_anchor_new __attribute((alias("IA__gtk_text_child_anchor_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEXT_LAYOUT_H__)
#if IN_FILE(__GTK_TEXT_CHILD_C__)
#undef gtk_text_child_anchor_queue_resize
extern __typeof (gtk_text_child_anchor_queue_resize) gtk_text_child_anchor_queue_resize __attribute((alias("IA__gtk_text_child_anchor_queue_resize"), visibility("default")));

#undef gtk_text_child_anchor_register_child
extern __typeof (gtk_text_child_anchor_register_child) gtk_text_child_anchor_register_child __attribute((alias("IA__gtk_text_child_anchor_register_child"), visibility("default")));

#undef gtk_text_child_anchor_unregister_child
extern __typeof (gtk_text_child_anchor_unregister_child) gtk_text_child_anchor_unregister_child __attribute((alias("IA__gtk_text_child_anchor_unregister_child"), visibility("default")));

#undef gtk_text_anchored_child_set_layout
extern __typeof (gtk_text_anchored_child_set_layout) gtk_text_anchored_child_set_layout __attribute((alias("IA__gtk_text_anchored_child_set_layout"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEXT_DISPLAY_H__)
#if IN_FILE(__GTK_TEXT_DISPLAY_C__)
#undef gtk_text_layout_draw
extern __typeof (gtk_text_layout_draw) gtk_text_layout_draw __attribute((alias("IA__gtk_text_layout_draw"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEXT_SEGMENT_H__)
#if IN_FILE(__GTK_TEXT_SEGMENT_C__)
#undef gtk_text_line_segment_split
extern __typeof (gtk_text_line_segment_split) gtk_text_line_segment_split __attribute((alias("IA__gtk_text_line_segment_split"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEXT_MARK_H__)
#if IN_FILE(__GTK_TEXT_MARK_C__)
#undef gtk_text_mark_new
extern __typeof (gtk_text_mark_new) gtk_text_mark_new __attribute((alias("IA__gtk_text_mark_new"), visibility("default")));

#undef gtk_text_mark_get_buffer
extern __typeof (gtk_text_mark_get_buffer) gtk_text_mark_get_buffer __attribute((alias("IA__gtk_text_mark_get_buffer"), visibility("default")));

#undef gtk_text_mark_get_deleted
extern __typeof (gtk_text_mark_get_deleted) gtk_text_mark_get_deleted __attribute((alias("IA__gtk_text_mark_get_deleted"), visibility("default")));

#undef gtk_text_mark_get_left_gravity
extern __typeof (gtk_text_mark_get_left_gravity) gtk_text_mark_get_left_gravity __attribute((alias("IA__gtk_text_mark_get_left_gravity"), visibility("default")));

#undef gtk_text_mark_get_name
extern __typeof (gtk_text_mark_get_name) gtk_text_mark_get_name __attribute((alias("IA__gtk_text_mark_get_name"), visibility("default")));

#undef gtk_text_mark_get_type
extern __typeof (gtk_text_mark_get_type) gtk_text_mark_get_type __attribute((alias("IA__gtk_text_mark_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_text_mark_get_visible
extern __typeof (gtk_text_mark_get_visible) gtk_text_mark_get_visible __attribute((alias("IA__gtk_text_mark_get_visible"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEXT_MARK_H__)
#if IN_FILE(__GTK_TEXT_BTREE_C__)
#undef gtk_text_mark_set_visible
extern __typeof (gtk_text_mark_set_visible) gtk_text_mark_set_visible __attribute((alias("IA__gtk_text_mark_set_visible"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TEXT_VIEW_H__)
#if IN_FILE(__GTK_TEXT_VIEW_C__)
#undef gtk_text_view_add_child_at_anchor
extern __typeof (gtk_text_view_add_child_at_anchor) gtk_text_view_add_child_at_anchor __attribute((alias("IA__gtk_text_view_add_child_at_anchor"), visibility("default")));

#undef gtk_text_view_add_child_in_window
extern __typeof (gtk_text_view_add_child_in_window) gtk_text_view_add_child_in_window __attribute((alias("IA__gtk_text_view_add_child_in_window"), visibility("default")));

#undef gtk_text_view_backward_display_line
extern __typeof (gtk_text_view_backward_display_line) gtk_text_view_backward_display_line __attribute((alias("IA__gtk_text_view_backward_display_line"), visibility("default")));

#undef gtk_text_view_backward_display_line_start
extern __typeof (gtk_text_view_backward_display_line_start) gtk_text_view_backward_display_line_start __attribute((alias("IA__gtk_text_view_backward_display_line_start"), visibility("default")));

#undef gtk_text_view_buffer_to_window_coords
extern __typeof (gtk_text_view_buffer_to_window_coords) gtk_text_view_buffer_to_window_coords __attribute((alias("IA__gtk_text_view_buffer_to_window_coords"), visibility("default")));

#undef gtk_text_view_forward_display_line
extern __typeof (gtk_text_view_forward_display_line) gtk_text_view_forward_display_line __attribute((alias("IA__gtk_text_view_forward_display_line"), visibility("default")));

#undef gtk_text_view_forward_display_line_end
extern __typeof (gtk_text_view_forward_display_line_end) gtk_text_view_forward_display_line_end __attribute((alias("IA__gtk_text_view_forward_display_line_end"), visibility("default")));

#undef gtk_text_view_get_accepts_tab
extern __typeof (gtk_text_view_get_accepts_tab) gtk_text_view_get_accepts_tab __attribute((alias("IA__gtk_text_view_get_accepts_tab"), visibility("default")));

#undef gtk_text_view_get_border_window_size
extern __typeof (gtk_text_view_get_border_window_size) gtk_text_view_get_border_window_size __attribute((alias("IA__gtk_text_view_get_border_window_size"), visibility("default")));

#undef gtk_text_view_get_buffer
extern __typeof (gtk_text_view_get_buffer) gtk_text_view_get_buffer __attribute((alias("IA__gtk_text_view_get_buffer"), visibility("default")));

#undef gtk_text_view_get_cursor_visible
extern __typeof (gtk_text_view_get_cursor_visible) gtk_text_view_get_cursor_visible __attribute((alias("IA__gtk_text_view_get_cursor_visible"), visibility("default")));

#undef gtk_text_view_get_default_attributes
extern __typeof (gtk_text_view_get_default_attributes) gtk_text_view_get_default_attributes __attribute((alias("IA__gtk_text_view_get_default_attributes"), visibility("default")));

#undef gtk_text_view_get_editable
extern __typeof (gtk_text_view_get_editable) gtk_text_view_get_editable __attribute((alias("IA__gtk_text_view_get_editable"), visibility("default")));

#undef gtk_text_view_get_hadjustment
extern __typeof (gtk_text_view_get_hadjustment) gtk_text_view_get_hadjustment __attribute((alias("IA__gtk_text_view_get_hadjustment"), visibility("default")));

#undef gtk_text_view_get_indent
extern __typeof (gtk_text_view_get_indent) gtk_text_view_get_indent __attribute((alias("IA__gtk_text_view_get_indent"), visibility("default")));

#undef gtk_text_view_get_iter_at_location
extern __typeof (gtk_text_view_get_iter_at_location) gtk_text_view_get_iter_at_location __attribute((alias("IA__gtk_text_view_get_iter_at_location"), visibility("default")));

#undef gtk_text_view_get_iter_at_position
extern __typeof (gtk_text_view_get_iter_at_position) gtk_text_view_get_iter_at_position __attribute((alias("IA__gtk_text_view_get_iter_at_position"), visibility("default")));

#undef gtk_text_view_get_iter_location
extern __typeof (gtk_text_view_get_iter_location) gtk_text_view_get_iter_location __attribute((alias("IA__gtk_text_view_get_iter_location"), visibility("default")));

#undef gtk_text_view_get_justification
extern __typeof (gtk_text_view_get_justification) gtk_text_view_get_justification __attribute((alias("IA__gtk_text_view_get_justification"), visibility("default")));

#undef gtk_text_view_get_left_margin
extern __typeof (gtk_text_view_get_left_margin) gtk_text_view_get_left_margin __attribute((alias("IA__gtk_text_view_get_left_margin"), visibility("default")));

#undef gtk_text_view_get_line_at_y
extern __typeof (gtk_text_view_get_line_at_y) gtk_text_view_get_line_at_y __attribute((alias("IA__gtk_text_view_get_line_at_y"), visibility("default")));

#undef gtk_text_view_get_line_yrange
extern __typeof (gtk_text_view_get_line_yrange) gtk_text_view_get_line_yrange __attribute((alias("IA__gtk_text_view_get_line_yrange"), visibility("default")));

#undef gtk_text_view_get_overwrite
extern __typeof (gtk_text_view_get_overwrite) gtk_text_view_get_overwrite __attribute((alias("IA__gtk_text_view_get_overwrite"), visibility("default")));

#undef gtk_text_view_get_pixels_above_lines
extern __typeof (gtk_text_view_get_pixels_above_lines) gtk_text_view_get_pixels_above_lines __attribute((alias("IA__gtk_text_view_get_pixels_above_lines"), visibility("default")));

#undef gtk_text_view_get_pixels_below_lines
extern __typeof (gtk_text_view_get_pixels_below_lines) gtk_text_view_get_pixels_below_lines __attribute((alias("IA__gtk_text_view_get_pixels_below_lines"), visibility("default")));

#undef gtk_text_view_get_pixels_inside_wrap
extern __typeof (gtk_text_view_get_pixels_inside_wrap) gtk_text_view_get_pixels_inside_wrap __attribute((alias("IA__gtk_text_view_get_pixels_inside_wrap"), visibility("default")));

#undef gtk_text_view_get_right_margin
extern __typeof (gtk_text_view_get_right_margin) gtk_text_view_get_right_margin __attribute((alias("IA__gtk_text_view_get_right_margin"), visibility("default")));

#undef gtk_text_view_get_tabs
extern __typeof (gtk_text_view_get_tabs) gtk_text_view_get_tabs __attribute((alias("IA__gtk_text_view_get_tabs"), visibility("default")));

#undef gtk_text_view_get_type
extern __typeof (gtk_text_view_get_type) gtk_text_view_get_type __attribute((alias("IA__gtk_text_view_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_text_view_get_vadjustment
extern __typeof (gtk_text_view_get_vadjustment) gtk_text_view_get_vadjustment __attribute((alias("IA__gtk_text_view_get_vadjustment"), visibility("default")));

#undef gtk_text_view_get_visible_rect
extern __typeof (gtk_text_view_get_visible_rect) gtk_text_view_get_visible_rect __attribute((alias("IA__gtk_text_view_get_visible_rect"), visibility("default")));

#undef gtk_text_view_get_window
extern __typeof (gtk_text_view_get_window) gtk_text_view_get_window __attribute((alias("IA__gtk_text_view_get_window"), visibility("default")));

#undef gtk_text_view_get_window_type
extern __typeof (gtk_text_view_get_window_type) gtk_text_view_get_window_type __attribute((alias("IA__gtk_text_view_get_window_type"), visibility("default")));

#undef gtk_text_view_get_wrap_mode
extern __typeof (gtk_text_view_get_wrap_mode) gtk_text_view_get_wrap_mode __attribute((alias("IA__gtk_text_view_get_wrap_mode"), visibility("default")));

#undef gtk_text_view_im_context_filter_keypress
extern __typeof (gtk_text_view_im_context_filter_keypress) gtk_text_view_im_context_filter_keypress __attribute((alias("IA__gtk_text_view_im_context_filter_keypress"), visibility("default")));

#undef gtk_text_view_move_child
extern __typeof (gtk_text_view_move_child) gtk_text_view_move_child __attribute((alias("IA__gtk_text_view_move_child"), visibility("default")));

#undef gtk_text_view_move_mark_onscreen
extern __typeof (gtk_text_view_move_mark_onscreen) gtk_text_view_move_mark_onscreen __attribute((alias("IA__gtk_text_view_move_mark_onscreen"), visibility("default")));

#undef gtk_text_view_move_visually
extern __typeof (gtk_text_view_move_visually) gtk_text_view_move_visually __attribute((alias("IA__gtk_text_view_move_visually"), visibility("default")));

#undef gtk_text_view_new
extern __typeof (gtk_text_view_new) gtk_text_view_new __attribute((alias("IA__gtk_text_view_new"), visibility("default")));

#undef gtk_text_view_new_with_buffer
extern __typeof (gtk_text_view_new_with_buffer) gtk_text_view_new_with_buffer __attribute((alias("IA__gtk_text_view_new_with_buffer"), visibility("default")));

#undef gtk_text_view_place_cursor_onscreen
extern __typeof (gtk_text_view_place_cursor_onscreen) gtk_text_view_place_cursor_onscreen __attribute((alias("IA__gtk_text_view_place_cursor_onscreen"), visibility("default")));

#undef gtk_text_view_reset_im_context
extern __typeof (gtk_text_view_reset_im_context) gtk_text_view_reset_im_context __attribute((alias("IA__gtk_text_view_reset_im_context"), visibility("default")));

#undef gtk_text_view_scroll_mark_onscreen
extern __typeof (gtk_text_view_scroll_mark_onscreen) gtk_text_view_scroll_mark_onscreen __attribute((alias("IA__gtk_text_view_scroll_mark_onscreen"), visibility("default")));

#undef gtk_text_view_scroll_to_iter
extern __typeof (gtk_text_view_scroll_to_iter) gtk_text_view_scroll_to_iter __attribute((alias("IA__gtk_text_view_scroll_to_iter"), visibility("default")));

#undef gtk_text_view_scroll_to_mark
extern __typeof (gtk_text_view_scroll_to_mark) gtk_text_view_scroll_to_mark __attribute((alias("IA__gtk_text_view_scroll_to_mark"), visibility("default")));

#undef gtk_text_view_set_accepts_tab
extern __typeof (gtk_text_view_set_accepts_tab) gtk_text_view_set_accepts_tab __attribute((alias("IA__gtk_text_view_set_accepts_tab"), visibility("default")));

#undef gtk_text_view_set_border_window_size
extern __typeof (gtk_text_view_set_border_window_size) gtk_text_view_set_border_window_size __attribute((alias("IA__gtk_text_view_set_border_window_size"), visibility("default")));

#undef gtk_text_view_set_buffer
extern __typeof (gtk_text_view_set_buffer) gtk_text_view_set_buffer __attribute((alias("IA__gtk_text_view_set_buffer"), visibility("default")));

#undef gtk_text_view_set_cursor_visible
extern __typeof (gtk_text_view_set_cursor_visible) gtk_text_view_set_cursor_visible __attribute((alias("IA__gtk_text_view_set_cursor_visible"), visibility("default")));

#undef gtk_text_view_set_editable
extern __typeof (gtk_text_view_set_editable) gtk_text_view_set_editable __attribute((alias("IA__gtk_text_view_set_editable"), visibility("default")));

#undef gtk_text_view_set_indent
extern __typeof (gtk_text_view_set_indent) gtk_text_view_set_indent __attribute((alias("IA__gtk_text_view_set_indent"), visibility("default")));

#undef gtk_text_view_set_justification
extern __typeof (gtk_text_view_set_justification) gtk_text_view_set_justification __attribute((alias("IA__gtk_text_view_set_justification"), visibility("default")));

#undef gtk_text_view_set_left_margin
extern __typeof (gtk_text_view_set_left_margin) gtk_text_view_set_left_margin __attribute((alias("IA__gtk_text_view_set_left_margin"), visibility("default")));

#undef gtk_text_view_set_overwrite
extern __typeof (gtk_text_view_set_overwrite) gtk_text_view_set_overwrite __attribute((alias("IA__gtk_text_view_set_overwrite"), visibility("default")));

#undef gtk_text_view_set_pixels_above_lines
extern __typeof (gtk_text_view_set_pixels_above_lines) gtk_text_view_set_pixels_above_lines __attribute((alias("IA__gtk_text_view_set_pixels_above_lines"), visibility("default")));

#undef gtk_text_view_set_pixels_below_lines
extern __typeof (gtk_text_view_set_pixels_below_lines) gtk_text_view_set_pixels_below_lines __attribute((alias("IA__gtk_text_view_set_pixels_below_lines"), visibility("default")));

#undef gtk_text_view_set_pixels_inside_wrap
extern __typeof (gtk_text_view_set_pixels_inside_wrap) gtk_text_view_set_pixels_inside_wrap __attribute((alias("IA__gtk_text_view_set_pixels_inside_wrap"), visibility("default")));

#undef gtk_text_view_set_right_margin
extern __typeof (gtk_text_view_set_right_margin) gtk_text_view_set_right_margin __attribute((alias("IA__gtk_text_view_set_right_margin"), visibility("default")));

#undef gtk_text_view_set_tabs
extern __typeof (gtk_text_view_set_tabs) gtk_text_view_set_tabs __attribute((alias("IA__gtk_text_view_set_tabs"), visibility("default")));

#undef gtk_text_view_set_wrap_mode
extern __typeof (gtk_text_view_set_wrap_mode) gtk_text_view_set_wrap_mode __attribute((alias("IA__gtk_text_view_set_wrap_mode"), visibility("default")));

#undef gtk_text_view_starts_display_line
extern __typeof (gtk_text_view_starts_display_line) gtk_text_view_starts_display_line __attribute((alias("IA__gtk_text_view_starts_display_line"), visibility("default")));

#undef gtk_text_view_window_to_buffer_coords
extern __typeof (gtk_text_view_window_to_buffer_coords) gtk_text_view_window_to_buffer_coords __attribute((alias("IA__gtk_text_view_window_to_buffer_coords"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_THEMES_H__)
#if IN_FILE(__GTK_THEMES_C__)
#undef gtk_theme_engine_create_rc_style
extern __typeof (gtk_theme_engine_create_rc_style) gtk_theme_engine_create_rc_style __attribute((alias("IA__gtk_theme_engine_create_rc_style"), visibility("default")));

#undef gtk_theme_engine_get
extern __typeof (gtk_theme_engine_get) gtk_theme_engine_get __attribute((alias("IA__gtk_theme_engine_get"), visibility("default")));

#undef gtk_theme_engine_get_type
extern __typeof (gtk_theme_engine_get_type) gtk_theme_engine_get_type __attribute((alias("IA__gtk_theme_engine_get_type"), visibility("default"))) G_GNUC_CONST;

#endif
#endif
#if IN_HEADER(__GTK_TIPS_QUERY_H__)
#if IN_FILE(__GTK_TIPS_QUERY_C__)
#undef gtk_tips_query_get_type
extern __typeof (gtk_tips_query_get_type) gtk_tips_query_get_type __attribute((alias("IA__gtk_tips_query_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tips_query_new
extern __typeof (gtk_tips_query_new) gtk_tips_query_new __attribute((alias("IA__gtk_tips_query_new"), visibility("default")));

#undef gtk_tips_query_set_caller
extern __typeof (gtk_tips_query_set_caller) gtk_tips_query_set_caller __attribute((alias("IA__gtk_tips_query_set_caller"), visibility("default")));

#undef gtk_tips_query_set_labels
extern __typeof (gtk_tips_query_set_labels) gtk_tips_query_set_labels __attribute((alias("IA__gtk_tips_query_set_labels"), visibility("default")));

#undef gtk_tips_query_start_query
extern __typeof (gtk_tips_query_start_query) gtk_tips_query_start_query __attribute((alias("IA__gtk_tips_query_start_query"), visibility("default")));

#undef gtk_tips_query_stop_query
extern __typeof (gtk_tips_query_stop_query) gtk_tips_query_stop_query __attribute((alias("IA__gtk_tips_query_stop_query"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TOGGLE_ACTION_H__)
#if IN_FILE(__GTK_TOGGLE_ACTION_C__)
#undef gtk_toggle_action_get_active
extern __typeof (gtk_toggle_action_get_active) gtk_toggle_action_get_active __attribute((alias("IA__gtk_toggle_action_get_active"), visibility("default")));

#undef gtk_toggle_action_get_draw_as_radio
extern __typeof (gtk_toggle_action_get_draw_as_radio) gtk_toggle_action_get_draw_as_radio __attribute((alias("IA__gtk_toggle_action_get_draw_as_radio"), visibility("default")));

#undef gtk_toggle_action_get_type
extern __typeof (gtk_toggle_action_get_type) gtk_toggle_action_get_type __attribute((alias("IA__gtk_toggle_action_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_toggle_action_new
extern __typeof (gtk_toggle_action_new) gtk_toggle_action_new __attribute((alias("IA__gtk_toggle_action_new"), visibility("default")));

#undef gtk_toggle_action_set_active
extern __typeof (gtk_toggle_action_set_active) gtk_toggle_action_set_active __attribute((alias("IA__gtk_toggle_action_set_active"), visibility("default")));

#undef gtk_toggle_action_set_draw_as_radio
extern __typeof (gtk_toggle_action_set_draw_as_radio) gtk_toggle_action_set_draw_as_radio __attribute((alias("IA__gtk_toggle_action_set_draw_as_radio"), visibility("default")));

#undef gtk_toggle_action_toggled
extern __typeof (gtk_toggle_action_toggled) gtk_toggle_action_toggled __attribute((alias("IA__gtk_toggle_action_toggled"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TOGGLE_BUTTON_H__)
#if IN_FILE(__GTK_TOGGLE_BUTTON_C__)
#undef gtk_toggle_button_get_active
extern __typeof (gtk_toggle_button_get_active) gtk_toggle_button_get_active __attribute((alias("IA__gtk_toggle_button_get_active"), visibility("default")));

#undef gtk_toggle_button_get_inconsistent
extern __typeof (gtk_toggle_button_get_inconsistent) gtk_toggle_button_get_inconsistent __attribute((alias("IA__gtk_toggle_button_get_inconsistent"), visibility("default")));

#undef gtk_toggle_button_get_mode
extern __typeof (gtk_toggle_button_get_mode) gtk_toggle_button_get_mode __attribute((alias("IA__gtk_toggle_button_get_mode"), visibility("default")));

#undef gtk_toggle_button_get_type
extern __typeof (gtk_toggle_button_get_type) gtk_toggle_button_get_type __attribute((alias("IA__gtk_toggle_button_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_toggle_button_new
extern __typeof (gtk_toggle_button_new) gtk_toggle_button_new __attribute((alias("IA__gtk_toggle_button_new"), visibility("default")));

#undef gtk_toggle_button_new_with_label
extern __typeof (gtk_toggle_button_new_with_label) gtk_toggle_button_new_with_label __attribute((alias("IA__gtk_toggle_button_new_with_label"), visibility("default")));

#undef gtk_toggle_button_new_with_mnemonic
extern __typeof (gtk_toggle_button_new_with_mnemonic) gtk_toggle_button_new_with_mnemonic __attribute((alias("IA__gtk_toggle_button_new_with_mnemonic"), visibility("default")));

#undef gtk_toggle_button_set_active
extern __typeof (gtk_toggle_button_set_active) gtk_toggle_button_set_active __attribute((alias("IA__gtk_toggle_button_set_active"), visibility("default")));

#undef gtk_toggle_button_set_inconsistent
extern __typeof (gtk_toggle_button_set_inconsistent) gtk_toggle_button_set_inconsistent __attribute((alias("IA__gtk_toggle_button_set_inconsistent"), visibility("default")));

#undef gtk_toggle_button_set_mode
extern __typeof (gtk_toggle_button_set_mode) gtk_toggle_button_set_mode __attribute((alias("IA__gtk_toggle_button_set_mode"), visibility("default")));

#undef gtk_toggle_button_toggled
extern __typeof (gtk_toggle_button_toggled) gtk_toggle_button_toggled __attribute((alias("IA__gtk_toggle_button_toggled"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TOGGLE_TOOL_BUTTON_H__)
#if IN_FILE(__GTK_TOGGLE_TOOL_BUTTON_C__)
#undef gtk_toggle_tool_button_get_active
extern __typeof (gtk_toggle_tool_button_get_active) gtk_toggle_tool_button_get_active __attribute((alias("IA__gtk_toggle_tool_button_get_active"), visibility("default")));

#undef gtk_toggle_tool_button_get_type
extern __typeof (gtk_toggle_tool_button_get_type) gtk_toggle_tool_button_get_type __attribute((alias("IA__gtk_toggle_tool_button_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_toggle_tool_button_new
extern __typeof (gtk_toggle_tool_button_new) gtk_toggle_tool_button_new __attribute((alias("IA__gtk_toggle_tool_button_new"), visibility("default")));

#undef gtk_toggle_tool_button_new_from_stock
extern __typeof (gtk_toggle_tool_button_new_from_stock) gtk_toggle_tool_button_new_from_stock __attribute((alias("IA__gtk_toggle_tool_button_new_from_stock"), visibility("default")));

#undef gtk_toggle_tool_button_set_active
extern __typeof (gtk_toggle_tool_button_set_active) gtk_toggle_tool_button_set_active __attribute((alias("IA__gtk_toggle_tool_button_set_active"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TOOLBAR_H__)
#if IN_FILE(__GTK_TOOLBAR_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_toolbar_append_element
extern __typeof (gtk_toolbar_append_element) gtk_toolbar_append_element __attribute((alias("IA__gtk_toolbar_append_element"), visibility("default")));

#undef gtk_toolbar_set_icon_size
extern __typeof (gtk_toolbar_set_icon_size) gtk_toolbar_set_icon_size __attribute((alias("IA__gtk_toolbar_set_icon_size"), visibility("default")));

#undef gtk_toolbar_unset_icon_size
extern __typeof (gtk_toolbar_unset_icon_size) gtk_toolbar_unset_icon_size __attribute((alias("IA__gtk_toolbar_unset_icon_size"), visibility("default")));

#undef gtk_toolbar_append_item
extern __typeof (gtk_toolbar_append_item) gtk_toolbar_append_item __attribute((alias("IA__gtk_toolbar_append_item"), visibility("default")));

#undef gtk_toolbar_prepend_element
extern __typeof (gtk_toolbar_prepend_element) gtk_toolbar_prepend_element __attribute((alias("IA__gtk_toolbar_prepend_element"), visibility("default")));

#undef gtk_toolbar_prepend_item
extern __typeof (gtk_toolbar_prepend_item) gtk_toolbar_prepend_item __attribute((alias("IA__gtk_toolbar_prepend_item"), visibility("default")));

#undef gtk_toolbar_insert_element
extern __typeof (gtk_toolbar_insert_element) gtk_toolbar_insert_element __attribute((alias("IA__gtk_toolbar_insert_element"), visibility("default")));

#undef gtk_toolbar_insert_item
extern __typeof (gtk_toolbar_insert_item) gtk_toolbar_insert_item __attribute((alias("IA__gtk_toolbar_insert_item"), visibility("default")));

#undef gtk_toolbar_insert_space
extern __typeof (gtk_toolbar_insert_space) gtk_toolbar_insert_space __attribute((alias("IA__gtk_toolbar_insert_space"), visibility("default")));

#undef gtk_toolbar_insert_stock
extern __typeof (gtk_toolbar_insert_stock) gtk_toolbar_insert_stock __attribute((alias("IA__gtk_toolbar_insert_stock"), visibility("default")));

#undef gtk_toolbar_append_space
extern __typeof (gtk_toolbar_append_space) gtk_toolbar_append_space __attribute((alias("IA__gtk_toolbar_append_space"), visibility("default")));

#undef gtk_toolbar_prepend_space
extern __typeof (gtk_toolbar_prepend_space) gtk_toolbar_prepend_space __attribute((alias("IA__gtk_toolbar_prepend_space"), visibility("default")));

#undef gtk_toolbar_remove_space
extern __typeof (gtk_toolbar_remove_space) gtk_toolbar_remove_space __attribute((alias("IA__gtk_toolbar_remove_space"), visibility("default")));

#undef gtk_toolbar_append_widget
extern __typeof (gtk_toolbar_append_widget) gtk_toolbar_append_widget __attribute((alias("IA__gtk_toolbar_append_widget"), visibility("default")));

#undef gtk_toolbar_insert_widget
extern __typeof (gtk_toolbar_insert_widget) gtk_toolbar_insert_widget __attribute((alias("IA__gtk_toolbar_insert_widget"), visibility("default")));

#undef gtk_toolbar_prepend_widget
extern __typeof (gtk_toolbar_prepend_widget) gtk_toolbar_prepend_widget __attribute((alias("IA__gtk_toolbar_prepend_widget"), visibility("default")));

#endif
#undef gtk_toolbar_get_drop_index
extern __typeof (gtk_toolbar_get_drop_index) gtk_toolbar_get_drop_index __attribute((alias("IA__gtk_toolbar_get_drop_index"), visibility("default")));

#undef gtk_toolbar_get_icon_size
extern __typeof (gtk_toolbar_get_icon_size) gtk_toolbar_get_icon_size __attribute((alias("IA__gtk_toolbar_get_icon_size"), visibility("default")));

#undef gtk_toolbar_get_item_index
extern __typeof (gtk_toolbar_get_item_index) gtk_toolbar_get_item_index __attribute((alias("IA__gtk_toolbar_get_item_index"), visibility("default")));

#undef gtk_toolbar_get_n_items
extern __typeof (gtk_toolbar_get_n_items) gtk_toolbar_get_n_items __attribute((alias("IA__gtk_toolbar_get_n_items"), visibility("default")));

#undef gtk_toolbar_get_nth_item
extern __typeof (gtk_toolbar_get_nth_item) gtk_toolbar_get_nth_item __attribute((alias("IA__gtk_toolbar_get_nth_item"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_toolbar_get_orientation
extern __typeof (gtk_toolbar_get_orientation) gtk_toolbar_get_orientation __attribute((alias("IA__gtk_toolbar_get_orientation"), visibility("default")));

#endif
#undef gtk_toolbar_get_relief_style
extern __typeof (gtk_toolbar_get_relief_style) gtk_toolbar_get_relief_style __attribute((alias("IA__gtk_toolbar_get_relief_style"), visibility("default")));

#undef gtk_toolbar_get_show_arrow
extern __typeof (gtk_toolbar_get_show_arrow) gtk_toolbar_get_show_arrow __attribute((alias("IA__gtk_toolbar_get_show_arrow"), visibility("default")));

#undef gtk_toolbar_get_style
extern __typeof (gtk_toolbar_get_style) gtk_toolbar_get_style __attribute((alias("IA__gtk_toolbar_get_style"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_toolbar_get_tooltips
extern __typeof (gtk_toolbar_get_tooltips) gtk_toolbar_get_tooltips __attribute((alias("IA__gtk_toolbar_get_tooltips"), visibility("default")));

#endif
#undef gtk_toolbar_get_type
extern __typeof (gtk_toolbar_get_type) gtk_toolbar_get_type __attribute((alias("IA__gtk_toolbar_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_toolbar_insert
extern __typeof (gtk_toolbar_insert) gtk_toolbar_insert __attribute((alias("IA__gtk_toolbar_insert"), visibility("default")));

#undef gtk_toolbar_new
extern __typeof (gtk_toolbar_new) gtk_toolbar_new __attribute((alias("IA__gtk_toolbar_new"), visibility("default")));

#undef gtk_toolbar_set_drop_highlight_item
extern __typeof (gtk_toolbar_set_drop_highlight_item) gtk_toolbar_set_drop_highlight_item __attribute((alias("IA__gtk_toolbar_set_drop_highlight_item"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_toolbar_set_orientation
extern __typeof (gtk_toolbar_set_orientation) gtk_toolbar_set_orientation __attribute((alias("IA__gtk_toolbar_set_orientation"), visibility("default")));

#endif
#undef gtk_toolbar_set_show_arrow
extern __typeof (gtk_toolbar_set_show_arrow) gtk_toolbar_set_show_arrow __attribute((alias("IA__gtk_toolbar_set_show_arrow"), visibility("default")));

#undef gtk_toolbar_set_style
extern __typeof (gtk_toolbar_set_style) gtk_toolbar_set_style __attribute((alias("IA__gtk_toolbar_set_style"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_toolbar_set_tooltips
extern __typeof (gtk_toolbar_set_tooltips) gtk_toolbar_set_tooltips __attribute((alias("IA__gtk_toolbar_set_tooltips"), visibility("default")));

#endif
#undef gtk_toolbar_unset_style
extern __typeof (gtk_toolbar_unset_style) gtk_toolbar_unset_style __attribute((alias("IA__gtk_toolbar_unset_style"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TOOL_SHELL_H__)
#if IN_FILE(__GTK_TOOL_SHELL_C__)
#undef gtk_tool_shell_get_type
extern __typeof (gtk_tool_shell_get_type) gtk_tool_shell_get_type __attribute((alias("IA__gtk_tool_shell_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tool_shell_get_ellipsize_mode
extern __typeof (gtk_tool_shell_get_ellipsize_mode) gtk_tool_shell_get_ellipsize_mode __attribute((alias("IA__gtk_tool_shell_get_ellipsize_mode"), visibility("default")));

#undef gtk_tool_shell_get_icon_size
extern __typeof (gtk_tool_shell_get_icon_size) gtk_tool_shell_get_icon_size __attribute((alias("IA__gtk_tool_shell_get_icon_size"), visibility("default")));

#undef gtk_tool_shell_get_orientation
extern __typeof (gtk_tool_shell_get_orientation) gtk_tool_shell_get_orientation __attribute((alias("IA__gtk_tool_shell_get_orientation"), visibility("default")));

#undef gtk_tool_shell_get_style
extern __typeof (gtk_tool_shell_get_style) gtk_tool_shell_get_style __attribute((alias("IA__gtk_tool_shell_get_style"), visibility("default")));

#undef gtk_tool_shell_get_relief_style
extern __typeof (gtk_tool_shell_get_relief_style) gtk_tool_shell_get_relief_style __attribute((alias("IA__gtk_tool_shell_get_relief_style"), visibility("default")));

#undef gtk_tool_shell_get_text_alignment
extern __typeof (gtk_tool_shell_get_text_alignment) gtk_tool_shell_get_text_alignment __attribute((alias("IA__gtk_tool_shell_get_text_alignment"), visibility("default")));

#undef gtk_tool_shell_get_text_orientation
extern __typeof (gtk_tool_shell_get_text_orientation) gtk_tool_shell_get_text_orientation __attribute((alias("IA__gtk_tool_shell_get_text_orientation"), visibility("default")));

#undef gtk_tool_shell_get_text_size_group
extern __typeof (gtk_tool_shell_get_text_size_group) gtk_tool_shell_get_text_size_group __attribute((alias("IA__gtk_tool_shell_get_text_size_group"), visibility("default")));

#undef gtk_tool_shell_rebuild_menu
extern __typeof (gtk_tool_shell_rebuild_menu) gtk_tool_shell_rebuild_menu __attribute((alias("IA__gtk_tool_shell_rebuild_menu"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TOOL_BUTTON_H__)
#if IN_FILE(__GTK_TOOL_BUTTON_C__)
#undef gtk_tool_button_get_icon_name
extern __typeof (gtk_tool_button_get_icon_name) gtk_tool_button_get_icon_name __attribute((alias("IA__gtk_tool_button_get_icon_name"), visibility("default")));

#undef gtk_tool_button_get_icon_widget
extern __typeof (gtk_tool_button_get_icon_widget) gtk_tool_button_get_icon_widget __attribute((alias("IA__gtk_tool_button_get_icon_widget"), visibility("default")));

#undef gtk_tool_button_get_label
extern __typeof (gtk_tool_button_get_label) gtk_tool_button_get_label __attribute((alias("IA__gtk_tool_button_get_label"), visibility("default")));

#undef gtk_tool_button_get_label_widget
extern __typeof (gtk_tool_button_get_label_widget) gtk_tool_button_get_label_widget __attribute((alias("IA__gtk_tool_button_get_label_widget"), visibility("default")));

#undef gtk_tool_button_get_stock_id
extern __typeof (gtk_tool_button_get_stock_id) gtk_tool_button_get_stock_id __attribute((alias("IA__gtk_tool_button_get_stock_id"), visibility("default")));

#undef gtk_tool_button_get_type
extern __typeof (gtk_tool_button_get_type) gtk_tool_button_get_type __attribute((alias("IA__gtk_tool_button_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tool_button_get_use_underline
extern __typeof (gtk_tool_button_get_use_underline) gtk_tool_button_get_use_underline __attribute((alias("IA__gtk_tool_button_get_use_underline"), visibility("default")));

#undef gtk_tool_button_new
extern __typeof (gtk_tool_button_new) gtk_tool_button_new __attribute((alias("IA__gtk_tool_button_new"), visibility("default")));

#undef gtk_tool_button_new_from_stock
extern __typeof (gtk_tool_button_new_from_stock) gtk_tool_button_new_from_stock __attribute((alias("IA__gtk_tool_button_new_from_stock"), visibility("default")));

#undef gtk_tool_button_set_icon_name
extern __typeof (gtk_tool_button_set_icon_name) gtk_tool_button_set_icon_name __attribute((alias("IA__gtk_tool_button_set_icon_name"), visibility("default")));

#undef gtk_tool_button_set_icon_widget
extern __typeof (gtk_tool_button_set_icon_widget) gtk_tool_button_set_icon_widget __attribute((alias("IA__gtk_tool_button_set_icon_widget"), visibility("default")));

#undef gtk_tool_button_set_label
extern __typeof (gtk_tool_button_set_label) gtk_tool_button_set_label __attribute((alias("IA__gtk_tool_button_set_label"), visibility("default")));

#undef gtk_tool_button_set_label_widget
extern __typeof (gtk_tool_button_set_label_widget) gtk_tool_button_set_label_widget __attribute((alias("IA__gtk_tool_button_set_label_widget"), visibility("default")));

#undef gtk_tool_button_set_stock_id
extern __typeof (gtk_tool_button_set_stock_id) gtk_tool_button_set_stock_id __attribute((alias("IA__gtk_tool_button_set_stock_id"), visibility("default")));

#undef gtk_tool_button_set_use_underline
extern __typeof (gtk_tool_button_set_use_underline) gtk_tool_button_set_use_underline __attribute((alias("IA__gtk_tool_button_set_use_underline"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TOOL_ITEM_H__)
#if IN_FILE(__GTK_TOOL_ITEM_C__)
#undef gtk_tool_item_get_ellipsize_mode
extern __typeof (gtk_tool_item_get_ellipsize_mode) gtk_tool_item_get_ellipsize_mode __attribute((alias("IA__gtk_tool_item_get_ellipsize_mode"), visibility("default")));

#undef gtk_tool_item_get_expand
extern __typeof (gtk_tool_item_get_expand) gtk_tool_item_get_expand __attribute((alias("IA__gtk_tool_item_get_expand"), visibility("default")));

#undef gtk_tool_item_get_homogeneous
extern __typeof (gtk_tool_item_get_homogeneous) gtk_tool_item_get_homogeneous __attribute((alias("IA__gtk_tool_item_get_homogeneous"), visibility("default")));

#undef gtk_tool_item_get_icon_size
extern __typeof (gtk_tool_item_get_icon_size) gtk_tool_item_get_icon_size __attribute((alias("IA__gtk_tool_item_get_icon_size"), visibility("default")));

#undef gtk_tool_item_get_is_important
extern __typeof (gtk_tool_item_get_is_important) gtk_tool_item_get_is_important __attribute((alias("IA__gtk_tool_item_get_is_important"), visibility("default")));

#undef gtk_tool_item_get_orientation
extern __typeof (gtk_tool_item_get_orientation) gtk_tool_item_get_orientation __attribute((alias("IA__gtk_tool_item_get_orientation"), visibility("default")));

#undef gtk_tool_item_get_proxy_menu_item
extern __typeof (gtk_tool_item_get_proxy_menu_item) gtk_tool_item_get_proxy_menu_item __attribute((alias("IA__gtk_tool_item_get_proxy_menu_item"), visibility("default")));

#undef gtk_tool_item_get_relief_style
extern __typeof (gtk_tool_item_get_relief_style) gtk_tool_item_get_relief_style __attribute((alias("IA__gtk_tool_item_get_relief_style"), visibility("default")));

#undef gtk_tool_item_get_text_alignment
extern __typeof (gtk_tool_item_get_text_alignment) gtk_tool_item_get_text_alignment __attribute((alias("IA__gtk_tool_item_get_text_alignment"), visibility("default")));

#undef gtk_tool_item_get_text_orientation
extern __typeof (gtk_tool_item_get_text_orientation) gtk_tool_item_get_text_orientation __attribute((alias("IA__gtk_tool_item_get_text_orientation"), visibility("default")));

#undef gtk_tool_item_get_text_size_group
extern __typeof (gtk_tool_item_get_text_size_group) gtk_tool_item_get_text_size_group __attribute((alias("IA__gtk_tool_item_get_text_size_group"), visibility("default")));

#undef gtk_tool_item_get_toolbar_style
extern __typeof (gtk_tool_item_get_toolbar_style) gtk_tool_item_get_toolbar_style __attribute((alias("IA__gtk_tool_item_get_toolbar_style"), visibility("default")));

#undef gtk_tool_item_get_type
extern __typeof (gtk_tool_item_get_type) gtk_tool_item_get_type __attribute((alias("IA__gtk_tool_item_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tool_item_get_use_drag_window
extern __typeof (gtk_tool_item_get_use_drag_window) gtk_tool_item_get_use_drag_window __attribute((alias("IA__gtk_tool_item_get_use_drag_window"), visibility("default")));

#undef gtk_tool_item_get_visible_horizontal
extern __typeof (gtk_tool_item_get_visible_horizontal) gtk_tool_item_get_visible_horizontal __attribute((alias("IA__gtk_tool_item_get_visible_horizontal"), visibility("default")));

#undef gtk_tool_item_get_visible_vertical
extern __typeof (gtk_tool_item_get_visible_vertical) gtk_tool_item_get_visible_vertical __attribute((alias("IA__gtk_tool_item_get_visible_vertical"), visibility("default")));

#undef gtk_tool_item_new
extern __typeof (gtk_tool_item_new) gtk_tool_item_new __attribute((alias("IA__gtk_tool_item_new"), visibility("default")));

#undef gtk_tool_item_rebuild_menu
extern __typeof (gtk_tool_item_rebuild_menu) gtk_tool_item_rebuild_menu __attribute((alias("IA__gtk_tool_item_rebuild_menu"), visibility("default")));

#undef gtk_tool_item_retrieve_proxy_menu_item
extern __typeof (gtk_tool_item_retrieve_proxy_menu_item) gtk_tool_item_retrieve_proxy_menu_item __attribute((alias("IA__gtk_tool_item_retrieve_proxy_menu_item"), visibility("default")));

#undef gtk_tool_item_set_expand
extern __typeof (gtk_tool_item_set_expand) gtk_tool_item_set_expand __attribute((alias("IA__gtk_tool_item_set_expand"), visibility("default")));

#undef gtk_tool_item_set_homogeneous
extern __typeof (gtk_tool_item_set_homogeneous) gtk_tool_item_set_homogeneous __attribute((alias("IA__gtk_tool_item_set_homogeneous"), visibility("default")));

#undef gtk_tool_item_set_is_important
extern __typeof (gtk_tool_item_set_is_important) gtk_tool_item_set_is_important __attribute((alias("IA__gtk_tool_item_set_is_important"), visibility("default")));

#undef gtk_tool_item_set_proxy_menu_item
extern __typeof (gtk_tool_item_set_proxy_menu_item) gtk_tool_item_set_proxy_menu_item __attribute((alias("IA__gtk_tool_item_set_proxy_menu_item"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_tool_item_set_tooltip
extern __typeof (gtk_tool_item_set_tooltip) gtk_tool_item_set_tooltip __attribute((alias("IA__gtk_tool_item_set_tooltip"), visibility("default")));

#endif
#undef gtk_tool_item_set_tooltip_markup
extern __typeof (gtk_tool_item_set_tooltip_markup) gtk_tool_item_set_tooltip_markup __attribute((alias("IA__gtk_tool_item_set_tooltip_markup"), visibility("default")));

#undef gtk_tool_item_set_tooltip_text
extern __typeof (gtk_tool_item_set_tooltip_text) gtk_tool_item_set_tooltip_text __attribute((alias("IA__gtk_tool_item_set_tooltip_text"), visibility("default")));

#undef gtk_tool_item_set_use_drag_window
extern __typeof (gtk_tool_item_set_use_drag_window) gtk_tool_item_set_use_drag_window __attribute((alias("IA__gtk_tool_item_set_use_drag_window"), visibility("default")));

#undef gtk_tool_item_set_visible_horizontal
extern __typeof (gtk_tool_item_set_visible_horizontal) gtk_tool_item_set_visible_horizontal __attribute((alias("IA__gtk_tool_item_set_visible_horizontal"), visibility("default")));

#undef gtk_tool_item_set_visible_vertical
extern __typeof (gtk_tool_item_set_visible_vertical) gtk_tool_item_set_visible_vertical __attribute((alias("IA__gtk_tool_item_set_visible_vertical"), visibility("default")));

#undef gtk_tool_item_toolbar_reconfigured
extern __typeof (gtk_tool_item_toolbar_reconfigured) gtk_tool_item_toolbar_reconfigured __attribute((alias("IA__gtk_tool_item_toolbar_reconfigured"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TOOL_ITEM_GROUP_H__)
#if IN_FILE(__GTK_TOOL_ITEM_GROUP_C__)
#undef gtk_tool_item_group_get_type
extern __typeof (gtk_tool_item_group_get_type) gtk_tool_item_group_get_type __attribute((alias("IA__gtk_tool_item_group_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tool_item_group_new
extern __typeof (gtk_tool_item_group_new) gtk_tool_item_group_new __attribute((alias("IA__gtk_tool_item_group_new"), visibility("default")));

#undef gtk_tool_item_group_set_label
extern __typeof (gtk_tool_item_group_set_label) gtk_tool_item_group_set_label __attribute((alias("IA__gtk_tool_item_group_set_label"), visibility("default")));

#undef gtk_tool_item_group_set_label_widget
extern __typeof (gtk_tool_item_group_set_label_widget) gtk_tool_item_group_set_label_widget __attribute((alias("IA__gtk_tool_item_group_set_label_widget"), visibility("default")));

#undef gtk_tool_item_group_set_collapsed
extern __typeof (gtk_tool_item_group_set_collapsed) gtk_tool_item_group_set_collapsed __attribute((alias("IA__gtk_tool_item_group_set_collapsed"), visibility("default")));

#undef gtk_tool_item_group_set_ellipsize
extern __typeof (gtk_tool_item_group_set_ellipsize) gtk_tool_item_group_set_ellipsize __attribute((alias("IA__gtk_tool_item_group_set_ellipsize"), visibility("default")));

#undef gtk_tool_item_group_set_header_relief
extern __typeof (gtk_tool_item_group_set_header_relief) gtk_tool_item_group_set_header_relief __attribute((alias("IA__gtk_tool_item_group_set_header_relief"), visibility("default")));

#undef gtk_tool_item_group_get_label
extern __typeof (gtk_tool_item_group_get_label) gtk_tool_item_group_get_label __attribute((alias("IA__gtk_tool_item_group_get_label"), visibility("default")));

#undef gtk_tool_item_group_get_label_widget
extern __typeof (gtk_tool_item_group_get_label_widget) gtk_tool_item_group_get_label_widget __attribute((alias("IA__gtk_tool_item_group_get_label_widget"), visibility("default")));

#undef gtk_tool_item_group_get_ellipsize
extern __typeof (gtk_tool_item_group_get_ellipsize) gtk_tool_item_group_get_ellipsize __attribute((alias("IA__gtk_tool_item_group_get_ellipsize"), visibility("default")));

#undef gtk_tool_item_group_get_header_relief
extern __typeof (gtk_tool_item_group_get_header_relief) gtk_tool_item_group_get_header_relief __attribute((alias("IA__gtk_tool_item_group_get_header_relief"), visibility("default")));

#undef gtk_tool_item_group_insert
extern __typeof (gtk_tool_item_group_insert) gtk_tool_item_group_insert __attribute((alias("IA__gtk_tool_item_group_insert"), visibility("default")));

#undef gtk_tool_item_group_set_item_position
extern __typeof (gtk_tool_item_group_set_item_position) gtk_tool_item_group_set_item_position __attribute((alias("IA__gtk_tool_item_group_set_item_position"), visibility("default")));

#undef gtk_tool_item_group_get_item_position
extern __typeof (gtk_tool_item_group_get_item_position) gtk_tool_item_group_get_item_position __attribute((alias("IA__gtk_tool_item_group_get_item_position"), visibility("default")));

#undef gtk_tool_item_group_get_n_items
extern __typeof (gtk_tool_item_group_get_n_items) gtk_tool_item_group_get_n_items __attribute((alias("IA__gtk_tool_item_group_get_n_items"), visibility("default")));

#undef gtk_tool_item_group_get_nth_item
extern __typeof (gtk_tool_item_group_get_nth_item) gtk_tool_item_group_get_nth_item __attribute((alias("IA__gtk_tool_item_group_get_nth_item"), visibility("default")));

#undef gtk_tool_item_group_get_drop_item
extern __typeof (gtk_tool_item_group_get_drop_item) gtk_tool_item_group_get_drop_item __attribute((alias("IA__gtk_tool_item_group_get_drop_item"), visibility("default")));

#undef gtk_tool_item_group_get_collapsed
extern __typeof (gtk_tool_item_group_get_collapsed) gtk_tool_item_group_get_collapsed __attribute((alias("IA__gtk_tool_item_group_get_collapsed"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TOOL_PALETTE_H__)
#if IN_FILE(__GTK_TOOL_PALETTE_C__)
#undef gtk_tool_palette_get_type
extern __typeof (gtk_tool_palette_get_type) gtk_tool_palette_get_type __attribute((alias("IA__gtk_tool_palette_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tool_palette_new
extern __typeof (gtk_tool_palette_new) gtk_tool_palette_new __attribute((alias("IA__gtk_tool_palette_new"), visibility("default")));

#undef gtk_tool_palette_set_group_position
extern __typeof (gtk_tool_palette_set_group_position) gtk_tool_palette_set_group_position __attribute((alias("IA__gtk_tool_palette_set_group_position"), visibility("default")));

#undef gtk_tool_palette_set_exclusive
extern __typeof (gtk_tool_palette_set_exclusive) gtk_tool_palette_set_exclusive __attribute((alias("IA__gtk_tool_palette_set_exclusive"), visibility("default")));

#undef gtk_tool_palette_set_expand
extern __typeof (gtk_tool_palette_set_expand) gtk_tool_palette_set_expand __attribute((alias("IA__gtk_tool_palette_set_expand"), visibility("default")));

#undef gtk_tool_palette_get_group_position
extern __typeof (gtk_tool_palette_get_group_position) gtk_tool_palette_get_group_position __attribute((alias("IA__gtk_tool_palette_get_group_position"), visibility("default")));

#undef gtk_tool_palette_get_exclusive
extern __typeof (gtk_tool_palette_get_exclusive) gtk_tool_palette_get_exclusive __attribute((alias("IA__gtk_tool_palette_get_exclusive"), visibility("default")));

#undef gtk_tool_palette_get_expand
extern __typeof (gtk_tool_palette_get_expand) gtk_tool_palette_get_expand __attribute((alias("IA__gtk_tool_palette_get_expand"), visibility("default")));

#undef gtk_tool_palette_set_icon_size
extern __typeof (gtk_tool_palette_set_icon_size) gtk_tool_palette_set_icon_size __attribute((alias("IA__gtk_tool_palette_set_icon_size"), visibility("default")));

#undef gtk_tool_palette_unset_icon_size
extern __typeof (gtk_tool_palette_unset_icon_size) gtk_tool_palette_unset_icon_size __attribute((alias("IA__gtk_tool_palette_unset_icon_size"), visibility("default")));

#undef gtk_tool_palette_set_style
extern __typeof (gtk_tool_palette_set_style) gtk_tool_palette_set_style __attribute((alias("IA__gtk_tool_palette_set_style"), visibility("default")));

#undef gtk_tool_palette_unset_style
extern __typeof (gtk_tool_palette_unset_style) gtk_tool_palette_unset_style __attribute((alias("IA__gtk_tool_palette_unset_style"), visibility("default")));

#undef gtk_tool_palette_get_icon_size
extern __typeof (gtk_tool_palette_get_icon_size) gtk_tool_palette_get_icon_size __attribute((alias("IA__gtk_tool_palette_get_icon_size"), visibility("default")));

#undef gtk_tool_palette_get_style
extern __typeof (gtk_tool_palette_get_style) gtk_tool_palette_get_style __attribute((alias("IA__gtk_tool_palette_get_style"), visibility("default")));

#undef gtk_tool_palette_get_drop_item
extern __typeof (gtk_tool_palette_get_drop_item) gtk_tool_palette_get_drop_item __attribute((alias("IA__gtk_tool_palette_get_drop_item"), visibility("default")));

#undef gtk_tool_palette_get_drop_group
extern __typeof (gtk_tool_palette_get_drop_group) gtk_tool_palette_get_drop_group __attribute((alias("IA__gtk_tool_palette_get_drop_group"), visibility("default")));

#undef gtk_tool_palette_get_drag_item
extern __typeof (gtk_tool_palette_get_drag_item) gtk_tool_palette_get_drag_item __attribute((alias("IA__gtk_tool_palette_get_drag_item"), visibility("default")));

#undef gtk_tool_palette_set_drag_source
extern __typeof (gtk_tool_palette_set_drag_source) gtk_tool_palette_set_drag_source __attribute((alias("IA__gtk_tool_palette_set_drag_source"), visibility("default")));

#undef gtk_tool_palette_add_drag_dest
extern __typeof (gtk_tool_palette_add_drag_dest) gtk_tool_palette_add_drag_dest __attribute((alias("IA__gtk_tool_palette_add_drag_dest"), visibility("default")));

#undef gtk_tool_palette_get_hadjustment
extern __typeof (gtk_tool_palette_get_hadjustment) gtk_tool_palette_get_hadjustment __attribute((alias("IA__gtk_tool_palette_get_hadjustment"), visibility("default")));

#undef gtk_tool_palette_get_vadjustment
extern __typeof (gtk_tool_palette_get_vadjustment) gtk_tool_palette_get_vadjustment __attribute((alias("IA__gtk_tool_palette_get_vadjustment"), visibility("default")));

#undef gtk_tool_palette_get_drag_target_item
extern __typeof (gtk_tool_palette_get_drag_target_item) gtk_tool_palette_get_drag_target_item __attribute((alias("IA__gtk_tool_palette_get_drag_target_item"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tool_palette_get_drag_target_group
extern __typeof (gtk_tool_palette_get_drag_target_group) gtk_tool_palette_get_drag_target_group __attribute((alias("IA__gtk_tool_palette_get_drag_target_group"), visibility("default"))) G_GNUC_CONST;

#endif
#endif
#if IN_HEADER(__GTK_TOOLTIP_H__)
#if IN_FILE(__GTK_TOOLTIP_C__)
#undef gtk_tooltip_get_type
extern __typeof (gtk_tooltip_get_type) gtk_tooltip_get_type __attribute((alias("IA__gtk_tooltip_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tooltip_set_custom
extern __typeof (gtk_tooltip_set_custom) gtk_tooltip_set_custom __attribute((alias("IA__gtk_tooltip_set_custom"), visibility("default")));

#undef gtk_tooltip_set_icon
extern __typeof (gtk_tooltip_set_icon) gtk_tooltip_set_icon __attribute((alias("IA__gtk_tooltip_set_icon"), visibility("default")));

#undef gtk_tooltip_set_icon_from_stock
extern __typeof (gtk_tooltip_set_icon_from_stock) gtk_tooltip_set_icon_from_stock __attribute((alias("IA__gtk_tooltip_set_icon_from_stock"), visibility("default")));

#undef gtk_tooltip_set_icon_from_icon_name
extern __typeof (gtk_tooltip_set_icon_from_icon_name) gtk_tooltip_set_icon_from_icon_name __attribute((alias("IA__gtk_tooltip_set_icon_from_icon_name"), visibility("default")));

#undef gtk_tooltip_set_icon_from_gicon
extern __typeof (gtk_tooltip_set_icon_from_gicon) gtk_tooltip_set_icon_from_gicon __attribute((alias("IA__gtk_tooltip_set_icon_from_gicon"), visibility("default")));

#undef gtk_tooltip_set_markup
extern __typeof (gtk_tooltip_set_markup) gtk_tooltip_set_markup __attribute((alias("IA__gtk_tooltip_set_markup"), visibility("default")));

#undef gtk_tooltip_set_text
extern __typeof (gtk_tooltip_set_text) gtk_tooltip_set_text __attribute((alias("IA__gtk_tooltip_set_text"), visibility("default")));

#undef gtk_tooltip_set_tip_area
extern __typeof (gtk_tooltip_set_tip_area) gtk_tooltip_set_tip_area __attribute((alias("IA__gtk_tooltip_set_tip_area"), visibility("default")));

#undef gtk_tooltip_trigger_tooltip_query
extern __typeof (gtk_tooltip_trigger_tooltip_query) gtk_tooltip_trigger_tooltip_query __attribute((alias("IA__gtk_tooltip_trigger_tooltip_query"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TOOLTIPS_H__)
#if IN_FILE(__GTK_TOOLTIPS_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_tooltips_data_get
extern __typeof (gtk_tooltips_data_get) gtk_tooltips_data_get __attribute((alias("IA__gtk_tooltips_data_get"), visibility("default")));

#undef gtk_tooltips_disable
extern __typeof (gtk_tooltips_disable) gtk_tooltips_disable __attribute((alias("IA__gtk_tooltips_disable"), visibility("default")));

#undef gtk_tooltips_enable
extern __typeof (gtk_tooltips_enable) gtk_tooltips_enable __attribute((alias("IA__gtk_tooltips_enable"), visibility("default")));

#undef gtk_tooltips_force_window
extern __typeof (gtk_tooltips_force_window) gtk_tooltips_force_window __attribute((alias("IA__gtk_tooltips_force_window"), visibility("default")));

#undef gtk_tooltips_get_info_from_tip_window
extern __typeof (gtk_tooltips_get_info_from_tip_window) gtk_tooltips_get_info_from_tip_window __attribute((alias("IA__gtk_tooltips_get_info_from_tip_window"), visibility("default")));

#undef gtk_tooltips_get_type
extern __typeof (gtk_tooltips_get_type) gtk_tooltips_get_type __attribute((alias("IA__gtk_tooltips_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tooltips_new
extern __typeof (gtk_tooltips_new) gtk_tooltips_new __attribute((alias("IA__gtk_tooltips_new"), visibility("default")));

#undef gtk_tooltips_set_delay
extern __typeof (gtk_tooltips_set_delay) gtk_tooltips_set_delay __attribute((alias("IA__gtk_tooltips_set_delay"), visibility("default")));

#undef gtk_tooltips_set_tip
extern __typeof (gtk_tooltips_set_tip) gtk_tooltips_set_tip __attribute((alias("IA__gtk_tooltips_set_tip"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_TRAY_ICON_H__)
#if IN_FILE(__GTK_TRAY_ICON_X11_C__)
#ifndef G_OS_WIN32
#undef gtk_tray_icon_get_type
extern __typeof (gtk_tray_icon_get_type) gtk_tray_icon_get_type __attribute((alias("IA__gtk_tray_icon_get_type"), visibility("default"))) G_GNUC_CONST;

#endif
#endif
#endif
#if IN_HEADER(__GTK_TREE_H__)
#if IN_FILE(__GTK_TREE_C__)
#undef gtk_tree_append
extern __typeof (gtk_tree_append) gtk_tree_append __attribute((alias("IA__gtk_tree_append"), visibility("default")));

#undef gtk_tree_child_position
extern __typeof (gtk_tree_child_position) gtk_tree_child_position __attribute((alias("IA__gtk_tree_child_position"), visibility("default")));

#undef gtk_tree_clear_items
extern __typeof (gtk_tree_clear_items) gtk_tree_clear_items __attribute((alias("IA__gtk_tree_clear_items"), visibility("default")));

#undef gtk_tree_get_type
extern __typeof (gtk_tree_get_type) gtk_tree_get_type __attribute((alias("IA__gtk_tree_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_insert
extern __typeof (gtk_tree_insert) gtk_tree_insert __attribute((alias("IA__gtk_tree_insert"), visibility("default")));

#undef gtk_tree_prepend
extern __typeof (gtk_tree_prepend) gtk_tree_prepend __attribute((alias("IA__gtk_tree_prepend"), visibility("default")));

#undef gtk_tree_remove_item
extern __typeof (gtk_tree_remove_item) gtk_tree_remove_item __attribute((alias("IA__gtk_tree_remove_item"), visibility("default")));

#undef gtk_tree_remove_items
extern __typeof (gtk_tree_remove_items) gtk_tree_remove_items __attribute((alias("IA__gtk_tree_remove_items"), visibility("default")));

#undef gtk_tree_select_child
extern __typeof (gtk_tree_select_child) gtk_tree_select_child __attribute((alias("IA__gtk_tree_select_child"), visibility("default")));

#undef gtk_tree_select_item
extern __typeof (gtk_tree_select_item) gtk_tree_select_item __attribute((alias("IA__gtk_tree_select_item"), visibility("default")));

#undef gtk_tree_set_selection_mode
extern __typeof (gtk_tree_set_selection_mode) gtk_tree_set_selection_mode __attribute((alias("IA__gtk_tree_set_selection_mode"), visibility("default")));

#undef gtk_tree_set_view_lines
extern __typeof (gtk_tree_set_view_lines) gtk_tree_set_view_lines __attribute((alias("IA__gtk_tree_set_view_lines"), visibility("default")));

#undef gtk_tree_set_view_mode
extern __typeof (gtk_tree_set_view_mode) gtk_tree_set_view_mode __attribute((alias("IA__gtk_tree_set_view_mode"), visibility("default")));

#undef gtk_tree_unselect_child
extern __typeof (gtk_tree_unselect_child) gtk_tree_unselect_child __attribute((alias("IA__gtk_tree_unselect_child"), visibility("default")));

#undef gtk_tree_unselect_item
extern __typeof (gtk_tree_unselect_item) gtk_tree_unselect_item __attribute((alias("IA__gtk_tree_unselect_item"), visibility("default")));

#undef gtk_tree_new
extern __typeof (gtk_tree_new) gtk_tree_new __attribute((alias("IA__gtk_tree_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TREE_DND_H__)
#if IN_FILE(__GTK_TREE_DND_C__)
#undef gtk_tree_drag_dest_drag_data_received
extern __typeof (gtk_tree_drag_dest_drag_data_received) gtk_tree_drag_dest_drag_data_received __attribute((alias("IA__gtk_tree_drag_dest_drag_data_received"), visibility("default")));

#undef gtk_tree_drag_dest_get_type
extern __typeof (gtk_tree_drag_dest_get_type) gtk_tree_drag_dest_get_type __attribute((alias("IA__gtk_tree_drag_dest_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_drag_dest_row_drop_possible
extern __typeof (gtk_tree_drag_dest_row_drop_possible) gtk_tree_drag_dest_row_drop_possible __attribute((alias("IA__gtk_tree_drag_dest_row_drop_possible"), visibility("default")));

#undef gtk_tree_drag_source_drag_data_delete
extern __typeof (gtk_tree_drag_source_drag_data_delete) gtk_tree_drag_source_drag_data_delete __attribute((alias("IA__gtk_tree_drag_source_drag_data_delete"), visibility("default")));

#undef gtk_tree_drag_source_drag_data_get
extern __typeof (gtk_tree_drag_source_drag_data_get) gtk_tree_drag_source_drag_data_get __attribute((alias("IA__gtk_tree_drag_source_drag_data_get"), visibility("default")));

#undef gtk_tree_drag_source_get_type
extern __typeof (gtk_tree_drag_source_get_type) gtk_tree_drag_source_get_type __attribute((alias("IA__gtk_tree_drag_source_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_drag_source_row_draggable
extern __typeof (gtk_tree_drag_source_row_draggable) gtk_tree_drag_source_row_draggable __attribute((alias("IA__gtk_tree_drag_source_row_draggable"), visibility("default")));

#undef gtk_tree_get_row_drag_data
extern __typeof (gtk_tree_get_row_drag_data) gtk_tree_get_row_drag_data __attribute((alias("IA__gtk_tree_get_row_drag_data"), visibility("default")));

#undef gtk_tree_set_row_drag_data
extern __typeof (gtk_tree_set_row_drag_data) gtk_tree_set_row_drag_data __attribute((alias("IA__gtk_tree_set_row_drag_data"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TREE_ITEM_H__)
#if IN_FILE(__GTK_TREE_ITEM_C__)
#undef gtk_tree_item_collapse
extern __typeof (gtk_tree_item_collapse) gtk_tree_item_collapse __attribute((alias("IA__gtk_tree_item_collapse"), visibility("default")));

#undef gtk_tree_item_deselect
extern __typeof (gtk_tree_item_deselect) gtk_tree_item_deselect __attribute((alias("IA__gtk_tree_item_deselect"), visibility("default")));

#undef gtk_tree_item_expand
extern __typeof (gtk_tree_item_expand) gtk_tree_item_expand __attribute((alias("IA__gtk_tree_item_expand"), visibility("default")));

#undef gtk_tree_item_get_type
extern __typeof (gtk_tree_item_get_type) gtk_tree_item_get_type __attribute((alias("IA__gtk_tree_item_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_item_new
extern __typeof (gtk_tree_item_new) gtk_tree_item_new __attribute((alias("IA__gtk_tree_item_new"), visibility("default")));

#undef gtk_tree_item_new_with_label
extern __typeof (gtk_tree_item_new_with_label) gtk_tree_item_new_with_label __attribute((alias("IA__gtk_tree_item_new_with_label"), visibility("default")));

#undef gtk_tree_item_remove_subtree
extern __typeof (gtk_tree_item_remove_subtree) gtk_tree_item_remove_subtree __attribute((alias("IA__gtk_tree_item_remove_subtree"), visibility("default")));

#undef gtk_tree_item_select
extern __typeof (gtk_tree_item_select) gtk_tree_item_select __attribute((alias("IA__gtk_tree_item_select"), visibility("default")));

#undef gtk_tree_item_set_subtree
extern __typeof (gtk_tree_item_set_subtree) gtk_tree_item_set_subtree __attribute((alias("IA__gtk_tree_item_set_subtree"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TREE_MODEL_H__)
#if IN_FILE(__GTK_TREE_MODEL_C__)
#undef gtk_tree_iter_copy
extern __typeof (gtk_tree_iter_copy) gtk_tree_iter_copy __attribute((alias("IA__gtk_tree_iter_copy"), visibility("default")));

#undef gtk_tree_iter_free
extern __typeof (gtk_tree_iter_free) gtk_tree_iter_free __attribute((alias("IA__gtk_tree_iter_free"), visibility("default")));

#undef gtk_tree_iter_get_type
extern __typeof (gtk_tree_iter_get_type) gtk_tree_iter_get_type __attribute((alias("IA__gtk_tree_iter_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_model_foreach
extern __typeof (gtk_tree_model_foreach) gtk_tree_model_foreach __attribute((alias("IA__gtk_tree_model_foreach"), visibility("default")));

#undef gtk_tree_model_get
extern __typeof (gtk_tree_model_get) gtk_tree_model_get __attribute((alias("IA__gtk_tree_model_get"), visibility("default")));

#undef gtk_tree_model_get_column_type
extern __typeof (gtk_tree_model_get_column_type) gtk_tree_model_get_column_type __attribute((alias("IA__gtk_tree_model_get_column_type"), visibility("default")));

#undef gtk_tree_model_get_flags
extern __typeof (gtk_tree_model_get_flags) gtk_tree_model_get_flags __attribute((alias("IA__gtk_tree_model_get_flags"), visibility("default")));

#undef gtk_tree_model_get_iter
extern __typeof (gtk_tree_model_get_iter) gtk_tree_model_get_iter __attribute((alias("IA__gtk_tree_model_get_iter"), visibility("default")));

#undef gtk_tree_model_get_iter_first
extern __typeof (gtk_tree_model_get_iter_first) gtk_tree_model_get_iter_first __attribute((alias("IA__gtk_tree_model_get_iter_first"), visibility("default")));

#undef gtk_tree_model_get_iter_from_string
extern __typeof (gtk_tree_model_get_iter_from_string) gtk_tree_model_get_iter_from_string __attribute((alias("IA__gtk_tree_model_get_iter_from_string"), visibility("default")));

#undef gtk_tree_model_get_n_columns
extern __typeof (gtk_tree_model_get_n_columns) gtk_tree_model_get_n_columns __attribute((alias("IA__gtk_tree_model_get_n_columns"), visibility("default")));

#undef gtk_tree_model_get_path
extern __typeof (gtk_tree_model_get_path) gtk_tree_model_get_path __attribute((alias("IA__gtk_tree_model_get_path"), visibility("default")));

#undef gtk_tree_model_get_string_from_iter
extern __typeof (gtk_tree_model_get_string_from_iter) gtk_tree_model_get_string_from_iter __attribute((alias("IA__gtk_tree_model_get_string_from_iter"), visibility("default")));

#undef gtk_tree_model_get_type
extern __typeof (gtk_tree_model_get_type) gtk_tree_model_get_type __attribute((alias("IA__gtk_tree_model_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_model_get_valist
extern __typeof (gtk_tree_model_get_valist) gtk_tree_model_get_valist __attribute((alias("IA__gtk_tree_model_get_valist"), visibility("default")));

#undef gtk_tree_model_get_value
extern __typeof (gtk_tree_model_get_value) gtk_tree_model_get_value __attribute((alias("IA__gtk_tree_model_get_value"), visibility("default")));

#undef gtk_tree_model_iter_children
extern __typeof (gtk_tree_model_iter_children) gtk_tree_model_iter_children __attribute((alias("IA__gtk_tree_model_iter_children"), visibility("default")));

#undef gtk_tree_model_iter_has_child
extern __typeof (gtk_tree_model_iter_has_child) gtk_tree_model_iter_has_child __attribute((alias("IA__gtk_tree_model_iter_has_child"), visibility("default")));

#undef gtk_tree_model_iter_n_children
extern __typeof (gtk_tree_model_iter_n_children) gtk_tree_model_iter_n_children __attribute((alias("IA__gtk_tree_model_iter_n_children"), visibility("default")));

#undef gtk_tree_model_iter_next
extern __typeof (gtk_tree_model_iter_next) gtk_tree_model_iter_next __attribute((alias("IA__gtk_tree_model_iter_next"), visibility("default")));

#undef gtk_tree_model_iter_nth_child
extern __typeof (gtk_tree_model_iter_nth_child) gtk_tree_model_iter_nth_child __attribute((alias("IA__gtk_tree_model_iter_nth_child"), visibility("default")));

#undef gtk_tree_model_iter_parent
extern __typeof (gtk_tree_model_iter_parent) gtk_tree_model_iter_parent __attribute((alias("IA__gtk_tree_model_iter_parent"), visibility("default")));

#undef gtk_tree_model_ref_node
extern __typeof (gtk_tree_model_ref_node) gtk_tree_model_ref_node __attribute((alias("IA__gtk_tree_model_ref_node"), visibility("default")));

#undef gtk_tree_model_row_changed
extern __typeof (gtk_tree_model_row_changed) gtk_tree_model_row_changed __attribute((alias("IA__gtk_tree_model_row_changed"), visibility("default")));

#undef gtk_tree_model_row_deleted
extern __typeof (gtk_tree_model_row_deleted) gtk_tree_model_row_deleted __attribute((alias("IA__gtk_tree_model_row_deleted"), visibility("default")));

#undef gtk_tree_model_row_has_child_toggled
extern __typeof (gtk_tree_model_row_has_child_toggled) gtk_tree_model_row_has_child_toggled __attribute((alias("IA__gtk_tree_model_row_has_child_toggled"), visibility("default")));

#undef gtk_tree_model_row_inserted
extern __typeof (gtk_tree_model_row_inserted) gtk_tree_model_row_inserted __attribute((alias("IA__gtk_tree_model_row_inserted"), visibility("default")));

#undef gtk_tree_model_rows_reordered
extern __typeof (gtk_tree_model_rows_reordered) gtk_tree_model_rows_reordered __attribute((alias("IA__gtk_tree_model_rows_reordered"), visibility("default")));

#undef gtk_tree_model_unref_node
extern __typeof (gtk_tree_model_unref_node) gtk_tree_model_unref_node __attribute((alias("IA__gtk_tree_model_unref_node"), visibility("default")));

#undef gtk_tree_path_append_index
extern __typeof (gtk_tree_path_append_index) gtk_tree_path_append_index __attribute((alias("IA__gtk_tree_path_append_index"), visibility("default")));

#undef gtk_tree_path_compare
extern __typeof (gtk_tree_path_compare) gtk_tree_path_compare __attribute((alias("IA__gtk_tree_path_compare"), visibility("default")));

#undef gtk_tree_path_copy
extern __typeof (gtk_tree_path_copy) gtk_tree_path_copy __attribute((alias("IA__gtk_tree_path_copy"), visibility("default")));

#undef gtk_tree_path_down
extern __typeof (gtk_tree_path_down) gtk_tree_path_down __attribute((alias("IA__gtk_tree_path_down"), visibility("default")));

#undef gtk_tree_path_free
extern __typeof (gtk_tree_path_free) gtk_tree_path_free __attribute((alias("IA__gtk_tree_path_free"), visibility("default")));

#undef gtk_tree_path_get_depth
extern __typeof (gtk_tree_path_get_depth) gtk_tree_path_get_depth __attribute((alias("IA__gtk_tree_path_get_depth"), visibility("default")));

#undef gtk_tree_path_get_indices
extern __typeof (gtk_tree_path_get_indices) gtk_tree_path_get_indices __attribute((alias("IA__gtk_tree_path_get_indices"), visibility("default")));

#undef gtk_tree_path_get_indices_with_depth
extern __typeof (gtk_tree_path_get_indices_with_depth) gtk_tree_path_get_indices_with_depth __attribute((alias("IA__gtk_tree_path_get_indices_with_depth"), visibility("default")));

#undef gtk_tree_path_get_type
extern __typeof (gtk_tree_path_get_type) gtk_tree_path_get_type __attribute((alias("IA__gtk_tree_path_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_path_is_ancestor
extern __typeof (gtk_tree_path_is_ancestor) gtk_tree_path_is_ancestor __attribute((alias("IA__gtk_tree_path_is_ancestor"), visibility("default")));

#undef gtk_tree_path_is_descendant
extern __typeof (gtk_tree_path_is_descendant) gtk_tree_path_is_descendant __attribute((alias("IA__gtk_tree_path_is_descendant"), visibility("default")));

#undef gtk_tree_path_new
extern __typeof (gtk_tree_path_new) gtk_tree_path_new __attribute((alias("IA__gtk_tree_path_new"), visibility("default")));

#undef gtk_tree_path_new_first
extern __typeof (gtk_tree_path_new_first) gtk_tree_path_new_first __attribute((alias("IA__gtk_tree_path_new_first"), visibility("default")));

#undef gtk_tree_path_new_from_indices
extern __typeof (gtk_tree_path_new_from_indices) gtk_tree_path_new_from_indices __attribute((alias("IA__gtk_tree_path_new_from_indices"), visibility("default")));

#undef gtk_tree_path_new_from_string
extern __typeof (gtk_tree_path_new_from_string) gtk_tree_path_new_from_string __attribute((alias("IA__gtk_tree_path_new_from_string"), visibility("default")));

#undef gtk_tree_path_next
extern __typeof (gtk_tree_path_next) gtk_tree_path_next __attribute((alias("IA__gtk_tree_path_next"), visibility("default")));

#undef gtk_tree_path_prepend_index
extern __typeof (gtk_tree_path_prepend_index) gtk_tree_path_prepend_index __attribute((alias("IA__gtk_tree_path_prepend_index"), visibility("default")));

#undef gtk_tree_path_prev
extern __typeof (gtk_tree_path_prev) gtk_tree_path_prev __attribute((alias("IA__gtk_tree_path_prev"), visibility("default")));

#undef gtk_tree_path_to_string
extern __typeof (gtk_tree_path_to_string) gtk_tree_path_to_string __attribute((alias("IA__gtk_tree_path_to_string"), visibility("default")));

#undef gtk_tree_path_up
extern __typeof (gtk_tree_path_up) gtk_tree_path_up __attribute((alias("IA__gtk_tree_path_up"), visibility("default")));

#undef gtk_tree_row_reference_copy
extern __typeof (gtk_tree_row_reference_copy) gtk_tree_row_reference_copy __attribute((alias("IA__gtk_tree_row_reference_copy"), visibility("default")));

#undef gtk_tree_row_reference_deleted
extern __typeof (gtk_tree_row_reference_deleted) gtk_tree_row_reference_deleted __attribute((alias("IA__gtk_tree_row_reference_deleted"), visibility("default")));

#undef gtk_tree_row_reference_free
extern __typeof (gtk_tree_row_reference_free) gtk_tree_row_reference_free __attribute((alias("IA__gtk_tree_row_reference_free"), visibility("default")));

#undef gtk_tree_row_reference_get_model
extern __typeof (gtk_tree_row_reference_get_model) gtk_tree_row_reference_get_model __attribute((alias("IA__gtk_tree_row_reference_get_model"), visibility("default")));

#undef gtk_tree_row_reference_get_path
extern __typeof (gtk_tree_row_reference_get_path) gtk_tree_row_reference_get_path __attribute((alias("IA__gtk_tree_row_reference_get_path"), visibility("default")));

#undef gtk_tree_row_reference_get_type
extern __typeof (gtk_tree_row_reference_get_type) gtk_tree_row_reference_get_type __attribute((alias("IA__gtk_tree_row_reference_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_row_reference_inserted
extern __typeof (gtk_tree_row_reference_inserted) gtk_tree_row_reference_inserted __attribute((alias("IA__gtk_tree_row_reference_inserted"), visibility("default")));

#undef gtk_tree_row_reference_new
extern __typeof (gtk_tree_row_reference_new) gtk_tree_row_reference_new __attribute((alias("IA__gtk_tree_row_reference_new"), visibility("default")));

#undef gtk_tree_row_reference_new_proxy
extern __typeof (gtk_tree_row_reference_new_proxy) gtk_tree_row_reference_new_proxy __attribute((alias("IA__gtk_tree_row_reference_new_proxy"), visibility("default")));

#undef gtk_tree_row_reference_reordered
extern __typeof (gtk_tree_row_reference_reordered) gtk_tree_row_reference_reordered __attribute((alias("IA__gtk_tree_row_reference_reordered"), visibility("default")));

#undef gtk_tree_row_reference_valid
extern __typeof (gtk_tree_row_reference_valid) gtk_tree_row_reference_valid __attribute((alias("IA__gtk_tree_row_reference_valid"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TREE_MODEL_SORT_H__)
#if IN_FILE(__GTK_TREE_MODEL_SORT_C__)
#undef gtk_tree_model_sort_clear_cache
extern __typeof (gtk_tree_model_sort_clear_cache) gtk_tree_model_sort_clear_cache __attribute((alias("IA__gtk_tree_model_sort_clear_cache"), visibility("default")));

#undef gtk_tree_model_sort_convert_child_iter_to_iter
extern __typeof (gtk_tree_model_sort_convert_child_iter_to_iter) gtk_tree_model_sort_convert_child_iter_to_iter __attribute((alias("IA__gtk_tree_model_sort_convert_child_iter_to_iter"), visibility("default")));

#undef gtk_tree_model_sort_convert_child_path_to_path
extern __typeof (gtk_tree_model_sort_convert_child_path_to_path) gtk_tree_model_sort_convert_child_path_to_path __attribute((alias("IA__gtk_tree_model_sort_convert_child_path_to_path"), visibility("default")));

#undef gtk_tree_model_sort_convert_iter_to_child_iter
extern __typeof (gtk_tree_model_sort_convert_iter_to_child_iter) gtk_tree_model_sort_convert_iter_to_child_iter __attribute((alias("IA__gtk_tree_model_sort_convert_iter_to_child_iter"), visibility("default")));

#undef gtk_tree_model_sort_convert_path_to_child_path
extern __typeof (gtk_tree_model_sort_convert_path_to_child_path) gtk_tree_model_sort_convert_path_to_child_path __attribute((alias("IA__gtk_tree_model_sort_convert_path_to_child_path"), visibility("default")));

#undef gtk_tree_model_sort_get_model
extern __typeof (gtk_tree_model_sort_get_model) gtk_tree_model_sort_get_model __attribute((alias("IA__gtk_tree_model_sort_get_model"), visibility("default")));

#undef gtk_tree_model_sort_get_type
extern __typeof (gtk_tree_model_sort_get_type) gtk_tree_model_sort_get_type __attribute((alias("IA__gtk_tree_model_sort_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_model_sort_iter_is_valid
extern __typeof (gtk_tree_model_sort_iter_is_valid) gtk_tree_model_sort_iter_is_valid __attribute((alias("IA__gtk_tree_model_sort_iter_is_valid"), visibility("default")));

#undef gtk_tree_model_sort_new_with_model
extern __typeof (gtk_tree_model_sort_new_with_model) gtk_tree_model_sort_new_with_model __attribute((alias("IA__gtk_tree_model_sort_new_with_model"), visibility("default")));

#undef gtk_tree_model_sort_reset_default_sort_func
extern __typeof (gtk_tree_model_sort_reset_default_sort_func) gtk_tree_model_sort_reset_default_sort_func __attribute((alias("IA__gtk_tree_model_sort_reset_default_sort_func"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TREE_MODEL_FILTER_H__)
#if IN_FILE(__GTK_TREE_MODEL_FILTER_C__)
#undef gtk_tree_model_filter_clear_cache
extern __typeof (gtk_tree_model_filter_clear_cache) gtk_tree_model_filter_clear_cache __attribute((alias("IA__gtk_tree_model_filter_clear_cache"), visibility("default")));

#undef gtk_tree_model_filter_convert_child_iter_to_iter
extern __typeof (gtk_tree_model_filter_convert_child_iter_to_iter) gtk_tree_model_filter_convert_child_iter_to_iter __attribute((alias("IA__gtk_tree_model_filter_convert_child_iter_to_iter"), visibility("default")));

#undef gtk_tree_model_filter_convert_child_path_to_path
extern __typeof (gtk_tree_model_filter_convert_child_path_to_path) gtk_tree_model_filter_convert_child_path_to_path __attribute((alias("IA__gtk_tree_model_filter_convert_child_path_to_path"), visibility("default")));

#undef gtk_tree_model_filter_convert_iter_to_child_iter
extern __typeof (gtk_tree_model_filter_convert_iter_to_child_iter) gtk_tree_model_filter_convert_iter_to_child_iter __attribute((alias("IA__gtk_tree_model_filter_convert_iter_to_child_iter"), visibility("default")));

#undef gtk_tree_model_filter_convert_path_to_child_path
extern __typeof (gtk_tree_model_filter_convert_path_to_child_path) gtk_tree_model_filter_convert_path_to_child_path __attribute((alias("IA__gtk_tree_model_filter_convert_path_to_child_path"), visibility("default")));

#undef gtk_tree_model_filter_get_model
extern __typeof (gtk_tree_model_filter_get_model) gtk_tree_model_filter_get_model __attribute((alias("IA__gtk_tree_model_filter_get_model"), visibility("default")));

#undef gtk_tree_model_filter_get_type
extern __typeof (gtk_tree_model_filter_get_type) gtk_tree_model_filter_get_type __attribute((alias("IA__gtk_tree_model_filter_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_model_filter_new
extern __typeof (gtk_tree_model_filter_new) gtk_tree_model_filter_new __attribute((alias("IA__gtk_tree_model_filter_new"), visibility("default")));

#undef gtk_tree_model_filter_refilter
extern __typeof (gtk_tree_model_filter_refilter) gtk_tree_model_filter_refilter __attribute((alias("IA__gtk_tree_model_filter_refilter"), visibility("default")));

#undef gtk_tree_model_filter_set_modify_func
extern __typeof (gtk_tree_model_filter_set_modify_func) gtk_tree_model_filter_set_modify_func __attribute((alias("IA__gtk_tree_model_filter_set_modify_func"), visibility("default")));

#undef gtk_tree_model_filter_set_visible_column
extern __typeof (gtk_tree_model_filter_set_visible_column) gtk_tree_model_filter_set_visible_column __attribute((alias("IA__gtk_tree_model_filter_set_visible_column"), visibility("default")));

#undef gtk_tree_model_filter_set_visible_func
extern __typeof (gtk_tree_model_filter_set_visible_func) gtk_tree_model_filter_set_visible_func __attribute((alias("IA__gtk_tree_model_filter_set_visible_func"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TREE_SELECTION_H__)
#if IN_FILE(__GTK_TREE_SELECTION_C__)
#undef gtk_tree_selection_count_selected_rows
extern __typeof (gtk_tree_selection_count_selected_rows) gtk_tree_selection_count_selected_rows __attribute((alias("IA__gtk_tree_selection_count_selected_rows"), visibility("default")));

#undef gtk_tree_selection_get_mode
extern __typeof (gtk_tree_selection_get_mode) gtk_tree_selection_get_mode __attribute((alias("IA__gtk_tree_selection_get_mode"), visibility("default")));

#undef gtk_tree_selection_get_select_function
extern __typeof (gtk_tree_selection_get_select_function) gtk_tree_selection_get_select_function __attribute((alias("IA__gtk_tree_selection_get_select_function"), visibility("default")));

#undef gtk_tree_selection_get_selected
extern __typeof (gtk_tree_selection_get_selected) gtk_tree_selection_get_selected __attribute((alias("IA__gtk_tree_selection_get_selected"), visibility("default")));

#undef gtk_tree_selection_get_selected_rows
extern __typeof (gtk_tree_selection_get_selected_rows) gtk_tree_selection_get_selected_rows __attribute((alias("IA__gtk_tree_selection_get_selected_rows"), visibility("default")));

#undef gtk_tree_selection_get_tree_view
extern __typeof (gtk_tree_selection_get_tree_view) gtk_tree_selection_get_tree_view __attribute((alias("IA__gtk_tree_selection_get_tree_view"), visibility("default")));

#undef gtk_tree_selection_get_type
extern __typeof (gtk_tree_selection_get_type) gtk_tree_selection_get_type __attribute((alias("IA__gtk_tree_selection_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_selection_get_user_data
extern __typeof (gtk_tree_selection_get_user_data) gtk_tree_selection_get_user_data __attribute((alias("IA__gtk_tree_selection_get_user_data"), visibility("default")));

#undef gtk_tree_selection_iter_is_selected
extern __typeof (gtk_tree_selection_iter_is_selected) gtk_tree_selection_iter_is_selected __attribute((alias("IA__gtk_tree_selection_iter_is_selected"), visibility("default")));

#undef gtk_tree_selection_path_is_selected
extern __typeof (gtk_tree_selection_path_is_selected) gtk_tree_selection_path_is_selected __attribute((alias("IA__gtk_tree_selection_path_is_selected"), visibility("default")));

#undef gtk_tree_selection_select_all
extern __typeof (gtk_tree_selection_select_all) gtk_tree_selection_select_all __attribute((alias("IA__gtk_tree_selection_select_all"), visibility("default")));

#undef gtk_tree_selection_selected_foreach
extern __typeof (gtk_tree_selection_selected_foreach) gtk_tree_selection_selected_foreach __attribute((alias("IA__gtk_tree_selection_selected_foreach"), visibility("default")));

#undef gtk_tree_selection_select_iter
extern __typeof (gtk_tree_selection_select_iter) gtk_tree_selection_select_iter __attribute((alias("IA__gtk_tree_selection_select_iter"), visibility("default")));

#undef gtk_tree_selection_select_path
extern __typeof (gtk_tree_selection_select_path) gtk_tree_selection_select_path __attribute((alias("IA__gtk_tree_selection_select_path"), visibility("default")));

#undef gtk_tree_selection_select_range
extern __typeof (gtk_tree_selection_select_range) gtk_tree_selection_select_range __attribute((alias("IA__gtk_tree_selection_select_range"), visibility("default")));

#undef gtk_tree_selection_set_mode
extern __typeof (gtk_tree_selection_set_mode) gtk_tree_selection_set_mode __attribute((alias("IA__gtk_tree_selection_set_mode"), visibility("default")));

#undef gtk_tree_selection_set_select_function
extern __typeof (gtk_tree_selection_set_select_function) gtk_tree_selection_set_select_function __attribute((alias("IA__gtk_tree_selection_set_select_function"), visibility("default")));

#undef gtk_tree_selection_unselect_all
extern __typeof (gtk_tree_selection_unselect_all) gtk_tree_selection_unselect_all __attribute((alias("IA__gtk_tree_selection_unselect_all"), visibility("default")));

#undef gtk_tree_selection_unselect_iter
extern __typeof (gtk_tree_selection_unselect_iter) gtk_tree_selection_unselect_iter __attribute((alias("IA__gtk_tree_selection_unselect_iter"), visibility("default")));

#undef gtk_tree_selection_unselect_path
extern __typeof (gtk_tree_selection_unselect_path) gtk_tree_selection_unselect_path __attribute((alias("IA__gtk_tree_selection_unselect_path"), visibility("default")));

#undef gtk_tree_selection_unselect_range
extern __typeof (gtk_tree_selection_unselect_range) gtk_tree_selection_unselect_range __attribute((alias("IA__gtk_tree_selection_unselect_range"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TREE_SORTABLE_H__)
#if IN_FILE(__GTK_TREE_SORTABLE_C__)
#undef gtk_tree_sortable_get_sort_column_id
extern __typeof (gtk_tree_sortable_get_sort_column_id) gtk_tree_sortable_get_sort_column_id __attribute((alias("IA__gtk_tree_sortable_get_sort_column_id"), visibility("default")));

#undef gtk_tree_sortable_get_type
extern __typeof (gtk_tree_sortable_get_type) gtk_tree_sortable_get_type __attribute((alias("IA__gtk_tree_sortable_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_sortable_has_default_sort_func
extern __typeof (gtk_tree_sortable_has_default_sort_func) gtk_tree_sortable_has_default_sort_func __attribute((alias("IA__gtk_tree_sortable_has_default_sort_func"), visibility("default")));

#undef gtk_tree_sortable_set_default_sort_func
extern __typeof (gtk_tree_sortable_set_default_sort_func) gtk_tree_sortable_set_default_sort_func __attribute((alias("IA__gtk_tree_sortable_set_default_sort_func"), visibility("default")));

#undef gtk_tree_sortable_set_sort_column_id
extern __typeof (gtk_tree_sortable_set_sort_column_id) gtk_tree_sortable_set_sort_column_id __attribute((alias("IA__gtk_tree_sortable_set_sort_column_id"), visibility("default")));

#undef gtk_tree_sortable_set_sort_func
extern __typeof (gtk_tree_sortable_set_sort_func) gtk_tree_sortable_set_sort_func __attribute((alias("IA__gtk_tree_sortable_set_sort_func"), visibility("default")));

#undef gtk_tree_sortable_sort_column_changed
extern __typeof (gtk_tree_sortable_sort_column_changed) gtk_tree_sortable_sort_column_changed __attribute((alias("IA__gtk_tree_sortable_sort_column_changed"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TREE_STORE_H__)
#if IN_FILE(__GTK_TREE_STORE_C__)
#undef gtk_tree_store_append
extern __typeof (gtk_tree_store_append) gtk_tree_store_append __attribute((alias("IA__gtk_tree_store_append"), visibility("default")));

#undef gtk_tree_store_clear
extern __typeof (gtk_tree_store_clear) gtk_tree_store_clear __attribute((alias("IA__gtk_tree_store_clear"), visibility("default")));

#undef gtk_tree_store_get_type
extern __typeof (gtk_tree_store_get_type) gtk_tree_store_get_type __attribute((alias("IA__gtk_tree_store_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_store_insert
extern __typeof (gtk_tree_store_insert) gtk_tree_store_insert __attribute((alias("IA__gtk_tree_store_insert"), visibility("default")));

#undef gtk_tree_store_insert_after
extern __typeof (gtk_tree_store_insert_after) gtk_tree_store_insert_after __attribute((alias("IA__gtk_tree_store_insert_after"), visibility("default")));

#undef gtk_tree_store_insert_before
extern __typeof (gtk_tree_store_insert_before) gtk_tree_store_insert_before __attribute((alias("IA__gtk_tree_store_insert_before"), visibility("default")));

#undef gtk_tree_store_insert_with_values
extern __typeof (gtk_tree_store_insert_with_values) gtk_tree_store_insert_with_values __attribute((alias("IA__gtk_tree_store_insert_with_values"), visibility("default")));

#undef gtk_tree_store_insert_with_valuesv
extern __typeof (gtk_tree_store_insert_with_valuesv) gtk_tree_store_insert_with_valuesv __attribute((alias("IA__gtk_tree_store_insert_with_valuesv"), visibility("default")));

#undef gtk_tree_store_is_ancestor
extern __typeof (gtk_tree_store_is_ancestor) gtk_tree_store_is_ancestor __attribute((alias("IA__gtk_tree_store_is_ancestor"), visibility("default")));

#undef gtk_tree_store_iter_depth
extern __typeof (gtk_tree_store_iter_depth) gtk_tree_store_iter_depth __attribute((alias("IA__gtk_tree_store_iter_depth"), visibility("default")));

#undef gtk_tree_store_iter_is_valid
extern __typeof (gtk_tree_store_iter_is_valid) gtk_tree_store_iter_is_valid __attribute((alias("IA__gtk_tree_store_iter_is_valid"), visibility("default")));

#undef gtk_tree_store_move_after
extern __typeof (gtk_tree_store_move_after) gtk_tree_store_move_after __attribute((alias("IA__gtk_tree_store_move_after"), visibility("default")));

#undef gtk_tree_store_move_before
extern __typeof (gtk_tree_store_move_before) gtk_tree_store_move_before __attribute((alias("IA__gtk_tree_store_move_before"), visibility("default")));

#undef gtk_tree_store_new
extern __typeof (gtk_tree_store_new) gtk_tree_store_new __attribute((alias("IA__gtk_tree_store_new"), visibility("default")));

#undef gtk_tree_store_newv
extern __typeof (gtk_tree_store_newv) gtk_tree_store_newv __attribute((alias("IA__gtk_tree_store_newv"), visibility("default")));

#undef gtk_tree_store_prepend
extern __typeof (gtk_tree_store_prepend) gtk_tree_store_prepend __attribute((alias("IA__gtk_tree_store_prepend"), visibility("default")));

#undef gtk_tree_store_remove
extern __typeof (gtk_tree_store_remove) gtk_tree_store_remove __attribute((alias("IA__gtk_tree_store_remove"), visibility("default")));

#undef gtk_tree_store_reorder
extern __typeof (gtk_tree_store_reorder) gtk_tree_store_reorder __attribute((alias("IA__gtk_tree_store_reorder"), visibility("default")));

#undef gtk_tree_store_set
extern __typeof (gtk_tree_store_set) gtk_tree_store_set __attribute((alias("IA__gtk_tree_store_set"), visibility("default")));

#undef gtk_tree_store_set_column_types
extern __typeof (gtk_tree_store_set_column_types) gtk_tree_store_set_column_types __attribute((alias("IA__gtk_tree_store_set_column_types"), visibility("default")));

#undef gtk_tree_store_set_valist
extern __typeof (gtk_tree_store_set_valist) gtk_tree_store_set_valist __attribute((alias("IA__gtk_tree_store_set_valist"), visibility("default")));

#undef gtk_tree_store_set_value
extern __typeof (gtk_tree_store_set_value) gtk_tree_store_set_value __attribute((alias("IA__gtk_tree_store_set_value"), visibility("default")));

#undef gtk_tree_store_set_valuesv
extern __typeof (gtk_tree_store_set_valuesv) gtk_tree_store_set_valuesv __attribute((alias("IA__gtk_tree_store_set_valuesv"), visibility("default")));

#undef gtk_tree_store_swap
extern __typeof (gtk_tree_store_swap) gtk_tree_store_swap __attribute((alias("IA__gtk_tree_store_swap"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TREE_VIEW_COLUMN_H__)
#if IN_FILE(__GTK_TREE_VIEW_COLUMN_C__)
#undef gtk_tree_view_column_add_attribute
extern __typeof (gtk_tree_view_column_add_attribute) gtk_tree_view_column_add_attribute __attribute((alias("IA__gtk_tree_view_column_add_attribute"), visibility("default")));

#undef gtk_tree_view_column_cell_get_position
extern __typeof (gtk_tree_view_column_cell_get_position) gtk_tree_view_column_cell_get_position __attribute((alias("IA__gtk_tree_view_column_cell_get_position"), visibility("default")));

#undef gtk_tree_view_column_cell_get_size
extern __typeof (gtk_tree_view_column_cell_get_size) gtk_tree_view_column_cell_get_size __attribute((alias("IA__gtk_tree_view_column_cell_get_size"), visibility("default")));

#undef gtk_tree_view_column_cell_is_visible
extern __typeof (gtk_tree_view_column_cell_is_visible) gtk_tree_view_column_cell_is_visible __attribute((alias("IA__gtk_tree_view_column_cell_is_visible"), visibility("default")));

#undef gtk_tree_view_column_cell_set_cell_data
extern __typeof (gtk_tree_view_column_cell_set_cell_data) gtk_tree_view_column_cell_set_cell_data __attribute((alias("IA__gtk_tree_view_column_cell_set_cell_data"), visibility("default")));

#undef gtk_tree_view_column_clear
extern __typeof (gtk_tree_view_column_clear) gtk_tree_view_column_clear __attribute((alias("IA__gtk_tree_view_column_clear"), visibility("default")));

#undef gtk_tree_view_column_clear_attributes
extern __typeof (gtk_tree_view_column_clear_attributes) gtk_tree_view_column_clear_attributes __attribute((alias("IA__gtk_tree_view_column_clear_attributes"), visibility("default")));

#undef gtk_tree_view_column_clicked
extern __typeof (gtk_tree_view_column_clicked) gtk_tree_view_column_clicked __attribute((alias("IA__gtk_tree_view_column_clicked"), visibility("default")));

#undef gtk_tree_view_column_focus_cell
extern __typeof (gtk_tree_view_column_focus_cell) gtk_tree_view_column_focus_cell __attribute((alias("IA__gtk_tree_view_column_focus_cell"), visibility("default")));

#undef gtk_tree_view_column_get_alignment
extern __typeof (gtk_tree_view_column_get_alignment) gtk_tree_view_column_get_alignment __attribute((alias("IA__gtk_tree_view_column_get_alignment"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_tree_view_column_get_cell_renderers
extern __typeof (gtk_tree_view_column_get_cell_renderers) gtk_tree_view_column_get_cell_renderers __attribute((alias("IA__gtk_tree_view_column_get_cell_renderers"), visibility("default")));

#endif
#undef gtk_tree_view_column_get_clickable
extern __typeof (gtk_tree_view_column_get_clickable) gtk_tree_view_column_get_clickable __attribute((alias("IA__gtk_tree_view_column_get_clickable"), visibility("default")));

#undef gtk_tree_view_column_get_expand
extern __typeof (gtk_tree_view_column_get_expand) gtk_tree_view_column_get_expand __attribute((alias("IA__gtk_tree_view_column_get_expand"), visibility("default")));

#undef gtk_tree_view_column_get_fixed_width
extern __typeof (gtk_tree_view_column_get_fixed_width) gtk_tree_view_column_get_fixed_width __attribute((alias("IA__gtk_tree_view_column_get_fixed_width"), visibility("default")));

#undef gtk_tree_view_column_get_max_width
extern __typeof (gtk_tree_view_column_get_max_width) gtk_tree_view_column_get_max_width __attribute((alias("IA__gtk_tree_view_column_get_max_width"), visibility("default")));

#undef gtk_tree_view_column_get_min_width
extern __typeof (gtk_tree_view_column_get_min_width) gtk_tree_view_column_get_min_width __attribute((alias("IA__gtk_tree_view_column_get_min_width"), visibility("default")));

#undef gtk_tree_view_column_get_reorderable
extern __typeof (gtk_tree_view_column_get_reorderable) gtk_tree_view_column_get_reorderable __attribute((alias("IA__gtk_tree_view_column_get_reorderable"), visibility("default")));

#undef gtk_tree_view_column_get_resizable
extern __typeof (gtk_tree_view_column_get_resizable) gtk_tree_view_column_get_resizable __attribute((alias("IA__gtk_tree_view_column_get_resizable"), visibility("default")));

#undef gtk_tree_view_column_get_sizing
extern __typeof (gtk_tree_view_column_get_sizing) gtk_tree_view_column_get_sizing __attribute((alias("IA__gtk_tree_view_column_get_sizing"), visibility("default")));

#undef gtk_tree_view_column_get_sort_column_id
extern __typeof (gtk_tree_view_column_get_sort_column_id) gtk_tree_view_column_get_sort_column_id __attribute((alias("IA__gtk_tree_view_column_get_sort_column_id"), visibility("default")));

#undef gtk_tree_view_column_get_sort_indicator
extern __typeof (gtk_tree_view_column_get_sort_indicator) gtk_tree_view_column_get_sort_indicator __attribute((alias("IA__gtk_tree_view_column_get_sort_indicator"), visibility("default")));

#undef gtk_tree_view_column_get_sort_order
extern __typeof (gtk_tree_view_column_get_sort_order) gtk_tree_view_column_get_sort_order __attribute((alias("IA__gtk_tree_view_column_get_sort_order"), visibility("default")));

#undef gtk_tree_view_column_get_spacing
extern __typeof (gtk_tree_view_column_get_spacing) gtk_tree_view_column_get_spacing __attribute((alias("IA__gtk_tree_view_column_get_spacing"), visibility("default")));

#undef gtk_tree_view_column_get_title
extern __typeof (gtk_tree_view_column_get_title) gtk_tree_view_column_get_title __attribute((alias("IA__gtk_tree_view_column_get_title"), visibility("default")));

#undef gtk_tree_view_column_get_tree_view
extern __typeof (gtk_tree_view_column_get_tree_view) gtk_tree_view_column_get_tree_view __attribute((alias("IA__gtk_tree_view_column_get_tree_view"), visibility("default")));

#undef gtk_tree_view_column_get_type
extern __typeof (gtk_tree_view_column_get_type) gtk_tree_view_column_get_type __attribute((alias("IA__gtk_tree_view_column_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_view_column_get_visible
extern __typeof (gtk_tree_view_column_get_visible) gtk_tree_view_column_get_visible __attribute((alias("IA__gtk_tree_view_column_get_visible"), visibility("default")));

#undef gtk_tree_view_column_get_widget
extern __typeof (gtk_tree_view_column_get_widget) gtk_tree_view_column_get_widget __attribute((alias("IA__gtk_tree_view_column_get_widget"), visibility("default")));

#undef gtk_tree_view_column_get_width
extern __typeof (gtk_tree_view_column_get_width) gtk_tree_view_column_get_width __attribute((alias("IA__gtk_tree_view_column_get_width"), visibility("default")));

#undef gtk_tree_view_column_new
extern __typeof (gtk_tree_view_column_new) gtk_tree_view_column_new __attribute((alias("IA__gtk_tree_view_column_new"), visibility("default")));

#undef gtk_tree_view_column_new_with_attributes
extern __typeof (gtk_tree_view_column_new_with_attributes) gtk_tree_view_column_new_with_attributes __attribute((alias("IA__gtk_tree_view_column_new_with_attributes"), visibility("default"))) G_GNUC_NULL_TERMINATED;

#undef gtk_tree_view_column_pack_end
extern __typeof (gtk_tree_view_column_pack_end) gtk_tree_view_column_pack_end __attribute((alias("IA__gtk_tree_view_column_pack_end"), visibility("default")));

#undef gtk_tree_view_column_queue_resize
extern __typeof (gtk_tree_view_column_queue_resize) gtk_tree_view_column_queue_resize __attribute((alias("IA__gtk_tree_view_column_queue_resize"), visibility("default")));

#undef gtk_tree_view_column_pack_start
extern __typeof (gtk_tree_view_column_pack_start) gtk_tree_view_column_pack_start __attribute((alias("IA__gtk_tree_view_column_pack_start"), visibility("default")));

#undef gtk_tree_view_column_set_alignment
extern __typeof (gtk_tree_view_column_set_alignment) gtk_tree_view_column_set_alignment __attribute((alias("IA__gtk_tree_view_column_set_alignment"), visibility("default")));

#undef gtk_tree_view_column_set_attributes
extern __typeof (gtk_tree_view_column_set_attributes) gtk_tree_view_column_set_attributes __attribute((alias("IA__gtk_tree_view_column_set_attributes"), visibility("default")));

#undef gtk_tree_view_column_set_cell_data_func
extern __typeof (gtk_tree_view_column_set_cell_data_func) gtk_tree_view_column_set_cell_data_func __attribute((alias("IA__gtk_tree_view_column_set_cell_data_func"), visibility("default")));

#undef gtk_tree_view_column_set_clickable
extern __typeof (gtk_tree_view_column_set_clickable) gtk_tree_view_column_set_clickable __attribute((alias("IA__gtk_tree_view_column_set_clickable"), visibility("default")));

#undef gtk_tree_view_column_set_expand
extern __typeof (gtk_tree_view_column_set_expand) gtk_tree_view_column_set_expand __attribute((alias("IA__gtk_tree_view_column_set_expand"), visibility("default")));

#undef gtk_tree_view_column_set_fixed_width
extern __typeof (gtk_tree_view_column_set_fixed_width) gtk_tree_view_column_set_fixed_width __attribute((alias("IA__gtk_tree_view_column_set_fixed_width"), visibility("default")));

#undef gtk_tree_view_column_set_max_width
extern __typeof (gtk_tree_view_column_set_max_width) gtk_tree_view_column_set_max_width __attribute((alias("IA__gtk_tree_view_column_set_max_width"), visibility("default")));

#undef gtk_tree_view_column_set_min_width
extern __typeof (gtk_tree_view_column_set_min_width) gtk_tree_view_column_set_min_width __attribute((alias("IA__gtk_tree_view_column_set_min_width"), visibility("default")));

#undef gtk_tree_view_column_set_reorderable
extern __typeof (gtk_tree_view_column_set_reorderable) gtk_tree_view_column_set_reorderable __attribute((alias("IA__gtk_tree_view_column_set_reorderable"), visibility("default")));

#undef gtk_tree_view_column_set_resizable
extern __typeof (gtk_tree_view_column_set_resizable) gtk_tree_view_column_set_resizable __attribute((alias("IA__gtk_tree_view_column_set_resizable"), visibility("default")));

#undef gtk_tree_view_column_set_sizing
extern __typeof (gtk_tree_view_column_set_sizing) gtk_tree_view_column_set_sizing __attribute((alias("IA__gtk_tree_view_column_set_sizing"), visibility("default")));

#undef gtk_tree_view_column_set_sort_column_id
extern __typeof (gtk_tree_view_column_set_sort_column_id) gtk_tree_view_column_set_sort_column_id __attribute((alias("IA__gtk_tree_view_column_set_sort_column_id"), visibility("default")));

#undef gtk_tree_view_column_set_sort_indicator
extern __typeof (gtk_tree_view_column_set_sort_indicator) gtk_tree_view_column_set_sort_indicator __attribute((alias("IA__gtk_tree_view_column_set_sort_indicator"), visibility("default")));

#undef gtk_tree_view_column_set_sort_order
extern __typeof (gtk_tree_view_column_set_sort_order) gtk_tree_view_column_set_sort_order __attribute((alias("IA__gtk_tree_view_column_set_sort_order"), visibility("default")));

#undef gtk_tree_view_column_set_spacing
extern __typeof (gtk_tree_view_column_set_spacing) gtk_tree_view_column_set_spacing __attribute((alias("IA__gtk_tree_view_column_set_spacing"), visibility("default")));

#undef gtk_tree_view_column_set_title
extern __typeof (gtk_tree_view_column_set_title) gtk_tree_view_column_set_title __attribute((alias("IA__gtk_tree_view_column_set_title"), visibility("default")));

#undef gtk_tree_view_column_set_visible
extern __typeof (gtk_tree_view_column_set_visible) gtk_tree_view_column_set_visible __attribute((alias("IA__gtk_tree_view_column_set_visible"), visibility("default")));

#undef gtk_tree_view_column_set_widget
extern __typeof (gtk_tree_view_column_set_widget) gtk_tree_view_column_set_widget __attribute((alias("IA__gtk_tree_view_column_set_widget"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TREE_VIEW_H__)
#if IN_FILE(__GTK_TREE_VIEW_C__)
#undef gtk_tree_view_append_column
extern __typeof (gtk_tree_view_append_column) gtk_tree_view_append_column __attribute((alias("IA__gtk_tree_view_append_column"), visibility("default")));

#undef gtk_tree_view_collapse_all
extern __typeof (gtk_tree_view_collapse_all) gtk_tree_view_collapse_all __attribute((alias("IA__gtk_tree_view_collapse_all"), visibility("default")));

#undef gtk_tree_view_collapse_row
extern __typeof (gtk_tree_view_collapse_row) gtk_tree_view_collapse_row __attribute((alias("IA__gtk_tree_view_collapse_row"), visibility("default")));

#undef gtk_tree_view_convert_bin_window_to_tree_coords
extern __typeof (gtk_tree_view_convert_bin_window_to_tree_coords) gtk_tree_view_convert_bin_window_to_tree_coords __attribute((alias("IA__gtk_tree_view_convert_bin_window_to_tree_coords"), visibility("default")));

#undef gtk_tree_view_convert_bin_window_to_widget_coords
extern __typeof (gtk_tree_view_convert_bin_window_to_widget_coords) gtk_tree_view_convert_bin_window_to_widget_coords __attribute((alias("IA__gtk_tree_view_convert_bin_window_to_widget_coords"), visibility("default")));

#undef gtk_tree_view_convert_tree_to_bin_window_coords
extern __typeof (gtk_tree_view_convert_tree_to_bin_window_coords) gtk_tree_view_convert_tree_to_bin_window_coords __attribute((alias("IA__gtk_tree_view_convert_tree_to_bin_window_coords"), visibility("default")));

#undef gtk_tree_view_convert_tree_to_widget_coords
extern __typeof (gtk_tree_view_convert_tree_to_widget_coords) gtk_tree_view_convert_tree_to_widget_coords __attribute((alias("IA__gtk_tree_view_convert_tree_to_widget_coords"), visibility("default")));

#undef gtk_tree_view_convert_widget_to_bin_window_coords
extern __typeof (gtk_tree_view_convert_widget_to_bin_window_coords) gtk_tree_view_convert_widget_to_bin_window_coords __attribute((alias("IA__gtk_tree_view_convert_widget_to_bin_window_coords"), visibility("default")));

#undef gtk_tree_view_convert_widget_to_tree_coords
extern __typeof (gtk_tree_view_convert_widget_to_tree_coords) gtk_tree_view_convert_widget_to_tree_coords __attribute((alias("IA__gtk_tree_view_convert_widget_to_tree_coords"), visibility("default")));

#undef gtk_tree_view_create_row_drag_icon
extern __typeof (gtk_tree_view_create_row_drag_icon) gtk_tree_view_create_row_drag_icon __attribute((alias("IA__gtk_tree_view_create_row_drag_icon"), visibility("default")));

#undef gtk_tree_view_enable_model_drag_dest
extern __typeof (gtk_tree_view_enable_model_drag_dest) gtk_tree_view_enable_model_drag_dest __attribute((alias("IA__gtk_tree_view_enable_model_drag_dest"), visibility("default")));

#undef gtk_tree_view_enable_model_drag_source
extern __typeof (gtk_tree_view_enable_model_drag_source) gtk_tree_view_enable_model_drag_source __attribute((alias("IA__gtk_tree_view_enable_model_drag_source"), visibility("default")));

#undef gtk_tree_view_expand_all
extern __typeof (gtk_tree_view_expand_all) gtk_tree_view_expand_all __attribute((alias("IA__gtk_tree_view_expand_all"), visibility("default")));

#undef gtk_tree_view_expand_row
extern __typeof (gtk_tree_view_expand_row) gtk_tree_view_expand_row __attribute((alias("IA__gtk_tree_view_expand_row"), visibility("default")));

#undef gtk_tree_view_expand_to_path
extern __typeof (gtk_tree_view_expand_to_path) gtk_tree_view_expand_to_path __attribute((alias("IA__gtk_tree_view_expand_to_path"), visibility("default")));

#undef gtk_tree_view_get_background_area
extern __typeof (gtk_tree_view_get_background_area) gtk_tree_view_get_background_area __attribute((alias("IA__gtk_tree_view_get_background_area"), visibility("default")));

#undef gtk_tree_view_get_bin_window
extern __typeof (gtk_tree_view_get_bin_window) gtk_tree_view_get_bin_window __attribute((alias("IA__gtk_tree_view_get_bin_window"), visibility("default")));

#undef gtk_tree_view_get_cell_area
extern __typeof (gtk_tree_view_get_cell_area) gtk_tree_view_get_cell_area __attribute((alias("IA__gtk_tree_view_get_cell_area"), visibility("default")));

#undef gtk_tree_view_get_column
extern __typeof (gtk_tree_view_get_column) gtk_tree_view_get_column __attribute((alias("IA__gtk_tree_view_get_column"), visibility("default")));

#undef gtk_tree_view_get_columns
extern __typeof (gtk_tree_view_get_columns) gtk_tree_view_get_columns __attribute((alias("IA__gtk_tree_view_get_columns"), visibility("default")));

#undef gtk_tree_view_get_cursor
extern __typeof (gtk_tree_view_get_cursor) gtk_tree_view_get_cursor __attribute((alias("IA__gtk_tree_view_get_cursor"), visibility("default")));

#undef gtk_tree_view_get_dest_row_at_pos
extern __typeof (gtk_tree_view_get_dest_row_at_pos) gtk_tree_view_get_dest_row_at_pos __attribute((alias("IA__gtk_tree_view_get_dest_row_at_pos"), visibility("default")));

#undef gtk_tree_view_get_drag_dest_row
extern __typeof (gtk_tree_view_get_drag_dest_row) gtk_tree_view_get_drag_dest_row __attribute((alias("IA__gtk_tree_view_get_drag_dest_row"), visibility("default")));

#undef gtk_tree_view_get_enable_search
extern __typeof (gtk_tree_view_get_enable_search) gtk_tree_view_get_enable_search __attribute((alias("IA__gtk_tree_view_get_enable_search"), visibility("default")));

#undef gtk_tree_view_get_enable_tree_lines
extern __typeof (gtk_tree_view_get_enable_tree_lines) gtk_tree_view_get_enable_tree_lines __attribute((alias("IA__gtk_tree_view_get_enable_tree_lines"), visibility("default")));

#undef gtk_tree_view_get_expander_column
extern __typeof (gtk_tree_view_get_expander_column) gtk_tree_view_get_expander_column __attribute((alias("IA__gtk_tree_view_get_expander_column"), visibility("default")));

#undef gtk_tree_view_get_fixed_height_mode
extern __typeof (gtk_tree_view_get_fixed_height_mode) gtk_tree_view_get_fixed_height_mode __attribute((alias("IA__gtk_tree_view_get_fixed_height_mode"), visibility("default")));

#undef gtk_tree_view_get_grid_lines
extern __typeof (gtk_tree_view_get_grid_lines) gtk_tree_view_get_grid_lines __attribute((alias("IA__gtk_tree_view_get_grid_lines"), visibility("default")));

#undef gtk_tree_view_get_hadjustment
extern __typeof (gtk_tree_view_get_hadjustment) gtk_tree_view_get_hadjustment __attribute((alias("IA__gtk_tree_view_get_hadjustment"), visibility("default")));

#undef gtk_tree_view_get_headers_clickable
extern __typeof (gtk_tree_view_get_headers_clickable) gtk_tree_view_get_headers_clickable __attribute((alias("IA__gtk_tree_view_get_headers_clickable"), visibility("default")));

#undef gtk_tree_view_get_headers_visible
extern __typeof (gtk_tree_view_get_headers_visible) gtk_tree_view_get_headers_visible __attribute((alias("IA__gtk_tree_view_get_headers_visible"), visibility("default")));

#undef gtk_tree_view_get_hover_expand
extern __typeof (gtk_tree_view_get_hover_expand) gtk_tree_view_get_hover_expand __attribute((alias("IA__gtk_tree_view_get_hover_expand"), visibility("default")));

#undef gtk_tree_view_get_hover_selection
extern __typeof (gtk_tree_view_get_hover_selection) gtk_tree_view_get_hover_selection __attribute((alias("IA__gtk_tree_view_get_hover_selection"), visibility("default")));

#undef gtk_tree_view_get_level_indentation
extern __typeof (gtk_tree_view_get_level_indentation) gtk_tree_view_get_level_indentation __attribute((alias("IA__gtk_tree_view_get_level_indentation"), visibility("default")));

#undef gtk_tree_view_get_model
extern __typeof (gtk_tree_view_get_model) gtk_tree_view_get_model __attribute((alias("IA__gtk_tree_view_get_model"), visibility("default")));

#undef gtk_tree_view_get_path_at_pos
extern __typeof (gtk_tree_view_get_path_at_pos) gtk_tree_view_get_path_at_pos __attribute((alias("IA__gtk_tree_view_get_path_at_pos"), visibility("default")));

#undef gtk_tree_view_get_reorderable
extern __typeof (gtk_tree_view_get_reorderable) gtk_tree_view_get_reorderable __attribute((alias("IA__gtk_tree_view_get_reorderable"), visibility("default")));

#undef gtk_tree_view_get_row_separator_func
extern __typeof (gtk_tree_view_get_row_separator_func) gtk_tree_view_get_row_separator_func __attribute((alias("IA__gtk_tree_view_get_row_separator_func"), visibility("default")));

#undef gtk_tree_view_get_rubber_banding
extern __typeof (gtk_tree_view_get_rubber_banding) gtk_tree_view_get_rubber_banding __attribute((alias("IA__gtk_tree_view_get_rubber_banding"), visibility("default")));

#undef gtk_tree_view_get_rules_hint
extern __typeof (gtk_tree_view_get_rules_hint) gtk_tree_view_get_rules_hint __attribute((alias("IA__gtk_tree_view_get_rules_hint"), visibility("default")));

#undef gtk_tree_view_get_search_column
extern __typeof (gtk_tree_view_get_search_column) gtk_tree_view_get_search_column __attribute((alias("IA__gtk_tree_view_get_search_column"), visibility("default")));

#undef gtk_tree_view_get_search_entry
extern __typeof (gtk_tree_view_get_search_entry) gtk_tree_view_get_search_entry __attribute((alias("IA__gtk_tree_view_get_search_entry"), visibility("default")));

#undef gtk_tree_view_get_search_equal_func
extern __typeof (gtk_tree_view_get_search_equal_func) gtk_tree_view_get_search_equal_func __attribute((alias("IA__gtk_tree_view_get_search_equal_func"), visibility("default")));

#undef gtk_tree_view_get_search_position_func
extern __typeof (gtk_tree_view_get_search_position_func) gtk_tree_view_get_search_position_func __attribute((alias("IA__gtk_tree_view_get_search_position_func"), visibility("default")));

#undef gtk_tree_view_get_selection
extern __typeof (gtk_tree_view_get_selection) gtk_tree_view_get_selection __attribute((alias("IA__gtk_tree_view_get_selection"), visibility("default")));

#undef gtk_tree_view_get_show_expanders
extern __typeof (gtk_tree_view_get_show_expanders) gtk_tree_view_get_show_expanders __attribute((alias("IA__gtk_tree_view_get_show_expanders"), visibility("default")));

#undef gtk_tree_view_get_tooltip_column
extern __typeof (gtk_tree_view_get_tooltip_column) gtk_tree_view_get_tooltip_column __attribute((alias("IA__gtk_tree_view_get_tooltip_column"), visibility("default")));

#undef gtk_tree_view_get_tooltip_context
extern __typeof (gtk_tree_view_get_tooltip_context) gtk_tree_view_get_tooltip_context __attribute((alias("IA__gtk_tree_view_get_tooltip_context"), visibility("default")));

#undef gtk_tree_view_get_type
extern __typeof (gtk_tree_view_get_type) gtk_tree_view_get_type __attribute((alias("IA__gtk_tree_view_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_tree_view_get_vadjustment
extern __typeof (gtk_tree_view_get_vadjustment) gtk_tree_view_get_vadjustment __attribute((alias("IA__gtk_tree_view_get_vadjustment"), visibility("default")));

#undef gtk_tree_view_get_visible_range
extern __typeof (gtk_tree_view_get_visible_range) gtk_tree_view_get_visible_range __attribute((alias("IA__gtk_tree_view_get_visible_range"), visibility("default")));

#undef gtk_tree_view_get_visible_rect
extern __typeof (gtk_tree_view_get_visible_rect) gtk_tree_view_get_visible_rect __attribute((alias("IA__gtk_tree_view_get_visible_rect"), visibility("default")));

#undef gtk_tree_view_insert_column
extern __typeof (gtk_tree_view_insert_column) gtk_tree_view_insert_column __attribute((alias("IA__gtk_tree_view_insert_column"), visibility("default")));

#undef gtk_tree_view_insert_column_with_attributes
extern __typeof (gtk_tree_view_insert_column_with_attributes) gtk_tree_view_insert_column_with_attributes __attribute((alias("IA__gtk_tree_view_insert_column_with_attributes"), visibility("default"))) G_GNUC_NULL_TERMINATED;

#undef gtk_tree_view_insert_column_with_data_func
extern __typeof (gtk_tree_view_insert_column_with_data_func) gtk_tree_view_insert_column_with_data_func __attribute((alias("IA__gtk_tree_view_insert_column_with_data_func"), visibility("default")));

#undef gtk_tree_view_is_rubber_banding_active
extern __typeof (gtk_tree_view_is_rubber_banding_active) gtk_tree_view_is_rubber_banding_active __attribute((alias("IA__gtk_tree_view_is_rubber_banding_active"), visibility("default")));

#undef gtk_tree_view_map_expanded_rows
extern __typeof (gtk_tree_view_map_expanded_rows) gtk_tree_view_map_expanded_rows __attribute((alias("IA__gtk_tree_view_map_expanded_rows"), visibility("default")));

#undef gtk_tree_view_move_column_after
extern __typeof (gtk_tree_view_move_column_after) gtk_tree_view_move_column_after __attribute((alias("IA__gtk_tree_view_move_column_after"), visibility("default")));

#undef gtk_tree_view_new
extern __typeof (gtk_tree_view_new) gtk_tree_view_new __attribute((alias("IA__gtk_tree_view_new"), visibility("default")));

#undef gtk_tree_view_new_with_model
extern __typeof (gtk_tree_view_new_with_model) gtk_tree_view_new_with_model __attribute((alias("IA__gtk_tree_view_new_with_model"), visibility("default")));

#undef gtk_tree_view_remove_column
extern __typeof (gtk_tree_view_remove_column) gtk_tree_view_remove_column __attribute((alias("IA__gtk_tree_view_remove_column"), visibility("default")));

#undef gtk_tree_view_row_activated
extern __typeof (gtk_tree_view_row_activated) gtk_tree_view_row_activated __attribute((alias("IA__gtk_tree_view_row_activated"), visibility("default")));

#undef gtk_tree_view_row_expanded
extern __typeof (gtk_tree_view_row_expanded) gtk_tree_view_row_expanded __attribute((alias("IA__gtk_tree_view_row_expanded"), visibility("default")));

#undef gtk_tree_view_scroll_to_cell
extern __typeof (gtk_tree_view_scroll_to_cell) gtk_tree_view_scroll_to_cell __attribute((alias("IA__gtk_tree_view_scroll_to_cell"), visibility("default")));

#undef gtk_tree_view_scroll_to_point
extern __typeof (gtk_tree_view_scroll_to_point) gtk_tree_view_scroll_to_point __attribute((alias("IA__gtk_tree_view_scroll_to_point"), visibility("default")));

#undef gtk_tree_view_set_column_drag_function
extern __typeof (gtk_tree_view_set_column_drag_function) gtk_tree_view_set_column_drag_function __attribute((alias("IA__gtk_tree_view_set_column_drag_function"), visibility("default")));

#undef gtk_tree_view_set_cursor
extern __typeof (gtk_tree_view_set_cursor) gtk_tree_view_set_cursor __attribute((alias("IA__gtk_tree_view_set_cursor"), visibility("default")));

#undef gtk_tree_view_set_cursor_on_cell
extern __typeof (gtk_tree_view_set_cursor_on_cell) gtk_tree_view_set_cursor_on_cell __attribute((alias("IA__gtk_tree_view_set_cursor_on_cell"), visibility("default")));

#undef gtk_tree_view_set_destroy_count_func
extern __typeof (gtk_tree_view_set_destroy_count_func) gtk_tree_view_set_destroy_count_func __attribute((alias("IA__gtk_tree_view_set_destroy_count_func"), visibility("default")));

#undef gtk_tree_view_set_drag_dest_row
extern __typeof (gtk_tree_view_set_drag_dest_row) gtk_tree_view_set_drag_dest_row __attribute((alias("IA__gtk_tree_view_set_drag_dest_row"), visibility("default")));

#undef gtk_tree_view_set_enable_search
extern __typeof (gtk_tree_view_set_enable_search) gtk_tree_view_set_enable_search __attribute((alias("IA__gtk_tree_view_set_enable_search"), visibility("default")));

#undef gtk_tree_view_set_enable_tree_lines
extern __typeof (gtk_tree_view_set_enable_tree_lines) gtk_tree_view_set_enable_tree_lines __attribute((alias("IA__gtk_tree_view_set_enable_tree_lines"), visibility("default")));

#undef gtk_tree_view_set_expander_column
extern __typeof (gtk_tree_view_set_expander_column) gtk_tree_view_set_expander_column __attribute((alias("IA__gtk_tree_view_set_expander_column"), visibility("default")));

#undef gtk_tree_view_set_fixed_height_mode
extern __typeof (gtk_tree_view_set_fixed_height_mode) gtk_tree_view_set_fixed_height_mode __attribute((alias("IA__gtk_tree_view_set_fixed_height_mode"), visibility("default")));

#undef gtk_tree_view_set_grid_lines
extern __typeof (gtk_tree_view_set_grid_lines) gtk_tree_view_set_grid_lines __attribute((alias("IA__gtk_tree_view_set_grid_lines"), visibility("default")));

#undef gtk_tree_view_set_hadjustment
extern __typeof (gtk_tree_view_set_hadjustment) gtk_tree_view_set_hadjustment __attribute((alias("IA__gtk_tree_view_set_hadjustment"), visibility("default")));

#undef gtk_tree_view_set_headers_clickable
extern __typeof (gtk_tree_view_set_headers_clickable) gtk_tree_view_set_headers_clickable __attribute((alias("IA__gtk_tree_view_set_headers_clickable"), visibility("default")));

#undef gtk_tree_view_set_headers_visible
extern __typeof (gtk_tree_view_set_headers_visible) gtk_tree_view_set_headers_visible __attribute((alias("IA__gtk_tree_view_set_headers_visible"), visibility("default")));

#undef gtk_tree_view_set_hover_expand
extern __typeof (gtk_tree_view_set_hover_expand) gtk_tree_view_set_hover_expand __attribute((alias("IA__gtk_tree_view_set_hover_expand"), visibility("default")));

#undef gtk_tree_view_set_hover_selection
extern __typeof (gtk_tree_view_set_hover_selection) gtk_tree_view_set_hover_selection __attribute((alias("IA__gtk_tree_view_set_hover_selection"), visibility("default")));

#undef gtk_tree_view_set_level_indentation
extern __typeof (gtk_tree_view_set_level_indentation) gtk_tree_view_set_level_indentation __attribute((alias("IA__gtk_tree_view_set_level_indentation"), visibility("default")));

#undef gtk_tree_view_set_model
extern __typeof (gtk_tree_view_set_model) gtk_tree_view_set_model __attribute((alias("IA__gtk_tree_view_set_model"), visibility("default")));

#undef gtk_tree_view_set_reorderable
extern __typeof (gtk_tree_view_set_reorderable) gtk_tree_view_set_reorderable __attribute((alias("IA__gtk_tree_view_set_reorderable"), visibility("default")));

#undef gtk_tree_view_set_row_separator_func
extern __typeof (gtk_tree_view_set_row_separator_func) gtk_tree_view_set_row_separator_func __attribute((alias("IA__gtk_tree_view_set_row_separator_func"), visibility("default")));

#undef gtk_tree_view_set_rubber_banding
extern __typeof (gtk_tree_view_set_rubber_banding) gtk_tree_view_set_rubber_banding __attribute((alias("IA__gtk_tree_view_set_rubber_banding"), visibility("default")));

#undef gtk_tree_view_set_rules_hint
extern __typeof (gtk_tree_view_set_rules_hint) gtk_tree_view_set_rules_hint __attribute((alias("IA__gtk_tree_view_set_rules_hint"), visibility("default")));

#undef gtk_tree_view_set_search_column
extern __typeof (gtk_tree_view_set_search_column) gtk_tree_view_set_search_column __attribute((alias("IA__gtk_tree_view_set_search_column"), visibility("default")));

#undef gtk_tree_view_set_search_entry
extern __typeof (gtk_tree_view_set_search_entry) gtk_tree_view_set_search_entry __attribute((alias("IA__gtk_tree_view_set_search_entry"), visibility("default")));

#undef gtk_tree_view_set_search_equal_func
extern __typeof (gtk_tree_view_set_search_equal_func) gtk_tree_view_set_search_equal_func __attribute((alias("IA__gtk_tree_view_set_search_equal_func"), visibility("default")));

#undef gtk_tree_view_set_search_position_func
extern __typeof (gtk_tree_view_set_search_position_func) gtk_tree_view_set_search_position_func __attribute((alias("IA__gtk_tree_view_set_search_position_func"), visibility("default")));

#undef gtk_tree_view_set_show_expanders
extern __typeof (gtk_tree_view_set_show_expanders) gtk_tree_view_set_show_expanders __attribute((alias("IA__gtk_tree_view_set_show_expanders"), visibility("default")));

#undef gtk_tree_view_set_tooltip_row
extern __typeof (gtk_tree_view_set_tooltip_row) gtk_tree_view_set_tooltip_row __attribute((alias("IA__gtk_tree_view_set_tooltip_row"), visibility("default")));

#undef gtk_tree_view_set_tooltip_cell
extern __typeof (gtk_tree_view_set_tooltip_cell) gtk_tree_view_set_tooltip_cell __attribute((alias("IA__gtk_tree_view_set_tooltip_cell"), visibility("default")));

#undef gtk_tree_view_set_tooltip_column
extern __typeof (gtk_tree_view_set_tooltip_column) gtk_tree_view_set_tooltip_column __attribute((alias("IA__gtk_tree_view_set_tooltip_column"), visibility("default")));

#undef gtk_tree_view_set_vadjustment
extern __typeof (gtk_tree_view_set_vadjustment) gtk_tree_view_set_vadjustment __attribute((alias("IA__gtk_tree_view_set_vadjustment"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_tree_view_tree_to_widget_coords
extern __typeof (gtk_tree_view_tree_to_widget_coords) gtk_tree_view_tree_to_widget_coords __attribute((alias("IA__gtk_tree_view_tree_to_widget_coords"), visibility("default")));

#endif
#undef gtk_tree_view_unset_rows_drag_dest
extern __typeof (gtk_tree_view_unset_rows_drag_dest) gtk_tree_view_unset_rows_drag_dest __attribute((alias("IA__gtk_tree_view_unset_rows_drag_dest"), visibility("default")));

#undef gtk_tree_view_unset_rows_drag_source
extern __typeof (gtk_tree_view_unset_rows_drag_source) gtk_tree_view_unset_rows_drag_source __attribute((alias("IA__gtk_tree_view_unset_rows_drag_source"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_tree_view_widget_to_tree_coords
extern __typeof (gtk_tree_view_widget_to_tree_coords) gtk_tree_view_widget_to_tree_coords __attribute((alias("IA__gtk_tree_view_widget_to_tree_coords"), visibility("default")));

#endif
#undef gtk_tree_view_columns_autosize
extern __typeof (gtk_tree_view_columns_autosize) gtk_tree_view_columns_autosize __attribute((alias("IA__gtk_tree_view_columns_autosize"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_TYPE_UTILS_H__)
#if IN_FILE(__GTK_TYPE_UTILS_C__)
#undef gtk_identifier_get_type
extern __typeof (gtk_identifier_get_type) gtk_identifier_get_type __attribute((alias("IA__gtk_identifier_get_type"), visibility("default"))) G_GNUC_CONST;

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_type_enum_find_value
extern __typeof (gtk_type_enum_find_value) gtk_type_enum_find_value __attribute((alias("IA__gtk_type_enum_find_value"), visibility("default")));

#undef gtk_type_enum_get_values
extern __typeof (gtk_type_enum_get_values) gtk_type_enum_get_values __attribute((alias("IA__gtk_type_enum_get_values"), visibility("default")));

#undef gtk_type_flags_find_value
extern __typeof (gtk_type_flags_find_value) gtk_type_flags_find_value __attribute((alias("IA__gtk_type_flags_find_value"), visibility("default")));

#undef gtk_type_flags_get_values
extern __typeof (gtk_type_flags_get_values) gtk_type_flags_get_values __attribute((alias("IA__gtk_type_flags_get_values"), visibility("default")));

#undef gtk_type_init
extern __typeof (gtk_type_init) gtk_type_init __attribute((alias("IA__gtk_type_init"), visibility("default")));

#undef gtk_type_new
extern __typeof (gtk_type_new) gtk_type_new __attribute((alias("IA__gtk_type_new"), visibility("default")));

#undef gtk_type_unique
extern __typeof (gtk_type_unique) gtk_type_unique __attribute((alias("IA__gtk_type_unique"), visibility("default")));

#undef gtk_type_class
extern __typeof (gtk_type_class) gtk_type_class __attribute((alias("IA__gtk_type_class"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_UI_MANAGER_H__)
#if IN_FILE(__GTK_UI_MANAGER_C__)
#undef gtk_ui_manager_add_ui
extern __typeof (gtk_ui_manager_add_ui) gtk_ui_manager_add_ui __attribute((alias("IA__gtk_ui_manager_add_ui"), visibility("default")));

#ifndef _WIN64
#undef gtk_ui_manager_add_ui_from_file
extern __typeof (gtk_ui_manager_add_ui_from_file) gtk_ui_manager_add_ui_from_file __attribute((alias("IA__gtk_ui_manager_add_ui_from_file"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_ui_manager_add_ui_from_file_utf8
extern __typeof (gtk_ui_manager_add_ui_from_file_utf8) gtk_ui_manager_add_ui_from_file_utf8 __attribute((alias("IA__gtk_ui_manager_add_ui_from_file_utf8"), visibility("default")));

#endif
#undef gtk_ui_manager_add_ui_from_string
extern __typeof (gtk_ui_manager_add_ui_from_string) gtk_ui_manager_add_ui_from_string __attribute((alias("IA__gtk_ui_manager_add_ui_from_string"), visibility("default")));

#undef gtk_ui_manager_ensure_update
extern __typeof (gtk_ui_manager_ensure_update) gtk_ui_manager_ensure_update __attribute((alias("IA__gtk_ui_manager_ensure_update"), visibility("default")));

#undef gtk_ui_manager_get_accel_group
extern __typeof (gtk_ui_manager_get_accel_group) gtk_ui_manager_get_accel_group __attribute((alias("IA__gtk_ui_manager_get_accel_group"), visibility("default")));

#undef gtk_ui_manager_get_action
extern __typeof (gtk_ui_manager_get_action) gtk_ui_manager_get_action __attribute((alias("IA__gtk_ui_manager_get_action"), visibility("default")));

#undef gtk_ui_manager_get_action_groups
extern __typeof (gtk_ui_manager_get_action_groups) gtk_ui_manager_get_action_groups __attribute((alias("IA__gtk_ui_manager_get_action_groups"), visibility("default")));

#undef gtk_ui_manager_get_add_tearoffs
extern __typeof (gtk_ui_manager_get_add_tearoffs) gtk_ui_manager_get_add_tearoffs __attribute((alias("IA__gtk_ui_manager_get_add_tearoffs"), visibility("default")));

#undef gtk_ui_manager_get_toplevels
extern __typeof (gtk_ui_manager_get_toplevels) gtk_ui_manager_get_toplevels __attribute((alias("IA__gtk_ui_manager_get_toplevels"), visibility("default")));

#undef gtk_ui_manager_get_type
extern __typeof (gtk_ui_manager_get_type) gtk_ui_manager_get_type __attribute((alias("IA__gtk_ui_manager_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_ui_manager_get_ui
extern __typeof (gtk_ui_manager_get_ui) gtk_ui_manager_get_ui __attribute((alias("IA__gtk_ui_manager_get_ui"), visibility("default")));

#undef gtk_ui_manager_get_widget
extern __typeof (gtk_ui_manager_get_widget) gtk_ui_manager_get_widget __attribute((alias("IA__gtk_ui_manager_get_widget"), visibility("default")));

#undef gtk_ui_manager_insert_action_group
extern __typeof (gtk_ui_manager_insert_action_group) gtk_ui_manager_insert_action_group __attribute((alias("IA__gtk_ui_manager_insert_action_group"), visibility("default")));

#undef gtk_ui_manager_new
extern __typeof (gtk_ui_manager_new) gtk_ui_manager_new __attribute((alias("IA__gtk_ui_manager_new"), visibility("default")));

#undef gtk_ui_manager_new_merge_id
extern __typeof (gtk_ui_manager_new_merge_id) gtk_ui_manager_new_merge_id __attribute((alias("IA__gtk_ui_manager_new_merge_id"), visibility("default")));

#undef gtk_ui_manager_remove_action_group
extern __typeof (gtk_ui_manager_remove_action_group) gtk_ui_manager_remove_action_group __attribute((alias("IA__gtk_ui_manager_remove_action_group"), visibility("default")));

#undef gtk_ui_manager_remove_ui
extern __typeof (gtk_ui_manager_remove_ui) gtk_ui_manager_remove_ui __attribute((alias("IA__gtk_ui_manager_remove_ui"), visibility("default")));

#undef gtk_ui_manager_set_add_tearoffs
extern __typeof (gtk_ui_manager_set_add_tearoffs) gtk_ui_manager_set_add_tearoffs __attribute((alias("IA__gtk_ui_manager_set_add_tearoffs"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_VBOX_H__)
#if IN_FILE(__GTK_VBOX_C__)
#undef gtk_vbox_get_type
extern __typeof (gtk_vbox_get_type) gtk_vbox_get_type __attribute((alias("IA__gtk_vbox_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_vbox_new
extern __typeof (gtk_vbox_new) gtk_vbox_new __attribute((alias("IA__gtk_vbox_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_VBBOX_H__)
#if IN_FILE(__GTK_VBBOX_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_vbutton_box_get_layout_default
extern __typeof (gtk_vbutton_box_get_layout_default) gtk_vbutton_box_get_layout_default __attribute((alias("IA__gtk_vbutton_box_get_layout_default"), visibility("default")));

#undef gtk_vbutton_box_get_spacing_default
extern __typeof (gtk_vbutton_box_get_spacing_default) gtk_vbutton_box_get_spacing_default __attribute((alias("IA__gtk_vbutton_box_get_spacing_default"), visibility("default")));

#undef gtk_vbutton_box_set_layout_default
extern __typeof (gtk_vbutton_box_set_layout_default) gtk_vbutton_box_set_layout_default __attribute((alias("IA__gtk_vbutton_box_set_layout_default"), visibility("default")));

#undef gtk_vbutton_box_set_spacing_default
extern __typeof (gtk_vbutton_box_set_spacing_default) gtk_vbutton_box_set_spacing_default __attribute((alias("IA__gtk_vbutton_box_set_spacing_default"), visibility("default")));

#endif
#undef gtk_vbutton_box_get_type
extern __typeof (gtk_vbutton_box_get_type) gtk_vbutton_box_get_type __attribute((alias("IA__gtk_vbutton_box_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_vbutton_box_new
extern __typeof (gtk_vbutton_box_new) gtk_vbutton_box_new __attribute((alias("IA__gtk_vbutton_box_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_VIEWPORT_H__)
#if IN_FILE(__GTK_VIEWPORT_C__)
#undef gtk_viewport_get_hadjustment
extern __typeof (gtk_viewport_get_hadjustment) gtk_viewport_get_hadjustment __attribute((alias("IA__gtk_viewport_get_hadjustment"), visibility("default")));

#undef gtk_viewport_get_shadow_type
extern __typeof (gtk_viewport_get_shadow_type) gtk_viewport_get_shadow_type __attribute((alias("IA__gtk_viewport_get_shadow_type"), visibility("default")));

#undef gtk_viewport_get_type
extern __typeof (gtk_viewport_get_type) gtk_viewport_get_type __attribute((alias("IA__gtk_viewport_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_viewport_get_vadjustment
extern __typeof (gtk_viewport_get_vadjustment) gtk_viewport_get_vadjustment __attribute((alias("IA__gtk_viewport_get_vadjustment"), visibility("default")));

#undef gtk_viewport_get_bin_window
extern __typeof (gtk_viewport_get_bin_window) gtk_viewport_get_bin_window __attribute((alias("IA__gtk_viewport_get_bin_window"), visibility("default")));

#undef gtk_viewport_get_view_window
extern __typeof (gtk_viewport_get_view_window) gtk_viewport_get_view_window __attribute((alias("IA__gtk_viewport_get_view_window"), visibility("default")));

#undef gtk_viewport_new
extern __typeof (gtk_viewport_new) gtk_viewport_new __attribute((alias("IA__gtk_viewport_new"), visibility("default")));

#undef gtk_viewport_set_hadjustment
extern __typeof (gtk_viewport_set_hadjustment) gtk_viewport_set_hadjustment __attribute((alias("IA__gtk_viewport_set_hadjustment"), visibility("default")));

#undef gtk_viewport_set_shadow_type
extern __typeof (gtk_viewport_set_shadow_type) gtk_viewport_set_shadow_type __attribute((alias("IA__gtk_viewport_set_shadow_type"), visibility("default")));

#undef gtk_viewport_set_vadjustment
extern __typeof (gtk_viewport_set_vadjustment) gtk_viewport_set_vadjustment __attribute((alias("IA__gtk_viewport_set_vadjustment"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_VOLUME_BUTTON_H__)
#if IN_FILE(__GTK_VOLUME_BUTTON_C__)
#undef gtk_volume_button_get_type
extern __typeof (gtk_volume_button_get_type) gtk_volume_button_get_type __attribute((alias("IA__gtk_volume_button_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_volume_button_new
extern __typeof (gtk_volume_button_new) gtk_volume_button_new __attribute((alias("IA__gtk_volume_button_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_VPANED_H__)
#if IN_FILE(__GTK_VPANED_C__)
#undef gtk_vpaned_get_type
extern __typeof (gtk_vpaned_get_type) gtk_vpaned_get_type __attribute((alias("IA__gtk_vpaned_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_vpaned_new
extern __typeof (gtk_vpaned_new) gtk_vpaned_new __attribute((alias("IA__gtk_vpaned_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_VRULER_H__)
#if IN_FILE(__GTK_VRULER_C__)
#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_vruler_get_type
extern __typeof (gtk_vruler_get_type) gtk_vruler_get_type __attribute((alias("IA__gtk_vruler_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_vruler_new
extern __typeof (gtk_vruler_new) gtk_vruler_new __attribute((alias("IA__gtk_vruler_new"), visibility("default")));

#endif
#endif
#endif
#if IN_HEADER(__GTK_VSCALE_H__)
#if IN_FILE(__GTK_VSCALE_C__)
#undef gtk_vscale_get_type
extern __typeof (gtk_vscale_get_type) gtk_vscale_get_type __attribute((alias("IA__gtk_vscale_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_vscale_new
extern __typeof (gtk_vscale_new) gtk_vscale_new __attribute((alias("IA__gtk_vscale_new"), visibility("default")));

#undef gtk_vscale_new_with_range
extern __typeof (gtk_vscale_new_with_range) gtk_vscale_new_with_range __attribute((alias("IA__gtk_vscale_new_with_range"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_VSCROLLBAR_H__)
#if IN_FILE(__GTK_VSCROLLBAR_C__)
#undef gtk_vscrollbar_get_type
extern __typeof (gtk_vscrollbar_get_type) gtk_vscrollbar_get_type __attribute((alias("IA__gtk_vscrollbar_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_vscrollbar_new
extern __typeof (gtk_vscrollbar_new) gtk_vscrollbar_new __attribute((alias("IA__gtk_vscrollbar_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_VSEPARATOR_H__)
#if IN_FILE(__GTK_VSEPARATOR_C__)
#undef gtk_vseparator_get_type
extern __typeof (gtk_vseparator_get_type) gtk_vseparator_get_type __attribute((alias("IA__gtk_vseparator_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_vseparator_new
extern __typeof (gtk_vseparator_new) gtk_vseparator_new __attribute((alias("IA__gtk_vseparator_new"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_WIDGET_H__)
#if IN_FILE(__GTK_WIDGET_C__)
#undef gtk_requisition_copy
extern __typeof (gtk_requisition_copy) gtk_requisition_copy __attribute((alias("IA__gtk_requisition_copy"), visibility("default")));

#undef gtk_requisition_free
extern __typeof (gtk_requisition_free) gtk_requisition_free __attribute((alias("IA__gtk_requisition_free"), visibility("default")));

#undef gtk_requisition_get_type
extern __typeof (gtk_requisition_get_type) gtk_requisition_get_type __attribute((alias("IA__gtk_requisition_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_widget_activate
extern __typeof (gtk_widget_activate) gtk_widget_activate __attribute((alias("IA__gtk_widget_activate"), visibility("default")));

#undef gtk_widget_is_composited
extern __typeof (gtk_widget_is_composited) gtk_widget_is_composited __attribute((alias("IA__gtk_widget_is_composited"), visibility("default")));

#undef gtk_widget_add_accelerator
extern __typeof (gtk_widget_add_accelerator) gtk_widget_add_accelerator __attribute((alias("IA__gtk_widget_add_accelerator"), visibility("default")));

#undef gtk_widget_add_events
extern __typeof (gtk_widget_add_events) gtk_widget_add_events __attribute((alias("IA__gtk_widget_add_events"), visibility("default")));

#undef gtk_widget_add_mnemonic_label
extern __typeof (gtk_widget_add_mnemonic_label) gtk_widget_add_mnemonic_label __attribute((alias("IA__gtk_widget_add_mnemonic_label"), visibility("default")));

#undef gtk_widget_can_activate_accel
extern __typeof (gtk_widget_can_activate_accel) gtk_widget_can_activate_accel __attribute((alias("IA__gtk_widget_can_activate_accel"), visibility("default")));

#undef gtk_widget_child_focus
extern __typeof (gtk_widget_child_focus) gtk_widget_child_focus __attribute((alias("IA__gtk_widget_child_focus"), visibility("default")));

#undef gtk_widget_child_notify
extern __typeof (gtk_widget_child_notify) gtk_widget_child_notify __attribute((alias("IA__gtk_widget_child_notify"), visibility("default")));

#undef gtk_widget_class_find_style_property
extern __typeof (gtk_widget_class_find_style_property) gtk_widget_class_find_style_property __attribute((alias("IA__gtk_widget_class_find_style_property"), visibility("default")));

#undef gtk_widget_class_install_style_property
extern __typeof (gtk_widget_class_install_style_property) gtk_widget_class_install_style_property __attribute((alias("IA__gtk_widget_class_install_style_property"), visibility("default")));

#undef gtk_widget_class_install_style_property_parser
extern __typeof (gtk_widget_class_install_style_property_parser) gtk_widget_class_install_style_property_parser __attribute((alias("IA__gtk_widget_class_install_style_property_parser"), visibility("default")));

#undef gtk_widget_class_list_style_properties
extern __typeof (gtk_widget_class_list_style_properties) gtk_widget_class_list_style_properties __attribute((alias("IA__gtk_widget_class_list_style_properties"), visibility("default")));

#undef gtk_widget_class_path
extern __typeof (gtk_widget_class_path) gtk_widget_class_path __attribute((alias("IA__gtk_widget_class_path"), visibility("default")));

#undef gtk_widget_create_pango_context
extern __typeof (gtk_widget_create_pango_context) gtk_widget_create_pango_context __attribute((alias("IA__gtk_widget_create_pango_context"), visibility("default")));

#undef gtk_widget_create_pango_layout
extern __typeof (gtk_widget_create_pango_layout) gtk_widget_create_pango_layout __attribute((alias("IA__gtk_widget_create_pango_layout"), visibility("default")));

#undef gtk_widget_destroy
extern __typeof (gtk_widget_destroy) gtk_widget_destroy __attribute((alias("IA__gtk_widget_destroy"), visibility("default")));

#undef gtk_widget_destroyed
extern __typeof (gtk_widget_destroyed) gtk_widget_destroyed __attribute((alias("IA__gtk_widget_destroyed"), visibility("default")));

#undef gtk_widget_ensure_style
extern __typeof (gtk_widget_ensure_style) gtk_widget_ensure_style __attribute((alias("IA__gtk_widget_ensure_style"), visibility("default")));

#undef gtk_widget_error_bell
extern __typeof (gtk_widget_error_bell) gtk_widget_error_bell __attribute((alias("IA__gtk_widget_error_bell"), visibility("default")));

#undef gtk_widget_event
extern __typeof (gtk_widget_event) gtk_widget_event __attribute((alias("IA__gtk_widget_event"), visibility("default")));

#undef gtk_widget_freeze_child_notify
extern __typeof (gtk_widget_freeze_child_notify) gtk_widget_freeze_child_notify __attribute((alias("IA__gtk_widget_freeze_child_notify"), visibility("default")));

#undef gtk_widget_get_accessible
extern __typeof (gtk_widget_get_accessible) gtk_widget_get_accessible __attribute((alias("IA__gtk_widget_get_accessible"), visibility("default")));

#undef gtk_widget_get_allocation
extern __typeof (gtk_widget_get_allocation) gtk_widget_get_allocation __attribute((alias("IA__gtk_widget_get_allocation"), visibility("default")));

#undef gtk_widget_get_ancestor
extern __typeof (gtk_widget_get_ancestor) gtk_widget_get_ancestor __attribute((alias("IA__gtk_widget_get_ancestor"), visibility("default")));

#undef gtk_widget_get_app_paintable
extern __typeof (gtk_widget_get_app_paintable) gtk_widget_get_app_paintable __attribute((alias("IA__gtk_widget_get_app_paintable"), visibility("default")));

#undef gtk_widget_get_child_requisition
extern __typeof (gtk_widget_get_child_requisition) gtk_widget_get_child_requisition __attribute((alias("IA__gtk_widget_get_child_requisition"), visibility("default")));

#undef gtk_widget_get_child_visible
extern __typeof (gtk_widget_get_child_visible) gtk_widget_get_child_visible __attribute((alias("IA__gtk_widget_get_child_visible"), visibility("default")));

#undef gtk_widget_get_clipboard
extern __typeof (gtk_widget_get_clipboard) gtk_widget_get_clipboard __attribute((alias("IA__gtk_widget_get_clipboard"), visibility("default")));

#undef gtk_widget_get_colormap
extern __typeof (gtk_widget_get_colormap) gtk_widget_get_colormap __attribute((alias("IA__gtk_widget_get_colormap"), visibility("default")));

#undef gtk_widget_get_composite_name
extern __typeof (gtk_widget_get_composite_name) gtk_widget_get_composite_name __attribute((alias("IA__gtk_widget_get_composite_name"), visibility("default")));

#undef gtk_widget_get_default_colormap
extern __typeof (gtk_widget_get_default_colormap) gtk_widget_get_default_colormap __attribute((alias("IA__gtk_widget_get_default_colormap"), visibility("default")));

#undef gtk_widget_get_default_direction
extern __typeof (gtk_widget_get_default_direction) gtk_widget_get_default_direction __attribute((alias("IA__gtk_widget_get_default_direction"), visibility("default")));

#undef gtk_widget_get_default_style
extern __typeof (gtk_widget_get_default_style) gtk_widget_get_default_style __attribute((alias("IA__gtk_widget_get_default_style"), visibility("default")));

#undef gtk_widget_get_default_visual
extern __typeof (gtk_widget_get_default_visual) gtk_widget_get_default_visual __attribute((alias("IA__gtk_widget_get_default_visual"), visibility("default")));

#undef gtk_widget_get_direction
extern __typeof (gtk_widget_get_direction) gtk_widget_get_direction __attribute((alias("IA__gtk_widget_get_direction"), visibility("default")));

#undef gtk_widget_get_display
extern __typeof (gtk_widget_get_display) gtk_widget_get_display __attribute((alias("IA__gtk_widget_get_display"), visibility("default")));

#undef gtk_widget_get_double_buffered
extern __typeof (gtk_widget_get_double_buffered) gtk_widget_get_double_buffered __attribute((alias("IA__gtk_widget_get_double_buffered"), visibility("default")));

#undef gtk_widget_get_events
extern __typeof (gtk_widget_get_events) gtk_widget_get_events __attribute((alias("IA__gtk_widget_get_events"), visibility("default")));

#undef gtk_widget_get_extension_events
extern __typeof (gtk_widget_get_extension_events) gtk_widget_get_extension_events __attribute((alias("IA__gtk_widget_get_extension_events"), visibility("default")));

#undef gtk_widget_get_has_tooltip
extern __typeof (gtk_widget_get_has_tooltip) gtk_widget_get_has_tooltip __attribute((alias("IA__gtk_widget_get_has_tooltip"), visibility("default")));

#undef gtk_widget_get_modifier_style
extern __typeof (gtk_widget_get_modifier_style) gtk_widget_get_modifier_style __attribute((alias("IA__gtk_widget_get_modifier_style"), visibility("default")));

#undef gtk_widget_get_name
extern __typeof (gtk_widget_get_name) gtk_widget_get_name __attribute((alias("IA__gtk_widget_get_name"), visibility("default")));

#undef gtk_widget_get_no_show_all
extern __typeof (gtk_widget_get_no_show_all) gtk_widget_get_no_show_all __attribute((alias("IA__gtk_widget_get_no_show_all"), visibility("default")));

#undef gtk_widget_get_pango_context
extern __typeof (gtk_widget_get_pango_context) gtk_widget_get_pango_context __attribute((alias("IA__gtk_widget_get_pango_context"), visibility("default")));

#undef gtk_widget_get_parent
extern __typeof (gtk_widget_get_parent) gtk_widget_get_parent __attribute((alias("IA__gtk_widget_get_parent"), visibility("default")));

#undef gtk_widget_get_parent_window
extern __typeof (gtk_widget_get_parent_window) gtk_widget_get_parent_window __attribute((alias("IA__gtk_widget_get_parent_window"), visibility("default")));

#undef gtk_widget_get_pointer
extern __typeof (gtk_widget_get_pointer) gtk_widget_get_pointer __attribute((alias("IA__gtk_widget_get_pointer"), visibility("default")));

#undef gtk_widget_get_receives_default
extern __typeof (gtk_widget_get_receives_default) gtk_widget_get_receives_default __attribute((alias("IA__gtk_widget_get_receives_default"), visibility("default")));

#undef gtk_widget_get_root_window
extern __typeof (gtk_widget_get_root_window) gtk_widget_get_root_window __attribute((alias("IA__gtk_widget_get_root_window"), visibility("default")));

#undef gtk_widget_get_requisition
extern __typeof (gtk_widget_get_requisition) gtk_widget_get_requisition __attribute((alias("IA__gtk_widget_get_requisition"), visibility("default")));

#undef gtk_widget_get_screen
extern __typeof (gtk_widget_get_screen) gtk_widget_get_screen __attribute((alias("IA__gtk_widget_get_screen"), visibility("default")));

#undef gtk_widget_get_sensitive
extern __typeof (gtk_widget_get_sensitive) gtk_widget_get_sensitive __attribute((alias("IA__gtk_widget_get_sensitive"), visibility("default")));

#undef gtk_widget_get_settings
extern __typeof (gtk_widget_get_settings) gtk_widget_get_settings __attribute((alias("IA__gtk_widget_get_settings"), visibility("default")));

#undef gtk_widget_get_size_request
extern __typeof (gtk_widget_get_size_request) gtk_widget_get_size_request __attribute((alias("IA__gtk_widget_get_size_request"), visibility("default")));

#undef gtk_widget_get_snapshot
extern __typeof (gtk_widget_get_snapshot) gtk_widget_get_snapshot __attribute((alias("IA__gtk_widget_get_snapshot"), visibility("default")));

#undef gtk_widget_get_state
extern __typeof (gtk_widget_get_state) gtk_widget_get_state __attribute((alias("IA__gtk_widget_get_state"), visibility("default")));

#undef gtk_widget_get_style
extern __typeof (gtk_widget_get_style) gtk_widget_get_style __attribute((alias("IA__gtk_widget_get_style"), visibility("default")));

#undef gtk_widget_get_tooltip_markup
extern __typeof (gtk_widget_get_tooltip_markup) gtk_widget_get_tooltip_markup __attribute((alias("IA__gtk_widget_get_tooltip_markup"), visibility("default")));

#undef gtk_widget_get_tooltip_text
extern __typeof (gtk_widget_get_tooltip_text) gtk_widget_get_tooltip_text __attribute((alias("IA__gtk_widget_get_tooltip_text"), visibility("default")));

#undef gtk_widget_get_tooltip_window
extern __typeof (gtk_widget_get_tooltip_window) gtk_widget_get_tooltip_window __attribute((alias("IA__gtk_widget_get_tooltip_window"), visibility("default")));

#undef gtk_widget_get_toplevel
extern __typeof (gtk_widget_get_toplevel) gtk_widget_get_toplevel __attribute((alias("IA__gtk_widget_get_toplevel"), visibility("default")));

#undef gtk_widget_get_type
extern __typeof (gtk_widget_get_type) gtk_widget_get_type __attribute((alias("IA__gtk_widget_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_widget_get_visible
extern __typeof (gtk_widget_get_visible) gtk_widget_get_visible __attribute((alias("IA__gtk_widget_get_visible"), visibility("default")));

#undef gtk_widget_get_visual
extern __typeof (gtk_widget_get_visual) gtk_widget_get_visual __attribute((alias("IA__gtk_widget_get_visual"), visibility("default")));

#undef gtk_widget_grab_default
extern __typeof (gtk_widget_grab_default) gtk_widget_grab_default __attribute((alias("IA__gtk_widget_grab_default"), visibility("default")));

#undef gtk_widget_grab_focus
extern __typeof (gtk_widget_grab_focus) gtk_widget_grab_focus __attribute((alias("IA__gtk_widget_grab_focus"), visibility("default")));

#undef gtk_widget_has_default
extern __typeof (gtk_widget_has_default) gtk_widget_has_default __attribute((alias("IA__gtk_widget_has_default"), visibility("default")));

#undef gtk_widget_has_focus
extern __typeof (gtk_widget_has_focus) gtk_widget_has_focus __attribute((alias("IA__gtk_widget_has_focus"), visibility("default")));

#undef gtk_widget_has_grab
extern __typeof (gtk_widget_has_grab) gtk_widget_has_grab __attribute((alias("IA__gtk_widget_has_grab"), visibility("default")));

#undef gtk_widget_has_rc_style
extern __typeof (gtk_widget_has_rc_style) gtk_widget_has_rc_style __attribute((alias("IA__gtk_widget_has_rc_style"), visibility("default")));

#undef gtk_widget_has_screen
extern __typeof (gtk_widget_has_screen) gtk_widget_has_screen __attribute((alias("IA__gtk_widget_has_screen"), visibility("default")));

#undef gtk_widget_hide
extern __typeof (gtk_widget_hide) gtk_widget_hide __attribute((alias("IA__gtk_widget_hide"), visibility("default")));

#undef gtk_widget_hide_all
extern __typeof (gtk_widget_hide_all) gtk_widget_hide_all __attribute((alias("IA__gtk_widget_hide_all"), visibility("default")));

#undef gtk_widget_hide_on_delete
extern __typeof (gtk_widget_hide_on_delete) gtk_widget_hide_on_delete __attribute((alias("IA__gtk_widget_hide_on_delete"), visibility("default")));

#undef gtk_widget_intersect
extern __typeof (gtk_widget_intersect) gtk_widget_intersect __attribute((alias("IA__gtk_widget_intersect"), visibility("default")));

#undef gtk_widget_is_ancestor
extern __typeof (gtk_widget_is_ancestor) gtk_widget_is_ancestor __attribute((alias("IA__gtk_widget_is_ancestor"), visibility("default")));

#undef gtk_widget_is_focus
extern __typeof (gtk_widget_is_focus) gtk_widget_is_focus __attribute((alias("IA__gtk_widget_is_focus"), visibility("default")));

#undef gtk_widget_is_sensitive
extern __typeof (gtk_widget_is_sensitive) gtk_widget_is_sensitive __attribute((alias("IA__gtk_widget_is_sensitive"), visibility("default")));

#undef gtk_widget_is_toplevel
extern __typeof (gtk_widget_is_toplevel) gtk_widget_is_toplevel __attribute((alias("IA__gtk_widget_is_toplevel"), visibility("default")));

#undef gtk_widget_is_drawable
extern __typeof (gtk_widget_is_drawable) gtk_widget_is_drawable __attribute((alias("IA__gtk_widget_is_drawable"), visibility("default")));

#undef gtk_widget_keynav_failed
extern __typeof (gtk_widget_keynav_failed) gtk_widget_keynav_failed __attribute((alias("IA__gtk_widget_keynav_failed"), visibility("default")));

#undef gtk_widget_list_accel_closures
extern __typeof (gtk_widget_list_accel_closures) gtk_widget_list_accel_closures __attribute((alias("IA__gtk_widget_list_accel_closures"), visibility("default")));

#undef gtk_widget_list_mnemonic_labels
extern __typeof (gtk_widget_list_mnemonic_labels) gtk_widget_list_mnemonic_labels __attribute((alias("IA__gtk_widget_list_mnemonic_labels"), visibility("default")));

#undef gtk_widget_map
extern __typeof (gtk_widget_map) gtk_widget_map __attribute((alias("IA__gtk_widget_map"), visibility("default")));

#undef gtk_widget_mnemonic_activate
extern __typeof (gtk_widget_mnemonic_activate) gtk_widget_mnemonic_activate __attribute((alias("IA__gtk_widget_mnemonic_activate"), visibility("default")));

#undef gtk_widget_modify_base
extern __typeof (gtk_widget_modify_base) gtk_widget_modify_base __attribute((alias("IA__gtk_widget_modify_base"), visibility("default")));

#undef gtk_widget_modify_bg
extern __typeof (gtk_widget_modify_bg) gtk_widget_modify_bg __attribute((alias("IA__gtk_widget_modify_bg"), visibility("default")));

#undef gtk_widget_modify_fg
extern __typeof (gtk_widget_modify_fg) gtk_widget_modify_fg __attribute((alias("IA__gtk_widget_modify_fg"), visibility("default")));

#undef gtk_widget_modify_cursor
extern __typeof (gtk_widget_modify_cursor) gtk_widget_modify_cursor __attribute((alias("IA__gtk_widget_modify_cursor"), visibility("default")));

#undef gtk_widget_modify_font
extern __typeof (gtk_widget_modify_font) gtk_widget_modify_font __attribute((alias("IA__gtk_widget_modify_font"), visibility("default")));

#undef gtk_widget_modify_style
extern __typeof (gtk_widget_modify_style) gtk_widget_modify_style __attribute((alias("IA__gtk_widget_modify_style"), visibility("default")));

#undef gtk_widget_modify_text
extern __typeof (gtk_widget_modify_text) gtk_widget_modify_text __attribute((alias("IA__gtk_widget_modify_text"), visibility("default")));

#undef gtk_widget_new
extern __typeof (gtk_widget_new) gtk_widget_new __attribute((alias("IA__gtk_widget_new"), visibility("default")));

#undef gtk_widget_path
extern __typeof (gtk_widget_path) gtk_widget_path __attribute((alias("IA__gtk_widget_path"), visibility("default")));

#undef gtk_widget_pop_colormap
extern __typeof (gtk_widget_pop_colormap) gtk_widget_pop_colormap __attribute((alias("IA__gtk_widget_pop_colormap"), visibility("default")));

#undef gtk_widget_pop_composite_child
extern __typeof (gtk_widget_pop_composite_child) gtk_widget_pop_composite_child __attribute((alias("IA__gtk_widget_pop_composite_child"), visibility("default")));

#undef gtk_widget_push_colormap
extern __typeof (gtk_widget_push_colormap) gtk_widget_push_colormap __attribute((alias("IA__gtk_widget_push_colormap"), visibility("default")));

#undef gtk_widget_push_composite_child
extern __typeof (gtk_widget_push_composite_child) gtk_widget_push_composite_child __attribute((alias("IA__gtk_widget_push_composite_child"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_widget_draw
extern __typeof (gtk_widget_draw) gtk_widget_draw __attribute((alias("IA__gtk_widget_draw"), visibility("default")));

#undef gtk_widget_queue_clear
extern __typeof (gtk_widget_queue_clear) gtk_widget_queue_clear __attribute((alias("IA__gtk_widget_queue_clear"), visibility("default")));

#undef gtk_widget_queue_clear_area
extern __typeof (gtk_widget_queue_clear_area) gtk_widget_queue_clear_area __attribute((alias("IA__gtk_widget_queue_clear_area"), visibility("default")));

#undef gtk_widget_set
extern __typeof (gtk_widget_set) gtk_widget_set __attribute((alias("IA__gtk_widget_set"), visibility("default")));

#undef gtk_widget_set_uposition
extern __typeof (gtk_widget_set_uposition) gtk_widget_set_uposition __attribute((alias("IA__gtk_widget_set_uposition"), visibility("default")));

#undef gtk_widget_set_usize
extern __typeof (gtk_widget_set_usize) gtk_widget_set_usize __attribute((alias("IA__gtk_widget_set_usize"), visibility("default")));

#undef gtk_widget_ref
extern __typeof (gtk_widget_ref) gtk_widget_ref __attribute((alias("IA__gtk_widget_ref"), visibility("default")));

#undef gtk_widget_unref
extern __typeof (gtk_widget_unref) gtk_widget_unref __attribute((alias("IA__gtk_widget_unref"), visibility("default")));

#endif
#undef gtk_widget_queue_draw
extern __typeof (gtk_widget_queue_draw) gtk_widget_queue_draw __attribute((alias("IA__gtk_widget_queue_draw"), visibility("default")));

#undef gtk_widget_queue_draw_area
extern __typeof (gtk_widget_queue_draw_area) gtk_widget_queue_draw_area __attribute((alias("IA__gtk_widget_queue_draw_area"), visibility("default")));

#undef gtk_widget_queue_resize
extern __typeof (gtk_widget_queue_resize) gtk_widget_queue_resize __attribute((alias("IA__gtk_widget_queue_resize"), visibility("default")));

#undef gtk_widget_queue_resize_no_redraw
extern __typeof (gtk_widget_queue_resize_no_redraw) gtk_widget_queue_resize_no_redraw __attribute((alias("IA__gtk_widget_queue_resize_no_redraw"), visibility("default")));

#undef gtk_widget_realize
extern __typeof (gtk_widget_realize) gtk_widget_realize __attribute((alias("IA__gtk_widget_realize"), visibility("default")));

#undef gtk_widget_region_intersect
extern __typeof (gtk_widget_region_intersect) gtk_widget_region_intersect __attribute((alias("IA__gtk_widget_region_intersect"), visibility("default")));

#undef gtk_widget_remove_accelerator
extern __typeof (gtk_widget_remove_accelerator) gtk_widget_remove_accelerator __attribute((alias("IA__gtk_widget_remove_accelerator"), visibility("default")));

#undef gtk_widget_remove_mnemonic_label
extern __typeof (gtk_widget_remove_mnemonic_label) gtk_widget_remove_mnemonic_label __attribute((alias("IA__gtk_widget_remove_mnemonic_label"), visibility("default")));

#undef gtk_widget_render_icon
extern __typeof (gtk_widget_render_icon) gtk_widget_render_icon __attribute((alias("IA__gtk_widget_render_icon"), visibility("default")));

#undef gtk_widget_reparent
extern __typeof (gtk_widget_reparent) gtk_widget_reparent __attribute((alias("IA__gtk_widget_reparent"), visibility("default")));

#undef gtk_widget_reset_rc_styles
extern __typeof (gtk_widget_reset_rc_styles) gtk_widget_reset_rc_styles __attribute((alias("IA__gtk_widget_reset_rc_styles"), visibility("default")));

#undef gtk_widget_reset_shapes
extern __typeof (gtk_widget_reset_shapes) gtk_widget_reset_shapes __attribute((alias("IA__gtk_widget_reset_shapes"), visibility("default")));

#undef gtk_widget_send_expose
extern __typeof (gtk_widget_send_expose) gtk_widget_send_expose __attribute((alias("IA__gtk_widget_send_expose"), visibility("default")));

#undef gtk_widget_send_focus_change
extern __typeof (gtk_widget_send_focus_change) gtk_widget_send_focus_change __attribute((alias("IA__gtk_widget_send_focus_change"), visibility("default")));

#undef gtk_widget_set_accel_path
extern __typeof (gtk_widget_set_accel_path) gtk_widget_set_accel_path __attribute((alias("IA__gtk_widget_set_accel_path"), visibility("default")));

#undef gtk_widget_set_allocation
extern __typeof (gtk_widget_set_allocation) gtk_widget_set_allocation __attribute((alias("IA__gtk_widget_set_allocation"), visibility("default")));

#undef gtk_widget_set_app_paintable
extern __typeof (gtk_widget_set_app_paintable) gtk_widget_set_app_paintable __attribute((alias("IA__gtk_widget_set_app_paintable"), visibility("default")));

#undef gtk_widget_set_child_visible
extern __typeof (gtk_widget_set_child_visible) gtk_widget_set_child_visible __attribute((alias("IA__gtk_widget_set_child_visible"), visibility("default")));

#undef gtk_widget_set_colormap
extern __typeof (gtk_widget_set_colormap) gtk_widget_set_colormap __attribute((alias("IA__gtk_widget_set_colormap"), visibility("default")));

#undef gtk_widget_set_composite_name
extern __typeof (gtk_widget_set_composite_name) gtk_widget_set_composite_name __attribute((alias("IA__gtk_widget_set_composite_name"), visibility("default")));

#undef gtk_widget_set_default_colormap
extern __typeof (gtk_widget_set_default_colormap) gtk_widget_set_default_colormap __attribute((alias("IA__gtk_widget_set_default_colormap"), visibility("default")));

#undef gtk_widget_set_default_direction
extern __typeof (gtk_widget_set_default_direction) gtk_widget_set_default_direction __attribute((alias("IA__gtk_widget_set_default_direction"), visibility("default")));

#undef gtk_widget_set_direction
extern __typeof (gtk_widget_set_direction) gtk_widget_set_direction __attribute((alias("IA__gtk_widget_set_direction"), visibility("default")));

#undef gtk_widget_set_double_buffered
extern __typeof (gtk_widget_set_double_buffered) gtk_widget_set_double_buffered __attribute((alias("IA__gtk_widget_set_double_buffered"), visibility("default")));

#undef gtk_widget_set_events
extern __typeof (gtk_widget_set_events) gtk_widget_set_events __attribute((alias("IA__gtk_widget_set_events"), visibility("default")));

#undef gtk_widget_set_extension_events
extern __typeof (gtk_widget_set_extension_events) gtk_widget_set_extension_events __attribute((alias("IA__gtk_widget_set_extension_events"), visibility("default")));

#undef gtk_widget_set_has_tooltip
extern __typeof (gtk_widget_set_has_tooltip) gtk_widget_set_has_tooltip __attribute((alias("IA__gtk_widget_set_has_tooltip"), visibility("default")));

#undef gtk_widget_set_name
extern __typeof (gtk_widget_set_name) gtk_widget_set_name __attribute((alias("IA__gtk_widget_set_name"), visibility("default")));

#undef gtk_widget_set_no_show_all
extern __typeof (gtk_widget_set_no_show_all) gtk_widget_set_no_show_all __attribute((alias("IA__gtk_widget_set_no_show_all"), visibility("default")));

#undef gtk_widget_set_parent
extern __typeof (gtk_widget_set_parent) gtk_widget_set_parent __attribute((alias("IA__gtk_widget_set_parent"), visibility("default")));

#undef gtk_widget_set_parent_window
extern __typeof (gtk_widget_set_parent_window) gtk_widget_set_parent_window __attribute((alias("IA__gtk_widget_set_parent_window"), visibility("default")));

#undef gtk_widget_set_receives_default
extern __typeof (gtk_widget_set_receives_default) gtk_widget_set_receives_default __attribute((alias("IA__gtk_widget_set_receives_default"), visibility("default")));

#undef gtk_widget_set_redraw_on_allocate
extern __typeof (gtk_widget_set_redraw_on_allocate) gtk_widget_set_redraw_on_allocate __attribute((alias("IA__gtk_widget_set_redraw_on_allocate"), visibility("default")));

#undef gtk_widget_set_scroll_adjustments
extern __typeof (gtk_widget_set_scroll_adjustments) gtk_widget_set_scroll_adjustments __attribute((alias("IA__gtk_widget_set_scroll_adjustments"), visibility("default")));

#undef gtk_widget_set_sensitive
extern __typeof (gtk_widget_set_sensitive) gtk_widget_set_sensitive __attribute((alias("IA__gtk_widget_set_sensitive"), visibility("default")));

#undef gtk_widget_set_size_request
extern __typeof (gtk_widget_set_size_request) gtk_widget_set_size_request __attribute((alias("IA__gtk_widget_set_size_request"), visibility("default")));

#undef gtk_widget_set_state
extern __typeof (gtk_widget_set_state) gtk_widget_set_state __attribute((alias("IA__gtk_widget_set_state"), visibility("default")));

#undef gtk_widget_set_style
extern __typeof (gtk_widget_set_style) gtk_widget_set_style __attribute((alias("IA__gtk_widget_set_style"), visibility("default")));

#undef gtk_widget_set_tooltip_markup
extern __typeof (gtk_widget_set_tooltip_markup) gtk_widget_set_tooltip_markup __attribute((alias("IA__gtk_widget_set_tooltip_markup"), visibility("default")));

#undef gtk_widget_set_tooltip_text
extern __typeof (gtk_widget_set_tooltip_text) gtk_widget_set_tooltip_text __attribute((alias("IA__gtk_widget_set_tooltip_text"), visibility("default")));

#undef gtk_widget_set_tooltip_window
extern __typeof (gtk_widget_set_tooltip_window) gtk_widget_set_tooltip_window __attribute((alias("IA__gtk_widget_set_tooltip_window"), visibility("default")));

#undef gtk_widget_set_visible
extern __typeof (gtk_widget_set_visible) gtk_widget_set_visible __attribute((alias("IA__gtk_widget_set_visible"), visibility("default")));

#undef gtk_widget_set_window
extern __typeof (gtk_widget_set_window) gtk_widget_set_window __attribute((alias("IA__gtk_widget_set_window"), visibility("default")));

#undef gtk_widget_shape_combine_mask
extern __typeof (gtk_widget_shape_combine_mask) gtk_widget_shape_combine_mask __attribute((alias("IA__gtk_widget_shape_combine_mask"), visibility("default")));

#undef gtk_widget_input_shape_combine_mask
extern __typeof (gtk_widget_input_shape_combine_mask) gtk_widget_input_shape_combine_mask __attribute((alias("IA__gtk_widget_input_shape_combine_mask"), visibility("default")));

#undef gtk_widget_show
extern __typeof (gtk_widget_show) gtk_widget_show __attribute((alias("IA__gtk_widget_show"), visibility("default")));

#undef gtk_widget_show_all
extern __typeof (gtk_widget_show_all) gtk_widget_show_all __attribute((alias("IA__gtk_widget_show_all"), visibility("default")));

#undef gtk_widget_show_now
extern __typeof (gtk_widget_show_now) gtk_widget_show_now __attribute((alias("IA__gtk_widget_show_now"), visibility("default")));

#undef gtk_widget_size_allocate
extern __typeof (gtk_widget_size_allocate) gtk_widget_size_allocate __attribute((alias("IA__gtk_widget_size_allocate"), visibility("default")));

#undef gtk_widget_size_request
extern __typeof (gtk_widget_size_request) gtk_widget_size_request __attribute((alias("IA__gtk_widget_size_request"), visibility("default")));

#undef gtk_widget_style_attach
extern __typeof (gtk_widget_style_attach) gtk_widget_style_attach __attribute((alias("IA__gtk_widget_style_attach"), visibility("default")));

#undef gtk_widget_style_get
extern __typeof (gtk_widget_style_get) gtk_widget_style_get __attribute((alias("IA__gtk_widget_style_get"), visibility("default"))) G_GNUC_NULL_TERMINATED;

#undef gtk_widget_style_get_property
extern __typeof (gtk_widget_style_get_property) gtk_widget_style_get_property __attribute((alias("IA__gtk_widget_style_get_property"), visibility("default")));

#undef gtk_widget_style_get_valist
extern __typeof (gtk_widget_style_get_valist) gtk_widget_style_get_valist __attribute((alias("IA__gtk_widget_style_get_valist"), visibility("default")));

#undef gtk_widget_thaw_child_notify
extern __typeof (gtk_widget_thaw_child_notify) gtk_widget_thaw_child_notify __attribute((alias("IA__gtk_widget_thaw_child_notify"), visibility("default")));

#undef gtk_widget_translate_coordinates
extern __typeof (gtk_widget_translate_coordinates) gtk_widget_translate_coordinates __attribute((alias("IA__gtk_widget_translate_coordinates"), visibility("default")));

#undef gtk_widget_trigger_tooltip_query
extern __typeof (gtk_widget_trigger_tooltip_query) gtk_widget_trigger_tooltip_query __attribute((alias("IA__gtk_widget_trigger_tooltip_query"), visibility("default")));

#undef gtk_widget_unmap
extern __typeof (gtk_widget_unmap) gtk_widget_unmap __attribute((alias("IA__gtk_widget_unmap"), visibility("default")));

#undef gtk_widget_unparent
extern __typeof (gtk_widget_unparent) gtk_widget_unparent __attribute((alias("IA__gtk_widget_unparent"), visibility("default")));

#undef gtk_widget_unrealize
extern __typeof (gtk_widget_unrealize) gtk_widget_unrealize __attribute((alias("IA__gtk_widget_unrealize"), visibility("default")));

#undef gtk_widget_get_window
extern __typeof (gtk_widget_get_window) gtk_widget_get_window __attribute((alias("IA__gtk_widget_get_window"), visibility("default")));

#undef gtk_widget_set_has_window
extern __typeof (gtk_widget_set_has_window) gtk_widget_set_has_window __attribute((alias("IA__gtk_widget_set_has_window"), visibility("default")));

#undef gtk_widget_get_has_window
extern __typeof (gtk_widget_get_has_window) gtk_widget_get_has_window __attribute((alias("IA__gtk_widget_get_has_window"), visibility("default")));

#undef gtk_widget_set_can_default
extern __typeof (gtk_widget_set_can_default) gtk_widget_set_can_default __attribute((alias("IA__gtk_widget_set_can_default"), visibility("default")));

#undef gtk_widget_get_can_default
extern __typeof (gtk_widget_get_can_default) gtk_widget_get_can_default __attribute((alias("IA__gtk_widget_get_can_default"), visibility("default")));

#undef gtk_widget_set_can_focus
extern __typeof (gtk_widget_set_can_focus) gtk_widget_set_can_focus __attribute((alias("IA__gtk_widget_set_can_focus"), visibility("default")));

#undef gtk_widget_get_can_focus
extern __typeof (gtk_widget_get_can_focus) gtk_widget_get_can_focus __attribute((alias("IA__gtk_widget_get_can_focus"), visibility("default")));

#undef gtk_widget_set_realized
extern __typeof (gtk_widget_set_realized) gtk_widget_set_realized __attribute((alias("IA__gtk_widget_set_realized"), visibility("default")));

#undef gtk_widget_get_realized
extern __typeof (gtk_widget_get_realized) gtk_widget_get_realized __attribute((alias("IA__gtk_widget_get_realized"), visibility("default")));

#undef gtk_widget_set_mapped
extern __typeof (gtk_widget_set_mapped) gtk_widget_set_mapped __attribute((alias("IA__gtk_widget_set_mapped"), visibility("default")));

#undef gtk_widget_get_mapped
extern __typeof (gtk_widget_get_mapped) gtk_widget_get_mapped __attribute((alias("IA__gtk_widget_get_mapped"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_WINDOW_H__)
#if IN_FILE(__GTK_WINDOW_C__)
#undef gtk_window_activate_default
extern __typeof (gtk_window_activate_default) gtk_window_activate_default __attribute((alias("IA__gtk_window_activate_default"), visibility("default")));

#undef gtk_window_activate_focus
extern __typeof (gtk_window_activate_focus) gtk_window_activate_focus __attribute((alias("IA__gtk_window_activate_focus"), visibility("default")));

#undef gtk_window_activate_key
extern __typeof (gtk_window_activate_key) gtk_window_activate_key __attribute((alias("IA__gtk_window_activate_key"), visibility("default")));

#undef gtk_window_add_accel_group
extern __typeof (gtk_window_add_accel_group) gtk_window_add_accel_group __attribute((alias("IA__gtk_window_add_accel_group"), visibility("default")));

#undef gtk_window_add_embedded_xid
extern __typeof (gtk_window_add_embedded_xid) gtk_window_add_embedded_xid __attribute((alias("IA__gtk_window_add_embedded_xid"), visibility("default")));

#undef gtk_window_add_mnemonic
extern __typeof (gtk_window_add_mnemonic) gtk_window_add_mnemonic __attribute((alias("IA__gtk_window_add_mnemonic"), visibility("default")));

#undef gtk_window_begin_move_drag
extern __typeof (gtk_window_begin_move_drag) gtk_window_begin_move_drag __attribute((alias("IA__gtk_window_begin_move_drag"), visibility("default")));

#undef gtk_window_begin_resize_drag
extern __typeof (gtk_window_begin_resize_drag) gtk_window_begin_resize_drag __attribute((alias("IA__gtk_window_begin_resize_drag"), visibility("default")));

#undef gtk_window_deiconify
extern __typeof (gtk_window_deiconify) gtk_window_deiconify __attribute((alias("IA__gtk_window_deiconify"), visibility("default")));

#undef gtk_window_fullscreen
extern __typeof (gtk_window_fullscreen) gtk_window_fullscreen __attribute((alias("IA__gtk_window_fullscreen"), visibility("default")));

#undef gtk_window_get_accept_focus
extern __typeof (gtk_window_get_accept_focus) gtk_window_get_accept_focus __attribute((alias("IA__gtk_window_get_accept_focus"), visibility("default")));

#undef gtk_window_get_decorated
extern __typeof (gtk_window_get_decorated) gtk_window_get_decorated __attribute((alias("IA__gtk_window_get_decorated"), visibility("default")));

#undef gtk_window_get_deletable
extern __typeof (gtk_window_get_deletable) gtk_window_get_deletable __attribute((alias("IA__gtk_window_get_deletable"), visibility("default")));

#undef gtk_window_get_default_widget
extern __typeof (gtk_window_get_default_widget) gtk_window_get_default_widget __attribute((alias("IA__gtk_window_get_default_widget"), visibility("default")));

#undef gtk_window_get_default_icon_list
extern __typeof (gtk_window_get_default_icon_list) gtk_window_get_default_icon_list __attribute((alias("IA__gtk_window_get_default_icon_list"), visibility("default")));

#undef gtk_window_get_default_icon_name
extern __typeof (gtk_window_get_default_icon_name) gtk_window_get_default_icon_name __attribute((alias("IA__gtk_window_get_default_icon_name"), visibility("default")));

#undef gtk_window_get_default_size
extern __typeof (gtk_window_get_default_size) gtk_window_get_default_size __attribute((alias("IA__gtk_window_get_default_size"), visibility("default")));

#undef gtk_window_get_destroy_with_parent
extern __typeof (gtk_window_get_destroy_with_parent) gtk_window_get_destroy_with_parent __attribute((alias("IA__gtk_window_get_destroy_with_parent"), visibility("default")));

#undef gtk_window_get_focus
extern __typeof (gtk_window_get_focus) gtk_window_get_focus __attribute((alias("IA__gtk_window_get_focus"), visibility("default")));

#undef gtk_window_get_focus_on_map
extern __typeof (gtk_window_get_focus_on_map) gtk_window_get_focus_on_map __attribute((alias("IA__gtk_window_get_focus_on_map"), visibility("default")));

#undef gtk_window_get_gravity
extern __typeof (gtk_window_get_gravity) gtk_window_get_gravity __attribute((alias("IA__gtk_window_get_gravity"), visibility("default")));

#undef gtk_window_get_group
extern __typeof (gtk_window_get_group) gtk_window_get_group __attribute((alias("IA__gtk_window_get_group"), visibility("default")));

#undef gtk_window_get_icon
extern __typeof (gtk_window_get_icon) gtk_window_get_icon __attribute((alias("IA__gtk_window_get_icon"), visibility("default")));

#undef gtk_window_get_icon_list
extern __typeof (gtk_window_get_icon_list) gtk_window_get_icon_list __attribute((alias("IA__gtk_window_get_icon_list"), visibility("default")));

#undef gtk_window_get_icon_name
extern __typeof (gtk_window_get_icon_name) gtk_window_get_icon_name __attribute((alias("IA__gtk_window_get_icon_name"), visibility("default")));

#undef gtk_window_get_mnemonic_modifier
extern __typeof (gtk_window_get_mnemonic_modifier) gtk_window_get_mnemonic_modifier __attribute((alias("IA__gtk_window_get_mnemonic_modifier"), visibility("default")));

#undef gtk_window_get_mnemonics_visible
extern __typeof (gtk_window_get_mnemonics_visible) gtk_window_get_mnemonics_visible __attribute((alias("IA__gtk_window_get_mnemonics_visible"), visibility("default")));

#undef gtk_window_get_modal
extern __typeof (gtk_window_get_modal) gtk_window_get_modal __attribute((alias("IA__gtk_window_get_modal"), visibility("default")));

#undef gtk_window_get_position
extern __typeof (gtk_window_get_position) gtk_window_get_position __attribute((alias("IA__gtk_window_get_position"), visibility("default")));

#undef gtk_window_get_resizable
extern __typeof (gtk_window_get_resizable) gtk_window_get_resizable __attribute((alias("IA__gtk_window_get_resizable"), visibility("default")));

#undef gtk_window_get_role
extern __typeof (gtk_window_get_role) gtk_window_get_role __attribute((alias("IA__gtk_window_get_role"), visibility("default")));

#undef gtk_window_get_screen
extern __typeof (gtk_window_get_screen) gtk_window_get_screen __attribute((alias("IA__gtk_window_get_screen"), visibility("default")));

#undef gtk_window_get_size
extern __typeof (gtk_window_get_size) gtk_window_get_size __attribute((alias("IA__gtk_window_get_size"), visibility("default")));

#undef gtk_window_get_skip_pager_hint
extern __typeof (gtk_window_get_skip_pager_hint) gtk_window_get_skip_pager_hint __attribute((alias("IA__gtk_window_get_skip_pager_hint"), visibility("default")));

#undef gtk_window_get_skip_taskbar_hint
extern __typeof (gtk_window_get_skip_taskbar_hint) gtk_window_get_skip_taskbar_hint __attribute((alias("IA__gtk_window_get_skip_taskbar_hint"), visibility("default")));

#undef gtk_window_get_urgency_hint
extern __typeof (gtk_window_get_urgency_hint) gtk_window_get_urgency_hint __attribute((alias("IA__gtk_window_get_urgency_hint"), visibility("default")));

#undef gtk_window_get_title
extern __typeof (gtk_window_get_title) gtk_window_get_title __attribute((alias("IA__gtk_window_get_title"), visibility("default")));

#undef gtk_window_get_transient_for
extern __typeof (gtk_window_get_transient_for) gtk_window_get_transient_for __attribute((alias("IA__gtk_window_get_transient_for"), visibility("default")));

#undef gtk_window_get_type
extern __typeof (gtk_window_get_type) gtk_window_get_type __attribute((alias("IA__gtk_window_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_window_get_type_hint
extern __typeof (gtk_window_get_type_hint) gtk_window_get_type_hint __attribute((alias("IA__gtk_window_get_type_hint"), visibility("default")));

#undef gtk_window_get_window_type
extern __typeof (gtk_window_get_window_type) gtk_window_get_window_type __attribute((alias("IA__gtk_window_get_window_type"), visibility("default")));

#undef gtk_window_group_add_window
extern __typeof (gtk_window_group_add_window) gtk_window_group_add_window __attribute((alias("IA__gtk_window_group_add_window"), visibility("default")));

#undef gtk_window_group_get_type
extern __typeof (gtk_window_group_get_type) gtk_window_group_get_type __attribute((alias("IA__gtk_window_group_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_window_group_new
extern __typeof (gtk_window_group_new) gtk_window_group_new __attribute((alias("IA__gtk_window_group_new"), visibility("default")));

#undef gtk_window_group_remove_window
extern __typeof (gtk_window_group_remove_window) gtk_window_group_remove_window __attribute((alias("IA__gtk_window_group_remove_window"), visibility("default")));

#undef gtk_window_group_list_windows
extern __typeof (gtk_window_group_list_windows) gtk_window_group_list_windows __attribute((alias("IA__gtk_window_group_list_windows"), visibility("default")));

#undef gtk_window_group_get_current_grab
extern __typeof (gtk_window_group_get_current_grab) gtk_window_group_get_current_grab __attribute((alias("IA__gtk_window_group_get_current_grab"), visibility("default")));

#undef gtk_window_has_group
extern __typeof (gtk_window_has_group) gtk_window_has_group __attribute((alias("IA__gtk_window_has_group"), visibility("default")));

#undef gtk_window_has_toplevel_focus
extern __typeof (gtk_window_has_toplevel_focus) gtk_window_has_toplevel_focus __attribute((alias("IA__gtk_window_has_toplevel_focus"), visibility("default")));

#undef gtk_window_iconify
extern __typeof (gtk_window_iconify) gtk_window_iconify __attribute((alias("IA__gtk_window_iconify"), visibility("default")));

#undef gtk_window_is_active
extern __typeof (gtk_window_is_active) gtk_window_is_active __attribute((alias("IA__gtk_window_is_active"), visibility("default")));

#undef gtk_window_list_toplevels
extern __typeof (gtk_window_list_toplevels) gtk_window_list_toplevels __attribute((alias("IA__gtk_window_list_toplevels"), visibility("default")));

#undef gtk_window_maximize
extern __typeof (gtk_window_maximize) gtk_window_maximize __attribute((alias("IA__gtk_window_maximize"), visibility("default")));

#undef gtk_window_mnemonic_activate
extern __typeof (gtk_window_mnemonic_activate) gtk_window_mnemonic_activate __attribute((alias("IA__gtk_window_mnemonic_activate"), visibility("default")));

#undef gtk_window_move
extern __typeof (gtk_window_move) gtk_window_move __attribute((alias("IA__gtk_window_move"), visibility("default")));

#undef gtk_window_new
extern __typeof (gtk_window_new) gtk_window_new __attribute((alias("IA__gtk_window_new"), visibility("default")));

#undef gtk_window_parse_geometry
extern __typeof (gtk_window_parse_geometry) gtk_window_parse_geometry __attribute((alias("IA__gtk_window_parse_geometry"), visibility("default")));

#undef gtk_window_present
extern __typeof (gtk_window_present) gtk_window_present __attribute((alias("IA__gtk_window_present"), visibility("default")));

#undef gtk_window_present_with_time
extern __typeof (gtk_window_present_with_time) gtk_window_present_with_time __attribute((alias("IA__gtk_window_present_with_time"), visibility("default")));

#undef gtk_window_propagate_key_event
extern __typeof (gtk_window_propagate_key_event) gtk_window_propagate_key_event __attribute((alias("IA__gtk_window_propagate_key_event"), visibility("default")));

#undef gtk_window_remove_accel_group
extern __typeof (gtk_window_remove_accel_group) gtk_window_remove_accel_group __attribute((alias("IA__gtk_window_remove_accel_group"), visibility("default")));

#undef gtk_window_remove_embedded_xid
extern __typeof (gtk_window_remove_embedded_xid) gtk_window_remove_embedded_xid __attribute((alias("IA__gtk_window_remove_embedded_xid"), visibility("default")));

#undef gtk_window_remove_mnemonic
extern __typeof (gtk_window_remove_mnemonic) gtk_window_remove_mnemonic __attribute((alias("IA__gtk_window_remove_mnemonic"), visibility("default")));

#undef gtk_window_reshow_with_initial_size
extern __typeof (gtk_window_reshow_with_initial_size) gtk_window_reshow_with_initial_size __attribute((alias("IA__gtk_window_reshow_with_initial_size"), visibility("default")));

#undef gtk_window_resize
extern __typeof (gtk_window_resize) gtk_window_resize __attribute((alias("IA__gtk_window_resize"), visibility("default")));

#undef gtk_window_set_accept_focus
extern __typeof (gtk_window_set_accept_focus) gtk_window_set_accept_focus __attribute((alias("IA__gtk_window_set_accept_focus"), visibility("default")));

#undef gtk_window_set_auto_startup_notification
extern __typeof (gtk_window_set_auto_startup_notification) gtk_window_set_auto_startup_notification __attribute((alias("IA__gtk_window_set_auto_startup_notification"), visibility("default")));

#undef gtk_window_set_decorated
extern __typeof (gtk_window_set_decorated) gtk_window_set_decorated __attribute((alias("IA__gtk_window_set_decorated"), visibility("default")));

#undef gtk_window_set_deletable
extern __typeof (gtk_window_set_deletable) gtk_window_set_deletable __attribute((alias("IA__gtk_window_set_deletable"), visibility("default")));

#undef gtk_window_set_default
extern __typeof (gtk_window_set_default) gtk_window_set_default __attribute((alias("IA__gtk_window_set_default"), visibility("default")));

#undef gtk_window_set_default_icon
extern __typeof (gtk_window_set_default_icon) gtk_window_set_default_icon __attribute((alias("IA__gtk_window_set_default_icon"), visibility("default")));

#ifndef _WIN64
#undef gtk_window_set_default_icon_from_file
extern __typeof (gtk_window_set_default_icon_from_file) gtk_window_set_default_icon_from_file __attribute((alias("IA__gtk_window_set_default_icon_from_file"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_window_set_default_icon_from_file_utf8
extern __typeof (gtk_window_set_default_icon_from_file_utf8) gtk_window_set_default_icon_from_file_utf8 __attribute((alias("IA__gtk_window_set_default_icon_from_file_utf8"), visibility("default")));

#endif
#undef gtk_window_set_default_icon_list
extern __typeof (gtk_window_set_default_icon_list) gtk_window_set_default_icon_list __attribute((alias("IA__gtk_window_set_default_icon_list"), visibility("default")));

#undef gtk_window_set_default_icon_name
extern __typeof (gtk_window_set_default_icon_name) gtk_window_set_default_icon_name __attribute((alias("IA__gtk_window_set_default_icon_name"), visibility("default")));

#undef gtk_window_set_default_size
extern __typeof (gtk_window_set_default_size) gtk_window_set_default_size __attribute((alias("IA__gtk_window_set_default_size"), visibility("default")));

#undef gtk_window_set_destroy_with_parent
extern __typeof (gtk_window_set_destroy_with_parent) gtk_window_set_destroy_with_parent __attribute((alias("IA__gtk_window_set_destroy_with_parent"), visibility("default")));

#undef gtk_window_set_focus
extern __typeof (gtk_window_set_focus) gtk_window_set_focus __attribute((alias("IA__gtk_window_set_focus"), visibility("default")));

#undef gtk_window_set_focus_on_map
extern __typeof (gtk_window_set_focus_on_map) gtk_window_set_focus_on_map __attribute((alias("IA__gtk_window_set_focus_on_map"), visibility("default")));

#undef gtk_window_set_startup_id
extern __typeof (gtk_window_set_startup_id) gtk_window_set_startup_id __attribute((alias("IA__gtk_window_set_startup_id"), visibility("default")));

#undef gtk_window_set_geometry_hints
extern __typeof (gtk_window_set_geometry_hints) gtk_window_set_geometry_hints __attribute((alias("IA__gtk_window_set_geometry_hints"), visibility("default")));

#undef gtk_window_set_gravity
extern __typeof (gtk_window_set_gravity) gtk_window_set_gravity __attribute((alias("IA__gtk_window_set_gravity"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_window_get_frame_dimensions
extern __typeof (gtk_window_get_frame_dimensions) gtk_window_get_frame_dimensions __attribute((alias("IA__gtk_window_get_frame_dimensions"), visibility("default")));

#undef gtk_window_set_frame_dimensions
extern __typeof (gtk_window_set_frame_dimensions) gtk_window_set_frame_dimensions __attribute((alias("IA__gtk_window_set_frame_dimensions"), visibility("default")));

#undef gtk_window_get_has_frame
extern __typeof (gtk_window_get_has_frame) gtk_window_get_has_frame __attribute((alias("IA__gtk_window_get_has_frame"), visibility("default")));

#undef gtk_window_set_has_frame
extern __typeof (gtk_window_set_has_frame) gtk_window_set_has_frame __attribute((alias("IA__gtk_window_set_has_frame"), visibility("default")));

#endif
#undef gtk_window_set_icon
extern __typeof (gtk_window_set_icon) gtk_window_set_icon __attribute((alias("IA__gtk_window_set_icon"), visibility("default")));

#ifndef _WIN64
#undef gtk_window_set_icon_from_file
extern __typeof (gtk_window_set_icon_from_file) gtk_window_set_icon_from_file __attribute((alias("IA__gtk_window_set_icon_from_file"), visibility("default")));

#endif
#ifdef G_OS_WIN32
#undef gtk_window_set_icon_from_file_utf8
extern __typeof (gtk_window_set_icon_from_file_utf8) gtk_window_set_icon_from_file_utf8 __attribute((alias("IA__gtk_window_set_icon_from_file_utf8"), visibility("default")));

#endif
#undef gtk_window_set_icon_list
extern __typeof (gtk_window_set_icon_list) gtk_window_set_icon_list __attribute((alias("IA__gtk_window_set_icon_list"), visibility("default")));

#undef gtk_window_set_icon_name
extern __typeof (gtk_window_set_icon_name) gtk_window_set_icon_name __attribute((alias("IA__gtk_window_set_icon_name"), visibility("default")));

#undef gtk_window_set_keep_above
extern __typeof (gtk_window_set_keep_above) gtk_window_set_keep_above __attribute((alias("IA__gtk_window_set_keep_above"), visibility("default")));

#undef gtk_window_set_keep_below
extern __typeof (gtk_window_set_keep_below) gtk_window_set_keep_below __attribute((alias("IA__gtk_window_set_keep_below"), visibility("default")));

#undef gtk_window_set_mnemonic_modifier
extern __typeof (gtk_window_set_mnemonic_modifier) gtk_window_set_mnemonic_modifier __attribute((alias("IA__gtk_window_set_mnemonic_modifier"), visibility("default")));

#undef gtk_window_set_mnemonics_visible
extern __typeof (gtk_window_set_mnemonics_visible) gtk_window_set_mnemonics_visible __attribute((alias("IA__gtk_window_set_mnemonics_visible"), visibility("default")));

#undef gtk_window_set_modal
extern __typeof (gtk_window_set_modal) gtk_window_set_modal __attribute((alias("IA__gtk_window_set_modal"), visibility("default")));

#undef gtk_window_set_opacity
extern __typeof (gtk_window_set_opacity) gtk_window_set_opacity __attribute((alias("IA__gtk_window_set_opacity"), visibility("default")));

#undef gtk_window_get_opacity
extern __typeof (gtk_window_get_opacity) gtk_window_get_opacity __attribute((alias("IA__gtk_window_get_opacity"), visibility("default")));

#ifndef GTK_DISABLE_DEPRECATED
#undef gtk_window_set_policy
extern __typeof (gtk_window_set_policy) gtk_window_set_policy __attribute((alias("IA__gtk_window_set_policy"), visibility("default")));

#endif
#undef gtk_window_set_position
extern __typeof (gtk_window_set_position) gtk_window_set_position __attribute((alias("IA__gtk_window_set_position"), visibility("default")));

#undef gtk_window_set_resizable
extern __typeof (gtk_window_set_resizable) gtk_window_set_resizable __attribute((alias("IA__gtk_window_set_resizable"), visibility("default")));

#undef gtk_window_set_role
extern __typeof (gtk_window_set_role) gtk_window_set_role __attribute((alias("IA__gtk_window_set_role"), visibility("default")));

#undef gtk_window_set_screen
extern __typeof (gtk_window_set_screen) gtk_window_set_screen __attribute((alias("IA__gtk_window_set_screen"), visibility("default")));

#undef gtk_window_set_skip_pager_hint
extern __typeof (gtk_window_set_skip_pager_hint) gtk_window_set_skip_pager_hint __attribute((alias("IA__gtk_window_set_skip_pager_hint"), visibility("default")));

#undef gtk_window_set_skip_taskbar_hint
extern __typeof (gtk_window_set_skip_taskbar_hint) gtk_window_set_skip_taskbar_hint __attribute((alias("IA__gtk_window_set_skip_taskbar_hint"), visibility("default")));

#undef gtk_window_set_urgency_hint
extern __typeof (gtk_window_set_urgency_hint) gtk_window_set_urgency_hint __attribute((alias("IA__gtk_window_set_urgency_hint"), visibility("default")));

#undef gtk_window_set_title
extern __typeof (gtk_window_set_title) gtk_window_set_title __attribute((alias("IA__gtk_window_set_title"), visibility("default")));

#undef gtk_window_set_transient_for
extern __typeof (gtk_window_set_transient_for) gtk_window_set_transient_for __attribute((alias("IA__gtk_window_set_transient_for"), visibility("default")));

#undef gtk_window_set_type_hint
extern __typeof (gtk_window_set_type_hint) gtk_window_set_type_hint __attribute((alias("IA__gtk_window_set_type_hint"), visibility("default")));

#undef gtk_window_set_wmclass
extern __typeof (gtk_window_set_wmclass) gtk_window_set_wmclass __attribute((alias("IA__gtk_window_set_wmclass"), visibility("default")));

#undef gtk_window_stick
extern __typeof (gtk_window_stick) gtk_window_stick __attribute((alias("IA__gtk_window_stick"), visibility("default")));

#undef gtk_window_unfullscreen
extern __typeof (gtk_window_unfullscreen) gtk_window_unfullscreen __attribute((alias("IA__gtk_window_unfullscreen"), visibility("default")));

#undef gtk_window_unmaximize
extern __typeof (gtk_window_unmaximize) gtk_window_unmaximize __attribute((alias("IA__gtk_window_unmaximize"), visibility("default")));

#undef gtk_window_unstick
extern __typeof (gtk_window_unstick) gtk_window_unstick __attribute((alias("IA__gtk_window_unstick"), visibility("default")));

#endif
#endif
#if IN_HEADER(__GTK_WIN32_EMBED_WIDGET_H__)
#if IN_FILE(__GTK_WIN32_EMBED_WIDGET_C__)
#ifdef G_OS_WIN32
#undef gtk_win32_embed_widget_get_type
extern __typeof (gtk_win32_embed_widget_get_type) gtk_win32_embed_widget_get_type __attribute((alias("IA__gtk_win32_embed_widget_get_type"), visibility("default"))) G_GNUC_CONST;

#endif
#endif
#endif
#if IN_HEADER(__GTK_INFO_BAR_H__)
#if IN_FILE(__GTK_INFO_BAR_C__)
#undef gtk_info_bar_get_type
extern __typeof (gtk_info_bar_get_type) gtk_info_bar_get_type __attribute((alias("IA__gtk_info_bar_get_type"), visibility("default"))) G_GNUC_CONST;

#undef gtk_info_bar_new
extern __typeof (gtk_info_bar_new) gtk_info_bar_new __attribute((alias("IA__gtk_info_bar_new"), visibility("default")));

#undef gtk_info_bar_new_with_buttons
extern __typeof (gtk_info_bar_new_with_buttons) gtk_info_bar_new_with_buttons __attribute((alias("IA__gtk_info_bar_new_with_buttons"), visibility("default")));

#undef gtk_info_bar_get_action_area
extern __typeof (gtk_info_bar_get_action_area) gtk_info_bar_get_action_area __attribute((alias("IA__gtk_info_bar_get_action_area"), visibility("default")));

#undef gtk_info_bar_get_content_area
extern __typeof (gtk_info_bar_get_content_area) gtk_info_bar_get_content_area __attribute((alias("IA__gtk_info_bar_get_content_area"), visibility("default")));

#undef gtk_info_bar_add_action_widget
extern __typeof (gtk_info_bar_add_action_widget) gtk_info_bar_add_action_widget __attribute((alias("IA__gtk_info_bar_add_action_widget"), visibility("default")));

#undef gtk_info_bar_add_button
extern __typeof (gtk_info_bar_add_button) gtk_info_bar_add_button __attribute((alias("IA__gtk_info_bar_add_button"), visibility("default")));

#undef gtk_info_bar_add_buttons
extern __typeof (gtk_info_bar_add_buttons) gtk_info_bar_add_buttons __attribute((alias("IA__gtk_info_bar_add_buttons"), visibility("default")));

#undef gtk_info_bar_set_response_sensitive
extern __typeof (gtk_info_bar_set_response_sensitive) gtk_info_bar_set_response_sensitive __attribute((alias("IA__gtk_info_bar_set_response_sensitive"), visibility("default")));

#undef gtk_info_bar_set_default_response
extern __typeof (gtk_info_bar_set_default_response) gtk_info_bar_set_default_response __attribute((alias("IA__gtk_info_bar_set_default_response"), visibility("default")));

#undef gtk_info_bar_response
extern __typeof (gtk_info_bar_response) gtk_info_bar_response __attribute((alias("IA__gtk_info_bar_response"), visibility("default")));

#undef gtk_info_bar_set_message_type
extern __typeof (gtk_info_bar_set_message_type) gtk_info_bar_set_message_type __attribute((alias("IA__gtk_info_bar_set_message_type"), visibility("default")));

#undef gtk_info_bar_get_message_type
extern __typeof (gtk_info_bar_get_message_type) gtk_info_bar_get_message_type __attribute((alias("IA__gtk_info_bar_get_message_type"), visibility("default")));

#endif
#endif
#endif /* G_HAVE_GNUC_VISIBILITY */
#endif /* DISABLE_VISIBILITY */
